/* Prepare input for GMCT from a continuum electrostatics / molecular mechanics model
   sidechain rotamers of amino acid sidechains can be generated automatically
     
   gcem  -chmtop filename -chmpar filename \
                  -rotlibname name -eps_ff eps \
                  -empirical_energy -filter_ctof energy \ 
                  -epsin1 eps -epsin2 eps -epsin3 eps\
                  -epscore eps -epshead eps -epsext eps \
                  -eps1set name -eps2set name -eps3set name \
                  -channelset name -elycavset name \
                  -membz lower_core_boundary upper_core_boundary \
                         lower_head_boundary upper_head_boundary \
                  -membhole radius xcoord ycoord \
                  molname
                                      
   commandline options:
   
   mandatory      -chmtop string -chmpar string \
                  -rotlibname string name
   
   optional       -meadpath string -gcemdir string \
                  -filter_ctof float -eps_ff float \
                  -empirical_energy -print_chm \
                  -skip_stat_mmg -skip_stat_mmint \
                  -cap_max_nin int -keep_predef \
                  -write_binary_rotlib
                  
   others (opt.)  -T float -solrad float -sterln float \
                  -kBolt float -conconv float -econv float \
                  -bohr_radius float -proton_charge float \
                  -blab1 -blab2 -blab3                 

   others (GCEM)  -epsin1 float -epsin2 float -epsin3 float \
                  -epscore float -epshead float -epsext float \
                  -eps1set string -eps2set string eps3set string \
                  -membz float float float float -grid_detail int \
   
   optional       -elycavset string -channelset string \
                  -membhole float float float -capacitance \
                  -inside int \
                  
   The program generates sidechain rotamers for a given pqr structure
   ,generates the MM part of the intrinsic and interaction energy
   matrices for GMCT and the GCEM input (structures + input script)
   
   molname.rot   defines the flexible sites for which rotamers 
                 are to be generated or accepted as predefined
		 in molname.pqr (instance ID non-zero)

   sitename.est  defines the atoms of each site with their respective
                 charge forms/instances 
   
   The sidechain torsion angles are taken as those found as population 
   maxima in the SQUIRELL backbone dependent rotamer library.
   The conformational intrinsic energy is either calculated from the 
   torsion energy terms of the CHARMM force field or from a knowledge
   based potential using the population probabilities given in the
   rotamer library. The CHARMM Lennard-Jones interaction energy of 
   the rotamer with the background is added.
   
   Lennard-Jones interaction energies of each flexible site with each
   other flexible site are also calculated
   
   -print_chm triggers output of read CHARMM parameters and topologies
              as well as separate energy terms for each instance

   -empirical_energy triggers the use of the knowledge based potential
   
   -eps_ff  (default 1) is the dielectric constant of the environment
            in which the molecular mechanics force field was
            parametrized, used in the Coulomb potential calculations
            and as reference homogeneous dielectric environment for
            the calculation of conformational energies

   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Thomas Ullmann, 2007
   
   /usr/bin/g++-4.1 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O3 gcem.cc
   /usr/bin/g++-4.1 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O3  -o gcem gcem.o -L../../libmead_e -lmead -lm

*/


#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <fstream>
using std::ifstream;
#include <string>
using std::string;
#include <sys/times.h> // I want to do some benchmarking of options...
#include <unistd.h>

#include "MEAD_e/PhysCond.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/ChargeDist.h"
#include "MEAD_e/AtomChargeSet.h"
#include "MEAD_e/DielectricEnvironment.h"
#include "MEAD_e/DielByAtoms.h"
#include "MEAD_e/DielMembAtoms.h"
#include "MEAD_e/UniformDielectric.h"

#include "MEAD_e/ElectrolyteEnvironment.h"
#include "MEAD_e/ElectrolyteByAtoms.h"
#include "MEAD_e/ElectrolyteMembAtoms.h"

#include "MEAD_e/FinDiffMethod.h"
#include "MEAD_e/ElstatPot.h"
#include "MEAD_e/FDGridLevel.h"

#include "MEAD_e/GcemMolecule.h"
#include <boost/regex.hpp>

int main(int argc, char* argv[])
{
  // Only for benchmarking
  struct tms start, end,  // start and end of program
    bpreproc,  apreproc,  // preprocessing GcemMolecule
    bpostproc, apostproc, // postprocessing GcemMolecule
    bfilter,   afilter,   // eliminate energetically unfavorable conformers
    brotlib,   arotlib;   // write rotamer library to binary file
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);

  string molname;
  string rotlibname = "rotlib";
  string chmtop = "top.inp";
  string chmpar = "par.inp";
  string gcemdir = "gcem_input";
  string meadpath = "~/my_mead";
  
  // hand over to GCEM for the LPBE solvers
  string eps1set;
  string eps2set;
  string eps3set;
  string channelset;
  string elycavset;
  
  float epsvac = 1.0;
  float eps_ff = 1.0;  // Dielectric constant of the homogeneous dielectric
  float epsin1 = 0.0;  // These zeros are default states.  A specification of
  float epsin2 = 0.0;  // epsin1, epsin2, epsin3, epshead, epscore
  float epsin3 = 0.0;  // on the command line triggers handing over to GCEM
  float epshead = 0.0; // Dielectric constant of the membrane headgroups
  float epscore = 0.0; // Dielectric constant of the membrane core
                       
  float istr1   = 0.0; // Ionic strength of the upper solvent region
  float istr2   = 0.0; // Ionic strength of the lower solvent region

  float zlowercore = 0, zuppercore = 0, zlowerhead = 0, zupperhead = 0; // membrane dimensions
  float holerad = 0, holex = 0, holey = 0; // cylindrical hole through membrane

  float filter_ctof    = std::numeric_limits<float>::max(); // cutoff for lowest energy instance of conformer vs lowest energy instance
  float filter_ctof_gs = std::numeric_limits<float>::max(); // cutoff for modified goldstein criterium (Goldstein criterium + filter_ctof_gs)

  int   inside  = 1;   // 1 means inside is upper membrane side 
                       // 0 means inside is lower membrane side
  int cap_max_nin = -1; // cap maximum number of instances in a site
  int grid_detail =  2; // 0 - coarse, 1 - economic (similar to settings in multiflex example),
                        // 2 - fine (ensures converged electrostatics energies), 3 - very fine, 4 - ultra fine

  short ismemb = 0; // membrane specific input options trigger this switch
                    // for handover of options to GCEM
  
  bool do_postproc         = false; // do post-processing, read MEAD output files, write GMCT compatible energy matrices
  bool do_empirical        = false; // use knowledge based potential
  bool print_chm           = false; // print detailed CHARMM energy terms, parameters and topology
  bool print_debug         = false; // print intermediate results in GCEM molecule
  bool use_filter          = false; // filter instances of a site for intrinsic energy cutoff
  bool do_cap              = false; // calculate capacitive energy and capacitance
  bool keep_predef         = false; // do not filter predefined instances/conformers
  bool skip_stat_mmg       = false; // do not calculate Lennard-Jones interactions among non-flexible sites
  bool skip_stat_mmint     = false; // do not calculate intrinsic mm energies
  bool write_binary_rotlib = false; // write rotamer library to binary file
  bool got_istr            = false; // option ionicstr  specified by user
  bool got_istr1           = false; // option ionicstr2 specified by user
  bool got_istr2           = false; // option ionicstr2 specified by user

  ////////////// Parse the command line /////////////////////////////////

  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+1==argc) {  // This is the last one, must be molname
      molname = argv[iarg];
    }
    else if ((string) argv[iarg] == (string) "-epsin1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin1 option");
      epsin1 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin2 option");
      epsin2 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin3") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin3");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin3 option");
      epsin3 = f;
    }
    //////// My options /////////
    else if ((string) argv[iarg] == (string) "-eps1set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps1set");
      eps1set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps2set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps2set");
      eps2set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps3set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps3set");
      eps3set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-channelset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -channelset");
      channelset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-elycavset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -elycavset");
      elycavset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps_ff" || (string) argv[iarg] == (string) "-epshomo") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps_ff or -epshomo");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -eps_ff option");
      eps_ff = f;
    }
    /////////////////////////////
    else if ((string) argv[iarg] == (string) "-epshead") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epshead");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epshead option");
      epshead = f;
    }
    else if ((string) argv[iarg] == (string) "-epscore") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epscore");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epscore option");
      epscore = f;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr1 option");
      istr1 = f;
      got_istr1 = true;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr2 option");
      istr2 = f;
      got_istr2 = true;
    }
    else if ((string) argv[iarg] == (string) "-inside") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -inside");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -inside option");
      inside = i;
    }
    else if ((string) argv[iarg] == (string) "-grid_detail") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -grid_detail");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -grid_detail option");
      grid_detail = i;
    }
    else if ((string) argv[iarg] == (string) "-cap_max_nin") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -cap_max_nin");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -cap_max_nin option");
      cap_max_nin = i;
    }
    else if ((string) argv[iarg] == (string) "-epsext") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsext");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsext option");
      PhysCond::set_epsext(f);
    }
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -solrad");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -solrad option");
      PhysCond::set_solrad(f);
    }
    else if ((string) argv[iarg] == (string) "-sterln") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -sterln");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -sterln option");
      PhysCond::set_sterln(f);
    }
    else if ((string) argv[iarg] == (string) "-ionicstr") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr option");
      PhysCond::set_ionicstr(f);
      got_istr = true;
    }
    else if ((string) argv[iarg] == (string) "-T") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -T");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -T option");
      PhysCond::set_T(f);
    }
    else if ((string) argv[iarg] == (string) "-kBolt") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -kBolt");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -kBolt option");
      PhysCond::set_kBolt(f);
    }
    else if ((string) argv[iarg] == (string) "-conconv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -conconv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -conconv option");
      PhysCond::set_conconv(f);
    }
    else if ((string) argv[iarg] == (string) "-econv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -econv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -econv option");
      PhysCond::set_econv(f);
    }
    else if ((string) argv[iarg] == (string) "-bohr_radius") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -bohr_radius");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -bohr_radius option");
      PhysCond::set_bohr_radius(f);
    }
    else if ((string) argv[iarg] == (string) "-proton_charge") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -proton_charge");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -proton_charge opt.");
      PhysCond::set_proton_charge(f);
    }
    // process the form "-membz float float float float"
    else if ((string) argv[iarg] == (string) "-membz") {
      ismemb = 1;
      if (iarg+2>=argc) error ("main: ERROR, not enough args after -membz");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membz opt.");
      zlowercore = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -membz opt.");
      zuppercore = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -membz opt.");
      zlowerhead = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -membz opt.");
      zupperhead = f;
    }
    // process the form "-membhole float [float float]"
    else if ((string) argv[iarg] == (string) "-membhole") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -membhole");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membhole opt.");
      holerad = f;
      if (iarg+2 < argc) { // There _could_ be two more floats...
	istringstream istx(argv[iarg+1]);
	istx >> f;
	if (istx.good() || istx.eof()) { // Yes, expect this float and another.
	  holex = f;
	  iarg+=2;
	  istringstream isty(argv[iarg]);
	  isty >> f;
	  if (!(isty.good() || isty.eof()))
	    error("main: ERROR, only two floats seen after -membhole.\n",
		  "One or three expected.\n");
	  else
	    holey = f;
	}
      }
    }
    else if ((string) argv[iarg] == (string) "-capacitance" || (string) argv[iarg] == (string) "-capacity") {
      do_cap = true;
    }
    else if ((string) argv[iarg] == (string) "-keep_predef") {
      keep_predef = true;
    }
    else if ((string) argv[iarg] == (string) "-write_binary_rotlib") {
      write_binary_rotlib = true;
    }
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-empirical_energy") {
      do_empirical = true;
    }
    else if ((string) argv[iarg] == (string) "-print_chm") {
      print_chm = true;
    }
    else if ((string) argv[iarg] == (string) "-print_debug") {
      print_debug = true;
    }
    else if ((string) argv[iarg] == (string) "-skip_stat_mmg") {
      skip_stat_mmg = true;
    }
    else if ((string) argv[iarg] == (string) "-skip_stat_mmint") {
      skip_stat_mmint = true;
    }
    else if ((string) argv[iarg] == (string) "-chmtop") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -chmtop");
      chmtop = argv[++iarg];
    }    
    else if ((string) argv[iarg] == (string) "-chmpar") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -chmpar");
      chmpar = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-rotlib") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -rotlib");
      rotlibname = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-gcemdir" || (string) argv[iarg] == (string) "-qmpbdir") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -gcemdir");
      gcemdir = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-meadpath") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -meadpath");
      meadpath = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-filter_ctof") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -filter_ctof");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -filter_ctof opt.");
      if (f < 0) f = -f;
      filter_ctof = f;
      use_filter = true;
    }
    else if ((string) argv[iarg] == (string) "-filter_ctof_gs") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -filter_ctof_gs");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -filter_ctof_gs opt.");
      if (f < 0) f = -f;
      filter_ctof_gs = f;
      use_filter = true;
    }
    else if ((string) argv[iarg] == (string) "-p" || (string) argv[iarg] == (string) "-post") {
      do_postproc = true;
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

  /////////////// Checks & Statistics ////////////////////////////////

  if (molname == (string) ""){
    error("main: ERROR, molname not given on command line");
  }
  else if (chmtop == (string) "" && !do_postproc){
    cerr << "main: Warning, -chmtop not specified on command line" << endl
         << "      Will use default name \"top.inp\" in case of flexible sites.";
    chmtop = "top.inp";
  }
  else if (chmpar == (string) "" && !do_postproc){
    cerr << "main: Warning, -chmpar not specified on command line" << endl
         << "      Will use default name \"par.inp\" in case of flexible sites.";
    chmpar = "par.inp";
  }
  else if (rotlibname == (string) "" && !do_postproc){
    cerr << "main: WARNING, -rotlib not specified" << endl
         << "      Will use default name \"rotlib\" in case of flexible sites.";
    rotlibname = "rotlib";
  }
    
  if (epsin1 == 0.0 && epsin2 == 0.0 && epsin3 == 0.0 && !do_postproc)
    cerr << "main: WARNING, -epsin1, -epsin2 and -epsin3 values not specified." << endl <<
            "      Will use default options in QMPB input script.";
  else if (epsin1 == 0.0 && eps1set != (string) "")
    ::error("main: ERROR, -epsin1 value not specified while -epsin1 value is.");
  else if (epsin2 == 0.0 && eps2set != (string) "")
    ::error("main: ERROR, -epsin1 value not specified while -epsin2 value is.");
  else if (epsin3 == 0.0 && eps3set != (string) "")
    ::error("main: ERROR, -epsin3 value not specified while -epsin3 value is.");
  else if (eps1set == (string) "" && eps2set == (string) "" && eps3set == (string) "" && !do_postproc)
    cerr << "main: WARNING, -eps1set, eps2set and -eps3set values not specified." << endl <<
            "      Will use default options in QMPB input script.";

  if (eps1set == (string) ""){
    if (epsin1 != 0){
      ::error("main: ERROR, eps1set value not specified.",
            "   This flag is now mandatory if epsin1 is set.");
    } else {epsin1 = 1;} // dummy value
  } // will be checked by DielectricEnvironment
  if (eps2set == (string) ""){
    if (epsin2 != 0){
      ::error("main: ERROR, eps2set value not specified.",
            "   This flag is now mandatory if epsin2 is set.");
    } else {epsin2 = 4;} // dummy value
  } // will be checked by DielectricEnvironment
  if (eps3set == (string) ""){
    if (epsin3 != 0){
      ::error("main: ERROR, eps3set value not specified.",
            "   This flag is now mandatory if epsin3 is set.");
    } else {epsin3 = 80;} // dummy value
  } // will be checked by DielectricEnvironment   
  
  // this check will be done again in calc_cuberep
  // of Electrolyte and Dielectric Environments
 
  if (zuppercore > zupperhead){
    ::error("main: ERROR, zuppercore > zupperhead");  
  }
  else if (zlowercore > zupperhead){
    ::error("main: ERROR, zlowercore > zupperhead");  
  }
  else if (zlowerhead > zupperhead){
    ::error("main: ERROR, zlowerhead > zupperhead");  
  }
  else if (zlowercore > zuppercore){
    ::error("main: ERROR, zlowercore > zuppercore");  
  }
  else if (zlowerhead > zuppercore){
    ::error("main: ERROR, zlowerhead > zuppercore");  
  }
  else if (zlowerhead > zlowercore){
    ::error("main: ERROR, zlowerhead > zlowercore");  
  }

  if (inside != 1 && inside != 0){
   ::error("main: ERROR, inside != 1 or 0");
  }

  if (grid_detail < 0 || grid_detail > 4){
   ::error("main: ERROR, grid_detail not within {0,1,2,3,4}");
  }
  
  // check which type of ElectrolyteEnvironment is needed
  // if elycavset and channelset -> TwoElectrolyteMembThreeAtoms
  
  if (elycavset != (string) "" && channelset != (string) ""){
    ismemb=3;
  }// if only elycavset -> TwoElectrolyteMembTwoAtoms
  else if (elycavset != (string) "" && channelset != (string) ""){
    ismemb=2;
  }// if none -> TwoElectrolyteMembAtoms
  else if (epscore != 0) {
    ismemb=1;
  }
  else ismemb = 0;
  
  // ionicstr is used as default ionic strength of the reference
  // environment for the model compounds, therefore it has also
  // an effect for membrane intrinsic receptors

  // if ionicstr1 or ionicstr2 are not specified on commandline
  // but ionicstr is specified then the respective solvent region 
  // is assigned ionicstr  
  if (ismemb && istr1 == 0.0 && !got_istr1 && got_istr){
   istr1 = PhysCond::get_ionicstr();
  }
  if (ismemb && istr2 == 0.0 && !got_istr2 && got_istr){
   istr2 = PhysCond::get_ionicstr();
  }

  // if ionicstr was not specified on commandline but
  // ionicstr1 or ionicstr2 was, ionicstr is assigned
  // the first specified value
  if        (ismemb && got_istr1 && !got_istr){
   PhysCond::set_ionicstr(istr1);
  } else if (ismemb && got_istr2 && !got_istr){
   PhysCond::set_ionicstr(istr2);
  }

  Coord c(holex, holey, 0.0); // position of cylindrical hole in membrane

  //////////////// Done with comand line processing ////////////////
  
  /////// Print summary of input info from the command line: ///////
  
  cout << "Starting " << argv[0] << " for molecule named " << molname << endl;
  cout << "with grid detail :" << grid_detail << endl;
  cout << "using the following physical conditions:" << endl;
  PhysCond::print();
  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ////////////////// Setting up calculation ////////////////////////

  /////// create and setup GCEM molecule ////////
  
  GcemMolecule mol(molname, rotlibname, chmtop, chmpar);
  
  mol.set_gcemdir(gcemdir);
  mol.set_meadpath(meadpath);
  mol.set_eps_ff(eps_ff);
  mol.set_epsin1(epsin1);
  mol.set_epsin2(epsin2);
  mol.set_epsin3(epsin3);
  mol.set_epsext(PhysCond::get_epsext());
  mol.set_ionicstr(PhysCond::get_ionicstr());
  mol.set_grid_detail(grid_detail);
  if (ismemb > 0) {
   mol.set_has_memb(true);
   mol.set_zmaxh(zupperhead);
   mol.set_zmaxc(zuppercore);
   mol.set_zminc(zlowercore);
   mol.set_zminh(zlowerhead);
   mol.set_ionicstr1(istr1);
   mol.set_ionicstr2(istr2);
   mol.set_epscore(epscore);
   mol.set_epshead(epshead);
   mol.set_membholer(holerad);
   mol.set_membholex(holex);
   mol.set_membholey(holey);
   mol.set_eps1set(eps1set);
   mol.set_eps2set(eps2set);
   mol.set_eps3set(eps3set);
   mol.set_elycavset(elycavset);
   mol.set_channelset(channelset);
   mol.set_eps3set(eps3set);
   mol.set_inside(inside);
   mol.set_do_cap(do_cap);   
  } else mol.set_has_memb(false);
  // post-processing flag
  mol.set_do_postproc(do_postproc);
  // output flags
  mol.set_print_chm(print_chm);
  mol.set_print_debug(print_debug);
  // scaling factor for 1-4 coulomb interactions
  // for some applications in CHARMM 0.75 default 1.00 
  mol.set_coulomb_14_scalefac(1.0);
  // filter cutoffs for intrinsic energy based filtering and
  // Goldstein like filtering after interaction energy calculation 
  mol.set_filter_cutoff(filter_ctof);
  mol.set_filter_cutoff_gs(filter_ctof_gs);
  // use knowledge based potential for intrinsic energy of Site
  mol.set_use_empirical(do_empirical);
  mol.set_use_filter(use_filter);
  // terms involving ligand atoms and no model atoms
  mol.set_use_term_l_e(true);
  mol.set_use_term_l_lj(true);
  mol.set_use_term_l_bond(false);
  mol.set_use_term_l_angle(false);
  mol.set_use_term_l_dihedral(true);
  mol.set_use_term_l_improper(true);
  // terms involving ligand atoms and model atoms (ligand model interactions belongs to mmback)
  mol.set_use_term_ml_e(false); // done later with LPBE (G_model_back in GCEM)
  mol.set_use_term_ml_lj(true);
  mol.set_use_term_ml_bond(false);
  mol.set_use_term_ml_angle(false);
  mol.set_use_term_ml_dihedral(true);
  mol.set_use_term_ml_improper(true);
  // terms involving no ligand and no model atoms (internal site)
  mol.set_use_term_s_e(true); // floored to minimum term of all instances of same form to avoid double counting in Gmodel
  mol.set_use_term_s_lj(true);
  mol.set_use_term_s_bond(false);
  mol.set_use_term_s_angle(false);
  mol.set_use_term_s_dihedral(true);
  mol.set_use_term_s_improper(true);
  // terms involving no ligand but model atoms (site/model interaction mmback)
  mol.set_use_term_ms_e(false); // done later with LPBE (G_model_back in GCEM), and should also be floored there to be consistent!?
  mol.set_use_term_ms_lj(true);
  mol.set_use_term_ms_bond(false);
  mol.set_use_term_ms_angle(false);
  mol.set_use_term_ms_dihedral(true);
  mol.set_use_term_ms_improper(true);

  // skip Lennard-Jones interactions among pairs of non-flexible sites
  mol.set_skip_stat_pair_mmg(skip_stat_mmg);
  // skip intrinsic energies of non-flexible sites
  mol.set_skip_stat_mmint(skip_stat_mmint);
  
  ///// create rotamers, calculate energies, write GCEM input //////
 
  if (!do_postproc){
   times(&bpreproc);
   mol.prepare_gcem(molname, rotlibname, chmtop, chmpar, keep_predef);
   times(&apreproc);
   cout << "Preprocessing runtime: " << ((float) (apreproc.tms_utime-bpreproc.tms_utime))/ticks << " s user " << ((float) (apreproc.tms_stime-bpreproc.tms_stime))/ticks <<  " s system " << endl;
  }
  else {
   times(&bpostproc);
   mol.postproc_gcem();
   times(&apostproc);
   cout << "Postprocessing runtime: " << ((float) (apostproc.tms_utime-bpostproc.tms_utime))/ticks << " s user " << ((float) (apostproc.tms_stime-bpostproc.tms_stime))/ticks <<  " s system " << endl;
   if (use_filter) {
    // values  of preprocessor run were read from restart file
    times(&bfilter);
    mol.set_grid_detail(grid_detail);
    mol.set_use_filter(use_filter);
    mol.set_filter_cutoff(filter_ctof);
    mol.set_filter_cutoff_gs(filter_ctof_gs);
    mol.prepare_filtered_gcem(cap_max_nin, keep_predef);
    times(&afilter);
    cout << "Filtering runtime: " << ((float) (afilter.tms_utime-bfilter.tms_utime))/ticks << " s user " << ((float) (afilter.tms_stime-bfilter.tms_stime))/ticks <<  " s system " << endl;
   }
  }
  
  if (write_binary_rotlib){
    times(&brotlib);
    mol.rlib.write_binary();
    times(&arotlib);
    cout << "Writing rotlib to binary archive runtime: " << ((float) (brotlib.tms_utime-arotlib.tms_utime))/ticks << " s user " << ((float) (brotlib.tms_stime-arotlib.tms_stime))/ticks <<  " s system " << endl;  
  }

  //////////////////////////////////////////////////////////////////

  times(&end);
  cout << "Total runtime: " << ((float) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((float) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;

  return 0;
} // END main
