/* make dummy-atom files representing cavities and channels

   cavities buried inside the protein: cavset
   cavities/invaginations within the membrane boundaries, but accessible from the solvent: elycavset
   solvent accessible volume within the membrane boundaries and accessible from the solvent phases of both membrane sides: elycavset

   get-cavities   [opts]   molname

   commandline options:

   mandatory      -membz float float float float   (membrane boundaries: zlowercore, zuppercore, zlowerhead zupperhead)

   optional       -cavset string -elycavset string -channelset string \  (filename prefixes, defaults: cav, elycav, channel)
                  -space float \ (grid spacing)
                  -solrad float -sterln float \ (solvent probe radius, and Stern layer radius (spelling sterln historical in MEAD))
                  -do_shrink \ (find dummy atoms which are entirely covered by their neighbors and remove them, takes 8 times as long, gain is moderate)
                  -write_dx \ (write an OpenDX file with the dielectric regions)
                  -blab1 -blab2 -blab3 \  (flags that set the blab/verbosity level)


   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   R. Thomas Ullmann, 2014

   /usr/bin/g++-3.4 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 get-cavities.cc
   /usr/bin/g++-3.4 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o get-cavities get-cavities.o -L../../libmead -lmead  -lm

*/


#include <limits>
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <fstream>
using std::ifstream;
#include <string>
using std::string;
#include <sys/times.h> // I want to do some benchmarking of options...
#include <unistd.h>

using namespace std;

#include "MEAD/PhysCond.h"
#include "MEAD/AtomSet.h"
#include "MEAD/ChargeDist.h"
#include "MEAD/AtomChargeSet.h"
#include "MEAD/MembChargeSet.h"
#include "MEAD/DielectricEnvironment.h"
#include "MEAD/DielByAtoms.h"
#include "MEAD/DielMembAtoms.h"
#include "MEAD/UniformDielectric.h"

#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/ElectrolyteByAtoms.h"
#include "MEAD/ElectrolyteMembAtoms.h"

#include "MEAD/FinDiffMethod.h"
#include "MEAD/ElstatPot.h"
#include "MEAD/FDGridLevel.h"

// Write region map to (dielectric constants) to an OpenDX file
// OpenDX expects the lower corner as origin, not the grid center
void write_dx_file(const unsigned short * const arr, const unsigned int nx, const unsigned int ny, const unsigned int nz,
                   const float space, const Coord &origin, const string &file_name)
{
  blab3 << "Entering write_dx() ..." << endl;

  int ncube = nx * ny * nz;
  Coord center((float)((nx-1)/2) * space + origin.x, (float)((ny-1)/2) * space + origin.y, (float)((nz-1)/2) * space + origin.z);
  ofstream of(file_name.c_str());
  if (!of.good())
    ::error("write_dx failed to open for writing, ",
            file_name.c_str());
  blab1 << "write_dx(): Writing " << file_name.c_str() << " in OpenDX format...\n ";
  of << "# "<< file_name << " " << nx << " " << ny << " " << nz << " " << space << " " << center << endl;
  of << "#\n# DielectricRegions witten by MEAD get-cavities\n#" << endl;
  of << "object 1 class gridpositions counts " << nx << " " << ny << " " << nz << endl;
  of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
  of << "object 2 class gridconnections counts " << nx << " " << ny << " " << nz << endl;
  of << "object 3 class array type double rank 0 times " << ncube << endl;
  // Now the actual values
  // Don't know if I could write more than three columns according to OpenDX format,
  // but the TCL reader can only read three columns or less...
  int h = 2;
  for (h=2; h<ncube; h+=3) {
    of << arr[h-2] << " " << arr[h-1] << " " << arr[h] << endl;
  }
  bool isfirst = true;
  for (; h<ncube; ++h) {
    if (isfirst) {
      of        << arr[h] << endl;
      isfirst = false;
    } else {
      of << " " << arr[h] << endl;
    }
  }
  // Now some blurb at the end
  of << "attribute \"dep\" string \"positions\"" << endl;
  of << "object \"regular positions regular connections\" class field" << endl;
  of << "component \"positions\" value 1" << endl;
  of << "component \"connections\" value 2" << endl;
  of << "component \"data\" value 3" << endl;
  if (! of.good())
    ::error("write_dx() failed to write OpenDX file");
  of.close();

  blab2 << "... exiting from write_dx()" << endl;
}

int main(int argc, char* argv[])
{
  // Only for benchmarking
  struct tms start, end, // start and end of program
    beps1set, aeps1set,  // reading or calculating eps1set
    bgetcav, agetcav;    // finding the cavities
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);

  string molname;
  string cavset("cav");
  string channelset("channel");
  string elycavset("elycav");
  string membcoreset("");
  string membheadset("");
  // values don't matter here, just need to differ from all others
  float epsvac = 1.0;
  float epsin1 = 4.0;
  float epsin2 = 8.0;
  float epsin3 = 12.0;
  float epshead = 20.0; // Dielectric constant of the membrane headgroups
  float epscore = 2.0; // Dielectric constant of the membrane core
  float epsprehead = 30; // Dielectric constant of the membrane core
  float epsprecore = 3.0; // Dielectric constant of the membrane core
  unsigned char uc_epsvac = static_cast<unsigned char>(epsvac);
  unsigned char uc_epsin1 = static_cast<unsigned char>(epsin1);
  unsigned char uc_epsin2 = static_cast<unsigned char>(epsin2);
  unsigned char uc_epsin3 = static_cast<unsigned char>(epsin3);
  unsigned char uc_epshead = static_cast<unsigned char>(epshead);
  unsigned char uc_epscore = static_cast<unsigned char>(epscore);
  unsigned char uc_epsprehead = static_cast<unsigned char>(epsprehead);
  unsigned char uc_epsprecore = static_cast<unsigned char>(epsprecore);

  float epslowerfill = 78;
  float epsupperfill = 77;
  float epsbothfill  = 60;
  float epssol  = 80;
  PhysCond::set_epsext(epssol);
  unsigned char uc_epslowerfill = static_cast<unsigned char>(epslowerfill);
  unsigned char uc_epsupperfill = static_cast<unsigned char>(epsupperfill);
  unsigned char uc_epsbothfill = static_cast<unsigned char>(epsbothfill);
  unsigned char uc_epssol =  static_cast<unsigned char>(epssol);

  float istr1   = 0.0; // Ionic strength of the upper solvent region
  float istr2   = 0.0; // Ionic strength of the lower solvent region

  // flags determinining if analytical ion accessible volumes could be read
  bool got_iav1 = false;
  bool do_shrink = false;
  bool fuse_cav = false;
  bool write_dx = false;

  int count, countx, county, countz; //gridpoints
  float x, y, z; // grid center x, y, z, grid spacing
  float space = 1.0; // grid center x, y, z, grid spacing
  int ismemb = 0;
  float zlowercore = 0, zuppercore = 0, zlowerhead = 0, zupperhead = 0;

  ////////////// Parse the command line /////////////////////////////////
  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+1==argc) {  // These are the last ones, must be sitename molname
      molname = argv[iarg];
    }
    //////// My options /////////
    else if ((string) argv[iarg] == (string) "-cavset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -cavset");
      cavset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-channelset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -channelset");
      channelset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-elycavset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -elycavset");
      elycavset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-membcoreset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -membcoreset");
      membcoreset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-membheadset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -membheadset");
      membheadset = argv[++iarg];
    }
    /////////////////////////////
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -solrad");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -solrad option");
      PhysCond::set_solrad(f);
    }
    else if ((string) argv[iarg] == (string) "-sterln") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -sterln");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -sterln option");
      PhysCond::set_sterln(f);
    }
    else if ((string) argv[iarg] == (string) "-ionicstr") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr option");
      PhysCond::set_ionicstr(f);
    }
    // process the form "-membz float float float float"
    else if ((string) argv[iarg] == (string) "-membz") {
      ismemb = 1;
      if (iarg+2>=argc) error ("main: ERROR, not enough args after -membz");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membz opt.");
      zlowercore = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -membz opt.");
      zuppercore = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -membz opt.");
      zlowerhead = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -membz opt.");
      zupperhead = f;
    }
    else if ((string) argv[iarg] == (string) "-space") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -space");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -space option");
      space = f;
    }
    else if ((string) argv[iarg] == (string) "-fuse_cav") {
      fuse_cav = true;
    }
    else if ((string) argv[iarg] == (string) "-do_shrink") {
      do_shrink = true;
    }
    else if ((string) argv[iarg] == (string) "-write_dx") {
      write_dx = true;
    }
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

  /////////////// Checks & Statistics ////////////////////////////////

  // this check will be done again in calc_cuberep
  // of Electrolyte and Dielectric Environments

  if (zuppercore > zupperhead){
    ::error("main: ERROR, zuppercore > zupperhead");
  }
  else if (zlowercore > zupperhead){
    ::error("main: ERROR, zlowercore > zupperhead");
  }
  else if (zlowerhead > zupperhead){
    ::error("main: ERROR, zlowerhead > zupperhead");
  }
  else if (zlowercore > zuppercore){
    ::error("main: ERROR, zlowercore > zuppercore");
  }
  else if (zlowerhead > zuppercore){
    ::error("main: ERROR, zlowerhead > zuppercore");
  }
  else if (zlowerhead > zlowercore){
    ::error("main: ERROR, zlowerhead > zlowercore");
  }

  // check which type of ElectrolyteEnvironment is needed
  ismemb=3;

  // we need ionic strength != 0 to discriminate ion-accessible from ion-inaccessible points
  PhysCond::set_ionicstr(1.0f);
  istr1 = PhysCond::get_ionicstr();
  istr2 = PhysCond::get_ionicstr();

  // if OpenDX output is requested we need to know
  // grid spacing and grid length, the grid center defaults to 0,0,0
  // in principle one could also read an .ogm
  // but usually we use multiple grids

  if (space == 0.0){
   ::error("main: ERROR, OpenDX output requested via",
           "-write_eps|ely flag, but space value not specified.");
  }

  //////////////// Done with comand line processing ////////////////

  /////// Print summary of input info from the command line: ///////

  cout << "Starting " << argv[0] << " for molecule named " << molname << "\n";
  cout << "using the following conditions:" << endl
       << "Membrane headgroup region lower boundary = " << zlowerhead << endl
       << "Membrane headgroup region upper boundary = " << zupperhead << endl
       << "Membrane core region lower boundary = " << zlowercore << endl
       << "Membrane core region upper boundary = " << zuppercore << endl
       << endl;
  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ////////////////// Setting up calculation ////////////////////////

  // background charge set
  AtomSet a1(molname);
  a1.read();

  // cavset
  AtomSet acav(cavset);

  // elycavset
  AtomSet aelycav(elycavset);

  // channelset
  AtomSet achannel(channelset);

  // elycavset
  AtomSet ahead(membheadset);

  // channelset
  AtomSet acore(membcoreset);

  /////// calculate or read precalculated analytic volumes ////////

  // use pqr2SolvAccVol, pqr2IonAccVol for precalculation

  SolvAccVol sav1(a1);
  // dummy
  SolvAccVol sav2(acav);
  SolvAccVol sav3(acav);

  // Takes the name given to construct AtomSet,
  // appends ".dat" or ".txt" respectively.
  // Read functions return zero on error!
  times(&beps1set);
  blab1 << "Try to read in SolvAccVol for " << molname << ", else calculate it" << endl;
  if (molname != (string) ""){
    if (sav1.read_top_in_binary(molname + DOT + "dat") == 0) {
      blab1 << "Unable to read in SolvAccVol from binary file for " << molname << endl;
      if (sav1.read_top_in_ascii(molname + DOT + "txt") == 0) {
        blab1 << "Unable to read in SolvAccVol from ASCII file for " << molname << endl;
        blab1 << "Calculating SolvAccVol from AtomSet for eps1set" << endl;
        sav1.anal_calc(); // this is a bit costly for big molecules
      }
    }
  }
  times(&aeps1set);

  blab3 << "Eps1set runtime: " << ((float) (aeps1set.tms_utime-beps1set.tms_utime))/ticks << " s user " << ((float) (aeps1set.tms_stime-beps1set.tms_stime))/ticks <<  " s system " << endl;

  // Now I should have analytical surfaces either loaded or calculated
  // Constructor called with solvent/ion accessible volume instead of atom
  // set as done in solinprot!

  blab3 << "Computing cavities for a SixValueDielMembThreeAtoms" << endl;

  // protein qm/mm, protein mm,
  // cavities / invaginations / channel volume (can be zero / no atoms used in definition)
  // membrane with cylindrical hole (can have zero width)
  DielectricEnvironment_lett *diel = new SixValueDielMembThreeAtoms(sav1, sav2, sav3, zlowerhead, zupperhead,
                                     zlowercore, zuppercore, Coord(0,0,0), 0.0f, epsprehead, epsprecore, epsin1, epsin2, epsin3, PhysCond::get_epsext());
  DielectricEnvironment eps(diel);

  ///////////// find the cavities //////////////

  if (a1.size() != 0){

    times(&bgetcav);

    blab1 << "Determining dielectric regions on a grid ..." << endl;

    blab2 << "compute grid size ..." << endl;

    float minx = numeric_limits<float>::max();
    float miny = numeric_limits<float>::max();
    float minz = numeric_limits<float>::max();
    float maxx = numeric_limits<float>::min();
    float maxy = numeric_limits<float>::min();
    float maxz = numeric_limits<float>::min();
    for (AtomSet::const_iterator aiter = a1.begin(); aiter != a1.end(); ++aiter) {
      if (aiter->second.coord.x < minx) minx = aiter->second.coord.x;
      if (aiter->second.coord.y < miny) miny = aiter->second.coord.y;
      if (aiter->second.coord.z < minz) minz = aiter->second.coord.z;
      if (aiter->second.coord.x > maxx) maxx = aiter->second.coord.x;
      if (aiter->second.coord.y > maxy) maxy = aiter->second.coord.y;
      if (aiter->second.coord.z > maxz) maxz = aiter->second.coord.z;
    }
    // increase the grid size to ensure that we have "bulk" points around the molecule
    float two_space = 2 * space;
    minx -= two_space + 5;
    miny -= two_space + 5;
    minz -= two_space + 5;
    maxx += two_space + 5;
    maxy += two_space + 5;
    maxz += two_space + 5;

    // compute the remaining grid parameters
    float x = (maxx - minx) * 0.5 + minx;
    float y = (maxy - miny) * 0.5 + miny;
    float z = (maxz - minz) * 0.5 + minz;
    Coord center(x, y, z);

    // the grid edge lengths
    unsigned int countx = static_cast<unsigned int>(ceil(maxx - minx)/space);
    unsigned int county = static_cast<unsigned int>(ceil(maxy - miny)/space);
    unsigned int countz = static_cast<unsigned int>(ceil(maxz - minz)/space);
    // need odd numbers
    if ((countx % 2) == 0) countx++;
    if ((county % 2) == 0) county++;
    if ((countz % 2) == 0) countz++;
    unsigned int countcube = countx;
    if (county > countcube) countcube = county;
    if (countz > countcube) countcube = countz;
    // redefine the grid boundaries interms of integer multiples of the grid spacing
    minx = center.x - static_cast<float>((countx - 1) / 2) * space;
    maxx = center.x + static_cast<float>((countx - 1) / 2) * space;
    miny = center.y - static_cast<float>((county - 1) / 2) * space;
    maxy = center.y + static_cast<float>((county - 1) / 2) * space;
    minz = center.z - static_cast<float>((countz - 1) / 2) * space;
    maxz = center.z + static_cast<float>((countz - 1) / 2) * space;
    
    float lengthx = static_cast<float>(ceil(maxx - minx)/space) * space;
    float lengthy = static_cast<float>(ceil(maxy - miny)/space) * space;
    float lengthz = static_cast<float>(ceil(maxz - minz)/space) * space;
    Coord origin(center.x - (((countx - 1) * space) / 2),
                 center.y - (((county - 1) * space) / 2),
                 center.z - (((countz - 1) * space) / 2));
    Coord origincube(center.x - (((countcube - 1) * space) / 2),
                     center.y - (((countcube - 1) * space) / 2),
                     center.z - (((countcube - 1) * space) / 2));

    int nsq = countz * county;
    int npoints = nsq * countx;
    int npointscube = countcube * countcube * countcube;

    blab3 << " Grid center: " << center << endl;
    blab3 << " Grid origin: " << origin << endl;
    blab3 << " Grid size  : " << countx << ", " << county << ", " << countz << endl;

    blab3 << "Preparing temporary array of accessibility values ..." << endl;

    // need to inflate our rectangular grid to a cube here
    // the cubic array is destroyed immediately again
    unsigned int ncube = countcube * countcube * countcube;
    unsigned int ngrid = countx    * county    * countz;
    unsigned int startiix = (countcube - countx) / 2;
    unsigned int startiiy = (countcube - county) / 2;
    unsigned int startiiz = (countcube - countz) / 2;
    CubeLatSpec tmpcls(countcube,space,center);
    DielCubeRep *dcr = new DielCubeRep(eps.get_cuberep(tmpcls));
    // this is a necessary workaround to allow direct access to the internal data
    dcr->declare_undefined();
    // the work array
    unsigned short int *work_arr = new unsigned short int[ngrid];
    unsigned short int *wrad_arr = new unsigned short int[ngrid];
    // transcribe the eps values from the DielCubeRep to the ideally leaner work array
    unsigned int h  = 0;        // index in the work array
    unsigned int hh = startiix * (countcube * countcube) + startiiy * countcube + startiiz; // index in the cubic array
    for (unsigned int ix = 0; ix < countx; ++ix) {
      for (unsigned int iy = 0; iy < county; ++iy) {
        for (unsigned int iz = 0; iz < countz; ++iz) {
//          cout << " ix, iy, iz = " << ix << ", " << iy << ", " << iz << endl;
          work_arr[h] = static_cast<short unsigned int>((*dcr)[hh]);
//          cout << " dcr[" << hh <<  "] = " << (*dcr)[hh] << ", work[" << h << "] = " << work_arr[h] << endl;
          h++;
          hh++;
        }
        hh += 2 * startiiz;
      }
      hh += (2 * startiiy) * countcube;
      hh += (2 * startiix) * countcube * countcube; 
    }
    delete dcr;

    blab3 << "Processing temporary array of accessibility values ..." << endl;

    // now determine the true membrane core region by exchanging continuous areas of epsprecore to epscore
    // initiate the true membrane core by priming the outer rim of the grid in the membrane core boundaries
    unsigned int zlowercore_in_grid = floor((zlowercore - origin.z) / space);
    unsigned int zuppercore_in_grid = ceil((zuppercore - origin.z) / space);
    h  = 0; // index in the work array
    for (unsigned int ix = 0; ix < countx; ++ix) {
      if (ix == 0 || ix == (countx - 1)) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = zlowercore_in_grid; iz <= zuppercore_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            work_arr[h] = uc_epscore;
          }
        }
      } else {
        for (unsigned int iy = 0; iy < county; iy += (county - 1)) {
          for (unsigned int iz = zlowercore_in_grid; iz <= zuppercore_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            work_arr[h] = uc_epscore;
          }
        }      
      }
    }
    // gather points for a continuous membrane core
    bool got_change = true;
    bool is_assigned = false;
    unsigned int niter = 0;
    while (got_change) {
      got_change = false;
      niter++;
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = zlowercore_in_grid; iz <= zuppercore_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            // preassigned core points are guaranteed to be a) inside the membrane core boundaries and b) outside the molecule
            // now we need to discriminate true membrane core poinst from cavities and the like
            if (work_arr[h] == uc_epsprecore || work_arr[h] == uc_epsprehead) {
              is_assigned = false;
              for (unsigned int iix = (ix > 0 ? ix - 1 : ix); iix <= (ix < (countx - 1) ? ix + 1 : ix); ++iix) {
                for (unsigned int iiy = (iy > 0 ? iy - 1 : iy); iiy <= (iy < (county - 1) ? iy + 1 : iy); ++iiy) {
                  for (unsigned int iiz = (iz > zlowercore_in_grid ? iz - 1 : iz); iiz <= (iz < (zuppercore_in_grid) ? iz + 1 : iz); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a neighboring membrane core point
                    if (work_arr[hh] == uc_epscore) {
                      work_arr[h] = uc_epscore;
                      got_change = true;
                      is_assigned = true;
                      break;
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }
        }      
      }
    }
    blab3 << "Finished gathering of membrane core region after " << niter << " iterations." << endl;

    // now we add the true head group region as any point pre-assigned as possible headgroup point and
    // within headgroup region thickness of a true membrane core point
    // the headgroup region thickness may differ between the membrane leaflets!
    float lower_headwidth = zlowercore - zlowerhead;
    float upper_headwidth = zupperhead - zuppercore;
    unsigned int half_memb_width_in_grid = ceil((zupperhead - zlowerhead) * 0.5);
    unsigned int core_width_in_grid = ceil(zuppercore - zlowercore);
    unsigned int upper_headwidth_in_grid = ceil(upper_headwidth / space);
    unsigned int lower_headwidth_in_grid = ceil(lower_headwidth / space);
    unsigned int zupperhead_in_grid = ceil((zupperhead - origin.z) / space);
    unsigned int zlowerhead_in_grid = floor((zlowerhead - origin.z) / space);
    float sq_space = space * space;
    unsigned int sq_upper_headwidth_in_grid = ceil(upper_headwidth * upper_headwidth / sq_space);
    unsigned int sq_lower_headwidth_in_grid = ceil(lower_headwidth * lower_headwidth / sq_space);

    blab3 << "zupperhead_in_grid = " << zupperhead_in_grid << endl;
    blab3 << "zuppercore_in_grid = " << zuppercore_in_grid << endl;
    blab3 << "zlowercore_in_grid = " << zlowercore_in_grid << endl;
    blab3 << "zlowerhead_in_grid = " << zlowerhead_in_grid << endl;
    blab3 << "upper_headwidth_in_grid = " << upper_headwidth_in_grid << endl;
    blab3 << "lower_headwidth_in_grid = " << lower_headwidth_in_grid << endl;

    unsigned int distsq_in_grid = 0;
    h  = 0; // index in the work array
    niter = 0;
    got_change = true;
    while (got_change) {
      got_change = false;
      niter++;
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = zlowerhead_in_grid; iz <= zlowercore_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epsprehead) {
              is_assigned = false;
              for (unsigned int iix = (ix > lower_headwidth_in_grid ? ix - lower_headwidth_in_grid : 0); iix <= (ix + lower_headwidth_in_grid  < (countx - 1) ? ix + lower_headwidth_in_grid : (countx - 1)); ++iix) {
                for (unsigned int iiy = (iy > lower_headwidth_in_grid ? iy - lower_headwidth_in_grid : 0); iiy <= (iy + lower_headwidth_in_grid  < (county - 1) ? iy + lower_headwidth_in_grid : (county - 1)); ++iiy) {
                  for (unsigned int iiz = ((iz - lower_headwidth_in_grid) >= zlowerhead_in_grid ? iz - lower_headwidth_in_grid : zlowerhead_in_grid); iiz <= (iz + lower_headwidth_in_grid  <= zuppercore_in_grid ? iz + lower_headwidth_in_grid : zuppercore_in_grid); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a membrane core point in appropriate distance
                    if (work_arr[hh] == uc_epscore) {
                      unsigned int sq_dist_in_grid = ((iix - ix) * (iix - ix))
                                                   + ((iiy - iy) * (iiy - iy))  
                                                   + ((iiz - iz) * (iiz - iz));
                      if (sq_dist_in_grid <= sq_lower_headwidth_in_grid) {
                        work_arr[h] = uc_epshead;
                        got_change = true;
                        is_assigned = true;
                        break;
                      }
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }
          for (unsigned int iz = zuppercore_in_grid; iz <= zupperhead_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epsprehead) {
              is_assigned = false;
              for (unsigned int iix = (ix > upper_headwidth_in_grid ? ix - upper_headwidth_in_grid : 0); iix <= ((ix + upper_headwidth_in_grid)  < (countx - 1) ? ix + upper_headwidth_in_grid : (countx - 1)); ++iix) {
                for (unsigned int iiy = (iy > upper_headwidth_in_grid ? iy - upper_headwidth_in_grid : 0); iiy <= ((iy + upper_headwidth_in_grid)  < (county - 1) ? iy + upper_headwidth_in_grid : (county - 1)); ++iiy) {
                  for (unsigned int iiz = ((iz - upper_headwidth_in_grid) >= zlowercore_in_grid ? iz - upper_headwidth_in_grid : zlowercore_in_grid); iiz <= ((iz + upper_headwidth_in_grid) <= zupperhead_in_grid ? iz + upper_headwidth_in_grid : zupperhead_in_grid); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a neighboring membrane core point
                    if (work_arr[hh] == uc_epscore) {
                      unsigned int sq_dist_in_grid = ((iix - ix) * (iix - ix))
                                                   + ((iiy - iy) * (iiy - iy))  
                                                   + ((iiz - iz) * (iiz - iz));
                      if (sq_dist_in_grid <= sq_upper_headwidth_in_grid) {
                        work_arr[h] = uc_epshead;
                        got_change = true;
                        is_assigned = true;
                        break;
                      }
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }        
        }
      }
    }
    blab3 << "Finished gathering of membrane headgroup regions after " << niter << " iterations." << endl;

    // now we gather all remaining points connected to one of the solvent phases
    // points accessible from both solvent phases are categorized as belonging
    //  to a channel -> channelset
    // points accessible from one side only are categorized as belonging to an
    //  electrolyte accessible cavity / cleft etc. -> elycavset
    // points not accessible from either side are categorized as belonging to an
    //  internal cavity -> cavset
    h  = 0; // index in the work array
    niter = 0;
    
    // first fill from the lower side
    got_change = true;
    while (got_change) {
      got_change = false;
      niter++;
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          // need to start from zero because the molecule might extend far from the membrane
          for (unsigned int iz = 0; iz <= zuppercore_in_grid; ++iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epsprehead || work_arr[h] == uc_epsprecore) {
              is_assigned = false;
              for (unsigned int iix = (ix > 0 ? ix - 1 : 0); iix <= (ix < (countx - 1) ? ix + 1 : (countx - 1)); ++iix) {
                for (unsigned int iiy = (iy > 0 ? iy - 1 : 0); iiy <= (iy < (county - 1) ? iy + 1 : (county - 1)); ++iiy) {
                  for (unsigned int iiz = (iz > 0 ? iz - 1 : 0); iiz <= (iz < zupperhead_in_grid ? iz + 1 : zupperhead_in_grid); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a solvent point in appropriate distance
                    if (work_arr[hh] == uc_epssol || work_arr[hh] == uc_epslowerfill) {
                      work_arr[h] = uc_epslowerfill;
                      got_change = true;
                      is_assigned = true;
                      break;
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }
        }
      }
    }
    // now fill from above
    got_change = true;
    while (got_change) {
      got_change = false;
      niter++;
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = (countz - 1); iz >= zlowerhead_in_grid; --iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epsprehead || work_arr[h] == uc_epsprecore) {
              is_assigned = false;
              for (unsigned int iix = (ix > 0 ? ix - 1 : 0); iix <= (ix < (countx - 1) ? ix + 1 : (countx - 1)); ++iix) {
                for (unsigned int iiy = (iy > 0 ? iy - 1 : 0); iiy <= (iy < (county - 1) ? iy + 1 : (county - 1)); ++iiy) {
                  for (unsigned int iiz = (iz > 0 ? iz - 1 : 0); iiz <= (iz < zupperhead_in_grid ? iz + 1 : zupperhead_in_grid); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a solvent point in appropriate distance
                    if (work_arr[hh] == uc_epslowerfill || work_arr[hh] == uc_epsupperfill ||
                        work_arr[hh] == uc_epssol || work_arr[hh] == uc_epsbothfill) {
                      work_arr[h] = uc_epsupperfill;
                      got_change = true;
                      is_assigned = true;
                      break;
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }        
        }
      }
    }
    // now join connected cavities to channels
    got_change = true;
    while (got_change) {
      got_change = false;
      niter++;
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = (countz - 1); iz >= zlowerhead_in_grid; --iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epslowerfill || work_arr[h] == uc_epsupperfill) {
              is_assigned = false;
              for (unsigned int iix = (ix > 0 ? ix - 1 : 0); iix <= (ix < (countx - 1) ? ix + 1 : (countx - 1)); ++iix) {
                for (unsigned int iiy = (iy > 0 ? iy - 1 : 0); iiy <= (iy < (county - 1) ? iy + 1 : (county - 1)); ++iiy) {
                  for (unsigned int iiz = (iz > 0 ? iz - 1 : 0); iiz <= (iz < zupperhead_in_grid ? iz + 1 : zupperhead_in_grid); ++iiz) {
                    hh = (iix * county * countz) + (iiy * countz) + iiz;
                    // gather points that already have a solvent point in appropriate distance
                    if        (work_arr[h] == uc_epsupperfill && (work_arr[hh] == uc_epslowerfill || work_arr[hh] == uc_epsbothfill)) {
                      work_arr[h] = uc_epsbothfill;
                      got_change = true;
                      is_assigned = true;
                      break;
                    } else if (work_arr[h] == uc_epslowerfill && (work_arr[hh] == uc_epsupperfill || work_arr[hh] == uc_epsbothfill)) {
                      work_arr[h] = uc_epsbothfill;
                      got_change = true;
                      is_assigned = true;
                      break;
                    }
                  }
                  // don't need to do further checks if the point is already assigned
                  if (is_assigned) {
                    break;
                  }
                }
                // don't need to do further checks if the point is already assigned
                if (is_assigned) {
                  break;
                }
              }
            }
          }        
        }
      }
    }
    blab3 << "Finished gathering of channel/elycav/cav regions after " << niter << " iterations." << endl;

    if (fuse_cav) {
      blab2 << endl << "Fusing all cavities in one categorie as requested via -fuse_cav ..." << endl;
      // consider all solvent filled regions within the membrane boundaries as a single category
      for (unsigned int ix = 0; ix < countx; ++ix) {
        for (unsigned int iy = 0; iy < county; ++iy) {
          for (unsigned int iz = (countz - 1); iz >= zlowerhead_in_grid; --iz) {
            h = (ix * county * countz) + (iy * countz) + iz;
            if (work_arr[h] == uc_epslowerfill || work_arr[h] == uc_epsupperfill ||
                work_arr[h] == uc_epsbothfill  || work_arr[h] == uc_epsprecore   || work_arr[h] == uc_epsprehead) {
              // will be written as elycav, Stern Layer will still apply -> small cavities pose no problem
              work_arr[h] = uc_epsupperfill;
            }
          }        
        }
      }
    }

    blab3 << endl << "Reducing the number of dummy atoms ..." << endl;

    // now try to join the volume in fewer dummy atoms
    // any dummy entirely surrounded by other dummy atoms can be grown to encompass the other dummy atoms
    // initialize wrad_arr
    unsigned int solrad_in_grid = static_cast<unsigned int>(ceil(PhysCond::get_solrad() / space));
    unsigned int sq_solrad_in_grid = solrad_in_grid * solrad_in_grid;
    for (unsigned int ix = 0; ix < countx; ++ix) {
      for (unsigned int iy = 0; iy < county; ++iy) {
        for (unsigned int iz = 0; iz < countz; ++iz) {
          h = (ix * county * countz) + (iy * countz) + iz;
          wrad_arr[h] = 0;
        }
      }
    }
    for (unsigned int ix = 0; ix < countx; ++ix) {
      for (unsigned int iy = 0; iy < county; ++iy) {
        for (unsigned int iz = zlowerhead_in_grid; iz <= zupperhead_in_grid; ++iz) {
          h = (ix * county * countz) + (iy * countz) + iz;
          wrad_arr[h] = 1;
        }
      }
    }

    bool got_change2   = true;
    bool is_head1      = false;
    bool is_core1      = false;
    bool is_cav1       = false;
    bool is_elycav1    = false;
    bool is_channel1   = false;
    bool is_solvent1   = false;
    bool is_interior1  = false;
    bool is_memb1      = false;
    bool is_head2      = false;
    bool is_core2      = false;
    bool is_cav2       = false;
    bool is_elycav2    = false;
    bool is_channel2   = false;
    bool is_solvent2   = false;
    bool is_interior2  = false;
    bool is_memb2      = false;
    bool is_head3      = false;
    bool is_core3      = false;
    bool is_cav3       = false;
    bool is_elycav3    = false;
    bool is_channel3   = false;
    bool is_solvent3   = false;
    bool is_interior3  = false;
    bool is_memb3      = false;
    bool is_surrounded = true;
    bool is_same_type  = true;

    blab3 << "core_width_in_grid = " << half_memb_width_in_grid << endl;
//    unsigned int istep = half_memb_width_in_grid;
    int istep = 3;
    int istep_max = core_width_in_grid;
    int maxrad = 0;
    int sq_istep = istep * istep;
    unsigned int nremoved = 0;
    unsigned int last_nremoved = 1;
    unsigned int nshrinked = 0;
    niter = 0;
    unsigned int noiter = 0;
    while (nremoved != last_nremoved) {
      noiter++;
      blab3 << "Cycle " << noiter << ": last_nremoved = " << last_nremoved << ", nremoved = " << nremoved << endl;
      last_nremoved = nremoved;
      got_change2  = false;
      istep = 2;
      while (istep <= istep_max) {
        got_change  = true;
        blab3 << "Reducing the number of dummy atoms. stepsize = " << istep << endl;
        while (got_change) {
          got_change = false;
          niter++;
          if ((niter % 10) == 0) blab3 << "Cycle " << noiter << ", iteration = " << niter << " stepsize = " << istep << ", nremoved = " << nremoved<< ", nshrinked = " << nshrinked << endl;
          for (int ix = 0; ix < countx; ++ix) {
            for (int iy = 0; iy < county; ++iy) {
              for (int iz = zlowerhead_in_grid; iz <= zupperhead_in_grid; ++iz) {
                h = (ix * county * countz) + (iy * countz) + iz;
                is_surrounded = false;
                // an already discarded or for this istep already enlarged dummy atom
                if (wrad_arr[h] == 0 || wrad_arr[h] == istep) continue;
                is_solvent1  = false;
                is_cav1      = false;              
                is_elycav1   = false;              
                is_channel1  = false;              
                is_interior1 = false;
                is_memb1     = false;              
                is_head1     = false;
                is_core1     = false;
                if        (work_arr[h] == uc_epsin1 || work_arr[h] == uc_epsin2) {
                  is_interior1 = true;
                } else if (work_arr[h] == uc_epshead || work_arr[h] == uc_epscore) {
                  is_memb1 = true;
                  if        (work_arr[h] == uc_epshead) {
                    is_head1 = true;
                    // don't need to process the headgroup region if it's not used later on
                    if (membheadset == "") continue;
                  } else if (work_arr[h] == uc_epscore) {
                    is_core1 = true;
                    // don't need to process the core region if it's not used later on
                    if (membcoreset == "") continue;
                  }
                } else if (work_arr[h] == uc_epsprehead || work_arr[h] == uc_epsprecore ||
                           work_arr[h] == uc_epsupperfill || work_arr[h] == uc_epslowerfill ||
                           work_arr[h] == uc_epsbothfill) {
                  is_solvent1 = true;
                  if        (work_arr[h] == uc_epsprehead || work_arr[h] == uc_epsprecore) {
                    is_cav1 = true;
                  } else if (work_arr[h] == uc_epsupperfill || work_arr[h] == uc_epslowerfill) {
                    is_elycav1 = true;
                  } else if (work_arr[h] == uc_epsbothfill) {
                    is_channel1 = true;
                  }
                }
                if (is_memb1) {
                  if (is_head1) {
                    if      (iz >= zuppercore) {
                      if ((iz + istep) > zupperhead_in_grid || (iz - istep) < zuppercore_in_grid) continue;
                    }
                    else if (iz <= zlowercore) {
                      if ((iz - istep) < zlowerhead_in_grid || (iz + istep) > zlowercore_in_grid) continue;
                    }
                  } else if (is_core1) {
                      if ((iz + istep) > zuppercore_in_grid || (iz - istep) < zlowercore_in_grid) continue;            
                  }
                }
                if (!is_interior1) {
                  is_surrounded = true;
                  for (int iix = (ix > istep ? (ix - istep) : 0); iix <= ((ix + istep) < (countx - 1) ? (ix + istep) : (countx - 1)); ++iix) {
                    for (int iiy = (iy > istep ? (iy - istep) : 0); iiy <= ((iy + istep) < (county - 1) ? (iy + istep) : (county - 1)); ++iiy) {
                      for (int iiz = (iz > (istep + zlowerhead_in_grid) ? (iz - istep) : zlowerhead_in_grid); iiz <= ((iz + istep) < zupperhead_in_grid ? (iz + istep) : zupperhead_in_grid); ++iiz) {
                        hh = (iix * county * countz) + (iiy * countz) + iiz;
                        if (hh == h) {
                          continue;
                        }
                        is_solvent2  = false;
                        is_cav2      = false;              
                        is_elycav2   = false;              
                        is_channel2  = false;              
                        is_interior2 = false;
                        is_memb2     = false;              
                        is_head2     = false;
                        is_core2     = false;
                        if        (work_arr[hh] == uc_epsin1 || work_arr[hh] == uc_epsin2) {
                          is_interior2 = true;
                        } else if (work_arr[hh] == uc_epshead || work_arr[hh] == uc_epscore) {
                          is_memb2 = true;
                          if        (work_arr[hh] == uc_epshead) {
                            is_head2 = true;
                          } else if (work_arr[hh] == uc_epscore) {
                            is_core2 = true;
                          }
                        } else if (work_arr[hh] == uc_epsprehead || work_arr[hh] == uc_epsprecore ||
                                   work_arr[hh] == uc_epsupperfill || work_arr[hh] == uc_epslowerfill ||
                                   work_arr[hh] == uc_epsbothfill) {
                          is_solvent2 = true;
                          if        (work_arr[hh] == uc_epsprehead || work_arr[hh] == uc_epsprecore) {
                            is_cav2 = true;
                          } else if (work_arr[hh] == uc_epsupperfill || work_arr[hh] == uc_epslowerfill) {
                            is_elycav2 = true;
                          } else if (work_arr[hh] == uc_epsbothfill) {
                            is_channel2 = true;
                          }
                        }
                        // the last line describes overlap of the regions within the membrane with interior
                        // regions. since the latter override the former, it is OK to have overlaps here and we can still
                        // increase the size of the dummy atom
                        is_same_type = false;
                        if ((is_cav1 && is_cav2) || (is_elycav1 && is_elycav2) || (is_channel1 && is_channel2) ||
                            (is_interior1 && is_interior2) ||
                            (is_head1 && is_head2) || (is_core1 && is_core2) ||
                            (is_solvent1 && is_interior2) || (is_memb1 && is_interior2)) {
                          is_same_type = true;
                          continue;
                        }
                        // we have to touch the circumsphere of another dummy atom of different type to discard the atom
                        int sq_dist_in_grid = ((iix - ix) * (iix - ix))
                                            + ((iiy - iy) * (iiy - iy))  
                                            + ((iiz - iz) * (iiz - iz));
                        int this_dist = static_cast<int>(floor(sqrt(static_cast<float>(sq_dist_in_grid)))
                                      - static_cast<int>(wrad_arr[hh]) - static_cast<int>(istep));
                        if (this_dist <= 0) {
                          // exclude points that are not entirely surrounded by points of the same type
                          is_surrounded = false;
                          break;
                        }
                      }                      
                      // don't need to do further checks if the point is already assigned
                      if (!is_surrounded) {
                        break;
                      }
                    }
                    // don't need to do further checks if the point is already assigned
                    if (!is_surrounded) {
                      break;
                    }
                  }
                  if (is_surrounded) {
                    got_change   = true;
                    wrad_arr[h] = istep;
                    if (istep > maxrad) maxrad = istep;
                    for (int iix = (ix > istep ? (ix - istep) : 0); iix <= ((ix + istep) < (countx - 1) ? (ix + istep) : (countx - 1)); ++iix) {
                      for (int iiy = (iy > istep ? (iy - istep) : 0); iiy <= ((iy + istep) < (county - 1) ? (iy + istep) : (county - 1)); ++iiy) {
                        for (int iiz = (iz > (istep + zlowerhead_in_grid) ? (iz - istep) : zlowerhead_in_grid); iiz <= ((iz + istep) < zupperhead_in_grid ? (iz + istep) : zupperhead_in_grid); ++iiz) {
                          hh = (iix * county * countz) + (iiy * countz) + iiz;
                          if (hh == h || wrad_arr[hh] == 0) continue;
                          is_solvent2  = false;
                          is_cav2      = false;              
                          is_elycav2   = false;              
                          is_channel2  = false;              
                          is_interior2 = false;
                          is_memb2     = false;              
                          is_head2     = false;
                          is_core2     = false;
                          if        (work_arr[hh] == uc_epsin1 || work_arr[hh] == uc_epsin2) {
                            is_interior2 = true;
                          } else if (work_arr[hh] == uc_epshead || work_arr[hh] == uc_epscore) {
                            is_memb2 = true;
                            if        (work_arr[hh] == uc_epshead) {
                              is_head2 = true;
                            } else if (work_arr[hh] == uc_epscore) {
                              is_core2 = true;
                            }
                          } else if (work_arr[hh] == uc_epsprehead || work_arr[hh] == uc_epsprecore ||
                                     work_arr[hh] == uc_epsupperfill || work_arr[hh] == uc_epslowerfill ||
                                     work_arr[hh] == uc_epsbothfill) {
                            is_solvent2 = true;
                            if        (work_arr[hh] == uc_epsprehead || work_arr[hh] == uc_epsprecore) {
                              is_cav2 = true;
                            } else if (work_arr[hh] == uc_epsupperfill || work_arr[hh] == uc_epslowerfill) {
                              is_elycav2 = true;
                            } else if (work_arr[hh] == uc_epsbothfill) {
                              is_channel2 = true;
                            }
                          }
                          // the last line describes overlap of the solvent regions within the membrane with interior
                          // regions. since the latter override the former, it is OK to have overlaps here and we can still
                          // increase the size of the dummy atom
                          if ((is_cav1 && is_cav2) || (is_elycav1 && is_elycav2) || (is_channel1 && is_channel2) ||
                              (is_interior1 && is_interior2) ||
                              (is_head1 && is_head2) || (is_core1 && is_core2) ||
                              (is_solvent1 && is_interior2)) {
                            is_same_type = true;
                          }
                          if (!is_same_type) continue;
                          int sq_dist_in_grid = ((iix - ix) * (iix - ix))
                                              + ((iiy - iy) * (iiy - iy))  
                                              + ((iiz - iz) * (iiz - iz));
                          int this_dist = static_cast<int>(floor(sqrt(static_cast<float>(sq_dist_in_grid)))
                                        + static_cast<int>(wrad_arr[hh]) - static_cast<int>(istep));
                          if        (this_dist <= 0) {
                            // hh is completely included in the enlarged h
                            wrad_arr[hh] = 0;
                            nremoved++;
                          } else if (this_dist <= istep && do_shrink && noiter > 1) {
                            // the center point of hh is included in h but hh's radius prevents full inclusion
                            // check if this is a fully internal point which can be shrinked to be fully included
                            // without changing the grid coverage
                            bool is_covered = true;
                            bool did_shrink = false;
                            int tmp_wrad_hh = wrad_arr[hh];
                            while (is_covered && wrad_arr[hh] > 0) {
                              tmp_wrad_hh--;
                              is_covered = false;
                              for (int iiix = (iix > istep ? (iix - istep) : 0); iiix <= ((iix + istep) < (countx - 1) ? (iix + istep) : (countx - 1)); ++iiix) {
                                for (int iiiy = (iiy > istep ? (iiy - istep) : 0); iiiy <= ((iiy + istep) < (county - 1) ? (iiy + istep) : (county - 1)); ++iiiy) {
                                  for (int iiiz = (iiz > (istep + zlowerhead_in_grid) ? (iiz - istep) : zlowerhead_in_grid); iiiz <= ((iiz + istep) < zupperhead_in_grid ? (iiz + istep) : zupperhead_in_grid); ++iiiz) {
                                    int hhh = (iiix * county * countz) + (iiiy * countz) + iiiz;
                                    if (hhh == hh || wrad_arr[hhh] == 0) continue;
                                    is_solvent3  = false;
                                    is_cav3      = false;              
                                    is_elycav3   = false;              
                                    is_channel3  = false;              
                                    is_interior3 = false;
                                    is_memb3     = false;              
                                    is_head3     = false;
                                    is_core3     = false;
                                    is_same_type = false;
                                    if        (work_arr[hhh] == uc_epsin1 || work_arr[hhh] == uc_epsin2) {
                                      is_interior3 = true;
                                    } else if (work_arr[hhh] == uc_epshead || work_arr[hhh] == uc_epscore) {
                                      is_memb3 = true;
                                      if        (work_arr[hhh] == uc_epshead) {
                                        is_head3 = true;
                                      } else if (work_arr[hhh] == uc_epscore) {
                                        is_core3 = true;
                                      }
                                    } else if (work_arr[hhh] == uc_epsprehead || work_arr[hhh] == uc_epsprecore ||
                                               work_arr[hhh] == uc_epsupperfill || work_arr[hhh] == uc_epslowerfill ||
                                               work_arr[hhh] == uc_epsbothfill) {
                                      is_solvent3 = true;
                                      if        (work_arr[hhh] == uc_epsprehead || work_arr[hhh] == uc_epsprecore) {
                                        is_cav3 = true;
                                      } else if (work_arr[hhh] == uc_epsupperfill || work_arr[hhh] == uc_epslowerfill) {
                                        is_elycav3 = true;
                                      } else if (work_arr[hhh] == uc_epsbothfill) {
                                        is_channel3 = true;
                                      }
                                    }
                                    if ((is_cav1 && is_cav2) || (is_elycav1 && is_elycav2) || (is_channel1 && is_channel2) ||
                                        (is_interior1 && is_interior2) ||
                                        (is_head1 && is_head2) || (is_core1 && is_core2)) {
                                      is_same_type = true;
                                    } else {
                                      continue;
                                    }
                                    int sq_dist_in_grid2 = ((iix - ix) * (iix - ix))
                                                         + ((iiy - iy) * (iiy - iy)) 
                                                         + ((iiz - iz) * (iiz - iz));
                                    int this_dist2 = static_cast<int>(floor(sqrt(static_cast<float>(sq_dist_in_grid2)))
                                                   + static_cast<int>(tmp_wrad_hh) - static_cast<int>(wrad_arr[hhh]));
                                    if (this_dist2 <= 0) {
                                      is_covered = true;
                                      break;
                                    }                                  
                                  }
                                  if (is_covered) break;
                                }    
                                if (is_covered) break;
                              }
                              if (is_covered) {
                                wrad_arr[hh] = tmp_wrad_hh;
                                did_shrink = true;
                              }
                            }
                            if (did_shrink) {
                              got_change  = true;
                              got_change2 = true;
                              if (wrad_arr[hh] == 0) { nremoved++; }
                              else { nshrinked++; }
                            }
                          }
                        }
                      }
                    }
                    // we increased an atom radius at h, now it is advantageous to move at least istep from the last point on the grid
                    if        (iz + istep <= zupperhead_in_grid) {
                      iz += istep;
                    } else if (iy + istep <= county) {
                      iy += istep;
                    } else if (ix + istep <= countx) {
                      ix += istep;
                    } else if (iy > istep) {
                      iy -= istep;
                    } else if (ix > istep) {
                      ix -= istep;
                    }
                  }
                }
              }        
            }
          }
        }
        istep = ceil(sqrt(static_cast<float>(3 * istep * istep)));
        sq_istep = istep * istep;
      }
    }
    blab2 << "Finished reducing the number of dummy atoms after " << niter << " iterations (outer iterations = " << noiter << ") with maximum radius " << static_cast<float>(maxrad) * space << "." << endl;
    blab2 << "Removed " << nremoved << " superfluous dummy atoms in total." << endl << endl;

    // now we write the dummy atoms
    for (unsigned int ix = 0; ix < countx; ++ix) {
      for (unsigned int iy = 0; iy < county; ++iy) {
        for (unsigned int iz = zlowerhead_in_grid; iz <= zupperhead_in_grid; ++iz) {
          h = (ix * county * countz) + (iy * countz) + iz;
          if (wrad_arr[h] == 0) continue;
          if (work_arr[h] == uc_epsprehead  || work_arr[h] == uc_epsprecore ||
              work_arr[h] == uc_epsbothfill ||
              work_arr[h] == uc_epsupperfill || work_arr[h] == uc_epslowerfill ||
              work_arr[h] == uc_epshead || work_arr[h] == uc_epscore) {
            Atom tmpatom;
            tmpatom.coord.x = origin.x + (ix * space);
            tmpatom.coord.y = origin.y + (iy * space);
            tmpatom.coord.z = origin.z + (iz * space);
            tmpatom.charge  = 0;
            tmpatom.rad     = (static_cast<float>(wrad_arr[h]) * space) + PhysCond::get_solrad();
            if (work_arr[h] == uc_epsprehead  || work_arr[h] == uc_epsprecore) {
              tmpatom.atname  = "CAV";
              tmpatom.resname = "CAV";
              tmpatom.resnum  = acav.size() + 1;
              acav.insert(tmpatom);
            }
            else if (work_arr[h] == uc_epsbothfill) {
              tmpatom.atname  = "CNL";
              tmpatom.resname = "CNL";
              tmpatom.resnum  = achannel.size() + 1;
              achannel.insert(tmpatom);
            }
            else if (work_arr[h] == uc_epslowerfill || work_arr[h] == uc_epsupperfill) {
              tmpatom.atname  = "ECV";
              tmpatom.resname = "ECV";
              tmpatom.resnum  = aelycav.size() + 1;
              aelycav.insert(tmpatom);
            }
            else if (work_arr[h] == uc_epshead && membheadset != "") {
              tmpatom.atname  = "LHG";
              tmpatom.resname = "LHG";
              tmpatom.resnum  = ahead.size() + 1;
              ahead.insert(tmpatom);
            }
            else if (work_arr[h] == uc_epscore && membcoreset != "") {
              tmpatom.atname  = "LTL";
              tmpatom.resname = "LTL";
              tmpatom.resnum  = acore.size() + 1;
              acore.insert(tmpatom);
            }
          }
        }
      }
    }
    
    blab3 << "Created " << acav.size() << " dummy atoms representing cavities." << endl;
    blab3 << "Created " << aelycav.size() << " dummy atoms representing cavities/gorges/clefts connected to the bulk solvent." << endl;
    blab3 << "Created " << achannel.size() << " dummy atoms representing channels connecting both solvent phases." << endl;
    blab3 << "Created " << ahead.size() << " dummy atoms representing the membrane headgroup regions." << endl;
    blab3 << "Created " << acore.size() << " dummy atoms representing the membrane core." << endl;
    acav.write_pqr();
    aelycav.write_pqr();
    achannel.write_pqr();
    if (membheadset != "") {
     ahead.write_pqr();
    }
    if (membcoreset != "") {
     acore.write_pqr();
    }

    // for debugging
    if (write_dx) {
      write_dx_file(work_arr, countx, county, countz, space, origin, "regions.dx");
    }

    delete [] work_arr;

    times(&agetcav);

    blab3 << "Cavity finding runtime: " << ((float) (agetcav.tms_utime-bgetcav.tms_utime))/ticks << " s user " << ((float) (agetcav.tms_stime-bgetcav.tms_stime))/ticks <<  " s system " << endl;

  }

  delete diel;

  times(&end);
  cout << "Total runtime: " << ((float) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((float) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;

  return 0;
} // END main
