/* Generate N-dimensional curves from Monte Carlo simulation data
   generated with the program gmct from the MCT program suite
     
   get-curves-gmct molname
   
   commandline options: none   
                  
   input options are read from a setup file named molname.setup
   if no setup file is found default options apply. In this case only 
   occupation probability curves for each instance are written.
   
   input options:
   
   
   
   The pqr files can be in my extended pqr format, containing four extra
   columns (integer values) after the columns of a normal pqr file.
   They can be used to identify atoms uniquely. For standard MEAD, two
   atoms are identical, when atom name and residue name are the same.
   This is not sufficient, when working with chains and multiple instances
   of a site.
   The intended usage (but not required) is conformer number, chain number,
   site number and instance number.

   Additionally to the features needed to calculate energies in the three
   dielectric environment, this program has also the option to calculate
   the energy of the site in a homogeneous dielectric of given value.
   The purpose for this is discussed in the appropriate section below.

   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Thomas Ullmann, 2008
   
   /usr/bin/g++-4.1 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 mcdock.cc
   /usr/bin/g++-4.1 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o mcdock mcdock.o -L../../libmead_e -lmead -lm

*/

#include <iostream>
#include <iomanip>
using std::cout;
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <fstream>
using std::ifstream;
#include <string>
using std::string;
using std::ofstream;
using std::fstream;
#include <sys/times.h> // I want to do some benchmarking of options...
#include <time.h>      // for random number seed
#include <unistd.h>
#include <stdio.h>

#include "get-curves-gmct.h"

int main(int argc, char* argv[])
{

  ///////////////////// Variable declaration, ... ///////////////////////

  // Only for benchmarking
  struct tms start, end, // start and end of program
    bread, aread,        // reading GMCT data
    bcrv, acrv,          // writing curve files
    btrj, atrj;          // writing trajectory files
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);
  
  using namespace::std;
  
  CurvePara para;                     // input parameters
  string filename;                    // filename, multiple usage
  set<SiteID> siteset;                // bookkeeping set sites
  map<string,int>          confindex; // index of global conformations
  map<InstanceID,InstData> instindex; // index + bound ligands of insts.
  map<muvec,int>          muvecindex; // index of chem. pot. vectors
  double ***prob;                     // actual occupation probabilities
                                      // [instance index][mu index]
  ////////////////////// Parse the command line /////////////////////////

  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+1==argc) {  // This is the last one, must be protname
      para.molname = argv[iarg];
    }
  }
   
  //////////////// Done with comand line processing /////////////////////

  // Check command line input
  
  if (para.molname == (string) "")
    error("main: ERROR, protname not given on command line");

  // Read input file
  
  read_setup(para);  
  
  //////////////////// Done with input processing //////////////////////
  
  ////// Print summary of input from setup file/command line: //////////
  
  cout << "Starting " << argv[0] << " for molecule named " << para.molname << "\n";
  cout << "using the following settings:"   << endl;
  cout << "gcem_dir   : " << para.gcem_dir  << endl;
  cout << "curve_dir  : " << para.curve_dir << endl;
  cout << "gmct_out   : " << para.gmct_out  << endl;
  cout << "iprob      : " << para.iprob     << endl;
  cout << "ifprob     : " << para.ifprob    << endl;
  cout << "ibprob     : " << para.ibprob    << endl;
  cout << "icprob     : " << para.icprob    << endl;
  cout << "ibmean     : " << para.ibmean    << endl;
  cout << "imaxp      : " << para.imaxp     << endl;
  cout << "ifmaxp     : " << para.ifmaxp    << endl;
  cout << "ibmaxp     : " << para.ibmaxp    << endl;
  cout << "icmaxp     : " << para.icmaxp    << endl;
  cout << "ihalf      : " << para.ihalf     << endl;
  cout << "ifhalf     : " << para.ifhalf    << endl;
  cout << "ibhalf     : " << para.ibhalf    << endl;
  cout << "ichalf     : " << para.ichalf    << endl;
  cout << "iginst     : " << para.iginst    << endl;
  cout << "igform     : " << para.igform    << endl;
  cout << "igbind     : " << para.igbind    << endl;
  cout << "igconf     : " << para.igconf    << endl;
  cout << "itraj      : " << para.itraj     << endl;
  cout << "ipmf       : " << para.ipmf      << endl;
  cout << "conversion factors and decimal digits for chemical potentials:" << endl;
  for (int i = 0; i<para.conv.size(); i++)
   cout << setw(9) << para.conv[i] << " " << setw(6) << para.prec[i] <<endl;


  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ///////////////////// Setting up calculation ////////////////////////

  // read GMCT output
  
  times(&bread);
  prob = read_gmct_out(siteset,confindex,instindex,muvecindex,para);
  times(&aread);
  blab2 << "read gmct output runtime: " << ((double) (aread.tms_utime-bread.tms_utime))/ticks << " s user " << ((double) (aread.tms_stime-bread.tms_stime))/ticks <<  " s system " << endl;
  
  // write curves  
  times(&bcrv);  
  if (para.iprob == 1 || para.ibprob == 1 || para.icprob == 1 || para.ibmean == 1 || para.ibmaxp == 1 || para.icmaxp == 1 || para.ibhalf == 1 || para.ichalf == 1 || para.igbind == 1 || para.igconf == 1)
   make_curves(siteset,confindex,instindex,muvecindex,para,prob);
  times(&acrv);  
  blab2 << "curves runtime: " << ((double) (acrv.tms_utime-bcrv.tms_utime))/ticks << " s user " << ((double) (acrv.tms_stime-bcrv.tms_stime))/ticks <<  " s system " << endl;

  // write structures of most highly populated states
  times(&btrj);  
  //if (para.itraj == 1)
  // make_traj(siteset,confindex,instindex,muvecindex,para,prob);
  times(&atrj);
  blab2 << "structure trajectory runtime: " << ((double) (atrj.tms_utime-btrj.tms_utime))/ticks << " s user " << ((double) (atrj.tms_stime-btrj.tms_stime))/ticks <<  " s system " << endl;
  
  /*
  if (para.itraj == 1)
   make_traj(para.gcem_dir,para.traj_dir,siteset,confindex,instindex,muvecindex,prob);
  */
  
  deallocate_prob(confindex,instindex,muvecindex,prob);


  //////////// now the actual grid levels are defined /////////////

    

  times(&end);
  cout << "Total runtime: " << ((double) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((double) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;

  return 0;
} // END main

///////////////////////////////////////// function declaration ///////////////////////////////////////////////////

/* .....................................................*/
/* reads the setup parameters from a file               */
/*                                                      */
void read_setup(CurvePara &para)
{    
    time_t tim;
    string fn_string = (para).molname + ".gc.setup";
    FILE *file;
    char line[LINELENGTH],
         s1[LINELENGTH], s2[LINELENGTH],
         s3[LINELENGTH], s4[LINELENGTH];

    // set default values

    (para).temp        = 298.15;                       // temperature in K
    (para).seed        = -time(&tim);                  // seed for random number generator
    (para).ibmean      = 0;                            // flag for writing nD mean number of bound ligands curves 0 - no, 1 - yes
    (para).imaxp       = 0;                            // flag for writing nD most highly occupied instance curves 0 - no, 1 - yes
    (para).ifmaxp      = 0;                            // flag for writing nD most highly occupied form curves 0 - no, 1 - yes
    (para).ibmaxp      = 0;                            // flag for writing nD most highly occupied number of bound ligands curves 0 - no, 1 - yes
    (para).icmaxp      = 0;                            // flag for writing nD most highly occupied conformation curves 0 - no, 1 - yes
    (para).iprob       = 0;                            // flag for writing nD instance occupation probability curves 0 - no, 1 - yes
    (para).ifprob      = 1;                            // flag for writing nD form occupation probability curves 0 - no, 1 - yes
    (para).ibprob      = 1;                            // flag for writing nD binding probability curves 0 - no, 1 - yes
    (para).icprob      = 0;                            // flag for writing nD conformational probability curves 0 - no, 1 - yes
    (para).ihalf       = 0;                            // flag for writing nD instance midpoint/intersection curves 0 - no, 1 - yes
    (para).ifhalf      = 0;                            // flag for writing nD form midpoint/intersection curves 0 - no, 1 - yes
    (para).ibhalf      = 0;                            // flag for writing (n-1)D binding midpoint/intersection curves 0 - no, 1 - yes
    (para).ichalf      = 0;                            // flag for writing nD conformational midpoint/intersection curves 0 - no, 1 - yes
    (para).iginst      = 0;                            // flag for writing nD instance free energy curves 0 - no, 1 - yes
    (para).igform      = 0;                            // flag for writing nD form free energy curves 0 - no, 1 - yes
    (para).igbind      = 0;                            // flag for writing nD binding free energy curves 0 - no, 1 - yes
    (para).igconf      = 0;                            // flag for writing nD conformational free energy curves 0 - no, 1 - yes
    (para).itraj       = 0;                            // flag for writing structures of most highly populated states 0 - no, 1 - yes
    (para).ipmf        = 0;                            // flag for writing the curves as nmu+1-D curves vs mu1-mu2 + membpot, mu1 is expected to be pHin mu2 pHout
    (para).curve_dir   = (string) "my_curves";         // directory for writing curves and structures
    (para).gcem_dir    = (string) "..";                // directory with .est s and subdir (gcem|qmpb)_input with QMPB input structures,
                                                       // subdirs (gcem|qmpb)_dir/conf with the conformation names are expected if there
                                                       // are multiple conformations 
    (para).gmct_out    = (para).molname + ".gmct.out"; // name of GMCT output file
    
    const char *filename = fn_string.c_str();
    if ((file = fopen(filename,"r")) == 0)
        ::error("Could not open %s. Abort.\n", filename);        

    /* read setup file */
    while (fgets(line,LINELENGTH,file)!=NULL){
         if (line[0]=='!') /* comment line */
            s1[0]='!'; /* just to do something */
         else if (0==sscanf(line, "%s",s1)) /* empty line */
            s1[0]='!'; /* just to do something */
         else if (1==sscanf(line, "%s %s",s1, s2))
            error("Problem in ReadSetup() - parameter has no value");
         else if (0==strcmp(s1,"temp")){
            (para).temp=atof(s2);
            PhysCond::set_T(atof(s2));
         }
         else if (0==strcmp(s1,"ipmf"))
            (para).ipmf=atoi(s2);
         else if (0==strcmp(s1,"itraj"))
            (para).itraj=atoi(s2);
         else if (0==strcmp(s1,"iginst"))
            (para).iginst=atoi(s2);
         else if (0==strcmp(s1,"igform"))
            (para).igform=atoi(s2);
         else if (0==strcmp(s1,"igbind"))
            (para).igbind=atoi(s2);
         else if (0==strcmp(s1,"igconf"))
            (para).igconf=atoi(s2);
         else if (0==strcmp(s1,"ibmean"))
            (para).ibmean=atoi(s2);
         else if (0==strcmp(s1,"imaxp"))
            (para).imaxp=atoi(s2);
         else if (0==strcmp(s1,"ifmaxp"))
            (para).ifmaxp=atoi(s2);
         else if (0==strcmp(s1,"ibmaxp"))
            (para).ibmaxp=atoi(s2);
         else if (0==strcmp(s1,"icmaxp"))
            (para).icmaxp=atoi(s2);
         else if (0==strcmp(s1,"ihalf"))
            (para).ihalf=atoi(s2);
         else if (0==strcmp(s1,"ifhalf"))
            (para).ifhalf=atoi(s2);
         else if (0==strcmp(s1,"ibhalf"))
            (para).ibhalf=atoi(s2);
         else if (0==strcmp(s1,"ichalf"))
            (para).ichalf=atoi(s2);
         else if (0==strcmp(s1,"iprob"))
            (para).iprob=atoi(s2);
         else if (0==strcmp(s1,"ifprob"))
            (para).ifprob=atoi(s2);
         else if (0==strcmp(s1,"ibprob"))
            (para).ibprob=atoi(s2);
         else if (0==strcmp(s1,"icprob"))
            (para).icprob=atoi(s2);
         else if (0==strcmp(s1,"blab"))
             {int tmpint = atoi(s2);
                 if (tmpint > 0) blab1pt = &cout;
                 if (tmpint > 1) blab2pt = &cout;
                 if (tmpint > 2) blab3pt = &cout;
             }
         else if (0==strcmp(s1,"seed"))
             {(para).seed=(long)atoi(s2);
              if ((para).seed>0)
                 (para).seed=-(para).seed;
             }
         else if (0==strcmp(s1,"gcem_dir") || 0==strcmp(s1,"qmpb_dir"))
            (para).gcem_dir=std::string(s2);
         else if (0==strcmp(s1,"curve_dir"))
            (para).curve_dir=std::string(s2);
         else if (0==strcmp(s1,"gmct_outfile"))
            (para).gmct_out=std::string(s2);         
         else if (0==strcmp(s1,"nformat")){
              int nfmt =atoi(s2);
              for (int i=1; i<=nfmt; i++){
                    if (fgets(line,LINELENGTH,file)==NULL)\
			         ::error("InputFile too short.");
                    if (line[0]=='!'){ // comment line
                       i--; continue;
                    }
                    else if (0==sscanf(line, "%s",s1)) // empty line
                             i--;
                    else if (3!=sscanf(line, "%s %s %s",s1, s2, s3)){
                           fprintf(stderr,"%s",line);
                           ::error("too few words in format");
                    }
                    if (0==strcmp(s1,"format")){
                      para.conv.push_back(atof(s2));
                      para.prec.push_back(atoi(s3));
                      //if (para.conv[i-1] > 0) para.conv[i-1] *= -1;
                      if (para.prec[i-1] < 0) para.prec[i-1] *= -1;                      
                    }
                    else ::error("too few format descriptors");
              }
         }
         else {fprintf(stderr,"%s\n",s1);
               error("Unknown word in ReadSetup(): %s", s1);
         }
    }
    fclose(file); /* setup file */
    
    // fill default format descriptors if none are given in setup file
    // first pK units / pH
    // second meV
    if (para.conv.size()==0) {
     para.conv.push_back((double) - PhysCond::get_pkconv()/PhysCond::get_econv()); 
     para.conv.push_back(-4184000/(PhysCond::get_Avogadro_Number()*PhysCond::get_elementary_charge())); 
     para.prec.push_back((int) 1);
     para.prec.push_back((int) 0);
    }    
    
}

/* .....................................................*/
/* reads the GMCT data from a file                      */
/*                                                      */
void deallocate_prob(const map<string,int>& confindex, 
                     const map<InstanceID,InstData>& instindex, 
                     const map<muvec,int>& muvecindex,
                     double ***prob){
  blab2 << "deallocate_prob: starting\n";
  int tmpi1 = confindex.size();
  int tmpi2 = instindex.size();
  for (int ii = 0; ii<tmpi1;ii++){
   for (int jj = 0; jj<tmpi2;jj++){
    delete [] prob[ii][jj];
    prob[ii][jj] = 0;
   }  
   delete [] prob[ii];
   prob[ii] = 0;  
  }
  delete [] prob;
  prob = 0;
  blab1 << "deallocate_prob: exiting\n";
}


/* .................................................................*/
/* reads the GMCT data from a file safe variant with regex          */
/*                                                                  */
double*** read_gmct_out( 
                   set<SiteID>& siteset, map<string,int>& confindex,
                   map<InstanceID,InstData>& instindex,
                   map<muvec,int>& muvecindex,CurvePara& para){
  
  blab2 << "read_gmct_out: starting\n";
  
  string filename = para.gmct_out;
    
  double ***prob;
  string line, tmpstr0, tmpstr1, tmpstr2, tmpstr3, tmpstr4, tmpstr5, tmpstr6, conf;
  int    tmpi, tmpi1, tmpi2, tmpi3;
  double tmpd, tmpprob;
  SiteID tmpsiteid;
  InstanceID tmpinstid;
  string::const_iterator start, end;
  muvec lastmv;

  // dont want:
  // ! ...
  // blank line
  // chemical potential  -0.272850  -13.836330   vmemb 0.000000
  // want:
  // conf_NTala-0-1_NTP-1           1.000000e+00 (-2.728497e-01 1 ...) vmemb 0.000000e+00
  
  boost::regex re_prob("^\\s*([^\\s]+)_([^\\s]+)-([^\\s]+)-([^\\s]+)_([^\\s]+)-([^\\s]+)\\s+([\\d.+-eE]+)\\s+(.+)\\s+vmemb\\s+([\\d.+-eE]+)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;  

  const char *fn = filename.c_str();
  std::ifstream file(fn);
  if (!file)
    ::error ("CharmmPara::read_top: Cannot open input file, ", fn, "\n");
    
  while (! file.eof() ){
      
   getline (file, line);

   start = line.begin(); end = line.end();
   
   if (boost::regex_search(start, end, match, re_prob, flags)){
    vector<double> tmpvec; // chemical potential vector
    InstData tmpdata;  // contains numerical instance id and ligand vector
    // this match should contain nmu pairs of chemical potential and
    // corresponding number of bound ligands, i.e., double int, ...
    tmpstr1 = std::string(match[8].first, match[8].second);
    stringstream ss1(tmpstr1);
    while (ss1){
     ss1 >> tmpd >> tmpi;
     if (ss1.fail()){break;} else{
      tmpvec.push_back(tmpd); // chemical potential ...
      // ... and corresponding number of bound ligands
      tmpdata.ligands.push_back(tmpi);
     }
    }
    // add membrane potential to muvec
    // membrane potential does not correspond to ligand
    tmpdata.ligands.push_back(0);
    tmpstr6 = string(match[9].first, match[9].second);
    stringstream ss2(tmpstr6);
    ss2 >> tmpd;
    tmpvec.push_back(tmpd);
    muvec tmpmv(tmpvec);
    if (tmpmv!=lastmv){
     muvecindex.insert(map<muvec,int>::value_type(tmpmv,muvecindex.size()));
     lastmv = tmpmv;
    }
    
    // confs, sites, instances are the same for each chemical potential
    // avoid costly map search/insertion operations
    if (muvecindex.size()>1) continue;
    conf    = std::string(match[1].first, match[1].second);     
    confindex.insert(map<string,int>::value_type(conf,confindex.size()));

    tmpstr1 = string(match[2].first, match[2].second);
    tmpstr2 = string(match[3].first, match[3].second);
    tmpstr3 = string(match[4].first, match[4].second);
    // the conformation is not important for the InstID and SiteID here
    // since conf is indexed separately -> set conf zero in InstID and SiteID
    tmpsiteid = SiteID(tmpstr1,0,atoi(tmpstr2.c_str()),atoi(tmpstr3.c_str()));
    siteset.insert(tmpsiteid);
    
    tmpstr1 = string(match[5].first, match[5].second);
    tmpstr2 = string(match[6].first, match[6].second);

    tmpinstid = InstanceID(tmpsiteid.get_sitename(), tmpstr1,
                           0, tmpsiteid.get_chain(), tmpsiteid.get_site(),
                           atoi(tmpstr2.c_str()));     
    tmpdata.id = instindex.size(); 
    instindex.insert(map<InstanceID,InstData>::value_type(tmpinstid,tmpdata));
            
   } // end regex match
      
  } // end while
  
  
  blab3 << "read_gmct_out: read " << confindex.size()  << " unique conformations." << endl;
  blab3 << "read_gmct_out: read " << siteset.size()    << " unique sites." << endl;
  blab3 << "read_gmct_out: read " << instindex.size()  << " unique insts." << endl;
  blab3 << "read_gmct_out: read " << muvecindex.size() << " unique chemical potential vectors." << endl;
  
  // if vmemb is always zero and no pmf is requested omit it altogether
  bool has_vmemb = false;
  for (map<muvec,int>::const_iterator iter=muvecindex.begin();iter!=muvecindex.end();iter++)
   if (iter->first.get_mu(iter->first.get_nmu() - 1) != 0)
    has_vmemb = true;
  if (has_vmemb) blab3 << "read_gmct_out: found non-zero membrane potential " << endl;
  if (has_vmemb) para.has_vmemb = 1;
  else para.has_vmemb = 0;
  
  if (para.ipmf==1){
   // add pmf as last potential, remove pH out if present
   map<muvec,int> tmp_muvecindex(muvecindex);
   map<InstanceID,InstData> tmp_instindex(instindex);
   muvecindex.clear();
   instindex.clear();
   for (map<InstanceID,InstData>::const_iterator iter=tmp_instindex.begin();iter!=tmp_instindex.end();iter++){
    InstData tmpdata;
    tmpdata.id = iter->second.id;
    for (int i=0;i<(iter->second.ligands.size());i++)
     // if we have pH-in and pH-out 
     //  omit pH-out
     //  add number of bound protons from outside to pH-in
     if (!(i==1 && iter->second.ligands.size()>2))
      tmpdata.ligands.push_back(iter->second.ligands[i]);
     else
      tmpdata.ligands[0] += iter->second.ligands[i];
    instindex.insert(map<InstanceID,InstData>::value_type(iter->first,tmpdata));  
   }
   for (map<muvec,int>::const_iterator iter=tmp_muvecindex.begin();iter!=tmp_muvecindex.end();iter++){
    vector<double> tmpvec;
    int i = iter->first.get_nmu();
    double pmf = 0;
    if      (i<2) 
      ::error("Error in read_gmct_out: pmf requested by input parameters but no chemical potentials found.");
    else if (i==2){
      if (iter == tmp_muvecindex.begin())
        cerr << "Warning in read_gmct_out: pmf requested by input parameters but only 1 chemical potential found." 
             << endl << "Did you miss pH-out or pH-in?" << endl;
      tmpvec.push_back(iter->first.get_mu(int(0)));
      // last potential should be the membrane potential
      pmf = -iter->first.get_mu(1);
      tmpvec.push_back(pmf);
    }
    else if (i>2){
      // i-1 should be the membrane potential, 0 should be pH-in, 1 should be pH-out
      pmf = iter->first.get_mu(1) - iter->first.get_mu(0) - iter->first.get_mu(i-1);
      tmpvec.push_back(iter->first.get_mu(int(0)));
      // keep all potentials except pH-out and membrane potential
      for (int i=2;i<(iter->first.get_nmu()-1);i++)
       tmpvec.push_back(iter->first.get_mu(i));
      tmpvec.push_back(pmf);
    }
        
    muvec tmpmv(tmpvec);
    muvecindex.insert(map<muvec,int>::value_type(tmpmv,muvecindex.size()));
   }
  }

  // delete membrane potential from muvecs if it is always zero and no pmf is requested
  else if (!has_vmemb && para.ipmf==0){
   map<muvec,int> tmp_muvecindex(muvecindex);
   map<InstanceID,InstData> tmp_instindex(instindex);
   instindex.clear();
   muvecindex.clear();
   for (map<InstanceID,InstData>::const_iterator iter=tmp_instindex.begin();iter!=tmp_instindex.end();iter++){
    InstData tmpdata;
    tmpdata.id = iter->second.id;
    for (int i=0;i<(iter->second.ligands.size()-1);i++)
     tmpdata.ligands.push_back(iter->second.ligands[i]);
    instindex.insert(map<InstanceID,InstData>::value_type(iter->first,tmpdata));  
   }
   for (map<muvec,int>::const_iterator iter=tmp_muvecindex.begin();iter!=tmp_muvecindex.end();iter++){
    vector<double> tmpvec;
    for (int i=0;i<(iter->first.get_nmu()-1);i++){    
     tmpvec.push_back(iter->first.get_mu(i));
    }
    muvec tmpmv(tmpvec);
    muvecindex.insert(map<muvec,int>::value_type(tmpmv,muvecindex.size()));
   }
  }

  // allocate memory for the protonation probabilities
  tmpi1 = confindex.size();
  tmpi2 = instindex.size();
  tmpi3 = muvecindex.size();
  
  try{
   prob       = new double**[tmpi1];  
   for (int ii = 0; ii<tmpi1;ii++){
    prob[ii] = new double*[tmpi2];
    for (int jj = 0; jj<tmpi2;jj++){
     prob[ii][jj] = new double[tmpi3];
    }  
   }
  } catch(bad_alloc){
    cerr << "Insufficent memory for tensor!" << endl;
    exit(0);
  };

  for (int ii = 0; ii<tmpi1;ii++){
   for (int jj = 0; jj<tmpi2;jj++){
    for (int kk = 0; kk<tmpi3;kk++){
     prob[ii][jj][kk] = 0;
    }
   }
  }

  // FIXME: why does this not work
  // go back to begin of file
  //file.seekg(0, ios::beg); 
  
  file.close();
  file.open(fn);
  
  while (! file.eof() ){
      
   getline (file, line); 
   start = line.begin(); end = line.end();
   
   if (boost::regex_search(start, end, match, re_prob, flags)){
    
    tmpstr0 = std::string(match[7].first, match[7].second);
    stringstream ss0(tmpstr0);
    ss0 >> tmpprob;
    int i = 0;
    double pmf = 0;

    vector<double> tmpvec; // chemical potential vector
    // this match should contain nmu pairs of chemical potential and
    // corresponding number of bound ligands, i.e., double int, ...
    tmpstr1 = std::string(match[8].first, match[8].second);
    stringstream ss1(tmpstr1);
    while (ss1){
     ss1 >> tmpd >> tmpi;
     if (ss1.fail()){break;} 
     else {
      if (!(i==1 && para.ipmf==1))
       tmpvec.push_back(tmpd);
     }
     // calculate dpH = pH-out - pH-in
     if      (i==1) pmf += tmpd;
     else if (i==0) pmf -= tmpd;
     i++;     
    } 
    // if only one pH is given dpH=0
    if (i<2) pmf = 0;
    if ((has_vmemb && para.ipmf == 0) || para.ipmf == 1){
     tmpstr6 = string(match[9].first, match[9].second);
     stringstream ss2(tmpstr6);
     ss2 >> tmpd;
     // if no pmf is requested add membrane potential to muvec
     if (para.ipmf == 0)
      tmpvec.push_back(tmpd);
     // if pmf is requested add membrane potential to pmf and pmf to muvec
     else {
      tmpvec.push_back((pmf-tmpd));
     }
    }
    muvec tmpmv(tmpvec);

    conf    = std::string(match[1].first, match[1].second);    
    map<string,int>::const_iterator citer = confindex.find(conf);
    if (citer != confindex.end())
      tmpi1 = citer->second;
    else
      ::error("Error in read_gmct_out: could not find conf in index.");
    map<muvec,int>::const_iterator mviter = muvecindex.find(tmpmv);
    if (mviter != muvecindex.end())
      tmpi3 = mviter->second;    
    else
      ::error("Error in read_gmct_out: could not find muvec in index.");

    tmpstr1 = string(match[2].first, match[2].second);
    tmpstr2 = string(match[3].first, match[3].second);
    tmpstr3 = string(match[4].first, match[4].second);
    
    tmpstr4 = string(match[5].first, match[5].second);
    tmpstr5 = string(match[6].first, match[6].second);    
    
    // conf is always set to 0 in IDs
    tmpinstid = InstanceID(tmpstr1,tmpstr4,0,atoi(tmpstr2.c_str()),
                           atoi(tmpstr3.c_str()),atoi(tmpstr5.c_str()));     
    map<InstanceID,InstData>::const_iterator iiter = instindex.find(tmpinstid);
    if (iiter != instindex.end())
      tmpi2 = iiter->second.id;
    else
      ::error("Error in read_gmct_out: could not find instance in index.");
    
    prob[tmpi1][tmpi2][tmpi3] = tmpprob;
  
            
   } // end regex match
      
  } // end while
  
  file.close();
        
  // FIXME! why is this not working when called from main
  // seems like the array is not returned in contrast to
  // all the other items
  //deallocate_prob(confindex,instindex,muvecindex,prob);     
  return prob;
  
  blab1 << "read_gmct_out: exiting\n";
}

/* .................................................................*/
/* reads the GMCT data from a file,                                 */
/* variant with strings/stringstreams, unexpectedly slower than the */
/* regex variant                                                    */
/*                                                                  */
/* DEPRECATED
double*** read_gmct_out_noregex(const std::string& filename, 
                   set<SiteID>& siteset, map<string,int>& confindex,
                   map<InstanceID,InstData>& instindex,
                   map<muvec,int>& muvecindex){
  
  blab2 << "read_gmct_out: starting\n";
    
  double ***prob;
  string line, tmpstr0, tmpstr1, tmpstr2, tmpstr3, tmpstr4, tmpstr5, conf;
  string muline = "chemical potential";
  char cline[LINELENGTH], s1[LINELENGTH], s2[LINELENGTH], s3[LINELENGTH];
  int    tmpi, tmpi1, tmpi2, tmpi3, count, rdcnt;
  double tmpd, tmpprob;
  SiteID tmpsiteid;
  InstanceID tmpinstid;
  
  //
  // dont want:
  // ! ...
  // blank line
  // chemical potential  -0.272850  -13.836330   vmemb 0.000000
  // want:
  // conf_NTala-0-1_NTP-1           1.000000e+00 (-2.728497e-01 1 ...) vmemb 0.000000e+00
  //  
 
  const char *fn = filename.c_str();
  ifstream file(fn);
  if (!file)
    ::error ("CharmmPara::read_top: Cannot open input file, ", fn, "\n");
    
  while (!file.eof()){
      
   getline (file, line);   
   string::const_iterator iter = line.begin();
   // skip comment lines, lines beginning with "chemical potential"
   // empty lines are excluded later, performance should not suffer
   // because empty lines are expected to be rare
   if (*iter=='!')
    continue;
   else {
    string::const_iterator iter2 = line.begin();
    string::const_iterator iter3 = muline.begin();
    while (iter3!=muline.end() && iter2 !=line.end() && *iter2==*iter3){    
     iter2++; iter3++;
    } iter3++;
    if (iter3==muline.end()){continue;}
   }
   
   // check if we get at least the pattern "string float float int"
   stringstream ss1(line);
   ss1 >> tmpstr1 >> tmpprob >> tmpd >> tmpi;
   // skip non-matching line
   if (ss1.fail()){continue;}
   else{
    // split first string in line at '_' into conf, site, inst
    int start, end, count;
    start = end = count = 0;
    for (string::iterator iter=tmpstr1.begin();iter!=tmpstr1.end();iter++){
     if (*iter=='_'){
      count++;
      if      (count==1)
       conf = tmpstr1.substr(start,end);
      else if (count==2)
       tmpstr2 = tmpstr1.substr(start,end);
      start = end + 2;
     }
     end++;
    }
    tmpstr3 = tmpstr1.substr(start);
    if (count!=2) continue;
    // save conformation
    confindex.insert(map<string,int>::value_type(conf,confindex.size()));

    // split sitelabel at '-' into sitename, chainid, siteid
    for (string::iterator iter=tmpstr2.begin();iter!=tmpstr2.end();iter++)
     if (*iter=='-') *iter = ' ';
    
    stringstream ss2(tmpstr2);
    ss2 >> tmpstr1 >> tmpi1 >> tmpi2;
    if (ss2.fail()){continue;}

    // split instlabel at '-' into sitename, chainid, siteid
    for (string::iterator iter=tmpstr3.begin();iter!=tmpstr3.end();iter++)
     if (*iter=='-') *iter = ' ';

    stringstream ss3(tmpstr3);
    ss3 >> tmpstr2 >> tmpi3;
    if (ss3.fail()){continue;}
    ss3 >> tmpi3;
    // instance id/number is not mandatory in sitelabel
    if (ss3.fail()){tmpi3 = 1;}    
    
    // save site and instance information
    //  the conformation is not important for the InstID and SiteID here
    //  since conf is indexed separately -> set conf zero in InstID and SiteID
    tmpsiteid = SiteID(tmpstr1,0,tmpi1,tmpi2);
    siteset.insert(tmpsiteid);
    tmpinstid = InstanceID(tmpstr1,tmpstr2,0,tmpi1,tmpi2,tmpi3);     
    // save chemical potentials and corresponding numbers of bound ligands
    vector<double> tmpvec; // chemical potential vector
    tmpvec.push_back(tmpd);
    InstData tmpdata;  // contains numerical instance id and ligand vector
    tmpdata.ligands.push_back(tmpi);
    // ss1 should still contain (nmu-1) pairs of chemical potential and
    // corresponding number of bound ligands, i.e., double int, ...
    while (ss1){
     ss1 >> tmpd >> tmpi;
     if (ss1.fail()){break;} else{
      tmpvec.push_back(tmpd); // chemical potential ...
      // ... and corresponding number of bound ligands
      tmpdata.ligands.push_back(tmpi);
     }
    }
    tmpdata.id = instindex.size(); 
    instindex.insert(map<InstanceID,InstData>::value_type(tmpinstid,tmpdata));
    muvec tmpmv(tmpvec);
    muvecindex.insert(map<muvec,int>::value_type(tmpmv,muvecindex.size()));
                
   } // end match
      
  } // end while
    
  blab3 << "read_gmct_out: read " << confindex.size()  << " unique conformations." << endl;
  blab3 << "read_gmct_out: read " << siteset.size()    << " unique sites." << endl;
  blab3 << "read_gmct_out: read " << instindex.size()  << " unique insts." << endl;
  blab3 << "read_gmct_out: read " << muvecindex.size() << " unique chemical potential vectors." << endl;
  
  // allocate memory for the protonation probabilities
  tmpi1 = confindex.size();
  tmpi2 = instindex.size();
  tmpi3 = muvecindex.size();
  
  try{
   prob       = new double**[tmpi1];  
   for (int ii = 0; ii<tmpi1;ii++){
    prob[ii] = new double*[tmpi2];
    for (int jj = 0; jj<tmpi2;jj++){
     prob[ii][jj] = new double[tmpi3];
    }  
   }
  } catch(bad_alloc){
    cerr << "Insufficent memory for tensor!" << endl;
    exit(0);
  };

   for (int ii = 0; ii<tmpi1;ii++){
    for (int jj = 0; jj<tmpi2;jj++){
     for (int kk = 0; kk<tmpi3;kk++){
      prob[ii][jj][kk] = 0;
     }
    }
   }

  // FIXME: why does this not work
  // go back to begin of file
  //file.seekg(0, ios::beg); 
  
  file.close();
  file.open(fn);
  
  while (! file.eof() ){
      
   getline (file, line);   
   string::const_iterator iter = line.begin();
   // skip comment lines, lines beginning with "chemical potential"
   // empty lines are excluded later, performance should not suffer
   // because empty lines are expected to be rare
   if (*iter=='!')
    continue;
   else {
    string::const_iterator iter2 = line.begin();
    string::const_iterator iter3 = muline.begin();
    while (iter3!=muline.end() && iter2 !=line.end() && *iter2==*iter3){    
     iter2++; iter3++;
    } iter3++;
    if (iter3==muline.end()) continue;
   }
      
   // check if we get at least the pattern "string float float int"
   stringstream ss1(line);
   ss1 >> tmpstr1 >> tmpprob >> tmpd >> tmpi;
   // skip non-matching line
   if (ss1.fail()){continue;}
   else{
    // split first string in line at '_' into conf, site, inst
    int start, end, count;
    start = end = count = 0;
    for (string::iterator iter=tmpstr1.begin();iter!=tmpstr1.end();iter++){
     if (*iter=='_'){
      count++;
      if      (count==1)
       conf = tmpstr1.substr(start,end);
      else if (count==2)
       tmpstr2 = tmpstr1.substr(start,end);
      start = end + 2;
     }
     end++;
    }
    tmpstr3 = tmpstr1.substr(start);
    if (count!=2) continue;

    // split sitelabel at '-' into sitename, chainid, siteid
    for (string::iterator iter=tmpstr2.begin();iter!=tmpstr2.end();iter++)
     if (*iter=='-') *iter = ' ';
    
    stringstream ss2(tmpstr2);
    ss2 >> tmpstr1 >> tmpi1 >> tmpi2;
    if (ss2.fail()){continue;}

    // split instlabel at '-' into sitename, chainid, siteid
    for (string::iterator iter=tmpstr3.begin();iter!=tmpstr3.end();iter++)
     if (*iter=='-') *iter = ' ';

    stringstream ss3(tmpstr3);
    ss3 >> tmpstr2 >> tmpi3;
    if (ss3.fail()){continue;}
    ss3 >> tmpi3;
    // instance id/number is not mandatory in sitelabel
    if (ss3.fail()){tmpi3 = 1;}    
    
    // save site and instance information
    //  the conformation is not important for the InstID and SiteID here
    //  since conf is indexed separately -> set conf zero in InstID and SiteID
    tmpinstid = InstanceID(tmpstr1,tmpstr2,0,tmpi1,tmpi2,tmpi3);     
    
    // save chemical potentials and corresponding numbers of bound ligands
    vector<double> tmpvec; // chemical potential vector
    tmpvec.push_back(tmpd);
    // ss1 should still contain (nmu-1) pairs of chemical potential and
    // corresponding number of bound ligands, i.e., double int, ...
    while (ss1){
     ss1 >> tmpd >> tmpi;
     if (ss1.fail()){break;} else{
      tmpvec.push_back(tmpd); // chemical potential ...
     }
    }
    muvec tmpmv(tmpvec);
   
    // find indices corresponding to conf, index, muvec
    map<string,int>::const_iterator citer = confindex.find(conf);
    if (citer != confindex.end())
      tmpi1 = citer->second;
    else
      ::error("Error in read_gmct_out: could not find conf in index.");
    map<InstanceID,InstData>::const_iterator iiter = instindex.find(tmpinstid);
    if (iiter != instindex.end())
      tmpi2 = iiter->second.id;
    else
      ::error("Error in read_gmct_out: could not find instance in index.");
    map<muvec,int>::const_iterator mviter = muvecindex.find(tmpmv);
    if (mviter != muvecindex.end())
      tmpi3 = mviter->second;    
    else
      ::error("Error in read_gmct_out: could not find muvec in index.");
        
    // save probability
    prob[tmpi1][tmpi2][tmpi3] = tmpprob;  
            
   } // end regex match
      
  } // end while
  
  file.close();
        
  // FIXME! why is this not working than called from main
  // seems like the array is not returned in contrast to
  // all the other items
  //deallocate_prob(confindex,instindex,muvecindex,prob);
  return prob;
  
  blab1 << "read_gmct_out: exiting\n";
}
*/

/* .................................................................*/
/* writes probability curves for each 0,1,...,n ligands of one      */
/* species bound without regard to other ligand species             */
/*                                                                  */
void make_curves(const set<SiteID>& siteset,
                 const map<string,int>& confindex,
                 const map<InstanceID,InstData>& instindex,
                 const map<muvec,int>& muvecindex,
                 const CurvePara& para, 
                 double ***prob){

  blab2 << "make_curves: starting\n";
 
  int i, j, k;

  string prob_curve_dir  = para.curve_dir + "/inst_prob";  
  string fprob_curve_dir = para.curve_dir + "/form_prob";  
  string bprob_curve_dir = para.curve_dir + "/bind_prob";
  string cprob_curve_dir = para.curve_dir + "/conf_prob";
  string bmean_curve_dir = para.curve_dir + "/bind_mean";
  string maxp_curve_dir  = para.curve_dir + "/inst_maxp";
  string fmaxp_curve_dir = para.curve_dir + "/form_maxp";
  string bmaxp_curve_dir = para.curve_dir + "/bind_maxp";
  string cmaxp_curve_dir = para.curve_dir + "/conf_maxp";
  string half_curve_dir  = para.curve_dir + "/inst_half";
  string fhalf_curve_dir = para.curve_dir + "/form_half";
  string bhalf_curve_dir = para.curve_dir + "/bind_half";
  string chalf_curve_dir = para.curve_dir + "/conf_half";
  string ginst_curve_dir = para.curve_dir + "/ginst";
  string gform_curve_dir = para.curve_dir + "/gform";
  string gbind_curve_dir = para.curve_dir + "/gbind";
  string gconf_curve_dir = para.curve_dir + "/gconf";
  string form_index_dir  = para.curve_dir + "/form_index";
  string conf_index_dir  = para.curve_dir + "/conf_index";
  string traj_dir        = para.curve_dir + "/traj";

  // create directories  
  if (para.iprob == 1 || para.ibprob == 1 || para.icprob == 1 || para.ibmean == 1 || para.ibmaxp == 1 || para.icmaxp == 1 || para.ibhalf == 1 || para.ichalf == 1 || para.igbind == 1 || para.igconf == 1){
   if (!fs::is_directory(para.curve_dir.c_str())){
    if (fs::exists(para.curve_dir.c_str())) fs::remove_all(para.curve_dir.c_str());
    fs::create_directory(para.curve_dir.c_str());
   }
  }
  if (para.iprob == 1){   
   if (!fs::is_directory(prob_curve_dir.c_str())){
    if (fs::exists(prob_curve_dir.c_str())) fs::remove_all(prob_curve_dir.c_str());
    fs::create_directory(prob_curve_dir.c_str());
   }
  }
  if (para.ifprob == 1){   
   if (!fs::is_directory(fprob_curve_dir.c_str())){
    if (fs::exists(fprob_curve_dir.c_str())) fs::remove_all(fprob_curve_dir.c_str());
    fs::create_directory(fprob_curve_dir.c_str());
   }
  }
  if (para.ibprob == 1){   
   if (!fs::is_directory(bprob_curve_dir.c_str())){
    if (fs::exists(bprob_curve_dir.c_str())) fs::remove_all(bprob_curve_dir.c_str());
    fs::create_directory(bprob_curve_dir.c_str());
   }
  }  if (para.icprob == 1){   
   if (!fs::is_directory(cprob_curve_dir.c_str())){
    if (fs::exists(cprob_curve_dir.c_str())) fs::remove_all(cprob_curve_dir.c_str());
    fs::create_directory(cprob_curve_dir.c_str());
   }
  }
  if (para.ibmean == 1){   
   if (!fs::is_directory(bmean_curve_dir.c_str())){
    if (fs::exists(bmean_curve_dir.c_str())) fs::remove_all(bmean_curve_dir.c_str());
    fs::create_directory(bmean_curve_dir.c_str());
   }
  }
  if (para.imaxp == 1){   
   if (!fs::is_directory(maxp_curve_dir.c_str())){
    if (fs::exists(maxp_curve_dir.c_str())) fs::remove_all(maxp_curve_dir.c_str());
    fs::create_directory(maxp_curve_dir.c_str());
   }
  }
  if (para.ifmaxp == 1){   
   if (!fs::is_directory(fmaxp_curve_dir.c_str())){
    if (fs::exists(fmaxp_curve_dir.c_str())) fs::remove_all(fmaxp_curve_dir.c_str());
    fs::create_directory(fmaxp_curve_dir.c_str());
   }
  }
  if (para.ibmaxp == 1){   
   if (!fs::is_directory(bmaxp_curve_dir.c_str())){
    if (fs::exists(bmaxp_curve_dir.c_str())) fs::remove_all(bmaxp_curve_dir.c_str());
    fs::create_directory(bmaxp_curve_dir.c_str());
   }
  }
  if (para.icmaxp == 1){   
   if (!fs::is_directory(cmaxp_curve_dir.c_str())){
    if (fs::exists(cmaxp_curve_dir.c_str())) fs::remove_all(cmaxp_curve_dir.c_str());
    fs::create_directory(cmaxp_curve_dir.c_str());
   }
  }
  if (para.ihalf == 1){   
   if (!fs::is_directory(half_curve_dir.c_str())){
    if (fs::exists(half_curve_dir.c_str())) fs::remove_all(half_curve_dir.c_str());
    fs::create_directory(half_curve_dir.c_str());
   }
  }
  if (para.ibhalf == 1){   
   if (!fs::is_directory(bhalf_curve_dir.c_str())){
    if (fs::exists(bhalf_curve_dir.c_str())) fs::remove_all(bhalf_curve_dir.c_str());
    fs::create_directory(bhalf_curve_dir.c_str());
   }
  }
  if (para.ifhalf == 1){   
   if (!fs::is_directory(fhalf_curve_dir.c_str())){
    if (fs::exists(fhalf_curve_dir.c_str())) fs::remove_all(fhalf_curve_dir.c_str());
    fs::create_directory(fhalf_curve_dir.c_str());
   }
  }
  if (para.ichalf == 1){   
   if (!fs::is_directory(chalf_curve_dir.c_str())){
    if (fs::exists(chalf_curve_dir.c_str())) fs::remove_all(chalf_curve_dir.c_str());
    fs::create_directory(chalf_curve_dir.c_str());
   }
  }
  if (para.iginst == 1){   
   if (!fs::is_directory(ginst_curve_dir.c_str())){
    if (fs::exists(ginst_curve_dir.c_str())) fs::remove_all(ginst_curve_dir.c_str());
    fs::create_directory(ginst_curve_dir.c_str());
   }
  }
  if (para.igform == 1){   
   if (!fs::is_directory(gform_curve_dir.c_str())){
    if (fs::exists(gform_curve_dir.c_str())) fs::remove_all(gform_curve_dir.c_str());
    fs::create_directory(gform_curve_dir.c_str());
   }
  }
  if (para.igbind == 1){   
   if (!fs::is_directory(gbind_curve_dir.c_str())){
    if (fs::exists(gbind_curve_dir.c_str())) fs::remove_all(gbind_curve_dir.c_str());
    fs::create_directory(gbind_curve_dir.c_str());
   }
  }
  if (para.igconf == 1){   
   if (!fs::is_directory(gconf_curve_dir.c_str())){
    if (fs::exists(gconf_curve_dir.c_str())) fs::remove_all(gconf_curve_dir.c_str());
    fs::create_directory(gconf_curve_dir.c_str());
   }
  }
  if (para.igform == 1 || para.ifprob == 1 || para.ifmaxp == 1 || para.ifhalf == 1){   
   if (!fs::is_directory(form_index_dir.c_str())){
    if (fs::exists(form_index_dir.c_str())) fs::remove_all(form_index_dir.c_str());
    fs::create_directory(form_index_dir.c_str());
   }
  }
  if (para.igconf == 1 || para.icprob == 1 || para.icmaxp == 1 || para.ichalf == 1){   
   if (!fs::is_directory(conf_index_dir.c_str())){
    if (fs::exists(conf_index_dir.c_str())) fs::remove_all(conf_index_dir.c_str());
    fs::create_directory(conf_index_dir.c_str());
   }
  }
  if (para.itraj == 1){   
   if (!fs::is_directory(traj_dir.c_str())){
    if (fs::exists(traj_dir.c_str())) fs::remove_all(traj_dir.c_str());
    fs::create_directory(traj_dir.c_str());
   }
  }
    
  // determine number of chemical potentials 
  map<muvec,int>::const_iterator tmpmiter = muvecindex.begin();
  int nmu = tmpmiter->first.get_nmu();  

  // prepare format parameters
  vector<double> conv(para.conv);
  vector<int>   prec(para.prec); 
  if (conv.size()<nmu){
   int nfill = nmu - conv.size();
   for (int i=0; i<nfill;i++){
    conv.push_back((double) 1.0);
   }
  }
  if (prec.size()<nmu){
   int nfill = nmu - prec.size();
   for (int i=0; i<nfill;i++){
    prec.push_back((int) 2);
   }
  }

  // make instance related curves
  if (para.iprob == 1  || para.iginst == 1|| para.ihalf == 1  || para.imaxp  == 1 || 
      para.ifprob == 1 || para.igform == 1|| para.ifhalf == 1 || para.ifmaxp == 1 ||
      para.icprob == 1 || para.igconf == 1|| para.ichalf == 1 || para.icmaxp == 1 || para.itraj == 1){

   blab2 << "writing curves for individual instances and forms\n";   
   // sum up probabilities of instances over conformations
   
   double **sprob;
   int tmpi1 = instindex.size();
   int tmpi2 = muvecindex.size();
   int tmpi3 = confindex.size();
   
   // allocate memory for summed up probabilities and initialize to zero
   try{
    sprob = new double*[tmpi1];
    for (int ii=0;ii<tmpi1;ii++){
     sprob[ii]  = new double[tmpi2];
     for (int jj=0;jj<tmpi2;jj++)
      sprob[ii][jj] = 0;
    }
   } catch(bad_alloc){
     cerr << "Error in make_curves: Insufficent memory for tensor sprob!" << endl;
     exit(0);
   }
   
   // sum up probabilities
    for (int ii=0;ii<tmpi1;ii++)
     for (int jj=0;jj<tmpi2;jj++)
      for (int hh=0;hh<tmpi3;hh++)
       sprob[ii][jj] += prob[hh][ii][jj];
   
   // write maximum populated state strutures
   if (para.itraj == 1){

    blab3 << "writing maximum populated state structures\n";
        
    string tmpstr = traj_dir;
     
    for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
     int kk = miter->second;
     string fn_str = tmpstr + "/";
     for (int i=0;i<nmu;i++){
      stringstream ss1;
      string tmpstr1;
      double ten_p_prec = pow((double)10,prec[i]);
      double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec;
      if (val==0) val*= -1;
      if    (prec[i] != 0) ss1 << setfill('0') << setw(5+prec[i]) << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
      else                 ss1 << setfill('0') << setw(5+prec[i]) << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
      ss1 >> tmpstr1;
      if (i!=0) fn_str += "_"; 
      fn_str += tmpstr1;        
     }
     fn_str += ".pqr";
     AtomSet mpa, background;
     string bgname;
     string maxconfstr;
     // for each site find the most highly populated instance, read and add its structure to mpa
     for (set<SiteID>::const_iterator siter=siteset.begin();siter!=siteset.end();siter++){    
      // determine highest populated instance jmp with index jj
      InstanceID jmp;
      int        jjmp;
      double maxp = 0;
      for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
       // check if instance belongs to site
       if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
        continue;
       int jj = iiter->second.id;
       if (sprob[jj][kk]>maxp){
        maxp=sprob[jj][kk];
        jmp = iiter->first;
        jjmp = iiter->second.id;
       }
      }
      // find out which global conformation ist most populated
      if (siter==siteset.begin() && confindex.size()>1){
       double maxp = 0;       
       for (map<string,int>::const_iterator citer= confindex.begin();citer!=confindex.end();citer++){
        int cc=citer->second; 
        if (prob[cc][jjmp][kk]>maxp){
         maxconfstr = citer->first;
        }
       }
      }
      string pqrname;
      if (confindex.size()>1)
       pqrname = para.gcem_dir + "/" + maxconfstr + "/qmpb_input/";
      else
       pqrname = para.gcem_dir + "/gcem_input/";
      bgname = pqrname + para.molname + ".background.pqr";
      string tmpstr1, tmpstr2, tmpstr3;
      stringstream ss1, ss2, ss3;
      ss1 << jmp.get_chain();    ss1 >> tmpstr1;
      ss2 << jmp.get_site();     ss2 >> tmpstr2;
      ss3 << jmp.get_instance(); ss3 >> tmpstr3;
      pqrname += jmp.get_sitename() + "-" + tmpstr1 + "-" + tmpstr2 
              + "_" + jmp.get_instname() + "-" + tmpstr3 + ".pqr";
      AtomSet a1;
      a1.read(pqrname);
      for (AtomSet::const_iterator i=a1.begin();i!=a1.end();i++) 
       mpa.insert(i->second);

     }
     // read background
     background.read(bgname);
     for (AtomSet::const_iterator i=background.begin();i!=background.end();i++) 
      mpa.insert(i->second);
     
     // extended pqr format is not read correctly by vmd
     //mpa.write_pqr(fn_str);

     FILE * pqr;
     pqr = fopen (fn_str.c_str(),"w");
     if (pqr == NULL)
      ::error ("ERROR in write_curves: while writing pqr: Cannot open file, ", fn_str.c_str(), "\n");   
     int c = 1;
     for (AtomSet::const_iterator i=mpa.begin();i!=mpa.end();i++) {
      if (i->second.atname.length() == 4){
       fprintf (pqr, "ATOM %6d %-4s %-3s%6d%10.3f%8.3f%8.3f  %9.6f  %6.3f\n", 
                c, i->second.atname.c_str(), i->second.resname.c_str(), i->second.resnum, 
                i->second.coord.x, i->second.coord.y, i->second.coord.z, i->second.charge, 
                i->second.rad); 
      }
      else {
       fprintf (pqr, "ATOM %6d  %-3.4s %-3s%6d%10.3f%8.3f%8.3f  %9.6f  %6.3f\n", 
                c, i->second.atname.c_str(), i->second.resname.c_str(), i->second.resnum, 
                i->second.coord.x, i->second.coord.y, i->second.coord.z, i->second.charge, 
                i->second.rad); 
      }
      c++; 
     }
     fclose(pqr);
      
    }
   
   } // end write maximum populated state structures
   
   // write instance probability curves   
   if (para.iprob == 1){

    blab3 << "writing instance occupation probability curves\n";
    
    string tmpstr = prob_curve_dir;
    
     for (map<InstanceID,InstData>::const_iterator iiter = instindex.begin(); iiter != instindex.end(); iiter++){
      int jj = iiter->second.id;
      string tmpstr1, tmpstr2, tmpstr3;
      stringstream ss1, ss2, ss3;
      ss1 << iiter->first.get_chain();    ss1 >> tmpstr1;
      ss2 << iiter->first.get_site();     ss2 >> tmpstr2;
      ss3 << iiter->first.get_instance(); ss3 >> tmpstr3;
      string fn_str = tmpstr + "/" + iiter->first.get_sitename() + "-"
                                   + tmpstr1 + "-" + tmpstr2 + "_"
                                   + iiter->first.get_instname() + "-" + tmpstr3 + ".txt";
      std::ofstream of(fn_str.c_str());
      if (!of.good())
       ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
     
      for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
       int kk = miter->second;
       for (int i=0;i<nmu;i++){
        double ten_p_prec = pow((double)10,prec[i]);
        double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
        if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
        else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
       }  
       of << setprecision(6) << scientific << sprob[jj][kk] << endl;
      }
     
      if (!of.good())
       ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
      of.close();
     
     }
   
   } // end instance probability curves
      
   // write maximum populated instance curves
   if (para.imaxp == 1){

    blab3 << "writing maximum populated instance curves\n";
        
    for (set<SiteID>::const_iterator siter=siteset.begin();siter!=siteset.end();siter++){    

     string tmpstr = maxp_curve_dir;
     string tmpstr1, tmpstr2;
     stringstream ss1, ss2;
     ss1 << (*siter).get_chain(); ss1 >> tmpstr1;
     ss2 << (*siter).get_site();  ss2 >> tmpstr2;
     string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                  + tmpstr1 + "-" + tmpstr2 + ".txt";     
     std::ofstream of(fn_str.c_str());
     if (!of.good())
      ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
     
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      int kk = miter->second;
      // determine highest populated instance jmp with index jj
      int jmp = 0;
      double maxp = 0;
      for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
       // check if instance belongs to site
       if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
        continue;
       int jj = iiter->second.id;
       if (sprob[jj][kk]>maxp){
        maxp=sprob[jj][kk];
        jmp = iiter->first.get_instance();
       }
      }
      
      for (int i=0;i<nmu;i++){
       double ten_p_prec = pow((double)10,prec[i]);
       double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
       if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
       else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
      }  
      of << jmp << endl; 
             
      }     

      if (!of.good())
       ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
      of.close();
      
    }
   
   } // end write maximum populated instance curves
      
   // write instance free energy curves   
   if (para.iginst == 1){

    blab3 << "writing instance free energy curves\n";

    const double RT(PhysCond::get_Boltzmann_constant() * PhysCond::get_Avogadro_Number() * PhysCond::get_T() / 4184);
    string tmpstr = ginst_curve_dir;
        
    for (set<SiteID>::const_iterator siter=siteset.begin();siter!=siteset.end();siter++){    
          
     double *maxp = new double[muvecindex.size()];
     // for each muvec determine the maximum population of an instance belonging to this site
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      int kk = miter->second;
      maxp[kk] = 0;
      for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
       // check if instance belongs to site
       if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
        continue;
       int jj = iiter->second.id;
       if (sprob[jj][kk]>maxp[kk])
        maxp[kk] = sprob[jj][kk];
      }
     }

     for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
      if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
       continue;
      int jj = iiter->second.id;
      
      string tmpstr1, tmpstr2, tmpstr3;
      stringstream ss1, ss2, ss3;
      ss1 << (*siter).get_chain();        ss1 >> tmpstr1;
      ss2 << (*siter).get_site();         ss2 >> tmpstr2;
      ss3 << iiter->first.get_instance(); ss3 >> tmpstr3;
      string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                  + tmpstr1 + "-" + tmpstr2 + "_" 
                                  + iiter->first.get_instname() + "-"
                                  + tmpstr3 + ".txt";     
      std::ofstream of(fn_str.c_str());
      if (!of.good())
       ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
      
      for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      int kk = miter->second;      
       for (int i=0;i<nmu;i++){
        double ten_p_prec = pow((double)10,prec[i]);
        double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
        if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
        else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
       }  
       double ginst;  
       if (sprob[jj][kk]>0)
        ginst = -RT * log(sprob[jj][kk]/maxp[kk]);
       else
        ginst = 9999;
       // ginst always > 0 since relative to lowest energy / highest populated
       // gets minus sign due to constant if zero
       if (ginst<=0) ginst *= -1;
       of << setprecision(6) << fixed << ginst << endl;             
      }
 
      if (!of.good())
       ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
      of.close();
           
     }     
    
     delete [] maxp;
    
    }

   } // end write instance free energy curves
   
   // create instance intersection curves
   if (para.ihalf == 1){

    blab3 << "writing instance half/intersection curves\n";
    
    string tmpstr = half_curve_dir;
    for (int ii=0;ii<nmu;ii++){
     int tmpi3 = muvecindex.size();
     // make index of chemical potential combinations without the
     // variable dimension for which we wish to determine the half point
     map<muvec,int> submvi; // index for subspace of muvecs with mu of ligand excluded
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      vector<double> tmpmv;
      for (int oo=0;oo<nmu;oo++){
       if (oo==ii) continue;
       tmpmv.push_back(miter->first.get_mu(oo));
      }
      map<muvec,int>::const_iterator tmiter = submvi.find(tmpmv);
      if (tmiter == submvi.end())
       submvi.insert(map<muvec,int>::value_type(tmpmv,submvi.size()));
     }
     // among all muvec in submvi make an index
     // of mu values for ligand ii
     map<double,int> iimvi;
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      double tmpd = 0;
      for (int oo=0;oo<nmu;oo++){
       if (oo==ii) tmpd = miter->first.get_mu(oo);
      }
      map<double,int>::const_iterator tmi = iimvi.find(tmpd);
      if (tmi==iimvi.end()) iimvi.insert(map<double,int>::value_type(tmpd,iimvi.size()));
     }
            
     // allocate memory for subset of summed up probabilities
     // subprob[inst][submvi][iimvi], initialize to zero
     double ***subsprob;
     try{
      int tmpi1 = instindex.size();
      int tmpi2 = submvi.size();
      int tmpi3 = iimvi.size();
      subsprob  = new double**[tmpi1];
      for (int hh=0;hh<tmpi1;hh++){
       subsprob[hh] = new double*[tmpi2];
       for (int gg=0;gg<tmpi2;gg++){
        subsprob[hh][gg] = new double[tmpi3];
        // zero out subsprob
        for (int ff=0;ff<tmpi3;ff++){
         subsprob[hh][gg][ff] = 0;
        }
       }
      }
     } catch(bad_alloc){
       cerr << "Error in make_curves: Insufficent memory for tensor subsprob!" << endl;
       exit(0);
     };
     
     // fill subsprob with actual values
     for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
      int gg = smiter->second;
      for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
       int ff = iimiter->second;
       vector<double> tmpmv;
       int pp = 0;
       for (int oo=0;oo<nmu;oo++){
        if (oo==ii){tmpmv.push_back(iimiter->first); pp--;}
        else tmpmv.push_back(smiter->first.get_mu(pp));
        pp++;
       }
       map<muvec,int>::const_iterator miter = muvecindex.find(muvec(tmpmv));
       // grid does not always contain all possible combinations of potential values
       if (miter==muvecindex.end()) continue;
       int kk = miter->second;
       for (int hh=0;hh<instindex.size();hh++){
        subsprob[hh][gg][ff] = sprob[hh][kk];
       }
      }
     }
     
     // for each point submuveci and each combination jj,jj2 of instances
     // interpolate the mu for ligand type ii at which prob(jj*ii) = prob(jj2*ii)       
     string tmpstr5;
     stringstream ss5;
     ss5 << ii;  ss5 >> tmpstr5;
     for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
      int jj = iiter->second.id;
      string tmpstr1, tmpstr2, tmpstr3;
      stringstream ss1, ss2, ss3;
      ss1 << iiter->first.get_chain();    ss1 >> tmpstr1;
      ss2 << iiter->first.get_site();     ss2 >> tmpstr2;
      ss3 << iiter->first.get_instance(); ss3 >> tmpstr3;      

      map<InstanceID,InstData>::const_iterator iiter2=iiter;
      for (iiter2++;iiter2!=instindex.end();iiter2++){
       // check if the instances belong to the same site
       if (iiter->first.get_sitename()!=iiter2->first.get_sitename() || iiter->first.get_chain()!=iiter2->first.get_chain() || iiter->first.get_site()!=iiter2->first.get_site())
        continue;       
       int jj2 = iiter2->second.id;
       string tmpstr4;
       stringstream ss4;
       ss4 << iiter2->first.get_instance(); ss4 >> tmpstr4;

       string fn_str = tmpstr + "/" + iiter->first.get_sitename() + "-"
                                    + tmpstr1 + "-" + tmpstr2 + "_" 
                                    + iiter->first.get_instname()  + "-" + tmpstr3 + "-"
                                    + iiter2->first.get_instname() + "-" + tmpstr4 + "_"
                                    + tmpstr5 + ".txt";      
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
       
       for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
       // multiple intersections or none at all could occur
        vector<double> isp;
        int gg = smiter->second;
        // traverse chemical potential belonging to ligand ii, 
        // search for pairs of successive points in which p1-p2 changes sign
        for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
         map<double,int>::const_iterator iimiter2 = iimiter; iimiter2++;
         if (iimiter2==iimvi.end()) break;
         int ff  = iimiter->second;
         int ff2 = iimiter2->second;
         // do not accept intersections in insignificant probability < 1%
         if ((subsprob[jj][gg][ff]<1.0E-02 && subsprob[jj][gg][ff2]<1.0E-02) || 
             (subsprob[jj2][gg][ff]<1.0E-02 && subsprob[jj2][gg][ff2]<1.0E-02<1.0E-02)){
          continue;
         }
         if ((subsprob[jj][gg][ff]>=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]<=subsprob[jj2][gg][ff2]) ||
             (subsprob[jj][gg][ff]<=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]>=subsprob[jj2][gg][ff2])){
          // calculate mu at intersection point assuming an intersection of two lines through
          // the points (iimiter->first, corresp sprob(jj));  (iimiter2->first, corresp. sprob(jj)) and
          // the points (iimiter->first, corresp sprob(jj2)); (iimiter2->first, corresp. sprob(jj2))
          double tmpisp  = (subsprob[jj][gg][ff] - subsprob[jj2][gg][ff]) * (iimiter2->first - iimiter->first); 
                 tmpisp /= subsprob[jj2][gg][ff2] - subsprob[jj][gg][ff2]
                         + subsprob[jj][gg][ff]   - subsprob[jj2][gg][ff];
                 tmpisp  += iimiter->first;
          isp.push_back(tmpisp);
         }
        }
        if (isp.size()!=0){
         // write the intersections to file
         // first chemical potentials          
         int j = 0;
         for (int i=0;i<nmu;i++){
          if (i==ii) continue;
          double ten_p_prec = pow((double)10,prec[i]);
          double val = round(ten_p_prec * smiter->first.get_mu(j) * conv[i]) / ten_p_prec;
          // avoid minus sign before zero potential
          if (val == 0) val = 0; 
          if (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
          else              of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
          j++;           
         }
         // then all intersections
         for (int vi=0;vi<isp.size();vi++){
          double ten_p_prec = pow((double)10,prec[ii]+1);
          double val = round(ten_p_prec * isp[vi] * conv[ii]) / ten_p_prec;            
          if (prec[ii] != 0) of << setprecision(prec[ii]+1) << fixed <<   showpoint << val;
          else               of << setprecision(prec[ii]+1) << fixed << noshowpoint << val;
          if (vi<(isp.size()-1)){of << " ";}
         }
         of << endl;
        }
       }
       
       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
       of.close();         
      }
     } // end intersection/half-points
            
     // deallocate memory for summed up probabilities
     int tmpi1 = instindex.size();
     int tmpi2 = submvi.size();
     for (int hh=0;hh<tmpi1;hh++){
      for (int gg=0;gg<tmpi2;gg++){
       delete [] subsprob[hh][gg];
      }
      delete [] subsprob[hh];
     }        
     delete [] subsprob;

    }  
    
   } // end instance intersection points / half points
   
   if (para.ifprob == 1 || para.igform == 1|| para.ifhalf == 1 || para.ifmaxp == 1 ||
       para.icprob == 1 || para.igconf == 1|| para.ichalf == 1 || para.icmaxp == 1){
    
    for (set<SiteID>::const_iterator siter=siteset.begin();siter!=siteset.end();siter++){
     
     // make index of forms in this site
     map<string,int> formindex;
     string firstform("");
     for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
      // check if the instances belong to the same site
      if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
       continue;       
      formindex.insert(map<string,int>::value_type(iiter->first.get_instname(),formindex.size()));
      if (firstform == "") firstform = iiter->first.get_instname();
     }
     // get number of conformers in this site
     int nconf = 0;
     vector<list<InstanceID> > cinst;
     list<InstanceID> instlist;
     for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
      // check if the instances belong to the same site
      if (iiter->first.get_sitename()!=(*siter).get_sitename() || iiter->first.get_chain()!=(*siter).get_chain() || iiter->first.get_site()!=(*siter).get_site())
       continue;       
      if (iiter->first.get_instname()==firstform){
       nconf++;
       if (instlist.size()>0){
        cinst.push_back(instlist);
        instlist.clear();     
       }
      }
      instlist.push_back(iiter->first);
     }
     if (instlist.size()>0){
      cinst.push_back(instlist);
      instlist.clear();     
     }
     
     double **sfprob, **scprob;
 
     // write forms and their indices in the curves to file
     if (para.ifprob == 1 || para.igform == 1|| para.ifhalf == 1 || para.ifmaxp == 1){
      
      string tmpstr = form_index_dir;
      
      string tmpstr1, tmpstr2;
      stringstream ss1, ss2;
      ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
      ss2 << (*siter).get_site();     ss2 >> tmpstr2;
      string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                    + tmpstr1 + "-" + tmpstr2 + ".txt";
      std::ofstream of(fn_str.c_str());

      for (map<string,int>::const_iterator fiter = formindex.begin(); fiter != formindex.end(); fiter++)
       of << fiter->first << " " << fiter->second << endl;
      
      if (!of.good())
       ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
     }

     // write associated instances of conformations and their indices in the curves to file
     if (para.icprob == 1 || para.igconf == 1|| para.ichalf == 1 || para.icmaxp == 1){
      
      string tmpstr = conf_index_dir;      
      
      string tmpstr1, tmpstr2;
      stringstream ss1, ss2;
      ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
      ss2 << (*siter).get_site();     ss2 >> tmpstr2;
      string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                    + tmpstr1 + "-" + tmpstr2 + ".txt";
      std::ofstream of(fn_str.c_str());

      for (int c=0;c<nconf;c++){
       for (list<InstanceID>::const_iterator citer=cinst[c].begin();citer!=cinst[c].end();citer++) 
        of << (*citer).get_instname() << "-" << (*citer).get_instance() << " ";
       of << "  " << c << endl;
      }
      if (!of.good())
       ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
     }
     
     if (para.ifprob == 1 || para.igform == 1|| para.ifhalf == 1 || para.ifmaxp == 1){
      
      int tmpi1 = formindex.size();
      int tmpi2 = muvecindex.size();
      // allocate memory for summed up form probabilities and initialize to zero
      try{
       sfprob = new double*[tmpi1];
       for (int ii=0;ii<tmpi1;ii++){
        sfprob[ii]  = new double[tmpi2];
        for (int jj=0;jj<tmpi2;jj++)
         sfprob[ii][jj] = 0;
       }
      } catch(bad_alloc){
        cerr << "Error in make_curves: Insufficent memory for tensor sfprob!" << endl;
        exit(0);
      }
            
      // sum up probabilities
      for (map<string,int>::const_iterator fiter=formindex.begin();fiter!=formindex.end();fiter++){
       int ff = fiter->second;
       for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin();iiter!=instindex.end();iiter++){
        if (iiter->first.get_sitename()==(*siter).get_sitename() && iiter->first.get_chain()==(*siter).get_chain() && iiter->first.get_site()==(*siter).get_site()){
         if (fiter->first == iiter->first.get_instname()){
          int ii = iiter->second.id;
          for (map<muvec,int>::const_iterator miter=muvecindex.begin();miter!=muvecindex.end();miter++){
           int jj = miter->second;
           sfprob[ff][jj] += sprob[ii][jj];
          }
         }
        }
       }
      }
           
     } // end sfprob
     
     if (para.icprob == 1 || para.igconf == 1|| para.ichalf == 1 || para.icmaxp == 1){
      
      int tmpi1 = nconf;
      int tmpi2 = muvecindex.size();
      // allocate memory for summed up form probabilities and initialize to zero
      try{
       scprob = new double*[tmpi1];
       for (int ii=0;ii<tmpi1;ii++){
        scprob[ii]  = new double[tmpi2];
        for (int jj=0;jj<tmpi2;jj++)
         scprob[ii][jj] = 0;
       }
      } catch(bad_alloc){
        cerr << "Error in make_curves: Insufficent memory for tensor sfprob!" << endl;
        exit(0);
      }
         
      // sum up probabilities      
      for (int c=0;c<nconf;c++){
       for (list<InstanceID>::const_iterator liter=cinst[c].begin();liter!=cinst[c].end();liter++){
        map<InstanceID,InstData>::const_iterator iiter = instindex.find(*liter);
        if (iiter==instindex.end()){
         cerr << *liter << endl;
         ::error("Error could not find instance in index.");
        }
        int ii = iiter->second.id;
        for (map<muvec,int>::const_iterator miter=muvecindex.begin();miter!=muvecindex.end();miter++){
         int jj = miter->second;
         scprob[c][jj] += sprob[ii][jj];
        } 
       }
      }
          
     } // end scprob

     // write form probability curves    
     if (para.ifprob == 1){
      
      string tmpstr = fprob_curve_dir;
      
       for (map<string,int>::const_iterator fiter = formindex.begin(); fiter != formindex.end(); fiter++){
        int jj = fiter->second;
        string tmpstr1, tmpstr2;
        stringstream ss1, ss2;
        ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
        ss2 << (*siter).get_site();     ss2 >> tmpstr2;
        string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                     + tmpstr1 + "-" + tmpstr2 + "_"
                                     + fiter->first + ".txt";
        std::ofstream of(fn_str.c_str());
        if (!of.good())
         ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
        
        for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
         int kk = miter->second;
         for (int i=0;i<nmu;i++){
          double ten_p_prec = pow((double)10,prec[i]);
          double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
          if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
          else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
         }  
         of << setprecision(6) << scientific << sfprob[jj][kk] << endl;
        }
       
        if (!of.good())
         ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
        of.close();
       
       }
     
     } // end form probability curves
             
     // write maximum populated form curves
     if (para.ifmaxp == 1){

      string tmpstr = fmaxp_curve_dir;
      string tmpstr1, tmpstr2;
      stringstream ss1, ss2;
      ss1 << (*siter).get_chain(); ss1 >> tmpstr1;
      ss2 << (*siter).get_site();  ss2 >> tmpstr2;
      string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                   + tmpstr1 + "-" + tmpstr2 + ".txt";     
      std::ofstream of(fn_str.c_str());
      if (!of.good())
       ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
              
      for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
       int kk = miter->second;
       // determine highest populated instance jmp with index jj
       int jmp = 0;
       double maxp = 0;
       for (map<string,int>::const_iterator fiter=formindex.begin();fiter!=formindex.end();fiter++){
        int jj = fiter->second;
        if (sfprob[jj][kk]>maxp){
         maxp = sfprob[jj][kk];
         jmp  = fiter->second;
        }
       }
       
       for (int i=0;i<nmu;i++){
        double ten_p_prec = pow((double)10,prec[i]);
        double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
        if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
        else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
       }  
       of << jmp << endl; 
              
      }     

      if (!of.good())
       ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
      of.close();
     
     } // end write maximum populated form curves
        
     // write form free energy curves   
     if (para.igform == 1){

      const double RT(PhysCond::get_Boltzmann_constant() * PhysCond::get_Avogadro_Number() * PhysCond::get_T() / 4184);
      string tmpstr = gform_curve_dir;
          
      double *maxp = new double[muvecindex.size()];
      // for each muvec determine the maximum population of an form belonging to this site
      for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
       int kk = miter->second;
       maxp[kk] = 0;
       for (map<string,int>::const_iterator fiter=formindex.begin();fiter!=formindex.end();fiter++){
        int jj = fiter->second;
        if (sfprob[jj][kk]>maxp[kk])
         maxp[kk] = sfprob[jj][kk];
       }
      }

      for (map<string,int>::const_iterator fiter=formindex.begin();fiter!=formindex.end();fiter++){
       int jj = fiter->second;
       
       string tmpstr1, tmpstr2;
       stringstream ss1, ss2;
       ss1 << (*siter).get_chain();        ss1 >> tmpstr1;
       ss2 << (*siter).get_site();         ss2 >> tmpstr2;
       string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                   + tmpstr1 + "-" + tmpstr2 + "_" 
                                   + fiter->first + ".txt";     
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
       
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        int kk = miter->second;      
        for (int i=0;i<nmu;i++){
         double ten_p_prec = pow((double)10,prec[i]);
         double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
         if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
         else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
        }  
        double gform;  
        if (sprob[jj][kk]>0)
         gform = -RT * log(sfprob[jj][kk]/maxp[kk]);
        else
         gform = 9999;
        // gform always > 0 since relative to lowest energy / highest populated
        // gets minus sign due to constant if zero
        if (gform<=0) gform *= -1;
        of << setprecision(6) << fixed << gform << endl;             
       }
   
       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
       of.close();
            
      }     
      
      delete [] maxp;

     } // end write form free energy curves     
     
     // create form intersection curves
     if (para.ifhalf == 1){
      
      string tmpstr = fhalf_curve_dir;
      for (int ii=0;ii<nmu;ii++){
       int tmpi3 = muvecindex.size();
       // make index of chemical potential combinations without the
       // variable dimension for which we wish to determine the half point
       map<muvec,int> submvi; // index for subspace of muvecs with mu of ligand excluded
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        vector<double> tmpmv;
        for (int oo=0;oo<nmu;oo++){
         if (oo==ii) continue;
         tmpmv.push_back(miter->first.get_mu(oo));
        }
        map<muvec,int>::const_iterator tmiter = submvi.find(tmpmv);
        if (tmiter == submvi.end())
         submvi.insert(map<muvec,int>::value_type(tmpmv,submvi.size()));
       }
       // among all muvec in submvi make an index
       // of mu values for ligand ii
       map<double,int> iimvi;
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        double tmpd = 0;
        for (int oo=0;oo<nmu;oo++){
         if (oo==ii) tmpd = miter->first.get_mu(oo);
        }
        map<double,int>::const_iterator tmi = iimvi.find(tmpd);
        if (tmi==iimvi.end()) iimvi.insert(map<double,int>::value_type(tmpd,iimvi.size()));
       }
              
       // allocate memory for subset of summed up probabilities
       // subprob[inst][submvi][iimvi], initialize to zero
       double ***subsprob;
       try{
        int tmpi1 = formindex.size();
        int tmpi2 = submvi.size();
        int tmpi3 = iimvi.size();
        subsprob  = new double**[tmpi1];
        for (int hh=0;hh<tmpi1;hh++){
         subsprob[hh] = new double*[tmpi2];
         for (int gg=0;gg<tmpi2;gg++){
          subsprob[hh][gg] = new double[tmpi3];
          // zero out subsprob
          for (int ff=0;ff<tmpi3;ff++){
           subsprob[hh][gg][ff] = 0;
          }
         }
        }
       } catch(bad_alloc){
         cerr << "Error in make_curves: Insufficent memory for tensor subsprob!" << endl;
         exit(0);
       };
       
       // fill subsprob with actual values
       for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
        int gg = smiter->second;
        for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
         int ff = iimiter->second;
         vector<double> tmpmv;
         int pp = 0;
         for (int oo=0;oo<nmu;oo++){
          if (oo==ii){tmpmv.push_back(iimiter->first); pp--;}
          else tmpmv.push_back(smiter->first.get_mu(pp));
          pp++;
         }
         map<muvec,int>::const_iterator miter = muvecindex.find(muvec(tmpmv));
         // grid does not always contain all possible combinations of potential values
         if (miter==muvecindex.end()) continue;
         int kk = miter->second;
         for (int hh=0;hh<formindex.size();hh++){
          subsprob[hh][gg][ff] = sfprob[hh][kk];
         }
        }
       }
       
       // for each point submuveci and each combination jj,jj2 of forms
       // interpolate the mu for ligand type ii at which prob(jj*ii) = prob(jj2*ii)       
       string tmpstr3;
       stringstream ss3;
       ss3 << ii;  ss3 >> tmpstr3;
       for (map<string,int>::const_iterator fiter=formindex.begin();fiter!=formindex.end();fiter++){
        int jj = fiter->second;
        string tmpstr1, tmpstr2;
        stringstream ss1, ss2;
        ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
        ss2 << (*siter).get_site();     ss2 >> tmpstr2;

        map<string,int>::const_iterator fiter2=fiter;
        for (fiter2++;fiter2!=formindex.end();fiter2++){
         int jj2 = fiter2->second;

         string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                      + tmpstr1 + "-" + tmpstr2 + "_" 
                                      + fiter->first + "-" + fiter2->first
                                      + "_" + tmpstr3 + ".txt";      
         std::ofstream of(fn_str.c_str());
         if (!of.good())
          ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
         
         for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
         // multiple intersections or none at all could occur
          vector<double> isp;
          int gg = smiter->second;
          // traverse chemical potential belonging to ligand ii, 
          // search for pairs of successive points in which p1-p2 changes sign
          for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
           map<double,int>::const_iterator iimiter2 = iimiter; iimiter2++;
           if (iimiter2==iimvi.end()) break;
           int ff  = iimiter->second;
           int ff2 = iimiter2->second;
           // do not accept intersections in insignificant probability < 1%
           if ((subsprob[jj][gg][ff]<1.0E-02 && subsprob[jj][gg][ff2]<1.0E-02) || 
               (subsprob[jj2][gg][ff]<1.0E-02 && subsprob[jj2][gg][ff2]<1.0E-02<1.0E-02)){
            continue;
           }
           if ((subsprob[jj][gg][ff]>=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]<=subsprob[jj2][gg][ff2]) ||
               (subsprob[jj][gg][ff]<=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]>=subsprob[jj2][gg][ff2])){
            // calculate mu at intersection point assuming an intersection of two lines through
            // the points (iimiter->first, corresp sprob(jj));  (iimiter2->first, corresp. sprob(jj)) and
            // the points (iimiter->first, corresp sprob(jj2)); (iimiter2->first, corresp. sprob(jj2))
            double tmpisp  = (subsprob[jj][gg][ff] - subsprob[jj2][gg][ff]) * (iimiter2->first - iimiter->first); 
                   tmpisp /= subsprob[jj2][gg][ff2] - subsprob[jj][gg][ff2]
                           + subsprob[jj][gg][ff]   - subsprob[jj2][gg][ff];
                   tmpisp  += iimiter->first;
            isp.push_back(tmpisp);
           }
          }
          if (isp.size()!=0){
           // write the intersections to file
           // first chemical potentials          
           int j = 0;
           for (int i=0;i<nmu;i++){
            if (i==ii) continue;
            double ten_p_prec = pow((double)10,prec[i]);
            double val = round(ten_p_prec * smiter->first.get_mu(j) * conv[i]) / ten_p_prec;
            // avoid minus sign before zero potential
            if (val == 0) val = 0; 
            if (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
            else              of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
            j++;           
           }
           // then all intersections
           for (int vi=0;vi<isp.size();vi++){
            double ten_p_prec = pow((double)10,prec[ii]+1);
            double val = round(ten_p_prec * isp[vi] * conv[ii]) / ten_p_prec;            
            if (prec[ii] != 0) of << setprecision(prec[ii]+1) << fixed <<   showpoint << val;
            else               of << setprecision(prec[ii]+1) << fixed << noshowpoint << val;
            if (vi<(isp.size()-1)){of << " ";}
           }
           of << endl;
          }
         }
         
         if (!of.good())
          ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
         of.close();         
        }
       } // end intersection/half-points
              
       // deallocate memory for summed up probabilities
       int tmpi1 = formindex.size();
       int tmpi2 = submvi.size();
       for (int hh=0;hh<tmpi1;hh++){
        for (int gg=0;gg<tmpi2;gg++){
         delete [] subsprob[hh][gg];
        }
        delete [] subsprob[hh];
       }        
       delete [] subsprob;

      }  
      
     } // end form intersection points / half points

     // write conformation probability curves    
     if (para.icprob == 1){
      
      string tmpstr = cprob_curve_dir;
      
      for (int jj=0;jj<nconf;jj++){
       string tmpstr1, tmpstr2, tmpstr3;
       stringstream ss1, ss2, ss3;
       ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
       ss2 << (*siter).get_site();     ss2 >> tmpstr2;
       ss2 << jj;                      ss2 >> tmpstr3;
       string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                    + tmpstr1 + "-" + tmpstr2 + "_"
                                    + tmpstr3 + ".txt";
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
      
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        int kk = miter->second;
        for (int i=0;i<nmu;i++){
         double ten_p_prec = pow((double)10,prec[i]);
         double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
         if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
         else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
        }  
        of << setprecision(6) << scientific << scprob[jj][kk] << endl;
       }
      
       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
       of.close();
      
      }
     
     } // end conformation probability curves
             
     // write maximum populated conformation curves
     if (para.ifmaxp == 1){

       string tmpstr = cmaxp_curve_dir;
       string tmpstr1, tmpstr2;
       stringstream ss1, ss2;
       ss1 << (*siter).get_chain(); ss1 >> tmpstr1;
       ss2 << (*siter).get_site();  ss2 >> tmpstr2;
       string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                    + tmpstr1 + "-" + tmpstr2 + ".txt";     
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
               
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        int kk = miter->second;
        // determine highest populated instance jmp with index jj
        int jmp = 0;
        double maxp = 0;
        for (int jj=0;jj<nconf;jj++){
         if (scprob[jj][kk]>maxp){
          maxp = scprob[jj][kk];
          jmp  = jj;
         }
        }
        
        for (int i=0;i<nmu;i++){
         double ten_p_prec = pow((double)10,prec[i]);
         double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
         if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
         else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
        }  
        of << jmp << endl; 
               
       }     

       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
       of.close();
     
     } // end write maximum populated conformation curves
        
     // write conformation free energy curves   
     if (para.igconf == 1){

      const double RT(PhysCond::get_Boltzmann_constant() * PhysCond::get_Avogadro_Number() * PhysCond::get_T() / 4184);
      string tmpstr = gconf_curve_dir;
          
      double *maxp = new double[muvecindex.size()];
      // for each muvec determine the maximum population of an form belonging to this site
      for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
       int kk = miter->second;
       maxp[kk] = 0;
       for (int jj=0;jj<nconf;jj++){
        if (scprob[jj][kk]>maxp[kk])
         maxp[kk] = scprob[jj][kk];
       }
      }

      for (int jj=0;jj<nconf;jj++){
       
       string tmpstr1, tmpstr2, tmpstr3;
       stringstream ss1, ss2, ss3;
       ss1 << (*siter).get_chain(); ss1 >> tmpstr1;
       ss2 << (*siter).get_site();  ss2 >> tmpstr2;
       ss3 << jj;                   ss3 >> tmpstr3;
       string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                   + tmpstr1 + "-" + tmpstr2 + "_" 
                                   + tmpstr3 + ".txt";     
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());      
       
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
       int kk = miter->second;      
        for (int i=0;i<nmu;i++){
         double ten_p_prec = pow((double)10,prec[i]);
         double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
         if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
         else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
        }  
        double gconf;  
        if (sprob[jj][kk]>0)
         gconf = -RT * log(scprob[jj][kk]/maxp[kk]);
        else
         gconf = 9999;
        // gconf always > 0 since relative to lowest energy / highest populated
        // gets minus sign due to constant if zero
        if (gconf<=0) gconf *= -1;
        of << setprecision(6) << fixed << gconf << endl;             
       }
   
       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());     
       of.close();
            
      }     
      
      delete [] maxp;

     } // end write conformation free energy curves     
     
     // create conformation intersection curves
     if (para.ichalf == 1){
      
      string tmpstr = chalf_curve_dir;
      for (int ii=0;ii<nmu;ii++){
       int tmpi3 = muvecindex.size();
       // make index of chemical potential combinations without the
       // variable dimension for which we wish to determine the half point
       map<muvec,int> submvi; // index for subspace of muvecs with mu of ligand excluded
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        vector<double> tmpmv;
        for (int oo=0;oo<nmu;oo++){
         if (oo==ii) continue;
         tmpmv.push_back(miter->first.get_mu(oo));
        }
        map<muvec,int>::const_iterator tmiter = submvi.find(tmpmv);
        if (tmiter == submvi.end())
         submvi.insert(map<muvec,int>::value_type(tmpmv,submvi.size()));
       }
       // among all muvec in submvi make an index
       // of mu values for ligand ii
       map<double,int> iimvi;
       for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
        double tmpd = 0;
        for (int oo=0;oo<nmu;oo++){
         if (oo==ii) tmpd = miter->first.get_mu(oo);
        }
        map<double,int>::const_iterator tmi = iimvi.find(tmpd);
        if (tmi==iimvi.end()) iimvi.insert(map<double,int>::value_type(tmpd,iimvi.size()));
       }
              
       // allocate memory for subset of summed up probabilities
       // subprob[inst][submvi][iimvi], initialize to zero
       double ***subsprob;
       try{
        int tmpi1 = nconf;
        int tmpi2 = submvi.size();
        int tmpi3 = iimvi.size();
        subsprob  = new double**[tmpi1];
        for (int hh=0;hh<tmpi1;hh++){
         subsprob[hh] = new double*[tmpi2];
         for (int gg=0;gg<tmpi2;gg++){
          subsprob[hh][gg] = new double[tmpi3];
          // zero out subsprob
          for (int ff=0;ff<tmpi3;ff++){
           subsprob[hh][gg][ff] = 0;
          }
         }
        }
       } catch(bad_alloc){
         cerr << "Error in make_curves: Insufficent memory for tensor subsprob!" << endl;
         exit(0);
       };
       
       // fill subsprob with actual values
       for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
        int gg = smiter->second;
        for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
         int ff = iimiter->second;
         vector<double> tmpmv;
         int pp = 0;
         for (int oo=0;oo<nmu;oo++){
          if (oo==ii){tmpmv.push_back(iimiter->first); pp--;}
          else tmpmv.push_back(smiter->first.get_mu(pp));
          pp++;
         }
         map<muvec,int>::const_iterator miter = muvecindex.find(muvec(tmpmv));
         // grid does not always contain all possible combinations of potential values
         if (miter==muvecindex.end()) continue;
         int kk = miter->second;
         for (int hh=0;hh<nconf;hh++){
          subsprob[hh][gg][ff] = scprob[hh][kk];
         }
        }
       }
       
       // for each point submuveci and each combination jj,jj2 of forms
       // interpolate the mu for ligand type ii at which prob(jj*ii) = prob(jj2*ii)       
       string tmpstr5;
       stringstream ss5;
       ss5 << ii;  ss5 >> tmpstr5;
       for (int jj=0;jj<nconf;jj++){
        string tmpstr1, tmpstr2, tmpstr3, tmpstr4;
        stringstream ss1, ss2, ss3, ss4;
        ss1 << (*siter).get_chain();    ss1 >> tmpstr1;
        ss2 << (*siter).get_site();     ss2 >> tmpstr2;

        for (int jj2=jj+1;jj2<nconf;jj2++){
         ss4 << jj2; ss4 >> tmpstr4;

         string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-"
                                      + tmpstr1 + "-" + tmpstr2 + "_" 
                                      + tmpstr3 + "-" + tmpstr4
                                      + "_" + tmpstr5 + ".txt";      
         std::ofstream of(fn_str.c_str());
         if (!of.good())
          ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
         
         for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
         // multiple intersections or none at all could occur
          vector<double> isp;
          int gg = smiter->second;
          // traverse chemical potential belonging to ligand ii, 
          // search for pairs of successive points in which p1-p2 changes sign
          for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
           map<double,int>::const_iterator iimiter2 = iimiter; iimiter2++;
           if (iimiter2==iimvi.end()) break;
           int ff  = iimiter->second;
           int ff2 = iimiter2->second;
           // do not accept intersections in insignificant probability < 1%
           if ((subsprob[jj][gg][ff]<1.0E-02 && subsprob[jj][gg][ff2]<1.0E-02) || 
               (subsprob[jj2][gg][ff]<1.0E-02 && subsprob[jj2][gg][ff2]<1.0E-02<1.0E-02)){
            continue;
           }
           if ((subsprob[jj][gg][ff]>=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]<=subsprob[jj2][gg][ff2]) ||
               (subsprob[jj][gg][ff]<=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]>=subsprob[jj2][gg][ff2])){
            // calculate mu at intersection point assuming an intersection of two lines through
            // the points (iimiter->first, corresp sprob(jj));  (iimiter2->first, corresp. sprob(jj)) and
            // the points (iimiter->first, corresp sprob(jj2)); (iimiter2->first, corresp. sprob(jj2))
            double tmpisp  = (subsprob[jj][gg][ff] - subsprob[jj2][gg][ff]) * (iimiter2->first - iimiter->first); 
                   tmpisp /= subsprob[jj2][gg][ff2] - subsprob[jj][gg][ff2]
                           + subsprob[jj][gg][ff]   - subsprob[jj2][gg][ff];
                   tmpisp  += iimiter->first;
            isp.push_back(tmpisp);
           }
          }
          if (isp.size()!=0){
           // write the intersections to file
           // first chemical potentials          
           int j = 0;
           for (int i=0;i<nmu;i++){
            if (i==ii) continue;
            double ten_p_prec = pow((double)10,prec[i]);
            double val = round(ten_p_prec * smiter->first.get_mu(j) * conv[i]) / ten_p_prec;
            // avoid minus sign before zero potential
            if (val == 0) val = 0; 
            if (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
            else              of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
            j++;           
           }
           // then all intersections
           for (int vi=0;vi<isp.size();vi++){
            double ten_p_prec = pow((double)10,prec[ii]+1);
            double val = round(ten_p_prec * isp[vi] * conv[ii]) / ten_p_prec;            
            if (prec[ii] != 0) of << setprecision(prec[ii]+1) << fixed <<   showpoint << val;
            else               of << setprecision(prec[ii]+1) << fixed << noshowpoint << val;
            if (vi<(isp.size()-1)){of << " ";}
           }
           of << endl;
          }
         }
         
         if (!of.good())
          ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
         of.close();         
        }
       } // end intersection/half-points
              
       // deallocate memory for summed up probabilities
       int tmpi1 = nconf;
       int tmpi2 = submvi.size();
       for (int hh=0;hh<tmpi1;hh++){
        for (int gg=0;gg<tmpi2;gg++){
         delete [] subsprob[hh][gg];
        }
        delete [] subsprob[hh];
       }        
       delete [] subsprob;

      }  
      
     } // end conformation intersection points / half points      
    
     // deallocate sfprob
     if (para.ifprob == 1 || para.igform == 1|| para.ifhalf == 1 || para.ifmaxp == 1){      
      int tmpi1 = formindex.size();
      for (int ii=0;ii<tmpi1;ii++)
       delete [] sfprob[ii];
      delete [] sfprob;
     }
     
     // deallocate scprob
     if (para.icprob == 1 || para.igconf == 1|| para.ichalf == 1 || para.icmaxp == 1){      
      int tmpi1 = nconf;
      for (int ii=0;ii<tmpi1;ii++)
       delete [] scprob[ii];
      delete [] scprob;
     }
     
            
    } // end sites loop
   
   } // end form / site conformation curves

   // deallocate array sprob
   tmpi1 = instindex.size();
   for (int ii=0;ii<tmpi1;ii++){
    delete [] sprob[ii];
   }
   delete [] sprob;
   

  } // end instance curves

  ///--------------------------------------------///

  // make global conformational curves
  if (confindex.size()>1 && (para.icprob == 1 || para.igconf == 1|| para.ichalf == 1)){

   blab2 << "writing curves for global conformations\n";
   
   // allocate memory for summed up probabilities
   double **cprob;
   cprob = new double*[confindex.size()];
   for (int i=0; i<confindex.size();i++)
    cprob[i] = new double[muvecindex.size()]; 
   
   // init cprob to zero
   for (int i=0; i<confindex.size();i++){
    for (int j=0; j<muvecindex.size();j++){
     cprob[i][j] = 0;
    }
   }
   
   // sum up probabilities of global conformations
   // must be equal for all site so we simply choose the first
   set<SiteID>::const_iterator siter = siteset.begin();
   for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin(); iiter!=instindex.end(); iiter++){
    // search instances belonging to this site
    if ((*siter).get_sitename() == iiter->first.get_sitename() && (*siter).get_chain() == iiter->first.get_chain() && (*siter).get_site() == iiter->first.get_site()){
     int j = iiter->second.id;
     for (int i=0; i<confindex.size();i++){
      for (int k=0; k<muvecindex.size();k++){
       cprob[i][k] += prob[i][j][k];
      }
     }
    }
   }

   // write global conformational probability curves   
   if (para.icprob == 1){

    blab3 << "writing global conformational probability curves\n";
    
    string tmpstr = cprob_curve_dir;
    
    for (map<string,int>::const_iterator citer = confindex.begin(); citer != confindex.end(); citer++){    
     int ii = citer->second;
     string fn_str = tmpstr + "/" + citer->first + ".txt";
     std::ofstream of(fn_str.c_str());
     if (!of.good())
      ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());

     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      int jj = miter->second;
      for (int i=0;i<nmu;i++){
       double ten_p_prec = pow((double)10,prec[i]);
       double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
       if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
       else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
      }  
      of << setprecision(6) << scientific << cprob[ii][jj] << endl; 
     }
     
     if (!of.good())
      ::error("Error in make_curves: failed to write file, ", fn_str.c_str());
     
     of.close();
       
    }  
   
   } // end global conformational probability curves

   // write maximum populated global conformation curves
   if (para.icmaxp == 1){

    blab3 << "writing maximum populated global conformational curves\n";
    
    string tmpstr = cmaxp_curve_dir;
    string fn_str = tmpstr + "/" + "global_conf.txt";
    std::ofstream of(fn_str.c_str());
    if (!of.good())
     ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
    
    for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
     int jj = miter->second;
     // search for global conformation with maximum population
     int cc = 0;
     double maxpop = 0;
     for (map<string,int>::const_iterator citer = confindex.begin(); citer != confindex.end(); citer++){    
      int ii = citer->second;
      if (cprob[ii][jj]>maxpop){
       cc = citer->second;
       maxpop = cprob[ii][jj];
      }
     } 
     for (int i=0;i<nmu;i++){
      double ten_p_prec = pow((double)10,prec[i]);
      double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec; 
      if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
      else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
     }  
     of << cc << endl; 
     
            
    }  
    
    if (!of.good())
     ::error("Error in make_curves: failed to write file, ", fn_str.c_str());
    
    of.close();
   
   } // end maximum populated global conformation curves
      
   // write global conformational free energy curves   
   if (para.igconf == 1){

    blab3 << "writing global conformational free energy curves\n";
    
    string tmpstr = gconf_curve_dir;
    const double RT(PhysCond::get_Boltzmann_constant() * PhysCond::get_Avogadro_Number() * PhysCond::get_T() / 4184);
    for (map<string,int>::const_iterator citer = confindex.begin(); citer != confindex.end(); citer++){    
     int ii = citer->second;
     string fn_str = tmpstr + "/" + citer->first + ".txt";
     std::ofstream of(fn_str.c_str());
     if (!of.good())
      ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());

     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      int jj = miter->second;
      // find largest population
      double maxcprob = 0;
      for (map<string,int>::const_iterator citer2 = confindex.begin(); citer2 != confindex.end(); citer2++){    
       int ii2 = citer2->second;
       if (cprob[ii2][jj] > maxcprob) maxcprob = cprob[ii2][jj];
      }
      for (int i=0;i<nmu;i++){
       double ten_p_prec = pow((double)10,prec[i]);
       double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec;
       // avoid minus sign before zero potential
       if (val == 0) val = 0; 
       if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
       else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";
      }
      double gconf;  
      if (cprob[ii][jj]>0)
       gconf = -RT * log(cprob[ii][jj]/maxcprob);
      else
       gconf = 9999;
      // gconf always > 0 since relative to lowest energy / highest populated
      // gets minus sign due to constant if zero
      if (gconf<=0) gconf *= -1;
      of << setprecision(6) << fixed << gconf << endl; 
     }
     
     if (!of.good())
      ::error("Error in make_curves: failed to write file, ", fn_str.c_str());
     
     of.close();
     
    } // end global conformational free energy curves
     
   } // end write global conformational free energy curves
   
   // create intersection curves varying one potential with the others fixed
   if (para.ichalf == 1){

    blab3 << "writing global conformational half/intersection curves\n";
    
    string tmpstr = chalf_curve_dir;
    for (int ii=0;ii<nmu;ii++){
     int tmpi3 = muvecindex.size();
     // make index of chemical potential combinations without the
     // variable dimension for which we wish to determine the half point
     map<muvec,int> submvi; // index for subspace of muvecs with mu of ligand excluded
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      vector<double> tmpmv;
      for (int oo=0;oo<nmu;oo++){
       if (oo==ii) continue;
       tmpmv.push_back(miter->first.get_mu(oo));
      }
      map<muvec,int>::const_iterator tmiter = submvi.find(tmpmv);
      if (tmiter == submvi.end())
       submvi.insert(map<muvec,int>::value_type(tmpmv,submvi.size()));
     }
     // among all muvec in submvi make an index
     // of mu values for ligand ii
     map<double,int> iimvi;
     for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
      double tmpd = 0;
      for (int oo=0;oo<nmu;oo++){
       if (oo==ii) tmpd = miter->first.get_mu(oo);
      }
      map<double,int>::const_iterator tmi = iimvi.find(tmpd);
      if (tmi==iimvi.end()) iimvi.insert(map<double,int>::value_type(tmpd,iimvi.size()));
     }
            
     // allocate memory for subset of summed up probabilities
     // subprob[nbound][submvi][iimvi], initialize to zero
     double ***subcprob;
     try{
      int tmpi1 = confindex.size();
      int tmpi2 = submvi.size();
      int tmpi3 = iimvi.size();
      subcprob  = new double**[tmpi1];
      for (int hh=0;hh<tmpi1;hh++){
       subcprob[hh] = new double*[tmpi2];
       for (int gg=0;gg<tmpi2;gg++){
        subcprob[hh][gg] = new double[tmpi3];
        // zero out subcprob
        for (int ff=0;ff<tmpi3;ff++){
         subcprob[hh][gg][ff] = 0;
        }
       }
      }
     } catch(bad_alloc){
       cerr << "Error in make_curves: Insufficent memory for tensor subcprob!" << endl;
       exit(0);
     };
     
     // fill subcprob with actual values
     for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
      int gg = smiter->second;
      for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
       int ff = iimiter->second;
       vector<double> tmpmv;
       int pp = 0;
       for (int oo=0;oo<nmu;oo++){
        if (oo==ii){tmpmv.push_back(iimiter->first); pp--;}
        else tmpmv.push_back(smiter->first.get_mu(pp));
        pp++;
       }
       map<muvec,int>::const_iterator miter = muvecindex.find(muvec(tmpmv));
       // grid does not always contain all possible combinations of potential values
       if (miter==muvecindex.end()) continue;

       int kk = miter->second;
       for (int hh=0;hh<confindex.size();hh++){
        subcprob[hh][gg][ff] = cprob[hh][kk];
       }
      }
     }
     
     // for each point submuveci and each combination jj!=jj2 of global conformations 
     // interpolate the mu for ligand type ii at which prob(jj*ii) = prob(jj2*ii)       
     string tmpstr3;
     stringstream ss1;
     ss1 << ii;  ss1 >> tmpstr3;
     for (map<string,int>::const_iterator citer=confindex.begin();citer!=confindex.end();citer++){
      int jj = citer->second;
      string tmpstr1(citer->first);
      map<string,int>::const_iterator citer2=citer;
      for (citer2++;citer2!=confindex.end();citer2++){
       int jj2 = citer2->second;
       string tmpstr2(citer2->first);
       string fn_str = tmpstr + "/" + tmpstr1 + "-" + tmpstr2 + "_" + tmpstr3 + ".txt";
       std::ofstream of(fn_str.c_str());
       if (!of.good())
        ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
       
       for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
       // multiple intersections or none at all could occur
        vector<double> isp;
        int gg = smiter->second;
        // traverse chemical potential belonging to ligand ii, 
        // search for pairs of successive points in which p1-p2 changes sign
        for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
         map<double,int>::const_iterator iimiter2 = iimiter; iimiter2++;
         if (iimiter2==iimvi.end()) break;
         int ff  = iimiter->second;
         int ff2 = iimiter2->second;
         // do not accept intersections in insignificant probability < 1%
         if ((subcprob[jj][gg][ff]<1.0E-02 && subcprob[jj][gg][ff2]<1.0E-02) || 
             (subcprob[jj2][gg][ff]<1.0E-02 && subcprob[jj2][gg][ff2]<1.0E-02<1.0E-02)){
          continue;
         }
         if ((subcprob[jj][gg][ff]>=subcprob[jj2][gg][ff] && subcprob[jj][gg][ff2]<=subcprob[jj2][gg][ff2]) ||
             (subcprob[jj][gg][ff]<=subcprob[jj2][gg][ff] && subcprob[jj][gg][ff2]>=subcprob[jj2][gg][ff2])){
          // calculate mu at intersection point assuming an intersection of two lines through
          // the points (iimiter->first, corresp sprob(jj));  (iimiter2->first, corresp. sprob(jj)) and
          // the points (iimiter->first, corresp sprob(jj2)); (iimiter2->first, corresp. sprob(jj2))
          double tmpisp  = (subcprob[jj][gg][ff] - subcprob[jj2][gg][ff]) * (iimiter2->first - iimiter->first); 
                 tmpisp /= subcprob[jj2][gg][ff2] - subcprob[jj][gg][ff2]
                         + subcprob[jj][gg][ff]   - subcprob[jj2][gg][ff];
                 tmpisp  += iimiter->first;
          isp.push_back(tmpisp);
         }
        }
        if (isp.size()!=0){
         // write the intersections to file
         // first chemical potentials          
         int j = 0;
         for (int i=0;i<nmu;i++){
          if (i==ii) continue;
          double ten_p_prec = pow((double)10,prec[i]);
          double val = round(ten_p_prec * smiter->first.get_mu(j) * conv[i]) / ten_p_prec;
          // avoid minus sign before zero potential
          if (val == 0) val = 0; 
          if (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
          else              of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
          j++;           
         }
         // then all intersections
         for (int vi=0;vi<isp.size();vi++){
          double ten_p_prec = pow((double)10,prec[ii]+1);
          double val = round(ten_p_prec * isp[vi] * conv[ii]) / ten_p_prec;            
          if (prec[ii] != 0) of << setprecision(prec[ii]+1) << fixed <<   showpoint << val;
          else               of << setprecision(prec[ii]+1) << fixed << noshowpoint << val;
          if (vi<(isp.size()-1)){of << " ";}
         }
         of << endl;
        }
       }
       
       if (!of.good())
        ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
       of.close();         
      }
     } // end intersection/half-points
            
     // deallocate memory for summed up probabilities
     int tmpi1 = confindex.size();
     int tmpi2 = submvi.size();
     for (int hh=0;hh<tmpi1;hh++){
      for (int gg=0;gg<tmpi2;gg++){
       delete [] subcprob[hh][gg];
      }
      delete [] subcprob[hh];
     }        
     delete [] subcprob;

    }  
    
   } // end global conformational intersection points / half points
         
   // deallocate memory for summed up probabilities
   for (int i=0; i<confindex.size();i++)
    delete [] cprob[i];
   delete [] cprob; 

  } // end global conformational curves
  
  ///--------------------------------------------///

  if (para.iprob == 1 || para.ibprob == 1 || para.icprob == 1 || para.ibmean == 1 || para.ibmaxp == 1 || para.icmaxp == 1 || para.ibhalf == 1 || para.ichalf == 1 || para.igbind == 1 || para.igconf == 1){
   
   // make site curves 
   
   blab2 << "writing ligand binding curves for sites\n";
   
   int *maxbound = new int[nmu];
   int *minbound = new int[nmu];
   int *doesbindthislig = new int[nmu];
   
   for (set<SiteID>::const_iterator siter=siteset.begin(); siter!=siteset.end(); siter++){
    
    // determine minimum and maximum number of bound ligands of each type
    for (int ii = 0; ii<nmu; ii++){
      minbound[ii] = 9999;
      maxbound[ii] = 0;
      doesbindthislig[ii] = 0;
    }   
    for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin(); iiter!=instindex.end(); iiter++){
     // search instances belonging to this site
     if ((*siter).get_sitename() == iiter->first.get_sitename() && (*siter).get_chain() == iiter->first.get_chain() && (*siter).get_site() == iiter->first.get_site()){
      for (int ii = 0; ii<nmu; ii++){
       if (iiter->second.ligands[ii] < minbound[ii]) minbound[ii] = iiter->second.ligands[ii];
       if (iiter->second.ligands[ii] > maxbound[ii]) maxbound[ii] = iiter->second.ligands[ii];
      }
     }  
    }
    bool doesbind = false;
    for (int ii = 0; ii<nmu; ii++){
     if (maxbound[ii] - minbound[ii] > 0){
      doesbind = true;
      doesbindthislig[ii] = 1;
     }
    }
    
    // sum up probability for each type and number of bound ligands of that type
    if (doesbind && nmu > 0 && (para.ibprob == 1 || para.ibhalf == 1 || para.igbind == 1)){
     // allocate memory for summed up probabilities
     double ***sprob;
     try{
      sprob = new double**[nmu];  
      for (int ii=0;ii<nmu;ii++){
       int tmpi2 = maxbound[ii] - minbound[ii] + 1;
       if (tmpi2 > 0){
        sprob[ii]  = new double*[tmpi2];
        sprob[ii] -= minbound[ii]; // move starting index
        int tmpi3 = muvecindex.size();
        for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
         sprob[ii][jj] = new double[tmpi3];
        }  
       }
      }
     } catch(bad_alloc){
       cerr << "Error in make_curves: Insufficent memory for tensor sprob!" << endl;
       exit(0);
     };
     
     // zero out sprob    
     for (int ii=0;ii<nmu;ii++){
      int tmpi2 = maxbound[ii] - minbound[ii] + 1;
      if (tmpi2 > 0){
       int tmpi3 = muvecindex.size();
       for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
        for (int kk=0;kk<tmpi3;kk++){
         sprob[ii][jj][kk] = 0;
        }
       }  
      }
     }       
  
     // sum up the probabilities
     int tmpi1 = confindex.size();
     int tmpi3 = muvecindex.size();
     for (int ii=0;ii<nmu;ii++){
      int tmpi2 = maxbound[ii] - minbound[ii] + 1;
      if (tmpi2 > 1){
       for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
        for (map<InstanceID,InstData>::const_iterator iiter=instindex.begin(); iiter!=instindex.end(); iiter++){
         // does instance belong to this site
         if ((*siter).get_sitename() == iiter->first.get_sitename() && (*siter).get_chain() == iiter->first.get_chain() && (*siter).get_site() == iiter->first.get_site()){
          // correct number of ligands bound?
          if (iiter->second.ligands[ii] == jj){
           // sum up over all instances with the correct number of ligands ...
           int b = iiter->second.id; 
           // ... over all conformations ...
           for (int a=0;a<tmpi1;a++){
            // ... and for each muvec
            for (int kk=0;kk<tmpi3;kk++){
             sprob[ii][jj][kk] += prob[a][b][kk];
            }        
           }
          }
         }
        }
       }
      }
     }   
     
     // create ligand binding probability curves
     if (para.ibprob == 1){

      string tmpstr = bprob_curve_dir;
      for (int ii=0;ii<nmu;ii++){
       if (doesbindthislig[ii] == 0) continue;
       int tmpi2 = maxbound[ii] - minbound[ii] + 1;
       if (tmpi2 > 0){
        int tmpi3 = muvecindex.size();
        for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
         string tmpstr1, tmpstr2, tmpstr4, tmpstr5;
         stringstream ss1, ss2, ss4, ss5;
         ss1 << ii;  ss1 >> tmpstr1;
         ss2 << jj;  ss2 >> tmpstr2;
         ss4 << (*siter).get_chain(); ss4 >> tmpstr4;
         ss5 << (*siter).get_site();  ss5 >> tmpstr5;
         string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-" 
                                      + tmpstr4 + "-"
                                      + tmpstr5 + "_"
                                      + tmpstr1 + "_"
                                      + tmpstr2
                                      + ".txt";
         std::ofstream of(fn_str.c_str());
         if (!of.good())
          ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
         for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
          int kk = miter->second;
          for (int i=0;i<nmu;i++){
           double ten_p_prec = pow((double)10,prec[i]);
           double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec;
           // avoid minus sign before zero potential
           if (val == 0) val = 0; 
           if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
           else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
          }
          of << setprecision(6) << scientific << sprob[ii][jj][kk] << endl;
         }
         if (!of.good())
          ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
         of.close();         
         
        }  
       }
      }  
      
     } // end ligand binding probability curves

     // create binding energy curves for binding steps n to n+1
     if (para.igbind == 1){
      
      string tmpstr = gbind_curve_dir;
      const double RT(PhysCond::get_Boltzmann_constant() * PhysCond::get_Avogadro_Number() * PhysCond::get_T() / 4184);
      for (int ii=0;ii<nmu;ii++){
       if (doesbindthislig[ii] == 0) continue;
       int tmpi2 = maxbound[ii] - minbound[ii] + 1;
       if (tmpi2 > 0){
        int tmpi3 = muvecindex.size();
        for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
         for (int jj2=jj+1;jj2<=maxbound[ii];jj2++){
          string tmpstr1, tmpstr2, tmpstr3, tmpstr4, tmpstr5;
          stringstream ss1, ss2, ss3, ss4, ss5;
          ss1 << ii;  ss1 >> tmpstr1;
          ss2 << jj;  ss2 >> tmpstr2;
          ss3 << jj2; ss3 >> tmpstr3;
          ss4 << (*siter).get_chain(); ss4 >> tmpstr4;
          ss5 << (*siter).get_site();  ss5 >> tmpstr5;
          string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-" 
                                       + tmpstr4 + "-"
                                       + tmpstr5 + "_"
                                       + tmpstr1 + "_"
                                       + tmpstr2 + "-" + tmpstr3
                                       + ".txt";
          std::ofstream of(fn_str.c_str());
          if (!of.good())
           ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
          for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
           int kk = miter->second;
           for (int i=0;i<nmu;i++){
            double ten_p_prec = pow((double)10,prec[i]);
            double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec;
            // avoid minus sign before zero potential
            if (val == 0) val = 0; 
            if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
            else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
           }
           double prob1 = sprob[ii][jj][kk];
           double prob2 = sprob[ii][jj2][kk];
           double gbind;
           if (prob1 != 0 && prob2 != 0)
            gbind = -RT * log(prob2/prob1);
           else if (prob1 == 0 && prob2 == 0)
            gbind = 0;
           else if (prob1 == 0 && prob2 != 0)
            gbind = -9999;
           else if (prob1 != 0 && prob2 == 0)
            gbind =  9999;
           // gets minus sign due to constant if zero 
           if (gbind==0) gbind *= -1;
           of << setprecision(6) << fixed << gbind << endl;
          }
          if (!of.good())
           ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
          of.close();         
         }
        }  
       }
      }  
      
     } // end ligand binding free energy curves

     
     // create half curves varying one potential with the others fixed
     // half points of transitions defined by intersections of curves
     // for binding n / n + 1 ligands should coincide with inflection 
     // points of these curves and that of the binding isotherm for
     // standard titration curves behaviour, the physical significance
     // of the intersection points lies in the equal free energy of
     // the states in question at these points in contrast to true
     // half points which have no general physical meaning
     
     if (para.ibhalf == 1){
      
      string tmpstr = bhalf_curve_dir;
      for (int ii=0;ii<nmu;ii++){
       if (doesbindthislig[ii] == 0) continue;

       int tmpi2 = maxbound[ii] - minbound[ii] + 1;
       if (tmpi2 > 1){
        int tmpi3 = muvecindex.size();
        // make index of chemical potential combinations without the
        // variable dimension for which we wish to determine the half point
        map<muvec,int> submvi; // index for subspace of muvecs with mu of ligand excluded
        for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
         vector<double> tmpmv;
         for (int oo=0;oo<nmu;oo++){
          if (oo==ii) continue;
          tmpmv.push_back(miter->first.get_mu(oo));
         }
         map<muvec,int>::const_iterator tmiter = submvi.find(tmpmv);
         if (tmiter == submvi.end())
          submvi.insert(map<muvec,int>::value_type(tmpmv,submvi.size()));
        }
               
        // among all muvec in submvi make an index
        // of mu values for ligand ii
        map<double,int> iimvi;
        for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
         double tmpd = 0;
         for (int oo=0;oo<nmu;oo++){
          if (oo==ii) tmpd = miter->first.get_mu(oo);
         }
         map<double,int>::const_iterator tmi = iimvi.find(tmpd);
         if (tmi==iimvi.end()) iimvi.insert(map<double,int>::value_type(tmpd,iimvi.size()));
        }
               
        // allocate memory for subset of summed up probabilities
        // subprob[nbound][submvi][iimvi], initialize to zero
        double ***subsprob;
        try{
         int tmpi1 = maxbound[ii] - minbound[ii] + 1;
         int tmpi2 = submvi.size();
         int tmpi3 = iimvi.size();

         subsprob  = new double**[tmpi1];
         subsprob -= minbound[ii]; // move starting index
         for (int hh=minbound[ii];hh<=maxbound[ii];hh++){
          subsprob[hh] = new double*[tmpi2];
          for (int gg=0;gg<tmpi2;gg++){
           subsprob[hh][gg] = new double[tmpi3];
           // zero out subsprob
           for (int ff=0;ff<tmpi3;ff++){
            subsprob[hh][gg][ff] = 0;
           }
          }
         }
        } catch(bad_alloc){
          cerr << "Error in make_curves: Insufficent memory for tensor subsprob!" << endl;
          exit(0);
        };
        
        // fill subsprob with actual values
        for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
         int gg = smiter->second;

         for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){

          int ff = iimiter->second;
          vector<double> tmpmv;
          int pp = 0;
          for (int oo=0;oo<nmu;oo++){
           if (oo==ii){tmpmv.push_back(iimiter->first); pp--;}
           else tmpmv.push_back(smiter->first.get_mu(pp));
           pp++;
          }

          map<muvec,int>::const_iterator miter = muvecindex.find(muvec(tmpmv));

          // grid does not always contain all possible combinations of potential values
          if (miter==muvecindex.end()) continue;
          
          int kk = miter->second;
          for (int hh=minbound[ii];hh<=maxbound[ii];hh++){

           subsprob[hh][gg][ff] = sprob[ii][hh][kk];
           
          }
         }
        }
        
        // for each point submuveci and each combination jj,jj2 of numbers of 
        // bound ligands of type ii interpolate the mu for ligand type ii
        // at which prob(jj*ii) = prob(jj2*ii)       
        for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
         for (int jj2=jj+1;jj2<=maxbound[ii];jj2++){
          string tmpstr1, tmpstr2, tmpstr3, tmpstr4, tmpstr5;
          stringstream ss1, ss2, ss3, ss4, ss5;
          ss1 << ii;  ss1 >> tmpstr1;
          ss2 << jj;  ss2 >> tmpstr2;
          ss3 << jj2; ss3 >> tmpstr3;
          ss4 << (*siter).get_chain(); ss4 >> tmpstr4;
          ss5 << (*siter).get_site();  ss5 >> tmpstr5;
          string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-" 
                                       + tmpstr4 + "-"
                                       + tmpstr5 + "_"
                                       + tmpstr1 + "_"
                                       + tmpstr2 + "-" + tmpstr3
                                       + ".txt";
          std::ofstream of(fn_str.c_str());
          if (!of.good())
           ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
          
          for (map<muvec,int>::const_iterator smiter=submvi.begin();smiter!=submvi.end();smiter++){
          // multiple intersections or none at all could occur
           vector<double> isp;
           int gg = smiter->second;
           // traverse chemical potential belonging to ligand ii, 
           // search for pairs of successive points in which p1-p2 changes sign
           for (map<double,int>::const_iterator iimiter=iimvi.begin();iimiter!=iimvi.end();iimiter++){
            map<double,int>::const_iterator iimiter2 = iimiter; iimiter2++;
            if (iimiter2==iimvi.end()) break;
            int ff  = iimiter->second;
            int ff2 = iimiter2->second;
            // do not accept intersections in insignificant probability < 1%
            if ((subsprob[jj][gg][ff]<1.0E-02 && subsprob[jj][gg][ff2]<1.0E-02) || 
                (subsprob[jj2][gg][ff]<1.0E-02 && subsprob[jj2][gg][ff2]<1.0E-02<1.0E-02)){
             continue;
            }
            if ((subsprob[jj][gg][ff]>=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]<=subsprob[jj2][gg][ff2]) ||
                (subsprob[jj][gg][ff]<=subsprob[jj2][gg][ff] && subsprob[jj][gg][ff2]>=subsprob[jj2][gg][ff2])){
             // calculate mu at intersection point assuming an intersection of two lines through
             // the points (iimiter->first, corresp sprob(jj));  (iimiter2->first, corresp. sprob(jj)) and
             // the points (iimiter->first, corresp sprob(jj2)); (iimiter2->first, corresp. sprob(jj2))
             double tmpisp  = (subsprob[jj][gg][ff] - subsprob[jj2][gg][ff]) * (iimiter2->first - iimiter->first); 
                    tmpisp /= subsprob[jj2][gg][ff2] - subsprob[jj][gg][ff2]
                            + subsprob[jj][gg][ff]   - subsprob[jj2][gg][ff];
                    tmpisp  += iimiter->first;
             isp.push_back(tmpisp);
            }
           }           
           
           if (isp.size()!=0){
            // write the intersections to file
            // first chemical potentials          
            int j = 0;
            for (int i=0;i<nmu;i++){
             if (i==ii) continue;
             double ten_p_prec = pow((double)10,prec[i]);
             double val = round(ten_p_prec * smiter->first.get_mu(j) * conv[i]) / ten_p_prec;
             // avoid minus sign before zero potential
             if (val == 0) val = 0; 
             if (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
             else              of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
             j++;           
            }
            // then all intersections
            for (int vi=0;vi<isp.size();vi++){
             double ten_p_prec = pow((double)10,prec[ii]+1);
             double val = round(ten_p_prec * isp[vi] * conv[ii]) / ten_p_prec;            
             if (prec[ii] != 0) of << setprecision(prec[ii]+1) << fixed <<   showpoint << val;
             else               of << setprecision(prec[ii]+1) << fixed << noshowpoint << val;
             if (vi<(isp.size()-1)){of << " ";}
            }
            of << endl;
           }
          }
          
          if (!of.good())
           ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
          of.close();         
         }
        } // end intersection/half-points
               
        // deallocate memory for summed up probabilities
        for (int hh=minbound[ii];hh<=maxbound[ii];hh++){
         int tmpi2 = submvi.size();
         for (int gg=0;gg<tmpi2;gg++){
          delete [] subsprob[hh][gg];
         }
         delete [] subsprob[hh];
        }        
        subsprob += minbound[ii];
        delete [] subsprob;       
          
       }
      }  
      
     } // end ligand binding intersection points / half points

     // create mean number of bound ligand curves for each ligand type
     if (para.ibmean == 1){
      
      string tmpstr = bmean_curve_dir;
      for (int ii=0;ii<nmu;ii++){
       if (doesbindthislig[ii] == 0) continue;
       int tmpi2 = maxbound[ii] - minbound[ii] + 1;
       if (tmpi2 > 0){
        int tmpi3 = muvecindex.size();
        string tmpstr1, tmpstr4, tmpstr5;
        stringstream ss1, ss4, ss5;
        ss1 << ii;  ss1 >> tmpstr1;
        ss4 << (*siter).get_chain(); ss4 >> tmpstr4;
        ss5 << (*siter).get_site();  ss5 >> tmpstr5;
        string fn_str = tmpstr + "/" + (*siter).get_sitename() + "-" 
                                     + tmpstr4 + "-"
                                     + tmpstr5 + "_"
                                     + tmpstr1 + ".txt";
        std::ofstream of(fn_str.c_str());
        if (!of.good())
         ::error("Error in make_curves: failed to open file for writing, ", fn_str.c_str());
        for (map<muvec,int>::const_iterator miter=muvecindex.begin(); miter!=muvecindex.end(); miter++){
         int kk = miter->second;
         for (int i=0;i<nmu;i++){
          double ten_p_prec = pow((double)10,prec[i]);
          double val = round(ten_p_prec * miter->first.get_mu(i) * conv[i]) / ten_p_prec;
          // avoid minus sign before zero potential
          if (val == 0) val = 0; 
          if    (prec[i] != 0) of << setprecision(prec[i]) << fixed <<   showpoint << val << " ";
          else                 of << setprecision(prec[i]) << fixed << noshowpoint << val << " ";           
         }
         double meanj = 0;
         for (int jj=minbound[ii];jj<=maxbound[ii];jj++)
          meanj += sprob[ii][jj][kk] * jj;
         of << setprecision(2) << fixed << meanj << endl;
        }
        if (!of.good())
         ::error("Error in make_curves: failed to write file, ", fn_str.c_str());      
        of.close();         
       }
      }  
      
     } // end mean number of bound ligands curves     
     
     // deallocate memory for summed up probabilities
     for (int ii=0;ii<nmu;ii++){
      int tmpi2 = maxbound[ii] - minbound[ii] + 1;
      if (tmpi2 > 0){
       int tmpi3 = muvecindex.size();
       for (int jj=minbound[ii];jj<=maxbound[ii];jj++){
        delete [] sprob[ii][jj];
       }
       sprob[ii] += minbound[ii];
       delete [] sprob[ii];
      }
     }        
     delete [] sprob;  
                 
    } // end ligand binding curves
   
   } // end sites

   delete [] maxbound;
   delete [] minbound;
   delete [] doesbindthislig;

  } // end site specific curves  
                      
  blab1 << "make_curves: exiting\n";
   
}
