#define LINELENGTH 500
/* Fundamental Physical Constants CODATA 2006 values */
/* ----------------------------*/
/* # in kcal/mol */
#define h_PLANCK   	       1.5826109662654e-37	/* Planck constant  in [kcal s]		*/
#define h_DIV_2PI   	       2.5204868100000e-38	/* Planck constant  h/2 Pi in [kcal s]	*/
#define k_B		       3.2998336500000e-27 	/* Boltzmann constant  k_B [kcal/K]	*/
#define R_GAS		       1.9871653920000e-03	/* Gas constant [kcal/(K*mol)]		*/
/* ----------------------------*/
#define EPSILON_0	       8.854187817E-12   	/* Permittivity of vacuum [(As)/(Vm)]   */
#define ELEMENTARY_CHARGE      1.602176487E-19          /* Elementary charge e [C] 		*/
#define V_MOL		      22.413996                 /* Molar volume [m3/kmol]   		*/
#define N_A		       6.02214179E+23  		/* Avogadro number  N_A [particles/mol] */
#define SPEED_OF_LIGHT         299792458  		/* Speed of light [m/s]			*/
#define PERMEABILITY   	      12.566370614e-7 		/* Permeability of vacuum [T^2*m^3/J]   */
#define FINE_STRUCTURE_CONST 137.035999679 		/* inverse Fine structure constant      */
#define ELECTRON_REST_MASS     9.10938215e-31   	/* Electron rest mass [kg] 		*/
#define PROTON_REST_MASS       1.672621637e-27 		/* Proton rest mass [kg] 		*/
#define NEUTRON_REST_MASS      1.674927212e-27 		/* Neutron rest mass [kg] 		*/
#define BOHR_MAGNETON 	       9.27400915e-24 		/* Bohr magneton [J/T] 			*/
#define NUCLEAR_MAGNETON       5.05078324e-27 		/* nuclear magneton [J/T] 		*/
#define FREE_ELECTRON_G_FACTOR 2.0023193043622		/* negative Free electron g factor      */
#define AMU		       1.660538782 		/* unified atomic mass unit [kg]	*/ 
#define FARADAY		       9.64853399e4 		/* Faraday constant [C/mol] 		*/

#include "MEAD_e/PhysCond.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/SiteID.h"
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/ChargeDist.h"
#include "MEAD_e/AtomChargeSet.h"
#include "MEAD_e/muvec.h"
#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <vector>
#include <map>
#include <set>
#include <list>
#include <string>
#include <cmath>

using std::vector;
using std::list;
using std::map;
using std::set;
using std::pair;
using std::string;
using namespace::std;
using namespace boost;
namespace fs = boost::filesystem;
using namespace fs;

///////////////////////////////////////// structure declaration //////////////////////////////////////////////////

typedef struct {std::string molname;      // molecule name
                std::string gcem_dir;     // name of directory with (gcem|qmpb) input structures, subdirs with the
                                          // conformation names are expected if there are multiple conformations 
                std::string curve_dir;    // name of main directory for curves and trajectories
                std::string gmct_out;     // name of gmct output file
                std::vector<double> conv; // conversion factors for chemical potentials
                std::vector<int> prec;    // output decimal digits for chemical potentials
		double   temp;            // temperature in K
                long int seed;            // seed for random number generator
                int      iprob;           // flag for writing nmu-D probabilty curves for each instance 0 - no, 1 - yes
                int      ifprob;          // flag for writing nmu-D probabilty curves for each form 0 - no, 1 - yes
                int      ibprob;          // flag for writing nmu-D probabilty curves for each number of bound ligands 0 - no, 1 - yes
                int      icprob;          // flag for writing nD conformational probability curves 0 - no, 1 - yes
                int      ibmean;          // flag for writing nmu-D curves with number of mean number of bound ligands 0 - no, 1 - yes
                int      imaxp;           // flag for writing nmu-D curves with id of most highly populated instance 0 - no, 1 - yes
                int      ifmaxp;          // flag for writing nmu-D curves with number of most highly populated form 0 - no, 1 - yes
                int      ibmaxp;          // flag for writing nmu-D curves with number of most highly populated number of bound ligands 0 - no, 1 - yes
                int      icmaxp;          // flag for writing nmu-D curves with id of most highly populated conformation 0 - no, 1 - yes
                int      ihalf;           // flag for writing nmu-D midpoint/intersection curves between instances 0 - no, 1 - yes
                int      ifhalf;          // flag for writing nmu-D midpoint/intersection curves between forms 0 - no, 1 - yes
                int      ibhalf;          // flag for writing (nmu-1)-D binding midpoint curves 0 - no, 1 - yes
                int      ichalf;          // flag for writing nmu-D conformational midpoint curves 0 - no, 1 - yes
                int      iginst;          // flag for writing nmu-D free energy curves for instances 0 - no, 1 - yes
                int      igform;          // flag for writing nmu-D free energy curves for forms 0 - no, 1 - yes
                int      igbind;          // flag for writing nmu-D binding free energy curves 0 - no, 1 - yes
                int      igconf;          // flag for writing nmu-D conformational free energy curves 0 - no, 1 - yes
                int      itraj;           // flag for writing structure trajectories 0 - no, 1 - yes
                int      ipmf;            // flag for writing the curves as nmu+1-D curves vs mu1-mu2 + membpot, mu1 is expected to be pHin mu2 pHout 
                int      has_vmemb;       // flag indicating wheter a none-zero membrane potential occurs in the gmct output
               }CurvePara;

typedef struct {std::vector<short int> ligands; // bound ligands in order of chemical potentials    
                int id;                         // index
               }InstData;
typedef struct {double umu; // chemical potential closest above half point
                double lmu; // chemical potential closest below half point
                double u1p; // occupation probability of binding form n @ umu
                double l1p; // occupation probability of binding form n @ lmu
                double u2p; // occupation probability of binding form n+1 @ umu
                double l2p; // occupation probability of binding form n+1 @ lmu
               }IntersPoint;


///////////////////////////////////////// function declaration ///////////////////////////////////////////////////

void read_setup(CurvePara& para);
double*** read_gmct_out(std::set<SiteID>&,std::map<std::string,int>&,std::map<InstanceID,InstData>&,std::map<muvec,int>&,CurvePara&);
//double*** read_gmct_out_noregex(const std::string&,std::set<SiteID>&,std::map<std::string,int>&,std::map<InstanceID,InstData>&,std::map<muvec,int>&);
void make_curves(const std::set<SiteID>&,const std::map<std::string,int>&,const std::map<InstanceID,InstData>&,const std::map<muvec,int>&,const CurvePara&,double***);
void make_traj(const std::set<SiteID>&,const std::map<std::string,int>&,const std::map<InstanceID,InstData>&,const std::map<muvec,int>&,const CurvePara&,double***);
void deallocate_prob(const map<string,int>&,const map<InstanceID,InstData>&,const map<muvec,int>&,double***);
