/* Do a Simulated Annealing / Metropolis Monte Carlo Docking of two molecules
   ligand can be treated flexible, different possibilities to calculate energie
   of complex formation
     
   mcdock         -chmtop filename -chmpar filename \
                  -iflex 0/1 0 rigid/flexible ligand \
                  -imin run MC Simulated Annealing minimizatuion of lowest energy structures \
                  -ivtraj  write vector trajectory \
                  -nvwrite write vector trajectory after each N steps \
                  -ivtrajm write minimization vector trajectory \
                  -nvwritem write minimization vector trajectory after each N steps \
                  -vtx x -vty y -vtz z components of vector for trajectory \
                  -otx x -oty y -otz z initial origin of vector for trajectory \                  
                  -itraja  write trajectories for all replicas \
                  -itraj   write trajectory of MC simulation for Tmin replica \
                  -itrajm  write trajectory of MC minimization \
                  -nwrite  write trajectory each N steps \
                  -nwritem write minimization trajectory each N steps \
                  -seed random number seed \
                  -runs N independent runs \                                    
                  -nmcequi N equilibration replica exchange MC scans \
                  -nmcfull N replica exchange MC scans \
                  -nreps   N replicas between Tmin and Tmax with exponential T spacing \ 
                  -nstruct N minimum energy structures to keep for minimization \ 
                  -nsamin N SA steps during minimization \
                  -nmcmin N MC steps per SA step during minimization \
                  -nstructs N structures to keep for refinement \
                  -nrefsteps N MC steps for refinement \
                  -Tmin float -Tmax float T range for replica exchange MC, Tmin is start for SA \
                  -epsin1 eps -epsin2 eps -epsin3 eps \                  
                  -epscore eps -epshead eps -epsext eps \
                  -eps1set name -eps2set name -eps3set name \
                  -channelset name -elycavset name\
                  -membz lower_core_boundary upper_core_boundary \
                         lower_head_boundary upper_head_boundary \
                  -membhole radius xcoord ycoord \
                  -pkunits \
                  -fpt filename -pf -epshomo \
                  -write_eps -write_ely -write_pot \
                  -count gridpoints(x) [gridpoints(x) gridpoints(y)] \ 
                  -space spacing -x center.x -y center.y -z center.z
                  ligname protname
   
   commandline options:
   
   mandatory      -chmtop string -chmpar string
                  -epsin1|2|3 float -eps1|2|3set string               
   
   optional       -Tstart float -Tend float \
                  -runs int -sasteps int -mcsteps int \
                  -structs int nrefsteps int \
                  -epsin1 float -epsin2 float -epsin3 float \
                  -epscore float -epshead float -epsext float \
                  -eps1set string -eps2set string eps3set string \
                  -membz float float float float \
                  -elycavset string -channelset string \
                  -membhole float float float \          
                  -fpt string -pf string -epshomo float \
                  -write_eps -write_ely -write_pot \
                  -count int [int int] -space float \
                  -x float -y float -z float \-pkunits \
                  
   others (opt.)  -solrad float -sterln float -T float \
                  -kBolt float -conconv float -econv float \
                  -bohr_radius float -proton_charge float \
                  -blab1 -blab2 -blab3 -converge_oldway                 
                  
   The program uses a Simulated Annealing / Metropolis Monte Carlo 
   approach to dock a two molecules termed ligand and receptor. 
   The receptor molecule is kept fixed at its input coordinates
   while the ligand is moved around it. Several trajectories are
   run with random starting positions for the ligand. The 
   simulation starts at high temperature. After a certain number
   of moves the temperature is lowered. The simulation proceeds
   till the temperature reaches zero K. After some more moves 
   to minimize the complex structure, the final interaction 
   energie is computed and the final ligand coordinates are 
   written to a PQR file.
   
   The electrostatic potential of the receptor is precomputed in 
   absence of the ligand molecule, i.e., the effect of desolvation 
   upon encounter of receptor and ligand is neglected. The apolar 
   part of the interaction energy is represented using the CHARMM 
   Lennard-Jones potential. The Lennard-Jones term is calculated 
   grid-free after each trial move but a distance cutoff is used 
   to speed up the calculation (default 8 A).  
 
   After each trial move the interaction energy of ligand and receptor 
   is calculated. 
   
   To calculated the electrostatic part of the interaction energy, 
   the charge of each ligand atom is multiplied by the calculated 
   electrostatic potential of the receptor at its coordinates.
   
   The Lennard-Jones part of the interaction energy is calculated by 
   summing up the pairwise Lennard-Jones interaction energies 
   for each pair of ligand and receptor atoms.
   
   FIXME: add refinement of complex candidates with solvation energies?
          what about nonpolar contributions besides Lennard Jones?
          what about a volume term if the ligand is treated flexible?
   


   Details for the electrostatics part of the calculation:
   
   Solves the Poisson-Boltzmann in an 
   
   6-value dielectric environment and in an 
   2-value electrolyte environment (linear gradient in channel region) 
   
   qm region (eps1set epsin1), 
    eps1set should contain the qm region
   normal region (eps2set, epsin2),
    eps2set should contain the mm region
   cavity region (eps3set, epsin3) 
    eps3set should contain the cavity region
   membrane core region 
    (lower_core_boundary <= z <= upper_core_boundary, epscore)
   membrane headgroup region
    (lower_head_boundary <= z <  lower_core_boundary ||
     upper_core_boundary <  z <= upper_head_boundary, epscore)
   solvent region (epsext, z >=/< membrane center, ionicstr1/ionicstr2)
   channel region epsext, 
    lower_core_boundary <= z <= upper_core_boundary
       ionicstr1 * (z - lowerhead_boundary)
                 / (upper_head_boundary - lowerhead_boundary)
     + ionicstr2 * (1 - ((z - lowerhead_boundary)
                 / (upper_head_boundary - lowerhead_boundary)))
    z > upper_head_boundary ionicstr1
    z < lower_head_boundary ionicstr2

   eps1set and eps2set specify two filenames containing all atoms in
   the epsin1 and epsin2 region (pqr file, charge not used).
   If files with the same name and ending ".txt" or ".dat" exist, it
   is assumed, that they were generated by pqr2SolvAccVol and read
   additionally to the pqr file. This avoids calculation of the
   solvent accessible volume analytically for each site, which can be costly.
   
   Similarly the IonAccessibleVolume for the ElectrolyteEnvironment can be
   read from files with the name protname and ending ".iav.txt" (ASCII) 
   or ".iav.dat" (binary).

   The two pqr files given as input files are used to calculate the charge
   distributions of the ligand and the receptor.

   The option fpt points to an extended fpt-file, which can be read.
   At the coordinates, given in the file, the electrostatic potential
   is calculated.

   The extended fpt format has six columns:
   Site number, instance number, x, y, z, and charge.
   
   The pqr files can be in my extended pqr format, containing four extra
   columns (integer values) after the columns of a normal pqr file.
   They can be used to identify atoms uniquely. For standard MEAD, two
   atoms are identical, when atom name and residue name are the same.
   This is not sufficient, when working with chains and multiple instances
   of a site.
   The intended usage (but not required) is conformer number, chain number,
   site number and instance number.

   Additionally to the features needed to calculate energies in the three
   dielectric environment, this program has also the option to calculate
   the energy of the site in a homogeneous dielectric of given value.
   The purpose for this is discussed in the appropriate section below.

   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Thomas Ullmann, 2007
   
   /usr/bin/g++-4.1 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 mcdock.cc
   /usr/bin/g++-4.1 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o mcdock mcdock.o -L../../libmead_e -lmead -lm

*/

#include "mcdock.h"

int main(int argc, char* argv[])
{
  // Only for benchmarking
  struct tms start, end, // start and end of program
    beps1set, aeps1set,  // reading or calculating eps1set
    beps2set, aeps2set,  // reading or calculating eps2set
    beps3set, aeps3set,  // reading or calculating eps2set
    bely1set, aely1set,  // reading or calculating ely1set
    bely2set, aely2set,  // reading or calculating ely2set
    bely3set, aely3set,  // reading or calculating ely3set
    bepsdx, aepsdx,      // writing DielCubeRep OpenDX
    belydx, aelydx,      // writing ElyCubeRep OpenDX
    bpotdx, apotdx,      // writing potential to OpenDX        
    bpbe, apbe,          // calculating PBE
    bhomo, ahomo,        // homogeneous dielectric calc
    bfpt, afpt;          // reading FPT, writing results
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);

  string ligname;
  string protname;
  string eps1set;
  string eps2set;
  string eps3set;
  string channelset;
  string elycavset;
  string fptname;
  string pfname;
  string tmpstr;
  string chmtop;
  string chmpar;
  
  float epsvac = 1.0;
  float epsin  = 0.0;  // These zeros are error states.  A specification of
  float epsin1 = 0.0;  // epsin or
  float epsin2 = 0.0;  // epsin1, epsin2, epsin3, epshead, epscore on command line is required.
  float epsin3 = 0.0;  // the command line are  required.
  float epshead = 0.0; // Dielectric constant of the membrane headgroups
  float epscore = 0.0; // Dielectric constant of the membrane core
  float epshomo = 0.0; // Dielectric constant of the homogeneous dielectric

  float istr1   = 0.0; // Ionic strength of the upper solvent region
  float istr2   = 0.0; // Ionic strength of the lower solvent region

  bool do_fpt = false;
  bool DoProteinField = false;

  // flags determinining if analytical ion accessible volumes could be read
  bool got_iav1 = false;
  bool got_iav2 = false;
  bool got_iav3 = false;
  
  bool write_eps = false; // write DielCubeRep to OpenDX
  bool write_ely = false; // write ElyCubeRep  to OpenDX 
  bool write_pot = false; // write potential to OpenDX
  
  int nruns     = 1000;    // number of independent runs
  int nsasteps  = 50;      // number of Simulated Annealing steps
  int nmcsteps  = 1000;    // number of MC steps at each SA step
  int nstructs  = 10;      // number of complexes to keep for refinement
  int nrefsteps = 1000;    // number of MC steps for refinement
  
  float tstart = 1.0E+03; // starting temperature for SA
  float tend   = 5.0E-01; // final temperature for SA
  
  int count, countx, county, countz; //gridpoints
  float x, y, z, space; // grid center x, y, z, grid spacing
  int ismemb = 0;
  float zlowercore = 0, zuppercore = 0, zlowerhead = 0, zupperhead = 0;
  float holerad = 0, holex = 0, holey = 0;  

  ////////////// Parse the command line /////////////////////////////////

  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+2==argc) {  // These are the last ones, must be ligname protname
      ligname = argv[iarg];
      ++iarg;
      protname = argv[iarg];
    }
    else if ((string) argv[iarg] == (string) "-epsin") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin option");
      epsin = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin1 option");
      epsin1 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin2 option");
      epsin2 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin3") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin3");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin3 option");
      epsin3 = f;
    }
    //////// My options /////////
    else if ((string) argv[iarg] == (string) "-eps1set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps1set");
      eps1set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps2set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps2set");
      eps2set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps3set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps3set");
      eps3set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-channelset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -channelset");
      channelset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-elycavset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -elycavset");
      elycavset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-fpt") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -fpt");
      fptname = argv[++iarg];
      do_fpt = true;
    }
    else if ((string) argv[iarg] == (string) "-pf") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -pf");
      pfname = argv[++iarg];
      DoProteinField = true;
    }
    else if ((string) argv[iarg] == (string) "-epshomo") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epshomo");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epshomo option");
      epshomo = f;
    }
    /////////////////////////////
    else if ((string) argv[iarg] == (string) "-epshead") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epshead");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epshead option");
      epshead = f;
    }
    else if ((string) argv[iarg] == (string) "-epscore") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epscore");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epscore option");
      epscore = f;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr1 option");
      istr1 = f;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr2 option");
      istr2 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsext") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsext");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsext option");
      PhysCond::set_epsext(f);
    }
    else if ((string) argv[iarg] == (string) "-epsvac") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsvac");
      istringstream ist(argv[++iarg]);
      ist >> epsvac;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsvac option");
    }
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -solrad");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -solrad option");
      PhysCond::set_solrad(f);
    }
    else if ((string) argv[iarg] == (string) "-sterln") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -sterln");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -sterln option");
      PhysCond::set_sterln(f);
    }
    else if ((string) argv[iarg] == (string) "-ionicstr") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr option");
      PhysCond::set_ionicstr(f);
    }
    else if ((string) argv[iarg] == (string) "-T") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -T");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -T option");
      PhysCond::set_T(f);
    }
    else if ((string) argv[iarg] == (string) "-kBolt") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -kBolt");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -kBolt option");
      PhysCond::set_kBolt(f);
    }
    else if ((string) argv[iarg] == (string) "-conconv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -conconv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -conconv option");
      PhysCond::set_conconv(f);
    }
    else if ((string) argv[iarg] == (string) "-econv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -econv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -econv option");
      PhysCond::set_econv(f);
    }
    else if ((string) argv[iarg] == (string) "-pkunits") {
      PhysCond::set_econv(PhysCond::get_pkconv());
    }    
    else if ((string) argv[iarg] == (string) "-bohr_radius") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -bohr_radius");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -bohr_radius option");
      PhysCond::set_bohr_radius(f);
    }
    else if ((string) argv[iarg] == (string) "-proton_charge") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -proton_charge");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -proton_charge opt.");
      PhysCond::set_proton_charge(f);
    }
    // process the form "-membz float float float float"
    else if ((string) argv[iarg] == (string) "-membz") {
      ismemb = 1;
      if (iarg+2>=argc) error ("main: ERROR, not enough args after -membz");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membz opt.");
      zlowercore = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -membz opt.");
      zuppercore = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -membz opt.");
      zlowerhead = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -membz opt.");
      zupperhead = f;    }
    // process the form "-membhole float [float float]"
    else if ((string) argv[iarg] == (string) "-membhole") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -membhole");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membhole opt.");
      holerad = f;
      if (iarg+2 < argc) { // There _could_ be two more floats...
	istringstream istx(argv[iarg+1]);
	istx >> f;
	if (istx.good() || istx.eof()) { // Yes, expect this float and another.
	  holex = f;
	  iarg+=2;
	  istringstream isty(argv[iarg]);
	  isty >> f;
	  if (!(isty.good() || isty.eof()))
	    error("main: ERROR, only two floats seen after -membhole.\n",
		  "One or three expected.\n");
	  else
	    holey = f;
	}
      }
    }
    else if ((string) argv[iarg] == (string) "-write_eps") {
      write_eps = true;
    }
    else if ((string) argv[iarg] == (string) "-write_ely") {
      write_ely = true;
    }
    else if ((string) argv[iarg] == (string) "-write_pot") {
      write_pot = true;
    }
    else if ((string) argv[iarg] == (string) "-x") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -x");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -x option");
      x = f;
    }
    else if ((string) argv[iarg] == (string) "-y") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -y");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -y option");
      y = f;
    }
    else if ((string) argv[iarg] == (string) "-z") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -z");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -z option");
      z = f;
    }
    else if ((string) argv[iarg] == (string) "-space") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -space");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -space option");
      space = f;
    }
    // -count int [int int]
    else if ((string) argv[iarg] == (string) "-count") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -count");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -count option");
      count = i;
      countx = i;
      county = i;
      countz = i;
      if (iarg+2 < argc) { // There _could_ be two more ints...
	istringstream isty(argv[iarg+1]);
	isty >> i;
	if (isty.good() || isty.eof()) { // Yes, expect this float and another.
	  county = i;
	  iarg+=2;
	  istringstream istz(argv[iarg]);
	  istz >> i;
	  if (!(istz.good() || istz.eof()))
	    error("main: ERROR, only two floats seen after -count\n",
		  "One or three expected.\n");
	  else
	    countz = i;
	}
      }
    }
    else if ((string) argv[iarg] == (string) "-converge_oldway")
      FDGridLevel::set_use_fixed_maxrmsdiff(true);
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    // the following options refer to SA/MMC docking
    else if ((string) argv[iarg] == (string) "-Tstart") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -Tstart");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -Tstart option");
      tstart = f;
    }
    else if ((string) argv[iarg] == (string) "-Tend") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -Tend");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -Tend option");
      tend = f;
    }    
    else if ((string) argv[iarg] == (string) "-runs") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -runs");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -runs option");
      nruns = i;
    }    
    else if ((string) argv[iarg] == (string) "-sasteps") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -sasteps");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -sasteps option");
      nsasteps = i;
    }     
    else if ((string) argv[iarg] == (string) "-mcsteps") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -mcsteps");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -mcsteps option");
      nmcsteps = i;
    }
    else if ((string) argv[iarg] == (string) "-structs") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -structs");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -structs option");
      nstructs = i;
    }     
    else if ((string) argv[iarg] == (string) "-refsteps") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -refsteps");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -refsteps option");
      nrefsteps = i;
    }
    else if ((string) argv[iarg] == (string) "-chmtop") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -chmtop");
      chmtop = argv[++iarg];
    }    
    else if ((string) argv[iarg] == (string) "-chmpar") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -chmpar");
      chmpar = argv[++iarg];
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

  /////////////// Checks & Statistics ////////////////////////////////

  if (ligname == (string) "")
    error("main: ERROR, ligname not given on command line");
  if (protname == (string) "")
    error("main: ERROR, protname not given on command line");
  else if (ismemb != 0){
   if (epscore == 0.0)
     ::error("main: ERROR, epscore value not specified.",
             "   This flag is now mandatory.");
   else if (epshead == 0.0)
     ::error("main: ERROR, epshead value not specified.",
             "   This flag is now mandatory.");
   if (epsin1 == 0.0 && epsin2 == 0.0 && epsin3 == 0.0)
    ::error("main: ERROR, -epsin1, -epsin2 and -epsin3 values not specified.",
             "   One of these flags is now mandatory.");
   else if (epsin1 == 0.0 && eps1set != (string) "")
     ::error("main: ERROR, -epsin1 value not specified.",
             "   This flag is now mandatory if eps1set is given.");
   else if (epsin2 == 0.0 && eps2set != (string) "")
     ::error("main: ERROR, -epsin1 value not specified.",
             "   This flag is now mandatory if eps1set is given.");
   else if (epsin3 == 0.0 && eps3set != (string) "")
     ::error("main: ERROR, -epsin3 value not specified.",
             "   This flag is now mandatory.");
   else if (eps1set == (string) "" && eps2set == (string) "" && eps3set == (string) "")
     ::error("main: ERROR, -eps1set, eps2set and -eps3set values not specified.",
             "   One of these flags is now mandatory.");
   else if (eps1set == (string) ""){
     if (epsin1 != 0){
       ::error("main: ERROR, eps1set value not specified.",
             "   This flag is now mandatory if epsin1 is set.");
     } else {epsin1 = 1;} // dummy value
   } // will be checked by DielectricEnvironment
   else if (eps2set == (string) ""){
     if (epsin2 != 0){
       ::error("main: ERROR, eps2set value not specified.",
             "   This flag is now mandatory if epsin2 is set.");
     } else {epsin2 = 1;} // dummy value
   } // will be checked by DielectricEnvironment
   else if (eps3set == (string) ""){
     if (epsin3 != 0){
       ::error("main: ERROR, eps3set value not specified.",
             "   This flag is now mandatory if epsin3 is set.");
     } else {epsin3 = 80;} // dummy value
   } // will be checked by DielectricEnvironment   
  }
  else if (ismemb == 0){
   if (eps1set == (string) ""){
     if (epsin1 != 0){
       ::error("main: ERROR, eps1set value not specified.",
             "   This flag is now mandatory if epsin1 is set.");
     } else {epsin1 = 1;} // dummy value
   } // will be checked by DielectricEnvironment
   else if (eps2set == (string) ""){
     if (epsin2 != 0){
       ::error("main: ERROR, eps2set value not specified.",
             "   This flag is now mandatory if epsin2 is set.");
     } else {epsin2 = 1;} // dummy value
   } // will be checked by DielectricEnvironment
   
   if (eps1set == (string) "" && eps2set == (string) ""){
    if (epsin != 0){
      epsin2 = epsin;
    }
    else {
       ::error("main: ERROR, epsin value not specified.",
               "   This flag is now mandatory if epsin1/eps1set, epsin2/eps2set are not set.");     
    }
   }
  }
  
  // this check will be done again in calc_cuberep
  // of Electrolyte and Dielectric Environments
 
  if (ismemb != 0){ 
   if (zuppercore > zupperhead){
     ::error("main: ERROR, zuppercore > zupperhead");  
   }
   else if (zlowercore > zupperhead){
    ::error("main: ERROR, zlowercore > zupperhead");  
   }
   else if (zlowerhead > zupperhead){
     ::error("main: ERROR, zlowerhead > zupperhead");  
   }
   else if (zlowercore > zuppercore){
     ::error("main: ERROR, zlowercore > zuppercore");  
   }
   else if (zlowerhead > zuppercore){
     ::error("main: ERROR, zlowerhead > zuppercore");  
   }
   else if (zlowerhead > zlowercore){
     ::error("main: ERROR, zlowerhead > zlowercore");  
   } 
  
  // check which type of ElectrolyteEnvironment is needed
  // if elycavset and channelset -> TwoElectrolyteMembThreeAtoms
  
    if (elycavset != (string) "" && channelset != (string) ""){
      ismemb=3;
    }// if only elycavset -> TwoElectrolyteMembTwoAtoms
    else if (elycavset != (string) "" && channelset != (string) ""){
      ismemb=2;
    }// if none -> TwoElectrolyteMembAtoms
    else {
      ismemb=1;
    }
  
   // if ionicstr1 and ionicstr2 are not specified on commandline
   // but ionicstr is specified then both solvent regions are 
   // assigned ionicstr
  
   if (istr1 == 0.0 && istr2 == 0 && PhysCond::get_ionicstr() != 0 ){
    istr1 = PhysCond::get_ionicstr();
    istr2 = PhysCond::get_ionicstr();
   }
  } 
    
  // if OpenDX output is requested we need to know
  // grid spacing and grid length, the grid center defaults to 0,0,0 
  // in principle one could also read an .ogm 
  // but usually we use multiple grids
  
  if (write_eps || write_ely){
    if (space == 0.0){
     ::error("main: ERROR, OpenDX output requested via ",
             "-write_eps|ely flag, but space value not specified.");
    }
    else if (count == 0){
     ::error("main: ERROR, OpenDX output requested via ",
             "-write_eps|ely flag, but count value not specified.");
    }
  }
  
  // parameter verification for SA/MMC docking
  
  if (tstart < 0) {
    ::error("main: ERROR, starting temperature specified via ",
            "-Tstart flag < 0.");
  }
  else if (tend < 0) {
    ::error("main: ERROR, final temperature specified via ",
            "-Tend flag < 0.");
  }
  else if (tstart < tend) {
    ::error("main: ERROR, starting temperature specified via ",
            "-Tstart flag < final temperature specified via -Tend flag.");
  }
  else if (nstructs > nsasteps){
    ::error("main: ERROR, number of SA steps specified via ",
            "-sasteps flag < number of structures to keep specified via -structs flag.");  
  } 
  else if (chmtop == (string) ""){
    ::error("main: ERROR, name of CHARMM topology file not specified via ",
            "-chmtop flag.");  
  }
  else if (chmpar == (string) ""){
    ::error("main: ERROR, name of CHARMM parameter file not specified via ",
            "-chmpar flag.");  
  }
          
  Coord c(holex, holey, 0.0); // position of cylindrical hole in membrane 
  
  //////////////// Done with comand line processing ////////////////
  
  /////// Print summary of input info from the command line: ///////
  
  cout << "Starting " << argv[0] << " for receptor named " << protname << "\n";
  cout << "and ligand named " << ligname << "\n";
  cout << "using the following settings for SA/MMC:\n";
  cout << " Number of independent runs, nruns = " << nruns
    << "\n Number of Simulated annealing steps for each run, nsasteps = " << nsasteps 
    << "\n Number of Metropolis Monte Carlo steps for each SA step, nmcsteps = " << nmcsteps 
    << "\n Number of lowest energy structures to keep for refinement, nstructs = " << nstructs 
    << "\n Number of Metropolis Monte Carlo steps for refinement, refsteps = " << nrefsteps 
    << "\n Starting temperature for SA, Tstart = " << tstart 
    << "\n Final temperature for SA, Tend = " << tend
    << "\n CHARMM topology  file, chmtop = chmtop" << chmtop
    << "\n CHARMM parameter file, chmpar = chmpar" << chmpar
    << "\n" << endl; 
  cout << "using the following physical conditions for solving the LPBE:\n";
  PhysCond::print();
  cout << " Protein interior dielectric constant, epsin1 = " << epsin1
    << "\n Protein interior dielectric constant, epsin2 = " << epsin2
    << "\n Protein interior dielectric constant, epsin3 = " << epsin3
    << "\n Membrane headgroup region dielectric constant, epshead = " << epshead
    << "\n Membrane core region dielectric constant, epscore = " << epscore
    << "\n Vacuum dielectric constant, epsvac = " << epsvac
    << "\n Ionic strength in upper solvent region, ionicstr1 = " << istr1
    << "\n Ionic strength in lower solvent region, ionicstr2 = " << istr2    
    << "\n Membrane headgroup region lower boundary = " << zlowerhead
    << "\n Membrane headgroup region upper boundary = " << zupperhead
    << "\n Membrane core region lower boundary = " << zlowercore
    << "\n Membrane core region upper boundary = " << zuppercore    
    << "\n" << endl;
  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ////////////////// Setting up calculation ////////////////////////

  /////// read parameter files ////////

  CharmmPara chmpara("CHARMM parameters");
  chmpara.read(chmtop, chmpar);

  /////// read structure files ////////

  // charge set of ligand
  AtomSet a1(ligname);
  a1.read();
  ChargeDist rholgnd(new AtomChargeSet(a1));
  
  //////// TEST charmm parameter reading ///////
  //a1.read_charmm(chmtop, chmpar, a1);
  //////////////////////////////////////////////
  
  // background charge set
  AtomSet a2(protname);
  a2.read();
  ChargeDist rhoprot(new AtomChargeSet(a2));

  // eps1set
  AtomSet a3(eps1set);
  if (eps1set != ""){
    a3.read();
  }  
  
  // eps2set / protein
  if (eps2set != ""){
   tmpstr = eps2set;
  } else if (eps1set == "" && eps3set == "") {
   tmpstr = protname;
  } 
  AtomSet a4(tmpstr);
  if (tmpstr != ""){
   a4.read(); 
  }
  
  // eps3set
  AtomSet a5(eps3set);
  if (eps3set != ""){
    a5.read();
  }
  
  // elycavset
  AtomSet a6(elycavset);
  if (elycavset != ""){
    a6.read();
  }
 
  // channelset
  AtomSet a7(channelset);
  if (channelset != ""){
    a7.read();
  }

  // Make the union of a3, a4 and a5 for defining electrolyte boundary
  // and protein envelop for ensuring that no solvent is found 
  // at the QM/MM interface

  AtomSet u(a4);
  for (AtomSet::const_iterator i = a3.begin(); i!=a3.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (a4.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the mm and qm atom sets.\n"
          << "   For defining the Electrolyte boundary, the mm version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  // for TwoValueElectrolyteMembXAtoms    
  AtomSet qm_mm(u);
  for (AtomSet::const_iterator i = a5.begin(); i!=a5.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (qm_mm.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm and cav atom sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  // for TwoValueElectrolyteMembXAtoms    
  AtomSet qm_mm_cav(u);
  // for SixValueDielMembThreeAtoms we add also elycavset and channelset  
  for (AtomSet::const_iterator i = a6.begin(); i!=a6.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (qm_mm_cav.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm_cav and elycav sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm_cav version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }  
  for (AtomSet::const_iterator i = a7.begin(); i!=a7.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (u.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm_cav_elycav and channel sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm_cav_elycav version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  AtomSet all(u);
  
  // for SixValueDielMembThreeAtoms we add up all solvent filled regions 
  // within membrane cav (eps3set), elycavset and channelset
  AtomSet allcav(a5);
  // for SixValueDielMembThreeAtoms we add also elycavset and channelset  
  for (AtomSet::const_iterator i = a6.begin(); i!=a6.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (qm_mm_cav.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the cav and elycav sets.\n"
          << "   For defining the Electrolyte boundary, the cav version\n"
            << "   will be used."  << endl;
    }
    else {
      allcav.insert(a);
    }
  }  
  for (AtomSet::const_iterator i = a7.begin(); i!=a7.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (u.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the cav_elycav and channel sets.\n"
          << "   For defining the Electrolyte boundary, the cav_elycav version\n"
            << "   will be used."  << endl;
    }
    else {
      allcav.insert(a);
    }
  }
  
  /////// calculate or read precalculated analytic volumes ////////
 
  // use pqr2SolvAccVol, pqr2IonAccVol for precalculation
  
  SolvAccVol sav1(qm_mm);
  SolvAccVol sav2(a4);
  SolvAccVol sav3(allcav);
  
  IonAccVol iav1(qm_mm_cav);
  IonAccVol iav2(elycavset);
  IonAccVol iav3(channelset);

  // Takes the name given to construct AtomSet, 
  // appends ".dat" or ".txt" respectively.
  // Read functions return zero on error!
  times(&beps1set);
  if (eps1set != ""){
   tmpstr = eps1set;
  } else {tmpstr = protname;}
  blab1 << "Try to read in SolvAccVol for eps1set, else calculate it" << endl;
  if (sav1.read_top_in_binary(tmpstr + DOT + "dat") == 0) {
    blab1 << "Unable to read in SolvAccVol from binary file for eps1set" << endl;
    if (sav1.read_top_in_ascii(tmpstr + DOT + "txt") == 0) {
      blab1 << "Unable to read in SolvAccVol from ASCII file for eps1set" << endl;
      blab1 << "Calculating SolvAccVol from AtomSet for eps1set" << endl;
      sav1.anal_calc(); // this is a bit costly for big molecules
    }
  }
  times(&aeps1set);

  times(&beps2set);
  blab1 << "Try to read in SolvAccVol for eps2set, else calculate it" << endl;
  if (eps2set != ""){
   tmpstr = eps2set;
  } else {tmpstr = protname;}   
  if (sav2.read_top_in_binary(tmpstr + DOT + "dat") == 0) {
    blab1 << "Unable to read in SolvAccVol from binary file for eps2set" << endl;
    if (sav2.read_top_in_ascii(tmpstr + DOT + "txt") == 0) {
      blab1 << "Unable to read in SolvAccVol from ASCII file for eps2set" << endl;
      blab1 << "Calculating SolvAccVol from AtomSet for eps2set" << endl;
      sav2.anal_calc(); // this is a bit costly for big molecules
    }
  }
  times(&aeps2set);

  times(&beps3set);
  blab1 << "Try to read in SolvAccVol for eps3set, else calculate it" << endl;
  if (sav3.read_top_in_binary(eps3set + DOT + "dat") == 0) {
    blab1 << "Unable to read in SolvAccVol from binary file for eps3set" << endl;
    if (sav3.read_top_in_ascii(eps3set + DOT + "txt") == 0) {
      blab1 << "Unable to read in SolvAccVol from ASCII file for eps3set" << endl;
      blab1 << "Calculating SolvAccVol from AtomSet for eps3set" << endl;
      sav3.anal_calc(); // this is a bit costly for big molecules
    }
  }
  times(&aeps3set);

  times(&bely1set);
  blab1 << "Try to read in IonAccVol for eps1set+eps2set+eps3set, else calculate it" << endl;
  if (iav1.read_top_in_binary(protname + DOT + "iav" + DOT + "dat") == 0) {
    blab1 << "Unable to read in IonAccVol from binary file for eps1set+eps2set+eps3set" << endl;
    if (iav1.read_top_in_ascii(protname + DOT + "iav" + DOT + "txt") == 0) {
      blab1 << "Unable to read in IonAccVol from ASCII file for eps1set+eps2set+eps3set" << endl;
      blab1 << "Calculating IonAccVol from AtomSet for eps1set+eps2set+eps3set" << endl;
      //iav1.anal_calc(); // this is a bit costly for big molecules
      got_iav1 = false;
    } else got_iav1 = true;
  } else got_iav1 = true;
  times(&aely1set);

  times(&bely2set);
  if (elycavset != (string) ""){
   blab1 << "Try to read in IonAccVol for elycavset, else calculate it" << endl;
   if (iav2.read_top_in_binary(elycavset + DOT + "iav" + DOT + "dat") == 0) {
     blab1 << "Unable to read in IonAccVol from binary file for elycavset" << endl;
     if (iav2.read_top_in_ascii(elycavset + DOT + "iav" + DOT + "txt") == 0) {
       blab1 << "Unable to read in IonAccVol from ASCII file for elycavset" << endl;
       blab1 << "Calculating IonAccVol from AtomSet for elycavset" << endl;
       //iav2.anal_calc(); // this is a bit costly for big molecules
       got_iav2 = false;
     } else got_iav2 = true;
   } else got_iav2 = true;
  }
  times(&aely2set);

  times(&bely3set);
  if (channelset != (string) ""){
   blab1 << "Try to read in IonAccVol for channelset, else calculate it" << endl;
   if (iav3.read_top_in_binary(elycavset + DOT + "iav" + DOT + "dat") == 0) {
     blab1 << "Unable to read in IonAccVol from binary file for channelset" << endl;
     if (iav3.read_top_in_ascii(elycavset + DOT + "iav" + DOT + "txt") == 0) {
       blab1 << "Unable to read in IonAccVol from ASCII file for channelset" << endl;
       blab1 << "Calculating IonAccVol from AtomSet for channelset" << endl;
       //iav3.anal_calc(); // this is a bit costly for big molecules
       got_iav3 = false;
     } else got_iav3 = true;
   } else got_iav3 = true;
  }
  times(&aely3set);
    
  blab3 << "Eps1set runtime: " << ((float) (aeps1set.tms_utime-beps1set.tms_utime))/ticks << " s user " << ((float) (aeps1set.tms_stime-beps1set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Eps2set runtime: " << ((float) (aeps2set.tms_utime-beps2set.tms_utime))/ticks << " s user " << ((float) (aeps2set.tms_stime-beps2set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Eps3set runtime: " << ((float) (aeps3set.tms_utime-beps3set.tms_utime))/ticks << " s user " << ((float) (aeps3set.tms_stime-beps3set.tms_stime))/ticks <<  " s system " << endl;
  
  blab3 << "Ely1set  runtime: " << ((float) (aely1set.tms_utime-bely1set.tms_utime))/ticks << " s user " << ((float) (aely1set.tms_stime-bely1set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Ely2set  runtime: " << ((float) (aely2set.tms_utime-bely2set.tms_utime))/ticks << " s user " << ((float) (aely2set.tms_stime-bely2set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Ely3set  runtime: " << ((float) (aely3set.tms_utime-bely3set.tms_utime))/ticks << " s user " << ((float) (aely3set.tms_stime-bely3set.tms_stime))/ticks <<  " s system " << endl;

  // Now I should have analytical surfaces either loaded or calculated
  // Constructor called with solvent/ion accessible volume instead of atom
  // set as done in solinprot!

  // ismemb determines which kind of DielectricEnvironment is needed

  DielectricEnvironment_lett* tmpeps = 0;  

  if (ismemb == 0){
    // protein, no membrane
    blab3 << "DielectricEnvironment is a MyThreeValueDielectricByAtoms" << endl; 
    tmpeps = new MyThreeValueDielectricByAtoms(sav1, epsin1, sav2, epsin2);
  }  
  else { 
    // protein, membrane with cylindrical hole (can have zero width)
    // protein qm/mm, protein mm, 
    // cavities / invaginations / channel volume (can be zero / no atoms used in definition) 
    blab3 << "DielectricEnvironment is a SixValueDielMembThreeAtoms" << endl; 
    tmpeps = new SixValueDielMembThreeAtoms(sav1, sav2, sav3, zlowerhead, zupperhead, 
                                            zlowercore, zuppercore, c, holerad, epshead, 
                                            epscore, epsin1, epsin2, epsin3, PhysCond::get_epsext());
  }
    
  DielectricEnvironment eps(tmpeps);  
  
  // ismemb determines which kind of ElectrolyteEnvironment is needed
  
  ElectrolyteEnvironment_lett* tmpely = 0;  

  if (ismemb == 0){ 
    // protein, no membrane
    blab3 << "ElectrolyteEnvironment is an ElectrolyteByAtoms" << endl; 
    if (got_iav1) tmpely = new ElectrolyteByAtoms(iav1);
    else tmpely = new SimpleElectrolyteByAtoms(qm_mm_cav);
  }
  else if (ismemb == 1){
    // protein, membrane with cylindrical hole (can have zero width)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembAtoms" << endl;
    if (got_iav1)
     tmpely = new TwoElectrolyteMembAtoms(iav1, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembAtoms(qm_mm_cav, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
  }
  else if (ismemb == 2){
    // protein, membrane with electrolyte filled invaginations/cavities
    // (can be zero / no atoms used in definition)
    // and cylindrical hole (can have zero width)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembTwoAtoms" << endl;
    if (got_iav1 && got_iav2)
     tmpely = new TwoElectrolyteMembTwoAtoms(iav1, iav2, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembTwoAtoms(qm_mm_cav, elycavset, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
  }
  else {
    // protein, membrane with electrolyte filled invaginations/cavities
    // and transmembrane channel (can be zero / no atoms used in definition)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembThreeAtoms" << endl;
    if (got_iav1 && got_iav2 && got_iav3)
     tmpely = new TwoElectrolyteMembThreeAtoms(iav1, iav2, iav3, zlowerhead, zupperhead,
                  zlowercore, zuppercore, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembThreeAtoms(qm_mm_cav, elycavset, channelset, zlowerhead, zupperhead,
                  zlowercore, zuppercore, istr1, istr2);
  }
  
  ElectrolyteEnvironment ely(tmpely);
  
  ///// write CubeReps to file if OpenDX output was requested /////
  
  if (write_eps || write_ely){
    
    blab1 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);
    blab1 << "Calculating grid centered at " << center << endl;

    CubeLatSpec cls(count, space, center);
    blab1 << "Gridpoints " << count << ", spacing " << space << endl;
    
    if (write_eps) {
      times(&bepsdx);
      cout << "Writing DielectricEnvironment to OpenDX ..." << endl;
      eps.write_dx(cls, "eps" + DOT + protname);
      times(&aepsdx);
      blab3 << "Writing DielCubeRep to OpenDX runtime: " << ((float) (aepsdx.tms_utime-bepsdx.tms_utime))/ticks << " s user " << ((float) (aepsdx.tms_stime-bepsdx.tms_stime))/ticks <<  " s system " << endl;
    }  
    if (write_ely) {
      times(&belydx);
      cout << "Writing ElectrolyteEnvironment to OpenDX ..." << endl;
      ely.write_dx(cls, "ely" + DOT + protname);
      times(&aelydx);
      blab3 << "Writing ElyCubeRep to OpenDX runtime: " << ((float) (aelydx.tms_utime-belydx.tms_utime))/ticks << " s user " << ((float) (aelydx.tms_stime-belydx.tms_stime))/ticks <<  " s system " << endl;
    } 
  }   
  

  //////////// now the actual grid levels are defined /////////////

    
  FinDiffMethod fdm;
  // ogm file has name of instance of site
  string ogm_filename = protname + ".ogm";
  fdm.read(ogm_filename);
  // First focus on geometric center of protein (all QM or MM atoms)
  // to get a reasonable grid resolution everywhere (for fpt calculations)
  // Finally focus on given coordinate set, which is the geometric center
  // of the site (not only the present instance)
  // This coordinate has to be kept fixed for the site to cancel grid
  // artefacts!
  // (SPECIFIC is defined in CenteringStyle.h and used in CubeLatSpec::resolve)
  fdm.resolve(all.geom_cent(), a2.geom_cent());
  cout << "Using finite difference method with lattice levels:" << endl;
  cout << fdm;

  
  ////////////// Now we are ready for solving PBE... /////////////////////

  times(&bpbe);
  ElstatPot phi(fdm, eps, rhoprot, ely);
  phi.solve();
  float self = (phi * rhoprot) / 2 * PhysCond::get_econv();
  float back = phi * rholgnd * PhysCond::get_econv();

  cout << "\n\nSelf energy of " << protname << " in Dielectric environment = "
       << self << "\n(probably in kcal/mole, INCLUDING GRID ARTEFACTS)" << endl;
  cout << "\n\nInteraction energy of " << ligname << " with background charge set in Dielectric environment = "
       << back << "\n(probably in kcal/mole)" << endl;
  times(&apbe);
  blab3 << "PBE in Six Dielectric runtime: " << ((float) (apbe.tms_utime-bpbe.tms_utime))/ticks << " s user " << ((float) (apbe.tms_stime-bpbe.tms_stime))/ticks <<  " s system " << endl;

  ///////////////// write potential to OpenDX //////////////////
  
  if (write_pot){

    times(&bpotdx);

    blab3 << "Writing Potential to OpenDX ..." << endl;
    blab3 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);
    
    blab3 << "Calculating grid centered at " << center << endl;
    
    Coord origin(center.x - (((countx - 1) * space) / 2),
                 center.y - (((county - 1) * space) / 2),
                 center.z - (((countz - 1) * space) / 2));

    int nsq = countz * county;
    int ncube = nsq * countx;
    float lengthx = space * (countx - 1);
    float lengthy = space * (county - 1);
    float lengthz = space * (countz - 1);
    
    blab3 << "Preparing temporary array of potential values ..." << endl;
    
    float *tmparray = new float[ncube];  
    
    Coord tmpvec(origin);
    
    float kt_over_e_factor =    PhysCond::get_energy_unit()
                              / PhysCond::get_Boltzmann_constant()
                              / PhysCond::get_T();
    
    for (int k=0; k<countz; k++){
      for (int i=0; i<countx; i++){
        int iterm = i * nsq;
        for (int j=0; j<county; j++){
          int h = iterm + (j * countz) + k;
          tmparray[h] = phi.value(tmpvec) * kt_over_e_factor;
          tmpvec.y   += space;
        }
        tmpvec.x += space; 
        tmpvec.y  = origin.y;
      }
      tmpvec.x  = origin.x;
      tmpvec.y  = origin.y;
      tmpvec.z += space; 
    }

    // now we are ready to write to OpenDX

    string curve_name = "pot" + DOT + protname + DOT + "dx";
    
    ofstream of (curve_name.c_str());
    // Could I open file for reading?
    if (!of.good()){
      ::error(argv[0], "main: failed to open OpenDX file for writing, ",
      curve_name.c_str());
    }
    of << "# "<< curve_name << " " << countx << "x" 
                                   << county << "x"
                                   << countz << " " 
                                   << space << " " << center << endl;
    of << "#\n# Potential written by MEAD in kT/e.\n#" << endl;
    of << "object 1 class gridpositions counts " << countx << " " << county << " " << countz << endl;
    of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
    of << "object 2 class gridconnections counts " << countx << " " << county << " " << countz << endl;
    of << "object 3 class array type double rank 0 times " << ncube << endl;
    int cnt = 0;
    // Now the actual values
    // Don't know if I could write more than three columns according to OpenDX format,
    // but the TCL reader can only read three columns or less...
    for (int h=2; h<ncube; h+=3) {
      of << tmparray[h-2] << " " << tmparray[h-1] << " " << tmparray[h] << endl;
      cnt++;
    }
    if (ncube - (3 * cnt) == 2){
      of << tmparray[ncube-2] << " " << tmparray[ncube-1] << endl;
    }
    if (ncube - (3 * cnt) == 1){
      of << tmparray[ncube-1] << endl;
    }
    of << "attribute \"dep\" string \"positions\"" << endl;
    of << "object \"regular positions regular connections\" class field" << endl;
    of << "component \"positions\" value 1" << endl;
    of << "component \"connections\" value 2" << endl;
    of << "component \"data\" value 3" << endl;
    // check if file handle still open
    if (! of.good()){
      ::error(argv[0], "main: failed to write GMT curve file, ",
      curve_name.c_str());
    }
    of.close();
    
    delete [] tmparray;
    
    times(&apotdx);
   
    blab3 << "Writing Membrane Potential to OpenDX runtime: " << ((float) (apotdx.tms_utime-bpotdx.tms_utime))/ticks << " s user " << ((float) (apotdx.tms_stime-bpotdx.tms_stime))/ticks <<  " s system " << endl;
     
  }
/*  
  ////////////////////// Calculate W /////////////////////////////////////

  // Interaction energy between present instance of site and all instances
  // of all other sites
  // Read fpt file, calculate potential, multiply by charge, sum up for each 
  // instance of each site

  if (do_fpt) {
    times(&bfpt);
    cout << "\n\n\nInteraction energies of " << ligname << " with charges of given instances of sites (probably in kcal/mole).\nFormat: site instance energy\n" << endl;
    // The fpt file is for all instances of all sites the same file
    // Analysis software has to make sure not to include interaction of
    // instance with itself or site with itself!
    string fieldpoint_filename = fptname + ".fpt" ;
    ifstream fpt (fieldpoint_filename.c_str());
    // Could I open file for reading?
    if (!fpt.good())
      ::error(argv[0], "main: failed to open for reading, ",
              fieldpoint_filename.c_str());
    
    float energy = 0; // energy of instance of site
    int old_site=-1, old_instance=-1;
    // for each line in file
    while (fpt.good()) {
      Coord c;
      int site, instance;
      float charge;
      fpt >> site >> instance >> c >> charge;
      // Catch exceptions if End-Of-File or format error
      if (fpt.eof()) break;
      if (!fpt.good())
        ::error(argv[0], "main: input failure while reading ",
                fieldpoint_filename.c_str());

      // I guess there is a more elegant way of doing the following...

      // initialize old_site and old_instance in first round of loop
      if (old_site == -1 || old_instance == -1) {
	old_site = site;
	old_instance = instance;
      }
      // same instance of site as before increment energy by the energy
      // of the present charge
      if (old_site == site && old_instance == instance) {
	energy += phi.value(c) * charge;
      }
      // new instance or site, so write out what I have for the previous
      // and reset variables
      else {
	cout << old_site << "   " << old_instance << " : " << energy * PhysCond::get_econv() << endl;
	old_site = site;
	old_instance = instance;
	energy = phi.value(c) * charge;
      }
    } // WHILE
    // Don't forget also to write the last site!
    cout << old_site << "   " << old_instance << " : " << energy * PhysCond::get_econv() << endl;   

    times(&afpt);
    blab3 << "FPT runtime: " << ((float) (afpt.tms_utime-bfpt.tms_utime))/ticks << " s user " << ((float) (afpt.tms_stime-bfpt.tms_stime))/ticks <<  " s system " << endl;
  } // END do_fpt

*/

  times(&end);
  cout << "Total runtime: " << ((float) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((float) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;

  return 0;
} // END main
