/* Calculate electrostatic self/born energies and interaction energies
   for a integral membrane protein which may also possess cavities and
   a trans-membrane channel
   Main program based on Timm Essigkes my_x_diel_solver.

   my_memb_solver -epsin1 eps -epsin2 eps -epsin3 eps\
                  -epscore eps -epshead eps -epsext eps \
                  -eps1set name -eps2set name -eps3set name \
                  -channelset name -elycavset name\
                  -membz lower_core_boundary upper_core_boundary \
                         lower_head_boundary upper_head_boundary \
                  -membhole radius xcoord ycoord \
                  -pkunits \
                  -fpt filename -pf -epshomo \
                  -write_eps -write_ely -write_rho -write_pot \
                  -count gridpoints(x) [gridpoints(x) gridpoints(y)] \
                  -space spacing -x center.x -y center.y -z center.z
                  -gmt ptx pty ptz vecx vecy vecz spacing gridpoints {pt2x pt2y pt2z vec2x vec2y vec2}
                  sitename protname

   commandline options:

   mandatory      -epsin1 float -epsin2 float -epsin3 float \
                  -epscore float -epshead float -epsext float \
                  -eps1set string -eps2set string eps3set string \
                  -membz float float float float \

   optional       -elycavset string -channelset string \
                  -membhole float float float \
                  -fpt string -pf string -epshomo float \
                  -write_eps -write_ely -write_pot \
                  -count int [int int] -space float \
                  -x float -y float -z float \-pkunits \

   others (opt.)  -solrad float -sterln float -T float \
                  -kBolt float -conconv float -econv float \
                  -bohr_radius float -proton_charge float \
                  -blab1 -blab2 -blab3 -converge_oldway


   Solves the Poisson-Boltzmann in an

   6-value dielectric environment and in an
   2-value electrolyte environment (linear gradient in channel region)

   qm region (eps1set epsin1),
    eps1set should contain the qm region
   normal region (eps2set, epsin2),
    eps2set should contain the mm region
   cavity region (eps3set, epsin3)
    eps3set should contain the cavity region
   membrane core region
    (lower_core_boundary <= z <= upper_core_boundary, epscore)
   membrane headgroup region
    (lower_head_boundary <= z <  lower_core_boundary ||
     upper_core_boundary <  z <= upper_head_boundary, epscore)
   solvent region (epsext, z >=/< membrane center, ionicstr1/ionicstr2)
   channel region epsext,
    lower_core_boundary <= z <= upper_core_boundary
       ionicstr1 * (z - lowerhead_boundary)
                 / (upper_head_boundary - lowerhead_boundary)
     + ionicstr2 * (1 - ((z - lowerhead_boundary)
                 / (upper_head_boundary - lowerhead_boundary)))
    z > upper_head_boundary ionicstr1
    z < lower_head_boundary ionicstr2

   eps1set and eps2set specify two filenames containing all atoms in
   the epsin1 and epsin2 region (pqr file, charge not used).
   If files with the same name and ending ".txt" or ".dat" exist, it
   is assumed, that they were generated by pqr2SolvAccVol and read
   additionally to the pqr file. This avoids calculation of the
   solvent accessible volume analytically for each site, which can be costly.

   Similarly the IonAccessibleVolume for the ElectrolyteEnvironment can be
   read from files with the name protname and ending ".iav.txt" (ASCII)
   or ".iav.dat" (binary).

   The two pqr files given as input files are used to calculate the charge
   distributions of the site and the background charge set.

   The Born and background energies are computed (without canceling grid
   artefacts! Second calculation needs to be done with the same grid and
   energy differences between the two calculations can be interpreted!).

   The option fpt points to an extended fpt-file, which can be read.
   At the coordinates, given in the file, the electrostatic potential
   is calculated.

   The extended fpt format has six columns:
   Site number, instance number, x, y, z, and charge.

   The calculated potential is multiplied by the charge at the given
   coordinates. This interaction energy is summed up for each instance
   of each site and given in the output.

   The pqr files can be in my extended pqr format, containing four extra
   columns (integer values) after the columns of a normal pqr file.
   They can be used to identify atoms uniquely. For standard MEAD, two
   atoms are identical, when atom name and residue name are the same.
   This is not sufficient, when working with chains and multiple instances
   of a site.
   The intended usage (but not required) is conformer number, chain number,
   site number and instance number.

   Additionally to the features needed to calculate energies in the three
   dielectric environment, this program has also the option to calculate
   the energy of the site in a homogeneous dielectric of given value.
   The purpose for this is discussed in the appropriate section below.

   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Thomas Ullmann, 2007

   /usr/bin/g++-3.4 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 my_memb_solver.cc
   /usr/bin/g++-3.4 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o my_memb_solver my_memb_solver.o -L../../libmead -lmead  -lm

*/


#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <fstream>
using std::ifstream;
#include <string>
using std::string;
#include <sys/times.h> // I want to do some benchmarking of options...
#include <unistd.h>

#include "MEAD/PhysCond.h"
#include "MEAD/AtomSet.h"
#include "MEAD/ChargeDist.h"
#include "MEAD/AtomChargeSet.h"
#include "MEAD/MembChargeSet.h"
#include "MEAD/DielectricEnvironment.h"
#include "MEAD/DielByAtoms.h"
#include "MEAD/DielMembAtoms.h"
#include "MEAD/UniformDielectric.h"

#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/ElectrolyteByAtoms.h"
#include "MEAD/ElectrolyteMembAtoms.h"

#include "MEAD/FinDiffMethod.h"
#include "MEAD/ElstatPot.h"
#include "MEAD/FDGridLevel.h"

int main(int argc, char* argv[])
{
  // Only for benchmarking
  struct tms start, end, // start and end of program
    beps1set, aeps1set,  // reading or calculating eps1set
    beps2set, aeps2set,  // reading or calculating eps2set
    beps3set, aeps3set,  // reading or calculating eps2set
    bely1set, aely1set,  // reading or calculating ely1set
    bely2set, aely2set,  // reading or calculating ely2set
    bely3set, aely3set,  // reading or calculating ely3set
    bepsdx, aepsdx,      // writing DielCubeRep OpenDX
    belydx, aelydx,      // writing ElyCubeRep OpenDX
    bpotdx, apotdx,      // writing potential to OpenDX
    brhodx, arhodx,      // writing charge distribution to OpenDX
    bpotgmt, apotgmt,    // writing membrane potential to GMT
    bpbe, apbe,          // calculating PBE
    bhomo, ahomo,        // homogeneous dielectric calc
    bfpt, afpt,          // reading FPT, writing results
    bpf, apf;            // reading traditional FPT, writing results
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);

  string sitename;
  string protname;
  string eps1set;
  string eps2set;
  string eps3set;
  string channelset;
  string elycavset;
  string fptname;
  string pfname;
  float epsvac = 1.0;
  float epsin1 = 0.0;  // These zeros are error states.  A specification of
  float epsin2 = 0.0;  // epsin1, epsin2, epsin3, epshead, epscore
  float epsin3 = 0.0;  // on the command line is required.
  float epshead = 0.0; // Dielectric constant of the membrane headgroups
  float epscore = 0.0; // Dielectric constant of the membrane core
  float epshomo = 0.0; // Dielectric constant of the homogeneous dielectric

  float istr1   = 0.0; // Ionic strength of the upper solvent region
  float istr2   = 0.0; // Ionic strength of the lower solvent region

  bool do_fpt = false;
  bool DoProteinField = false;

  // flags determinining if analytical ion accessible volumes could be read
  bool got_iav1 = false;
  bool got_iav2 = false;
  bool got_iav3 = false;

  bool write_eps = false; // write DielCubeRep to OpenDX
  bool write_ely = false; // write ElyCubeRep  to OpenDX
  bool write_pot = false; // write potential to OpenDX
  bool write_rho = false; // write charge distribution to OpenDX
  bool DoGMT = false;          // write a gmt compatible 2D square curve in a plane
                               // with normal vector gmt_vec, spacing gmt_space and
                               // edge length gmt_count around a center point gmt_pt
                               // the upper edge is perpendicular to z
  bool do_2ndplane = false;    // trigger value calculation in second plane followed
                               // by back-projection on first plane

  int count, gmt_count, countx, county, countz; //gridpoints
  float x, y, z, space, gmt_space; // grid center x, y, z, grid spacing
  Coord gmt_pt, gmt_vec;           // base point and normal vector of plane
  Coord gmt_pt2, gmt_vec2;         // base point and normal vector of plane 2
  int ismemb = 0;
  float zlowercore = 0, zuppercore = 0, zlowerhead = 0, zupperhead = 0;
  float holerad = 0, holex = 0, holey = 0;

  ////////////// Parse the command line /////////////////////////////////
  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+2==argc) {  // These are the last ones, must be sitename protname
      sitename = argv[iarg];
      ++iarg;
      protname = argv[iarg];
    }
    else if ((string) argv[iarg] == (string) "-epsin1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin1 option");
      epsin1 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin2 option");
      epsin2 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsin3") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin3");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsin3 option");
      epsin3 = f;
    }
    //////// My options /////////
    else if ((string) argv[iarg] == (string) "-eps1set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps1set");
      eps1set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps2set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps2set");
      eps2set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-eps3set") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -eps3set");
      eps3set = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-channelset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -channelset");
      channelset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-elycavset") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -elycavset");
      elycavset = argv[++iarg];
    }
    else if ((string) argv[iarg] == (string) "-fpt") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -fpt");
      fptname = argv[++iarg];
      do_fpt = true;
    }
    else if ((string) argv[iarg] == (string) "-pf") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -pf");
      pfname = argv[++iarg];
      DoProteinField = true;
    }
    // process the form -gmt 7 float 1 int
    else if ((string) argv[iarg] == (string) "-gmt") {
      DoGMT = true;
      if (iarg+8>=argc)
      error ("main: ERROR, not enough args after -gmt");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -gmt opt.(pt.x)");
      gmt_pt.x = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -gmt opt. (pt.y)");
      gmt_pt.y = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -gmt opt. (pt.z)");
      gmt_pt.z = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -gmt opt. (vec.x)");
      gmt_vec.x = f;
      istringstream ist5(argv[++iarg]);
      ist5 >> f;
      if (!(ist5.good() || ist5.eof()))
	error ("main: FAILED trying to read fith value after -gmt opt. (vec.y)");
      gmt_vec.y = f;
      istringstream ist6(argv[++iarg]);
      ist6 >> f;
      if (!(ist6.good() || ist6.eof()))
	error ("main: FAILED trying to read sixth value after -gmt opt.(vec.z)");
      gmt_vec.z = f;
      istringstream ist7(argv[++iarg]);
      ist7 >> f;
      if (!(ist7.good() || ist7.eof()))
	error ("main: FAILED trying to read seventh value after -gmt opt.(space)");
      gmt_space = f;
      int i;
      istringstream ist8(argv[++iarg]);
      ist8 >> i;
      if (!(ist8.good() || ist8.eof()))
	error ("main: FAILED trying to read eighth value after -gmt opt.(count)");
      gmt_count = i;
      // try to read 6 more values giving the base point and normal vector of a second plane
      // values are calculated in this second plane then and projected onto the first plane
      istringstream ist9(argv[++iarg]);
      ist9 >> f;
      do_2ndplane = ((ist9.good() || ist9.eof()));
      if(do_2ndplane){
        if (iarg+5>=argc)
         error ("main: ERROR, not enough args after -gmt primary arguments, 8 or 14 values expected.");
        gmt_pt2.x = f;
        istringstream ist10(argv[++iarg]);
        ist10 >> f;
        if (!(ist10.good() || ist10.eof()))
 	 error ("main: FAILED trying to read 10th value after -gmt opt. (pt2.y)");
        gmt_pt2.y = f;
        istringstream ist11(argv[++iarg]);
        ist11 >> f;
        if (!(ist11.good() || ist11.eof()))
  	 error ("main: FAILED trying to read 11th value after -gmt opt. (pt2.z)");
        gmt_pt2.z = f;
        istringstream ist12(argv[++iarg]);
        ist12 >> f;
        if (!(ist12.good() || ist12.eof()))
 	 error ("main: FAILED trying to read 12th value after -gmt opt. (vec2.x)");
        gmt_vec2.x = f;
        istringstream ist13(argv[++iarg]);
        ist13 >> f;
        if (!(ist13.good() || ist13.eof()))
         error ("main: FAILED trying to read 13th value after -gmt opt. (vec2.y)");
        gmt_vec2.y = f;
        istringstream ist14(argv[++iarg]);
        ist14 >> f;
        if (!(ist14.good() || ist14.eof()))
  	 error ("main: FAILED trying to read 14th value after -gmt opt.(vec2.z)");
        gmt_vec2.z = f;
      } else {iarg--;}
    }
    else if ((string) argv[iarg] == (string) "-epshomo") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epshomo");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epshomo option");
      epshomo = f;
    }
    /////////////////////////////
    else if ((string) argv[iarg] == (string) "-epshead") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epshead");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epshead option");
      epshead = f;
    }
    else if ((string) argv[iarg] == (string) "-epscore") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epscore");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epscore option");
      epscore = f;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr1") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr1 option");
      istr1 = f;
    }
    else if ((string) argv[iarg] == (string) "-ionicstr2") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr2");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr2 option");
      istr2 = f;
    }
    else if ((string) argv[iarg] == (string) "-epsext") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsext");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsext option");
      PhysCond::set_epsext(f);
    }
    else if ((string) argv[iarg] == (string) "-epsvac") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsvac");
      istringstream ist(argv[++iarg]);
      ist >> epsvac;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -epsvac option");
    }
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -solrad");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -solrad option");
      PhysCond::set_solrad(f);
    }
    else if ((string) argv[iarg] == (string) "-sterln") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -sterln");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -sterln option");
      PhysCond::set_sterln(f);
    }
    else if ((string) argv[iarg] == (string) "-ionicstr") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -ionicstr");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -ionicstr option");
      PhysCond::set_ionicstr(f);
    }
    else if ((string) argv[iarg] == (string) "-T") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -T");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -T option");
      PhysCond::set_T(f);
    }
    else if ((string) argv[iarg] == (string) "-kBolt") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -kBolt");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -kBolt option");
      PhysCond::set_kBolt(f);
    }
    else if ((string) argv[iarg] == (string) "-conconv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -conconv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -conconv option");
      PhysCond::set_conconv(f);
    }
    else if ((string) argv[iarg] == (string) "-econv") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -econv");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -econv option");
      PhysCond::set_econv(f);
    }
    else if ((string) argv[iarg] == (string) "-pkunits") {
      PhysCond::set_econv(PhysCond::get_pkconv());
    }
    else if ((string) argv[iarg] == (string) "-bohr_radius") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -bohr_radius");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -bohr_radius option");
      PhysCond::set_bohr_radius(f);
    }
    else if ((string) argv[iarg] == (string) "-proton_charge") {
      if (iarg+1>=argc)
        error ("main: ERROR, not enough args after -proton_charge");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -proton_charge opt.");
      PhysCond::set_proton_charge(f);
    }
    // process the form "-membz float float float float"
    else if ((string) argv[iarg] == (string) "-membz") {
      ismemb = 1;
      if (iarg+2>=argc) error ("main: ERROR, not enough args after -membz");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membz opt.");
      zlowercore = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -membz opt.");
      zuppercore = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -membz opt.");
      zlowerhead = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -membz opt.");
      zupperhead = f;    }
    // process the form "-membhole float [float float]"
    else if ((string) argv[iarg] == (string) "-membhole") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -membhole");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -membhole opt.");
      holerad = f;
      if (iarg+2 < argc) { // There _could_ be two more floats...
	istringstream istx(argv[iarg+1]);
	istx >> f;
	if (istx.good() || istx.eof()) { // Yes, expect this float and another.
	  holex = f;
	  iarg+=2;
	  istringstream isty(argv[iarg]);
	  isty >> f;
	  if (!(isty.good() || isty.eof()))
	    error("main: ERROR, only two floats seen after -membhole.\n",
		  "One or three expected.\n");
	  else
	    holey = f;
	}
      }
    }
    else if ((string) argv[iarg] == (string) "-write_eps") {
      write_eps = true;
    }
    else if ((string) argv[iarg] == (string) "-write_ely") {
      write_ely = true;
    }
    else if ((string) argv[iarg] == (string) "-write_pot") {
      write_pot = true;
    }
    else if ((string) argv[iarg] == (string) "-write_rho") {
      write_rho = true;
    }
    else if ((string) argv[iarg] == (string) "-x") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -x");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -x option");
      x = f;
    }
    else if ((string) argv[iarg] == (string) "-y") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -y");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -y option");
      y = f;
    }
    else if ((string) argv[iarg] == (string) "-z") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -z");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -z option");
      z = f;
    }
    else if ((string) argv[iarg] == (string) "-space") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -space");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -space option");
      space = f;
    }
    // -count int [int int]
    else if ((string) argv[iarg] == (string) "-count") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -count");
      istringstream ist(argv[++iarg]);
      int i;
      ist >> i;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -count option");
      count = i;
      countx = i;
      county = i;
      countz = i;
      if (iarg+2 < argc) { // There _could_ be two more ints...
	istringstream isty(argv[iarg+1]);
	isty >> i;
	if (isty.good() || isty.eof()) { // Yes, expect this float and another.
	  county = i;
	  iarg+=2;
	  istringstream istz(argv[iarg]);
	  istz >> i;
	  if (!(istz.good() || istz.eof()))
	    error("main: ERROR, only two floats seen after -count\n",
		  "One or three expected.\n");
	  else
	    countz = i;
	}
      }
    }
    else if ((string) argv[iarg] == (string) "-converge_oldway")
      FDGridLevel::set_use_fixed_maxrmsdiff(true);
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

  /////////////// Checks & Statistics ////////////////////////////////

  if (sitename == (string) "")
    error("main: ERROR, sitename not given on command line");
  else if (protname == (string) "")
    error("main: ERROR, protname not given on command line");
  else if (epscore == 0.0)
    ::error("main: ERROR, epscore value not specified.",
            "   This flag is now mandatory.");
  else if (epshead == 0.0)
    ::error("main: ERROR, epshead value not specified.",
            "   This flag is now mandatory.");
  else if (ismemb == 0)
    ::error("main: ERROR, -zmemb zlowercore zuppercore zlowerhead zupperhead not specified.",
            "   This flag is now mandatory.");
  if (epsin1 == 0.0 && epsin2 == 0.0 && epsin3 == 0.0)
    ::error("main: ERROR, -epsin1, -epsin2 and -epsin3 values not specified.",
            "   One of these flags is now mandatory.");
  else if (epsin1 == 0.0 && eps1set != (string) "")
    ::error("main: ERROR, -epsin1 value not specified.",
            "   This flag is now mandatory if eps1set is given.");
  else if (epsin2 == 0.0 && eps2set != (string) "")
    ::error("main: ERROR, -epsin1 value not specified.",
            "   This flag is now mandatory if eps1set is given.");
  else if (epsin3 == 0.0 && eps3set != (string) "")
    ::error("main: ERROR, -epsin3 value not specified.",
            "   This flag is now mandatory.");
  else if (eps1set == (string) "" && eps2set == (string) "" && eps3set == (string) "")
    cerr << "main: WARNING, -eps1set, eps2set and -eps3set values not specified." << endl <<
            "      Will use default options in QMPB input script.";

  if (eps1set == (string) ""){
    if (epsin1 != 0){
      ::error("main: ERROR, eps1set value not specified.",
            "   This flag is now mandatory if epsin1 is set.");
    } else {epsin1 = 1;} // dummy value
  } // will be checked by DielectricEnvironment
  if (eps2set == (string) ""){
    if (epsin2 != 0){
      ::error("main: ERROR, eps2set value not specified.",
            "   This flag is now mandatory if epsin2 is set.");
    } else {epsin2 = 4;} // dummy value
  } // will be checked by DielectricEnvironment
  if (eps3set == (string) ""){
    if (epsin3 != 0){
      ::error("main: ERROR, eps3set value not specified.",
            "   This flag is now mandatory if epsin3 is set.");
    } else {epsin3 = 80;} // dummy value
  } // will be checked by DielectricEnvironment

  // this check will be done again in calc_cuberep
  // of Electrolyte and Dielectric Environments

  if (zuppercore > zupperhead){
    ::error("main: ERROR, zuppercore > zupperhead");
  }
  else if (zlowercore > zupperhead){
    ::error("main: ERROR, zlowercore > zupperhead");
  }
  else if (zlowerhead > zupperhead){
    ::error("main: ERROR, zlowerhead > zupperhead");
  }
  else if (zlowercore > zuppercore){
    ::error("main: ERROR, zlowercore > zuppercore");
  }
  else if (zlowerhead > zuppercore){
    ::error("main: ERROR, zlowerhead > zuppercore");
  }
  else if (zlowerhead > zlowercore){
    ::error("main: ERROR, zlowerhead > zlowercore");
  }

  // check which type of ElectrolyteEnvironment is needed
  // if channelset -> TwoElectrolyteMembThreeAtoms
  if (channelset != (string) ""){
    ismemb=3;
  }// if only elycavset -> TwoElectrolyteMembTwoAtoms
  else if (elycavset != (string) ""){
    ismemb=2;
  }// if none -> TwoElectrolyteMembAtoms
  else {
    ismemb=1;
  }

  // if ionicstr1 and ionicstr2 are not specified on commandline
  // but ionicstr is specified then both solvent regions are
  // assigned ionicstr

  if (istr1 == 0.0 && istr2 == 0 && PhysCond::get_ionicstr() != 0 ){
   istr1 = PhysCond::get_ionicstr();
   istr2 = PhysCond::get_ionicstr();
  }

  // if OpenDX output is requested we need to know
  // grid spacing and grid length, the grid center defaults to 0,0,0
  // in principle one could also read an .ogm
  // but usually we use multiple grids

  if (write_eps || write_ely){
    if (space == 0.0){
     ::error("main: ERROR, OpenDX output requested via",
             "-write_eps|ely flag, but space value not specified.");
    }
    else if (count == 0){
     ::error("main: ERROR, OpenDX output requested via",
             "-write_eps|ely flag, but count value not specified.");
    }
  }

  Coord c(holex, holey, 0.0); // position of cylindrical hole in membrane

  //////////////// Done with comand line processing ////////////////

  /////// Print summary of input info from the command line: ///////

  cout << "Starting " << argv[0] << " for molecule named " << sitename << "\n";
  cout << "using the following physical conditions:\n";
  PhysCond::print();
  cout << " Protein interior dielectric constant, epsin1 = " << epsin1
    << "\n  Protein interior dielectric constant, epsin2 = " << epsin2
    << "\n  Protein interior dielectric constant, epsin3 = " << epsin3
    << "\n  Membrane headgroup region dielectric constant, epshead = " << epshead
    << "\n  Membrane core region dielectric constant, epscore = " << epscore
    << "\n  Vacuum dielectric constant, epsvac = " << epsvac
    << "\n  Ionic strength in upper solvent region, ionicstr1 = " << istr1
    << "\n  Ionic strength in lower solvent region, ionicstr2 = " << istr2
    << "\n  Membrane headgroup region lower boundary = " << zlowerhead
    << "\n  Membrane headgroup region upper boundary = " << zupperhead
    << "\n  Membrane core region lower boundary = " << zlowercore
    << "\n  Membrane core region upper boundary = " << zuppercore
    << "\n" << endl;
  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ////////////////// Setting up calculation ////////////////////////

  // charge set of instance of site
  AtomSet a1(sitename);
  a1.read();
  ChargeDist rhosite(new AtomChargeSet(a1));
  // background charge set
  AtomSet a2(protname);
  a2.read();
  ChargeDist rhoback(new AtomChargeSet(a2));

  // eps1set
  AtomSet a3(eps1set);
  if (eps1set != ""){
    a3.read();
  }

  // eps2set
  AtomSet a4(eps2set);
  if (eps2set != ""){
    a4.read();
  }

  // eps3set
  AtomSet a5(eps3set);
  if (eps3set != ""){
    a5.read();
  }

  // elycavset
  AtomSet a6(elycavset);
  if (elycavset != ""){
    a6.read();
  }

  // channelset
  AtomSet a7(channelset);
  if (channelset != ""){
    a7.read();
  }

  // Make the union of a3, a4 and a5 for defining electrolyte boundary
  // and protein envelop for ensuring that no solvent is found
  // at the QM/MM interface

  AtomSet u(a4);
  for (AtomSet::const_iterator i = a3.begin(); i!=a3.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (a4.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the mm and qm atom sets.\n"
          << "   For defining the Electrolyte boundary, the mm version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  // for TwoValueElectrolyteMembXAtoms
  AtomSet qm_mm(u);
  for (AtomSet::const_iterator i = a5.begin(); i!=a5.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (qm_mm.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm and cav atom sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  // for TwoValueElectrolyteMembXAtoms
  AtomSet qm_mm_cav(u);
  // for SixValueDielMembThreeAtoms we add also elycavset and channelset
  for (AtomSet::const_iterator i = a6.begin(); i!=a6.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (qm_mm_cav.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm_cav and elycav sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm_cav version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  for (AtomSet::const_iterator i = a7.begin(); i!=a7.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (u.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the qm_mm_cav_elycav and channel sets.\n"
          << "   For defining the Electrolyte boundary, the qm_mm_cav_elycav version\n"
            << "   will be used."  << endl;
    }
    else {
      u.insert(a);
    }
  }
  AtomSet all(u);

  // for SixValueDielMembThreeAtoms we add up all solvent filled regions
  // within membrane cav (eps3set), elycavset and channelset
  AtomSet allcav(a5);
  // for SixValueDielMembThreeAtoms we add also elycavset and channelset
  for (AtomSet::const_iterator i = a6.begin(); i!=a6.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (allcav.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the cav and elycav sets.\n"
          << "   For defining the Electrolyte boundary, the cav version\n"
            << "   will be used."  << endl;
    }
    else {
      allcav.insert(a);
    }
  }
  for (AtomSet::const_iterator i = a7.begin(); i!=a7.end(); ++i) {
    const Atom& a = i->second;
    AtomID k = i->first;
    if (allcav.contains(k)) {
      cerr << argv[0] << ": WARNING:  The atom, \"" << k
        << "\",\n   occurs in both the cav_elycav and channel sets.\n"
          << "   For defining the Electrolyte boundary, the cav_elycav version\n"
            << "   will be used."  << endl;
    }
    else {
      allcav.insert(a);
    }
  }

  /////// calculate or read precalculated analytic volumes ////////

  // use pqr2SolvAccVol, pqr2IonAccVol for precalculation

  SolvAccVol sav1(qm_mm);
  SolvAccVol sav2(a4);
  SolvAccVol sav3(allcav);

  IonAccVol iav1(qm_mm_cav);
  IonAccVol iav2(elycavset);
  IonAccVol iav3(channelset);

  // Takes the name given to construct AtomSet,
  // appends ".dat" or ".txt" respectively.
  // Read functions return zero on error!
  times(&beps1set);
  blab1 << "Try to read in SolvAccVol for eps1set, else calculate it" << endl;
  if (eps1set != (string) ""){
    if (sav1.read_top_in_binary(eps1set + DOT + "dat") == 0) {
      blab1 << "Unable to read in SolvAccVol from binary file for eps1set" << endl;
      if (sav1.read_top_in_ascii(eps1set + DOT + "txt") == 0) {
        blab1 << "Unable to read in SolvAccVol from ASCII file for eps1set" << endl;
        blab1 << "Calculating SolvAccVol from AtomSet for eps1set" << endl;
        sav1.anal_calc(); // this is a bit costly for big molecules
      }
    }
  }
  times(&aeps1set);

  times(&beps2set);
  blab1 << "Try to read in SolvAccVol for eps2set, else calculate it" << endl;
  if (sav2.read_top_in_binary(eps2set + DOT + "dat") == 0) {
    blab1 << "Unable to read in SolvAccVol from binary file for eps2set" << endl;
    if (sav2.read_top_in_ascii(eps2set + DOT + "txt") == 0) {
      blab1 << "Unable to read in SolvAccVol from ASCII file for eps2set" << endl;
      blab1 << "Calculating SolvAccVol from AtomSet for eps2set" << endl;
      sav2.anal_calc(); // this is a bit costly for big molecules
    }
  }
  if (eps1set == (string) ""){
   sav1 = sav2;
  }
  times(&aeps2set);

  times(&beps3set);
  blab1 << "Try to read in SolvAccVol for eps3set, else calculate it" << endl;
  if (sav3.read_top_in_binary(eps3set + DOT + "dat") == 0) {
    blab1 << "Unable to read in SolvAccVol from binary file for eps3set" << endl;
    if (sav3.read_top_in_ascii(eps3set + DOT + "txt") == 0) {
      blab1 << "Unable to read in SolvAccVol from ASCII file for eps3set" << endl;
      blab1 << "Calculating SolvAccVol from AtomSet for eps3set" << endl;
      sav3.anal_calc(); // this is a bit costly for big molecules
    }
  }
  times(&aeps3set);

  times(&bely1set);
  blab1 << "Try to read in IonAccVol for eps1set+eps2set+eps3set, else calculate it" << endl;
  if (iav1.read_top_in_binary(protname + DOT + "iav" + DOT + "dat") == 0) {
    blab1 << "Unable to read in IonAccVol from binary file for eps1set+eps2set+eps3set" << endl;
    if (iav1.read_top_in_ascii(protname + DOT + "iav" + DOT + "txt") == 0) {
      blab1 << "Unable to read in IonAccVol from ASCII file for eps1set+eps2set+eps3set" << endl;
      blab1 << "Calculating IonAccVol from AtomSet for eps1set+eps2set+eps3set" << endl;
      //iav1.anal_calc(); // this is a bit costly for big molecules
      got_iav1 = false;
    } else got_iav1 = true;
  } else got_iav1 = true;
  times(&aely1set);

  times(&bely2set);
  if (elycavset != (string) ""){
   blab1 << "Try to read in IonAccVol for elycavset, else calculate it" << endl;
   if (iav2.read_top_in_binary(elycavset + DOT + "iav" + DOT + "dat") == 0) {
     blab1 << "Unable to read in IonAccVol from binary file for elycavset" << endl;
     if (iav2.read_top_in_ascii(elycavset + DOT + "iav" + DOT + "txt") == 0) {
       blab1 << "Unable to read in IonAccVol from ASCII file for elycavset" << endl;
       blab1 << "Calculating IonAccVol from AtomSet for elycavset" << endl;
       //iav2.anal_calc(); // this is a bit costly for big molecules
       got_iav2 = false;
     } else got_iav2 = true;
   } else got_iav2 = true;
  }
  times(&aely2set);

  times(&bely3set);
  if (channelset != (string) ""){
   blab1 << "Try to read in IonAccVol for channelset, else calculate it" << endl;
   if (iav3.read_top_in_binary(elycavset + DOT + "iav" + DOT + "dat") == 0) {
     blab1 << "Unable to read in IonAccVol from binary file for channelset" << endl;
     if (iav3.read_top_in_ascii(elycavset + DOT + "iav" + DOT + "txt") == 0) {
       blab1 << "Unable to read in IonAccVol from ASCII file for channelset" << endl;
       blab1 << "Calculating IonAccVol from AtomSet for channelset" << endl;
       //iav3.anal_calc(); // this is a bit costly for big molecules
       got_iav3 = false;
     } else got_iav3 = true;
   } else got_iav3 = true;
  }
  times(&aely3set);

  blab3 << "Eps1set runtime: " << ((float) (aeps1set.tms_utime-beps1set.tms_utime))/ticks << " s user " << ((float) (aeps1set.tms_stime-beps1set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Eps2set runtime: " << ((float) (aeps2set.tms_utime-beps2set.tms_utime))/ticks << " s user " << ((float) (aeps2set.tms_stime-beps2set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Eps3set runtime: " << ((float) (aeps3set.tms_utime-beps3set.tms_utime))/ticks << " s user " << ((float) (aeps3set.tms_stime-beps3set.tms_stime))/ticks <<  " s system " << endl;

  blab3 << "Ely1set  runtime: " << ((float) (aely1set.tms_utime-bely1set.tms_utime))/ticks << " s user " << ((float) (aely1set.tms_stime-bely1set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Ely2set  runtime: " << ((float) (aely2set.tms_utime-bely2set.tms_utime))/ticks << " s user " << ((float) (aely2set.tms_stime-bely2set.tms_stime))/ticks <<  " s system " << endl;
  blab3 << "Ely3set  runtime: " << ((float) (aely3set.tms_utime-bely3set.tms_utime))/ticks << " s user " << ((float) (aely3set.tms_stime-bely3set.tms_stime))/ticks <<  " s system " << endl;

  // Now I should have analytical surfaces either loaded or calculated
  // Constructor called with solvent/ion accessible volume instead of atom
  // set as done in solinprot!

  blab3 << "DielectricEnvironment is a SixValueDielMembThreeAtoms" << endl;

  // protein qm/mm, protein mm,
  // cavities / invaginations / channel volume (can be zero / no atoms used in definition)
  // membrane with cylindrical hole (can have zero width)

  DielectricEnvironment_lett *tmpeps = new SixValueDielMembThreeAtoms(sav1, sav2, sav3, zlowerhead, zupperhead,
                                                                      zlowercore, zuppercore, c, holerad,
                                                                      epshead, epscore, epsin1, epsin2, epsin3,
                                                                      PhysCond::get_epsext());
  DielectricEnvironment eps(tmpeps);

  // ismemb determines which kind of ElectrolyteEnvironment is needed

  ElectrolyteEnvironment_lett *tmpely = NULL;

  // check if analytical volumes could be read, otherwise do not calculate IAVs
  // but resort to faster inflated atom radii solution for performance reasons -
  // IAVs take awfully long if many atoms are present and or the atom density is
  // anomalously high (as in the case of calculations with multiple rotamers)
  if (ismemb == 1){
    // protein, membrane with cylindrical hole (can have zero width)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembAtoms" << endl;
    if (got_iav1)
     tmpely = new TwoElectrolyteMembAtoms(iav1, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembAtoms(qm_mm_cav, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
  }
  else if (ismemb == 2){
    // protein, membrane with electrolyte filled invaginations/cavities
    // (can be zero / no atoms used in definition)
    // and cylindrical hole (can have zero width)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembTwoAtoms" << endl;
    if (got_iav1 && got_iav2)
     tmpely = new TwoElectrolyteMembTwoAtoms(iav1, iav2, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembTwoAtoms(qm_mm_cav, a6, zlowerhead, zupperhead,
                  zlowercore, zuppercore, c, holerad, istr1, istr2);
  }
  else {
    // protein, membrane with electrolyte filled invaginations/cavities
    // and transmembrane channel (can be zero / no atoms used in definition)
    blab3 << "ElectrolyteEnvironment is a TwoElectrolyteMembThreeAtoms" << endl;
    if (got_iav1 && got_iav2 && got_iav3)
     tmpely = new TwoElectrolyteMembThreeAtoms(iav1, iav2, iav3, zlowerhead, zupperhead,
                  zlowercore, zuppercore, istr1, istr2);
    else
     tmpely = new TwoElectrolyteMembThreeAtoms(qm_mm_cav, a6, a7, zlowerhead, zupperhead,
                  zlowercore, zuppercore, istr1, istr2);
  }

  ElectrolyteEnvironment ely(tmpely);

  ///// write CubeReps to file if OpenDX output was requested /////

  if (write_eps || write_ely || write_rho){

    blab1 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);
    blab1 << "Calculating grid centered at " << center << endl;

    CubeLatSpec cls(count, space, center);
    blab1 << "Gridpoints " << count << ", spacing " << space << endl;

    if (write_eps) {
      times(&bepsdx);
      cout << "Writing DielectricEnvironment to OpenDX ..." << endl;
      eps.write_dx(cls, "eps" + DOT + sitename);
      times(&aepsdx);
      blab3 << "Writing DielCubeRep to OpenDX runtime: " << ((float) (aepsdx.tms_utime-bepsdx.tms_utime))/ticks << " s user " << ((float) (aepsdx.tms_stime-bepsdx.tms_stime))/ticks <<  " s system " << endl;
    }
    if (write_ely) {
      times(&belydx);
      cout << "Writing ElectrolyteEnvironment to OpenDX ..." << endl;
      ely.write_dx(cls, "ely" + DOT + sitename);
      times(&aelydx);
      blab3 << "Writing ElyCubeRep to OpenDX runtime: " << ((float) (aelydx.tms_utime-belydx.tms_utime))/ticks << " s user " << ((float) (aelydx.tms_stime-belydx.tms_stime))/ticks <<  " s system " << endl;
    }
    if (write_rho) {
      times(&brhodx);
      cout << "Writing protein ChargeDist to OpenDX ..." << endl;
      rhosite.write_dx(cls, "rho" + DOT + sitename);
      times(&arhodx);
      blab3 << "Writing protein ChargeCubeRep to OpenDX runtime: " << ((float) (arhodx.tms_utime-brhodx.tms_utime))/ticks << " s user " << ((float) (arhodx.tms_stime-brhodx.tms_stime))/ticks <<  " s system " << endl;
    }
  }


  //////////// now the actual grid levels are defined /////////////

  FinDiffMethod fdm;
  // ogm file has name of instance of site
  string ogm_filename = sitename + ".ogm";
  fdm.read(ogm_filename);
  // First focus on geometric center of protein (all QM or MM atoms)
  // to get a reasonable grid resolution everywhere (for fpt calculations)
  // Finally focus on given coordinate set, which is the geometric center
  // of the site (not only the present instance)
  // This coordinate has to be kept fixed for the site to cancel grid
  // artifacts!
  // (SPECIFIC is defined in CenteringStyle.h and used in CubeLatSpec::resolve)
  fdm.resolve(all.geom_cent(), a1.geom_cent());
  cout << "Using finite difference method with lattice levels:" << endl;
  cout << fdm;


  ////////////// Now we are ready for solving PBE... /////////////////////

  times(&bpbe);
  ElstatPot phi(fdm, eps, rhosite, ely);
  phi.solve();
  float self = (phi * rhosite) / 2 * PhysCond::get_econv();
  float back = phi * rhoback * PhysCond::get_econv();

  cout << "\n\nSelf energy of " << sitename << " in SixDielectric environment = "
       << self << "\n(probably in kcal/mole, INCLUDING GRID ARTEFACTS)" << endl;
  cout << "\n\nInteraction energy of " << sitename << " with background charge set in SixDielectric environment = "
       << back << "\n(probably in kcal/mole)" << endl;
  times(&apbe);
  blab3 << "PBE in Six Dielectric runtime: " << ((float) (apbe.tms_utime-bpbe.tms_utime))/ticks << " s user " << ((float) (apbe.tms_stime-bpbe.tms_stime))/ticks <<  " s system " << endl;

  ///////////////// Self energy in homogeneous Dielectric ////////////////

  // This is actually not the job for a "memb_solver", but it is not worth
  // writing a separate program...

  // Rotamers of MM sites are calculated by transfering them into a homo-
  // geneous dielectric, calculating the self energy, doing a conformational
  // change (i.e. internal (partial) CHARMM energy difference of site) and
  // transfering it back into a three dielectric environment.

  // No check is done, that site is actually in the same dielectric region
  // as given by epshomo. The preprocessor software has to take care of this.
  if (epshomo > 0.0) {
    times(&bhomo);
    ElectrolyteEnvironment homo_ely; // No electrolyte is the default
    DielectricEnvironment homo_eps(new UniformDielectric(epshomo));
    ElstatPot homo_phi(fdm, homo_eps, rhosite, homo_ely);
    homo_phi.solve();
    float homo_self = (homo_phi * rhosite) / 2 * PhysCond::get_econv();
    cout << "\n\nSelf energy of " << sitename << " in homogeneous dielectric environment = "
	 << homo_self << "\n(probably in kcal/mole, INCLUDING GRID ARTEFACTS)" << endl;
    times(&ahomo);
    blab3 << "Homogeneous Dielectric runtime: " << ((float) (ahomo.tms_utime-bhomo.tms_utime))/ticks << " s user " << ((float) (ahomo.tms_stime-bhomo.tms_stime))/ticks <<  " s system " << endl;
  }


  ////////////////////// Calculate W /////////////////////////////////////

  // Interaction energy between present instance of site and all instances
  // of all other sites
  // Read fpt file, calculate potential, multiply by charge, sum up for each
  // instance of each site

  if (do_fpt) {
    times(&bfpt);
    cout << "\n\n\nInteraction energies of " << sitename << " with charges of given instances of sites (probably in kcal/mole).\nFormat: site instance energy\n" << endl;
    // The fpt file is for all instances of all sites the same file
    // Analysis software has to make sure not to include interaction of
    // instance with itself or site with itself!
    string fieldpoint_filename = fptname + ".fpt" ;
    ifstream fpt (fieldpoint_filename.c_str());
    // Could I open file for reading?
    if (!fpt.good())
      ::error(argv[0], "main: failed to open for reading, ",
              fieldpoint_filename.c_str());

    float energy = 0; // energy of instance of site
    int old_site=-1, old_instance=-1;

    // while read commands are successful
    // FIXME? : this is potentially dangerous
    // line breaks are overread if a line does not contain the complete data
    // incorrect lines do not lead to an error
    Coord c;
    int site = -1, instance = -1;
    float charge, x, y, z;
    while (fpt >> site >> instance >> c >> charge) {

      // initialize old_site and old_instance in first round of loop
      if (old_site == -1 || old_instance == -1) {
	old_site = site;
	old_instance = instance;
      }
      // same instance of site as before increment energy by the energy
      // of the present charge
      if (old_site == site && old_instance == instance) {
	energy += phi.value(c) * charge;
      }
      // new instance or site, so write out what I have for the previous
      // and reset variables
      else {
	cout << old_site << "   " << old_instance << " : " << energy * PhysCond::get_econv() << endl;
	old_site = site;
	old_instance = instance;
	energy = phi.value(c) * charge;
      }
    } // WHILE
    // Don't forget also to write the last site!
    cout << old_site << "   " << old_instance << " : " << energy * PhysCond::get_econv() << endl;

    times(&afpt);
    blab3 << "FPT runtime: " << ((float) (afpt.tms_utime-bfpt.tms_utime))/ticks << " s user " << ((float) (afpt.tms_stime-bfpt.tms_stime))/ticks <<  " s system " << endl;
  } // END do_fpt

  times(&bpf);
  if (DoProteinField) {
    string fieldpoint_filename = pfname + ".fpt" ;
    ifstream fpt (fieldpoint_filename.c_str());
    // Could I open file for reading?
    if (!fpt.good())
      ::error(argv[0], "main: failed to open for reading, ",
              fieldpoint_filename.c_str());

    // while read commands are successful
    // FIXME? : this is potentially dangerous
    // line breaks are overread if a line does not contain the complete data
    // incorrect lines do not lead to an error
    Coord c;
    float charge;
    while (fpt >> c >> charge) {
      cout << "PF-TAG " << c << "   " << charge << "   " <<  phi.value(c) << "  : " << phi.value(c) * charge * PhysCond::get_econv() << "\n";
    }
  }
  times(&apf);
  blab3 << "ProteinField runtime: " << ((float) (apf.tms_utime-bpf.tms_utime))/ticks << " s user " << ((float) (apbe.tms_stime-bpbe.tms_stime))/ticks <<  " s system " << endl;

  ///// write ChargeCubeRep to file if OpenDX output was requested /////

  if (write_rho){

    blab1 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);
    blab1 << "Calculating grid centered at " << center << endl;

    CubeLatSpec cls(count, space, center);
    blab1 << "Gridpoints " << count << ", spacing " << space << endl;

    times(&brhodx);

    // fake a membrane potential inducing charge distribution to get the boundaries right
    ChargeDist_lett* tmprho = 0;
    if (ismemb == 1){
      // protein, membrane with cylindrical hole (can have zero width)
      blab3 << "ElectrolyteEnvironment is a MembChargeSetMembAtoms" << endl;
      if (got_iav1) tmprho = new MembChargeSetMembAtoms(iav1, zlowerhead, zupperhead,
                                 zlowercore, zuppercore, c, holerad, istr1, istr2, (float)1.0);
      else          tmprho = new MembChargeSetMembAtoms(qm_mm_cav, zlowerhead, zupperhead,
                                 zlowercore, zuppercore, c, holerad, istr1, istr2, (float)1.0);
    }
    else if (ismemb == 2){
      // protein, membrane with electrolyte filled invaginations/cavities
      // (can be zero / no atoms used in definition)
      // and cylindrical hole (can have zero width)
      blab3 << "ElectrolyteEnvironment is a MembChargeSetMembTwoAtoms" << endl;
      if (got_iav1 && got_iav2)
       tmprho = new MembChargeSetMembTwoAtoms(iav1, iav2, zlowerhead, zupperhead,
                    zlowercore, zuppercore, c, holerad, istr1, istr2, (float)1.0);
      else
       tmprho = new MembChargeSetMembTwoAtoms(qm_mm_cav, elycavset, zlowerhead, zupperhead,
                    zlowercore, zuppercore, c, holerad, istr1, istr2, (float)1.0);
    }
    else {
      // protein, membrane with electrolyte filled invaginations/cavities
      // and transmembrane channel (can be zero / no atoms used in definition)
      blab3 << "ElectrolyteEnvironment is a MembChargeSetMembThreeAtoms" << endl;
      if (got_iav1 && got_iav2 && got_iav3)
       tmprho = new MembChargeSetMembThreeAtoms(iav1, iav2, iav3, zlowerhead, zupperhead,
                    zlowercore, zuppercore, istr1, istr2, (float)1.0);
      else
       tmprho = new MembChargeSetMembThreeAtoms(qm_mm_cav, elycavset, channelset, zlowerhead, zupperhead,
                    zlowercore, zuppercore, istr1, istr2, (float)1.0);
    }
    ChargeDist rhomembpot(tmprho);

    cout << "Writing induced ChargeDist to OpenDX ..." << endl;

    ChargeCubeRep* rhomobile = rhomembpot.get_cuberep(cls, phi, (bool) false);
    rhomembpot.write_dx(cls, "rho" + DOT + "ind" + DOT + sitename, rhomobile);

    times(&arhodx);
    blab3 << "Writing membrane potential ChargeCubeReps to OpenDX runtime: " << ((float) (arhodx.tms_utime-brhodx.tms_utime))/ticks << " s user " << ((float) (arhodx.tms_stime-brhodx.tms_stime))/ticks <<  " s system " << endl;

  }

  ///////////// write electrostatic potential to OpenDX //////////////

  if (write_pot){

    times(&bpotdx);

    blab3 << "Writing Potential in kT/e to OpenDX ..." << endl;
    blab3 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);

    blab3 << "Calculating grid centered at " << center << endl;

    Coord origin(center.x - (((countx - 1) * space) / 2),
                 center.y - (((county - 1) * space) / 2),
                 center.z - (((countz - 1) * space) / 2));

    int nsq = countz * county;
    int ncube = nsq * countx;
    float lengthx = space * (countx - 1);
    float lengthy = space * (county - 1);
    float lengthz = space * (countz - 1);

    blab3 << "Preparing temporary array of potential values ..." << endl;

    float* tmparray = new float[ncube];

    Coord tmpvec(origin);

    float kt_over_e_factor =    PhysCond::get_energy_unit()
                              / PhysCond::get_Boltzmann_constant()
                              / PhysCond::get_T();

    for (int k=0; k<countz; k++){
      for (int i=0; i<countx; i++){
        int iterm = i * nsq;
        for (int j=0; j<county; j++){
          int h = iterm + (j * countz) + k;
          tmparray[h] = phi.value(tmpvec) * kt_over_e_factor;
          tmpvec.y   += space;
        }
        tmpvec.x += space;
        tmpvec.y  = origin.y;
      }
      tmpvec.x  = origin.x;
      tmpvec.y  = origin.y;
      tmpvec.z += space;
    }

    // now we are ready to write to OpenDX

    string curve_name = "pot" + DOT + sitename + DOT +"kT_over_e" + DOT + "dx";

    ofstream of (curve_name.c_str());
    // Could I open file for reading?
    if (!of.good()){
      ::error(argv[0], "main: failed to open GMT curve for writing, ",
      curve_name.c_str());
    }
    of << "# "<< curve_name << " " << countx << "x"
                                   << county << "x"
                                   << countz << " "
                                   << space << " " << center << endl;
    of << "#\n# Potential written by MEAD in kT/e.\n#" << endl;
    of << "object 1 class gridpositions counts " << countx << " " << county << " " << countz << endl;
    of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
    of << "object 2 class gridconnections counts " << countx << " " << county << " " << countz << endl;
    of << "object 3 class array type double rank 0 times " << ncube << endl;
    int cnt = 0;
    // Now the actual values
    // Don't know if I could write more than three columns according to OpenDX format,
    // but the TCL reader can only read three columns or less...
    for (int h=2; h<ncube; h+=3) {
      of << tmparray[h-2] << " " << tmparray[h-1] << " " << tmparray[h] << endl;
      cnt++;
    }
    if (ncube - (3 * cnt) == 2){
      of << tmparray[ncube-2] << " " << tmparray[ncube-1] << endl;
    }
    if (ncube - (3 * cnt) == 1){
      of << tmparray[ncube-1] << endl;
    }
    of << "attribute \"dep\" string \"positions\"" << endl;
    of << "object \"regular positions regular connections\" class field" << endl;
    of << "component \"positions\" value 1" << endl;
    of << "component \"connections\" value 2" << endl;
    of << "component \"data\" value 3" << endl;
    // check if file handle still open
    if (! of.good()){
      ::error(argv[0], "main: failed to write GMT curve file, ",
      curve_name.c_str());
    }
    of.close();

    delete [] tmparray;

    times(&apotdx);

    blab3 << "Writing Potential to OpenDX runtime: " << ((float) (apotdx.tms_utime-bpotdx.tms_utime))/ticks << " s user " << ((float) (apotdx.tms_stime-bpotdx.tms_stime))/ticks <<  " s system " << endl;

  }
  ///////////// write electrostatic potential to OpenDX //////////////

  if (write_pot){

    times(&bpotdx);

    blab3 << "Writing Potential in mV to OpenDX ..." << endl;
    blab3 << "Defining grid for OpenDX output ..." << endl;

    Coord center(x, y, z);

    blab3 << "Calculating grid centered at " << center << endl;

    Coord origin(center.x - (((countx - 1) * space) / 2),
                 center.y - (((county - 1) * space) / 2),
                 center.z - (((countz - 1) * space) / 2));

    int nsq = countz * county;
    int ncube = nsq * countx;
    float lengthx = space * (countx - 1);
    float lengthy = space * (county - 1);
    float lengthz = space * (countz - 1);

    blab3 << "Preparing temporary array of potential values ..." << endl;

    float* tmparray = new float[ncube];

    Coord tmpvec(origin);

    for (int k=0; k<countz; k++){
      for (int i=0; i<countx; i++){
        int iterm = i * nsq;
        for (int j=0; j<county; j++){
          int h = iterm + (j * countz) + k;
          tmparray[h] =  phi.value(tmpvec) * PhysCond::get_voltconv() * 1000;
          tmpvec.y   += space;
        }
        tmpvec.x += space;
        tmpvec.y  = origin.y;
      }
      tmpvec.x  = origin.x;
      tmpvec.y  = origin.y;
      tmpvec.z += space;
    }

    // now we are ready to write to OpenDX

    string curve_name =  "pot" + DOT + sitename + DOT + "dx";

    ofstream of (curve_name.c_str());
    // Could I open file for reading?
    if (!of.good()){
      ::error(argv[0], "main: failed to open GMT curve for writing, ",
      curve_name.c_str());
    }
    of << "# "<< curve_name << " " << countx << "x"
                                   << county << "x"
                                   << countz << " "
                                   << space << " " << center << endl;
    of << "#\n# Potential written by MEAD in mV.\n#" << endl;
    of << "object 1 class gridpositions counts " << countx << " " << county << " " << countz << endl;
    of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
    of << "object 2 class gridconnections counts " << countx << " " << county << " " << countz << endl;
    of << "object 3 class array type double rank 0 times " << ncube << endl;
    int cnt = 0;
    // Now the actual values
    // Don't know if I could write more than three columns according to OpenDX format,
    // but the TCL reader can only read three columns or less...
    for (int h=2; h<ncube; h+=3) {
      of << tmparray[h-2] << " " << tmparray[h-1] << " " << tmparray[h] << endl;
      cnt++;
    }
    if (ncube - (3 * cnt) == 2){
      of << tmparray[ncube-2] << " " << tmparray[ncube-1] << endl;
    }
    if (ncube - (3 * cnt) == 1){
      of << tmparray[ncube-1] << endl;
    }
    of << "attribute \"dep\" string \"positions\"" << endl;
    of << "object \"regular positions regular connections\" class field" << endl;
    of << "component \"positions\" value 1" << endl;
    of << "component \"connections\" value 2" << endl;
    of << "component \"data\" value 3" << endl;
    // check if file handle still open
    if (! of.good()){
      ::error(argv[0], "main: failed to write GMT curve file, ",
      curve_name.c_str());
    }
    of.close();

    delete [] tmparray;

    times(&apotdx);

    blab3 << "Writing Potential to OpenDX runtime: " << ((float) (apotdx.tms_utime-bpotdx.tms_utime))/ticks << " s user " << ((float) (apotdx.tms_stime-bpotdx.tms_stime))/ticks <<  " s system " << endl;

  }



  ////////// write GMT 2D-curve of potential ///////////

  if (DoGMT) {

    times(&bpotgmt);

    string curvedir_name = protname + "_potcurves" ;

    blab3 << "Writing 2D potential curve " << curvedir_name.c_str() << " in GMT format in units of mV ...\n ";
    blab3 << "Writing 2D induced mobile charge density curve " << curvedir_name.c_str() << " in GMT format in unit charges per cubic A ...\n ";
    blab3 << "base point of curve    : (" << gmt_pt.x << ", " << gmt_pt.y << ", "<< gmt_pt.z << ")\n";
    blab3 << "normal vector of curve : (" << gmt_vec.x << ", " << gmt_vec.y << ", "<< gmt_vec.z << ")\n";
    blab3 << "grid spacing           :  " << gmt_space  << "\n";
    blab3 << "grid points            :  " << gmt_count  << "\n";

    if (do_2ndplane){
     blab3 << "base point of value plane    : (" << gmt_pt2.x << ", " << gmt_pt2.y << ", "<< gmt_pt2.z << ")\n";
     blab3 << "normal vector of value plane : (" << gmt_vec2.x << ", " << gmt_vec2.y << ", "<< gmt_vec2.z << ")\n";
    }

    // make sure normal vector of second plane has unit length
    if (do_2ndplane) gmt_vec2 /= sqrt(gmt_vec2.x*gmt_vec2.x + gmt_vec2.y*gmt_vec2.y + gmt_vec2.z*gmt_vec2.z);

    Coord tmpvec, tmpvec2, ovec;
    Coord xvec, yvec; // perpendicular vectors in plane, x
                      // will be perpendicular to z by convention
    Coord yedgevec; // side of grid

    // find perpendicular vectors in plane to generate curve

    // make normal vector unit length
    gmt_vec /= sqrt((gmt_vec.x * gmt_vec.x) + (gmt_vec.y * gmt_vec.y) + (gmt_vec.z * gmt_vec.z));
    Coord gmt_nvec = gmt_vec;

    // check if the normal vector is the z-axis
    if (gmt_vec.x == 0 && gmt_vec.y == 0 && gmt_vec.z == 0){
     ::error("All normal vector components are zero.");
    }
    else if (gmt_vec.x == 0 && gmt_vec.y == 0){
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 1;
      yvec.z = 0;
    }
    // check if the normal vector is the x-axis
    else if (gmt_vec.y == 0 && gmt_vec.z == 0){
      xvec.x = 0;
      xvec.y = 1;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
    }
    // check if the normal vector is the y-axis
    else if (gmt_vec.x == 0 && gmt_vec.z == 0){
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
    }
    // we have a normal vector in the y-z plane
    else if (gmt_vec.x == 0){
      float tmpf1;
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      tmpvec = cross(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    // we have a normal vector in the x-z plane
    else if (gmt_vec.y == 0){
      float tmpf1;
      xvec.x = 0;
      xvec.y = 1;
      xvec.z = 0;
      tmpvec = cross(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    // we have a normal vector in the x-y plane
    else if (gmt_vec.z == 0){
      float tmpf1;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
      tmpvec = cross(yvec, gmt_vec);
      xvec = tmpvec / abs_value(tmpvec);
    }
    // or an arbitrary axis apart from the main axes with x and y-component
    // make xvec perpendicular to z-axis
    else if (gmt_vec.x != 0 && gmt_vec.y != 0){
      float tmpf1, tmpf2, tmpf3;
      tmpf1 = -gmt_vec.y/gmt_vec.x;
      tmpf2 = sqrt((1 + (tmpf1 * tmpf1)));
      tmpf1 /= tmpf2;
      tmpf2 = 1/tmpf2;
      xvec.x = tmpf1;
      xvec.y = tmpf2;
      xvec.z = 0;
      tmpvec = cross(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    else {
      ::error(argv[0], "unexpected case in construction of grid vectors\n");
    }
    
    // directionality of curve vectors
    if (xvec.x < 0) xvec *= -1;
    if (xvec.y < 0) xvec *= -1;
    if (yvec.z < 0) yvec *= -1;

    // make vectors in length spacing

    xvec    *= gmt_space;
    yvec    *= gmt_space;
    gmt_vec *= gmt_space;

    yedgevec = gmt_count * yvec;

    // this factor multiplied by the potential should give unit charge density per cubic A
    float qindu_fac =        -2 * 1e-27 * PhysCond::get_Avogadro_Number()
                                * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                              /  (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());

    // write 2D curves in slice planes along normal vector
    // now we are ready to write the GMT curve

    // dont know safe & direct way to check if curve directory exists,
    // remove if it exists
    // (re)create directory
    // FIXME ! that's ugly indeed

    //system ("perl -p -e \'if (!-d curvedir_name){mkdir(curvedir_name)}\'");

    // this gives all slices along the normal vector
    //for (int k=0; k<gmt_count; k+=4){

    // need only center slice

    int centerk = int((gmt_count - 1) * 0.5);

    for (int k=centerk; k<=centerk; k++){
    //int mid = (int) 0.5*(gmt_count-1);
    //for (int k=mid; k<mid+1; k++){
      // central point of plane slides along normal vector

      ovec   = gmt_pt - (0.5 * (gmt_count - 1) * xvec)
                      - (0.5 * (gmt_count - 1) * yvec)
                      - ((k - (0.5 *(gmt_count - 1))) * gmt_vec);
      /*
      ovec.x = gmt_pt.x - (0.5 * (gmt_count - 1) * xvec.x)
                        - (0.5 * (gmt_count - 1) * yvec.x);
      ovec.y = gmt_pt.y - (0.5 * (gmt_count - 1) * xvec.y)
                        - (0.5 * (gmt_count - 1) * yvec.y);
      ovec.z = gmt_pt.z - (0.5 * (gmt_count - 1) * xvec.z)
                        - (0.5 * (gmt_count - 1) * yvec.z);
      */

      blab3 << "X-vector of curve      : (" << xvec.x << ", " << xvec.y << ", "<< xvec.z << ")\n";
      blab3 << "Y-vector of curve      : (" << yvec.x << ", " << yvec.y << ", "<< yvec.z << ")\n";
      blab3 << "Origin of curve        : (" << ovec.x << ", " << ovec.y << ", "<< ovec.z << ")\n";

      // 2D curve
      std::stringstream fmt1;
      fmt1 << k;

      string curve_name  = fmt1.str() + DOT + "curve";
      string curve_name2 = fmt1.str() + DOT + "charge" + DOT + "curve";
      ofstream of (curve_name.c_str());
      ofstream of2(curve_name2.c_str());

      // Could I open 2D file for writing?
      if (!of.good()){
        ::error(argv[0], "main: failed to open GMT curve for writing, ",
        curve_name.c_str());
      }
      if (!of2.good()){
        ::error(argv[0], "main: failed to open GMT curve for writing, ",
        curve_name2.c_str());
      }

      // 1D curves

      float lx = (-0.5 * (gmt_count - 1) * gmt_space); // coordinates in the plane,
                                                       // tmpvec holds the real coordinates
      tmpvec = ovec;

      for (int i=0; i<gmt_count; i+=4){
        // 1D curve
        std::stringstream fmt2;
        fmt2 << i;
        string d_curve_name(fmt1.str() + DOT + fmt2.str() + DOT + "txt");
        string d_curve_name2(fmt1.str() + DOT + fmt2.str() + DOT + "charge" + DOT + "txt");
        ofstream d_of (d_curve_name.c_str());
        ofstream d_of2 (d_curve_name2.c_str());
        // Could I open 1D file for writing?
        if (!d_of.good()){
          ::error(argv[0], "main: failed to open 1D curve for writing, ",
          d_curve_name.c_str());
        }
        if (!d_of2.good()){
          ::error(argv[0], "main: failed to open 1D curve for writing, ",
          d_curve_name2.c_str());
        }
        float ly = (-0.5 * (gmt_count - 1) * gmt_space);
        for (int j=0; j<gmt_count; j++){
          float dsq, mdsq, tmpphi;
          bool isacc = true;
          Coord dv;
          if (do_2ndplane){
           // get vector to project tmpvec in plot plane into slice plane
           // dotproduct(point_in_plane - pt_to_project, nvec)
           tmpvec2 = gmt_vec2 * (((gmt_pt2.x - tmpvec.x) * gmt_vec2.x)
                                +((gmt_pt2.y - tmpvec.y) * gmt_vec2.y)
                                +((gmt_pt2.z - tmpvec.z) * gmt_vec2.z));
           tmpvec2 += tmpvec;
           tmpphi   = phi.value(tmpvec2);
           // use get_cuberep of Electrolyte Environment with one point grid to check ion accessibility
           CubeLatSpec tmpcls(1,1.0,tmpvec2);
           ElyCubeRep ecr(ely.get_cuberep(tmpcls));
           ecr.declare_undefined();
           if (ecr[0] == 0) isacc = false;
          } else {
           tmpphi = phi.value(tmpvec);
           // use get_cuberep of Electrolyte Environment with one point grid to check ion accessibility
           CubeLatSpec tmpcls(1,1.0,tmpvec2);
           ElyCubeRep ecr(ely.get_cuberep(tmpcls));
           ecr.declare_undefined();
           if (ecr[0] == 0) isacc = false;
          }

          d_of                 << ly << "   " << (tmpphi * PhysCond::get_voltconv() * 1000) << endl;

          if (!isacc)
           d_of2                << ly << "   " << 0 << endl;
          else
           d_of2                << ly << "   " << (tmpphi * qindu_fac) << endl;

          tmpvec.x += yvec.x;
          tmpvec.y += yvec.y;
          tmpvec.z += yvec.z;
          ly += gmt_space;
        }
        tmpvec.x -= yedgevec.x;
        tmpvec.y -= yedgevec.y;
        tmpvec.z -= yedgevec.z;
        tmpvec.x += xvec.x;
        tmpvec.y += xvec.y;
        tmpvec.z += xvec.z;
        lx += gmt_space;
        // is 1D filehandle still open?
        if (!d_of.good()){
          ::error(argv[0], "main: failed to write 1D curve, ",
          d_curve_name.c_str());
        }
        if (!d_of2.good()){
          ::error(argv[0], "main: failed to write 1D curve, ",
          d_curve_name2.c_str());
        }
        d_of.close();
        d_of2.close();
      }


      // 2D curve

      lx = (-0.5 * (gmt_count - 1) * gmt_space); // coordinates in the plane,
                                                 // tmpvec holds the real coordinates
      tmpvec = ovec;

      for (int i=0; i<gmt_count; i++){
        float ly = (-0.5 * (gmt_count - 1) * gmt_space);
        for (int j=0; j<gmt_count; j++){
          float dsq, mdsq, tmpphi, tmpchg;
          bool isacc = true;
          Coord dv;
          if (do_2ndplane){
           // get vector to project tmpvec in plot plane into slice plane
           // dotproduct(point_in_plane - pt_to_project, nvec)
           tmpvec2 = gmt_vec2 * (((gmt_pt2.x - tmpvec.x) * gmt_vec2.x)
                                +((gmt_pt2.y - tmpvec.y) * gmt_vec2.y)
                                +((gmt_pt2.z - tmpvec.z) * gmt_vec2.z));
           tmpvec2 += tmpvec;
           tmpphi   = phi.value(tmpvec2);
           // use get_cuberep of Electrolyte Environment with one point grid to check ion accessibility
           CubeLatSpec tmpcls(1,1.0,tmpvec2);
           ElyCubeRep ecr(ely.get_cuberep(tmpcls));
           ecr.declare_undefined();
           if (ecr[0] == 0) isacc = false;
          } else {
           tmpphi = phi.value(tmpvec);
           // use get_cuberep of Electrolyte Environment with one point grid to check ion accessibility
           CubeLatSpec tmpcls(1,1.0,tmpvec2);
           ElyCubeRep ecr(ely.get_cuberep(tmpcls));
           ecr.declare_undefined();
           if (ecr[0] == 0) isacc = false;
          }

          of    << lx << "   " << ly << "   " << (tmpphi * PhysCond::get_voltconv() * 1000) << endl;

          if (tmpphi == 0 || !isacc)
           tmpchg = 0;
          else
           tmpchg = tmpphi * qindu_fac;
          of2    << lx << "   " << ly << "   " << tmpchg << endl;
          tmpvec.x += yvec.x;
          tmpvec.y += yvec.y;
          tmpvec.z += yvec.z;
          ly += gmt_space;
        }
        tmpvec.x -= yedgevec.x;
        tmpvec.y -= yedgevec.y;
        tmpvec.z -= yedgevec.z;
        tmpvec.x += xvec.x;
        tmpvec.y += xvec.y;
        lx += gmt_space;
      }

      // is 2D filehandle still open?
      if (! of.good()){
        ::error(argv[0], "main: failed to write GMT 2D curve, ",
        curve_name.c_str());
      }
      if (! of2.good()){
        ::error(argv[0], "main: failed to write GMT 2D curve, ",
        curve_name2.c_str());
      }
      of.close();
      of2.close();
    }

    times(&apotgmt);
    blab3 << "Writing electrostatic potential to GMT runtime: " << ((float) (apotgmt.tms_utime-bpotgmt.tms_utime))/ticks << " s user " << ((float) (apotgmt.tms_stime-bpotgmt.tms_stime))/ticks <<  " s system " << endl;
  }

  times(&end);
  cout << "Total runtime: " << ((float) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((float) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;
  
  return 0;
} // END main
