/* Calculate the "Solvent Accessible Volume" from a PQR file 
   and write it into a file.


   Don provides already methods to read and write "sausages"
   in both ascii and binary (MEAD/SolvAccVol.cc).

   Dielectric Environments can be constructed either using
   "AtomSet"s or "SolvAccVol"s (MEAD/DielByAtoms.cc). 
   Assigning dielectric constants is
   done by calculating a grid and assigning a value dependent if
   the gridpoint is interior or exterior to the volume. This should
   not be time critical.

   The "sausages" contain analytical representations of the solvent
   accessible volume, while MEAD uses internally a cubic (grid based)
   representation.

   The probe radius can be set for each surface independently.

   Options: solrad, blab1, blab2, blab3 as other MEAD programs.
   Option ascii for output format in ASCII, binary else.

   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Timm Essigke, 2005

   /usr/bin/g++-3.4 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 pqr2SolvAccVol.cc
   /usr/bin/g++-3.4 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o pqr2SolvAccVol pqr2SolvAccVol.o -L../../libmead -lmead  -lm

*/



#include "MEAD/SolvAccVol.h"
#include "MEAD/PhysCond.h"
#include "MEAD/AtomSet.h"
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <fstream>
using std::ifstream;
#include <string>
using std::string;

int main(int argc, char* argv[])
{
  string molname; // name of the input and output file
  float radius = PhysCond::get_solrad ();   // probe radius
  bool ascii = 0; // Use ASCII output? Else, binary
  
  // Parse the command line
  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+1==argc) {  // These is the last one, must be molname
      molname = argv[iarg];
    }
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -solrad");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -solrad option");
      radius = f;
    }
    else if ((string) argv[iarg] == (string) "-ascii") {
      ascii = 1;
    }
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

   if (molname == (string) "")
    error("main: ERROR, molname not given on command line");

   AtomSet a(molname);
   a.read();
   SolvAccVol s(a, radius);
   s.anal_calc(); // this is a bit costly for big molecules
   if (ascii) {
     s.write_top_in_ascii(molname + ".txt");
   }
   else {
     s.write_top_in_binary(molname +".dat");
   }

   return 0;
}
