/* Create 2D curve of slice or whole molecule projection into a plane for use with GMT

   pqr2crv          -gmt ptx pty ptz vecx vecy vecz spacing gridpoints {pt2x pt2y pt2z vec2x vec2y vec2z} \
                     protname



   This program is based on MEAD by Don Bashford and licensed under
   the GPL.

   Thomas Ullmann, 2008

   /usr/bin/g++-3.4 -c -I. -I../.. -I.  -DUSE_EXCEPTIONS   -O2 my_membpot_solver.cc
   /usr/bin/g++-3.4 -I. -I../.. -I.  -DUSE_EXCEPTIONS    -O2  -o my_membpot_solver my_membpot_solver.o -L../../libmead -lmead  -lm

*/

#include "pqr2crv.h"

using namespace std;

int main(int argc, char* argv[])
{
  // Only for benchmarking
  struct tms start, end, // start and end of program
    bgmt, agmt;    // writing GMT curve
  long ticks = sysconf(_SC_CLK_TCK);
  times(&start);

  string protname;

  bool DoGMT = false;          // write a gmt compatible 2D square curve in a plane
                               // with normal vector gmt_vec, spacing gmt_space and
                               // edge length gmt_count around a center point gmt_pt
                               // the upper edge is perpendicular to z
  bool do_slice = false;       // trigger projection of slice of thickness gmt_slice 
                               // instead of whole molecule
  bool do_2ndplane = false;    // trigger value calculation in second plane followed
                               // by back-projection on first plane
  bool add_solrad  = false;

  int   gmt_count = 0;         // gridpoints
  float gmt_space = 0;         // grid spacing
  float gmt_slice = numeric_limits<float>::max(); // slice thickness
  float solrad = 1.4;          // solvent radius
  Coord gmt_pt,  gmt_vec;      // base point and normal vector of plane 2
  Coord gmt_pt2, gmt_vec2;     // base point and normal vector of plane 2

  ////////////// Parse the command line /////////////////////////////////
  for (int iarg=1; iarg<argc; ++iarg) {
    if (iarg+1==argc) {  // This is the last one, must be protname
      protname = argv[iarg];
    }
    else if ((string) argv[iarg] == (string) "-slice") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -slice option");
      gmt_slice = f * 0.5;
      do_slice = true;
    }
    else if ((string) argv[iarg] == (string) "-solrad") {
      if (iarg+1>=argc) error ("main: ERROR, not enough args after -epsin1");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
        error ("main: FAILED trying to read value after -slice option");
      solrad = f;
    }
    // process the form -gmt 7 float 1 int
    else if ((string) argv[iarg] == (string) "-gmt") {
      DoGMT = true;
      if (iarg+8>=argc)
       error ("main: ERROR, not enough args after -gmt");
      istringstream ist(argv[++iarg]);
      float f;
      ist >> f;
      if (!(ist.good() || ist.eof()))
	error ("main: FAILED trying to read first value after -gmt opt.(pt.x)");
      gmt_pt.x = f;
      istringstream ist2(argv[++iarg]);
      ist2 >> f;
      if (!(ist2.good() || ist2.eof()))
	error ("main: FAILED trying to read second value after -gmt opt. (pt.y)");
      gmt_pt.y = f;
      istringstream ist3(argv[++iarg]);
      ist3 >> f;
      if (!(ist3.good() || ist3.eof()))
	error ("main: FAILED trying to read third value after -gmt opt. (pt.z)");
      gmt_pt.z = f;
      istringstream ist4(argv[++iarg]);
      ist4 >> f;
      if (!(ist4.good() || ist4.eof()))
	error ("main: FAILED trying to read fourth value after -gmt opt. (vec.x)");
      gmt_vec.x = f;
      istringstream ist5(argv[++iarg]);
      ist5 >> f;
      if (!(ist5.good() || ist5.eof()))
	error ("main: FAILED trying to read fith value after -gmt opt. (vec.y)");
      gmt_vec.y = f;
      istringstream ist6(argv[++iarg]);
      ist6 >> f;
      if (!(ist6.good() || ist6.eof()))
	error ("main: FAILED trying to read sixth value after -gmt opt.(vec.z)");
      gmt_vec.z = f;
      istringstream ist7(argv[++iarg]);
      ist7 >> f;
      if (!(ist7.good() || ist7.eof()))
	error ("main: FAILED trying to read seventh value after -gmt opt.(space)");
      gmt_space = f;
      int i;
      istringstream ist8(argv[++iarg]);
      ist8 >> i;
      if (!(ist8.good() || ist8.eof()))
	error ("main: FAILED trying to read eighth value after -gmt opt.(count)");
      gmt_count = i;
      // try to read 6 more values giving the base point and normal vector of a second plane
      // values are calculated in this second plne then and projected onto the first plane
      istringstream ist9(argv[++iarg]);
      ist9 >> f;      
      do_2ndplane = ((ist9.good() || ist9.eof()));
      if(do_2ndplane){
        if (iarg+5>=argc)
         error ("main: ERROR, not enough args after -gmt primary arguments, 8 or 14 values expected.");
        gmt_pt2.x = f;
        istringstream ist10(argv[++iarg]);
        ist10 >> f;
        if (!(ist10.good() || ist10.eof()))
 	 error ("main: FAILED trying to read 10th value after -gmt opt. (pt2.y)");
        gmt_pt2.y = f;
        istringstream ist11(argv[++iarg]);
        ist11 >> f;
        if (!(ist11.good() || ist11.eof()))
  	 error ("main: FAILED trying to read 11th value after -gmt opt. (pt2.z)");
        gmt_pt2.z = f;
        istringstream ist12(argv[++iarg]);
        ist12 >> f;
        if (!(ist12.good() || ist12.eof()))
 	 error ("main: FAILED trying to read 12th value after -gmt opt. (vec2.x)");
        gmt_vec2.x = f;
        istringstream ist13(argv[++iarg]);
        ist13 >> f;
        if (!(ist13.good() || ist13.eof()))
         error ("main: FAILED trying to read 13th value after -gmt opt. (vec2.y)");
        gmt_vec2.y = f;
        istringstream ist14(argv[++iarg]);
        ist14 >> f;
        if (!(ist14.good() || ist14.eof()))
  	 error ("main: FAILED trying to read 14th value after -gmt opt.(vec2.z)");
        gmt_vec2.z = f;      
      } else {iarg--;}
    }
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-add_solrad") {
      add_solrad = true;
    }
    // Process -blab flags
    else if ((string) argv[iarg] == (string) "-blab1") {
      blab1pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab2") {
      blab1pt = &cout;
      blab2pt = &cout;
    }
    else if ((string) argv[iarg] == (string) "-blab3") {
      blab1pt = &cout;
      blab2pt = &cout;
      blab3pt = &cout;
    }
    else {
      cerr << "main: ERROR command option, " << argv[iarg]
        << ", not recognized" << endl;
      error();
    }
  }

  /////////////// Checks & Statistics ////////////////////////////////

  if (protname == (string) "")
    error("main: ERROR, protname not given on command line");
  else if (!DoGMT){
    cerr << "Syntax " << argv[0] << " -gmt base.x base.y base.z normalvec.x normalvec.y normalvec.z spacing griddim {-slice thickness} molname" << endl << endl;
    ::error("No parameters for GMT curve given on commandline.");
  }
  else if (solrad < 0.0)
    solrad *= -1;
  else if (gmt_slice < 0.0)
    gmt_slice *= -1;
  else if (gmt_space < 0.0)
    gmt_space *= -1;
  else if (gmt_count < 0.0)
    gmt_count *= -1;
  
  /////// Print summary of input info from the command line: ///////

  cout << "Starting " << argv[0] << " for molecule named " << protname << endl;
  if (blab3pt == &cout) cout << "Blab level set to " << 3 << endl;
  else if (blab2pt == &cout) cout << "Blab level set to " << 2 << endl;
  else if (blab1pt == &cout) cout << "Blab level set to " << 1 << endl;
  else cout << "No blab level set (so no blabbing)" << endl;

  ////////////////// Setting up calculation ////////////////////////

  // background charge set
  AtomSet a1(protname);
  a1.read();

  ////////////// write GMT 2D-curve of molecule shape //////////////

  if (DoGMT) {

    times(&bgmt);

    blab3 << "Writing 2D molecular shape in GMT format in units of Angstrom ...\n ";
    blab3 << "base point of curve    : (" << gmt_pt.x << ", " << gmt_pt.y << ", "<< gmt_pt.z << ")\n";
    blab3 << "normal vector of curve : (" << gmt_vec.x << ", " << gmt_vec.y << ", "<< gmt_vec.z << ")\n";
    blab3 << "grid spacing           :  " << gmt_space  << "\n";
    blab3 << "grid points            :  " << gmt_count  << "\n";
    blab3 << "solvent radius         :  " << solrad  << "\n";

    if (do_2ndplane){
     blab3 << "base point of value plane    : (" << gmt_pt2.x << ", " << gmt_pt2.y << ", "<< gmt_pt2.z << ")\n";
     blab3 << "normal vector of value plane : (" << gmt_vec2.x << ", " << gmt_vec2.y << ", "<< gmt_vec2.z << ")\n";    
    }
    
    // make sure normal vector of second plane has unit length
    if (do_2ndplane) gmt_vec2 /= abs_value(gmt_vec2);

    Coord tmpvec, ovec;
    Coord xvec, yvec; // perpendicular vectors in plane, x
                      // will be perpendicular to z by convention
    Coord yedgevec; // side of grid


    // make normal vector unit length
    gmt_vec /= sqrt((gmt_vec.x * gmt_vec.x) + (gmt_vec.y * gmt_vec.y) + (gmt_vec.z * gmt_vec.z));
    Coord gmt_nvec = gmt_vec;

    // check if there is a special case for the normal vector
    if (gmt_vec.x == 0 && gmt_vec.y == 0 && gmt_vec.z == 0){
     ::error("All normal vector components are zero.");
    }
    // check if the normal vector is the z-axis
    else if (gmt_vec.x == 0 && gmt_vec.y == 0){
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 1;
      yvec.z = 0;
    }
    // check if the normal vector is the x-axis
    else if (gmt_vec.y == 0 && gmt_vec.z == 0){
      xvec.x = 0;
      xvec.y = 1;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
    }
    // check if the normal vector is the y-axis
    else if (gmt_vec.x == 0 && gmt_vec.z == 0){
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
    }
    // we have a normal vector in the y-z plane
    else if (gmt_vec.x == 0){
      float tmpf1;
      xvec.x = 1;
      xvec.y = 0;
      xvec.z = 0;
      tmpvec = cross_p(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    // we have a normal vector in the x-z plane
    else if (gmt_vec.y == 0){
      float tmpf1;
      xvec.x = 0;
      xvec.y = 1;
      xvec.z = 0;
      tmpvec = cross_p(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    // we have a normal vector in the x-y plane
    else if (gmt_vec.z == 0){
      float tmpf1;
      yvec.x = 0;
      yvec.y = 0;
      yvec.z = 1;
      tmpvec = cross_p(yvec, gmt_vec);
      xvec = tmpvec / abs_value(tmpvec);
    }
    // or an arbitrary axis apart from the main axes with x and y-component
    // make xvec perpendicular to z-axis
    else if (gmt_vec.x != 0 && gmt_vec.y != 0){
      float tmpf1, tmpf2, tmpf3;
      tmpf1 = -gmt_vec.y/gmt_vec.x;
      tmpf2 = sqrt((1 + (tmpf1 * tmpf1)));
      tmpf1 /= tmpf2;
      tmpf2 = 1/tmpf2;
      xvec.x = tmpf1;
      xvec.y = tmpf2;
      xvec.z = 0;
      tmpvec = cross_p(xvec, gmt_vec);
      yvec = tmpvec / abs_value(tmpvec);
    }
    else {
      ::error(argv[0], "unexpected case in construction of grid vectors\n");
    }
    
    // directionality of curve vectors
    if (xvec.x < 0) xvec *= -1;
    if (xvec.y < 0) xvec *= -1;
    if (yvec.z < 0) yvec *= -1;
    
    // make vectors in length spacing
    xvec    *= gmt_space;
    yvec    *= gmt_space;
    gmt_vec *= gmt_space;

    yedgevec = gmt_count * yvec;

    // write 2D curves in slice planes along normal vector
    // now we are ready to write the GMT curve

    // need only center slice for the moment

    int centerk = int((gmt_count - 1) * 0.5);
    AtomSet unprojected_slice_set;

    for (int k=centerk; k<=centerk; k++){
      //int mid = (int) 0.5*(gmt_count-1);
      //for (int k=mid; k<mid+1; k++){
      // central point of plane slides along normal vector

      ovec   = gmt_pt - (0.5 * ((float)gmt_count - 1) * xvec)
                      - (0.5 * ((float)gmt_count - 1) * yvec)
                      - (((float)k - (0.5 *((float)gmt_count - 1))) * gmt_vec);

      blab3 << "X-vector of curve      : (" << xvec.x << ", " << xvec.y << ", "<< xvec.z << ")\n";
      blab3 << "Y-vector of curve      : (" << yvec.x << ", " << yvec.y << ", "<< yvec.z << ")\n";
      blab3 << "Origin of curve        : (" << ovec.x << ", " << ovec.y << ", "<< ovec.z << ")\n";

      // collect coordinates of atoms projected into this plane
      AtomSet slice_set;
      double dist = 0;
      long int num_pa = 0;
      for (AtomSet::const_iterator aiter = a1.begin(); aiter != a1.end(); aiter++){
        if (do_slice){
         if (do_2ndplane)
          dist = get_dist_p_plane(aiter->second.coord,gmt_pt2,gmt_vec2);
         else
          dist = get_dist_p_plane(aiter->second.coord,gmt_pt,gmt_nvec);
         if (dist < 0) dist = -dist; 
         if (dist > gmt_slice)
           continue;
        }
        if (blab3pt == &cout) unprojected_slice_set.insert(aiter->second);
        Atom tmpatom(aiter->second);
        tmpatom.coord = project_on_plane(gmt_nvec,aiter->second.coord,ovec);
        if (!slice_set.contains(aiter->first))
         slice_set.insert(tmpatom);
        num_pa++;
      }

      blab3 << "projected   " << num_pa << " of " << a1.size() << " atoms into plane " << k << endl;


      // 2D curve
      std::stringstream fmt1;
      fmt1 << k;

      string curve_name  = fmt1.str() + DOT + "shape.curve";
      ofstream of (curve_name.c_str());

      // Could I open 2D file for writing?
      if (!of.good()){
        ::error(argv[0], "main: failed to open GMT curve for writing, ",
        curve_name.c_str());
      }

      // 2D curve

      short int **result;
      if (add_solrad){
       result = new short int*[gmt_count];
       for (int i=0; i<gmt_count; i++){
        result[i] = new short int[gmt_count];
       }
      }

      // grid coordinates in the plane,
      double lx = (-0.5 * ((double)gmt_count - 1) * gmt_space);
      // tmpvec holds the real coordinates
      tmpvec = ovec;
      short int inside = 0;

      int num_inside = 0;
      AtomSet tmpatomset;
      int atcnt = 0;
      for (int i=0; i<gmt_count; i++){
        double ly = (-0.5 * ((double)gmt_count - 1) * gmt_space);
        for (int j=0; j<gmt_count; j++){
          // this is more expensive than incremental addition of xvec, yvec
          // but prevents skewed images caused by precision errors in case
          // of fine grid spacings and large grid lengths
          tmpvec = ovec + (i * xvec) + (j * yvec);
          inside = 0;          
          for (AtomSet::const_iterator aiter = slice_set.begin(); aiter != slice_set.end(); aiter++){
            if ((get_dist(aiter->second.coord,tmpvec) - aiter->second.rad - solrad) < gmt_space){
              inside = 1;
              num_inside++;
              break;
            }
          }
          if (blab3pt == &cout) {
            Atom tmpat(slice_set.begin()->second);
            atcnt++;
            tmpat.instance = atcnt;
            tmpat.coord = tmpvec;
            tmpat.charge = inside;
            tmpatomset.insert(tmpat);
          }
          if (!add_solrad){
           of << setw(8) << setprecision(2) << fixed << lx << "   " << ly << "   "
              << setw(2) << setprecision(0) << inside << endl;
          } else {
           result[i][j] = inside;
          }
/*
          tmpvec.x += yvec.x;
          tmpvec.y += yvec.y;
          tmpvec.z += yvec.z;
*/
          ly += gmt_space;
        }
/*
        tmpvec.x -= yedgevec.x;
        tmpvec.y -= yedgevec.y;
        tmpvec.z -= yedgevec.z;
        tmpvec.x += xvec.x;
        tmpvec.y += xvec.y;
        tmpvec.z += xvec.z;
*/
        lx += gmt_space;
      }
      
      if (blab3pt == &cout) {
        tmpatomset.write_pqr("plane.pqr");
        if (gmt_slice > 0 && do_2ndplane) unprojected_slice_set.write_pqr("slice.pqr");
      }

      if (add_solrad){
      
       // set all points within solrad of a solvent accessible point
       // to outside for better volume visualization --
       // whole atom volumes instead of volume occupied by atom centers
       const double solradsq = solrad * solrad;
       int solrad_in_grid = static_cast<int>(solrad / gmt_space) + 1;
       for (int i=0; i<gmt_count; i++){
        for (int j=0; j<gmt_count; j++){
         // do not check points already set to external
         if (result[i][j] != 1) continue;
         // check only points within a square of edge length
         // 2 * solrad arround the present point
         int startii = i - solrad_in_grid;
         if (startii < 0) startii = 0;
         int endii = i + solrad_in_grid;
         if (endii > gmt_count) endii = 0;
         int startjj = j - solrad_in_grid;
         if (startjj < 0) startjj = 0;
         int endjj = j + solrad_in_grid;
         if (endjj > gmt_count) endjj = 0;
         for (int ii=startii; ii<endii; ii++){
          for (int jj=startjj; jj<endjj; jj++){
           // check only around points originally set to external
           if (result[ii][jj] != 0) continue;
           double tmpdx = static_cast<double>((ii - i) * gmt_space);
           double tmpdy = static_cast<double>((jj - j) * gmt_space);
           double tmpdistsq = (tmpdx * tmpdx) + (tmpdy * tmpdy);
           // set new points of expansion to different value
           // to distinguish them from the original points
           if (tmpdistsq <= solradsq){
            result[i][j] = 2;
            goto END_II_LP;
           }
          }
         }
         END_II_LP:;
        }
       }
       // set all temporary results to the final value
       // and write data to curve file
       num_inside = 0;
       tmpvec = ovec;
       lx =  -0.5 * ((float)gmt_count - 1) * (float)gmt_space;
       for (int i=0; i<gmt_count; i++){
        float ly = -0.5 * ((float)gmt_count - 1) * (float)gmt_space;
        for (int j=0; j<gmt_count; j++){
         if        (result[i][j] == (short int)2){
          result[i][j] = (short int)0;
         }
         if (result[i][j] != (short int)1){
          result[i][j] = 0;
         } else {
          num_inside++;
         }         
         of << setw(8) << setprecision(2) << fixed << lx << "   " << ly << "   " << setw(2) << setprecision(0) << result[i][j] << endl;
/*
         tmpvec.x += yvec.x;
         tmpvec.y += yvec.y;
         tmpvec.z += yvec.z;
*/
         ly += gmt_space;         
        }
/*
        tmpvec.x -= yedgevec.x;
        tmpvec.y -= yedgevec.y;
        tmpvec.z -= yedgevec.z;
        tmpvec.x += xvec.x;
        tmpvec.y += xvec.y;
        tmpvec.z += xvec.z;
*/
        lx += gmt_space;
       }

       // deallocate temporary array
       for (int i=0; i<gmt_count; i++){
        delete [] result[i];
       }
       delete [] result;
      }

      // is 2D filehandle still open?
      if (! of.good()){
        ::error(argv[0], "main: failed to write GMT 2D curve, ",
        curve_name.c_str());
      }
      of.close();

      blab3 << "Found   " << num_inside << " of total " << gmt_count*gmt_count << " points within molecule in plane " << k << endl;

    }

    times(&agmt);
    blab3 << "Writing molecule shape to GMT runtime: " << ((float) (agmt.tms_utime-bgmt.tms_utime))/ticks << " s user " << ((float) (agmt.tms_stime-bgmt.tms_stime))/ticks <<  " s system " << endl;
  }

  times(&end);
  cout << "Total runtime: " << ((float) (end.tms_utime-start.tms_utime))/ticks << " s user " << ((float) (end.tms_stime-start.tms_stime))/ticks <<  " s system " << endl;

  return 0;
} // END main



////////////////////////// geometry functions ////////////////////////// 

// i,j,k,l, target torsion angle
inline Coord set_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const double t)
{  
  Coord a(k-j);
  Coord b(l-k);
  double d(abs_value(a));
  a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  double tmptor = get_tor(i,j,k,l) - t;
  double cos_t(cos(tmptor));
  double sin_t(sin(tmptor));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l, change in torsion angle
inline Coord change_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const double t)
{
  Coord a(k-j);
  Coord b(l-k);
  double d(abs_value(a));
  if (d==0)
   ::error("Error in get_tor(): points j and k have identical coordinates.");
   a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  double cos_t(cos(-t));
  double sin_t(sin(-t));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l 
inline double get_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l)
{
 Coord a(get_normal(i,j,k));
 Coord b(get_normal(j,k,l));
 double dp = dot_p(a,b);
 if      ((1-dp) < 1.0E-14) return 0;
 else if ((1+dp) < 1.0E-14) return pi;
 double t(acos(dp));
 double v(dot_p(k-j,cross_p(a,b)));
 if (v < (double) 0) t *= -1;
 return t;
}

// i,j,k 
inline double get_angle(const Coord& i, const Coord& j, const Coord& k)
{
  Coord a(i-j); a /= abs_value(a);
  Coord b(k-j); b /= abs_value(b);
  return acos(dot_p(a,b));
}

// point, point
inline double get_dist(const Coord& a, const Coord& b)
{return sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// point, point
inline double get_sqdist(const Coord& a, const Coord& b)
{return ((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// vector
inline double abs_value(const Coord& a)
{return sqrt(a.x*a.x + a.y*a.y + a.z*a.z);}

// vector, vector
inline double dot_p(const Coord& a, const Coord& b)
{return a.x*b.x + a.y*b.y + a.z*b.z;}
  
// vector, vector
inline Coord cross_p(const Coord& a, const Coord& b)
{Coord prod(a.y*b.z - a.z*b.y, b.x*a.z - a.x*b.z, a.x*b.y - b.x*a.y); return prod;}

// point in plane, point in plane, point in plane
inline Coord get_normal (const Coord& i, const Coord& j, const Coord& k)
{Coord n(cross_p(j-i,k-i)); n /= abs_value(n); return n;}

// unit normal vector of plane, point to project, point in plane
inline Coord project_on_plane (const Coord& n, const Coord& a, const Coord& b)
{return Coord(a - (n * get_dist_p_plane(a, b, n)));}

// point, point in plane, normal vector
inline double get_dist_p_plane (const Coord& a, const Coord& b, const Coord& n)
{return dot_p(a-b,n);}
 
// point, point on line, point on line
inline double get_dist_p_line (const Coord& a, const Coord& b, const Coord& c)
{return abs_value(cross_p(c-b,b-a)) / abs_value(c-b);}

inline int get_tor_bin(const double z)
{return static_cast<int>((floor(((z*(180/3.141593)) + 5) / 10) * 10));} 

// pqr2crv.h ends here
