#!/bin/tcsh

## set meaddir manually
#set meaddir="${HOME}/my_mead"

if    ( -d ${HOME}/my_mead.git ) then
 set meaddir="${HOME}/my_mead.git"
else if ( -d "/scratch/data/ullmannt/svn/mead/branches/ullmannt/current" ) then
 set meaddir="/scratch/data/ullmannt/svn/mead/branches/ullmannt/current"
else if ( -d "/sw/sci/app/mead_64/my_mead_09" ) then
 set meaddir="/sw/sci/app/mead_64/my_mead_09"
else if ( -d "/scratch/btbs08/my_mead_09" ) then
 set meaddir="/scratch/btbs08/my_mead_09"
else
 echo "Could not find MEAD directory, exiting ..."
 exit
endif

# added 4 colums to pqr format: conf, chain, site, inst

# QM region is empty (could also be omitted)
echo '' > qm_region.dat
# MM region, whole protein here
if ( -e mm_region.pqr ) rm -f mm_region.pqr
ln -s d.pqr mm_region.pqr

# ionic_str1/2 ionic strength in upper/lower phase, set equal here,
# ions are permitted to penetrate the headgroup region also
set ionic_str="0.15"
# only read if sidechain rotamers are requested via molname.rot
# prefix/name of Squirell backbone dependent rotamer library file of Roland Dunbrack
# if ${rotlib}.dat is found boost serialized binary is read (faster),
# otherwise ASCII file ${rotlib}.txt is read
set rotlib="rotlib"
# charmm topology and parameter file, my version all residues,
# bond order BOND/DOUBLE used (for topology determination via CharmmParaResTop)
set chmtop="top.inp"
set chmpar="par.inp"
# intrinsic energy cutoff. instances with
# intrinsic energy > 50 kcal + that of the lowest energy instance of the residue
# are discarded, mainly thought to remove clashing rotamers
set filter_ctof=50
set filter_ctof_gs=10
# MEAD path for solvers in job script,
# meant for convenience if solvers are run on a remote system/cluster
#set meadpath="/opt/btbs/mead/my_mead_07"
set meadpath=${meaddir}
############ the preprocessing run - calculate MM energies, set up electrostatic calculations

## options
# epsext        dielectric constant of solvent region
# epshead       dielectric constant of membrane headgroup region
# epscore       dielectric constant of membrane core region
# epsin1        dielectric constant of inner region 1 (intended to be used for QM region)
# epsin2        dielectric constant of inner region 2 (intended to be used for normal protein region)
# epsin3        dielectric constant of inner region 3 (intended to be used for cavities)
# eps1set       prefix of .pqr for inner region 1 (intended to be used for QM region)
# eps2set       prefix of .pqr for inner region 2 (intended to be used for normal protein region)
# eps3set       prefix of .pqr for inner region 3 (intended to be used for cavities)
# elycavset     prefix of .pqr for ion accessible cavities or depressions in the protein surface
#               reaching into the membrane boundaries (intended e.g. for gorges leading to a channel entrance)
# channelset    prefix of .pqr for a a channel reaching into the membrane boundaries
#               alternative for simple geometries -membhole
# membhole      -membhole rad x y (x, y are optional, default 0,0) radius, center coordinates
#               of cylindrical hole/channel/cavities through the membrane excluding volume occupied by
#               Atoms of eps(1|2|3)set, works seldom satisfactory for real, non-cylindrical shaped proteins
# capacitance   calculate capacitance of the system normalized to 1 V membrane potential,
#               capacitive energy term negligibly small under normal conditions of natural membrane potentials,
#               but depends quadratic on membrane potential also not needed if there is only one global conformation
#               CAUTION: depends on system size, use the same grid size for all global conformations

## input files
# file containing the SQUIRREL backbone dependent rotamer library
#  suffix .txt indicates ASCII, .dat indicates boost serialization binary (faster reading)
#  is only read if flexible sites are found in molname.rot
# molname.rot containing specification of the sites and their type
#   example:
#   site label (sitename-chain-resid)     site type
#   NTala-0-1       3
#   types
#   0 = non-titratable non-flexible (ignored as site put into background)
#   1 =     titratable     flexible
#   2 = non-titratable     flexible
#   3 =     titratable non-flexible
#   4 =     QM site
# molname.con containing specification of the site connectivities to the membrane sides
# only needed if membrane potential and/or chemical potential gradients are to be considered
# a site can only be connected to one membrane side in the current model!
#   example:
#   site label (sitename-chain-resid)     connectivity (same definition as in GMCT)
#   NTala-0-1       1
#   1 = inside
#   0 = outside

# for each sitename occurring in molname.rot a file sitename.est
#  files containing the model energies, definition of ligand binding numbers and atomic partial charges
#  of each form for titratable sites (e.g. protonated and deprotonated form of a protonatable site)

# grid definitions (optional):
#  if molname.ogm and molname.mgm exist, they are used to define the grids
#  otherwise grid centers/sizes are adjusted to molecule/site size,
#  grid sizes and centers are automatically determined for each site
#  grid size of innermost grid for each site is also adjusted in case of user defined grids
# detail of electrostatics calculations adjustable with autogenerated grids
# 0 = coarse, results will be afflicted with significant discretization error (errors in solvation energies can exceed ~ 1.5 kcal/mol)
# 1 = economic (similar to settings in old multiflex examples), results may partly be afflicted with significant discretization error (error in solvation energies up to 0.5 kcal/mol)
# 2 = fine, results should be fairly converged (default)
# 3 = very fine, ensures high-quality solvation and interaction energies also for membrane proteins
# 4 = ultra fine, should not result in significant differences relative to 3
# grid_detail = 2 or even better 3 (especially for membrane proteins) is the recommended setting
#  remember that the electrostatics is the basis of the following simulations and is done only once
# grid_detail = 0 should not be used for production calculations
set grid_detail=3

# molname.pqr
#  the molecule itself
# cav.pqr etc...
#  structure files with dummy atoms describing special solvent or ion accessible regions like cavities
# top.inp, par.inp
#  CHARMM topology and parameter files, must be consistent with naming in the molecule structure!

# use a pure continuum electrostatics model, more precisely
# skip all molecular mechanics terms for sites without conformational flexibility
# -skip_stat_mmg & skip_stat_mmint

## creates job scripts, grid and structure files for the MEAD calculations
#  in the directory qmpb_input
# writes a binary restart file for the postprocessing run molname.restart.dat
#  the restart file is not platform independent! (32/64 bit, endianness),
#  pre- and postprocessing run must be run on the same type of system, future versions may change this
# writes molname.mmint and molname.mmg containing the MM energy contributions
#  to intrinsic and interaction energies (no longer needed but kept for user reference)
# writes a QMPB input script gcem_input/molname.qmpb.in (not needed by GCEM but allows usage of QMPB)

if ( ! -f d.restart.dat ) then
 ${meaddir}/apps/gcem/gcem -blab2 -chmpar ${chmpar} -chmtop ${chmtop}  \
                           -rotlib ${rotlib} \
                           -filter_ctof ${filter_ctof} -filter_ctof_gs ${filter_ctof_gs} \
                           -meadpath ${meadpath} \
                           -epsext 80 -epsin1 4 -epsin2 4 \
                           -ionicstr ${ionic_str} \
                           -skip_stat_mmg -skip_stat_mmint \
                           -eps1set qm_region -eps2set mm_region \
                           -grid_detail ${grid_detail} d \
                           >& d.gcem.out
else
 # postprocessing run, read restart file & output of electrostatics solvers
 #  write GMCT input
 ${meaddir}/apps/gcem/gcem -p d >>& d.gcem.out
endif

if ( -f d.restart.dat && ! -d results ) then
 cd gcem_input/gcem
 # precompute solvent & ion inaccessible volumes to avoid recalculating them in each solver run
 ${meaddir}/apps/pqr2SolvAccVol/pqr2SolvAccVol epsin2_region
 rm -f background.iav.dat
 ${meaddir}/apps/pqr2IonAccVol/pqr2IonAccVol d.all && mv d.all.iav.dat background.iav.dat
 # you can set the number of threads to use at the beginning of the script in the hash %global,
 # otherwise the maximum number of (logical) processors is determined from /proc/cpuinfo
 ../../rst-gcem-here-mp.pl job.sh
 # as before but submit the jobs to a queueing system via qsub (options for PBS queueing system)
 # options, qsub command defined in head of script
 # ../../rst-gcem-cluster.pl job.sh
 cd ../..
endif
