#!/bin/tcsh

## set meaddir manually
#set meaddir="${HOME}/my_mead"

if    ( -d ${HOME}/my_mead.git ) then
 set meaddir="${HOME}/my_mead.git"
else if ( -d "/scratch/data/ullmannt/svn/mead/branches/ullmannt/current" ) then
 set meaddir="/scratch/data/ullmannt/svn/mead/branches/ullmannt/current"
else if ( -d "/sw/sci/app/mead_64/my_mead_09" ) then
 set meaddir="/sw/sci/app/mead_64/my_mead_09"
else if ( -d "/scratch/btbs08/my_mead_09" ) then
 set meaddir="/scratch/btbs08/my_mead_09"
else
 if ( ${meaddir} == "" ) then
  echo "Could not find MEAD directory, exiting ..."
  exit
 endif
endif

# added 4 colums to pqr format: conf, chain, site, inst

# QM region is empty (could also be omitted)
echo '' > qm_region.dat
# MM region, whole protein here
if ( -e mm_region.pqr ) rm -f mm_region.pqr
ln -s d.pqr mm_region.pqr

# create d.rot defining the sites
./make-rot.py d
# create d.rot defining the site connectivities
./make-con.py d

# visual inspection / OPM database -> membrane core region
# center 0, core thickness 30 A, headgroup region thickness 5 A
set membz="-15 15 -20 20"

# relative dielectric permittivities of the dielectric regions
# headgroup region
set epshead=20
# core region
set epscore=2
# solvent region
set epsext=80
# protein , QM region (empty here)
set epsin1=4
# protein, MM region (whole protein here)
set epsin2=4
# protein cavities, (would otherwise be filled by membrane dielectric if within membrane boundaries)
set epsin3=80
# inside 0 lower (z-coordinate) phase, 1 upper phase (default)
# needed for membrane potential
set inside=1
# ionic_str1/2 ionic strength in upper/lower phase, set equal here,
# ions are permitted to penetrate the headgroup region also
set ionic_str=0.15
# only read if sidechain rotamers are requested via molname.rot
# prefix/name of Squirell backbone dependent rotamer library file of Roland Dunbrack
# if ${rotlib}.dat is found boost serialized binary is read (faster),
# otherwise ASCII file ${rotlib}.txt is read
set rotlib="rotlib"
# charmm topology and parameter file, my version all residues,
# bond order BOND/DOUBLE used (for topology determination via CharmmParaResTop)
set chmtop="top.inp"
set chmpar="par.inp"
# intrinsic energy cutoff. instances with
# intrinsic energy > 100 kcal + that of the lowest energy instance of the residue
# are discarded, mainly thought to remove clashing rotamers
set filter_ctof=100
# MEAD path for solvers in job script,
# meant for convenience if solvers are run on a remote system/cluster
#set meadpath="/opt/btbs/mead/my_mead_07"
set meadpath=${meaddir}

############ the preprocessing run - calculate MM energies, set up electrostatic calculations

## options
# epsext        dielectric constant of solvent region
# epshead       dielectric constant of membrane headgroup region
# epscore       dielectric constant of membrane core region
# epsin1        dielectric constant of inner region 1 (intended to be used for QM region)
# epsin2        dielectric constant of inner region 2 (intended to be used for normal protein region)
# epsin3        dielectric constant of inner region 3 (intended to be used for cavities)
# eps1set       prefix of .pqr for inner region 1 (intended to be used for QM region)
# eps2set       prefix of .pqr for inner region 2 (intended to be used for normal protein region)
# eps3set       prefix of .pqr for inner region 3 (intended to be used for cavities)
# elycavset     prefix of .pqr for ion accessible cavities or depressions in the protein surface
#               reaching into the membrane boundaries (intended e.g. for gorges leading to a channel entrance)
# channelset    prefix of .pqr for a a channel reaching into the membrane boundaries
#               alternative for simple geometries -membhole
# membhole      -membhole rad x y (x, y are optional, default 0,0) radius, center coordinates
#               of cylindrical hole/channel/cavities through the membrane excluding volume occupied by
#               Atoms of eps(1|2|3)set, works seldom satisfactory for real, non-cylindrical shaped proteins
# capacitance   calculate capacitance of the system = accumulated charge in As per membrane potential in V,
#               capacitive energy term negligibly small under normal conditions of natural membrane potentials,
#               but depends quadratically on membrane potential also not needed if there is only one global conformation
#               CAUTION: depends on system size, use the same grid size for all global conformations

## input files
# file containing the SQUIRREL backbone dependent rotamer library
#  suffix .txt indicates ASCII, .dat indicates boost serialization binary (faster reading)
#  is only read if flexible sites are found in molname.rot
# molname.rot containing specification of the sites and their type
#   example:
#   site label (sitename-chain-resid)     site type
#   NTala-0-1       3
#   types
#   0 = non-titratable non-flexible (ignored as site put into background)
#   1 =     titratable     flexible
#   2 = non-titratable     flexible
#   3 =     titratable non-flexible
#   4 =     QM site
# molname.con containing specification of the site connectivities to the membrane sides
# only needed if membrane potential and/or chemical potential gradients are to be considered
# a site can only be connected to one membrane side in the current model!
#   example:
#   site label (sitename-chain-resid)     connectivity (same definition as in GMCT)
#   NTala-0-1       1
#   1 = inside
#   0 = outside

# for each sitename occurring in molname.rot a file sitename.est
#  files containing the model energies, definition of ligand binding numbers and atomic partial charges
#  of each form for titratable sites (e.g. protonated and deprotonated form of a protonatable site)

# grid definitions (optional):
#  if molname.ogm and molname.mgm exist, they are used to define the grids
#  otherwise grid centers/sizes are adjusted to molecule/site size,
#  grid sizes and centers are automatically determined for each site
#  grid size of innermost grid for each site is also adjusted in case of user defined grids
# detail of electrostatics calculations adjustable with autogenerated grids
# 0 = coarse, results will be afflicted with significant discretization error (errors in solvation energies can exceed ~ 1.5 kcal/mol)
# 1 = economic (similar to settings in old multiflex examples), results may partly be afflicted with significant discretization error (error in solvation energies up to 0.5 kcal/mol)
# 2 = fine, results should be fairly converged (default)
# 3 = very fine, ensures high-quality solvation and interaction energies also for membrane proteins
# 4 = ultra fine, should not result in significant differences relative to 3
# grid_detail = 2 or even better 3 (especially for membrane proteins) is the recommended setting
#  remember that the electrostatics is the basis of the following simulations and is done only once
# grid_detail = 0 should not be used for production calculations
set grid_detail=2

# molname.pqr
#  the molecule itself
# cav.pqr etc...
#  structure files with dummy atoms describing special solvent or ion accessible regions like cavities
# top.inp, par.inp
#  CHARMM topology and parameter files, must be consistent with naming in the molecule structure!

# use a pure continuum electrostatics model, more precisely
# skip all molecular mechanics terms for sites without conformational flexibility
# -skip_stat_mmg & skip_stat_mmint

## creates job scripts, grid and structure files for the MEAD calculations
#  in the directory gcem_input
# writes a binary restart file for the postprocessing run molname.restart.dat
#  the restart file is not platform independent! (32/64 bit, endianness),
#  pre- and postprocessing run must be run on the same type of system, future versions may change this
# writes molname.mmint and molname.mmg containing the MM energy contributions
#  to intrinsic and interaction energies (no longer needed but kept for user reference)

if ( ! -f d.restart.dat ) then
 ${meaddir}/apps/gcem/gcem -blab2 -chmpar ${chmpar} -chmtop ${chmtop}  \
                           -rotlib ${rotlib} \
                           -meadpath ${meadpath} \
                           -epsext ${epsext} -epshead ${epshead} -epscore ${epscore} \
                           -epsin1 ${epsin1} -epsin2 ${epsin2} -epsin3 ${epsin3} \
                           -ionicstr2 ${ionic_str} -ionicstr1 ${ionic_str} \
                           -membz ${membz} -inside ${inside} -capacitance \
                           -skip_stat_mmg -skip_stat_mmint  \
                           -eps1set qm_region -eps2set mm_region -eps3set cav \
                           -grid_detail ${grid_detail} d \
                           >& d.gcem.out
else
 # postprocessing run, read restart file & output of electrostatics solvers
 #  write GMCT input
 ${meaddir}/apps/gcem/gcem -p d >>& d.gcem.out
endif

if ( -f d.restart.dat && ! -d results ) then
 ############ after the preprocessing run - run electrostatic calculations
 cd gcem_input
  ## it's a good idea to run the commented out jobs at the beginning of the job script
  ## they generate binary files containing analytic descriptions
  ## of the dielectric regions and ion accessible regions
  ## the solvers can read these precomputed files instead of recalculating the volumes each time
  ## this speeds up the calculations considerably, especially in case of artifical volumes with
  ## densely spaced dummy atoms (e.g, cavities)!
  ## the binary files are not platform independent and thus
  ## must be created on the same kind of system as those running the solver jobs
  ## to be readable (32/64 bit, endianness)
  cd gcem
   # precompute solvent & ion inaccessible volumes to avoid recalculating them in each solver run
   rm -f background.iav.dat epsin2_region.dat cav.dat
   ${meaddir}/apps/pqr2SolvAccVol/pqr2SolvAccVol epsin2_region
   ${meaddir}/apps/pqr2SolvAccVol/pqr2SolvAccVol cav
   ${meaddir}/apps/pqr2IonAccVol/pqr2IonAccVol d.all && mv d.all.iav.dat background.iav.dat
  cd ..
  # if a membrane potential influence is to be considered
  #  (triggered by presence of a membrane and the molname.con file)
  cd membpot
   rm -f background.iav.dat epsin2_region.dat cav.dat
   ln -s ../gcem/*.dat .
   # you can set the number of threads to use at the beginning of the script in the hash %global,
   # otherwise the maximum number of (logical) processors is determined from /proc/cpuinfo
   tcsh job.sh
  cd ..
  cd gcem
   # run the solver jobs locally
   #  for each job check if the output file exists and if so check if it is complete
   #  before running the job
   # you can set the number of threads to use at the beginning of the script in the hash %global,
   # otherwise the maximum number of (logical) processors is determined from /proc/cpuinfo
   ../../rst-gcem-here-mp.pl job.sh
   # as before but submit the jobs to a queueing system via qsub (options for PBS queueing system)
   # options, qsub command defined in head of script
   # ../../rst-gcem-cluster.pl job.sh
  cd ..
 cd ..
endif


############ run GMCT programs

# make a directory where you put the directory "conf" and all GMCT input files (molname.rnames, molname.cnames etc.)
# create a setup file molname.setup (example d.setup) or
# use scripts like
#  run-gmct-pH-E-btrzx4.pl run-mct-coop-pH-E-btrzx4.pl run-mct-coop-pH-E-btrzx4.pl (for PBS queueing system)
# to distribute gmct/mct-coop to a cluster
# options are specified in the head of the scripts
# pH is the pH on the inside
# Eo is the redox potential
# pmf is delta pH + dpsi (Vmemb), (Vmemb is a fixed, additional membrane potential)
# pmftype gives specifies how the pmf is set up (pure dpH, pure dpsi or composite dpH+dpsi)
# if composite slopVmemb = dpsi/dpH specifies how large the contributions are
# beware: multidimensional titrations (pH vs. pmf, etc.) can take very long
# you need a rectangular grid with fixed increment for plotting with GMT (Generic Mapping Tools, google GMT SOEST)
# the conversion factors are used with such high precision to prevent problems in obtaining such a grid
# after conversion of the chemical potentials to the plotting units of pH / redox potential / pmf 
#
# the scripts are invoked with ./script.pl molname jobname_prefix
# each job gets his own directory and is submitted with the name named jobname_prefix.job_number
# the number of titration steps in each dimension per job can be varied with
# the options steps_in_script, redsteps_in_script, pmfsteps_in_script

############ VISUALIZATION examples

# the MEAD solvers can write OpenDX files for visualization in VMD, Pymol or the OpenDX viewer
# (useful for checking if everything is mapped as intended, examining the potential distribution)
# potential, dielectric constant, ionic strength, charge_distribution
# (also the ion distribution causing the membrane potential is written by my_membpot_solver then)
# -write_pot -write_eps -write_ely -write_rho
# x,y,z coordinates of grid center, grid spacing, number of grid points
# -x 0 -y 0 -z 1.881 -space 0.50 -count 201 
# the solvers can also write two-dimensional maps along slice planes for plotting
# (e.g. with GMT = Generic Mapping Tools, google GMT SOEST)
# -gmt 7 float 1 int {6 more floats}
# -gmt xcenter, ycenter, zcenter, xnormal, ynormal, znormal, grid spacing, number of grid points
#      {xcenter2, ycenter2, zcenter2, xnormal2, ynormal2, znormal2}
#      try to read 6 more values giving the base point and normal vector of a second plane
#      values are calculated in this second plane then and projected onto the first plane
#      (useful e.g. for a channel which is not perfectly normal to the membrane plane)
# example calls:
#   OpenDX + GMT potential,
# d.pqr is a prepared structure file with charges of a defined state,
# the structure with the most highly populated instance of each site can be obtained for example
# with get-curves-gmct from meaddir/apps/get-curves-gmct/ from the structures in gcem_input and a GMCT output file with the desired conditions
# d.gc.setup is an example setup file for this program, itraj 1 triggers the writing of structures in ./my_curves/traj
#  /scratch/data/ullmannt/svn/mead/branches/ullmannt/current/apps/my_memb_solver/my_memb_solver -blab3 -epsext 80 -epshead 20 -epscore 2 -epsin1 4 -epsin2 4 -epsin3 80 -ionicstr2 ${istr2} -ionicstr1 ${istr1} -membz -11.869 16.881 -16.869 20.631 -elycavset elycav -eps1set qm_region -eps2set mm_region -eps3set cav -x 0 -y 0 -z 1.881 -space 0.50 -count 201 -write_pot  -gmt 2.37832 -3.9938 1.881  -0.8 1.0 0  0.25 401  2.37832 -3.9938 1.881  -0.8 1.0 0.36 d d >& d.curve_tilt.out
#  /scratch/data/ullmannt/svn/mead/branches/ullmannt/current/apps/my_memb_solver/my_memb_solver -blab3 -epsext 80 -epshead 20 -epscore 2 -epsin1 4 -epsin2 4 -epsin3 80 -ionicstr2 ${istr2} -ionicstr1 ${istr1} -membz -11.869 16.881 -16.869 20.631 -elycavset elycav -eps1set qm_region -eps2set mm_region -eps3set cav -x 0 -y 0 -z 1.881 -space 0.50 -count 201 -write_pot -write_rho -gmt 2.37832 -3.9938 1.881  -0.8 1.0 0  0.25 401 d d >& d.curve.out
#   OpenDX membrane potential
#  /scratch/data/ullmannt/svn/mead/branches/ullmannt/current/apps/my_membpot_solver/my_membpot_solver -blab3 -epsext 80 -epshead 20 -epscore 2 -epsin1 4 -epsin2 4 -epsin3 80 -ionicstr2 ${istr2} -ionicstr1 ${istr1} -membpot 1.0  -membz -11.869 16.881 -16.869 20.631 -eps1set qm_region -eps2set mm_region -eps3set cav -fpt site -x 0 -y 0 -z 1.881 -space 0.50 -count 201 -write_pot -write_eps -write_ely -write_rho -inside 0 d >& out_${H}_${I}_elycav/d.out.vmemb

# example GMT scripts for (membrane) potential plotting:
#  plot-potential-0-1.csh plot-potential.csh
# pqr2curve in meaddir/apps/pqr2crv can be used to project the protein shape (or that of a slice)
# -slice float  gives the thickness of the slice plane in A normal to the slice plane
# solrad can be used to influence the smoothness of the outline
# a second tilted plane can be projected in a reference plene as before, the slice applies to the second, tilted plane
# example calls:
#/scratch/data/ullmannt/svn/mead/branches/ullmannt/current/apps/pqr2crv/pqr2crv -blab3 -solrad 0.0 -gmt 2.37832 -3.9938 1.881  -0.8 1.0 0  0.1 1001  his >& d.curve.his.out
#/scratch/data/ullmannt/svn/mead/branches/ullmannt/current/apps/pqr2crv/pqr2crv -blab3 -solrad 1 -gmt 2.37832 -3.9938 1.881  -0.8 1.0 0  0.25 401  2.37832 -3.9938 1.881  -0.8 1.0 0.36 -slice 1.5 d >& d.curve.overall_shape_slice_tilt.out

