#include "MEAD/AnalyMemb.h"
#include "MEAD/DielectricSlab.h"
#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/ChargeDist.h"
#include "MEAD/PhysCond.h"
#include "MEAD/globals.h"
#include <math.h>


AnalyMemb::AnalyMemb (DielectricSlab* e, ChargeDist_lett *r,
			  ElectrolyteEnvironment_lett* ely)
: AnalyticEP(e, r, ely), eps(e), rho(r), electrolyte(ely)
{
  if (ely->ionic_strength() == 0) {
    error ("AnalyMemb constructor: Transmembrane Potential with zero ionic strength\n",
	   "not implemented. Sorry.\n");
  }

  blab3 << "AnalyMemb:: initializing phi from analytical solution." << endl;

  solved = 0;
  epsslab = eps->epsslab_value();
  epsext = eps->epsext_value();
  zupper = eps->zupper_value();
  zlower = eps->zlower_value();
  istr = ely->ionic_strength();
  membpothalf = PhysCond::get_membpot() * 0.5 / PhysCond::get_voltconv();
  
  zmiddle  = (zlower + zupper) / 2.0;
  membl    = zupper - zlower;
  lhalf    = membl * 0.5;
  em_by_es = epsslab / epsext;

  kappam = sqrt(((istr * 2 * PhysCond::get_Avogadro_Number() * PhysCond::get_elementary_charge() * PhysCond::get_elementary_charge())
           / (PhysCond::get_kBolt() * PhysCond::get_T() * PhysCond::get_vacuum_permittivity() * epsext)));
  kappas = sqrt(((istr * 2 * PhysCond::get_Avogadro_Number() * PhysCond::get_elementary_charge() * PhysCond::get_elementary_charge())
           / (PhysCond::get_kBolt() * PhysCond::get_T() * PhysCond::get_vacuum_permittivity() * epsslab)));
  
  fac0 = (1 / ((em_by_es * kappas * membl) + 2));
  
  fac1 = PhysCond::get_membpot() * fac0 * em_by_es;  
  fac2 = PhysCond::get_membpot() * fac0;

}

void AnalyMemb::solve()
{
  if (solved) {
    std::cerr << "WARNING: AnalyMemb::solve called a second time"
	      << std::endl;
    return;
  }
  solved = 1;
}


float AnalyMemb::value(Coord c) const
{
  if (!solved) {
    error ("ERROR: AnalyMemb::value:\n",
	   "Setup not done yet, must call solve first\n");
  }

    float z = c.z - zmiddle;
    float phi_tot = 0;
    
    // membrane core
    if (z <= lhalf && z >= -lhalf){
      phi_tot = fac1 * (kappas * z);
    } // upper solvent region
    else if (z > lhalf){
      phi_tot =  membpothalf - (fac2 * exp((-kappas * (z - lhalf)))); 
    } // lower solvent region
    else if (z < lhalf){
      phi_tot = -membpothalf + (fac2 * exp((kappas * (z + lhalf)))); 
    }
    
   return phi_tot;
   
   //return 0.0;
}

Coord AnalyMemb::field(Coord c) const
{
  ::error("SORRY, AnalyMemb::field not implemented yet.  Exiting\n");
  return Coord(0,0,0);
}

Coord AnalyMemb::displacement(Coord c) const
{
  ::error("SORRY, AnalyMemb::displacement not implemented yet.  Exiting\n");
  return Coord(0,0,0);
}

// AnalyMemb.cc ends here
