#include "MEAD/AnalyMemb2.h"
#include "MEAD/DielectricSlab.h"
#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/ChargeDist.h"
#include "MEAD/PhysCond.h"
#include "MEAD/globals.h"
#include <math.h>


AnalyMemb2::AnalyMemb2 (DielectricSlab* e, ChargeDist_lett *r,
			  ElectrolyteEnvironment_lett* ely)
: AnalyticEP(e, r, ely), eps(e), rho(r), electrolyte(ely)
{
  if (ely->ionic_strength() == 0) {
    error ("AnalyMemb2 constructor: Transmembrane Potential with zero ionic strength\n",
	   "not implemented. Sorry.\n");
  }

  blab3 << "AnalyMemb2:: initializing phi from analytical solution." << endl;

  solved = 0;
  epsslab = eps->epsslab_value();
  epsext = eps->epsext_value();
  zupper = eps->zupper_value();
  zlower = eps->zlower_value();
  istr = ely->ionic_strength();
  membpothalf = PhysCond::get_membpot() * 0.5 / PhysCond::get_voltconv();
  
  zmiddle  = (zlower + zupper)/2.0;
  membl    = zupper - zlower;
  lhalf    = membl * 0.5;
  em_by_es = epsslab / epsext;

  kappam = sqrt(((istr * 2 * PhysCond::get_Avogadro_Number() * PhysCond::get_elementary_charge() * PhysCond::get_elementary_charge())
           / (PhysCond::get_kBolt() * PhysCond::get_T() * PhysCond::get_vacuum_permittivity() * epsext)));
  kappas = sqrt(((istr * 2 * PhysCond::get_Avogadro_Number() * PhysCond::get_elementary_charge() * PhysCond::get_elementary_charge())
           / (PhysCond::get_kBolt() * PhysCond::get_T() * PhysCond::get_vacuum_permittivity() * epsslab)));
  
  em_by_es_times_kappam_by_kappas = (em_by_es * kappam) / kappas;
  e_kappam_l = exp((kappam * membl));
  
  fac1 = membpothalf / (e_kappam_l - 1);  

}

void AnalyMemb2::solve()
{
  if (solved) {
    std::cerr << "WARNING: AnalyMemb2::solve called a second time"
	      << std::endl;
    return;
  }
  solved = 1;
}


float AnalyMemb2::value(Coord c) const
{
  if (!solved) {
    error ("ERROR: AnalyMemb2::value:\n",
	   "Setup not done yet, must call solve first\n");
  }

    float z = c.z - zmiddle;
    float phi_tot = 0;
    
    // membrane core
    if (z <= zupper && z >= zlower){
      phi_tot = fac1 * (exp((kappam * (z+lhalf))) - exp((-kappam * (z - lhalf))));
    } // upper solvent region
    else if (z > zupper){
      phi_tot =  membpothalf * (exp((-kappas * (z - lhalf)))); 
    } // lower solvent region
    else if (z > zupper){
      phi_tot = -membpothalf * (exp(( kappas * (z + lhalf)))); 
    }
    

   return phi_tot;
}

Coord AnalyMemb2::field(Coord c) const
{
  ::error("SORRY, AnalyMemb2::field not implemented yet.  Exiting\n");
  return Coord(0,0,0);
}

Coord AnalyMemb2::displacement(Coord c) const
{
  ::error("SORRY, AnalyMemb2::displacement not implemented yet.  Exiting\n");
  return Coord(0,0,0);
}

// AnalyMemb2.cc ends here
