/*  Object containing CHARMM parameters and methods for lookup
    
    lists of CHARMM parameters associated with corresponding
    atom type(s): bond, angle, dihedral, improper, nonbonded
    
    I/O functions for reading CHARMM topology and parameter files
    
    lookup functions for finding parameters given CHARMM 
    atom type(s) or CharmmID(s)

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _CharmmPara_h
#define _CharmmPara_h 1

#include <map>
#include <list>
#include <string>
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/CharmmParaBond.h"
#include "MEAD_e/CharmmParaAngle.h"
#include "MEAD_e/CharmmParaDihedral.h"
#include "MEAD_e/CharmmParaImproper.h"
#include "MEAD_e/CharmmParaNB.h"
#include "MEAD_e/CharmmParaResTop.h"
#include "MEAD_e/globals.h"
#include <vector>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class CharmmPara {
public:
  CharmmPara();
  CharmmPara (const string& name);
  CharmmPara (const CharmmPara& a);
  ~CharmmPara () {}
  //!nowrap!+
  CharmmPara& operator=(const CharmmPara& a);

  //!nowrap!-
  void read();
  // read from CHARMM topology and parameter file
  void read (const string& fn_string1, const string& fn_string2);
  void read_top (const string& fn_string);
  void read_par (const string& fn_string);
  void print() const;
  void print_typeindex() const;
  void print_chargeindex() const;
  void print_Topology() const;
  void print_NB() const;
  void print_Bond() const;
  void print_Angle() const;
  void print_Improper() const;
  void print_Dihedral() const;
  string get_name() const {return name;}

  std::map<CharmmID,string> pres_typeindex;
  std::map<CharmmID,string> typeindex;
  std::map<CharmmID,float> chargeindex;

  std::list<CharmmParaResTop> Topology;
  
  std::list<CharmmParaNB> NB;
  std::list<CharmmParaBond> Bond;
  std::list<CharmmParaAngle> Angle;
  std::list<CharmmParaDihedral> Dihedral;
  std::list<CharmmParaImproper> Improper;

  // lookup, insert residue topology CharmmResTop by resname

  // lookup, insert attype by CharmmID or resname/atname pair
  string get_attype(const CharmmID&) const;
  string get_attype(const string&, const string&) const;
  string get_pres_attype(const CharmmID&) const;
  string get_pres_attype(const string&, const string&) const;
  string get_pres_attype(const string&) const;
  bool contains_attype(const CharmmID&) const;
  bool contains_attype(const string&, const string&) const;
  bool contains_pres_attype(const CharmmID&) const;
  bool contains_pres_attype(const string&, const string&) const;
  bool contains_pres_attype(const string&) const;
  pair<std::map<CharmmID,string>::iterator,bool> insert_attype(const CharmmID&, const string&);
  pair<std::map<CharmmID,string>::iterator,bool> insert_attype(const string&, const string&, const string&);
  pair<std::map<CharmmID,string>::iterator,bool> insert_pres_attype(const CharmmID&, const string&);
  pair<std::map<CharmmID,string>::iterator,bool> insert_pres_attype(const string&, const string&, const string&);

  // lookup, insert charge by CharmmID or resname/atname pair
  float get_charge(const CharmmID&) const;
  float get_charge(const string&, const string&) const;
  bool contains_charge(const CharmmID&) const;
  bool contains_charge(const string&, const string&) const;
  pair<std::map<CharmmID,float>::iterator,bool> insert_charge(const CharmmID&, const float);
  pair<std::map<CharmmID,float>::iterator,bool> insert_charge(const string&, const string&, const float);

  /* check if parameter is in list by CharmmID(s) or attype(s) or resname atname pair(s) */

  bool contains_Topology(const string&) const;
  
  bool contains_NB(const string&) const;
  bool contains_NB(const string&, const string&) const;
  bool contains_NB(const CharmmID&) const;

  bool contains_Bond(const string&, const string&) const;
  bool contains_Bond(const string&, const string&, 
                     const string&, const string&) const;
  bool contains_Bond(const CharmmID&, const CharmmID&) const;

  bool contains_Angle(const string&, const string&, const string&) const;
  bool contains_Angle(const string&, const string&, 
                      const string&, const string&, 
                      const string&, const string&) const;
  bool contains_Angle(const CharmmID&, const CharmmID&, const CharmmID&) const;

  bool contains_Dihedral(const string&, const string&, const string&, const string&) const;
  bool contains_Dihedral(const string&, const string&, 
                         const string&, const string&, 
                         const string&, const string&, 
                         const string&, const string&) const;
  bool contains_Dihedral(const CharmmID&, const CharmmID&, const CharmmID&, const CharmmID&) const;

  bool contains_Improper(const string&, const string&, const string&, const string&) const;
  bool contains_Improper(const string&, const string&, 
                         const string&, const string&, 
                         const string&, const string&, 
                         const string&, const string&) const;
  bool contains_Improper(const CharmmID&, const CharmmID&, const CharmmID&, const CharmmID&) const;
  
  /* retrieve parameter by CharmmID(s) or attype(s) or resname atname pair(s) */

  CharmmParaResTop get_Topology(const string&) const;

  CharmmParaNB get_NB(const string&) const;
  CharmmParaNB get_NB(const string&, const string&) const;
  CharmmParaNB get_NB(const CharmmID&) const;

  CharmmParaBond get_Bond(const string&, const string&) const;
  CharmmParaBond get_Bond(const string&, const string&, 
                          const string&, const string&) const;
  CharmmParaBond get_Bond(const CharmmID&, const CharmmID&) const;

  CharmmParaAngle get_Angle(const string&, const string&, const string&) const;
  CharmmParaAngle get_Angle(const string&, const string&, 
                            const string&, const string&, 
                            const string&, const string&) const;
  CharmmParaAngle get_Angle(const CharmmID&, const CharmmID&, const CharmmID&) const;

  list<CharmmParaDihedral> get_Dihedral(const string&, const string&, const string&, const string&) const;
  list<CharmmParaDihedral> get_Dihedral(const string&, const string&, 
                                        const string&, const string&, 
                                        const string&, const string&, 
                                        const string&, const string&) const;
  list<CharmmParaDihedral> get_Dihedral(const CharmmID&, const CharmmID&, const CharmmID&, const CharmmID&) const;

  CharmmParaImproper get_Improper(const string&, const string&, const string&, const string&) const;
  CharmmParaImproper get_Improper(const string&, const string&, 
                                  const string&, const string&, 
                                  const string&, const string&, 
                                  const string&, const string&) const;
  CharmmParaImproper get_Improper(const CharmmID&, const CharmmID&, const CharmmID&, const CharmmID&) const;

  /* insert parameter in list, return true if successfull */

  std::list<CharmmParaResTop>::iterator insert_Topology(const CharmmParaResTop&);

  std::list<CharmmParaNB>::iterator insert_NB(const CharmmParaNB&);
  std::list<CharmmParaNB>::iterator insert_NB(const string&, const float, const float, const float, 
                                              const float, const float);
  std::list<CharmmParaBond>::iterator insert_Bond(const CharmmParaBond&);
  std::list<CharmmParaBond>::iterator insert_Bond(const string&, const string&, const float, const float);
  std::list<CharmmParaAngle>::iterator insert_Angle(const CharmmParaAngle&);
  std::list<CharmmParaAngle>::iterator insert_Angle(const string&, const string&, const string&, 
                                                 const float, const float, const float, const float);
  std::list<CharmmParaDihedral>::iterator insert_Dihedral(const CharmmParaDihedral&);
  std::list<CharmmParaDihedral>::iterator insert_Dihedral(const string&, const string&, const string&, const string&, 
                                                    const int, const float, const float);
  std::list<CharmmParaImproper>::iterator insert_Improper(const CharmmParaImproper&);
  std::list<CharmmParaImproper>::iterator insert_Improper(const string&, const string&, const string&, const string&, 
                                                          const float, const float);

  // def values of elements in CHARMM
  bool contains_element_NB(const string&) const;
  CharmmParaNB get_element_NB(const string&) const;

  //!nowrap!-
private:
  string name;
};

inline bool CharmmPara::contains_element_NB(const std::string& s) const
{
 bool found = false;
 if (s == std::string("C")   || s == std::string("O")   ||
     s == std::string("H")   || s == std::string("N")   ||
     s == std::string("S")   || s == std::string("P")   ||
     s == std::string("CU")  || s == std::string("FE")  ||
     s == std::string("CA")  || s == std::string("NA")  ||
     s == std::string("K")   || s == std::string("MG")  ||
     s == std::string("F")   || s == std::string("CL")  ||
     s == std::string("CES") || s == std::string("ZN")  ||
     s == std::string("HE")  || s == std::string("NE")  ||
     s == std::string("HE")  || s == std::string("NE")
     ){found = true;}
 return found;
}

inline CharmmParaNB CharmmPara::get_element_NB(const std::string& s) const
{
 bool found(contains_element_NB(s));
 if(!found)
    ::error("CharmmPara::get_element_NB: key not found");
 CharmmParaNB par;
 if        (s == std::string("C")){
  par.ljr = 2.0; par.lje = -0.07; 
 } else if (s == std::string("O")){
  par.ljr = 1.7; par.lje = -0.120000; 
 } else if (s == std::string("H")){
  par.ljr = 0.224500; par.lje = -0.046000;
 } else if (s == std::string("N")){
  par.ljr = 1.850000; par.lje = -0.200000;
 } else if (s == std::string("S")){
  par.ljr = 2.0; par.lje = -0.45; 
 } else if (s == std::string("P")){
  par.ljr = 2.15; par.lje = -0.585;
 } else if (s == std::string("CU")){
  par.ljr = 1.85; par.lje = -0.2; 
 } else if (s == std::string("FE")){
  par.ljr = 0.650000; par.lje = -0.100;
 } else if (s == std::string("CA")){
  par.ljr = 1.367; par.lje = -0.120;
 } else if (s == std::string("NA")){
  par.ljr = 1.36375; par.lje = -0.0469;
 } else if (s == std::string("K")){
  par.ljr = 1.76375; par.lje = -0.0870;
 } else if (s == std::string("MG")){
  par.ljr = 1.18500; par.lje = -0.0150;
 } else if (s == std::string("F")){
  par.ljr = 1.70; par.lje = -0.09;
 } else if (s == std::string("CL")){
  par.ljr = 2.27; par.lje = -0.150;
 } else if (s == std::string("CES")){
  par.ljr = 2.100; par.lje = -0.1900;
 } else if (s == std::string("ZN")){
  par.ljr = 1.090000; par.lje = -0.250000;
 } else if (s == std::string("HE")){
  par.ljr = 1.4800; par.lje = -0.021270;
 } else if (s == std::string("NE")){
  par.ljr = 1.5300; par.lje = -0.086000;
 } else {
    ::error("CharmmPara::get_element_NB: key not found"); 
 } 
 return par;
}

// typeindex functions

/*
inline string CharmmPara::get_attype(const CharmmID& k) const
{
  std::string result; 
  std::map<CharmmID,string>::const_iterator i=typeindex.find(k);
  if (i == typeindex.end())
    ::error("CharmmPara::get_attype: key not found");
  return i->second;
}
*/

inline string CharmmPara::get_attype(const CharmmID& k) const
{
  std::string result; 
  std::map<CharmmID,string>::const_iterator i=typeindex.find(k);
  if (i == typeindex.end()){
   if (contains_pres_attype(k.get_atname())) 
    result = get_pres_attype(k.get_atname());
   else {
    cerr << k << endl;
    ::error("CharmmPara::get_attype: key not found");
   }
  }
  else {result = i->second;}
  return result;
}

inline string CharmmPara::get_attype(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::get_attype(k);
}

inline string CharmmPara::get_pres_attype(const CharmmID& k) const
{
  std::string result; 
  std::map<CharmmID,string>::const_iterator i=pres_typeindex.find(k);
  if (i == pres_typeindex.end())
    ::error("CharmmPara::get_pres_attype: key not found");
  return i->second;
}

inline string CharmmPara::get_pres_attype(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::get_pres_attype(k);
}

inline string CharmmPara::get_pres_attype(const string& an) const
{
  bool found = false;
  std::string result;
  for (std::map<CharmmID,string>::const_iterator i=pres_typeindex.begin(); i!=pres_typeindex.end(); ++i){
   if (i->first.get_atname() == an){
    found = true;
    result = i->second;
    break;
   }
  }
  if (!found){
    ::error("CharmmPara::get_pres_attype: key not found");
  }
  return result;
}

// Existence of element (not STL-like

inline bool CharmmPara::contains_attype(const CharmmID& k) const
{
  std::map<CharmmID,string>::const_iterator i=typeindex.find(k);
  return i != typeindex.end();
}

inline bool CharmmPara::contains_attype(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::contains_attype(k);
}

inline bool CharmmPara::contains_pres_attype(const CharmmID& k) const
{
  std::map<CharmmID,string>::const_iterator i=pres_typeindex.find(k);
  return i != pres_typeindex.end();
}

inline bool CharmmPara::contains_pres_attype(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::contains_pres_attype(k);
}

inline bool CharmmPara::contains_pres_attype(const string& an) const
{
  bool found = false;
  for (std::map<CharmmID,string>::const_iterator i=pres_typeindex.begin(); i!=pres_typeindex.end(); ++i){
   if (i->first.get_atname() == an){
    found = true;
    break;
   }
  }
  return found;
}

// easier to use insert (not quite STL-like)
inline pair<std::map<CharmmID,string>::iterator,bool> CharmmPara::insert_attype (const string& rn, const string& an, const string& at){
  CharmmID k(rn, an);
  if (contains_attype(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_attype: attempt to insert a duplicate attype for atom\n");
  }
  return typeindex.insert(std::pair<CharmmID,string>(k, at));
}

inline pair<std::map<CharmmID,string>::iterator,bool> CharmmPara::insert_attype (const CharmmID& k, const string& at){
  if (contains_attype(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_attype: attempt to insert a duplicate attype for atom\n");
  }
  return typeindex.insert(std::pair<CharmmID,string>(k, at));
}

inline pair<std::map<CharmmID,string>::iterator,bool> CharmmPara::insert_pres_attype (const string& rn, const string& an, const string& at){
  CharmmID k(rn, an);
  if (contains_pres_attype(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_attype: attempt to insert a duplicate pres_attype for atom\n");
  }
  return pres_typeindex.insert(std::pair<CharmmID,string>(k, at));
}

inline pair<std::map<CharmmID,string>::iterator,bool> CharmmPara::insert_pres_attype (const CharmmID& k, const string& at){
  if (contains_pres_attype(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_attype: attempt to insert a duplicate pres_attype for atom\n");
  }
  return pres_typeindex.insert(std::pair<CharmmID,string>(k, at));
}

// chargeindex functions

inline float CharmmPara::get_charge(const CharmmID& k) const
{
  std::map<CharmmID,float>::const_iterator i=chargeindex.find(k);
  if (i == chargeindex.end())
    ::error("AtomSet::get_charge: key not found");
  return i->second;
}

inline float CharmmPara::get_charge(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::get_charge(k);
}

// Existence of element (not STL-like
inline bool CharmmPara::contains_charge(const CharmmID& k) const
{
  std::map<CharmmID,float>::const_iterator i=chargeindex.find(k);
  return i != chargeindex.end();
}

inline bool CharmmPara::contains_charge(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return CharmmPara::contains_charge(k);
}

// easier to use insert (not quite STL-like)
inline pair<std::map<CharmmID,float>::iterator,bool> CharmmPara::insert_charge (const string& rn, const string& an, const float q){
  CharmmID k(rn, an);
  if (contains_charge(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_charge: attempt to insert a duplicate charge for atom\n");
  }
  return chargeindex.insert(std::pair<CharmmID,float>(k, q));
}

inline pair<std::map<CharmmID,float>::iterator,bool> CharmmPara::insert_charge (const CharmmID& k, const float q){
  if (contains_charge(k)) {
    cerr << k << endl;
    ::error("CharmmPara::insert_charge: attempt to insert a duplicate charge for atom\n");
  }
  return chargeindex.insert(std::pair<CharmmID,float>(k, q));
}


/////  existence of element  /////

// -- NB -- //

inline bool CharmmPara::contains_NB(const string& at) const
{
  bool result = false;
  for (std::list<CharmmParaNB>::const_iterator i = NB.begin(); i != NB.end(); ++i) {
   if ((*i).attype == at){
    result = true;
    break;
   } 
  }    
  return result;
}
inline bool CharmmPara::contains_NB(const string& rn, const string& an) const
{
  string at = get_attype(rn, an);
  return contains_NB(at);
}
inline bool CharmmPara::contains_NB(const CharmmID& cid) const
{
  string at = get_attype(cid);
  return contains_NB(at); 
}

// -- Bond -- //

inline bool CharmmPara::contains_Bond(const string& at1, const string& at2) const
{
  bool result = false;
  for (std::list<CharmmParaBond>::const_iterator i = Bond.begin(); i != Bond.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2) || 
       ((*i).attype1 == at2 && (*i).attype2 == at1)){
    result = true;
    break;    
   } 
  }    
  return result;
}
inline bool CharmmPara::contains_Bond(const string& rn1, const string& an1, 
                                      const string& rn2, const string& an2) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  return contains_Bond(at1, at2);
}

inline bool CharmmPara::contains_Bond(const CharmmID& cid1, const CharmmID& cid2) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  return contains_Bond(at1, at2); 
}

// -- Angle -- //

inline bool CharmmPara::contains_Angle(const string& at1, const string& at2, const string& at3) const
{
  bool result = false;
  for (std::list<CharmmParaAngle>::const_iterator i = Angle.begin(); i != Angle.end(); ++i) {
   if ((*i).attype2 == at2){
    if (((*i).attype1 == at1 && (*i).attype3 == at3) || 
        ((*i).attype1 == at3 && (*i).attype3 == at1)){
     result = true;
     break;   
    }
   }
  }    
  return result;
}
inline bool CharmmPara::contains_Angle(const string& rn1, const string& an1, 
                                       const string& rn2, const string& an2,
                                       const string& rn3, const string& an3) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  return contains_Angle(at1, at2, at3);
}
inline bool CharmmPara::contains_Angle(const CharmmID& cid1, const CharmmID& cid2, const CharmmID& cid3) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  return contains_Angle(at1, at2, at3);
}

// -- Dihedral -- //

inline bool CharmmPara::contains_Dihedral(const string& at1, const string& at2, const string& at3, const string& at4) const
{
  bool result = false;
  for (std::list<CharmmParaDihedral>::const_iterator i = Dihedral.begin(); i != Dihedral.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
       ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1) ||
       ((*i).attype1 == "X" && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == "X") ||
       ((*i).attype1 == "X" && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == "X")){
    result = true;
    break;
   }
  }    
  return result;
}

inline bool CharmmPara::contains_Dihedral(const string& rn1, const string& an1, 
                                          const string& rn2, const string& an2, 
                                          const string& rn3, const string& an3, 
                                          const string& rn4, const string& an4) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  string at4 = get_attype(rn4, an4);
  return contains_Dihedral(at1, at2, at3, at4);
}

inline bool CharmmPara::contains_Dihedral(const CharmmID& cid1, const CharmmID& cid2, 
                                          const CharmmID& cid3, const CharmmID& cid4) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  string at4 = get_attype(cid4);
  return contains_Dihedral(at1, at2, at3, at4);
}

// -- Improper -- //

inline bool CharmmPara::contains_Improper(const string& at1, const string& at2, const string& at3, const string& at4) const
{
  bool result = false;
  for (std::list<CharmmParaImproper>::const_iterator i = Improper.begin(); i != Improper.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
       ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1) ||
       ((*i).attype1 == at1 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at4) ||
       ((*i).attype1 == at4 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at1)){
    result = true;
    break;
   }
  }    
  return result;
}

inline bool CharmmPara::contains_Improper(const string& rn1, const string& an1, 
                                          const string& rn2, const string& an2, 
                                          const string& rn3, const string& an3, 
                                          const string& rn4, const string& an4) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  string at4 = get_attype(rn4, an4);
  return contains_Improper(at1, at2, at3, at4);
}

inline bool CharmmPara::contains_Improper(const CharmmID& cid1, const CharmmID& cid2, 
                                          const CharmmID& cid3, const CharmmID& cid4) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  string at4 = get_attype(cid4);
  return contains_Improper(at1, at2, at3, at4);
}

inline bool CharmmPara::contains_Topology(const string& n) const {
  bool found = false;
  for (std::list<CharmmParaResTop>::const_iterator i = Topology.begin(); i != Topology.end(); ++i) {
   if ((*i).get_name() == n){
     found = true;
     break;        
   }
  }
  return found;
}

///// list insert functions /////

inline std::list<CharmmParaNB>::iterator CharmmPara::insert_NB (const CharmmParaNB& par){
  string at = par.attype;
  for (std::list<CharmmParaNB>::iterator i = NB.begin(); i != NB.end(); ++i) {
   if ((*i).attype == at){
    cerr << "WARNING: attempt to insert duplicate NONBOND parameter for atom type " 
         << at << "\nLast one will overide previous" << endl;
    NB.erase(i);
    break;
   }
  }    
  NB.push_back(par);
  return NB.end();
}

inline std::list<CharmmParaNB>::iterator CharmmPara::insert_NB (const string& at, const float f1, const float f2, 
                                                                const float f3, const float f4, const float f5){
 CharmmParaNB par;
 par.attype = at;
 par.lje = f1;
 par.ljr = f2;
 par.lje14 = f3;
 par.ljr14 = f4;
 par.charge = f5;
 insert_NB(par);
}

inline std::list<CharmmParaBond>::iterator CharmmPara::insert_Bond (const CharmmParaBond& par){
  string at1 = par.attype1;
  string at2 = par.attype2;
  for (std::list<CharmmParaBond>::iterator i = Bond.begin(); i != Bond.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2) || 
       ((*i).attype1 == at2 && (*i).attype2 == at1)){
    cerr << "WARNING: attempt to insert duplicate BOND parameter for atom types " 
         << at1 << " " << at2 << "\nLast one will overide previous" << endl;
    Bond.erase(i);
    break;
   }
  }    
  Bond.push_back(par);
  return Bond.end();
}

inline std::list<CharmmParaBond>::iterator CharmmPara::insert_Bond (const string& at1, const string& at2, 
                                                                    const float f1, const float f2){
 CharmmParaBond par;
 par.attype1 = at1;
 par.attype2 = at2;
 par.k = f1;
 par.r = f2;
 insert_Bond(par);
}

inline std::list<CharmmParaAngle>::iterator CharmmPara::insert_Angle (const CharmmParaAngle& par){
  string at1 = par.attype1;
  string at2 = par.attype2;
  string at3 = par.attype3;
  for (std::list<CharmmParaAngle>::iterator i = Angle.begin(); i != Angle.end(); ++i) {
   if ((*i).attype2 == at2){
    if (((*i).attype1 == at1 && (*i).attype3 == at3) || 
        ((*i).attype1 == at3 && (*i).attype3 == at1)){
     cerr << "WARNING: attempt to insert duplicate Angle parameter for atom types " 
          << at1 << " " << at2 << "\nLast one will overide previous" << endl;
     Angle.erase(i);
     break;
    }
   }
  }    
  Angle.push_back(par);
  return Angle.end();
}

inline std::list<CharmmParaAngle>::iterator CharmmPara::insert_Angle (const string& at1, const string& at2,
 const string& at3, const float f1, const float f2, const float f3, const float f4){
 CharmmParaAngle par;
 par.attype1 = at1;
 par.attype2 = at2;
 par.attype3 = at3;
 par.kphi = f1;
 par.phi = f2;
 par.kub = f3;
 par.rub = f4;
 insert_Angle(par);
}

inline std::list<CharmmParaDihedral>::iterator CharmmPara::insert_Dihedral (const CharmmParaDihedral& par){
  string at1 = par.attype1;
  string at2 = par.attype2;
  string at3 = par.attype3;
  string at4 = par.attype4;
  float n = par.n;
  for (std::list<CharmmParaDihedral>::iterator i = Dihedral.begin(); i != Dihedral.end(); ++i) {
   if ((((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
        ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1))
        && (*i).n == n){
    cerr << "WARNING: attempt to insert duplicate DIHEDRAL parameter for atom types " 
         << at1 << " " << at2 << " " << at3 << " " << at4 << " " << n <<
         "\nLast one will overide previous" << endl;
    Dihedral.erase(i);
    break;
   }
  }    
  Dihedral.push_back(par);
  return Dihedral.end();
}

inline std::list<CharmmParaDihedral>::iterator CharmmPara::insert_Dihedral (const string& at1, const string& at2,
 const string& at3, const string& at4, const int i1, const float f1, const float f2){
 CharmmParaDihedral par;
 par.attype1 = at1;
 par.attype2 = at2;
 par.attype3 = at3;
 par.attype4 = at4;
 par.n = i1;
 par.k = f1;
 par.delta = f2;
 insert_Dihedral(par);
}

inline std::list<CharmmParaImproper>::iterator CharmmPara::insert_Improper (const CharmmParaImproper& par){
  string at1 = par.attype1;
  string at2 = par.attype2;
  string at3 = par.attype3;
  string at4 = par.attype4;
  for (std::list<CharmmParaImproper>::iterator i = Improper.begin(); i != Improper.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
       ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1) ||
       ((*i).attype1 == at1 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at4) ||
       ((*i).attype1 == at4 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at1)){
    cerr << "WARNING: attempt to insert duplicate Improper parameter for atom types " 
         << at1 << " " << at2 << " " << at3 << " " << at4 << 
         "\nLast one will overide previous" << endl;
    Improper.erase(i);
    break;
   }
  }    
  Improper.push_back(par);
  return Improper.end();
}

inline std::list<CharmmParaImproper>::iterator CharmmPara::insert_Improper (const string& at1, const string& at2,
 const string& at3, const string& at4, const float f1, const float f2){
 CharmmParaImproper par;
 par.attype1 = at1;
 par.attype2 = at2;
 par.attype3 = at3;
 par.attype4 = at4;
 par.kpsi = f1;
 par.psi = f2;
 insert_Improper(par);
}

inline std::list<CharmmParaResTop>::iterator CharmmPara::insert_Topology(const CharmmParaResTop& par) {
  for (std::list<CharmmParaResTop>::iterator i = Topology.begin(); i != Topology.end(); ++i) {
   if ((*i).get_name() == par.get_name()){
    cerr << "WARNING: attempt to insert duplicate Residue Topology for resname " 
         << par.get_name() << ".\nLast one will override previous." << endl;             
    Topology.erase(i);
    break;
   }
  }
  Topology.push_back(par);    
  return Topology.end();
}

///// list element retrieval functions functions /////

//-- NB --//

inline CharmmParaNB CharmmPara::get_NB(const string& at) const {
  CharmmParaNB par;
  for (std::list<CharmmParaNB>::const_iterator i = NB.begin(); i != NB.end(); ++i) {
   if ((*i).attype == at){
    par = (*i);
    break;
   }
   else if (i == NB.end())
    ::error("CharmmPara::get_NB: key not found");
  }    
  return par;
}
inline CharmmParaNB CharmmPara::get_NB(const string& rn, const string& an) const
{
  string at = get_attype(rn, an);
  return get_NB(at);
}
inline CharmmParaNB CharmmPara::get_NB(const CharmmID& cid) const
{
  string at = get_attype(cid);
  return get_NB(at);
}

// -- Bond -- //

inline CharmmParaBond CharmmPara::get_Bond(const string& at1, const string& at2) const {
  CharmmParaBond par;
  for (std::list<CharmmParaBond>::const_iterator i = Bond.begin(); i != Bond.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2) || 
       ((*i).attype1 == at2 && (*i).attype2 == at1)){
    par = (*i);
    break;
   }
   else if (i == Bond.end())
    ::error("CharmmPara::get_Bond: key not found");
  }    
  return par;
}

inline CharmmParaBond CharmmPara::get_Bond(const string& rn1, const string& an1, 
                                           const string& rn2, const string& an2) const 
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  return get_Bond(at1, at2);
}

inline CharmmParaBond CharmmPara::get_Bond(const CharmmID& cid1, const CharmmID& cid2) const 

{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  return get_Bond(at1, at2);
}

//-- Angle --//

inline CharmmParaAngle CharmmPara::get_Angle(const string& at1, const string& at2, const string& at3) const {
  CharmmParaAngle par;
  for (std::list<CharmmParaAngle>::const_iterator i = Angle.begin(); i != Angle.end(); ++i) {
   if ((*i).attype2 == at2){
    if (((*i).attype1 == at1 && (*i).attype3 == at3) || 
        ((*i).attype1 == at3 && (*i).attype3 == at1)){
     par = (*i);
     break;    
    }
    else if (i == Angle.end())
     ::error("CharmmPara::get_Angle: key not found");
   }
   else if (i == Angle.end())
    ::error("CharmmPara::get_Angle: key not found");
  }    
  return par;
}

inline CharmmParaAngle CharmmPara::get_Angle(const string& rn1, const string& an1, 
                                             const string& rn2, const string& an2, 
                                             const string& rn3, const string& an3) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  return get_Angle(at1, at2, at3);
}

inline CharmmParaAngle CharmmPara::get_Angle(const CharmmID& cid1, const CharmmID& cid2,
                                             const CharmmID& cid3) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  return get_Angle(at1, at2, at3);
}

//-- Dihedral --//

inline list<CharmmParaDihedral> CharmmPara::get_Dihedral(const string& at1, const string& at2, 
                                                   const string& at3, const string& at4) const
{
  bool found = false; 
  list<CharmmParaDihedral> par;
  for (std::list<CharmmParaDihedral>::const_iterator i = Dihedral.begin(); i != Dihedral.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
       ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1) ||
       ((*i).attype1 == "X" && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == "X") ||
       ((*i).attype1 == "X" && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == "X")){
    par.push_back(*i);
    found = true;
   }
  }
  if (!found)
    ::error("CharmmPara::get_Dihedral: key not found");
  return par;
}

inline list<CharmmParaDihedral> CharmmPara::get_Dihedral(const string& rn1, const string& an1, 
                                          const string& rn2, const string& an2, 
                                          const string& rn3, const string& an3, 
                                          const string& rn4, const string& an4) const
{
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  string at4 = get_attype(rn4, an4);
  return get_Dihedral(at1, at2, at3, at4);
}

inline list<CharmmParaDihedral> CharmmPara::get_Dihedral(const CharmmID& cid1, const CharmmID& cid2, 
                                                   const CharmmID& cid3, const CharmmID& cid4) const
{
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  string at4 = get_attype(cid4);
  return get_Dihedral(at1, at2, at3, at4);
}

//-- Improper --//

inline CharmmParaImproper CharmmPara::get_Improper(const string& at1, const string& at2, 
                                                   const string& at3, const string& at4) const
{
  CharmmParaImproper par;
  for (std::list<CharmmParaImproper>::const_iterator i = Improper.begin(); i != Improper.end(); ++i) {
   if (((*i).attype1 == at1 && (*i).attype2 == at2 && (*i).attype3 == at3 && (*i).attype4 == at4) || 
       ((*i).attype1 == at4 && (*i).attype2 == at3 && (*i).attype3 == at2 && (*i).attype4 == at1) ||
       ((*i).attype1 == at1 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at4) ||
       ((*i).attype1 == at4 && (*i).attype2 == "X" && (*i).attype3 == "X" && (*i).attype4 == at1)){
     par = (*i);
     break;        
   }
   else if (i == Improper.end())
    ::error("CharmmPara::get_Improper: key not found");
  }    
  return par;
}

inline CharmmParaImproper CharmmPara::get_Improper(const string& rn1, const string& an1, 
                                          const string& rn2, const string& an2, 
                                          const string& rn3, const string& an3, 
                                          const string& rn4, const string& an4) const {
  string at1 = get_attype(rn1, an1);
  string at2 = get_attype(rn2, an2);
  string at3 = get_attype(rn3, an3);
  string at4 = get_attype(rn4, an4);
  return get_Improper(at1, at2, at3, at4);
}

inline CharmmParaImproper CharmmPara::get_Improper(const CharmmID& cid1, const CharmmID& cid2, 
                                                   const CharmmID& cid3, const CharmmID& cid4) const {
  string at1 = get_attype(cid1);
  string at2 = get_attype(cid2);
  string at3 = get_attype(cid3);
  string at4 = get_attype(cid4);
  return get_Improper(at1, at2, at3, at4);
}

//-- Topology --//

inline CharmmParaResTop CharmmPara::get_Topology(const string& n) const {
  CharmmParaResTop par;
  for (std::list<CharmmParaResTop>::const_iterator i = Topology.begin(); i != Topology.end(); ++i) {
   if ((*i).get_name() == n){
     par = (*i);
     break;        
   }
   else if (i == Topology.end())
    ::error("CharmmPara::get_Improper: key not found");
  }    
  return par;
}

#endif

// CharmmPara.h ends here
