/*  Topology of a Charmm Residue

    bonds, impropers are predefined as read from CHARMM topology file
    angles and dihedrals are generated from bonds

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de
    
    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/CharmmParaResTop.h"
#include "MEAD_e/globals.h"
#include <stdio.h>

using namespace std;
using namespace boost;

CharmmParaResTop::CharmmParaResTop(const string& nm, const list<string> a, 
                                   const map<pair<string, string>, int> b, const list<vector<string> > i)
: name(nm), atoms(a), bonds(b), impropers(i)
{
 generate_top();
}

CharmmParaResTop::CharmmParaResTop(const CharmmParaResTop& a)
: name(a.name), atoms(a.atoms), bonded_list(a.bonded_list), bonds(a.bonds), angles(a.angles), 
  dihedrals(a.dihedrals), impropers(a.impropers), sidechain_tor(a.sidechain_tor)
{
}

CharmmParaResTop::CharmmParaResTop()
: name("")
{
}

CharmmParaResTop::CharmmParaResTop (const string& nm)
: name(nm)
{
}

void CharmmParaResTop::check_bonds() const
{ // check if all atoms in bonds occur also in the atom name list
  //blab3 << "Entering CharmmParaResTop::check_bonds()" << endl;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   // +/- mark atoms of next/previous residue
   if ((((i->first).first).find("+") != 0 && ((i->first).first).find("-") != 0) && !contains_atom((i->first).first)){
    cerr << "residue name: " << name << " atom name: " << (i->first).first << endl;
    ::error("CharmmParaResTop::check_bonds: atom in bond does not exist in residue definition\n");        
   }
   else if ((((i->first).second).find("+") != 0 && ((i->first).second).find("-") != 0) && !contains_atom((i->first).second)){
    cerr << "residue name: " << name << " atom name: " << (i->first).second << endl;
    ::error("CharmmParaResTop::check_bonds: atom in bond does not exist in residue definition\n");            
   }
  }
  //blab3 << "Exiting CharmmParaResTop::check_bonds()" << endl;
}

void CharmmParaResTop::check_impropers() const
{ // check if all vectors contain 4 atoms and if all 4 names occur in the atom name list
  //blab3 << "Entering CharmmParaResTop::check_impropers()" << endl;
  for (std::list<vector<string> >::const_iterator i = impropers.begin(); i != impropers.end(); ++i){
   if ((*i).size() != 4){
    ::error("CharmmParaResTop::check_improper: incorrect vector size in impropers.\n");     
   }
   // +/- mark atoms of next/previous residue
   if (((*i)[0].find("+") != 0 && (*i)[0].find("-") != 0) && !contains_atom((*i)[0])){
    cerr << "residue name: " << name << " atom name: " << (*i)[0] << endl;
    ::error("CharmmParaResTop::check_impropers: atom in improper does not exist in residue definition\n");   
   }
   else if (((*i)[1].find("+") != 0 && (*i)[1].find("-") != 0) && !contains_atom((*i)[1])){
    cerr << "residue name: " << name << " atom name: " << (*i)[1] << endl;
    ::error("CharmmParaResTop::check_impropers: atom in improper does not exist in residue definition\n");   
   }
   else if (((*i)[2].find("+") != 0 && (*i)[2].find("-") != 0) && !contains_atom((*i)[2])){
    cerr << "residue name: " << name << " atom name: " << (*i)[2] << endl;
    ::error("CharmmParaResTop::check_impropers: atom in improper does not exist in residue definition\n");   
   }
   else if (((*i)[3].find("+") != 0 && (*i)[3].find("-") != 0) && !contains_atom((*i)[3])){
    cerr << "residue name: " << name << " atom name: " << (*i)[3] << endl;
    ::error("CharmmParaResTop::check_impropers: atom in improper does not exist in residue definition\n");   
   }
  }
  //blab3 << "Exiting CharmmParaResTop::check_impropers()" << endl;
}

void CharmmParaResTop::generate_top()
{
 // verify that each atom in a bond is also on the atom list
 check_bonds();
 // verify that each improper string vector contains only 4 elements
 check_impropers();
 generate_bonded_list(); 
 generate_angles();
 generate_dihedrals();
 generate_sidechain_tor();
 //print();
}

void CharmmParaResTop::generate_bonded_list()
{
  // bonded list is a map of all atoms and a list of atoms bonded to them
  //blab3 << "Entering CharmmParaResTop::generate_bonded_list()" << endl;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   if (contains_bonded_entry((i->first).first)){    
    for (map<string,list<string> >::iterator j = bonded_list.begin(); j != bonded_list.end(); ++j){
     if (j->first == (i->first).first){
      (j->second).push_back((i->first).second);
     }
    }
   }
   else {
    list<string> tmpl;
    tmpl.push_back((i->first).second);
    pair<string,list<string> > tmppair((i->first).first, tmpl);
    bonded_list.insert(tmppair);
   }
   if (contains_bonded_entry((i->first).second)){    
    for (map<string,list<string> >::iterator j = bonded_list.begin(); j != bonded_list.end(); ++j){
     if (j->first == (i->first).second){
      (j->second).push_back((i->first).first);
     }
    }
   }
   else {
    list<string> tmpl;
    tmpl.push_back((i->first).first);
    pair<string,list<string> > tmppair((i->first).second, tmpl);
    bonded_list.insert(tmppair);
   }
  }
  //blab3 << "Exiting CharmmParaResTop::generate_bonded_list()" << endl;
}

void CharmmParaResTop::generate_angles(){
  // angles is a list of tripels of atoms bonded successively and thus linked in an angle
  //blab3 << "Entering CharmmParaResTop::generate_angles()" << endl;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   for (map<pair<string,string>,int>::const_iterator j = i; j != bonds.end(); ++j){
    if ((i->first).first == (j->first).first && (i->first).second != (j->first).second){
     if (!contains_angle((i->first).second, (i->first).first, (j->first).second)){
      insert_angle((i->first).second, (i->first).first, (j->first).second);
     }
    }
    else if ((i->first).first == (j->first).second && (i->first).second != (j->first).first){
     if (!contains_angle((i->first).second, (i->first).first, (j->first).first)){
      insert_angle((i->first).second, (i->first).first, (j->first).first);
     }
    } 
    else if ((i->first).second == (j->first).second && (i->first).first != (j->first).first){
     if (!contains_angle((i->first).first, (i->first).second, (j->first).first)){
      insert_angle((i->first).first, (i->first).second, (j->first).first);
     }
    } 
    else if ((i->first).second == (j->first).first && (i->first).first != (j->first).second){
     if (!contains_angle((i->first).first, (i->first).second, (j->first).second)){
      insert_angle((i->first).first, (i->first).second, (j->first).second);
     }
    } 
   }
  }
  //blab3 << "Exiting CharmmParaResTop::generate_angles()" << endl;
}

void CharmmParaResTop::generate_dihedrals()
{
  // dihedrals is a list of quadrupels of atoms bonded successively and thus linked in an angle
  // no double baonds are to be involved in autogenerated dihedrals
  //blab3 << "Entering CharmmParaResTop::generate_dihedrals()" << endl;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   for (map<pair<string,string>,int>::const_iterator j = bonds.begin(); j != bonds.end(); ++j){
    for (map<pair<string,string>,int>::const_iterator k = bonds.begin(); k != bonds.end(); ++k){
     // here we iterate necessarily twice through each pair, 
     // but we save some time by considering only one direction of the dihedrals
     // first check if all three bonds are different
     // at least one atom must differ between a pair of bonds
     if(i->second != 1 || j->second != 1 || k-> second != 1                              ||
        ((i->first).first == (j->first).first && (i->first).second == (j->first).second) ||
        ((i->first).first == (j->first).second && (i->first).second == (j->first).first) ||
        ((i->first).first == (k->first).first && (i->first).second == (k->first).second) ||
        ((i->first).first == (k->first).second && (i->first).second == (k->first).first) ||
        ((j->first).first == (k->first).first && (j->first).second == (k->first).second) ||
        ((j->first).first == (k->first).second && (j->first).second == (k->first).first)){
      continue;
     }
     // here we can be sure to have 3 different single bonds
     // eight possible overlapping orientations of the bonds are possible for
     // each of 3 different linear arrangements of the bonds
     // abc
     else if ((i->first).second == (j->first).first && (j->first).second == (k->first).first){
      if (!contains_dihedral((i->first).first, (i->first).second, (j->first).second, (k->first).second)){
             insert_dihedral((i->first).first, (i->first).second, (j->first).second, (k->first).second);
      }
     }
     else if ((i->first).second == (j->first).first && (j->first).second == (k->first).second){
      if (!contains_dihedral((i->first).first, (i->first).second, (j->first).second, (k->first).first)){
             insert_dihedral((i->first).first, (i->first).second, (j->first).second, (k->first).first);
      }
     }
     else if ((i->first).second == (j->first).second && (j->first).first == (k->first).first){
      if (!contains_dihedral((i->first).first, (i->first).second, (j->first).first, (k->first).second)){
             insert_dihedral((i->first).first, (i->first).second, (j->first).first, (k->first).second);
      }
     }
     else if ((i->first).first == (j->first).first && (j->first).second == (k->first).first){
      if (!contains_dihedral((i->first).second, (i->first).first, (j->first).second, (k->first).second)){
             insert_dihedral((i->first).second, (i->first).first, (j->first).second, (k->first).second);
      }
     }
     else if ((i->first).second == (j->first).second && (j->first).first == (k->first).second){
      if (!contains_dihedral((i->first).first, (i->first).second, (j->first).first, (k->first).first)){
             insert_dihedral((i->first).first, (i->first).second, (j->first).first, (k->first).first);
      }
     }
     else if ((i->first).first == (j->first).first && (j->first).second == (k->first).second){
      if (!contains_dihedral((i->first).second, (i->first).first, (j->first).second, (k->first).first)){
             insert_dihedral((i->first).second, (i->first).first, (j->first).second, (k->first).first);
      }
     }
     else if ((i->first).first == (j->first).second && (j->first).first == (k->first).first){
      if (!contains_dihedral((i->first).second, (i->first).first, (j->first).first, (k->first).second)){
             insert_dihedral((i->first).second, (i->first).first, (j->first).first, (k->first).second);
      }
     }
     else if ((i->first).first == (j->first).second && (j->first).first == (k->first).second){
      if (!contains_dihedral((i->first).second, (i->first).first, (j->first).first, (k->first).first)){
             insert_dihedral((i->first).second, (i->first).first, (j->first).first, (k->first).first);
      }
     }
     // acb
     else if ((i->first).second == (k->first).first && (k->first).second == (j->first).first){
      if (!contains_dihedral((i->first).first, (i->first).second, (k->first).second, (j->first).second)){
             insert_dihedral((i->first).first, (i->first).second, (k->first).second, (j->first).second);
      }
     }
     else if ((i->first).second == (k->first).first && (k->first).second == (j->first).second){
      if (!contains_dihedral((i->first).first, (i->first).second, (k->first).second, (j->first).first)){
             insert_dihedral((i->first).first, (i->first).second, (k->first).second, (j->first).first);
      }
     }
     else if ((i->first).second == (k->first).second && (k->first).first == (j->first).first){
      if (!contains_dihedral((i->first).first, (i->first).second, (k->first).first, (j->first).second)){
             insert_dihedral((i->first).first, (i->first).second, (k->first).first, (j->first).second);
      }
     }
     else if ((i->first).first == (k->first).first && (k->first).second == (j->first).first){
      if (!contains_dihedral((i->first).second, (i->first).first, (k->first).second, (j->first).second)){
             insert_dihedral((i->first).second, (i->first).first, (k->first).second, (j->first).second);
      }
     }
     else if ((i->first).second == (k->first).second && (k->first).first == (j->first).second){
      if (!contains_dihedral((i->first).first, (i->first).second, (k->first).first, (j->first).first)){
             insert_dihedral((i->first).first, (i->first).second, (k->first).first, (j->first).first);
      }
     }
     else if ((i->first).first == (k->first).first && (k->first).second == (j->first).second){
      if (!contains_dihedral((i->first).second, (i->first).first, (k->first).second, (j->first).first)){
             insert_dihedral((i->first).second, (i->first).first, (k->first).second, (j->first).first);
      }
     }
     else if ((i->first).first == (k->first).second && (k->first).first == (j->first).first){
      if (!contains_dihedral((i->first).second, (i->first).first, (k->first).first, (j->first).second)){
             insert_dihedral((i->first).second, (i->first).first, (k->first).first, (j->first).second);
      }
     }
     else if ((i->first).first == (k->first).second && (k->first).first == (j->first).second){
      if (!contains_dihedral((i->first).second, (i->first).first, (k->first).first, (j->first).first)){
             insert_dihedral((i->first).second, (i->first).first, (k->first).first, (j->first).first);
      }
     }     
     // bac
     else if ((j->first).second == (i->first).first && (i->first).second == (k->first).first){
      if (!contains_dihedral((j->first).first, (j->first).second, (i->first).second, (k->first).second)){
             insert_dihedral((j->first).first, (j->first).second, (i->first).second, (k->first).second);
      }
     }
     else if ((j->first).second == (i->first).first && (i->first).second == (k->first).second){
      if (!contains_dihedral((j->first).first, (j->first).second, (i->first).second, (k->first).first)){
             insert_dihedral((j->first).first, (j->first).second, (i->first).second, (k->first).first);
      }
     }
     else if ((j->first).second == (i->first).second && (i->first).first == (k->first).first){
      if (!contains_dihedral((j->first).first, (j->first).second, (i->first).first, (k->first).second)){
             insert_dihedral((j->first).first, (j->first).second, (i->first).first, (k->first).second);
      }
     }
     else if ((j->first).first == (i->first).first && (i->first).second == (k->first).first){
      if (!contains_dihedral((j->first).second, (j->first).first, (i->first).second, (k->first).second)){
             insert_dihedral((j->first).second, (j->first).first, (i->first).second, (k->first).second);
      }
     }
     else if ((j->first).second == (i->first).second && (i->first).first == (k->first).second){
      if (!contains_dihedral((j->first).first, (j->first).second, (i->first).first, (k->first).first)){
             insert_dihedral((j->first).first, (j->first).second, (i->first).first, (k->first).first);
      }
     }
     else if ((j->first).first == (i->first).first && (i->first).second == (k->first).second){
      if (!contains_dihedral((j->first).second, (j->first).first, (i->first).second, (k->first).first)){
             insert_dihedral((j->first).second, (j->first).first, (i->first).second, (k->first).first);
      }
     }
     else if ((j->first).first == (i->first).second && (i->first).first == (k->first).first){
      if (!contains_dihedral((j->first).second, (j->first).first, (i->first).first, (k->first).second)){
             insert_dihedral((j->first).second, (j->first).first, (i->first).first, (k->first).second);
      }
     }
     else if ((j->first).first == (i->first).second && (i->first).first == (k->first).second){
      if (!contains_dihedral((j->first).second, (j->first).first, (i->first).first, (k->first).first)){
             insert_dihedral((j->first).second, (j->first).first, (i->first).first, (k->first).first);
      }
     }    
    }
   }
  }
  //blab3 << "Exiting CharmmParaResTop::generate_dihedrals()" << endl;
}

// considers only amino acids 
// recognized by full CHARMM nomenclature backbone
// (not proline, pyroglutamate and alike)
void CharmmParaResTop::generate_sidechain_tor()
{
  //blab3 << "Entering CharmmParaResTop::generate_sidechain_tor for res " << name << endl;
  
  // check if all backbone atoms are present to make sure we have an amino acid
  // and not GLY, PRO, pyroglutamate or alike, 
  // more general search for torsions will be implemented later
 
  if (contains_atom("CA") && contains_atom("HA") && (contains_atom("N") || contains_atom("NH")) &&
      contains_atom("HN") && contains_atom("C") && contains_atom("O") && contains_atom("CB")){
       
   list<string> tmpalist;
   list<vector<string> > tlist;
   vector<vector<string> > tor;

   // make a list of dihedrals involving only heavy atoms   
   for (list<vector<string> >::const_iterator i = dihedrals.begin(); i != dihedrals.end(); ++i){   
    if (get_element((*i)[0]) != "H" && get_element((*i)[1]) != "H" &&
        get_element((*i)[2]) != "H" && get_element((*i)[3]) != "H"){
     tlist.push_back(*i);
    }
   }
   
   if(tlist.size() == 0){goto END;}
   
   // search for 1st torsion N CA CB X1
   list<vector<string> > tlist1;
   // make list of candidates
   for (list<vector<string> >::const_iterator i = tlist.begin(); i != tlist.end(); ++i){   
    if (((*i)[0] == "N" || (*i)[0] == "NH") && (*i)[1] == "CA" && (*i)[2] == "CB"){
     tlist1.push_back(*i);      
    }        
    if (((*i)[3] == "N" || (*i)[3] == "NH") && (*i)[2] == "CA" && (*i)[1] == "CB"){
     vector<string> tmpvec;
     tmpvec.push_back((*i)[3]);
     tmpvec.push_back((*i)[2]);
     tmpvec.push_back((*i)[1]);
     tmpvec.push_back((*i)[0]);
     tlist1.push_back(tmpvec);
    }        
   }   
   if (tlist1.size() == 0){
    goto END;
   }
   else if (tlist1.size() == 1){
    pair<list<string>, bool> tmppair(
      get_moving_ats((*tlist1.begin())[0], (*tlist1.begin())[1], (*tlist1.begin())[2], (*tlist1.begin())[3]));
    if (tmppair.second){
     goto END;
    }
    else {
     tor.push_back((*tlist1.begin()));
     tmpalist = tmppair.first;
     sidechain_tor.push_back(pair<vector<string>,list<string> >(*(tor.end()-1),tmppair.first));
    }
   }
   else{
   // determine highest_priority_torsion
    vector<string> tmpvec = get_hp_tor(tlist1);
    pair<list<string>, bool> tmppair(
      get_moving_ats(tmpvec[0], tmpvec[1], tmpvec[2], tmpvec[3]));
    if (tmppair.second){
     goto END;
    }
    else {
     tor.push_back(tmpvec);
     tmpalist = tmppair.first;
     sidechain_tor.push_back(pair<vector<string>,list<string> >(*(tor.end()-1),tmppair.first));
    }    
   }  

  // cout << "TORSION " << tor.size() << " of residue " << name << " " 
  //      << (*(tor.end()-1))[0] << " " << (*(tor.end()-1))[1] << " " 
  //      << (*(tor.end()-1))[2] << " " << (*(tor.end()-1))[3] << endl;      

   SEARCHT:
   
   tlist1.clear();
   
   // search for 2nd torsion CA CB X1 X2
   // make list of candidates
   for (list<vector<string> >::const_iterator i = tlist.begin(); i != tlist.end(); ++i){   
    if ((*i)[0] == (*(tor.end()-1))[1] && (*i)[1] == (*(tor.end()-1))[2] && (*i)[2] == (*(tor.end()-1))[3]){
     tlist1.push_back(*i);      
    }        
    if ((*i)[3] == (*(tor.end()-1))[1] && (*i)[2] == (*(tor.end()-1))[2] && (*i)[1] == (*(tor.end()-1))[3]){
     vector<string> tmpvec;
     tmpvec.push_back((*i)[3]);
     tmpvec.push_back((*i)[2]);
     tmpvec.push_back((*i)[1]);
     tmpvec.push_back((*i)[0]);
     tlist1.push_back(tmpvec);
    }        
   }   
   if (tlist1.size() == 0){
    goto END2;
   }
   else if (tlist1.size() == 1){
    pair<list<string>, bool> tmppair(
      get_moving_ats((*tlist1.begin())[0], (*tlist1.begin())[1], (*tlist1.begin())[2], (*tlist1.begin())[3], tmpalist));
    if (tmppair.second){
     goto END2;
    }
    else {
     tor.push_back(*tlist1.begin());
     tmpalist = tmppair.first;
     sidechain_tor.push_back(pair<vector<string>,list<string> >(*(tor.end()-1),tmppair.first));
    }
   }
   else{
    SEARCHT2:
    // determine highest_priority_torsion
    vector<string> tmpvec = get_hp_tor(tlist1);
    pair<list<string>, bool> tmppair(
      get_moving_ats(tmpvec[0], tmpvec[1], tmpvec[2], tmpvec[3], tmpalist));
    if (tmppair.second){
     // remove highest priority torsion because in cycle
     for (list<vector<string> >::iterator i = tlist1.begin(); i != tlist1.end(); ++i){
      if (*i == tmpvec){
       tlist1.erase(i);
       break;
      }
     }       
     goto SEARCHT2;
    }
    else {
     tor.push_back(tmpvec);
     tmpalist = tmppair.first;
     sidechain_tor.push_back(pair<vector<string>,list<string> >(*(tor.end()-1),tmppair.first));
    }    
   }
   
  // cout << "TORSION " << tor.size() << " of residue " << name << " " 
  //      << (*(tor.end()-1))[0] << " " << (*(tor.end()-1))[1] << " " 
  //      << (*(tor.end()-1))[2] << " " << (*(tor.end()-1))[3] << endl;      
   
   goto SEARCHT;
   

  }
  else {
   END: 
   blab3 << "CharmmParaResTop::generate_sidechain_tor does not handle residue " << name << endl;
  }
  
  END2:
 //blab3 << "Exiting CharmmParaResTop::generate_sidechain_tor" << endl;
  return;
  
}



void CharmmParaResTop::print() const
{
 //blab3 << "Entering CharmmParaResTop::print()" << endl;
 cout << endl << "!--------------------------------" << endl << endl;
 cout << "Printing topology of residue " << name << endl << endl;
 print_atoms();
 print_bonds();
 print_bonded_list();
 print_angles();
 print_dihedrals();
 print_impropers();
 print_sidechain_tor();
 //blab3 << "Exiting CharmmParaResTop::print()" << endl;
}

void CharmmParaResTop::print_atoms() const
{
  cout << "\natoms of residue named " << name << ":\n" << endl;
  for (list<string>::const_iterator i = atoms.begin(); i != atoms.end(); ++i){   
   cout << *i << endl;
  }
}

void CharmmParaResTop::print_bonds() const
{
  cout << "\nbonds of residue named " << name << ":\n" << endl;
  cout << "atom name : atom name : bond order (1,2,3 = BOND,DOUBLE,TRIPLE 4 = aromatic)" << endl;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){   
   cout << (i->first).first << " : " << (i->first).second << " : " << (i->second) << endl;
  }
}

void CharmmParaResTop::print_bonded_list() const
{
  cout << "\nbonded_list of residue named " << name << ":\n" << endl;
  cout << "atom name : names of bonded atoms" << endl;
  for (map<string,list<string> >::const_iterator i = bonded_list.begin(); i != bonded_list.end(); ++i){
   cout << i->first << " :";
   for (list<string>::const_iterator j = (i->second).begin(); j != (i->second).end(); ++j){   
    cout << " " << *j;
   }
   cout << endl;
  }
}

void CharmmParaResTop::print_angles() const
{
  cout << "\nangles of residue named " << name << ":\n" << endl;
  cout << "atom name : atom name : atom name" << endl;
  for (list<vector<string> >::const_iterator i = angles.begin(); i != angles.end(); ++i){   
   cout << (*i)[0] << " : " << (*i)[1] << " : " << (*i)[2] << endl;
  }
}

void CharmmParaResTop::print_dihedrals() const
{
  cout << "\ndihedrals of residue named " << name << ":\n" << endl;
  cout << "atom name : atom name : atom name : atom name" << endl;
  for (list<vector<string> >::const_iterator i = dihedrals.begin(); i != dihedrals.end(); ++i){   
   cout << (*i)[0] << " : " << (*i)[1] << " : " << (*i)[2] << " : " << (*i)[3] << endl;
  }
}

void CharmmParaResTop::print_impropers() const
{
  cout << "\nimpropers of residue named " << name << ":\n" << endl;
  cout << "atom name : atom name : atom name : atom name" << endl;
  for (list<vector<string> >::const_iterator i = impropers.begin(); i != impropers.end(); ++i){   
   cout << (*i)[0] << " : " << (*i)[1] << " : " << (*i)[2] << " : " << (*i)[3] << endl;
  }
}

void CharmmParaResTop::print_sidechain_tor() const
{
  cout << "\nsidechain torsion angles of residue named " << name << ":\n" << endl;
  cout << "atom names in torsion : list of atoms to move" << endl;
  for (vector<pair<vector<string>,list<string> > >::const_iterator i = sidechain_tor.begin(); i != sidechain_tor.end(); ++i){
   cout << (*i).first[0] << " - " << (*i).first[1] << " - " 
        << (*i).first[2] << " - " << (*i).first[3] << " - " << " :";
   for (list<string>::const_iterator j = (*i).second.begin(); j != (*i).second.end(); ++j){   
    cout << " " << *j;
   }
   cout << endl;
  }
}

// CharmmParaResTop.cc ends here
