#include "MEAD/ElectrolyteByAtoms.h"
#include "MEAD/AtomSet.h"
#include "MEAD/CubeLatSpec.h"
#include "MEAD/ElySlab.h"
#include "MEAD/SolvAccVol.h"
#include "MEAD/PhysCond.h"
#include <stdio.h>
#include <iostream>

/* 
   added, TRU 2007
   This is a 1-value electrolyte defined by 1 atom set
*/


SimpleElectrolyteByAtoms::SimpleElectrolyteByAtoms (const AtomSet & a)
  : ElectrolyteEnvironment_lett(), _ats(a), _ats_defined(true),
    //solv_acc(a, PhysCond::get_solrad()), 
    ionic_str(PhysCond::get_ionicstr()), radius(PhysCond::get_solrad()), sternl(PhysCond::get_sterln())
{
}

SimpleElectrolyteByAtoms::SimpleElectrolyteByAtoms (const AtomSet & a1, float ionic, const SolvAccVol& a2)
 : ElectrolyteEnvironment_lett(), _ats(a1), _ats_defined(true),
   ionic_str(ionic), sternl(PhysCond::get_sterln())
{
  solv_acc = new SolvAccVol(a2);
}

// For GB/DL's sake
const AtomSet& SimpleElectrolyteByAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("SimpleElectrolyteByAtoms::get_atomset: atom set not defined");
  return _ats;
}



ElyCubeRep SimpleElectrolyteByAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "SimpleElectrolyteByAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
 
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;          
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  int grid_dim_plus_stli = grid_dim + stli;

  ElyCubeRep dcr(cls);
  AccTag* acc_array = new AccTag[ncube];

  if (_ats_defined){

    // initialize acc_array
    
    for (int h=0; h<ncube; ++h) {
      acc_array[h] = exterior;
    }

    for(AtomSet::const_iterator ip = _ats.begin(); ip!=_ats.end(); ++ip ) {
      
      int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
      int radstlsq = rad * rad;
      Coord atom_coord = ip->second.coord - grid_origin_in_space;
      int k = (int) round((atom_coord.z/spacing));
      int i = (int) round((atom_coord.x/spacing));
      int j = (int) round((atom_coord.y/spacing));
      int h = (i * nsq) + (j * grid_dim) + k;
      // skip atom if not in grid
      if (k < stlin || i < stlin || j < stlin ||
          k > grid_dim_plus_stli || 
          i > grid_dim_plus_stli || 
          j > grid_dim_plus_stli){continue;}
      int lower_l = k - rad;
      int lower_m = i - rad;
      int lower_n = j - rad;
      int upper_l = k + rad + 1;
      int upper_m = i + rad + 1;
      int upper_n = j + rad + 1;
      bool fastswitch = true;
      if (lower_l < 0) {lower_l = 0; fastswitch = false;}
      if (lower_m < 0) {lower_m = 0; fastswitch = false;}
      if (lower_n < 0) {lower_n = 0; fastswitch = false;}
      if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
      if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
      if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
      if (fastswitch) {
        // in the inner grid we have a full sphere around the point
        // and can save time by computing by doing loops and distance 
        // only in +,+,+ octand of sphere
        // tried to improve by computing dist only for 1/48th sphere
        // also permutating x,y,z and doing loops and distance only
        // in in x >= y >= z but result is slower       
        int dl = -1;
        int lterm = -grid_dim;
        for (int l=k; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          lterm += grid_dim;
          int lsqterm = dl*nsq;
          int dm = -1;
          int mterm = -grid_dim; 
          for (int m=i; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            // out of sphere already in plane
            if (dmlsq > radstlsq){break;}
            mterm += grid_dim;
            int msqterm = dm * nsq;
            int dn = -1;
            int nterm = -grid_dim;
            for (int n=j; n<upper_n; ++n) {                 
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq > radstlsq){break;}
              int nsqterm = dn*nsq;
              nterm += grid_dim;              
              // 1: x, y, z
              // here we permutate +/- signs in the three components
              // + + +
              int hh = h + msqterm + nterm + dl;
              acc_array[hh] = interior;
              // + + -
              hh = hh - dl  - dl;
              acc_array[hh] = interior;
              // + - -
              hh = hh - nterm - nterm;
              acc_array[hh] = interior;
              // + - +
              hh = hh + dl + dl;
              acc_array[hh] = interior;
              // - - +
              hh = hh - msqterm - msqterm;                    
              acc_array[hh] = interior;
              // - - -
              hh = hh - dl - dl;              
              acc_array[hh] = interior;
              // - + -
              hh = hh + nterm + nterm;                
              acc_array[hh] = interior;               
              // - + +
              hh = hh + dl + dl;              
              acc_array[hh] = interior;
            }
          }
        }        
      }
      else {
        // these are atoms at the grid border
        int dl = -1;
        for (int l=lower_l; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          int dm = -1;
          for (int m=lower_m; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            if (dmlsq > radstlsq){break;}
            int msqterm = m*nsq;
            int dn = -1;
            for (int n=lower_n; n<upper_n; ++n) {                 
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq <= radstlsq){
                int hh = msqterm + n*grid_dim + l;            
                acc_array[hh] = interior;
              }
            }
          }
        }      
      }
    }
  }
  // got a SolvAccVol, will use it as is
  else {
    if (solv_acc->check_is_calculated()) {
      blab2 << "SimpleElectrolyteByAtoms::get_cuberep using existing "
      << "analytical representation to make a cubic representation" << endl;
    }
    else {
       solv_acc->anal_calc();
    }
    solv_acc->calc_cuberep(cls, acc_array);
  }

  // now we are ready construct the ElyCubeRep, 
  //acc_array holds the ion accessible volume

  for (int h=0; h<ncube; ++h) {
    if (acc_array[h] == interior)
      dcr[h] = 0.0;
    else if (acc_array[h] == exterior)
      dcr[h] = ionic_str;
    else {
      cerr << "acc_array[" << h << "] has unexpected value, "
        << acc_array[h] << endl;
      ::error("ERROR in ");
    }
  }
  
  delete [] acc_array;      
  dcr.declare_defined();
  blab3 << "exit from SimpleElectrolyteByAtoms::get_cuberep\n" << endl;
  return dcr;

}


SimpleElectrolyteByAtoms::~SimpleElectrolyteByAtoms()
{
}



/* 
   added, TRU 2007   
   This is a 1-value electrolyte defined by 1 atom set 
   and a smooth Surface defined by SolvAccVol + Sternlayer.
*/


ElectrolyteByAtoms::ElectrolyteByAtoms (const AtomSet & a)
  : ElectrolyteEnvironment_lett(), _ats(a), _ats_defined(true),
    ion_acc(a, PhysCond::get_sterln()), ionic_str(PhysCond::get_ionicstr()), radius(PhysCond::get_solrad()), sternl(PhysCond::get_sterln())
{
}

ElectrolyteByAtoms::ElectrolyteByAtoms (const AtomSet & a, float ionic, float stl)
 : ElectrolyteEnvironment_lett(), _ats(a), _ats_defined(true),
    ion_acc(a, stl), ionic_str(ionic), radius(PhysCond::get_solrad()), sternl(stl)
{
}

ElectrolyteByAtoms::ElectrolyteByAtoms (const AtomSet & a, float ionic, float rad, float stl)
 : ElectrolyteEnvironment_lett(), _ats(a), _ats_defined(true),
    ion_acc(a, stl), ionic_str(ionic), radius(rad), sternl(stl)
{
}

ElectrolyteByAtoms::ElectrolyteByAtoms (const IonAccVol & a)
  : ElectrolyteEnvironment_lett(), _ats(), _ats_defined(false),
    ion_acc(a), ionic_str(PhysCond::get_ionicstr()), radius(PhysCond::get_solrad()), sternl(PhysCond::get_sterln())
{
}

ElectrolyteByAtoms::ElectrolyteByAtoms (const IonAccVol & a, float ionic, float stl)
 : ElectrolyteEnvironment_lett(), _ats(), _ats_defined(false),
    ion_acc(a), ionic_str(ionic), sternl(PhysCond::get_sterln())
{
}

ElectrolyteByAtoms::ElectrolyteByAtoms (const AtomSet & a1, float ionic, const IonAccVol & a2)
 : ElectrolyteEnvironment_lett(), _ats(a1), _ats_defined(true),
    ion_acc(a2), ionic_str(ionic), sternl(PhysCond::get_sterln())
{
}



// For GB/DL's sake
const AtomSet& ElectrolyteByAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("ElectrolyteByAtoms::get_atomset: atom set not defined");
  return _ats;
}



ElyCubeRep ElectrolyteByAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "ElectrolyteByAtoms::get_cuberep entered" << endl;

  if (ion_acc.check_is_calculated()) {
    blab2 << "ElectrolyteByAtoms::get_cuberep using existing "
      << "analytical representation to make a cubic representation" << endl;
  }
  else {
    ion_acc.anal_calc();
  }

  // we have to extend the solvent inaccessible volume by the Stern layer
 
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);          

  ElyCubeRep dcr(cls);
  AccTag* acc_array = new AccTag[ncube];
  //AccTag* tmp_array = new AccTag[ncube];  
  ion_acc.calc_cuberep(cls, acc_array);
  
/*
  ion_acc.calc_cuberep(cls, tmp_array);


  blab3 << "Will now extend SolvAccVol by Stern Layer " << endl;

  // initialize acc_array
  
  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }
 
  for (int k=0; k<grid_dim; ++k) {
    for (int i = 0; i<grid_dim; ++i) {
      int iterm = i*nsq;
      for (int j=0; j<grid_dim; ++j) {      
	int h = iterm + j*grid_dim + k;
        if (tmp_array[h] == interior) {
          acc_array[h] = interior;
          // also set points within solrad to interior        
          int lower_l = k - stli;
          int lower_m = i - stli;
          int lower_n = j - stli;
          int upper_l = k + stli;
          int upper_m = i + stli;
          int upper_n = j + stli;
	  bool dofast = true;
          if (lower_l < 0){lower_l = 0; dofast = false;}
          if (lower_m < 0){lower_m = 0; dofast = false;}
          if (lower_n < 0){lower_n = 0; dofast = false;}
          if (upper_l > grid_dim){upper_l = grid_dim; dofast = false;}
          if (upper_m > grid_dim){upper_m = grid_dim; dofast = false;}
          if (upper_n > grid_dim){upper_n = grid_dim; dofast = false;}                          
          if (dofast){
	    // if we have a full sphere around the point within the grid
	    // we can save time by computing dist only for 1/8th sphere 
	    for (int l=k; l<=upper_l; ++l) {
              int dl = (l - k);
              int dlsq = dl * dl;
	      int kterm = dl*nsq;
              for (int m=i; m<=upper_m; ++m) {
                int dm = (m - i);
                int dmsq = dm * dm;
		int mterm = dm*nsq;
                for (int n=j; n<=upper_n; ++n) {                 
                  int dn = (n - j);
		  int nterm = dn*nsq;
	   	  int dissq = (dn * dn) + dmsq + dlsq;
                  if (dissq <= stlsq){
		    // +x +y +z
		    int hh = h + mterm + dn*grid_dim + dl;
		    acc_array[hh] = interior;
		    // +x +y -z
		    hh = h + mterm + dn*grid_dim - dl;
		    acc_array[hh] = interior;
		    // +x -y -z
		    hh = h + mterm - dn*grid_dim - dl;
		    acc_array[hh] = interior;
		    // +x -y +z
		    hh = h + mterm - dn*grid_dim + dl;
		    acc_array[hh] = interior;
		    // -x -y +z
		    hh = h - mterm - dn*grid_dim + dl;		    
		    acc_array[hh] = interior;
		    // -x -y -z
		    hh = h - mterm - dn*grid_dim - dl;		    
		    acc_array[hh] = interior;
		    // -x +y -z
		    hh = h - mterm + dn*grid_dim - dl;		    
		    acc_array[hh] = interior;		    
		    // -x +y +z
		    hh = h - mterm + dn*grid_dim + dl;		    
		    acc_array[hh] = interior;
		  }
                }
              }
            }
	  }
	  else{
	    for (int l=lower_l; l<=upper_l; ++l) {
              int dlsq = (l - k) * (l - k);
              for (int m=lower_m; m<=upper_m; ++m) {
                int mterm = m*nsq;
		int dmsq = (m - i) * (m - i);
                for (int n=lower_n; n<=upper_n; ++n) {                 
		int dissq = ((n - j) * (n - j)) + dmsq + dlsq;
                  if (dissq <= stlsq){
                    int hh = mterm + n*grid_dim + l;		  
		    acc_array[hh] = interior;
		  }
                }
              }
            }
	  } 
        }
      }
    }
  }   
  delete [] tmp_array;
  */

  // now we are ready to construct the ElyCubeRep, 
  //acc_array holds the ion accessible volume

  for (int h=0; h<ncube; ++h) {
    if (acc_array[h] == interior){      
      dcr[h] = 0.0;
    }
    else if (acc_array[h] == exterior){
      dcr[h] = ionic_str;
    }
    else {
      cerr << "acc_array[" << h << "] has unexpected value, "
	<< acc_array[h] << endl;
      ::error("ERROR in ");
    }
  }
  delete [] acc_array;    
  dcr.declare_defined();
  blab3 << "exit from ElectrolyteByAtoms::get_cuberep\n" << endl;

  return dcr;

}


ElectrolyteByAtoms::~ElectrolyteByAtoms()
{
}

// ElectrolyteByAtoms.cc ends here
