/* Base class for Electrolyte environment and subclasses.
    
    Modified 2006--2007 by Thomas Ullmann, BISB Group, 
    Dep. of Biopolymers, University of Bayreuth, Bayreuth, Germany
    
    The Electrolyte is now represented like the Dielectric by a float
    array with explicit values at each grid point. 
    
    The ion inaccessible volume is built from the solvent inaccessible
    volume + Stern Layer instead of molecular volume + Stern Layer
    
    This representation allows to include multiple ionic
    solutions, ion gradients etc., but the preparation of the actual
    Finite Difference Calculation is more memory demanding and
    somewhat slower than before.

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: ElectrolyteEnvironment.cc,v 2.7 2004/12/06 17:58:12 bashford Exp $
*/
#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/UniformElectrolyte.h"
#include <iostream>


int ElectrolyteEnvironment::instanceCount = 0;

// Constructs an envelope with a ChargDist letter, that is, a default letter.
ElectrolyteEnvironment::ElectrolyteEnvironment()
{
  rep = new UniformElectrolyte(0.0);
}

// Ctor must have a valid envelope (non-zero rep) as its argument.
ElectrolyteEnvironment::ElectrolyteEnvironment(const ElectrolyteEnvironment & c)
{
  if(c.rep) { // Arg is an envelope.
    c.rep->referenceCount++;
    rep = c.rep;
  }
  else
    error("ERROR: ElectrolyteEnvironment copy constructor:\n",
	  "argument has zero rep pointer");
  ++instanceCount;
}

// Both both sides of assignment must be envelopes (non-zero rep)

ElectrolyteEnvironment& ElectrolyteEnvironment::operator=
(const ElectrolyteEnvironment& c)
{
  if (this != &c) {
    if (rep) {
      if (c.rep) {
        c.rep->referenceCount++;
        if (rep->referenceCount == 1)
          delete rep;
        else
          --rep->referenceCount;
        rep = c.rep;
      }
      else {
        error("ERROR: ElectrolyteEnvironment::operator=:\n",
	      "argument has zero rep");
      }
    }
    else {
      error("ERROR ElectrolyteEnvironment::operator=: this->rep is zero");
    }
  }
  return *this;
}

// TE
// Write ElectrolyteEnvironment (ionic strength) to OpenDX file
void ElectrolyteEnvironment::write_dx(const CubeLatSpec& cls, string file)
{
  blab3 << "ElectrolyteEnvironment::write_dx entered" << endl;
  //  rep->write_dx(cls, name);
    // run get_cuberep to get ionic strength mapped on grid
  ElyCubeRep dcr(get_cuberep(cls));
  // this is necessary to circumvent error in ElyCubeRep.h
  //   "ERROR: illegal application of operator[] to a ElyCubeRep",
  //      "with istr_array defined"
  // FixMe: find out what the check is good for
  // seems to make no sense at the moment to
  // request values without istr_array defined
  dcr.declare_undefined();
  int grid_dim = cls.get_grid_dim();
  int ncube = grid_dim*grid_dim*grid_dim;
  Coord center = cls.get_center();
  float space = cls.get_spacing();
  // OpenDX expects lower corner as origin, not center
  Coord origin(center.x-((grid_dim * space)/2),
               center.y-((grid_dim * space)/2),
               center.z-((grid_dim * space)/2));
  string file_name = file + ".dx";
  ofstream of(file_name.c_str());
  if (!of.good())
    ::error("ElectrolyteEnvironment:write_dx failed to open for writing, ",
            file_name.c_str());
  blab1 << "ElectrolyteEnvironment::write_dx: Writing " << file_name.c_str() << " in OpenDX format...\n ";
  of << "# "<< file << " " << grid_dim << " " << space << " " << center << endl;
  of << "#\n# ElectrolyteEnvironment witten by MEAD\n#" << endl;
  of << "object 1 class gridpositions counts " << grid_dim << " " << grid_dim << " " << grid_dim << endl;
  of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
  of << "object 2 class gridconnections counts " << grid_dim << " " << grid_dim << " " << grid_dim << endl;
  of << "object 3 class array type double rank 0 times " << ncube << endl;
  int count = 0;
  // Now the actual values
  // Don't know if I could write more than three columns according to OpenDX format,
  // but the TCL reader can only read three columns or less...
  for (int h=2; h<ncube; h+=3) {
    of << dcr[h-2] << " " << dcr[h-1] << " " << dcr[h] << endl;
    count++;
  }
  if (ncube - (3 * count) == 2)
    of << dcr[ncube-2] << " " << dcr[ncube-1] << endl;
  if (ncube - (3 * count) == 1)
    of << dcr[ncube-1] << endl;
  // Now some blurb at the end
  of << "attribute \"dep\" string \"positions\"" << endl;
  of << "object \"regular positions regular connections\" class field" << endl;
  of << "component \"positions\" value 1" << endl;
  of << "component \"connections\" value 2" << endl;
  of << "component \"data\" value 3" << endl;
  if (! of.good())
    ::error("ElectrolyteEnvironment:write_dx failed to write OpenDX file");
  of.close();
  blab3 << "exit from ElectrolyteEnvironment::write_dx with " << endl;
}
//TE


// This will create an envelope containing *cdp.
ElectrolyteEnvironment::ElectrolyteEnvironment(ElectrolyteEnvironment_lett * cdp)
{
  cdp->referenceCount++;
  rep = cdp;
  instanceCount++;
}

ElectrolyteEnvironment::~ElectrolyteEnvironment()
{
  if (rep) {
    blab2 << "ElectrolyteEnvironment destructor called with rep->referenceCount = " << rep->referenceCount << endl;
    if (rep->referenceCount == 1)
      delete rep;
    else
      --rep->referenceCount;
  }
  instanceCount--;
}

int ElectrolyteEnvironment_lett::instanceCount = 0;


ElectrolyteEnvironment_lett::ElectrolyteEnvironment_lett()
{
  blab2 << "ElectrolyteEnvironment_lett default constructor called" << endl;
  referenceCount=1;
  ++instanceCount;
}


ElectrolyteEnvironment_lett::ElectrolyteEnvironment_lett
(const ElectrolyteEnvironment_lett& a)
{
  blab2 << "ElectrolyteEnvironment_lett copy constructor called" << endl;
  referenceCount = 1;
  ++instanceCount;
}

// Don't delete this letter until all references have been released
void ElectrolyteEnvironment_lett::operator delete(void *p, size_t s)
{
  ElectrolyteEnvironment_lett *del = reinterpret_cast<ElectrolyteEnvironment_lett*> (p);
  blab2 << "ElectrolyteEnvironment_lett operator delete called with referenceCount = " << del->referenceCount << endl;
  if (del->referenceCount == 0) {
    blab2 << "ElectrolyteEnvironment_lett operator delete frees object" << endl;
    ::operator delete(p);
  }
}

ElectrolyteEnvironment_lett::~ElectrolyteEnvironment_lett()
{
  blab2 << "ElectrolyteEnvironment_lett destructor called with referenceCount = " << referenceCount << endl;
  if (--referenceCount)
    // It's now OK for this to happen
    blab2 << "WARNING: ElectrolyteEnvironment_lett destructor called\n"
      << "on object with referenceCount = " << referenceCount+1
	<< "\nas if the object is still in use somewhere" << endl;
  --instanceCount;
}

ElectrolyteEnvironment_lett&
ElectrolyteEnvironment_lett::operator=(const ElectrolyteEnvironment_lett& acs)
{return *this;}

// ElectrolyteEnvironment.cc ends here
