/* Base class for Elecly. Env. and envelope for subclasses -*- C++ -*-
    
    Modified 2006--2007 from DielectricEnvironment
    
    Thomas Ullmann, BISB Group, 
    Dep. of Biopolymers, University of Bayreuth, 
    Universitaetsstrasse 30 BGI, 95447 Bayreuth, Germany
    thomas.ullmann@uni-bayreuth.de
    
    The Electrolyte is now represented like the Dielectric by a float
    array with explicit values at each grid point. 
    
    The ion inaccessible volume is built from the solvent inaccessible
    volume + Stern Layer instead of molecular volume + Stern Layer
    
    This representation allows to include multiple ionic
    solutions, ion gradients etc., but the preparation of the actual
    Finite Difference Calculation is more memory demanding and
    somewhat slower than before.
    
    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: ElectrolyteEnvironment.h,v 2.10 2004/12/06 17:58:13 bashford Exp $
*/
#ifndef _ElectrolyteEnvironment_h
#define _ElectrolyteEnvironment_h 1


#include "MEAD/CubeLatSpec.h"
#include "MEAD/ElyCubeRep.h"
#include "MEAD/Coord.h"

class ElectrolyteEnvironment_lett;
class ElstatPot;

//DON'T !wrap!
class ElectrolyteEnvironment {
// LANGUAGE BUG!!  The following friend declaration is OK with gcc
// but not cfront.  According to the ARM Sec 11.4 p. 251, cfront is right!
//    friend ElstatPot::ElstatPot (ElectrolyteEnvironment, ChargeDist,
//                                 ElectrolyteEnvironment);
// So must get "overly friendly"...
  //!nowrap!+
  friend class ElstatPot;
  //!nowrap!-
public:
  // Canonical stuff
  //!nowrap!+
  ElectrolyteEnvironment();
  ElectrolyteEnvironment(const ElectrolyteEnvironment& d);
  //!nowrap!-
  ElectrolyteEnvironment(ElectrolyteEnvironment_lett*);
  //!nowrap!+
  ElectrolyteEnvironment& operator= (const ElectrolyteEnvironment&);
  virtual ~ElectrolyteEnvironment();
  virtual ElyCubeRep get_cuberep(const CubeLatSpec& cls);
  // TE
  virtual void write_dx(const CubeLatSpec& cls, string name);
  // TE
  virtual float ionic_strength() const;
  static int instances() {return instanceCount;}
  //!nowrap!-
private:
  ElectrolyteEnvironment_lett *rep;
  static int instanceCount;
};

//!wrap!
class ElectrolyteEnvironment_lett {
  //!nowrap!+
friend class ElectrolyteEnvironment;
  //!nowrap!-
public:
  //!nowrap!+
  ElectrolyteEnvironment_lett();
  ElectrolyteEnvironment_lett(const ElectrolyteEnvironment_lett&);
  ElectrolyteEnvironment_lett& operator= (const ElectrolyteEnvironment_lett&);
  virtual ~ElectrolyteEnvironment_lett();
  virtual ElyCubeRep get_cuberep(const CubeLatSpec& clsp) = 0;
  virtual float ionic_strength() const= 0;
  static int instances() {return instanceCount;}
  void operator delete(void *p, size_t s);
  //!nowrap!-
private:
  static int instanceCount;
  short referenceCount;
};


inline ElyCubeRep
ElectrolyteEnvironment::get_cuberep(const CubeLatSpec& cls)
{return rep->get_cuberep(cls);}

inline float
ElectrolyteEnvironment::ionic_strength() const
 {return rep->ionic_strength();}

#if SWIGPP_LITERAL_INCLUDE

%addmethods ElectrolyteEnvironment_lett {
//
// Return a Numeric_3D_Array_FLOAT of ionic_str values
// This is used by all the ElectrolyteEnvironment_lett derived classes,
// and works via an implicit downcasted self->get_cuberep() call.
//

  Numeric_3D_Array_FLOAT * get_cuberep(const CubeLatSpec& cls)
  {
    int ngr = cls.get_grid_dim();
    int nsq = ngr * ngr;
    int ncube = nsq * ngr;
    Numeric_3D_Array_FLOAT *array_data = new Numeric_3D_Array_FLOAT [ncube];
    if (array_data != 0) {
      const ElyCubeRep dcr = self->get_cuberep(cls);
      int h = 0;
      for (int i = 0; i < ngr; ++i) {
        for (int j = 0; j < ngr; ++j) {
          for (int k = 0; k < ngr; ++k) {
            int fortind = i + j*ngr + k*nsq;
            array_data[fortind] = dcr[h];
            ++h;
          }
        }
      }
    }
    return array_data;
  }
};

#endif // SWIGPP_LITERAL_INCLUDE


#endif

// ElectrolyteEnvironment.h ends here
