/* Electrolyte on a cubic lattice.

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: ElyCubeRep.cc,v 2.6 2004/12/06 17:58:10 bashford Exp $
*/
#include "MEAD/globals.h"
#include "MEAD/Bigmem.h"
#include "MEAD/ElyCubeRep.h"
#include <iostream>

ElyCubeRep::ElyCubeRep(const CubeLatSpec& cls)
{
  grid_dim = cls.get_grid_dim();
  nsq = grid_dim*grid_dim;
  ncube = grid_dim*grid_dim*grid_dim;
  istr_def =  0;
  rep = new ElyCubeRepRep;
  rep->referenceCount = 1;
  rep->istr_array = (float *) big_rigid_malloc((sizeof (float))*ncube);
  memset ((char *) rep->istr_array, 0, ncube * sizeof (float));
}

ElyCubeRep::ElyCubeRep(const ElyCubeRep& dcr)
{
  grid_dim = dcr.grid_dim;
  nsq = dcr.nsq;
  ncube = dcr.ncube;
  if (dcr.istr_def) { // The old one is not to change so safe to point to it.
    rep = dcr.rep;
    rep->referenceCount++;
  }
  else { // The old one or this one may change so make a copy
    rep = new ElyCubeRepRep;
    rep->referenceCount = 1;
    rep->istr_array = new float [ncube];
    for (int h = 0; h<ncube; ++h)
      rep->istr_array[h] = dcr.rep->istr_array[h];
  }
  istr_def = dcr.istr_def;
}

ElyCubeRep& ElyCubeRep::operator=(const ElyCubeRep& dcr)
{
  if (istr_def)
    ::error("Tryed to assign to a DieCubeRep with a fixed istr_array");
  else {
    if (rep->referenceCount != 1)
      ::error("ERROR: ElyCubeRep::operator=: this object has istr_def",
	      " but referenceCount != 1");
    grid_dim = dcr.grid_dim;
    nsq = dcr.nsq;
    // The usual decrement and test of referenceCount is not needed
    // since we know it is 1.
    if (dcr.istr_def) { // The old one is not to change so safe to point to it.
      big_rigid_free(rep->istr_array);
      delete rep;
      rep = dcr.rep;
      rep->referenceCount++;
    }
    else { // The old one or this one may change so make a copy
      // Use the old space if it is the right size
      if(dcr.ncube != ncube) {
	big_rigid_free (rep->istr_array);
	rep->istr_array = new float [dcr.ncube];
      }
      for (int h = 0; h<dcr.ncube; ++h)
	rep->istr_array[h] = dcr.rep->istr_array[h];
    }
    ncube = dcr.ncube;
    istr_def = dcr.istr_def;
  }
  return *this;
}

void ElyCubeRep::declare_defined()
{
  istr_def = 1;
}

void ElyCubeRep::declare_undefined()
{
  istr_def = 0;
}


ElyCubeRep::~ElyCubeRep()
{
  if(--rep->referenceCount == 0) {
    big_rigid_free(rep->istr_array);
    delete rep;
  }
}

// ElyCubeRep.cc ends here
