/*  class InstanceID is a key type for an associative array of 
    of atom sets each representing an instance of a site
    
    TRU 2007
    
    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

*/
#ifndef _InstanceID_h
#define _InstanceID_h 1

#include "MEAD_e/Atom.h"
#include <string>
#include <iostream>
#include <sstream>
#include <ostream>

using std::string;
using std::stringstream;
using std::ostream;

using namespace std;

//!wrap!
class InstanceID {
public:
  InstanceID ();
  InstanceID (const string&, const string&, const int, const int, const int, const int);
  InstanceID (const InstanceID&);
  ~InstanceID () {}
  //!nowrap!
  InstanceID& operator= (const InstanceID&);
  bool operator== (const InstanceID&) const;
  bool operator!= (const InstanceID& a) const;
  bool operator<(const InstanceID&) const;
  bool operator>(const InstanceID&) const;
  //!nowrap!+
  ostream& print(ostream&);
  void print2() const;
  int get_conformer() const {return conformer;}
  int get_chain() const {return chain;}
  int get_site() const {return site;}
  int get_instance() const {return instance;}
  string get_sitename() const {return sitename;}
  string get_instname() const {return instname;}
  string get_fn_prefix() const;
  //!nowrap!-
private:
  string sitename;
  string instname;
  int conformer;
  int chain;
  int site;
  int instance;
};

#if SWIGPP_LITERAL_INCLUDE

%addmethods InstanceID {
// Wrap these private members as read-only attributes for Python, but
// this is a bit of a hack (see below).
// It'd probably be better to just wrap the get_member() functions,
// but for Python's use it seems more natural to treat these simple types
// as attributes.
  string sitename;
  string instname;
  int confomer;
  int chain;
  int site;
  int instance;

// swigpp.el strips the default value of the string argument in the
// above constructor. So, wrap this constructor with the default
// argument removed and...
  InstanceID (const string& site_name);
};

%wrapper %{
// ...provide the wrapped constructor with the default argument supplied
InstanceID * new_InstanceID(int residue_number, const string& atom_name){
  return new InstanceID(residue_number, atom_name, "");
}
// Provide the read-only get/set functions for the aboved wrapped attributes
void InstanceID_sitename_set(InstanceID *self, string *sitename){
   _SWIG_exception(SWIG_ValueError, "The sitename attribute is read-only");
}
void InstanceID_instname_set(InstanceID *self, string *instname){
   _SWIG_exception(SWIG_ValueError, "The instname attribute is read-only");
}
int InstanceID_resnum_get(InstanceID *self){return self->get_resnum();}
// These strings are returned by reference, so create global placeholders
// to stash the results to be returned (read-only).
// Note that this works fine from Python since assignment of
// simple objects like numbers and strings effectively creates a copy.
// However, using these global references certainly is not thread safe!
string _InstanceID_sitename_readonly;
string _InstanceID_instname_readonly;
string * InstanceID_atname_get(InstanceID *self){_InstanceID_sitename_readonly = self->get_sitename(); return &_InstanceID_sitename_readonly;}
string * InstanceID_instname_get(InstanceID *self){_InstanceID_instname_readonly = self->get_instname(); return &_InstanceID_sitename_readonly;}
%}

#endif // SWIGPP_LITERAL_INCLUDE

inline InstanceID::InstanceID (const InstanceID& a)
  : sitename(a.sitename), instname(a.instname), conformer(a.conformer), chain(a.chain), site(a.site), instance(a.instance) {}

inline InstanceID& InstanceID::operator= (const InstanceID& a)
{sitename = a.sitename; instname = a.instname; conformer = a.conformer;
 chain = a.chain; site = a.site; instance = a.instance;
 return *this;}

inline bool InstanceID::operator== (const InstanceID& a) const
{return sitename == a.sitename && instname == a.instname && conformer == a.conformer && 
        chain == a.chain && site == a.site && instance == a.instance;}
 
inline bool InstanceID::operator!= (const InstanceID& a) const
{return !(*this == a);}

inline bool InstanceID::operator<(const InstanceID& x) const
{
    if (conformer == x.conformer) {
      if (chain == x.chain) {
        if (site == x.site) {
          if (instance == x.instance) {
            if (sitename == x.sitename) {
              return instname < x.instname;
            } else 
              return sitename < x.sitename;
        } else
          return instance < x.instance;
      } else
	return site<x.site;
    } else
      return chain<x.chain;
  } else
    return conformer<x.conformer;
}

inline bool InstanceID::operator>(const InstanceID& x) const
{
    if (conformer == x.conformer) {
      if (chain == x.chain) {
        if (site == x.site) {
          if (instance == x.instance) {
            if (sitename == x.sitename){
              return instname > x.instname;
            } else 
              return sitename > x.sitename;
        } else
          return instance > x.instance;
      } else
	return site>x.site;
    } else
      return chain>x.chain;
  } else
    return conformer>x.conformer;
}

inline ostream& operator<<(ostream& ost, InstanceID aid)
{return aid.print(ost);}

#endif

// InstanceID.h ends here
