/*
    Read dielectric regions from file
    filename is given with constructor
    file with name filename.diel is read by function read

    file format:
       prefix epsin solvent_probe_radius
          "       "            "

    For each solvent accessible volume it is tried to read a precalculated
    analytical representation from file.
    
    Binary file is read from prefix.dat if the file exists
    otherwise, ASCII file is read from prefix.txt if the file exists
    otherwise, the analytical SAV is recalculated from prefix.pqr
    
    Priority increases from the beginning to the end of the list.
    Volumes overide previous volumes on the list.
    
    e.g.
    mm_region_plus_qm_region     4 1.4
    surface_atoms_of_mm_region  20 1.4
    qm_region                    1 0.5

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/

#include "MEAD/NDielByFile.h"

NDielByFile::NDielByFile (const string& nm) : DielectricEnvironment_lett()
{
  blab2 << "Contructing an NDielByFile from file for " << nm << endl;
  name = nm;
  epsext = PhysCond::get_epsext();
  read();
}

NDielByFile::NDielByFile (const NDielByFile& a) : DielectricEnvironment_lett(), eps(a.eps), solv_acc_vol(a.solv_acc_vol), epsext(a.epsext), name (a.name)
{
}

void NDielByFile::read()
{
  string filename = name + DOT + "diel";
  read(filename);
}

void NDielByFile::read(const string& filename)
{
  // clear previous definitions
  eps.clear();
  solv_acc_vol.clear();

  float tmpeps, radius;
  string dielfile;
  
  // open file
  ifstream surfin (filename.c_str());
  if (!surfin)
   ::error ("NDielByFile:: Cannot open input file, ", filename.c_str(), "\n");
  
  // read lines with definitions
  while (surfin >> dielfile >> tmpeps >> radius) {
   AtomSet ats(dielfile);
   ats.read(); // read pqr
   eps.push_back(tmpeps);
   
   SolvAccVol tmpsav(ats, radius);
   solv_acc_vol.push_back(tmpsav);
   vector<SolvAccVol>::reverse_iterator a = solv_acc_vol.rbegin();

   // try to read precomputed analytical volume, otherwise recalculate
   blab1 << "Try to read in SolvAccVol for " << dielfile << ", if impossible calculate it" << endl;
   if (a->read_top_in_binary(dielfile + DOT + "dat") == 0) {
    blab2 << "Unable to read in SolvAccVol from binary file for " << dielfile << endl;
    if (a->read_top_in_ascii(dielfile + DOT + "txt") == 0) {
     blab2 << "Unable to read in SolvAccVol from ASCII file for " << dielfile << endl;
     blab2 << "Calculating SolvAccVol from AtomSet for " << dielfile << endl;
     a->anal_calc(); // this is a bit costly for big molecules
    }
   }

  }
}


DielCubeRep NDielByFile::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "NDielByFile::get_cuberep entered" << endl;

  int grid_dim = cls.get_grid_dim();
  int ncube = grid_dim*grid_dim*grid_dim;

  DielCubeRep dcr(cls);
  for (int h=0; h<ncube; ++h) {
   dcr[h] = epsext;
  }

  // each SolvAccVol is mapped to the grid, next overrides previous,
  // only exterior is mapped to grid
  vector<float>::const_iterator b = eps.begin();
  for (vector<SolvAccVol>::iterator a = solv_acc_vol.begin(); a!=solv_acc_vol.end(); ++a){

   AccTag* acc_array = new AccTag[ncube];
   a->calc_cuberep(cls, acc_array);

   // acc_array holds the ion accessible volume
   for (int h=0; h<ncube; ++h) {
     if        (acc_array[h] == interior){
       dcr[h] = *b;
     } else if (acc_array[h] != exterior) {
       cerr << "acc_array[" << h << "] has unexpected value, "
	 << acc_array[h] << endl;
       ::error("ERROR in ");
     }
   }
   delete [] acc_array;

   b++;
  }

  dcr.declare_defined();

  blab3 << "exit from NDielByFile::get_cuberep\n" << endl;
  return dcr;
}

// NDielByFile ends here
