/*
    Read electrolyte regions from file
    filename is given with constructor
    file with name filename.ely is read by function read

    file format:
       prefix ionic_strength ion_radius
       "      "              "

    For each ion inaccessible volume (IAV) it is tried to read a precalculated
    analytical representation from file.
    
    Binary file is read from prefix.dat if the file exists
    otherwise, ASCII file is read from prefix.txt if the file exists
    otherwise, the analytical IAV is recalculated from prefix.pqr
    
    Priority increases from the beginning to the end of the list.
    Volumes overide previous volumes on the list.

    A gridpoint is ion accessible if is outside one of gthe IAVs.
    
    e.g.
    membrane_plus_solvent_phase_2   ionic_strength_1   1.4
    membrane_plus_solvent_phase_1   ionic_strength_2   1.4

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/

#ifndef ElectrolyteByAtoms_h
#define ElectrolyteByAtoms_h 1

#include "MEAD/ElectrolyteEnvironment.h"
#include "MEAD/IonAccVol.h"
#include "MEAD/AtomSet.h"
#include "MEAD/CubeLatSpec.h"

//!wrap!
class NElectrolyteByFile : public ElectrolyteEnvironment_lett {
 public:
  //!nowrap!
  NElectrolyteByFile(const NElectrolyteByFile& a);
  NElectrolyteByFile(const string& nm);
  ~NElectrolyteByFile(){}
  void read();
  void read(const string& fn);
  // this is ill-defined here, but needed by analytical
  // initialization of electrostatic potential with Debye
  virtual float ionic_strength() const {return *ionic_str.begin();}
  //!nowrap!
  virtual ElyCubeRep get_cuberep(const CubeLatSpec& cls);
  string get_name(){return name;}
  void set_name(string a){name = a;}
 private:
  string name;
  vector<float> ionic_str;
  vector<IonAccVol> ion_acc_vol;
};

#endif

// NElectrolyteByFile.h ends here
