/*  A form of a QMPB Site 

    label, bound ligands, charge and occupancy of atoms 

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _QmpbFormDef_h
#define _QmpbFormDef_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/globals.h"

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class QmpbFormDef {
public:
  QmpbFormDef();
  QmpbFormDef (const string& name);
  QmpbFormDef (const string& name, const float gmodel);
  QmpbFormDef (const QmpbFormDef& a);
  ~QmpbFormDef () {}
  //!nowrap!+
  QmpbFormDef& operator=(const QmpbFormDef& a);

  //!nowrap!-

  void print_ligandmap();
  void print_atommap();

  string get_name() const {return name;}
  float get_gmodel() const {return gmodel;}
  void set_gmodel(const float);
      
  std::map<string,int> ligandmap;
  std::map<CharmmID,pair<float, int> > atommap;

  // lookup attype by CharmmID or resname/atname pair

  /* check if atom is in list by CharmmID(s) or resname atname pair(s) */
  /* check if ligand is in list by ligand label  */
  bool contains_ligand (const string&) const;
  bool contains_atom (const string&, const string&) const;
  bool contains_atom (const CharmmID&) const;

  /* retrieve parameter by CharmmID(s) or attype(s) or resname atname pair(s) */
  /* retrieve ligand number by ligand label  */
  pair<float,int> get_atom(const string&, const string&) const;
  pair<float,int> get_atom(const CharmmID&) const;  
  int get_ligand (const string&) const;

  /* insert in list, return true if successfull */
  std::pair<std::map<string,int>::iterator,bool> insert_ligand(const string&, const int);
  std::pair<std::map<CharmmID,pair<float,int> >::iterator,bool> insert_atom(const string&, const string&, const float, const int);
  std::pair<std::map<CharmmID,pair<float,int> >::iterator,bool> insert_atom(const CharmmID&, const float, const int);

  //!nowrap!-
private:
  string name;
  float gmodel;
};

inline void QmpbFormDef::set_gmodel(const float f){
 gmodel = f;
}

// atom and ligand map functions

inline pair<float,int> QmpbFormDef::get_atom(const CharmmID& k) const
{
  std::map<CharmmID,pair<float, int> >::const_iterator i=atommap.find(k);
  if (i == atommap.end())
    ::error("QmpbFormDef::get_atom: key not found");
  return i->second;
}

inline pair<float,int> QmpbFormDef::get_atom(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return QmpbFormDef::get_atom(k);
}

inline int QmpbFormDef::get_ligand(const string& k) const
{
  std::map<string,int>::const_iterator i=ligandmap.find(k);
  if (i == ligandmap.end())
    ::error("QmpbFormDef::get_ligand: key not found");
  return i->second;
}

// Existence of element (not STL-like

inline bool QmpbFormDef::contains_atom(const CharmmID& k) const
{
  std::map<CharmmID,std::pair<float,int> >::const_iterator i=atommap.find(k);
  return i != atommap.end();
}

inline bool QmpbFormDef::contains_atom(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return QmpbFormDef::contains_atom(k);
}

inline bool QmpbFormDef::contains_ligand(const string& k) const
{
  std::map<string,int>::const_iterator i=ligandmap.find(k);
  return i != ligandmap.end();
}

// easier to use insert (not quite STL-like)
inline pair<std::map<CharmmID,pair<float,int> >::iterator,bool> QmpbFormDef::insert_atom (const string& rn, const string& an, const float q, const int o) {
  CharmmID k(rn, an);
  if (contains_atom(k)) {
    cerr << k << ":" << name << endl;
    ::error("QmpbFormDef::insert_atom: attempt to insert a duplicate atom for form\n");
  }
  return atommap.insert(std::pair<CharmmID,pair<float, int> >(k, std::pair<float,int>(q,o)));
}

inline pair<std::map<CharmmID,pair<float,int> >::iterator,bool> QmpbFormDef::insert_atom (const CharmmID& k, const float q, const int o){
  if (contains_atom(k)) {
    cerr << k << ":" << name << endl;
    ::error("QmpbFormDef::insert_atom: attempt to insert a duplicate attype for atom\n");
  }
  return atommap.insert(std::pair<CharmmID,pair<float,int> >(k, std::pair<float,int>(q,o)));
}

inline pair<std::map<string,int>::iterator,bool> QmpbFormDef::insert_ligand (const string& k, const int n){
  if (contains_ligand(k)) {
    cerr << k << ":" << name << endl;
    ::error("QmpbFormDef::insert_ligand: attempt to insert a duplicate ligand\n");
  }
  return ligandmap.insert(std::pair<string,int>(k, n));
}

#endif

// QmpbFormDef.h ends here
