/*  A QMPB Site 

    label, forms
    
    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de
    
    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/QmpbInstance.h"
#include "MEAD_e/globals.h"
#include <stdio.h>
#include <string>

using namespace std;

QmpbInstance::QmpbInstance()
: name(""), instid(InstanceID()), ligandmap(),
  model_struct(AtomSet()), site_struct(AtomSet()),
  gmodel(0), gmmint(0), propensity(0),
  has_propensity(false), confnum(0)
{
}

QmpbInstance::QmpbInstance(const QmpbInstance& a)
: name(a.name), instid(a.instid), ligandmap(a.ligandmap),
  model_struct(a.model_struct), site_struct(a.site_struct),
  gmodel(a.gmodel), gmmint(a.gmmint), gmmback(a.gmmback), 
  gmmmodelback(a.gmmmodelback), propensity(a.propensity),
  has_propensity(a.has_propensity), confnum(a.confnum)
{
}

QmpbInstance::QmpbInstance (const InstanceID& id, const map<string,int>& lm, const AtomSet& a1, const AtomSet& a2, const int cn)
: name(id.get_instname()), sitename(id.get_sitename()), instid(id), ligandmap(lm), confnum(cn), 
  gmodel(0), propensity(0), has_propensity(false), gmmint(0), gmmback(0), gmmmodelback(0)
{
  // make sure that all atoms have IDs fitting to the provided InstanceID
  for (AtomSet::const_iterator i = a1.begin(); i != a1.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = id.get_instance();
   site_struct.insert(at);  
   // make sure that model has all atoms
   AtomID aid(id.get_site(), at.atname, id.get_conformer(),id.get_chain(),id.get_site(),id.get_instance());
   if (!model_struct.contains(aid))
    model_struct.insert(at);  
  }
  for (AtomSet::const_iterator i = a2.begin(); i != a2.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = int(0);
   model_struct.insert(at);  
  }
}

QmpbInstance::QmpbInstance (const InstanceID& id, const map<string,int>& lm, const AtomSet& a1, const AtomSet& a2, const int cn, const float gm)
: name(id.get_instname()), sitename(id.get_sitename()), instid(id), ligandmap(lm), confnum(cn), 
  gmodel(gm), propensity(0), has_propensity(false), gmmint(0), gmmback(0), gmmmodelback(0)
{
  // make sure that all atoms have IDs fitting to the provided InstanceID
  for (AtomSet::const_iterator i = a1.begin(); i != a1.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = id.get_instance();
   site_struct.insert(at);  
   // make sure that model has all atoms
   AtomID aid(id.get_site(), at.atname, id.get_conformer(),id.get_chain(),id.get_site(),id.get_instance());
   if (!model_struct.contains(aid))
    model_struct.insert(at);  
  }
  for (AtomSet::const_iterator i = a2.begin(); i != a2.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = int(0);
   model_struct.insert(at);  
  }
}

QmpbInstance::QmpbInstance (const InstanceID& id, const map<string,int>& lm, const AtomSet& a1, const AtomSet& a2, const int cn, const float gm, const float prop)
: name(id.get_instname()), sitename(id.get_sitename()), instid(id), ligandmap(lm), confnum(cn), gmodel(gm),
  propensity(prop), has_propensity(true), gmmint(0), gmmback(0), gmmmodelback(0)
{
  // make sure that all atoms have IDs fitting to the provided InstanceID
  for (AtomSet::const_iterator i = a1.begin(); i != a1.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = id.get_instance();
   site_struct.insert(at);  
   // make sure that model has all atoms
   AtomID aid(id.get_site(), at.atname, id.get_conformer(),id.get_chain(),id.get_site(),id.get_instance());
   if (!model_struct.contains(aid))
    model_struct.insert(at);  
  }
  for (AtomSet::const_iterator i = a2.begin(); i != a2.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = int(0);
   model_struct.insert(at);  
  }
}

QmpbInstance::QmpbInstance (const InstanceID& id, const map<string,int>& lm, const AtomSet& a1, const AtomSet& a2, const int cn, const float gm, const float prop, const float gmmi)
: name(id.get_instname()), sitename(id.get_sitename()), instid(id), ligandmap(lm), confnum(cn),
  gmodel(gm), gmmint(gmmi), propensity(prop), has_propensity(true), gmmback(0), gmmmodelback(0)
{
  // make sure that all atoms have IDs fitting to the provided InstanceID
  for (AtomSet::const_iterator i = a1.begin(); i != a1.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = id.get_instance();
   site_struct.insert(at);  
   // make sure that model has all atoms
   AtomID aid(id.get_site(), at.atname, id.get_conformer(),id.get_chain(),id.get_site(),id.get_instance());
   if (!model_struct.contains(aid))
    model_struct.insert(at);  
  }
  for (AtomSet::const_iterator i = a2.begin(); i != a2.end(); ++i){
   Atom at(i->second);
   at.conformer = id.get_conformer();
   at.chain = id.get_chain();
   at.site = id.get_site();
   at.instance = int(0);
   model_struct.insert(at);  
  }
}

void QmpbInstance::print() const
{
 cout << "-----------------------------------" << endl;
 cout << instid << " " << name << ":" << endl; 
 cout << "\nname and number of bound ligands:" << endl;
 for (std::map<string,int>::const_iterator j = ligandmap.begin(); j != ligandmap.end(); ++j){
   cout << j->first << "   " << j->second << endl;
 } cout << endl;
 cout << "\nsite structure:" << endl;
 site_struct.print();
 cout << "\nmodel compound structure:" << endl;
 model_struct.print();          
}

void QmpbInstance::write_struct()
{
 string fn1 = instid.get_fn_prefix() + ".pqr";
 string fn2 = instid.get_fn_prefix() + "_model.pqr";
 write_site_struct(fn1);          
 write_model_struct(fn2);
}

void QmpbInstance::write_struct(const string& fn1, const string& fn2)
{
 write_site_struct(fn1);          
 write_model_struct(fn2);
}

void QmpbInstance::write_site_struct()
{
 string fn = instid.get_fn_prefix() + ".pqr";
 write_site_struct(fn);          
}

void QmpbInstance::write_site_struct(const string& fn)
{
 site_struct.write_pqr(fn);
}

void QmpbInstance::write_model_struct()
{
 string fn = instid.get_fn_prefix() + "_model.pqr";
 write_model_struct(fn);          
}

void QmpbInstance::write_model_struct(const string& fn)
{
 model_struct.write_pqr(fn);
}


// QmpbInstance.cc ends here
