/*  A QMPB Instance

    label, forms
    
    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _QmpbInstance_h
#define _QmpbInstance_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/globals.h"

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class QmpbInstance {
public:
  QmpbInstance ();
  QmpbInstance (const QmpbInstance& a);
  QmpbInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&);
  QmpbInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int);
  QmpbInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float);
  QmpbInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float, const float);
  QmpbInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float, const float, const float);
  ~QmpbInstance () {}
  //!nowrap!+
  QmpbInstance& operator=(const QmpbInstance& a);

  //!nowrap!-
  void print() const;
  void write_struct();
  void write_struct(const string&, const string&);
  void write_model_struct();
  void write_model_struct(const string&);
  void write_site_struct();
  void write_site_struct(const string&);
  
  string get_name() const {return name;}
  InstanceID get_instid() const {return instid;}
  void set_instid(const InstanceID& id){instid = id;}
  float get_gmodel() const {return gmodel;}   
  void set_gmodel(const float f){gmodel = f;}
  float get_gmmint() const {return gmmint;}   
  float get_gmmmodelback() const {return gmmmodelback;}   
  void set_gmmmodelback(const float f){gmmmodelback = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_gmmsite() const {return gmmsite;}   
  void set_gmmsite(const float f){gmmsite = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_gmmback() const {return gmmback;}   
  void set_gmmback(const float f){gmmback = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_propensity() const {return propensity;}   
  void set_propensity(const float f){propensity = f;}
  bool get_has_propensity() const {return has_propensity;}
  void set_has_propensity(const bool f){has_propensity = f;}
  int get_confnum() const {return confnum;}
  void set_confnum(const int i){confnum = i;}
         
  AtomSet model_struct;
  AtomSet site_struct;
  map<string,int> ligandmap;

  // lookup attype by CharmmID or resname/atname pair

  /* check by label if form or ligand are already listed */
  bool contains_ligand (const string&) const;
  // return number of bound ligands of a certain type
  int ligands_bound (const string&) const;

  bool operator== (const QmpbInstance& a) const {return instid == a.get_instid();}
  bool operator!= (const QmpbInstance& a) const {return instid != a.get_instid();}
  bool operator< (const QmpbInstance& a) const {return instid < a.get_instid();}
  bool operator> (const QmpbInstance& a) const {return instid > a.get_instid();}

  //!nowrap!-
private:
  string sitename;
  string name;
  InstanceID instid;
  // total intrinsic MM-energy
  float gmmint;
  // interactions (site+ligand)-(background-model)
  float gmmback;
  // interactions (site+ligand)-(site+ligand)
  float gmmsite;
  float gmmmodelback;
  float gmodel;
  float propensity;
  bool has_propensity;
  int confnum;
};

// form list functions

inline bool QmpbInstance::contains_ligand(const string& k) const
{
  std::map<string,int>::const_iterator j(ligandmap.find(k));
  return j!=ligandmap.end();
}

inline int QmpbInstance::ligands_bound(const string& k) const
{
  int result;
  std::map<string,int>::const_iterator j(ligandmap.find(k));
  if (j!=ligandmap.end())
   result = j->second;
  else
   result = 0;
  return result;
}

#endif

// QmpbInstance.h ends here
