/*   A QMPB molecule containing titratable and/or flexible sites

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _QmpbMolecule_h
#define _QmpbMolecule_h 1

#include "MEAD_e/CharmmID.h"
#include "MEAD_e/Coord.h"
#include "MEAD_e/SiteID.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/CubeLatSpec.h"
#include "MEAD_e/FinDiffMethod.h"
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/QmpbSiteDef.h"
#include "MEAD_e/QmpbSiteDef.h"
#include "MEAD_e/QmpbSite.h"
#include "MEAD_e/CharmmPara.h"
#include "MEAD_e/Rotlib.h"
#include "MEAD_e/globals.h"
#include "MEAD_e/PhysCond.h"
#include <vector>
#include <map>
#include <list>
#include <string>
#include <cmath>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

using namespace std;

//!wrap!
class QmpbMolecule {
public:
  QmpbMolecule();
  QmpbMolecule (const string&);
  QmpbMolecule (const QmpbMolecule&);
  QmpbMolecule (const string&, const string&, const string&, const string&);
  QmpbMolecule (const string&, const string&, const string&, const string&, const bool);
  QmpbMolecule (const string&, const string&, const string&, const string&, const float);
  QmpbMolecule (const string&, const string&, const string&, const string&, const float, const bool);
  ~QmpbMolecule () {}
  //!nowrap!+
  QmpbMolecule& operator=(const QmpbMolecule& a);
 
  // set default behaviour and variables
  void set_default();

  // these subroutines control the work flow
  // read data
  // create structures of background and instances of sites,
  // compute MM background and interaction energies
  // write structure files and QMPB input file
  // the master subroutine
  void prepare_qmpb();
  void prepare_qmpb(const string&, const string&, const string&, const string&);
  // read site definitions for titratable sites from .est files
  // create instances from rotamers and or charge forms of sites
  void create_sites();
  // calculate MM part of background and interaction energies
  // use knowledge based potential for intrinsic energies or use pure CHARMM 
  // call writing of matrices to file
  void calc_mmint();
  void calc_mmint_propensities();
  void calc_mmint_charmm();
  void calc_mmg();
  void filter_instances();
  // write energy matrices to files .mmint and .mmg
  void write_mmint(const string&);
  void write_mmg(const string&);
  void print_mmint() const;
  void print_mmg() const;
  // write structure files and input script for QMPB
  void write_qmpb_input(const string&, const string&);
  
  // helper functions
  std::list<std::string> get_ligands();
  // read .ogm and .mgm if existent, otherwise create grids
  std::pair<FinDiffMethod,FinDiffMethod> get_fdm(const string&);
  // read .rot file to find out which sites are and how to treat --> rotmap
  void read_rot();
  void read_rot(const string&);  
  // read .est files
  void read_site_defs();
  void create_sitemap();
  // book keeping --> sitemap, write .sites file 
  // (actually not necessary but good check)
  void write_sites_file() const;
  void write_sites_file(const string&) const;
  
  // list functions
  bool in_list(const string&, const std::list<std::string>&) const;
  
  // geometry functions  
  // i,j,k,l, target torsion angle
  //!nowrap!-
  Coord set_tor(const Coord&, const Coord&, const Coord&, const Coord&, float) const;
  // i,j,k,l, change in torsion angle
  Coord change_tor(const Coord&, const Coord&, const Coord&, const Coord&, float) const;
  // i,j,k,l, 
  float get_tor(const Coord&, const Coord&, const Coord&, const Coord&) const;
  // i,j,k 
  float get_angle(const Coord&, const Coord&, const Coord&) const;
  // point, point
  float get_dist(const Coord&, const Coord&) const;
  float get_sqdist(const Coord&, const Coord&) const;
  // vector
  float abs_value(const Coord&) const;
  // vector, vector
  float dot_p(const Coord&, const Coord&) const;  
  // point, point point
  Coord get_normal(const Coord&, const Coord&, const Coord&) const;
  // vector, vector  
  Coord cross_p(const Coord&, const Coord&) const;
  // unit normal_vector of plane, point to project, point in plane
  Coord project_on_plane (const Coord&, const Coord&, const Coord&) const;
  // point, point in plane, normal vector
  float get_dist_p_plane (const Coord&, const Coord&, const Coord&) const; 
  // point, point on line, point on line
  float get_dist_p_line (const Coord&, const Coord&, const Coord&) const; 
  // get integer torsion bin (rotamer library) from torsion angle
  int get_tor_bin(const float) const; 
  //!nowrap!-
 
  //!nowrap!-

  // get/set variables determining properties or behaviour of the object
  
  string get_meadpath() const {return meadpath;}
  void set_meadpath(const string& n) {meadpath = n;}
  string get_eps1set() const {return eps1set;}
  void set_eps1set(const string& n) {eps1set = n;}
  string get_eps2set() const {return eps2set;}
  void set_eps2set(const string& n) {eps2set = n;}
  string get_eps3set() const {return eps3set;}
  void set_eps3set(const string& n) {eps3set = n;}
  string get_channelset() const {return channelset;}
  void set_channelset(const string& n) {channelset = n;}
  string get_elycavset() const {return elycavset;}
  void set_elycavset(const string& n) {elycavset = n;}
  string get_name() const {return name;}
  void set_name(const string& n) {name = n;}
  string get_qmpbdir() const {return qmpbdir;}
  void set_qmpbdir(const string& n) {qmpbdir = n;}
  // query/set usage of knowledge based potential
  bool get_has_memb() const {return has_memb;}
  void set_has_memb(const bool n) {has_memb = n;}
  bool get_use_prop() const {return use_prop;}
  void set_use_prop(const bool n) {use_prop = n;}
  // query/set usage of intrinsic energy filter
  bool get_use_filter() const {return use_filter;}
  void set_use_filter(const bool n) {use_filter = n;}
  // print separate energy terms
  bool get_print_chm() const {return print_chm;}
  void set_print_chm(const bool n) {print_chm = n;}
  bool get_print_debug() const {return print_debug;}
  void set_print_debug(const bool n) {print_debug = n; if (n) print_chm = n;}
  // get/set energy cutoff for intrinsic energies
  float get_filter_cutoff() const {return filter_cutoff;}
  void set_filter_cutoff (const float f) {filter_cutoff = f;}
  float get_zmaxh() const {return zmaxh;}
  void set_zmaxh (const float f) {zmaxh = f; has_memb = true;}
  float get_zmaxc() const {return zmaxc;}
  void set_zmaxc (const float f) {zmaxc = f; has_memb = true;}
  float get_zminc() const {return zminc;}
  void set_zminc (const float f) {zminc = f; has_memb = true;}
  float get_zminh() const {return zminh;}
  void set_zminh (const float f) {zminh = f; has_memb = true;}
  float get_membholer() const {return membholer;}
  void set_membholer (const float f) {membholer = f; has_memb = true;}
  float get_membholex() const {return membholex;}
  void set_membholex (const float f) {membholex = f; has_memb = true;}
  float get_membholey() const {return membholey;}
  void set_membholey (const float f) {membholey = f; has_memb = true;}
  float get_ionicstr() const {return ionicstr;}
  void set_ionicstr (const float f) {ionicstr = f;}
  float get_ionicstr1() const {return ionicstr1;}
  void set_ionicstr1 (const float f) {ionicstr1 = f; has_memb = true;}
  float get_ionicstr2() const {return ionicstr2;}
  void set_ionicstr2 (const float f) {ionicstr2 = f; has_memb = true;}
  float get_eps_ff() const {return eps_ff;}
  void set_eps_ff (const float f) {eps_ff = f;}
  float get_epsin1() const {return epsin1;}
  void set_epsin1 (const float f) {epsin1 = f;}
  float get_epsin2() const {return epsin2;}
  void set_epsin2 (const float f) {epsin2 = f;}
  float get_epsin3() const {return epsin3;}
  void set_epsin3 (const float f) {epsin3 = f;}
  float get_epsext() const {return epsext;}
  void set_epsext (const float f) {epsext = f;}
  float get_epscore() const {return epscore;}
  void set_epscore (const float f) {epscore = f; has_memb = true;}
  float get_epshead() const {return epshead;}
  void set_epshead (const float f) {epshead = f; has_memb = true;}
  // get/set which energy terms to use
  bool get_use_term_l_e() const {return use_term_l_e;}
  void set_use_term_l_e(const bool b) {use_term_l_e = b;}
  bool get_use_term_l_lj() const {return use_term_l_lj;}
  void set_use_term_l_lj(const bool b) {use_term_l_lj = b;}
  bool get_use_term_l_bond() const {return use_term_l_bond;}
  void set_use_term_l_bond(const bool b) {use_term_l_bond = b;}
  bool get_use_term_l_angle() const {return use_term_l_angle;}
  void set_use_term_l_angle(const bool b) {use_term_l_angle = b;}
  bool get_use_term_l_dihedral() const {return use_term_l_dihedral;}
  void set_use_term_l_dihedral(const bool b) {use_term_l_dihedral = b;}
  bool get_use_term_l_improper() const {return use_term_l_improper;}
  void set_use_term_l_improper(const bool b) {use_term_l_improper = b;}
  bool get_use_term_ml_e() const {return use_term_ml_e;}
  void set_use_term_ml_e(const bool b) {use_term_ml_e = b;}
  bool get_use_term_ml_lj() const {return use_term_ml_lj;}
  void set_use_term_ml_lj(const bool b) {use_term_ml_lj = b;}
  bool get_use_term_ml_bond() const {return use_term_ml_bond;}
  void set_use_term_ml_bond(const bool b) {use_term_ml_bond = b;}
  bool get_use_term_ml_angle() const {return use_term_ml_angle;}
  void set_use_term_ml_angle(const bool b) {use_term_ml_angle = b;}
  bool get_use_term_ml_dihedral() const {return use_term_ml_dihedral;}
  void set_use_term_ml_dihedral(const bool b) {use_term_ml_dihedral = b;}
  bool get_use_term_ml_improper() const {return use_term_ml_improper;}
  void set_use_term_ml_improper(const bool b) {use_term_ml_improper = b;}
  bool get_use_term_s_e() const {return use_term_s_e;}
  void set_use_term_s_e(const bool b) {use_term_s_e = b;}
  bool get_use_term_s_lj() const {return use_term_s_lj;}
  void set_use_term_s_lj(const bool b) {use_term_s_lj = b;}
  bool get_use_term_s_bond() const {return use_term_s_bond;}
  void set_use_term_s_bond(const bool b) {use_term_s_bond = b;}
  bool get_use_term_s_angle() const {return use_term_s_angle;}
  void set_use_term_s_angle(const bool b) {use_term_s_angle = b;}
  bool get_use_term_s_dihedral() const {return use_term_s_dihedral;}
  void set_use_term_s_dihedral(const bool b) {use_term_s_dihedral = b;}
  bool get_use_term_s_improper() const {return use_term_s_improper;}
  void set_use_term_s_improper(const bool b) {use_term_s_improper = b;}
  bool get_use_term_ms_e() const {return use_term_ms_e;}
  void set_use_term_ms_e(const bool b) {use_term_ms_e = b;}
  bool get_use_term_ms_lj() const {return use_term_ms_lj;}
  void set_use_term_ms_lj(const bool b) {use_term_ms_lj = b;}
  bool get_use_term_ms_bond() const {return use_term_ms_bond;}
  void set_use_term_ms_bond(const bool b) {use_term_ms_bond = b;}
  bool get_use_term_ms_angle() const {return use_term_ms_angle;}
  void set_use_term_ms_angle(const bool b) {use_term_ms_angle = b;}
  bool get_use_term_ms_dihedral() const {return use_term_ms_dihedral;}
  void set_use_term_ms_dihedral(const bool b) {use_term_ms_dihedral = b;}
  bool get_use_term_ms_improper() const {return use_term_ms_improper;}
  void set_use_term_ms_improper(const bool b) {use_term_ms_improper = b;}
      
  // holds contents of .rot files, defines sites and their type
  map<SiteID,int> rotmap;
  map<SiteID,pair<int,int> > sitemap;
  
  // this is parameter data read from file
  list<QmpbSiteDef> sitedef;
  CharmmPara chmpara;
  Rotlib rlib;

  // this is the derived data the QMPB representation of our molecule
  AtomSet input_pqr;
  AtomSet background;
  std::list<QmpbSite> sites;
  std::map<InstanceID,std::map<InstanceID, float> > mmg;
  
  // does only apply if the map does really contain an entry
  bool contains_mmg (const InstanceID&, const InstanceID&) const;
  // functions for mmg supporting primitive sparseness 
  // by simply returning zero if an element does not exist
  float get_mmg (const InstanceID&, const InstanceID&) const;
  // this one is complaining if no real entry is found
  float get_mmg_strict (const InstanceID&, const InstanceID&) const;
  std::pair<std::map<InstanceID, float>::iterator,bool> insert_mmg (const InstanceID&, const InstanceID&, const float);
  
  // verify uniqueness and insert in rotmap/sitemap, retrieve by SiteID
  std::pair<std::map<SiteID,int>::iterator,bool> insert_rotmap_entry(const SiteID&, const int);
  std::pair<std::map<SiteID,pair<int,int> >::iterator,bool> insert_sitemap_entry(const SiteID&, const int, const int);
  bool contains_rotmap_entry(const SiteID&) const;
  bool contains_sitemap_entry(const SiteID&) const;
  int get_rotmap_value(const SiteID&) const;
  pair<int,int> get_sitemap_value(const SiteID&) const;
  
  // sitedef functions
  bool contains_sitedef(const string&) const;
  QmpbSiteDef get_sitedef(const string&) const;
  list<QmpbSiteDef>::iterator insert_sitedef(const QmpbSiteDef&);

  // sites functions
  bool contains_site(const SiteID&) const;
  QmpbSite get_site(const SiteID&) const;
  list<QmpbSite>::iterator insert_site(const QmpbSite&);  
  pair<map<CharmmID,string>::iterator,bool> insert_site(const SiteID&, const string&);

  //!nowrap!-
private:
  std::string name;
  std::string qmpbdir;
  // terms involving ligand atoms and no model atoms
  bool use_term_l_e;
  bool use_term_l_lj;
  bool use_term_l_bond;
  bool use_term_l_angle;
  bool use_term_l_dihedral;
  bool use_term_l_improper;
  // terms involving ligand atoms and model atoms (ligand model interactions belongs to mmback)
  bool use_term_ml_e;
  bool use_term_ml_lj;
  bool use_term_ml_bond;
  bool use_term_ml_angle;
  bool use_term_ml_dihedral;
  bool use_term_ml_improper;
  // terms involving no ligand and no model atoms (internal site)
  bool use_term_s_e;
  bool use_term_s_lj;
  bool use_term_s_bond;
  bool use_term_s_angle;
  bool use_term_s_dihedral;
  bool use_term_s_improper;
  // terms involving no ligand but model atoms (site/model interaction mmback)
  bool use_term_ms_e;
  bool use_term_ms_lj;
  bool use_term_ms_bond;
  bool use_term_ms_angle;
  bool use_term_ms_dihedral;
  bool use_term_ms_improper;
  bool use_prop;
  bool use_filter;
  bool print_chm;
  bool print_debug;
  // energy cutoff for exclusion of instances with high intrinsic energy
  float filter_cutoff;
  // relative dielectric permittivity for which the forcefield was parametrized
  float eps_ff;
  float epsin1,epsin2,epsin3,epscore,epshead,epsext,zminh,zminc,zmaxc,zmaxh,
        membholer,membholex,membholey,ionicstr,ionicstr1,ionicstr2;
  bool has_memb;
  std::string meadpath,eps1set,eps2set,eps3set,channelset,elycavset;
};

// mmg functions

inline bool QmpbMolecule::contains_mmg(const InstanceID& k, const InstanceID& l) const
{
  bool found = false;
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator i=mmg.find(k);
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator ii=mmg.find(l);
  if (i  != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
   }
  }
  if (!found && ii != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=i->second.find(k);
   if (j != ii->second.end()){
    found = true;
   }
  }
  return found;
}

// BEWARE: this function does not complain about non-existing elements but
// simply returns zero to get the mmg map sparse
// calculation of all terms must be ensured by calc mmg!!!
inline float QmpbMolecule::get_mmg(const InstanceID& k, const InstanceID& l) const
{
  float result = 0;
  bool found = false;
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator i=mmg.find(k);
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator ii=mmg.find(l);
  if (i  != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  return result;
}

// this get function complains about a non-existing entry
inline float QmpbMolecule::get_mmg_strict(const InstanceID& k, const InstanceID& l) const
{
  float result;
  bool found = false;
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator i=mmg.find(k);
  std::map<InstanceID,std::map<InstanceID,float> >::const_iterator ii=mmg.find(l);
  if (i  != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != mmg.end()){
   std::map<InstanceID,float>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found)
    ::error ("QmpbMolecule::get_mmg_strict: could not find entry.\n");
  return result;
}

inline std::pair<std::map<InstanceID,float>::iterator,bool> QmpbMolecule::insert_mmg(const InstanceID& k, const InstanceID& l, const float f)
{
  pair<std::map<InstanceID,float>::iterator,bool> result;
  bool found   = false;
  bool foundi  = false;
  bool foundii = false;
  std::map<InstanceID,std::map<InstanceID,float> >::iterator i=mmg.find(k);
  std::map<InstanceID,std::map<InstanceID,float> >::iterator ii=mmg.find(l);
  std::map<InstanceID,float>::iterator j;
  if (i != mmg.end()){
   foundi = true;
   j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
   }
  }
  if (!found && ii != mmg.end()){
   foundii = true;
   j=i->second.find(k);
   if (j != ii->second.end()){
    found = true;
   }
  }
  if (found){
   j->second = f;
  } else if (foundi){
   result = i->second.insert(std::pair<InstanceID,float>(l,f));
  } else if (foundii){
   result = ii->second.insert(std::pair<InstanceID,float>(k,f));
  } else {
   std::map<InstanceID,float> tmpmap;
   tmpmap.insert(std::pair<InstanceID,float>(l,f));
   std::pair<std::map<InstanceID,std::map<InstanceID,float> >::iterator, bool> tmppair2 = 
        mmg.insert(std::pair<InstanceID,std::map<InstanceID,float> >(k,tmpmap));
   if(tmppair2.second){
    std::map<InstanceID,float>::iterator titr = tmppair2.first->second.find(k);
    bool tbool = titr != tmppair2.first->second.end();
    result = pair<std::map<InstanceID,float>::iterator,bool>(titr,tbool);
   }
  }
  return result;
}

// sitemap and rotmap functions

inline bool QmpbMolecule::contains_rotmap_entry(const SiteID& k) const
{
  std::map<SiteID,int>::const_iterator i=rotmap.find(k);
  return i != rotmap.end();
}

inline int QmpbMolecule::get_rotmap_value(const SiteID& k) const
{
  std::map<SiteID,int>::const_iterator i=rotmap.find(k);
  if (i != rotmap.end())
    return i->second;
  else
    ::error ("QmpbMolecule::get_rotmap_entry: could not find entry.\n");    
}

inline pair<map<SiteID,int>::iterator,bool> QmpbMolecule::insert_rotmap_entry(const SiteID& sid, const int i1)
{
  pair<map<SiteID,int>::iterator,bool> tp2;
  if (!contains_rotmap_entry(sid)){
    pair<SiteID,int> tp(sid, i1);
    tp2 = rotmap.insert(tp);
  }
  else {
    ::error ("QmpbMolecule::insert_rotmap_entry: attempt to insert duplicate key.\n");      
  }
  return tp2;
}


inline bool QmpbMolecule::contains_sitemap_entry(const SiteID& k) const
{
  std::map<SiteID,pair<int,int> >::const_iterator i=sitemap.find(k);
  return i != sitemap.end();
}

inline pair<int,int> QmpbMolecule::get_sitemap_value(const SiteID& k) const
{
  std::map<SiteID,pair<int,int> >::const_iterator i=sitemap.find(k);
  if (i != sitemap.end())
    return pair<int,int>(i->second.first, i->second.second);
  else
    ::error ("QmpbMolecule::get_sitemap_entry: could not find entry.\n");    
}

inline pair<map<SiteID,pair<int,int> >::iterator,bool> QmpbMolecule::insert_sitemap_entry(const SiteID& sid, const int i1, const int i2)
{

  if (!contains_sitemap_entry(sid)){
    pair<int,int> tp1(i1,i2);
    pair<SiteID,pair<int,int> > tp2(sid, tp1);
    return sitemap.insert(tp2);
  }
  else {
    ::error ("QmpbMolecule::insert_sitemap_entry: attempt to insert duplicate key.\n");      
  }

}

// sitedef functions

inline bool QmpbMolecule::contains_sitedef(const string& k) const
{
  bool found = false;
  for (list<QmpbSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if ((*i).get_name() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline QmpbSiteDef QmpbMolecule::get_sitedef(const string& k) const
{
  bool found = false;
  list<QmpbSiteDef>::const_iterator j;
  for (list<QmpbSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if ((*i).get_name() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
    ::error ("QmpbMolecule::get_sitedef: could not find key.\n");      
  else
    return *j;
}

inline list<QmpbSiteDef>::iterator QmpbMolecule::insert_sitedef(const QmpbSiteDef& a)
{
  if (contains_sitedef(a.get_name()))
    ::error ("QmpbMolecule::insert_sitedef: attempt to insert duplicate entry.\n");
  sitedef.push_back(a);
  return sitedef.end();
}

// site functions

inline bool QmpbMolecule::contains_site(const SiteID& k) const
{
  bool found = false;
  for (list<QmpbSite>::const_iterator i = sites.begin(); i != sites.end(); ++i){
   if ((*i).get_sid() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline QmpbSite QmpbMolecule::get_site(const SiteID& k) const
{
  bool found = false;
  list<QmpbSite>::const_iterator j;
  for (list<QmpbSite>::const_iterator i = sites.begin(); i != sites.end(); ++i){
   if ((*i).get_sid() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
    ::error ("QmpbMolecule::get_site: could not find key.\n");      
  else
    return *j;
}

inline list<QmpbSite>::iterator QmpbMolecule::insert_site(const QmpbSite& a)
{
  if (contains_site(a.get_sid()))
    ::error ("QmpbMolecule::insert_site: attempt to insert duplicate entry.\n");
  sites.push_back(a);
  return sites.end();
}

//////////////////////// list helper functions /////////////////////////

inline bool QmpbMolecule::in_list(const std::string& s, const std::list<std::string>& l) const
{
  bool found = false;
  for (std::list<std::string>::const_iterator i = l.begin(); i != l.end(); ++i){
   if (*i == s){
    found = true;
    break;
   }
  }
  return found;
}

////////////////////////// geometry functions ////////////////////////// 

// i,j,k,l, target torsion angle
inline Coord QmpbMolecule::set_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const float t) const
{  
  Coord a(k-j);
  Coord b(l-k);
  float d(abs_value(a));
  a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  float tmptor = get_tor(i,j,k,l) - t;
  float cos_t(cos(tmptor));
  float sin_t(sin(tmptor));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l, change in torsion angle
inline Coord QmpbMolecule::change_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const float t) const
{
  Coord a(k-j);
  Coord b(l-k);
  float d(abs_value(a));
  a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  float cos_t(cos(-t));
  float sin_t(sin(-t));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l 
inline float QmpbMolecule::get_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l) const
{
 Coord a(get_normal(i,j,k));
 Coord b(get_normal(j,k,l));
 float t(acos(dot_p(a,b)));
 float v(dot_p(k-j,cross_p(a,b)));
 if (v < float(0)) t *= -1;
 return t;
}

// i,j,k 
inline float QmpbMolecule::get_angle(const Coord& i, const Coord& j, const Coord& k) const
{
  Coord a(i-j); a /= abs_value(a);
  Coord b(k-j); b /= abs_value(b);
  return acos(dot_p(a,b));
}

// point, point
inline float QmpbMolecule::get_dist(const Coord& a, const Coord& b) const
{return sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// point, point
inline float QmpbMolecule::get_sqdist(const Coord& a, const Coord& b) const
{return ((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// vector
inline float QmpbMolecule::abs_value(const Coord& a) const
{return sqrt(a.x*a.x + a.y*a.y + a.z*a.z);}

// vector, vector
inline float QmpbMolecule::dot_p(const Coord& a, const Coord& b) const
{return a.x*b.x + a.y*b.y + a.z*b.z;}
  
// vector, vector
inline Coord QmpbMolecule::cross_p(const Coord& a, const Coord& b) const
{Coord prod(a.y*b.z - a.z*b.y, b.x*a.z - a.x*b.z, a.x*b.y - b.x*a.y); return prod;}

// point in plane, point in plane, point in plane
inline Coord QmpbMolecule::get_normal (const Coord& i, const Coord& j, const Coord& k) const
{Coord n(cross_p(j-i,k-i)); n /= abs_value(n); return n;}

// unit normal vector of plane, point to project, point in plane
inline Coord QmpbMolecule::project_on_plane (const Coord& n, const Coord& a, const Coord& b) const
{return Coord(a - (n * get_dist_p_plane(a, b, n)) - b);}

// point, point in plane, normal vector
inline float QmpbMolecule::get_dist_p_plane (const Coord& a, const Coord& b, const Coord& n) const
{return dot_p(a-b,n);}
 
// point, point on line, point on line
inline float QmpbMolecule::get_dist_p_line (const Coord& a, const Coord& b, const Coord& c) const
{return abs_value(cross_p(c-b,b-a)) / abs_value(c-b);}

inline int QmpbMolecule::get_tor_bin(const float z) const
{return static_cast<int>((floor(((z*(180/3.141593)) + 5) / 10) * 10));} 

#endif

// QmpbMolecule.h ends here
