/*  A QMPB Site 

    site label, instances (defined by charge form and conformer)
    
    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _QmpbSite_h
#define _QmpbSite_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/SiteID.h"
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/QmpbInstance.h"
#include "MEAD_e/globals.h"

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class QmpbSite {
public:
  QmpbSite();
  QmpbSite (const SiteID& id);
  QmpbSite (const QmpbSite& a);
  ~QmpbSite () {}
  //!nowrap!+
  QmpbSite& operator=(const QmpbSite& a);

  //!nowrap!-
  void print() const;
  void print_instances() const;
  void renumber_instances();
  
  list<QmpbInstance> instances;

  string get_name() const {return name;}
  SiteID get_sid() const {return sid;}
  int get_max_confnum() const;
  int get_nconf() const;
  int get_ninst() const {return instances.size();}

  /* check by label if form or ligand are already listed */
  bool contains_form (const string&) const;
  bool contains_instance (const InstanceID&) const;
  bool contains_ligand (const string&) const;
 
  /* retrieve list of instances by form label */
  list<QmpbInstance> get_form(const string&) const;
  /* retrieve specific instance by InstanceID */
  QmpbInstance get_instance(const InstanceID&) const;

  /* insert in map, InstanceID is generated automatically */
  list<QmpbInstance>::iterator insert_instance(const QmpbInstance&);
  list<QmpbInstance>::iterator delete_instance(const InstanceID&);
  InstanceID get_free_instid (const QmpbInstance&) const;

  bool operator== (const QmpbSite& a) const {return sid == a.get_sid();}
  bool operator!= (const QmpbSite& a) const {return sid != a.get_sid();}
  bool operator< (const QmpbSite& a) const {return sid < a.get_sid();}
  bool operator> (const QmpbSite& a) const {return sid > a.get_sid();}

  //!nowrap!-
private:
  string name;
  SiteID sid;
};

// form list functions

inline std::list<QmpbInstance> QmpbSite::get_form(const string& k) const
{
  bool found = false;
  std::list<QmpbInstance> result;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if ((*i).get_name() == k){
    found = true;
    result.push_back(*i);
   }
  }
  if (!found)
    ::error("QmpbSite::get_form: key not found");
  else
    return result;
}

inline QmpbInstance QmpbSite::get_instance(const InstanceID& k) const
{
  bool found = false;
  std::list<QmpbInstance>::const_iterator j;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if ((*i).get_instid() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
    ::error("QmpbSite::get_instance: key not found");
  else
    return (*j);
}

inline std::list<QmpbInstance>::iterator QmpbSite::delete_instance(const InstanceID& k)
{
  bool found = false;
  std::list<QmpbInstance>::iterator j;
  for (list<QmpbInstance>::iterator i = instances.begin(); i != instances.end(); ++i){
   if ((*i).get_instid() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
    ::error("QmpbSite::get_instance: key not found");
  else {
    instances.erase(j);
    return instances.begin();
  }
}

// Existence of element (not STL-like

inline bool QmpbSite::contains_form(const string& k) const
{
  bool found = false;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if ((*i).get_name() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline bool QmpbSite::contains_instance(const InstanceID& k) const
{
  bool found = false;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if ((*i).get_instid() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline bool QmpbSite::contains_ligand(const string& k) const
{
  bool found = false;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   for (map<string,int>::const_iterator j = (*i).ligandmap.begin(); j != (*i).ligandmap.end(); ++j){
    if (j->first == k && j->second > 0){
     found = true;
     break;
    }
   }
  }
  return found;
}

inline int QmpbSite::get_max_confnum() const
{
  int maxn = -100000000;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (maxn < (*i).get_confnum())
    maxn = (*i).get_confnum();
  }
  return maxn;
}

inline int QmpbSite::get_nconf() const
{
  std::list<int> cl;
  for (list<QmpbInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   cl.push_back((*i).get_confnum());
  }
  cl.sort();
  cl.unique();
  return cl.size();
}

// easier to use insert (not quite STL-like)
inline std::list<QmpbInstance>::iterator QmpbSite::insert_instance (const QmpbInstance& a){
  QmpbInstance b(a);
  InstanceID id = get_free_instid(b);
  b.set_instid(id);
  if (!contains_instance(id))
    instances.push_back(b);
  else
    ::error("QmpbSite::insert_instance: attempt to insert a duplicate instance\n");  
  return instances.end();
}

inline InstanceID QmpbSite::get_free_instid(const QmpbInstance& a) const {
  // start from 1 because instid 0 is meant to mark background parts
  int maxinst = int(1);
  InstanceID id(a.get_instid());
  while (contains_instance(id)){
   id = InstanceID(sid.get_sitename(), id.get_instname(), sid.get_conformer(), sid.get_chain(), sid.get_site(), maxinst);
   maxinst++;
  }
  return id;
}

#endif

// QmpbSite.h ends here
