/*  A QMPB Site 

    label, forms
    
    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann 
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _QmpbSiteDef_h
#define _QmpbSiteDef_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/QmpbFormDef.h"
#include "MEAD_e/globals.h"

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class QmpbSiteDef {
public:
  QmpbSiteDef();
  QmpbSiteDef (const string& name);
  QmpbSiteDef (const QmpbSiteDef& a);
  ~QmpbSiteDef () {}
  //!nowrap!+
  QmpbSiteDef& operator=(const QmpbSiteDef& a);

  //!nowrap!-
  void read_est();
  void read_est(const string&);
  void print_est();

  string get_name() const {return name;}
      
  std::list<QmpbFormDef> forms;

  // lookup attype by CharmmID or resname/atname pair

  /* check by label if form or ligand are already listed */
  bool contains_form (const string&) const;
  bool contains_ligand (const string&) const;
  /* check if atom occurs in any form */
  bool contains_atom (const string&, const string&) const;
  bool contains_atom (const CharmmID&) const;
  /* check if an AtomSet contains any atom occuring in any form */  
  bool atoms_complete (const AtomSet&) const;

  /* retrieve form by form label */  
  QmpbFormDef get_form(const string&) const;

  /* insert in list */
  std::list<QmpbFormDef>::iterator insert_form(const QmpbFormDef&);

  bool operator== (const QmpbSiteDef& a) const {return name == a.get_name();}
  bool operator!= (const QmpbSiteDef& a) const {return name != a.get_name();}
  bool operator< (const QmpbSiteDef& a) const {return name < a.get_name();}
  bool operator> (const QmpbSiteDef& a) const {return name > a.get_name();}

  //!nowrap!-
private:
  string name;
};

// form list functions

inline QmpbFormDef QmpbSiteDef::get_form(const string& k) const
{
  bool found = false;
  std::list<QmpbFormDef>::const_iterator j;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i){
   if ((*i).get_name() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (! found)
    ::error("QmpbSiteDef::get_form: key not found");
  else
    return (*j);
}

// Existence of element (not STL-like

inline bool QmpbSiteDef::contains_form(const string& k) const
{
  bool found = false;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(), !found; ++i){
   if ((*i).get_name() == k){
    found = true;
   }
  }
  return found;
}

inline bool QmpbSiteDef::contains_ligand(const string& k) const
{
  bool found = false;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(), !found; ++i){
   for (std::map<string,int>::const_iterator j = (*i).ligandmap.begin(); j != (*i).ligandmap.end(), !found; ++j){
    if (j->first == k)
     found = true;
   }
  }
  return found;
}

inline bool QmpbSiteDef::contains_atom(const string& k, const string& l) const
{
  bool found = false;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i){
   if ((*i).contains_atom(k,l)){
    found = true;
    break;
   }  
  }
  return found;
}

inline bool QmpbSiteDef::contains_atom(const CharmmID& k) const
{
  bool found = false;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(), !found; ++i){
   if ((*i).contains_atom(k)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool QmpbSiteDef::atoms_complete(const AtomSet& a) const
{
  bool complete = true;
  list<CharmmID> alist;
  for (std::list<QmpbFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i){   
   for (std::map<CharmmID,pair<float, int> >::const_iterator j=(*i).atommap.begin(); j!=(*i).atommap.end(); ++j){
    alist.push_back(j->first);
   }
  }
  alist.sort();
  alist.unique();
  for (std::list<CharmmID>::const_iterator i = alist.begin(); i != alist.end(); ++i){
   bool found = false;
   for (AtomSet::const_iterator j = a.begin(); j != a.end(); ++j){
    if (*i == CharmmID(j->second)){
     found = true;
     break;
    }
   }
   if (!found){
    complete = false;
    break;
   }
  }
  return complete;
}

// easier to use insert (not quite STL-like)
inline std::list<QmpbFormDef>::iterator QmpbSiteDef::insert_form (const QmpbFormDef& a){
  if (contains_form(a.get_name())) {
    cerr << name << " : " << a.get_name() << endl;
    ::error("QmpbSiteDef::insert_form: attempt to insert a duplicate form\n");
  }
  forms.push_back(a);
  return forms.end();
}


#endif

// QmpbSiteDef.h ends here
