/*  Rotlib is a representation of a backbone dependent rotamer library 
    
    short lists of sidechain rotamers represented by the container 
    RotData (sidechain torsion angles, occupation probability) 
    in an associative array keyed by RotID (resname, backbone torsions)
    
    TRU 2007

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.


$Id: Rotlib.cc,v 1.29 2005/11/09 23:18:24 bashford Exp $
*/
#include "MEAD_e/Rotlib.h"
#include "MEAD_e/globals.h"
#include <fstream>
#include <boost/regex.hpp>
#include <sstream>
using std::istringstream;

typedef std::list<Atom>::iterator alitr;
typedef std::list<Atom>::const_iterator calitr;

Rotlib::Rotlib(const Rotlib& a)
: map<RotID,list<RotData> >(a), name(a.name)
{
} 


Rotlib::Rotlib()
: name("")
{
}
Rotlib::Rotlib (const string& nm)
: name(nm)
{

}

Rotlib::Rotlib(const RotID& k, const list<RotData>& alin)
{
  insert(k, alin);
}

Rotlib& Rotlib::operator=(const Rotlib& a)
{
  map<RotID,list<RotData> >::operator=(a);
  name = a.name;
  return *this;
}


/* FIXME
   read is a temporary hack for testing purposes.
   Probably should redo this whole mess with flex.
*/
//#include <string.h>
#include <stdio.h>
void Rotlib::read (const string& fn_string)
{
  blab2 << "Rotlib::read: starting\n";
  int rcount = 0;

  const char *fn = fn_string.c_str();
  string line, tmpstr1, tmpstr2;
  string resname;
  int lastphi = -1000; int lastpsi = -1000;
  int phi, psi;
  float p, x1, x2, x3, x4;
  std::list<RotData> tmplist;
  
  boost::regex re_sq("^\\s*(\\w+)\\s+([\\d+-]+)\\s+([\\d+-]+)\\s+\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;  

  ifstream topfile(fn);
  
  if (!topfile)
    ::error ("Rotlib::read: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (! topfile.eof() ){
   
   std::string::const_iterator start, end;
   
   getline (topfile, line); 
   start = line.begin();
   end   = line.end();
   
   // look for SQUIRELL BB type line 
   if (boost::regex_search(start, end, match, re_sq, flags)){
    resname = std::string(match[1].first, match[1].second);
    istringstream ist1(std::string(match[2].first, match[2].second));
    ist1 >> phi;
    istringstream ist2(std::string(match[3].first, match[3].second));
    ist2 >> psi;
    istringstream ist3(std::string(match[4].first, match[4].second));
    ist3 >> p;
    istringstream ist4(std::string(match[5].first, match[5].second));
    ist4 >> x1;
    istringstream ist5(std::string(match[6].first, match[6].second));
    ist5 >> x2;
    istringstream ist6(std::string(match[7].first, match[7].second));
    ist6 >> x3;
    istringstream ist7(std::string(match[8].first, match[8].second));
    ist7 >> x4;
    
    RotData tmpdata(p, x1, x2, x3, x4);
    tmplist.push_back(tmpdata);
    
    if (phi != lastphi || psi != lastpsi){
     // first insert add old RotID and list of RotData to Rotlib ...
     if (!(lastphi == -1000 && lastpsi == -1000)){
      RotID tmpid(resname, lastphi, lastpsi);
      std::pair<std::map<RotID,std::list<RotData> >::iterator,bool> tmppair1 = insert(tmpid, tmplist);
      if (tmppair1.second){rcount++;}
      // ... then clear old data to begin new list 
      if (!tmplist.empty()){tmplist.clear();}
     }
     lastphi = phi; lastpsi = psi;
    }
   }
      
  } // end of parsing

  topfile.close();

  if (rcount != size()) {
    cerr << "WARNING Rotlib::read_top: Map length = " << size()
         << ", while in reading rcount = " << rcount << endl;
  }
  
  /////// debug
  
  //print();
  
  ///////// 
   
  blab1 << "Rotlib::read: finished with " << rcount << " entries" << endl;
}

void Rotlib::print(){
     cout << "\nBackbone Dependent Rotamer Library named " << name << " :\n" << endl;  
     cout << "resname   phi   psi   :   occupation probability   chi1   chi2   chi3   chi4" << endl;
     for (Rotlib::const_iterator j=begin(); j!=end(); ++j) {   
       RotID tmprid = j->first;       
       for (std::list<RotData>::const_iterator i = (j->second).begin(); i != (j->second).end(); ++i) { 
         cout << tmprid << "   :   " << (*i).p << "   " << 
                 (*i).x1 << "   " << (*i).x2 << "   " << (*i).x3 << "   " << (*i).x4 << "   " << endl; 

       }
     }
}

// Rotlib.cc ends here
