/*  Rotlib is a representation of a backbone dependent rotamer library 
    
    short lists of sidechain rotamers represented by the container 
    RotData (sidechain torsion angles, occupation probability) 
    in an associative array keyed by RotID (resname, backbone torsions)
    
    TRU 2007

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/

#ifndef _Rotlib_h
#define _Rotlib_h 1

#include <map>
#include <list>
#include <string>
#include "MEAD_e/RotData.h"
#include "MEAD_e/RotID.h"
#include "MEAD_e/globals.h"
#include <vector>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

class DeltaRotlib;

// Needed for swig typemap
typedef std::list<RotData> list_RotData;

//!wrap!
class Rotlib: public map<RotID, list<RotData> > {
public:
  Rotlib();
  Rotlib (const string& name);
  Rotlib (const Rotlib& a);
  Rotlib (const RotID& k, const list<RotData>& alin);
  ~Rotlib () {}
  //!nowrap!+
  Rotlib& operator=(const Rotlib& a);
  //!nowrap!-
  void read (const string& filename);
  void print ();
  string get_name() const {return name;}

  // some deviations from STL container functionality:

  // associative lookup, but don't create if missing
  const list<RotData>& operator[] (const RotID&) const;
  //!nowrap!-
  list<RotData>& operator[] (const RotID&);

  //!nowrap!+
  bool contains_residue(const string&) const;
  bool contains(const RotID&) const; // Existence of element (not STL-like)
  pair<iterator,bool> insert (const RotID&, const list<RotData>&); // easier-to-use insert
  //!nowrap!-
private:
  string name;
};

#if SWIGPP_LITERAL_INCLUDE

%addmethods Rotlib {
// Wrap this private member as a read-only attribute for Python, but
// this is a bit of a hack (see below).
// It'd probably be better to just wrap the get_name() function,
// but for Python it seems more natural to treat this simple type
// as an attribute.

  string name;
//
// Not really dict-like, but it's the only way to add new atoms
  void insert (const RotID& rid, const list<RotData>& rdt) { self->insert(rid, rdt); }
//
// Functions to emulate a Python dict
//
  void clear () { dict_clear(self); }
  Rotlib * copy() { return new Rotlib(*self); }
  int has_key (const RotID& k) { return dict_has_key(self, k); }
  int __len__ () { return dict___len__(self); }
  void __delitem__ (const RotID& k) { dict___delitem__(self, k); }
// Adds all elements from ats
  void update (const Rotlib& ats) { dict_update(self, ats); }
// This function will return a list of RotID keys from map_AtomID_Atom,
// which later is converted to a python list of keys by a typemap
//  list_RotID * keys() { return dict_keys(self); }
// This function will return a list of RotData lists
// which later is converted to a python list of values by a typemap
//  list_RotPtr * values() { return dict_values(self); }

};

%wrapper %{
//
// Provide the read-only get/set functions for the aboved wrapped attribute
void Rotlib_name_set(Rotlib *self, string *name){
   _SWIG_exception(SWIG_ValueError, "The name attribute is read-only");
}
// This type is returned by reference, so create a global placeholder
// to stash the result to be returned (read-only)
// Note that this works OK from Python since assignment of
// simple objects like numbers and strings effectively creates a copy.
// However, using this global reference certainly is not thread safe!
string _Rotlib_name_readonly;
string * Rotlib_name_get(Rotlib *self){_Rotlib_name_readonly = self->get_name(); return &_Rotlib_name_readonly;}

%}

#endif // SWIGPP_LITERAL_INCLUDE

// associative lookup, but don't create if missing
inline const list<RotData>& Rotlib::operator[] (const RotID& k) const
{
  const_iterator i = find(k);
  if (i == end())
    ::error("Rotlib::operator[]: key not found");
  return i->second;
}

inline list<RotData>& Rotlib::operator[] (const RotID& k)
{
  iterator i = find(k);
  if (i == end())
    ::error("Rotlib::operator[]: key not found");
  return i->second;
}

// Existence of element (not STL-like
inline bool Rotlib::contains(const RotID& k) const
{
  return find(k) != end();
}

// easier to use insert (not quite STL-like)
inline pair<Rotlib::iterator,bool> Rotlib::insert (const RotID& k, const list<RotData>& l){
  if (contains(k)) {
    ::error("Rotlib::insert: attempt to insert a duplicate rotamer entry\n");
  }
  return map<RotID,list<RotData> >::insert(value_type(k, l));
}

inline bool Rotlib::contains_residue(const string& k) const
{
  bool found = false;
  for (const_iterator i = begin(); i!=end(); ++i) {
    if (i->first.get_resname() == k){
     found = true;
     break;
    }
  }
  return found;
}


#endif

// Rotlib.h ends here
