/*  class SiteID is a key type for an associative array of 
    sites each containing an array of instances
    
    TRU 2007
    
    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

*/
#ifndef _SiteID_h
#define _SiteID_h 1

#include <string>
#include <iostream>
#include "MEAD_e/Atom.h"
using std::string;

//!wrap!
class SiteID {
public:
  SiteID ();
  SiteID (const string& site_name, const int conformer = 0, const int chain = 0, const int site = 0);
  //const string& chainid = "");
  SiteID (const SiteID&);
  SiteID (const Atom&);
  ~SiteID () {}
  //!nowrap!
  SiteID& operator= (const SiteID&);
  bool operator== (const SiteID&) const;
  bool operator!= (const SiteID& a) const;
  bool operator<(const SiteID&) const;
  bool operator>(const SiteID&) const;
  //!nowrap!+
  ostream& print(ostream&);
  int get_conformer() const {return conformer;}
  int get_chain() const {return chain;}
  int get_site() const {return site;}
  string get_sitename() const {return sitename;}
  //!nowrap!-
private:
  string sitename;
  int conformer;
  int chain;
  int site;
};

#if SWIGPP_LITERAL_INCLUDE

%addmethods SiteID {
// Wrap these private members as read-only attributes for Python, but
// this is a bit of a hack (see below).
// It'd probably be better to just wrap the get_member() functions,
// but for Python's use it seems more natural to treat these simple types
// as attributes.
  string sitename;
  int confomer;
  int chain;
  int site;

// swigpp.el strips the default value of the string argument in the
// above constructor. So, wrap this constructor with the default
// argument removed and...
  SiteID (const string& site_name);
};

%wrapper %{
// ...provide the wrapped constructor with the default argument supplied
SiteID * new_SiteID(int residue_number, const string& atom_name){
  return new SiteID(residue_number, atom_name, "");
}
// Provide the read-only get/set functions for the aboved wrapped attributes
void SiteID_sitename_set(SiteID *self, string *sitename){
   _SWIG_exception(SWIG_ValueError, "The sitename attribute is read-only");
}
int SiteID_resnum_get(SiteID *self){return self->get_resnum();}
// These strings are returned by reference, so create global placeholders
// to stash the results to be returned (read-only).
// Note that this works fine from Python since assignment of
// simple objects like numbers and strings effectively creates a copy.
// However, using these global references certainly is not thread safe!
string _SiteID_sitename_readonly;
string * SiteID_atname_get(SiteID *self){_SiteID_sitename_readonly = self->get_sitename(); return &_SiteID_sitename_readonly;}
%}

#endif // SWIGPP_LITERAL_INCLUDE

inline SiteID::SiteID (const SiteID& a)
  : sitename(a.sitename), conformer(a.conformer), chain(a.chain), site(a.site) {}

inline SiteID& SiteID::operator= (const SiteID& a)
{sitename = a.sitename; conformer = a.conformer; chain = a.chain; site = a.site;
 return *this;}

inline SiteID::SiteID (const Atom& a)
  : sitename(a.resname), conformer(a.conformer), chain(a.chain), site(a.site) {}
 
inline bool SiteID::operator== (const SiteID& a) const
{return sitename == a.sitename && conformer == a.conformer && chain == a.chain && site == a.site;}

inline bool SiteID::operator!= (const SiteID& a) const
{return !(*this == a);}

inline bool SiteID::operator<(const SiteID& x) const
{
    if (conformer == x.conformer) {
      if (chain == x.chain) {
        if (site == x.site) {
          return sitename < x.sitename;
      } else
	return site<x.site;
    } else
      return chain<x.chain;
  } else
    return conformer<x.conformer;
}

inline bool SiteID::operator>(const SiteID& x) const
{
    if (conformer == x.conformer) {
      if (chain == x.chain) {
        if (site == x.site) {
          return sitename > x.sitename;
      } else
	return site>x.site;
    } else
      return chain>x.chain;
  } else
    return conformer>x.conformer;
}

inline ostream& operator<<(ostream& ost, SiteID aid)
{return aid.print(ost);}

#endif

// SiteID.h ends here
