/* Definition of some global constants, variables and functions
    
    changes by TRU 2007 :
    
    recent values of constants from CODATA
    1 thermochemical calory = 4.184 Joule by definition 
    (previously international steam table calorie 4.1868)
    
    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: globals.cc,v 1.25 2004/12/06 17:58:27 bashford Exp $
*/



#include "MEAD/globals.h"
#include "MEAD/PhysCond.h"
#include <iostream>
#include <fstream>

#ifdef USE_EXCEPTIONS
#include "MEAD/MEADexcept.h"
#endif

// Strictly for debugging:
#ifdef POPULATION_COUNT
#include "MEAD/Vertex.h"
#include "MEAD/VertElem.h"
 int Vertex::population = 0;
 int Vertex::max_population = 0;
 int Vertex::num_created = 0;
 int Vertex::num_deleted = 0;
 int VertElem::population = 0;
 int VertElem::max_population = 0;
 int VertElem::num_created = 0;
 int VertElem::num_deleted = 0;
#endif

ofstream cnull ("/dev/null");
ostream *blab1pt = &cnull;
ostream *blab2pt = &cnull;
ostream *blab3pt = &cnull;




const float pi = 3.141592654;
const float ln10 = 2.3025851;
const double Boltzmann_constant = 1.3806505e-23;  // J/K
// UNUSED const double Gas_constant = 8.314472;  // J/mol/K
const double Avogadro_Number = 6.0221415e+23; // particles/mol
const double elementary_charge = 1.60217531e-19; // Coulombs
const double light_speed = 299792458.0;  // m/s
const double inv_4pi_vacuum_permittivity = light_speed*light_speed*1.0e-7;
                                        // F/m
// The standard energy unit internally to the program is
// charge units squared divided by length, or elem_ch**2/Angst
const double energy_unit = elementary_charge * elementary_charge / 1.0e-10 
                        * inv_4pi_vacuum_permittivity;

std::string DOT = ".";

float PhysCond::epsext = 80.0;
float PhysCond::solrad = 1.4;
float PhysCond::sterln = 2.0;
float PhysCond::ionicstr = 0.0;
float PhysCond::membpot = 0.0; // membrane potential (V)

float PhysCond::T = 298.15;
float PhysCond::Avogadro_Number = 6.0221415e+23;
float PhysCond::Boltzmann_constant = 1.3806505e-23;
float PhysCond::kBolt = Boltzmann_constant  * 1.0e-10 / (elementary_charge * elementary_charge) 
                        / (light_speed * light_speed * 1.0e-7);
                                   // (proton ch sq)/(Angst * Kelvin)
                                   // or is 5.98339e-6 a better value??
float PhysCond::conconv =  Avogadro_Number * 1.0e-27;
                                     // ((particles/cubic Ang)/(moles/liter))
//float PhysCond::econv = energy_unit * Avogadro_Number / 4184.0;
float PhysCond::econv =  1.60217531e-19 * 1.60217531e-19 / 1.0e-10 
                         * light_speed*light_speed * 1.0e-7
			 * Avogadro_Number / 4184.0;                             
                             // ((kcal/mole)/(proton ch sq/Angst))
float PhysCond::pkconv =  1.60217531e-19 * 1.60217531e-19 / 1.0e-10 
                         * light_speed*light_speed * 1.0e-7
			  / Boltzmann_constant / 298.15 / 2.302585093;                             
                             // ((pK units)/(proton ch sq/Angst))
float PhysCond::pkcfac =  1.60217531e-19 * 1.60217531e-19 / 1.0e-10 
                         * light_speed*light_speed * 1.0e-7
			  / Boltzmann_constant / 2.302585093;                             
                             // ((pK units * T)/(proton ch sq/Angst))
float PhysCond::voltconv = elementary_charge / 1.0e-10
                          * inv_4pi_vacuum_permittivity;
                             // ((Volt)/(proton ch /Angst))
float PhysCond::energy_unit = elementary_charge * elementary_charge / 1.0e-10
                          * inv_4pi_vacuum_permittivity;
float PhysCond::elementary_charge = 1.60217531e-19;
float PhysCond::vacuum_permittivity = 1 / (4 * pi * light_speed * light_speed * 1.0e-7);

			  
// The standard energy unit internally to the program is
// charge units squared divided by length, or elem_ch**2/Angst			  
float PhysCond::hueck = 8 * pi * PhysCond::conconv * PhysCond::ionicstr
                      / PhysCond::kBolt / PhysCond::T;
float PhysCond::kappasq = PhysCond::hueck / PhysCond::epsext;
float PhysCond::ln10kT = ln10 * PhysCond::kBolt * PhysCond::T;
float PhysCond::bohr_radius = 0.529177;
float PhysCond::proton_charge = 1.0;

void error (const char *s1, const char *s2, const char *s3)
{
  cerr << s1 << s2 << s3 << "\n";
#ifdef USE_EXCEPTIONS
  throw MEADexcept(s1, s2, s3);
#else
  exit(1);
#endif
}

void safeopen(ofstream& f, std::string filename_string)
{
  const char *filename = filename_string.c_str();
  if (fexists(filename)) {
    ::error("ERROR: safeopen: File, ", filename,
	    " already exists\n");
  }
#if defined(__sgi) && !defined(__GNUC__) && defined(_STANDARD_C_PLUS_PLUS)
  // Workaround for stupid bug in SGI version 7.0 C++ header file fstream
  // (basic_ofstream::open uses "in" where it should use "out")
  // Result is that open fails if file doesn't already exist.
  // FIXME!  Get rid of this after a while when SGI fixes this bug.
  {
    ofstream b(filename);  // touch the file so it exists.
  }
  // Now SGI's open will work in spite of the bug.
#endif
  f.open(filename, std::ios::out);
  if (!f) {
    ::error("ERROR: safeopen: File, ", filename,
	    " could not be opened for writing\n");
  }
}

// globals.cc ends here
