/*  Electrostatic potential for a membrane -*- C++ -*-

    Electrostatic transmembrane potential for a membrane with
    mobile ions also in the core / a wide pore in a membrane,
    filled with electrolyte

    TRU, 2007

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: AnalyMemb.h,v 2.7 2004/12/06 17:57:51 bashford Exp $
*/
#ifndef _AnalyMemb_h
#define _AnalyMemb_h 1

#include "MEAD_e/ElstatPot.h"

#define AnalyMemb_MAXTERM 20

class DielectricSlab;

//!wrap!
class AnalyMemb : public AnalyticEP {
public:
  //!nowrap!
//  AnalyMemb (DielectricSlab*, ChargeDist_lett*, ElectrolyteEnvironment_lett*,
//	int maxterm=AnalyMemb_MAXTERM);  // No. of terms in image expression.
  AnalyMemb (DielectricSlab*, ChargeDist_lett*, ElectrolyteEnvironment_lett*);
  ~AnalyMemb() {}
  virtual void solve();
  virtual float value(Coord) const ;
  virtual Coord field(Coord x) const ;
  virtual Coord displacement(Coord x) const ;
  float get_zlower() const;
  float get_zupper() const;
  int get_maxterm() const;
private:
  ChargeDist_lett* rho;
  DielectricSlab* eps;
  ElectrolyteEnvironment_lett* electrolyte;
  float epsslab, epsext;
  float zupper, zlower;

  float zmiddle, e_kappam_l, lhalf;
  float membl, istr1, istr2, istr, em_by_es, kappam, kappas;
  float kappam_by_kappas, em_by_es_times_kappam_by_kappas;
  float fac0, fac1, fac2, fac3, membpothalf;

  int maxterm;

  int solved;
};

inline float AnalyMemb::get_zlower() const {return zlower;}
inline float AnalyMemb::get_zupper() const {return zupper;}
inline int AnalyMemb::get_maxterm() const {return maxterm;}

#if SWIGPP_LITERAL_INCLUDE

%addmethods AnalyMemb {
//
// swigpp.el strips the default value of the maxterm argument in the
// above constructor. So, wrap this constructor with the default
// argument both present and removed.
  AnalyMemb (DielectricSlab*, ChargeDist_lett*, ElectrolyteEnvironment_lett*, int);
  AnalyMemb (DielectricSlab*, ChargeDist_lett*, ElectrolyteEnvironment_lett*);
//
// Multiply with a ChargeDist_lett derived class.
// (The __rmul__ version is defined in the ChargeDist_lett derived classes
//
  float operator* (const AnalyMemb& asl, const ChargeDist_lett& cdl);
//
// Scale operations
//
  ElstatPotCombination operator* (AnalyMemb& asl, float scale);
  ElstatPotCombination operator* (float scale, AnalyMemb& asl);
  ElstatPotCombination operator/ (AnalyMemb& asl, float scale);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Provide the wrapped constructors.
AnalyMemb * new_AnalyMemb (DielectricSlab* ds, ChargeDist_lett* cd, ElectrolyteEnvironment_lett* ee, int maxterm){
  return new AnalyMemb(ds, cd, ee, maxterm);
}
AnalyMemb * new_AnalyMemb (DielectricSlab* ds, ChargeDist_lett* cd, ElectrolyteEnvironment_lett* ee){
  return new AnalyMemb(ds, cd, ee, AnalyMemb_MAXTERM);
}

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE

#endif

// AnalyMemb.h ends here
