/*  Object containing CHARMM parameters and methods for lookup

    lists of CHARMM parameters associated with corresponding
    atom type(s): bond, angle, dihedral, improper, nonbonded

    I/O functions for reading CHARMM topology and parameter files

    lookup functions for finding parameters given CHARMM
    atom type(s) or CharmmID(s)

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/CharmmPara.h"
#include "MEAD_e/globals.h"
#include <fstream>
#include <boost/regex.hpp>
#include <sstream>
#include <stdio.h>
using std::istringstream;
using std::stringstream;


CharmmPara::CharmmPara(const CharmmPara& a)
: name(a.name), typeindex(a.typeindex),
  NB(a.NB), Bond(a.Bond), Angle(a.Angle),
  Dihedral(a.Dihedral),Improper(a.Improper)
{
}


CharmmPara::CharmmPara()
: name("")
{
}
CharmmPara::CharmmPara (const string& nm)
: name(nm)
{
}

void CharmmPara::read ()
{
  string top_fn = "top" + DOT + "inp";
  string par_fn = "par" + DOT + "inp";
  read_top(top_fn);
  read_par(par_fn);
}

void CharmmPara::read (const string& fn1, const string& fn2)
{
  read_top(fn1);
  read_par(fn2);
}

void CharmmPara::read_par (const string& fn_string){

  blab2 << "CharmmPara::read_par: starting\n";

  const char *fn = fn_string.c_str();
  string line, tmpstr1, tmpstr2;
  string attype1, attype2, attype3, attype4;
  float tmpf1, tmpf2, tmpf3, tmpf4;

  bool in_bond = false;
  bool in_angl = false;
  bool in_dihe = false;
  bool in_impr = false;
  bool in_nonb = false;

  int ntype = 0;
  int nbond = 0;
  int nangl = 0;
  int ndihe = 0;
  int nimpr = 0;
  int nnonb = 0;

  boost::regex re_st_bond("^\\s*BONDS\\s*");
  boost::regex re_st_angl("^\\s*ANGLES\\s*");
  boost::regex re_st_dihe("^\\s*DIHEDRALS\\s*");
  boost::regex re_st_impr("^\\s*IMPROPER\\s*");
  boost::regex re_st_nonb("^\\s*NONBONDED\\s*");

  boost::regex re_nb  ("^\\s*(\\w+)\\s+[\\d.+-eE]+\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::regex re_nb14("^\\s*(\\w+)\\s+[\\d.+-eE]+\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+[\\d.+-eE]+\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::regex re_bond("^\\s*(\\w+)\\s+(\\w+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::regex re_angl("^\\s*(\\w+)\\s+(\\w+)\\s+(\\w+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::regex re_anglub("^\\s*(\\w+)\\s+(\\w+)\\s+(\\w+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)\\s+([\\d.+-eE]+)");
  boost::regex re_dihe("^\\s*(\\w+)\\s*(\\w+)\\s+(\\w+)\\s+(\\w+)\\s+([\\d.+-eE]+)\\s+(\\d+)\\s+([\\d.+-eE]+)");
  boost::regex re_impr("^\\s*(\\w+)\\s*(\\w+)\\s+(\\w+)\\s+(\\w+)\\s+([\\d.+-eE]+)\\s+\\d{1}\\s+([\\d.+-eE]+)");

  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  ifstream parfile(fn);

  if (!parfile)
    ::error ("CharmmPara::read_par: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (! parfile.eof() ){

   std::string::const_iterator start, end;

   getline (parfile, line);
   start = line.begin();
   end   = line.end();

   // begin of section
   if (boost::regex_search(start, end, match, re_st_bond, flags)){
    in_bond = true;  in_angl = false; in_dihe = false; in_impr = false; in_nonb = false;
   } else if (boost::regex_search(start, end, match, re_st_angl, flags)){
    in_bond = false; in_angl = true;  in_dihe = false; in_impr = false; in_nonb = false;
   } else if (boost::regex_search(start, end, match, re_st_dihe, flags)){
    in_bond = false; in_angl = false; in_dihe = true;  in_impr = false; in_nonb = false;
   } else if (boost::regex_search(start, end, match, re_st_impr, flags)){
    in_bond = false; in_angl = false; in_dihe = false; in_impr = true;  in_nonb = false;
   } else if (boost::regex_search(start, end, match, re_st_nonb, flags)){
    in_bond = false; in_angl = false; in_dihe = false; in_impr = false; in_nonb = true;
   }

   if (in_nonb){
    // look for string attype, float ignored, float eps, float Rmin/2,
    // float ignored, {float eps 1-4, float Rmin/2 1-4}
    if (boost::regex_search(start, end, match, re_nb14, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     istringstream ist1(std::string(match[2].first, match[2].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[3].first, match[3].second));
     ist2 >> tmpf2;
     istringstream ist3(std::string(match[4].first, match[4].second));
     ist3 >> tmpf3;
     istringstream ist4(std::string(match[5].first, match[5].second));
     ist4 >> tmpf4;
     std::list<CharmmParaNB>::iterator tmpi = insert_NB(attype1, tmpf1, tmpf2, tmpf3, tmpf4, 0);
     nnonb++;
    }
    else if (boost::regex_search(start, end, match, re_nb, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     istringstream ist1(std::string(match[2].first, match[2].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[3].first, match[3].second));
     ist2 >> tmpf2;
     std::list<CharmmParaNB>::iterator tmpi = insert_NB(attype1, tmpf1, tmpf2, 0, 0, 0);
     nnonb++;
    }
   }
   else if (in_bond){
    // look for string attype1, string attype2, float Kb, float b0
    if (boost::regex_search(start, end, match, re_bond, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     attype2 = std::string(match[2].first, match[2].second);
     istringstream ist1(std::string(match[3].first, match[3].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[4].first, match[4].second));
     ist2 >> tmpf2;
     std::list<CharmmParaBond>::iterator tmpi = insert_Bond(attype1, attype2, tmpf1, tmpf2);
     nbond++;
    }
   }
   else if (in_angl){
    // look for string attype1, attype2, attype3,
    // float Ktheta, float Theta0, {float Kub, float S0}
    if (boost::regex_search(start, end, match, re_anglub, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     attype2 = std::string(match[2].first, match[2].second);
     attype3 = std::string(match[3].first, match[3].second);
     istringstream ist1(std::string(match[4].first, match[4].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[5].first, match[5].second));
     ist2 >> tmpf2;
     istringstream ist3(std::string(match[6].first, match[6].second));
     ist3 >> tmpf3;
     istringstream ist4(std::string(match[7].first, match[7].second));
     ist4 >> tmpf4;
     std::list<CharmmParaAngle>::iterator tmpi = insert_Angle(attype1, attype2, attype3, tmpf1, tmpf2, tmpf3, tmpf4);
     nangl++;
    }
    else if (boost::regex_search(start, end, match, re_angl, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     attype2 = std::string(match[2].first, match[2].second);
     attype3 = std::string(match[3].first, match[3].second);
     istringstream ist1(std::string(match[4].first, match[4].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[5].first, match[5].second));
     ist2 >> tmpf2;
     std::list<CharmmParaAngle>::iterator tmpi = insert_Angle(attype1, attype2, attype3, tmpf1, tmpf2, 0, 0);
     nangl++;
    }
   }
   else if (in_dihe){
    // look for string attype1, string attype2, string attype3, string attype4,
    // float Kchi, int n, float delta
    if (boost::regex_search(start, end, match, re_dihe, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     attype2 = std::string(match[2].first, match[2].second);
     attype3 = std::string(match[3].first, match[3].second);
     attype4 = std::string(match[4].first, match[4].second);
     istringstream ist1(std::string(match[5].first, match[5].second));
     ist1 >> tmpf1;
     int n;
     istringstream ist2(std::string(match[6].first, match[6].second));
     ist2 >> n;
     istringstream ist3(std::string(match[7].first, match[7].second));
     ist3 >> tmpf2;
     std::list<CharmmParaDihedral>::iterator tmpi = insert_Dihedral(attype1, attype2, attype3, attype4, n, tmpf1, tmpf2);
     ndihe++;
    }
   }
   else if (in_impr){
    // look for string attype1, string attype2, string attype3, string attype4,
    // float Kpsi, float psi0
    if (boost::regex_search(start, end, match, re_impr, flags)){
     attype1 = std::string(match[1].first, match[1].second);
     attype2 = std::string(match[2].first, match[2].second);
     attype3 = std::string(match[3].first, match[3].second);
     attype4 = std::string(match[4].first, match[4].second);
     istringstream ist1(std::string(match[5].first, match[5].second));
     ist1 >> tmpf1;
     istringstream ist2(std::string(match[6].first, match[6].second));
     ist2 >> tmpf2;
     std::list<CharmmParaImproper>::iterator tmpi = insert_Improper(attype1, attype2, attype3, attype4, tmpf1, tmpf2);
     nimpr++;
    }
   }

  } // end of parsing

  parfile.close();

  if (nnonb != NB.size()) {
    cerr << "WARNING CharmmPara::read_par: Map length of NB = " << NB.size()
         << ", while in attype assignment atom count = " << nnonb << endl;
  }
  if (nbond != Bond.size()) {
    cerr << "WARNING CharmmPara::read_par: Map length of Bond = " << Bond.size()
         << ", while in attype assignment atom count = " << nbond << endl;
  }
  if (nangl != Angle.size()) {
    cerr << "WARNING CharmmPara::read_par: Map length of Angle = " << Angle.size()
         << ", while in attype assignment atom count = " << nangl << endl;
  }
  if (ndihe != Dihedral.size()) {
    cerr << "WARNING CharmmPara::read_par: Map length of Dihedral = " << Dihedral.size()
         << ", while in attype assignment atom count = " << ndihe << endl;
  }
  if (nimpr != Improper.size()) {
    cerr << "WARNING CharmmPara::read_par: Map length of Improper = " << Improper.size()
         << ", while in attype assignment atom count = " << nimpr << endl;
  }

  /////// debug
  /*
  print_Bond();
  print_Angle();
  print_Dihedral();
  print_Improper();
  print_NB();
  */
  /////////

  blab2 << "CharmmPara::read_par: finished with " <<
  nbond << " BOND parameters, " <<
  nangl << " ANGLE parameters, " <<
  ndihe << " DIHEDRAL parameters, " <<
  nimpr << " IMPROPER parameters, " <<
  nnonb << " NONBONDED parameters" <<
  endl;

  blab1 << "CharmmPara::read_par: exiting\n";
}

void CharmmPara::read_top (const string& fn_string){

  blab2 << "CharmmPara::read_top: starting\n";
  const char *fn = fn_string.c_str();
  string line, tmpstr1, tmpstr2;
  string atname, attype, resname;
  map<pair<string, string>, int> bonds;
  list<vector<string> > impropers;
  list<string> atoms;
  float charge;
  bool inres  = false;
  bool inpres = false;

  boost::regex re_residue("^\\s*RESI\\w*\\s+([^\\s]+)");
  boost::regex re_pres("^\\s*PRES\\w*\\s+([^\\s]+)");
  boost::regex re_bond("^\\s*(BOND|DOUB|TRIP|AROM)\\w*\\s+([^\\s]+\\s+[^\\s]+[^!]*)");
  boost::regex re_impr("^\\s*(IMPR|IMPH)\\w*\\s+([^\\s]+\\s+[^\\s]+\\s+[^\\s]+\\s+[^\\s]+[^!]*)");
  boost::regex re_resend("^\\s*(IC|PRES|RESI|PATC)\\w*\\s+[^\\s]+");
  boost::regex re_presend("^\\s*(IC|RESI)\\w*\\s+[^\\s]+");
  boost::regex re_atom("^\\s*ATOM\\s+([^\\s]+)\\s+([^\\s]+)\\s+([\\d.+-eE]+)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  ifstream topfile(fn);

  if (!topfile)
    ::error ("CharmmPara::read_top: Cannot open input file, ", fn, "\n");

  int atcount     = 0;
  int chargecount = 0;
  int topcount    = 0;

  // start of parsing
  while (! topfile.eof() ){

   std::string::const_iterator start, end;

   getline (topfile, line);
   start = line.begin();
   end   = line.end();

   // end of atom/bond definitions
   if (inres){
    if (boost::regex_search(start, end, match, re_resend, flags)){
     inres = false;
     CharmmParaResTop tmpres(resname, atoms, bonds, impropers);
     insert_Topology(tmpres);
     topcount++;
     atoms.clear();
     bonds.clear();
     impropers.clear();
    }
   }
   if (inpres){
    if (boost::regex_search(start, end, match, re_presend, flags)){
     inpres = false;
    }
   }
   // begin of a new residue, update resname
   if (boost::regex_search(start, end, match, re_residue, flags)){
    inres = true; inpres = false;
    resname = std::string(match[1].first, match[1].second);
   }
   if (boost::regex_search(start, end, match, re_pres, flags)){
    inres = false; inpres = true;
    resname = std::string(match[1].first, match[1].second);
   }

   if (inres){
    // look for ATOM atname attype charge lines
    if (boost::regex_search(start, end, match, re_atom, flags)){
     atname = std::string(match[1].first, match[1].second);
     attype = std::string(match[2].first, match[2].second);
     istringstream ist(std::string(match[3].first, match[3].second));
     ist >> charge;
     pair<std::map<CharmmID,string>::iterator,bool> tmppair1 = insert_attype(resname, atname, attype);
     if (tmppair1.second){atcount++; atoms.push_back(atname);}
     pair<std::map<CharmmID,float>::iterator,bool> tmppair2 = insert_charge(resname, atname, charge);
     if (tmppair2.second){chargecount++;}
    }
    else if (boost::regex_search(start, end, match, re_bond, flags)){
     string tmpstr0, tmpstr1, tmpstr2;
     int n;
     tmpstr0 = std::string(match[2].first, match[2].second);
     stringstream ss1(tmpstr0);
     // bond order or 4 for aromatic
     tmpstr0 = std::string(match[1].first, match[1].second);
     if (tmpstr0 == "BOND"){
      n = int(1);
     } else if (tmpstr0 == "DOUB"){
      n = int(2);
     } else if (tmpstr0 == "TRIP"){
      n = int(3);
     } else if (tmpstr0 == "AROM"){
      n = int(4);
     }
     while (ss1){
      ss1 >> tmpstr1 >> tmpstr2;
      if (ss1.fail()){break;} else{
       pair<string,string> tmppair1(tmpstr1,tmpstr2);
       pair<pair<string,string>,int> tmppair2(tmppair1,n);
       bonds.insert(tmppair2);
      }
     }
    }
    else if (boost::regex_search(start, end, match, re_impr, flags)){
     string tmpstr1, tmpstr2, tmpstr3, tmpstr4;
     int n;
     tmpstr1 = std::string(match[1].first, match[1].second);
     stringstream ss1(tmpstr1);
     while (ss1){
      ss1 >> tmpstr1 >> tmpstr2 >> tmpstr3 >> tmpstr4;
      if (ss1.fail()){break;} else{
       vector<string> tmpvec;
       tmpvec.push_back(tmpstr1);
       tmpvec.push_back(tmpstr2);
       tmpvec.push_back(tmpstr3);
       tmpvec.push_back(tmpstr4);
       impropers.push_back(tmpvec);
      }
     }
    }
   }

   if (inpres){
    // look for ATOM atname attype charge lines
    if (boost::regex_search(start, end, match, re_atom, flags)){
     atname = std::string(match[1].first, match[1].second);
     attype = std::string(match[2].first, match[2].second);
     istringstream ist(std::string(match[3].first, match[3].second));
     ist >> charge;
     pair<std::map<CharmmID,string>::iterator,bool> tmppair1 = insert_pres_attype(resname, atname, attype);
    }
   }

  } // end of parsing

  topfile.close();

  if (atcount != typeindex.size()) {
    cerr << "WARNING CharmmPara::read_top: Map length of attypeindex = " << typeindex.size()
         << ", while in reading attype count = " << atcount << endl;
  }
  if (chargecount != chargeindex.size()) {
    cerr << "WARNING CharmmPara::read_top: Map length of chargeindex = " << chargeindex.size()
         << ", while in reading charge count = " << chargecount << endl;
  }
  if (topcount != Topology.size()) {
    cerr << "WARNING CharmmPara::read_top: Map length of Topology = " << Topology.size()
         << ", while in reading topology count = " << topcount << endl;
  }

  /////// debug
  /*
  print_typeindex();
  print_pres_typeindex();
  print_chargeindex();
  print_Topology();
  */
  /////////

  blab1 << "CharmmPara::read_top: finished with " << topcount << " residue topologies, "
                                                  << atcount << " atom types and "
                                                  << chargecount << " charges read." << endl;


}

void CharmmPara::print() const
{
  cout << endl << "*************************************************" << name << endl << endl;
  cout << endl << "printing CHARMM PARAMETER OBJECT named " << name << endl << endl;
  print_typeindex();
  print_chargeindex();
  print_Topology();
  print_NB();
  print_Bond();
  print_Angle();
  print_Dihedral();
  print_Improper();
}


void CharmmPara::print_typeindex() const
{
     cout << "Map of CharmmIDs and atom types:\n" << endl;
     cout << "RESNAME:ATNAME : ATTYPE\n" << endl;
     for (std::map<CharmmID,string>::const_iterator i = typeindex.begin(); i != typeindex.end(); ++i) {
       cout << i->first << " : " << i->second << endl;
     }
}

void CharmmPara::print_chargeindex() const
{
     cout << "Map of CharmmIDs and atom types:\n" << endl;
     cout << "RESNAME:ATNAME : CHARGE\n" << endl;
     for (std::map<CharmmID,float>::const_iterator i = chargeindex.begin(); i != chargeindex.end(); ++i) {
       cout << i->first << " : " << i->second << endl;
     }
}

void CharmmPara::print_Topology() const
{
     cout << "\nTopology:\n" << endl;
     cout << "\n!!!!!!!!!!!!!!!!!!!!!!!!!\n" << endl;
     for (std::list<CharmmParaResTop>::const_iterator i = Topology.begin(); i != Topology.end(); ++i) {
      (*i).print();
     }
}

void CharmmPara::print_NB() const
{
     cout << "\nNONBONDED:\n" << endl;
     cout << "atom type   eps   Rmin/2   eps,1-4   Rmin/2,1-4" << endl;
     for (std::list<CharmmParaNB>::const_iterator i = NB.begin(); i != NB.end(); ++i) {
       cout << (*i).attype << "   " << (*i).lje << "   " << (*i).ljr << "   " <<
               (*i).lje14 << "   " << (*i).ljr14 << endl;
     }
}

void CharmmPara::print_Bond() const
{
     cout << "\nBONDS:\n" << endl;
     cout << "atom type   atom type   Kb   b0" << endl;
     for (std::list<CharmmParaBond>::const_iterator i = Bond.begin(); i != Bond.end(); ++i) {
       cout << (*i).attype1 << "   " << (*i).attype2 << "   " <<
               (*i).k << "   " << (*i).r << endl;
     }
}

void CharmmPara::print_Angle() const
{
     cout << "\nAngles:\n" << endl;
     cout << "atom type   atom type   atom type   Ktheta   Theta0   Kub   S0" << endl;
     for (std::list<CharmmParaAngle>::const_iterator i = Angle.begin(); i != Angle.end(); ++i) {
       cout << (*i).attype1 << "   " << (*i).attype2 << "   " << (*i).attype3 << "   " <<
               (*i).kphi << "   " << (*i).phi << "   " << (*i).kub << "   " << (*i).rub << endl;
     }
}

void CharmmPara::print_Dihedral() const
{
     cout << "\nDihedrals:\n" << endl;
     cout << "atom type   atom type   atom type   atom type   Kchi   n   delta" << endl;
     for (std::list<CharmmParaDihedral>::const_iterator i = Dihedral.begin(); i != Dihedral.end(); ++i) {
       cout << (*i).attype1 << "   " << (*i).attype2 << "   " <<
               (*i).attype3 << "   " << (*i).attype4 << "   " <<
               (*i).k << "   " << (*i).n  << "   " << (*i).delta << endl;
     }
}

void CharmmPara::print_Improper() const
{
     cout << "\nIMPROPER:\n" << endl;
     cout << "atom type   atom type   atom type   atom type   Kpsi   psi0" << endl;
     for (std::list<CharmmParaImproper>::const_iterator i = Improper.begin(); i != Improper.end(); ++i) {
       cout << (*i).attype1 << "   " << (*i).attype2 << "   " <<
               (*i).attype3 << "   " << (*i).attype4 << "   " <<
               (*i).kpsi  << "   " << (*i).psi << endl;
     }
}


// CharmmPara.cc ends here
