/*  A set of CHARMM dihedral parameters -*- C++ -*-

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (C) 1990--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

*/
#ifndef _CharmmParaDihedral_h
#define _CharmmParaDihedral_h 1

#include <string>
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>

using std::string;
using std::ostream;
using std::istream;

//!wrap!
class CharmmParaDihedral {
public:
  string attype1;
  string attype2;
  string attype3;
  string attype4;
  float n;
  float delta;
  float k;
  CharmmParaDihedral ();
  CharmmParaDihedral (const CharmmParaDihedral &);
  ~CharmmParaDihedral() {}
  //!nowrap!
  CharmmParaDihedral& operator = (const CharmmParaDihedral& a);
  bool operator== (const CharmmParaDihedral&) const;
  bool operator!= (const CharmmParaDihedral&) const;
  bool operator> (const CharmmParaDihedral&) const;
  bool operator< (const CharmmParaDihedral&) const;
  //!nowrap!+
  void print() const;
  void print(ostream& os) const;
  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & attype1;
    ar & attype2;
    ar & attype3;
    ar & attype4;
    ar & n;
    ar & delta;
    ar & k;
  }
};

// torsions are equal when comparing i,j,k,l with either i,j,k,l or l,k,j,i
inline bool CharmmParaDihedral::operator== (const CharmmParaDihedral& b) const
{
 if ((attype1 == b.attype1 && attype2 == b.attype2 && attype3 == b.attype3 && attype4 == b.attype4) ||
     (attype1 == b.attype4 && attype2 == b.attype2 && attype3 == b.attype2 && attype4 == b.attype1)){
  return n == b.n && delta == b.delta && k == b.k;
 } else {return false;}
}

inline bool CharmmParaDihedral::operator!= (const CharmmParaDihedral& a) const
{
  return !(*this == a);
}

ostream& operator <<(ostream& os, const CharmmParaDihedral& a);

extern CharmmParaDihedral nil_CharmmParaDihedral;

inline bool CharmmParaDihedral::operator<(const CharmmParaDihedral& a) const
{
  // torsions are equal when comparing i,j,k,l with either i,j,k,l or l,k,j,i
 if ((a.attype1 == attype1 && a.attype2 == attype2 && a.attype3 == attype3 && a.attype4 == attype4) ||
     (a.attype4 == attype1 && a.attype3 == attype2 && a.attype2 == attype3 && a.attype1 == attype4)){
  if (a.n == n) {
   if (a.delta == delta) {
    return a.k < k;
   } else {return a.delta < delta;}
  } else {return a.n < n;}
 }
 // center atoms are equal when comparing i,j,k,l with i,j,k,l
 else if (a.attype2 == attype2 && a.attype3 == attype3){
  if      (a.attype1 < attype1 && a.attype4 < attype4){return true;}
  else if (a.attype1 == attype1){return a.attype4 < attype4;}
  else if (a.attype4 == attype4){return a.attype1 < attype1;}
  else if (a.attype1 < attype1){return true;}
  else {return false;}
 }
 // center atoms are equal when comparing i,j,k,l with l,k,j,i
 else if (a.attype2 == attype3 && a.attype3 == attype2){
  if      (a.attype4 < attype1 && a.attype1 < attype4){return true;}
  else if (a.attype4 == attype1){return a.attype1 < attype4;}
  else if (a.attype1 == attype4){return a.attype4 < attype1;}
  else if (a.attype1 < attype4){return true;}
  else {return false;}
 }
 // neither all nor only center atoms match
 // sort hierarchically, center atoms have higher priority
 else if (a.attype2 == attype2){return a.attype3 < attype3;}
 else if (a.attype3 == attype3){return a.attype2 < attype2;}
 else if (a.attype2 == attype3){return a.attype3 < attype2;}
 else if (a.attype3 == attype2){return a.attype2 < attype3;}
 else if ((a.attype2 < attype2 && a.attype3 < attype3) ||
          (a.attype2 < attype3 && a.attype3 < attype2)){return true;}
 else if ((a.attype2 > attype2 && a.attype3 > attype3) ||
          (a.attype2 > attype3 && a.attype3 > attype2)){return false;}
 // could not unambigously decide from center atoms -> try outer atoms
 else if (a.attype1 == attype1){
  if (a.attype4 == attype4){
   if (a.n == n) {
    if (a.delta == delta) {
     return a.k < k;
    } else {return a.delta < delta;}
   } else {return a.n < n;}
  }
  else if (a.attype4 < attype4){return true;}
  else if (a.attype4 > attype4){return false;}
 }
 else if (a.attype1 < attype1){return true;}
 else if (a.attype1 > attype1){return false;}
 return false;
}

inline bool CharmmParaDihedral::operator>(const CharmmParaDihedral& a) const
{
  // torsions are equal in when comparing i,j,k,l with either i,j,k,l or l,k,j,i
 if ((a.attype1 == attype1 && a.attype2 == attype2 && a.attype3 == attype3 && a.attype4 == attype4) ||
     (a.attype4 == attype1 && a.attype3 == attype2 && a.attype2 == attype3 && a.attype1 == attype4)){
  if (a.n == n) {
   if (a.delta == delta) {
    return a.k > k;
   } else {return a.delta > delta;}
  } else {return a.n > n;}
 }
 // center atoms are equal when comparing i,j,k,l with i,j,k,l
 else if (a.attype2 == attype2 && a.attype3 == attype3){
  if      (a.attype1 > attype1 && a.attype4 > attype4){return true;}
  else if (a.attype1 == attype1){return a.attype4 > attype4;}
  else if (a.attype4 == attype4){return a.attype1 > attype1;}
  else if (a.attype1 > attype1){return true;}
  else {return false;}
 }
 // center atoms are equal when comparing i,j,k,l with l,k,j,i
 else if (a.attype2 == attype3 && a.attype3 == attype2){
  if      (a.attype4 > attype1 && a.attype1 > attype4){return true;}
  else if (a.attype4 == attype1){return a.attype1 > attype4;}
  else if (a.attype1 == attype4){return a.attype4 > attype1;}
  else if (a.attype1 > attype4){return true;}
  else {return false;}
 }
 // neither all nor only center atoms match
 // sort hierarchically, center atoms have higher priority
 else if (a.attype2 == attype2){return a.attype3 > attype3;}
 else if (a.attype3 == attype3){return a.attype2 > attype2;}
 else if (a.attype2 == attype3){return a.attype3 > attype2;}
 else if (a.attype3 == attype2){return a.attype2 > attype3;}
 else if ((a.attype2 > attype2 && a.attype3 > attype3) ||
          (a.attype2 > attype3 && a.attype3 > attype2)){return true;}
 else if ((a.attype2 < attype2 && a.attype3 < attype3) ||
          (a.attype2 < attype3 && a.attype3 < attype2)){return false;}
 // could not unambigously decide from center atoms -> try outer atoms
 else if (a.attype1 == attype1){
  if (a.attype4 == attype4){
   if (a.n == n) {
    if (a.delta == delta) {
     return a.k > k;
    } else {return a.delta > delta;}
   } else {return a.n > n;}
  }
  else if (a.attype4 < attype4){return true;}
  else if (a.attype4 < attype4){return false;}
 }
 else if (a.attype1 > attype1){return true;}
 else if (a.attype1 > attype1){return false;}
 return false;
}

#if SWIGPP_LITERAL_INCLUDE

%addmethods CharmmParaDihedral {
   void write() { self->print(cout); }
   bool operator<(const CharmmParaDihedral& a, const CharmmParaDihedral& b);
   bool operator>(const CharmmParaDihedral& a, const CharmmParaDihedral& b);
};

#endif // SWIGPP_LITERAL_INCLUDE

#endif

// CharmmParaDihedral.h ends here
