/*  Topology of a Charmm Residue

    bonds, impropers are predefined as read from CHARMM topology file
    angles and dihedrals are generated from bonds

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _CharmmParaResTop_h
#define _CharmmParaResTop_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include <cctype>
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/GcemFormDef.h"
#include "MEAD_e/globals.h"
#include <sstream>
#include <boost/regex.hpp>
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/vector.hpp>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;
using std::stringstream;

using namespace std;
using namespace boost;

//!wrap!
class CharmmParaResTop {
public:
  CharmmParaResTop();
  CharmmParaResTop (const string&, const list<string>, const map<pair<string, string>, int>, const list<vector<string> >);
  CharmmParaResTop (const string&);
  CharmmParaResTop (const CharmmParaResTop&);
  ~CharmmParaResTop () {}
  //!nowrap!+
  CharmmParaResTop& operator=(const CharmmParaResTop& a);

  //!nowrap!-

  // generate topology of residue from bonding
  // fills bonded list
  // fills angles and dihedrals, but not impropers!
  void generate_top();
  void generate_bonded_list();
  void generate_angles();
  void generate_dihedrals();
  void generate_sidechain_tor();
  void check_bonds() const;
  void check_impropers() const;

  // print lists
  void print() const;
  void print_atoms() const;
  void print_bonds() const;
  void print_bonded_list() const;
  void print_angles() const;
  void print_dihedrals() const;
  void print_impropers() const;
  void print_sidechain_tor() const;

  string get_name() const {return name;}

  std::list<std::string> atoms;
  std::map<std::pair<std::string,string>,int> bonds;
  std::list<std::vector<std::string> > angles;
  std::list<std::vector<std::string> > dihedrals;
  std::list<std::vector<std::string> > impropers;

  std::map<string,list<string> > bonded_list;
  std::vector<pair<vector<string>,list<string> > > sidechain_tor;

  std::string get_element(const string&) const;
  float get_mass(const string&) const;

  /* check if element is already listed */
  bool contains_atom (const string&) const;
  bool contains_bond  (const string&, const string&) const;
  bool contains_bonded_entry  (const string&) const;
  bool contains_angle (const string&, const string&, const string&) const;
  bool contains_dihedral (const string&, const string&, const string&, const string&) const;
  bool contains_improper (const string&, const string&, const string&, const string&) const;
  bool contains_sidechain_tor (const string&, const string&, const string&, const string&) const;
  bool in_12 (const string&, const string&) const;
  bool in_13 (const string&, const string&) const;
  bool in_14 (const string&, const string&) const;

  // retrieve bond order; 0 means no bond exists 1,2,3 meant literally, 4 means aromatic
  // (FIXME: is AROMATIC ever actually used in CHARMM?)

  int get_bond_order(const string&, const string&) const;
  std::list<string> get_bond_partners(const string&) const;
  std::pair<list<string>, bool> get_moving_ats(const string&, const string&, const string&, const string&) const;
  std::pair<list<string>, bool> get_moving_ats(const string&, const string&, const string&, const string&,
                                               const list<string>&) const;
  bool in_list (const string&, const list<string>&) const;
  vector<string> get_hp_tor (const list<vector<string> >&) const;
  bool is_hp_at (const string&, const string&) const;
  bool moved_by_sctor(const string&) const;

  /* insert in list */
  std::list<string>::iterator insert_atom (const string&);
  std::pair<std::map<pair<string,string>,int>::iterator, bool> insert_bond(const string&, const string&, const int);
  std::list<vector<string> >::iterator insert_angle(const string&, const string&, const string&);
  std::list<vector<string> >::iterator insert_dihedral(const string&, const string&, const string&, const string&);
  std::list<vector<string> >::iterator insert_improper(const string&, const string&, const string&, const string&);

  std::list<vector<string> >::iterator insert_angle(const vector<string>&);
  std::list<vector<string> >::iterator insert_dihedral(const vector<string>&);
  std::list<vector<string> >::iterator insert_improper(const vector<string>&);

  bool operator== (const CharmmParaResTop&) const;
  bool operator!= (const CharmmParaResTop&) const;
  bool operator< (const CharmmParaResTop&) const;
  bool operator> (const CharmmParaResTop&) const;

  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & atoms;
    ar & bonds;
    ar & angles;
    ar & dihedrals;
    ar & impropers;
    ar & bonded_list;
    ar & sidechain_tor;
    ar & name;
  }
  string name;
};

inline CharmmParaResTop& CharmmParaResTop::operator= (const CharmmParaResTop& a)
{
 name = a.name; atoms = a.atoms; bonds = a.bonds; angles = a.angles; dihedrals = a.dihedrals;
 impropers = a.impropers; bonded_list = a.bonded_list; sidechain_tor =a.sidechain_tor;
 return *this;
}

inline bool CharmmParaResTop::operator== (const CharmmParaResTop& a) const
{
 return name != a.name;
}

inline bool CharmmParaResTop::operator!= (const CharmmParaResTop& a) const
{
 return name == a.name;
}

inline bool CharmmParaResTop::operator> (const CharmmParaResTop& a) const
{
 return name > a.name;
}

inline bool CharmmParaResTop::operator< (const CharmmParaResTop& a) const
{
 return name < a.name;
}

inline vector<string> CharmmParaResTop::get_hp_tor (const list<vector<string> >& a) const
{ // get torsion with highest priority on the list
  if (a.size() == 0) {
   return vector<string>();
  } else if (a.begin()->size() == 0) {
   return vector<string>();
  }
  vector<string> result(*a.begin());
  string tmpstr = (*a.begin())[3];
  for (list<vector<string> >::const_iterator j = a.begin(); j != a.end(); j++){
   if (is_hp_at((*j)[3],tmpstr)){
    tmpstr = (*j)[3];
    result = *j;
   }
  }
  return result;
}

inline bool CharmmParaResTop::is_hp_at (const string& a, const string& b) const
{
  bool result   = false;
  bool matched1 = true;
  bool matched2 = true;
  string at1, at2;
  int n1 = 0; int n2 = 0;

  boost::regex rex("^\\s*\\d*([^\\d^\\s]+)(\\d*)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  if (boost::regex_search(a.begin(), a.end(), match, rex, flags)){
     at1 = std::string(match[1].first, match[1].second);
     string tmpstr1 = std::string(match[2].first, match[2].second);
     stringstream ss1(tmpstr1);
     ss1 >> n1;
     if (ss1.fail()){n1 = 0;}
  }
  else {// only numbers found
   matched1 = false;
  }
  if (boost::regex_search(b.begin(), b.end(), match, rex, flags)){
     at2 = std::string(match[1].first, match[1].second);
     string tmpstr1 = std::string(match[2].first, match[2].second);
     stringstream ss1(tmpstr1);
     ss1 >> n2;
     if (ss1.fail()){n2 = 0;}
  }
  else {// only numbers found
   matched2 = false;
  }
  //both name strings contain only numbers
  if (!matched1 && !matched2){result = (a > b); cout << "one name strings contains only numbers, probably H" << endl;}
  //one name strings contains only numbers, probably H
  else if (matched1 && !matched2){result = true;}
  else if (matched2 && !matched1){result = false;}
  else {// both matched
   float m1 = get_mass(get_element(a));
   float m2 = get_mass(get_element(b));
   if      (m1 > m2){result = true;}
   else if (m1 < m2){result = false;}
   else {// same element, compare atname strings
    if      (at1 > at2){result = true;}
    else if (at1 < at2){result = false;}
    //same element and name, compare numbers
    else{
     if      (n1 < n2){result = true;}
     else if (n1 > n2){result = false;}
     else {result = false;}
    }
   }
  }
  return result;
}

inline std::pair<list<string>, bool> CharmmParaResTop::get_moving_ats(const string& a1, const string& a2,
                                                                      const string& a3, const string& a4) const
{
 return get_moving_ats(a1, a2, a3, a4, atoms);
}

inline std::pair<list<string>, bool> CharmmParaResTop::get_moving_ats(const string& a1, const string& a2,
                                                                      const string& a3, const string& a4,
                                                                      const list<string>& alist0) const
{// get atoms which move then changing the sidechain torsion and check if the torsion is within a cycle
  list<string> alist;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   if      (i->first.first == a3 && i->first.second != a2){
    alist.push_back(i->first.second);
   }
   else if ((i->first).second == a3 && (i->first).first != a2){
    alist.push_back(i->first.first);
   }
  }
  // gather all atoms bonded to the ones bonded to a3
  bool found = true;
  int i = 1;
  while(found){
   i++;
   found = false;
   for (list<string>::const_iterator j = alist.begin(); j != alist.end(); j++){
    for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
     if      (i->first.first == *j && i->first.second != a3 &&
              !in_list(i->first.second, alist) && in_list(i->first.second, alist0)){
      alist.push_back(i->first.second);
      found = true;
     }
     else if (i->first.second == *j && i->first.first != a3 &&
              !in_list(i->first.first, alist) && in_list(i->first.first, alist0)){
      alist.push_back(i->first.first);
      found = true;
     }
    }
   }
  }
  bool incycle = false;
  if (alist.size() == (alist0.size() - 1)){incycle = true;}
  return(pair<list<string>,bool>(alist,incycle));
}

inline bool CharmmParaResTop::in_list(const string& s, const list<string>& alist) const
{
   bool found = false;
   for (list<string>::const_iterator j = alist.begin(); j != alist.end(); j++){
    if (*j == s){
     found = true;
    }
   }
   return found;
}

inline float CharmmParaResTop::get_mass(const string& s) const
{
  float mass;

  if      (s == "C")  {mass = float(12.000000000);}
  else if (s == "O")  {mass = float(15.99491463);}
  else if (s == "H")  {mass = float(1.007825035);}
  else if (s == "N")  {mass = float(14.003074002);}
  else if (s == "S")  {mass = float(31.97207070);}
  else if (s == "P")  {mass = float(30.9737620);}
  else if (s == "AU") {mass = float(196.966552);}
  else if (s == "OS") {mass = float(191.961467);}
  else if (s == "RE") {mass = float(186.955744);}
  else if (s == "W")  {mass = float(185.954357);}
  else if (s == "BA") {mass = float(137.3277);}
  else if (s == "AG") {mass = float(107.86822);}
  else if (s == "RU") {mass = float(101.9043485);}
  else if (s == "MO") {mass = float(97.9054073);}
  else if (s == "ZN") {mass = float(63.9291448);}
  else if (s == "CU") {mass = float(62.9295989);}
  else if (s == "NI") {mass = float(57.9353462);}
  else if (s == "FE") {mass = float(55.9349393);}
  else if (s == "MN") {mass = float(54.9380471);}
  else if (s == "CR") {mass = float(51.9405098);}
  else if (s == "V")  {mass = float(50.94151);}
  else if (s == "CA") {mass = float(39.9625906);}
  else if (s == "MG") {mass = float(23.9850423);}
  else if (s == "NA") {mass = float(22.9897677);}
  else if (s == "BR") {mass = float(78.9183361);}
  else if (s == "SE") {mass = float(79.9165196);}
  else if (s == "K")  {mass = float(39.09831);}
  else if (s == "CL") {mass = float(34.968852721);}
  else if (s == "F")  {mass = float(18.99840322);}
  else {
   cerr << "CharmmParaResTop::get_mass: Warning unknown element" << s
        << ", assuming hydrogen." << endl;
   mass = float(1.007825035);
  }

  return mass;
}

inline string CharmmParaResTop::get_element(const string& s) const
{
  // not very rigorous, includes biologically important elements
  // otherwise defaults to H

  string num("0123456789");
  string tmpstr = s.substr(s.find_first_not_of(num),2);
  // clean c++ to upper from <cctype> by Danny Kalev @ Dev.Xcom
  // (int(*)(int)) toupper because there is one more std::toupper from <locale>
  // begin, end, begin of result, operation
  //std::transform(s.begin(), s.end(), s.begin(), (int(*)(int)) toupper);
  std::transform(tmpstr.begin(), tmpstr.end(), tmpstr.begin(), (int(*)(int)) toupper);
  string element;
  if        (tmpstr == "BR"){
   element = "BR";
  } else if (tmpstr == "SE"){
   element = "SE";
  } else if (tmpstr == "CL"){
   element = "CL";
  } else if (tmpstr == "CR"){
   element = "CR";
  } else if (tmpstr == "NI"){
   element = "NI";
  } else if (tmpstr == "ZN"){
   element = "ZN";
  } else if (tmpstr == "RU"){
   element = "RU";
  } else if (tmpstr == "RE"){
   element = "RE";
  } else if (tmpstr == "MO"){
   element = "MO";
  } else if (tmpstr == "MN"){
   element = "MN";
  } else if (tmpstr == "MG"){
   element = "MG";
  } else if (tmpstr == "CU"){
   element = "CU";
  } else if (tmpstr == "FE"){
   element = "FE";
  } else if (tmpstr == "CO"){
   element = "CO";
  } else if (tmpstr == "BA"){
   element = "BA";
  } else if (tmpstr == "AU"){
   element = "AU";
  } else if (tmpstr == "AG"){
   element = "AG";
  } else {tmpstr = tmpstr[0];}

  if (tmpstr == "H" || tmpstr == "D" || tmpstr == "T"){
    element = "H";
  } else if (tmpstr == "C"){
    element = "C";
  } else if (tmpstr == "O"){
    element = "O";
  } else if (tmpstr == "N"){
    element = "N";
  } else if (tmpstr == "S"){
    element = "S";
  } else if (tmpstr == "P"){
    element = "P";
  } else if (tmpstr == "W"){
    element = "W";
  } else if (tmpstr == "V"){
    element = "V";
  } else if (tmpstr == "K"){
    element = "K";
  } else {element = "H";}

  return element;
}


// list functions

inline list<string> CharmmParaResTop::get_bond_partners(const string& a) const
{
  list<string> result;
  bool found = false;
  for (std::map<string,list<string> >::const_iterator i = bonded_list.begin(); i != bonded_list.end(); ++i){
   if (i->first == a){
    result = i->second;
    found = true;
    break;
   }
  }
  if (!found)
  {
   ::error("CharmmParaResTop::get_bond_partners: atom not found");
  }
  return result;
}

inline int CharmmParaResTop::get_bond_order(const string& a, const string& b) const
{
  bool found = false;
  int order;
  for (std::map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   if (((i->first).first == a && (i->first).first == b) ||
       ((i->first).first == b && (i->first).first == a)){
    order = i->second;
    found = true;
    break;
   }
  }
  if (!found){
    if (!contains_atom(a) || !contains_atom(b)){
     cerr << "residue name " << name << "atoms " << a << " " << b << endl;
     ::error("CharmmParaResTop::get_bond_order: atom in bond not found");
    }
    else {
     order = 0;
    }
  }
  return order;
}

// Existence of element (not STL-like)

inline bool CharmmParaResTop::contains_atom(const string& k) const
{
  bool found = false;
  for (std::list<string>::const_iterator i = atoms.begin(); i != atoms.end(); ++i){
   if ((*i) == k){
     found = true;
     break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_bonded_entry(const string& k) const
{
  bool found = false;
  for (map<string,list<string> >::const_iterator i = bonded_list.begin(); i != bonded_list.end(); ++i){
   if (i->first == k){
    found = true;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_bond(const string& a, const string& b) const
{
  bool found = false;
  for (map<pair<string,string>,int>::const_iterator i = bonds.begin(); i != bonds.end(); ++i){
   if (((i->first).first == a && (i->first).first == b) ||
       ((i->first).first == b && (i->first).first == a)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_angle(const string& a, const string& b, const string& c) const
{
  bool found = false;
  for (std::list<vector<string> >::const_iterator i = angles.begin(); i != angles.end(); ++i){
   if (((*i)[0] == a && (*i)[1] == b && (*i)[2] == c) ||
       ((*i)[0] == c && (*i)[1] == b && (*i)[2] == a)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_dihedral(const string& a, const string& b, const string& c, const string& d) const
{
  bool found = false;
  for (std::list<vector<string> >::const_iterator i = dihedrals.begin(); i != dihedrals.end(); ++i){
   if (((*i)[0] == a && (*i)[1] == b && (*i)[2] == c && (*i)[3] == d) ||
       ((*i)[0] == d && (*i)[1] == c && (*i)[2] == b && (*i)[3] == a)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_improper(const string& a, const string& b, const string& c, const string& d) const
{
  bool found = false;
  for (std::list<vector<string> >::const_iterator i = impropers.begin(); i != impropers.end(); ++i){
   if (((*i)[0] == a && (*i)[1] == b && (*i)[2] == c && (*i)[3] == d) ||
       ((*i)[0] == a && (*i)[1] == d && (*i)[2] == c && (*i)[3] == b) ||
       ((*i)[0] == b && (*i)[1] == a && (*i)[2] == c && (*i)[3] == d) ||
       ((*i)[0] == b && (*i)[1] == d && (*i)[2] == c && (*i)[3] == a) ||
       ((*i)[0] == d && (*i)[1] == a && (*i)[2] == c && (*i)[3] == b) ||
       ((*i)[0] == d && (*i)[1] == b && (*i)[2] == c && (*i)[3] == a)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::contains_sidechain_tor(const string& a, const string& b, const string& c, const string& d) const
{
  bool found = false;
  for (std::vector<pair<vector<string>,list<string> > >::const_iterator i=sidechain_tor.begin();i!=sidechain_tor.end();++i){
   if (((*i).first[0] == a && (*i).first[1] == b && (*i).first[2] == c && (*i).first[3] == d) ||
       ((*i).first[0] == d && (*i).first[1] == c && (*i).first[2] == b && (*i).first[3] == a)){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::moved_by_sctor(const string& a) const
{
  bool found = false;
  for (std::vector<pair<vector<string>,list<string> > >::const_iterator i=sidechain_tor.begin();i!=sidechain_tor.end();++i){
   for (std::list<string>::const_iterator j=(*i).second.begin();j!=(*i).second.end();++j){
    if (*j == a){
     found = true;
     goto END;
    }
   }
  }
  END:
  return found;
}

// insert functions

inline std::list<string>::iterator CharmmParaResTop::insert_atom (const string& a){
  if (contains_atom(a)) {
    cerr << name << " : " << a << endl;
    ::error("CharmmParaResTop::insert_atom: attempt to insert a duplicate atom\n");
  }
  atoms.push_back(a);
  return atoms.end();
}

inline pair<map<pair<string,string>, int>::iterator,bool> CharmmParaResTop::insert_bond (const string& a, const string& b,
                                                                                         const int o) {
  if (contains_bond(a, b)) {
    cerr << name << " : " << a << " " << b << endl;
    ::error("CharmmParaResTop::insert_bond: attempt to insert a duplicate bond\n");
  }
  pair<string,string> c(a, b);
  return bonds.insert(make_pair(c, o));
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_angle (const string& a, const string& b, const string& c){
  if (contains_angle(a, b, c)) {
    cerr << name << " : " << a << " " << b << " " << c << endl;
    ::error("CharmmParaResTop::insert_angle: attempt to insert a duplicate angle\n");
  }
  vector<string> tmpvec;
  tmpvec.push_back(a);
  tmpvec.push_back(b);
  tmpvec.push_back(c);
  angles.push_back(tmpvec);
  return angles.end();
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_angle (const vector<string>& a){
  if (a.size() != 3) {
    ::error("CharmmParaResTop::insert_angle: incorrect input vector size.\n");
  }
  else if (contains_angle(a[0], a[1], a[3])) {
    cerr << name << " : " << a[0] << " " << a[1] << " " << a[3] << endl;
    ::error("CharmmParaResTop::insert_angle: attempt to insert a duplicate angle\n");
  }
  angles.push_back(a);
  return angles.end();
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_dihedral (const string& a, const string& b,
                                                                               const string& c, const string& d){
  if (contains_dihedral(a, b, c, d)) {
    cerr << name << " : " << a << " " << b << " " << c << endl;
    ::error("CharmmParaResTop::insert_dihedral: attempt to insert a duplicate dihedral\n");
  }
  vector<string> tmpvec;
  tmpvec.push_back(a);
  tmpvec.push_back(b);
  tmpvec.push_back(c);
  tmpvec.push_back(d);
  dihedrals.push_back(tmpvec);
  return dihedrals.end();
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_dihedral (const vector<string>& a){
  if (a.size() != 4) {
    ::error("CharmmParaResTop::insert_diheral: incorrect input vector size.\n");
  }
  else if (contains_dihedral(a[0], a[1], a[3], a[4])) {
    cerr << name << " : " << a[0] << " " << a[1] << " " << a[3] << " " << a[4] << endl;
    ::error("CharmmParaResTop::insert_dihedral: attempt to insert a duplicate dihedral\n");
  }
  dihedrals.push_back(a);
  return dihedrals.end();
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_improper (const string& a, const string& b,
                                                                               const string& c, const string& d){
  if (contains_improper(a, b, c, d)) {
    cerr << name << " : " << a << " " << b << " " << c << endl;
    ::error("CharmmParaResTop::insert_improper: attempt to insert a duplicate angle\n");
  }
  vector<string> tmpvec;
  tmpvec.push_back(a);
  tmpvec.push_back(b);
  tmpvec.push_back(c);
  tmpvec.push_back(d);
  impropers.push_back(tmpvec);
  return impropers.end();
}

inline std::list<vector<string> >::iterator CharmmParaResTop::insert_improper (const vector<string>& a){
  if (a.size() != 4) {
    ::error("CharmmParaResTop::insert_improper: incorrect input vector size.\n");
  }
  else if (contains_improper(a[0], a[1], a[3], a[4])) {
    cerr << name << " : " << a[0] << " " << a[1] << " " << a[3] << " " << a[4] << endl;
    ::error("CharmmParaResTop::insert_improper: attempt to insert a duplicate improper\n");
  }
  impropers.push_back(a);
  return impropers.end();
}

inline bool CharmmParaResTop::in_12(const string& a, const string& b) const
{
  bool found = false;
  for (map<string,list<string> >::const_iterator i = bonded_list.begin(); i != bonded_list.end(); ++i){
   if (i->first == a){
    for (list<string>::const_iterator j = (i->second).begin(); j != (i->second).end(); ++j){
     if (*j == b){
      found = true;
      goto END;
     }
    }
   }
   if (i->first == b){
    for (list<string>::const_iterator j = (i->second).begin(); j != (i->second).end(); ++j){
     if (*j == a){
      found = true;
      goto END;
     }
    }
   }
  }
  END: return found;
}

inline bool CharmmParaResTop::in_13(const string& a, const string& b) const
{
  bool found = false;
  for (list<vector<string> >::const_iterator i = angles.begin(); i != angles.end(); ++i){
   if ((*i)[0] == a && (*i)[2] == b){
    found = true;
    break;
   }
   if ((*i)[0] == b && (*i)[2] == a){
    found = true;
    break;
   }
  }
  return found;
}

inline bool CharmmParaResTop::in_14(const string& a, const string& b) const
{
  bool found = false;
  for (list<vector<string> >::const_iterator i = dihedrals.begin(); i != dihedrals.end(); ++i){
   if ((*i)[0] == a && (*i)[3] == b){
    found = true;
    break;
   }
   if ((*i)[0] == b && (*i)[3] == a){
    found = true;
    break;
   }
  }
  return found;
}


#endif

// CharmmParaResTop.h ends here
