/* -*- C++ -*-
Dielectric environment due to a membrane plus some Atoms,
eg. a membrane protein.

    Copyright (c) 1995 by Donald Bashford

    This source code file is part of the MEAD_e (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD_e is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: DielMembAtoms.h,v 2.7 2001/06/26 18:32:21 bergsma Exp $
*/

#ifndef DielMembAtoms_h
#define DielMembAtoms_h

#include "MEAD_e/DielectricSlab.h"
#include "MEAD_e/SolvAccVol.h"
#include "MEAD_e/AtomSet.h"

//!wrap!
class TwoValueDielMembAtoms : public DielectricEnvironment_lett {
public:
  TwoValueDielMembAtoms(const AtomSet& ats, float epsin, float zl, float zu, Coord c, float holrad);
  TwoValueDielMembAtoms(const SolvAccVol& sav, float epsin, float zl, float zu, Coord c, float holrad);
  ~TwoValueDielMembAtoms() {if (membranept != NULL) delete membranept;}
  //!nowrap!+
  DielCubeRep get_cuberep(const CubeLatSpec& cls);
  DielectricSlab* get_slab() {return membranept;}  // FIXME?  Dangerous?
  //!nowrap!-
  float epsext_value() const {return epsext;}
private:
  float epsin, epsext;
  float epsmemb;
  DielectricEnvironment membrane;
  DielectricSlab* membranept;
  Coord hole_center;
  float hole_radius;
// FIXME!  If this were a ref or ptr, not need include SolvAccVol.h
  SolvAccVol solv_acc;
};

// Given membrane boundaries and two AtomSets, region within atset1
// has ein1, region within atset2 (but outside of atset1) has ein2
// region outside of both atsets, but inside the membrane boundaries
// also has ein2, region outside of both atsets and outside membran
// boundaries gets epsext.
//!wrap!
class ThreeValueDielMembAtomsAtoms : public DielectricEnvironment_lett {
public:
  ThreeValueDielMembAtomsAtoms(const AtomSet& a1, float solrad1, const AtomSet& a2, float solrad2, float zl, float zu, Coord c, float holrad, float ein1, float ein2, float epsext);
  ~ThreeValueDielMembAtomsAtoms() {if (membranept != NULL) delete membranept;}
  //!nowrap!+
  DielCubeRep get_cuberep(const CubeLatSpec& cls);
  DielectricSlab* get_slab() {return membranept;}  // FIXME?  Dangerous?
  //!nowrap!-
  float epsext_value() const {return epsext;}
private:
  DielectricEnvironment membrane;
  DielectricSlab* membranept;
  Coord hole_center;
  float hole_radius;
  float epsin1, epsin2, epsext;
  SolvAccVol solv_acc1, solv_acc2;
};

/* TRU, 2006
SixValueDielMembThreeAtoms derived from ThreeValueDielMembAtomsAtoms
protein (QM sites, normal sites, cavities) (epsin1, epsin2, epsin3)
 embedded in 3-slab membrane model (headgroup regions and core region, epshead/epscore)
  embedded in solvent (epsext)

first a 3-slab 2-dielectric membrane is constructed and
then a 3-regions 3-dielectric protein from 3 atomsets
the protein dielectric regions override solvent and membrane regions

protein dielectric is constructed as follows:

 atomset 1 contains all atoms of protein (QM and normal sites) and cavities and is
  initially assigned epsin1 corresponding to QM dielectric
 atomset 2 contains normal sites and cavities,
  overrides space occupied by both atomset1 and atomset2 and
  is initially assigned epsin2 corresponding to normal protein dielectric
 atomset 3 contains cavities,
  overrides space occupied by the (atomset1 || atomset2) and atomset3
  and is assigned epsin3 corresponding to cavity dielectric

space within the membrane boundaries and inside the cylinder radius of hole_center
 is assigned epsext

this procedure ensures that:

 the two solvent regions are safely separated by the membrane,
 the protein is embedded in the membrane without artifactual
  solvent dielectric at the membrane-protein interface,
 there are no unwanted solvent artefacts in the protein regions
 that the QM region contains also the space between atoms of QM and normal sites and
 that the normal region contains also the space between cavities and normal region

the cylinder is retained to offer a simpler alternative to providing a file
 with dummy atoms representing protein cavities and probably a transmembrane channel
 the cavity region is not overwritten by the cylinder volume such that one can have
 a channel filled with normal solvent assigned epsext and the cavities assigned epsin3 if desired

begin added
*/

//!wrap!
class SixValueDielMembThreeAtoms : public DielectricEnvironment_lett {
public:
  SixValueDielMembThreeAtoms(const AtomSet& a1, float solrad1, const AtomSet& a2, float solrad2, const AtomSet& a3, float solrad3, float zlh, float zuh, float zlc, float zuc, Coord c, float holrad, float ehead, float ecore, float ein1, float ein2, float ein3, float epsext);
  SixValueDielMembThreeAtoms(const SolvAccVol& a1, const SolvAccVol& a2, const SolvAccVol& a3, float zlh, float zuh, float zlc, float zuc, Coord c, float holrad, float ehead, float ecore, float ein1, float ein2, float ein3, float epsext);
  ~SixValueDielMembThreeAtoms() {if (membranept != NULL) delete membranept;}
  //!nowrap!+
  DielCubeRep get_cuberep(const CubeLatSpec& cls);
  DielectricSlab* get_slab() {return membranept;}  // FIXME?  Dangerous?
  //!nowrap!-
  float epshead_value() const {return epshead;}
  float epscore_value() const {return epscore;}
  float epsin1_value() const {return epsin1;}
  float epsin2_value() const {return epsin2;}
  float epsin3_value() const {return epsin3;}
  float epsext_value() const {return epsext;}
  private:
  DielectricEnvironment membrane;
  DielectricSlab* membranept;
  Coord hole_center;
  float zlowerhead, zupperhead, zlowercore, zuppercore, hole_radius;
  float epshead, epscore, epsin1, epsin2, epsin3, epsext;
  SolvAccVol solv_acc1, solv_acc2, solv_acc3;
};

/* end added */
#endif

// DielMembAtoms.h ends here
