/* Base class for dielectric environment and subclasses.

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: DielectricEnvironment.cc,v 2.7 2004/12/06 17:58:12 bashford Exp $
*/
#include "MEAD_e/DielectricEnvironment.h"
#include "MEAD_e/UniformDielectric.h"
#include <iostream>


int DielectricEnvironment::instanceCount = 0;

// Constructs an envelope with a ChargDist letter, that is, a default letter.
DielectricEnvironment::DielectricEnvironment()
{
  rep = new UniformDielectric(1.0);
}

// Ctor must have a valid envelope (non-zero rep) as its argument.
DielectricEnvironment::DielectricEnvironment(const DielectricEnvironment & c)
{
  if(c.rep) { // Arg is an envelope.
    c.rep->referenceCount++;
    rep = c.rep;
  }
  else
    error("ERROR: DielectricEnvironment copy constructor:\n",
	  "argument has zero rep pointer");
  ++instanceCount;
}

// Both both sides of assignment must be envelopes (non-zero rep)

DielectricEnvironment& DielectricEnvironment::operator=
(const DielectricEnvironment& c)
{
  if (this != &c) {
    if (rep) {
      if (c.rep) {
        c.rep->referenceCount++;
        if (rep->referenceCount == 1)
          delete rep;
        else
          --rep->referenceCount;
        rep = c.rep;
      }
      else {
        error("ERROR: DielectricEnvironment::operator=:\n",
	      "argument has zero rep");
      }
    }
    else {
      error("ERROR DielectricEnvironment::operator=: this->rep is zero");
    }
  }
  return *this;
}

// TE
// Write DielectricEnvironment (dielectric constant) to OpenDX file
void DielectricEnvironment::write_dx(const CubeLatSpec& cls, string file)
{
 blab3 << "DielectricEnvironment::write_dx entered" << endl;
  //  rep->write_dx(cls, name);
    // run get_cuberep to get dielectric constants mapped on grid
  DielCubeRep dcr(get_cuberep(cls));
  // this is necessary to circumvent error in ElyCubeRep.h
  //   "ERROR: illegal application of operator[] to a DielCubeRep",
  //      "with eps_array defined"
  // FixMe: find out what the check is good for
  // seems to make no sense at the moment to
  // request values without eps_array defined
  dcr.declare_undefined();
  int grid_dim = cls.get_grid_dim();
  int ncube = grid_dim*grid_dim*grid_dim;
  Coord center = cls.get_center();
  float space = cls.get_spacing();
  // OpenDX expects lower corner as origin, not center
  Coord origin(center.x-((grid_dim * space)/2),
               center.y-((grid_dim * space)/2),
               center.z-((grid_dim * space)/2));
  string file_name = file + ".dx";
  ofstream of(file_name.c_str());
  if (!of.good())
    ::error("DielectricEnvironment:write_dx failed to open for writing, ",
            file_name.c_str());
  blab1 << "DielectricEnvironment::write_dx: Writing " << file_name.c_str() << " in OpenDX format...\n ";
  of << "# "<< file << " " << grid_dim << " " << space << " " << center << endl;
  of << "#\n# DielectricEnvironment witten by MEAD\n#" << endl;
  of << "object 1 class gridpositions counts " << grid_dim << " " << grid_dim << " " << grid_dim << endl;
  of << "origin " << origin.x << " " << origin.y << " " << origin.z << "\ndelta " << space << " 0.0 0.0\ndelta 0.0 " << space << " 0.0\ndelta 0.0 0.0 " << space << endl;
  of << "object 2 class gridconnections counts " << grid_dim << " " << grid_dim << " " << grid_dim << endl;
  of << "object 3 class array type double rank 0 times " << ncube << endl;
  int count = 0;
  // Now the actual values
  // Don't know if I could write more than three columns according to OpenDX format,
  // but the TCL reader can only read three columns or less...
  for (int h=2; h<ncube; h+=3) {
    of << dcr[h-2] << " " << dcr[h-1] << " " << dcr[h] << endl;
    count++;
  }
  if (ncube - (3 * count) == 2)
    of << dcr[ncube-2] << " " << dcr[ncube-1] << endl;
  if (ncube - (3 * count) == 1)
    of << dcr[ncube-1] << endl;
  // Now some blurb at the end
  of << "attribute \"dep\" string \"positions\"" << endl;
  of << "object \"regular positions regular connections\" class field" << endl;
  of << "component \"positions\" value 1" << endl;
  of << "component \"connections\" value 2" << endl;
  of << "component \"data\" value 3" << endl;
  if (! of.good())
    ::error("DielectricEnvironment:write_dx failed to write OpenDX file");
  of.close();
  blab3 << "exit from DielectricEnvironment::write_dx with " << endl;
}
//TE


// This will create an envelope containing *cdp.
DielectricEnvironment::DielectricEnvironment(DielectricEnvironment_lett * cdp)
{
  cdp->referenceCount++;
  rep = cdp;
  instanceCount++;
}

DielectricEnvironment::~DielectricEnvironment()
{
  if (rep) {
    blab2 << "DielectricEnvironment destructor called with rep->referenceCount = " << rep->referenceCount << endl;
    if (rep->referenceCount == 1)
      delete rep;
    else
      --rep->referenceCount;
  }
  instanceCount--;
}

int DielectricEnvironment_lett::instanceCount = 0;


DielectricEnvironment_lett::DielectricEnvironment_lett()
{
  blab2 << "DielectricEnvironment_lett default constructor called" << endl;
  referenceCount=1;
  ++instanceCount;
}


DielectricEnvironment_lett::DielectricEnvironment_lett
(const DielectricEnvironment_lett& a)
{
  blab2 << "DielectricEnvironment_lett copy constructor called" << endl;
  referenceCount = 1;
  ++instanceCount;
}

// Don't delete this letter until all references have been released
void DielectricEnvironment_lett::operator delete(void *p, size_t s)
{
  DielectricEnvironment_lett *del = reinterpret_cast<DielectricEnvironment_lett*> (p);
  blab2 << "DielectricEnvironment_lett operator delete called with referenceCount = " << del->referenceCount << endl;
  if (del->referenceCount == 0) {
    blab2 << "DielectricEnvironment_lett operator delete frees object" << endl;
    ::operator delete(p);
  }
}

DielectricEnvironment_lett::~DielectricEnvironment_lett()
{
  blab2 << "DielectricEnvironment_lett destructor called with referenceCount = " << referenceCount << endl;
  if (--referenceCount)
    // It's now OK for this to happen
    blab2 << "WARNING: DielectricEnvironment_lett destructor called\n"
      << "on object with referenceCount = " << referenceCount+1
	<< "\nas if the object is still in use somewhere" << endl;
  --instanceCount;
}

DielectricEnvironment_lett&
DielectricEnvironment_lett::operator=(const DielectricEnvironment_lett& acs)
{return *this;}

// DielectricEnvironment.cc ends here
