/*  
    helper object for GcemMolecule
    calculates MM energies using the CHARMM force field and an
    empirical energy term from propensities in a rotamer library

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemCharmmEnergyCalculator.h"
#include "MEAD_e/globals.h"
#include <stdio.h>
#include <fstream>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <sstream>
#include <iostream>
#include <string>

using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::stringstream;
using std::string;

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

void GcemCharmmEnergyCalculator::calc_mmint_propensities(GcemMolecule &molecule) const
{

  blab2 << "Entering GcemCharmmEnergyCalculator::calc_mmint_propensities" << endl;
  
  const double one_over_beta = PhysCond::get_Boltzmann_constant() * PhysCond::get_T()
                              * PhysCond::get_Avogadro_Number() / 4184;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   int stype(molecule.get_rotmap_value(i->get_sid()));
   if (stype != 1 && stype != 2 && molecule.get_skip_stat_mmint()){
    blab1 << "GcemCharmmEnergyCalculator::calc_mmint_propensities: site with name "
          << i->get_name() << " and ID "
          << i->get_sid()  << " is not flexible or flexible and titratable.\n"
          << "Will not calculate intrinsic MM energies for this site" << endl;
    continue;
   }
   if (!molecule.chmpara.contains_Topology(i->get_name())){
    cerr << "GcemCharmmEnergyCalculator::calc_mmint_propensities Warning: could not find topology entry in CHARMM topology for site name "
         << i->get_name() << ".\nWill not calculate intrinsic MM energy for this residue."
         << i->get_sid()  << endl;
    continue;
   }
   // retrieve topology
   CharmmParaResTop top(molecule.chmpara.get_Topology(i->get_name()));
   list<AtomID> aidlist;
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   list<string> ligats;
   if (stype == 1 || stype == 3){
    // this is a titratable and/or flexible site
    // collect all ligand atoms
    GcemSiteDef sdef(molecule.get_sitedef(i->get_name()));
    for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
     for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
      // search atom with zero occupancy
      if (l->second.second == 0){ligats.push_back(l->first.get_atname());}
     }
    }
    ligats.sort();
    ligats.unique();
   }

   map<InstanceID, map<string, double> > emap;
   map<string, double> min_e_map;

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    // all terms involving ligand + site atoms and molecule.background - model atoms
    double gmmback = 0;

    // all terms involving ligand atoms and only site atoms
    double gmml = 0;
    double gmml_e = 0;
    double gmml_lj = 0;
    double gmml_bond = 0;
    double gmml_angle = 0;
    double gmml_dihedral = 0;
    double gmml_improper = 0;

    // all terms not involving ligand atoms and only site+model atoms
    // are assumed to be accumulated in the rotamer propensity

    double gprop = 0;

    if (j->get_has_propensity()){
     // if the propensity is zero we would get an infinitly large gprop
     // Squirell has propensities in .6f so assume 5.0E-07
     double p = j->get_propensity();
     if (p > 0)
      gprop = -one_over_beta * log(p);
     else if (p == 0){
      gprop = -one_over_beta * log(5.0E-07);
      blab3 << j->get_instid() << endl;
      blab3 << "GcemCharmmEnergyCalculator::calc_mmint_propensities WARNING: Instance has zero propensity.\n"
            << "5.0E-07 is used instead." << endl;
     }
     else {
      cerr << j->get_instid() << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Instance has negative propensity.");
     }
    }
    else
     cerr << "GcemCharmmEnergyCalculator::calc_mmint_propensities: WARNING instance " << j->get_instid()
          << " has no propensity attached. Will not calculate complete intrinsic energy for this site" << endl;

    // all terms involving ligand atoms and also model atoms
    double mgmml = 0;
    double mgmml_e = 0;
    double mgmml_lj = 0;
    double mgmml_bond = 0;
    double mgmml_angle = 0;
    double mgmml_dihedral = 0;
    double mgmml_improper = 0;

    // calculate gmm_back

    for (AtomSet::const_iterator k=j->site_struct.begin(); k!=j->site_struct.end(); ++k){
     if (!molecule.chmpara.contains_attype(CharmmID(k->second)) && !molecule.chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: attype not found.");
     }
     CharmmParaNB nbk;
     if (molecule.chmpara.contains_NB(CharmmID(k->second))){
      nbk = molecule.chmpara.get_NB(CharmmID(k->second));
     }
     else if (k->second.atname == "OT1" || k->second.atname == "OT2" || k->second.atname == "OY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
     }
     else if (k->second.atname == "CAT" || k->second.atname == "CT" || k->second.atname == "CAY"){
      nbk = molecule.chmpara.get_NB(string("CT3"));
     }
     else if (k->second.atname == "CY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("C")));
     }
     else if (k->second.atname == "NT"){
      nbk = molecule.chmpara.get_NB(string("NH2"));
     }
     else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT" || k->second.atname == "HY1" || k->second.atname == "HY2" || k->second.atname == "HY3"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
     }
     else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
      nbk = molecule.chmpara.get_NB(string("H"));
     }
     else {
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Lennard Jones parameter missing.");
     }
     for (AtomSet::const_iterator l=molecule.background.begin(); l!=molecule.background.end(); ++l){

      ////////////////////////////////////////////////////////////////////////
      // FIXME: this is an ugly hack to be commented out in the trunk version
      // it allows to include my custom peptide flip residue in azurin
      if (k->second.resname == "PEP"){
       if (k->second.chain == l->second.chain && k->second.conformer == l->second.conformer){
        if (l->second.resnum == 35 || l->second.resnum == 39 || l->second.resnum == 87 || l->second.resnum == 91) continue;
/*
        // data gives first character
        if (l->second.resnum == 35 && k->second.atname.data() == "1"){
         if (l->second.atname == "C"){
          if        (k->second.atname == "1N"){
           continue;
          } else if (k->second.atname == "1CA"){
           continue;
          } else if (k->second.atname == "1HA"){
           continue;
          } else if (k->second.atname == "1CD"){
           continue;
          } else if (k->second.atname == "1HD1"){
           continue;
          } else if (k->second.atname == "1HD2"){
           continue;
          } else if (k->second.atname == "1C"){
           continue;
          } else if (k->second.atname == "1CB"){
           continue;
          } else if (k->second.atname == "1CG"){
           continue;
          }
         }
         if (l->second.atname == "O" || l->second.atname == "CA"){
          if        (k->second.atname == "1N"){
           continue;
          } else if (k->second.atname == "1CA"){
           continue;
          } else if (k->second.atname == "1CD"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB" || l->second.atname == "N"){
          if        (k->second.atname == "1N"){
           continue;
          }
         }
        }
        if (l->second.resnum == 87 && k->second.atname.data() == "4"){
         if (l->second.atname == "C"){
          if        (k->second.atname == "4N"){
           continue;
          } else if (k->second.atname == "4H"){
           continue;
          } else if (k->second.atname == "4HN"){
           continue;
          } else if (k->second.atname == "4CA"){
           continue;
          } else if (k->second.atname == "4HA1"){
           continue;
          } else if (k->second.atname == "4HA2"){
           continue;
          } else if (k->second.atname == "4C"){
           continue;
          }
         }
         if (l->second.atname == "O" || l->second.atname == "CA"){
          if        (k->second.atname == "4N"){
           continue;
          } else if (k->second.atname == "4CA"){
           continue;
          } else if (k->second.atname == "4HA1"){
           continue;
          } else if (k->second.atname == "4HA2"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB" || l->second.atname == "N"){
          if        (k->second.atname == "4N"){
           continue;
          }
         }
        }
        if (l->second.resnum == 39 && k->second.atname.data() == "3"){
         if (l->second.atname == "N"){
          if        (k->second.atname == "3N"){
           continue;
          } else if (k->second.atname == "3CA"){
           continue;
          } else if (k->second.atname == "3HA"){
           continue;
          } else if (k->second.atname == "3CB"){
           continue;
          } else if (k->second.atname == "3C"){
           continue;
          } else if (k->second.atname == "3O"){
           continue;
          }
         }
         if (l->second.atname == "H" || l->second.atname == "HN" || l->second.atname == "CA"){
          if        (k->second.atname == "3C"){
           continue;
          } else if (k->second.atname == "3O"){
           continue;
          } else if (k->second.atname == "3CA"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB"){
          if        (k->second.atname == "3C"){
           continue;
          }
         }
        }
        if (l->second.resnum == 91 && k->second.atname.data() == "6"){
         if (l->second.atname == "N"){
          if        (k->second.atname == "6N"){
           continue;
          } else if (k->second.atname == "6CA"){
           continue;
          } else if (k->second.atname == "6HA1"){
           continue;
          } else if (k->second.atname == "6HA2"){
           continue;
          } else if (k->second.atname == "6C"){
           continue;
          } else if (k->second.atname == "6O"){
           continue;
          }
         }
         if (l->second.atname == "H" || l->second.atname == "HN" || l->second.atname == "CA"){
          if        (k->second.atname == "6C"){
           continue;
          } else if (k->second.atname == "6O"){
           continue;
          } else if (k->second.atname == "6CA"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB"){
          if        (k->second.atname == "6C"){
           continue;
          }
         }
        }
*/
       }
      }
      ////////////////////////////////////////////////////////////////////////

      // skip model atoms, interactions within site are covered below
      if(j->model_struct.contains(l->first)){continue;}
      // get LJ parameters
      CharmmParaNB nbl;
      // molecule.background could contain unknown atom types
      if (molecule.chmpara.contains_attype(CharmmID(l->second))){
       if (molecule.chmpara.contains_NB(CharmmID(l->second))){
        nbl = molecule.chmpara.get_NB(CharmmID(l->second));
       }
      }
      else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
      }
      else if (l->second.atname == "CY"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("C")));
      }
      else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
       nbl = molecule.chmpara.get_NB(string("CT3"));
      }
      else if (l->second.atname == "NT"){
       nbl = molecule.chmpara.get_NB(string("NH2"));
      }
      else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
      }
      else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
       nbl = molecule.chmpara.get_NB(string("H"));
      }
      // sloppy search for matching atom name in patch residues
      else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
      }
      else if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname))){
       nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
      }
      else {
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Lennard Jones parameter missing.");
      }
      double dist = geom_calc.abs_value(k->second.coord - l->second.coord);
      double ljek = nbk.lje;
      double ljrk = nbk.ljr;
      double ljel = nbl.lje;
      double ljrl = nbl.ljr;
      if (ljek <= 0 && ljel <= 0){
       if ( dist != 0 ){
        double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
        gmmback += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
       } else {
        cerr << k->second << endl;
        cerr << l->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: two atoms have identical coordinates");
       }
      }
      else {
       if (ljek > 0 || ljel > 0){
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities atom has non-negative VDW well depth.");
       }
      }
     }
    }

    // calculate internal terms

    // first non-bonded terms

    for (AtomSet::const_iterator k=j->model_struct.begin(); k!=j->model_struct.end(); ++k){
     if (!molecule.chmpara.contains_attype(CharmmID(k->second)) && !molecule.chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: attype not found.");
     }
     AtomSet::const_iterator kplus = k; kplus++;
     for (AtomSet::const_iterator l=kplus; l!=j->model_struct.end(); ++l){
      if (!molecule.chmpara.contains_attype(CharmmID(l->second)) && !molecule.chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: attype not found.");
      }
      // skip term if no ligand atom is involved
      if (!molecule.in_list(k->second.atname,ligats) && !molecule.in_list(l->second.atname,ligats)) {continue;}

      // explore bonding situation
      bool i12 = top.in_12(k->second.atname,l->second.atname);
      bool i13 = top.in_13(k->second.atname,l->second.atname);
      bool i14 = top.in_14(k->second.atname,l->second.atname);
      if (!i12 && !i13){
       CharmmParaNB nbk, nbl;
       if (molecule.chmpara.contains_NB(CharmmID(k->second))){
        nbk = molecule.chmpara.get_NB(CharmmID(k->second));
       }
       else if (k->second.atname == "OT1" || k->second.atname == "OT2"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
       }
       else if (k->second.atname == "CAT" || k->second.atname == "CT"){
        nbk = molecule.chmpara.get_NB(string("CT3"));
       }
       else if (k->second.atname == "NT"){
        nbk = molecule.chmpara.get_NB(string("NH2"));
       }
       else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
       }
       else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
        nbk = molecule.chmpara.get_NB(string("H"));
       }
       else {
        cerr << k->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Lennard Jones parameter missing.");
       }
       if (molecule.chmpara.contains_NB(CharmmID(l->second))){
        nbl = molecule.chmpara.get_NB(CharmmID(l->second));
       }
       else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
       }
       else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
       }
       else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
        nbl = molecule.chmpara.get_NB(string("CT3"));
       }
       else if (l->second.atname == "CY"){
        nbl = molecule.chmpara.get_NB(string("C"));
       }
       else if (l->second.atname == "NT"){
        nbl = molecule.chmpara.get_NB(string("NH2"));
       }
       else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
        nbl = molecule.chmpara.get_NB(string("H"));
       }
       // sloppy search for matching atom name in patch residues
       else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
       }
       else if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname))){
        nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
       }
       else {
        cerr << l->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Lennard Jones parameter missing.");
       }
       double ljek, ljrk, ljel, ljrl;
       // 1-4 interaction may or may not have special LJ parameters
       if (i14){
        if (nbk.lje14!=0 && nbk.ljr14!=0){
         ljek = nbk.lje14;
         ljrk = nbk.ljr14;
        } else {
         ljek = nbk.lje;
         ljrk = nbk.ljr;
        }
        if (nbl.lje14!=0 && nbl.ljr14!=0){
         ljel = nbl.lje14;
         ljrl = nbl.ljr14;
        } else {
         ljel = nbl.lje;
         ljrl = nbl.ljr;
        }
       } else {
        ljek = nbk.lje;
        ljrk = nbk.ljr;
        ljel = nbl.lje;
        ljrl = nbl.ljr;
       }
       double tmpf1 = 0;
       double tmpf2 = 0;

       // Lennard Jones

       Coord dvec(k->second.coord - l->second.coord);
       double dist = geom_calc.abs_value(dvec);

       if (ljek <= 0 && ljel <= 0){
        if ( dist != 0 ){
         if (dist < 10){
           double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
          tmpf1 = sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
         }
        } else {
         cerr << k->second << endl;
         cerr << l->second << endl;
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: two atoms have identical coordinates");
        }
       }
       else {
        if (ljek > 0 || ljel > 0){
          ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities atom has non-negative VDW well depth.");
        }
       }

       // Coulomb Electrostatics
       tmpf2 = PhysCond::get_econv() * (k->second.charge * l->second.charge) / (dist * molecule.get_eps_ff());
       // Coulomb interactions may be scaled in CHARMM force field
       if (i14) tmpf2 *= molecule.get_coulomb_14_scalefac();

       if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first)){
        mgmml_lj += tmpf1;
        mgmml_e  += tmpf2;
       }
       else {
        gmml_lj += tmpf1;
        gmml_e  += tmpf2;
       }

      }
     }
    }

    // bond terms

    for (map<pair<string,string>,int>::const_iterator ii = top.bonds.begin(); ii != top.bonds.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!molecule.in_list(ii->first.first,ligats) && !molecule.in_list(ii->first.second,ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l;

     if (((ii->first).first).find("+") != 0 && ((ii->first).first).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.first,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((ii->first).first,ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.first,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && ii->first.first == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && ii->first.first == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((ii->first).first,ligats)){
       cerr << "non-ligand atom named " << (ii->first).first << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (((ii->first).second).find("+") != 0 && ((ii->first).second).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.second,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((ii->first).second,ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.second,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && ii->first.second == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && (ii->first.second == "HN" || ii->first.second == "HNT")){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((ii->first).second,ligats)){
       cerr << "non-ligand atom named " << (ii->first).second << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy

     CharmmParaBond tmppar;
     if (molecule.chmpara.contains_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second))){
      tmppar = molecule.chmpara.get_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second));
     }
     else if (molecule.chmpara.contains_Bond(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)))){
      tmppar = molecule.chmpara.get_Bond(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)));
     }
     else if (molecule.chmpara.contains_Bond(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname))){
      tmppar = molecule.chmpara.get_Bond(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Bond parameter missing.");
     }

     double dist = geom_calc.abs_value(k->second.coord - l->second.coord);
     double tmpf = tmppar.k * (dist - tmppar.r) * (dist - tmppar.r);

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first)){
      mgmml_bond += tmpf;
     }
     else {
      gmml_bond += tmpf;
     }

    }

    // angle terms

    for (list<vector<string> >::const_iterator ii = top.angles.begin(); ii != top.angles.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!molecule.in_list((*ii)[0],ligats) && !molecule.in_list((*ii)[1],ligats) && !molecule.in_list((*ii)[2],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){

      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "HN"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy

     CharmmParaAngle tmppar;

     if (molecule.chmpara.contains_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second))){
      tmppar = molecule.chmpara.get_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl << m->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Angle parameter missing.");
     }

     double angle  = geom_calc.get_angle(k->second.coord ,l->second.coord, m->second.coord);
     double phirad = tmppar.phi * (pi / 180);
     if (phirad < 0){phirad = -1 * phirad;}
     double tmpf = tmppar.kphi * (angle - phirad) * (angle - phirad);
     // Urey-Bradley term
     if (tmppar.kub != 0){
      double dist = geom_calc.abs_value(k->second.coord - m->second.coord);
      tmpf += tmppar.kub * (dist - tmppar.rub) * (dist - tmppar.rub);
     }

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first)){
       mgmml_angle += tmpf;
     }
     else {
      gmml_angle += tmpf;
     }

    }

    // dihedral terms

    for (list<vector<string> >::const_iterator ii = top.dihedrals.begin(); ii != top.dihedrals.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!molecule.in_list((*ii)[0],ligats) && !molecule.in_list((*ii)[1],ligats) && !molecule.in_list((*ii)[2],ligats) && !molecule.in_list((*ii)[3],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (n == j->model_struct.end() && molecule.in_list((*ii)[3],ligats)) {continue;}
      if (n == j->model_struct.end()){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "O"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "HN"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && !molecule.in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy
     list<CharmmParaDihedral> tmppar;

     if (molecule.chmpara.contains_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = molecule.chmpara.get_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Dihedral parameter missing.");
     }

     double dihe = geom_calc.get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     double tmpf = 0;
     for (list<CharmmParaDihedral>::const_iterator jj=tmppar.begin(); jj != tmppar.end(); ++jj){
      tmpf += jj->k * (1 + cos((jj->n * dihe) - (jj->delta * (pi / 180))));
     }

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first) || !j->site_struct.contains(n->first)){
      mgmml_dihedral += tmpf;
     }
     else {
      gmml_dihedral += tmpf;
     }

    }

    // improper terms

    for (list<vector<string> >::const_iterator ii = top.impropers.begin(); ii != top.impropers.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!molecule.in_list((*ii)[0],ligats) && !molecule.in_list((*ii)[1],ligats) && !molecule.in_list((*ii)[2],ligats) && !molecule.in_list((*ii)[3],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "HN"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "HN"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (n == j->model_struct.end() && molecule.in_list((*ii)[3],ligats)) {continue;}
      if (n == j->model_struct.end()){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "O"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "HN"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && !molecule.in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy
     CharmmParaImproper tmppar;

     if (molecule.chmpara.contains_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = molecule.chmpara.get_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities: Improper parameter missing.");
     }

     double imp = geom_calc.get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     double psirad = tmppar.psi * (pi / 180);
     double tmpf = tmppar.kpsi * (imp - psirad) * (imp - psirad);

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first) || !j->site_struct.contains(n->first)){
      mgmml_improper += tmpf;
     }
     else {
      gmml_improper += tmpf;
     }
    }

    gmml = gmml_e + gmml_lj + gmml_bond + gmml_angle + gmml_dihedral + gmml_improper;
    mgmml = mgmml_e + mgmml_lj + mgmml_bond + mgmml_angle + mgmml_dihedral + mgmml_improper;

    if (molecule.get_print_chm()){

     cout << endl << "---------------------------------------" << endl;

     cout << "Energies for instance " << (j->get_instid()).get_fn_prefix() << endl;

     cout << "gmmback "       << gmmback << endl;

     cout << "gmml "          << gmml << endl;
     cout << "gmml_e "        << gmml_e << endl;
     cout << "gmml_lj "       << gmml_lj << endl;
     cout << "gmml_bond "     << gmml_bond << endl;
     cout << "gmml_angle "    << gmml_angle << endl;
     cout << "gmml_dihedral " << gmml_dihedral << endl;
     cout << "gmml_improper " << gmml_improper << endl;

     cout << "mgmml "          << mgmml << endl;
     cout << "mgmml_e "        << mgmml_e << endl;
     cout << "mgmml_lj "       << mgmml_lj << endl;
     cout << "mgmml_bond "     << mgmml_bond << endl;
     cout << "mgmml_angle "    << mgmml_angle << endl;
     cout << "mgmml_dihedral " << mgmml_dihedral << endl;
     cout << "mgmml_improper " << mgmml_improper << endl;

     cout << "gprop "          << gprop << endl;

     cout << endl << "---------------------------------------" << endl;

    }

    map<InstanceID, map<string, double> >::iterator em_iter = emap.find(j->get_instid());
    // should never happen
    if (em_iter != emap.end()) {
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities non-unique InstanceID\n");
    }

    map<string,double> tmpmap;
    tmpmap.insert(pair<string,double>("gmmback",gmmback));
    pair<InstanceID,map<string, double> > tmppair1(j->get_instid(),tmpmap);

    pair<map<InstanceID, map<string, double> >::iterator, bool> tmpair2 = emap.insert(tmppair1);

    em_iter = tmpair2.first;

    em_iter->second.insert(pair<string,double>("gmml_e",gmml_e));
    em_iter->second.insert(pair<string,double>("gmml_lj",gmml_lj));
    em_iter->second.insert(pair<string,double>("gmml_bond",gmml_bond));
    em_iter->second.insert(pair<string,double>("gmml_angle",gmml_angle));
    em_iter->second.insert(pair<string,double>("gmml_dihedral",gmml_dihedral));
    em_iter->second.insert(pair<string,double>("gmml_improper",gmml_improper));

    em_iter->second.insert(pair<string,double>("mgmml_e",mgmml_e));
    em_iter->second.insert(pair<string,double>("mgmml_lj",mgmml_lj));
    em_iter->second.insert(pair<string,double>("mgmml_bond",mgmml_bond));
    em_iter->second.insert(pair<string,double>("mgmml_angle",mgmml_angle));
    em_iter->second.insert(pair<string,double>("mgmml_dihedral",mgmml_dihedral));
    em_iter->second.insert(pair<string,double>("mgmml_improper",mgmml_improper));

    em_iter->second.insert(pair<string,double>("gprop",gprop));

   }

   // search for minimum ligand-site/model energy instance of each form of the site

   map<string,double> min_gmml;

   for (map<InstanceID,map<string, double> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    double tmpf = 0;
    map<string, double>::const_iterator titr;
    string formname(eitr->first.get_instname());

    // ligand-site terms

    if (molecule.get_use_term_l_e()){
      titr = eitr->second.find(string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_lj()){
      titr = eitr->second.find(string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_bond()){
      titr = eitr->second.find(string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_angle()){
      titr = eitr->second.find(string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_dihedral()){
      titr = eitr->second.find(string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_improper()){
      titr = eitr->second.find(string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    // ligand-model terms

    if (molecule.get_use_term_ml_e()){
      titr = eitr->second.find(string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_lj()){
      titr = eitr->second.find(string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_bond()){
      titr = eitr->second.find(string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_angle()){
      titr = eitr->second.find(string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_dihedral()){
      titr = eitr->second.find(string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_improper()){
      titr = eitr->second.find(string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // create or update minimum energy entry for this form of the site

    map<string,double>::iterator meitr(min_gmml.find(formname));
    if (meitr == min_gmml.end()){
     pair<string,double> tmppair(formname,tmpf);
     min_gmml.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }
   }

   // now collect all energy terms in the final terms

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    double gmmint = 0;
    double gmmback = 0;
    double gmmsite = 0;
    double gprop = 0;
    double gmmmodelback = 0;

    map<InstanceID,map<string, double> >::const_iterator eitr(emap.find(j->get_instid()));
    if (eitr == emap.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities instance missing in energy term map\n");
    map<string, double>::const_iterator titr(eitr->second.find(string("gmmback")));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");

    gmmback = titr->second;

    titr = eitr->second.find(string("gprop"));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
    else
     gprop = titr->second;

    double min_l = 0;

    map<string, double>::const_iterator ml_itr(min_gmml.find(j->get_name()));
    if (ml_itr != min_gmml.end()) min_l = ml_itr->second;

    double gmml = 0;

    // ligand-site terms

    if (molecule.get_use_term_l_e()){
      titr = eitr->second.find(string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_lj()){
      titr = eitr->second.find(string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_bond()){
      titr = eitr->second.find(string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_angle()){
      titr = eitr->second.find(string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_dihedral()){
      titr = eitr->second.find(string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_improper()){
      titr = eitr->second.find(string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    double mgmml = 0;

    // ligand-modelback terms

    if (molecule.get_use_term_ml_e()){
      titr = eitr->second.find(string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_lj()){
      titr = eitr->second.find(string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_bond()){
      titr = eitr->second.find(string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_angle()){
      titr = eitr->second.find(string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_dihedral()){
      titr = eitr->second.find(string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_improper()){
      titr = eitr->second.find(string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    // minimum ligand-site interaction should be included in model energy
    // correct for this by considering only relative ligand site interactions within 1 form
    gmmsite      = gprop  + gmml - min_l;
    // interactions of non-ligand site atoms with modelback are assumed to be included in gprop
    gmmmodelback = mgmml;

    j->set_gmmsite(gmmsite);
    j->set_gmmmodelback(gmmmodelback);
    j->set_gmmback(gmmback);

   }



  }

  if (molecule.get_print_chm()) molecule.print_mmint();

  blab1 << "Exiting GcemCharmmEnergyCalculator::calc_mmint_propensities" << endl;

}

void GcemCharmmEnergyCalculator::calc_mmint_charmm(GcemMolecule &molecule) const
{

  blab2 << "Entering GcemCharmmEnergyCalculator::calc_mmint_charmm" << endl;
  
  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   int stype(molecule.get_rotmap_value(i->get_sid()));
   if (stype != 1 && stype != 2 && molecule.get_skip_stat_mmint()){
    blab1 << "GcemCharmmEnergyCalculator::calc_mmint_charmm: site with name "
          << i->get_name() << " and ID "
          << i->get_sid()  << " is not flexible or flexible and titratable.\n"
          << "Will not calculate intrinsic MM energies for this site" << endl;
    continue;
   }
   // QM sites treated separately
   if (stype == 4){
    continue;
   }
   if (!molecule.chmpara.contains_Topology(i->get_name())){
    cerr << "GcemCharmmEnergyCalculator::calc_mmint_charmm Warning: could not find topology entry in CHARMM topology for site name "
         << i->get_name() << ".\nWill not calculate intrinsic MM energy for this residue."
         << i->get_sid()  << endl;
    continue;
   }
   // retrieve topology
   CharmmParaResTop top(molecule.chmpara.get_Topology(i->get_name()));
   list<AtomID> aidlist;
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   list<string> ligats;
   if (stype == 1 || stype == 3){
    // this is a titratable and flexible site
    // collect all ligand atoms
    GcemSiteDef sdef(molecule.get_sitedef(i->get_name()));
    for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
     for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
      // search atom with zero occupancy
      if (l->second.second == 0){ligats.push_back(l->first.get_atname());}
     }
    }
    ligats.sort();
    ligats.unique();
   }

   map<InstanceID, map<string, double> > emap;
   map<string, double> min_e_map;

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    // all terms involving ligand + site atoms and molecule.background - model atoms
    double gmmback = 0;

    // all terms involving ligand atoms and only site atoms
    double gmml = 0;
    double gmml_e = 0;
    double gmml_lj = 0;
    double gmml_bond = 0;
    double gmml_angle = 0;
    double gmml_dihedral = 0;
    double gmml_improper = 0;
    // all terms not involving ligand atoms and only site atoms
    double gmmr = 0;
    double gmmr_e = 0;
    double gmmr_lj = 0;
    double gmmr_bond = 0;
    double gmmr_angle = 0;
    double gmmr_dihedral = 0;
    double gmmr_improper = 0;

    // all terms involving ligand atoms and also model atoms
    double mgmml = 0;
    double mgmml_e = 0;
    double mgmml_lj = 0;
    double mgmml_bond = 0;
    double mgmml_angle = 0;
    double mgmml_dihedral = 0;
    double mgmml_improper = 0;
    // all terms not involving ligand atoms but model_atoms
    double mgmmr = 0;
    double mgmmr_e = 0;
    double mgmmr_lj = 0;
    double mgmmr_bond = 0;
    double mgmmr_angle = 0;
    double mgmmr_dihedral = 0;
    double mgmmr_improper = 0;

    // calculate gmm_back
    if (molecule.get_print_debug()) cout << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing molecule.background terms" << endl;

    for (AtomSet::const_iterator k=j->site_struct.begin(); k!=j->site_struct.end(); ++k){
     if (!molecule.chmpara.contains_attype(CharmmID(k->second)) && !molecule.chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: attype not found.");
     }
     CharmmParaNB nbk;
     if (molecule.chmpara.contains_NB(CharmmID(k->second))){
      nbk = molecule.chmpara.get_NB(CharmmID(k->second));
     }
     else if (k->second.atname == "OT1" || k->second.atname == "OT2" || k->second.atname == "OY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
     }
     else if (k->second.atname == "CAT" || k->second.atname == "CT" || k->second.atname == "CAY"){
      nbk = molecule.chmpara.get_NB(string("CT3"));
     }
     else if (k->second.atname == "CY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("C")));
     }
     else if (k->second.atname == "NT"){
      nbk = molecule.chmpara.get_NB(string("NH2"));
     }
     else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT" || k->second.atname == "HY1" || k->second.atname == "HY2" || k->second.atname == "HY3"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
     }
     else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
      nbk = molecule.chmpara.get_NB(string("H"));
     }
     else {
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Lennard Jones parameter missing.");
     }
     for (AtomSet::const_iterator l=molecule.background.begin(); l!=molecule.background.end(); ++l){

      ////////////////////////////////////////////////////////////////////////
      // FIXME: this is an ugly hack to be commented out in the trunk version
      // it allows to include my custom peptide flip residue in azurin
      if (k->second.resname == "PEP"){
       if (k->second.chain == l->second.chain && k->second.conformer == l->second.conformer){
        // data gives first character
        if (l->second.resnum == 35 || l->second.resnum == 39 || l->second.resnum == 87 || l->second.resnum == 91) continue;
/*
        if (l->second.resnum == 35 && k->second.atname.data() == "1"){
         if (l->second.atname == "C"){
          if        (k->second.atname == "1N"){
           continue;
          } else if (k->second.atname == "1CA"){
           continue;
          } else if (k->second.atname == "1HA"){
           continue;
          } else if (k->second.atname == "1CD"){
           continue;
          } else if (k->second.atname == "1HD1"){
           continue;
          } else if (k->second.atname == "1HD2"){
           continue;
          } else if (k->second.atname == "1C"){
           continue;
          } else if (k->second.atname == "1CB"){
           continue;
          } else if (k->second.atname == "1CG"){
           continue;
          }
         }
         if (l->second.atname == "O" || l->second.atname == "CA"){
          if        (k->second.atname == "1N"){
           continue;
          } else if (k->second.atname == "1CA"){
           continue;
          } else if (k->second.atname == "1CD"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB" || l->second.atname == "N"){
          if        (k->second.atname == "1N"){
           continue;
          }
         }
        }
        if (l->second.resnum == 87 && k->second.atname.data() == "4"){
         if (l->second.atname == "C"){
          if        (k->second.atname == "4N"){
           continue;
          } else if (k->second.atname == "4H"){
           continue;
          } else if (k->second.atname == "4HN"){
           continue;
          } else if (k->second.atname == "4CA"){
           continue;
          } else if (k->second.atname == "4HA1"){
           continue;
          } else if (k->second.atname == "4HA2"){
           continue;
          } else if (k->second.atname == "4C"){
           continue;
          }
         }
         if (l->second.atname == "O" || l->second.atname == "CA"){
          if        (k->second.atname == "4N"){
           continue;
          } else if (k->second.atname == "4CA"){
           continue;
          } else if (k->second.atname == "4HA1"){
           continue;
          } else if (k->second.atname == "4HA2"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB" || l->second.atname == "N"){
          if        (k->second.atname == "4N"){
           continue;
          }
         }
        }
        if (l->second.resnum == 39 && k->second.atname.data() == "3"){
         if (l->second.atname == "N"){
          if        (k->second.atname == "3N"){
           continue;
          } else if (k->second.atname == "3CA"){
           continue;
          } else if (k->second.atname == "3HA"){
           continue;
          } else if (k->second.atname == "3CB"){
           continue;
          } else if (k->second.atname == "3C"){
           continue;
          } else if (k->second.atname == "3O"){
           continue;
          }
         }
         if (l->second.atname == "H" || l->second.atname == "HN" || l->second.atname == "CA"){
          if        (k->second.atname == "3C"){
           continue;
          } else if (k->second.atname == "3O"){
           continue;
          } else if (k->second.atname == "3CA"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB"){
          if        (k->second.atname == "3C"){
           continue;
          }
         }
        }
        if (l->second.resnum == 91 && k->second.atname.data() == "6"){
         if (l->second.atname == "N"){
          if        (k->second.atname == "6N"){
           continue;
          } else if (k->second.atname == "6CA"){
           continue;
          } else if (k->second.atname == "6HA1"){
           continue;
          } else if (k->second.atname == "6HA2"){
           continue;
          } else if (k->second.atname == "6C"){
           continue;
          } else if (k->second.atname == "6O"){
           continue;
          }
         }
         if (l->second.atname == "H" || l->second.atname == "HN" || l->second.atname == "CA"){
          if        (k->second.atname == "6C"){
           continue;
          } else if (k->second.atname == "6O"){
           continue;
          } else if (k->second.atname == "6CA"){
           continue;
          }
         }
         if (l->second.atname == "HA" || l->second.atname == "CB"){
          if        (k->second.atname == "6C"){
           continue;
          }
         }
        }
*/
       }
      }

      ////////////////////////////////////////////////////////////////////////

      if (!molecule.chmpara.contains_attype(CharmmID(l->second)) && !molecule.chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: attype not found.");
      }
      // skip model atoms, interactions within site are covered below
      if(j->model_struct.contains(l->first)){continue;}
      // get LJ parameters
      CharmmParaNB nbl;
      // molecule.background could contain unknown atom types
      if (molecule.chmpara.contains_attype(CharmmID(l->second))){
       if (molecule.chmpara.contains_NB(CharmmID(l->second))){
        nbl = molecule.chmpara.get_NB(CharmmID(l->second));
       }
      }
      else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
      }
      else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
      }
      else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
       nbl = molecule.chmpara.get_NB(string("CT3"));
      }
      else if (l->second.atname == "CY"){
       nbl = molecule.chmpara.get_NB(string("C"));
      }
      else if (l->second.atname == "NT"){
       nbl = molecule.chmpara.get_NB(string("NH2"));
      }
      else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
       nbl = molecule.chmpara.get_NB(string("H"));
      }
      // sloppy search for matching atom name in patch residues
      else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
      }
      else if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname))){
       nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
      }
      else {
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: Lennard Jones parameter missing.");
      }
      double dist = geom_calc.abs_value(k->second.coord - l->second.coord);
      double ljek = nbk.lje;
      double ljrk = nbk.ljr;
      double ljel = nbl.lje;
      double ljrl = nbl.ljr;
      if (ljek <= 0 && ljel <= 0){
       if ( dist != 0 ){
        double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
        gmmback += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
       } else {
        cerr << k->second << endl;
        cerr << l->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: two atoms have identical coordinates");
       }
      }
      else {
       if (ljek > 0 || ljel > 0){
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm atom has non-negative VDW well depth.");
       }
      }
     }
    }

    // calculate internal terms
    if (molecule.get_print_debug()) cout << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing internal terms" << endl;

    // first non-bonded terms
    if (molecule.get_print_debug())
     cout << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing non-bonded terms" << endl;

    for (AtomSet::const_iterator k=j->model_struct.begin(); k!=j->model_struct.end(); ++k){
     if (!molecule.chmpara.contains_attype(CharmmID(k->second)) && !molecule.chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: attype not found.");
     }
     AtomSet::const_iterator kplus = k; kplus++;
     for (AtomSet::const_iterator l=kplus; l!=j->model_struct.end(); ++l){
      if (!molecule.chmpara.contains_attype(CharmmID(l->second)) && !molecule.chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: attype not found.");
      }
      if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first)) {continue;}
      // explore bonding situation
      bool i12 = top.in_12(k->second.atname,l->second.atname);
      bool i13 = top.in_13(k->second.atname,l->second.atname);
      bool i14 = top.in_14(k->second.atname,l->second.atname);
      if (!i12 && !i13){
       CharmmParaNB nbk, nbl;
       if (molecule.chmpara.contains_attype(CharmmID(k->second))){
        if (molecule.chmpara.contains_NB(CharmmID(k->second))){
         nbk = molecule.chmpara.get_NB(CharmmID(k->second));
        }
       }
       else if (k->second.atname == "OT1" || k->second.atname == "OT2" || k->second.atname == "OY"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
       }
       else if (k->second.atname == "CAT" || k->second.atname == "CT" || k->second.atname == "CAY"){
        nbk = molecule.chmpara.get_NB(string("CT3"));
       }
       else if (k->second.atname == "CY"){
        nbl = molecule.chmpara.get_NB(string("C"));
       }
       else if (k->second.atname == "NT"){
        nbk = molecule.chmpara.get_NB(string("NH2"));
       }
       else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT" || k->second.atname == "HY1" || k->second.atname == "HY2" || k->second.atname == "HY3"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
       }
       else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
        nbk = molecule.chmpara.get_NB(string("H"));
       }
       else {
        cerr << k->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Lennard Jones parameter missing.");
       }
       if (molecule.chmpara.contains_attype(CharmmID(l->second))){
        if (molecule.chmpara.contains_NB(CharmmID(l->second))){
         nbl = molecule.chmpara.get_NB(CharmmID(l->second));
        }
       }
       else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
       }
       else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
       }
       else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
        nbl = molecule.chmpara.get_NB(string("CT3"));
       }
       else if (l->second.atname == "CY"){
        nbl = molecule.chmpara.get_NB(string("C"));
       }
       else if (l->second.atname == "NT"){
        nbl = molecule.chmpara.get_NB(string("NH2"));
       }
       else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
        nbl = molecule.chmpara.get_NB(string("H"));
       }
       // sloppy search for matching atom name in patch residues
       else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
        nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
       }
       else if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname))){
        nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
       }
       else {
        cerr << l->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Lennard Jones parameter missing.");
       }
       double ljek, ljrk, ljel, ljrl;
       // 1-4 interaction may or may not have special LJ parameters
       if (i14){
        if (nbk.lje14!=0 && nbk.ljr14!=0){
         ljek = nbk.lje14;
         ljrk = nbk.ljr14;
        } else {
         ljek = nbk.lje;
         ljrk = nbk.ljr;
        }
        if (nbl.lje14!=0 && nbl.ljr14!=0){
         ljel = nbl.lje14;
         ljrl = nbl.ljr14;
        } else {
         ljel = nbl.lje;
         ljrl = nbl.ljr;
        }
       } else {
        ljek = nbk.lje;
        ljrk = nbk.ljr;
        ljel = nbl.lje;
        ljrl = nbl.ljr;
       }
       double tmpf1 = 0;
       double tmpf2 = 0;

       // Lennard Jones

       Coord dvec(k->second.coord - l->second.coord);
       double dist = geom_calc.abs_value(dvec);

       if (ljek <= 0 && ljel <= 0){
        if ( dist != 0 ){
         if (dist < 10){
          double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
          tmpf1 = sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
         }
        } else {
         cerr << k->second << endl;
         cerr << l->second << endl;
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: two atoms have identical coordinates");
        }
       }
       else {
        if (ljek > 0 || ljel > 0){
          ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: atom has non-negative VDW well depth.");
        }
       }

       // Coulomb Electrostatics

       tmpf2 = PhysCond::get_econv() * (k->second.charge * l->second.charge) / (dist * molecule.get_eps_ff());
       // 1-4 Coulomb interactions may be scaled in CHARMM force field
       if (i14) tmpf2 *= molecule.get_coulomb_14_scalefac();

       if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first)){
        if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats)){
         mgmml_lj += tmpf1;
         mgmml_e  += tmpf2;
        } else {
         if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first)){continue;}
         mgmmr_lj += tmpf1;
         mgmmr_e  += tmpf2;
        }
       }
       else {
        if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats)){
         gmml_lj += tmpf1;
         gmml_e  += tmpf2;
        } else {
         gmmr_lj += tmpf1;
         gmmr_e  += tmpf2;
        }
       }

      }
     }
    }

    if (molecule.get_print_debug()) cout << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing bond terms" << endl;

    // bond terms

    for (map<pair<string,string>,int>::const_iterator ii = top.bonds.begin(); ii != top.bonds.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l;

     if (((ii->first).first).find("+") != 0 && ((ii->first).first).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.first,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((ii->first).first,ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.first,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && ii->first.first == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (ii->first.first == "HN" || ii->first.first == "HNT")){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((ii->first).first,ligats)){
       cerr << "non-ligand atom named " << (ii->first).first << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (((ii->first).second).find("+") != 0 && ((ii->first).second).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.second,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((ii->first).second,ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), ii->first.second,
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && ii->first.second == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && (ii->first.second == "HN" || ii->first.second == "HNT")){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((ii->first).second,ligats)){
       cerr << "non-ligand atom named " << (ii->first).second << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first)) {continue;}

     if (molecule.get_print_debug()) cout << "bond  " << endl << k->first << "   cid " << CharmmID(k->second)
                                       << endl << l->first << "   cid " << CharmmID(l->second)
                                       << endl;

    // retrieve parameters and calculate energy

     CharmmParaBond tmppar;
     if (molecule.chmpara.contains_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second))){
      tmppar = molecule.chmpara.get_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second));
     }
     else if (molecule.chmpara.contains_Bond(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)))){
      tmppar = molecule.chmpara.get_Bond(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)));
     }
     else if (molecule.chmpara.contains_Bond(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname))){
      tmppar = molecule.chmpara.get_Bond(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: Bond parameter missing.");
     }

     double dist = geom_calc.abs_value(k->second.coord - l->second.coord);
     double tmpf = tmppar.k * (dist - tmppar.r) * (dist - tmppar.r);

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first)){
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats)){
       mgmml_bond += tmpf;
      } else {
       if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first)){continue;}
       mgmmr_bond += tmpf;
      }
     }
     else {
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats)){
       gmml_bond += tmpf;
      } else {
       gmmr_bond += tmpf;
      }
     }

    }

    blab3 << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing angle terms" << endl;

    // angle terms

    for (list<vector<string> >::const_iterator ii = top.angles.begin(); ii != top.angles.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){

      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "H"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "HN"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (molecule.get_print_debug()) cout << "angle " << endl << k->first << "   cid " << CharmmID(k->second)
                                       << endl << l->first << "   cid " << CharmmID(l->second)
                                       << endl << m->first << "   cid " << CharmmID(m->second)
                                       << endl;

     // skip term if no site atom is involved
     if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first)) {continue;}

     // retrieve parameters and calculate energy

     CharmmParaAngle tmppar;

     if (molecule.chmpara.contains_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second))){
      tmppar = molecule.chmpara.get_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)));
     }
     else if (molecule.chmpara.contains_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname))){
      tmppar = molecule.chmpara.get_Angle(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl << m->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: Angle parameter missing.");
     }

     double angle  = geom_calc.get_angle(k->second.coord ,l->second.coord, m->second.coord);
     double phirad = tmppar.phi * (pi / 180);
     if (phirad < 0){phirad = -1 * phirad;}
     double tmpf = tmppar.kphi * (angle - phirad) * (angle - phirad);
     // Urey-Bradley term
     if (tmppar.kub != 0){
      double dist = geom_calc.abs_value(k->second.coord - m->second.coord);
      tmpf += tmppar.kub * (dist - tmppar.rub) * (dist - tmppar.rub);
     }

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first)){
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) || molecule.in_list(m->second.atname, ligats)){
       mgmml_angle += tmpf;
      } else {
       if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first)){continue;}
       mgmmr_angle += tmpf;
      }
     }
     else {
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) || molecule.in_list(m->second.atname, ligats)){
       gmml_angle += tmpf;
      } else {
       gmmr_angle += tmpf;
      }
     }

    }

    blab3 << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing dihedral terms" << endl;

    // dihedral terms

    for (list<vector<string> >::const_iterator ii = top.dihedrals.begin(); ii != top.dihedrals.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (n == j->model_struct.end() && molecule.in_list((*ii)[3],ligats)) {continue;}
      if (n == j->model_struct.end()){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "O"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "HN"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && !molecule.in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (molecule.get_print_debug()) cout << "dihedral " << endl << k->first << "   cid " << CharmmID(k->second)
                                          << endl << l->first << "   cid " << CharmmID(l->second)
                                          << endl << m->first << "   cid " << CharmmID(m->second)
                                          << endl << n->first << "   cid " << CharmmID(n->second)
                                          << endl;

     // skip term if no site atom is involved
     if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first) && !j->site_struct.contains(n->first)) {continue;}

     // retrieve parameters and calculate energy
     list<CharmmParaDihedral> tmppar;

     if (molecule.chmpara.contains_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = molecule.chmpara.get_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname))){
      tmppar = molecule.chmpara.get_Dihedral(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Dihedral parameter missing.");
     }
     /*
     if (k->second.coord == l->second.coord || k->second.coord == m->second.coord || k->second.coord == n->second.coord ||
         l->second.coord == m->second.coord || l->second.coord == n->second.coord ||
         m->second.coord == n->second.coord){
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       cerr << k->second.coord << endl << l->second.coord << endl << m->second.coord << endl << n->second.coord << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Two atoms have identical coordinates.");
     }
     */
     double dihe = geom_calc.get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     double tmpf = 0;
     int ccc = 0;
     for (list<CharmmParaDihedral>::const_iterator jj=tmppar.begin(); jj != tmppar.end(); ++jj){
      if (jj->k == 0) continue;
      tmpf += jj->k * (1 + cos((jj->n * dihe) - (jj->delta * (pi / 180))));
     }

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first) || !j->site_struct.contains(n->first)){
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) ||
          molecule.in_list(m->second.atname, ligats) || molecule.in_list(n->second.atname, ligats)){
       mgmml_dihedral += tmpf;
      } else {
       if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first) && !j->site_struct.contains(n->first)){continue;}
       mgmmr_dihedral += tmpf;
      }
     }
     else {
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) ||
          molecule.in_list(m->second.atname, ligats) || molecule.in_list(n->second.atname, ligats)){
       gmml_dihedral += tmpf;
      } else {
       gmmr_dihedral += tmpf;
      }
     }

    }

    blab3 << "GcemCharmmEnergyCalculator::calc_mmint_charmm computing improper terms" << endl;

    // improper terms

    for (list<vector<string> >::const_iterator ii = top.impropers.begin(); ii != top.impropers.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (k == j->model_struct.end() && molecule.in_list((*ii)[0],ligats)) {continue;}
      if (k == j->model_struct.end()){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[0],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "O"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && (*ii)[0] == "HN"){
       k = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (k == j->model_struct.end() && !molecule.in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (l == j->model_struct.end() && molecule.in_list((*ii)[1],ligats)) {continue;}
      if (l == j->model_struct.end()){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[1],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "O"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && (*ii)[1] == "HN"){
       l = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (l == j->model_struct.end() && !molecule.in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (m == j->model_struct.end() && molecule.in_list((*ii)[2],ligats)) {continue;}
      if (m == j->model_struct.end()){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[2],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "O"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && (*ii)[2] == "HN"){
       m = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (m == j->model_struct.end() && !molecule.in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), (j->get_instid()).get_instance()));
      if (n == j->model_struct.end() && molecule.in_list((*ii)[3],ligats)) {continue;}
      if (n == j->model_struct.end()){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), (*ii)[3],
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(), int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "O"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "OT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && (*ii)[3] == "HN"){
       n = j->model_struct.find(
          AtomID((j->get_instid()).get_site(), "HT1",
                 (j->get_instid()).get_conformer(), (j->get_instid()).get_chain(), (j->get_instid()).get_site(),
                 int(0)));
      }
      if (n == j->model_struct.end() && !molecule.in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << i->get_sid() << endl;
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm: IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first) && !j->site_struct.contains(n->first)) {continue;}

     // retrieve parameters and calculate energy
     CharmmParaImproper tmppar;

     if (molecule.chmpara.contains_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = molecule.chmpara.get_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_pres_attype(k->second.atname), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_pres_attype(l->second.atname), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_pres_attype(m->second.atname), molecule.chmpara.get_attype(CharmmID(n->second)));
     }
     else if (molecule.chmpara.contains_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname))){
      tmppar = molecule.chmpara.get_Improper(molecule.chmpara.get_attype(CharmmID(k->second)), molecule.chmpara.get_attype(CharmmID(l->second)), molecule.chmpara.get_attype(CharmmID(m->second)), molecule.chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint: Improper parameter missing.");
     }

     double imp = geom_calc.get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     double psirad = tmppar.psi * (pi / 180);
     double tmpf = tmppar.kpsi * (imp - psirad) * (imp - psirad);

     if (!j->site_struct.contains(k->first) || !j->site_struct.contains(l->first) || !j->site_struct.contains(m->first) || !j->site_struct.contains(n->first)){
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) ||
          molecule.in_list(m->second.atname, ligats) || molecule.in_list(n->second.atname, ligats)){
       mgmml_improper += tmpf;
      } else {
       if (!j->site_struct.contains(k->first) && !j->site_struct.contains(l->first) && !j->site_struct.contains(m->first) && !j->site_struct.contains(n->first)){continue;}
       mgmmr_improper += tmpf;
      }
     }
     else {
      if (molecule.in_list(k->second.atname, ligats) || molecule.in_list(l->second.atname, ligats) ||
          molecule.in_list(m->second.atname, ligats) || molecule.in_list(n->second.atname, ligats)){
       gmml_improper += tmpf;
      } else {
       gmmr_improper += tmpf;
      }
     }
    }

    gmml = gmml_e + gmml_lj + gmml_bond + gmml_angle + gmml_dihedral + gmml_improper;
    gmmr = gmmr_e + gmmr_lj + gmmr_bond + gmmr_angle + gmmr_dihedral + gmmr_improper;
    mgmml = mgmml_e + mgmml_lj + mgmml_bond + mgmml_angle + mgmml_dihedral + mgmml_improper;
    mgmmr = mgmmr_e + mgmmr_lj + mgmmr_bond + mgmmr_angle + mgmmr_dihedral + mgmmr_improper;

    if (molecule.get_print_chm()){

     cout << endl << "---------------------------------------" << endl;

     cout << "Energies for instance " << (j->get_instid()).get_fn_prefix() << endl;

     cout << "gmmback "       << gmmback << endl;

     cout << "gmml "          << gmml << endl;
     cout << "gmml_e "        << gmml_e << endl;
     cout << "gmml_lj "       << gmml_lj << endl;
     cout << "gmml_bond "     << gmml_bond << endl;
     cout << "gmml_angle "    << gmml_angle << endl;
     cout << "gmml_dihedral " << gmml_dihedral << endl;
     cout << "gmml_improper " << gmml_improper << endl;

     cout << "gmmr "          << gmmr << endl;
     cout << "gmmr_e "        << gmmr_e << endl;
     cout << "gmmr_lj "       << gmmr_lj << endl;
     cout << "gmmr_bond "     << gmmr_bond << endl;
     cout << "gmmr_angle "    << gmmr_angle << endl;
     cout << "gmmr_dihedral " << gmmr_dihedral << endl;
     cout << "gmmr_improper " << gmmr_improper << endl;

     cout << "mgmml "          << mgmml << endl;
     cout << "mgmml_e "        << mgmml_e << endl;
     cout << "mgmml_lj "       << mgmml_lj << endl;
     cout << "mgmml_bond "     << mgmml_bond << endl;
     cout << "mgmml_angle "    << mgmml_angle << endl;
     cout << "mgmml_dihedral " << mgmml_dihedral << endl;
     cout << "mgmml_improper " << mgmml_improper << endl;

     cout << "mgmmr "          << mgmmr << endl;
     cout << "mgmmr_e "        << mgmmr_e << endl;
     cout << "mgmmr_lj "       << mgmmr_lj << endl;
     cout << "mgmmr_bond "     << mgmmr_bond << endl;
     cout << "mgmmr_angle "    << mgmmr_angle << endl;
     cout << "mgmmr_dihedral " << mgmmr_dihedral << endl;
     cout << "mgmmr_improper " << mgmmr_improper << endl;

     cout << endl << "---------------------------------------" << endl;

    }

    map<InstanceID, map<string, double> >::iterator em_iter = emap.find(j->get_instid());
    // should never happen
    if (em_iter != emap.end()) {
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint non-unique InstanceID\n");
    }

    map<string,double> tmpmap;
    tmpmap.insert(pair<string,double>("gmmback",gmmback));
    pair<InstanceID,map<string, double> > tmppair1(j->get_instid(),tmpmap);

    pair<map<InstanceID, map<string, double> >::iterator, bool> tmpair2 = emap.insert(tmppair1);

    em_iter = tmpair2.first;

    em_iter->second.insert(pair<string,double>("gmml_e",gmml_e));
    em_iter->second.insert(pair<string,double>("gmml_lj",gmml_lj));
    em_iter->second.insert(pair<string,double>("gmml_bond",gmml_bond));
    em_iter->second.insert(pair<string,double>("gmml_angle",gmml_angle));
    em_iter->second.insert(pair<string,double>("gmml_dihedral",gmml_dihedral));
    em_iter->second.insert(pair<string,double>("gmml_improper",gmml_improper));

    em_iter->second.insert(pair<string,double>("gmmr_e",gmmr_e));
    em_iter->second.insert(pair<string,double>("gmmr_lj",gmmr_lj));
    em_iter->second.insert(pair<string,double>("gmmr_bond",gmmr_bond));
    em_iter->second.insert(pair<string,double>("gmmr_angle",gmmr_angle));
    em_iter->second.insert(pair<string,double>("gmmr_dihedral",gmmr_dihedral));
    em_iter->second.insert(pair<string,double>("gmmr_improper",gmmr_improper));

    em_iter->second.insert(pair<string,double>("mgmml_e",mgmml_e));
    em_iter->second.insert(pair<string,double>("mgmml_lj",mgmml_lj));
    em_iter->second.insert(pair<string,double>("mgmml_bond",mgmml_bond));
    em_iter->second.insert(pair<string,double>("mgmml_angle",mgmml_angle));
    em_iter->second.insert(pair<string,double>("mgmml_dihedral",mgmml_dihedral));
    em_iter->second.insert(pair<string,double>("mgmml_improper",mgmml_improper));

    em_iter->second.insert(pair<string,double>("mgmmr_e",mgmmr_e));
    em_iter->second.insert(pair<string,double>("mgmmr_lj",mgmmr_lj));
    em_iter->second.insert(pair<string,double>("mgmmr_bond",mgmmr_bond));
    em_iter->second.insert(pair<string,double>("mgmmr_angle",mgmmr_angle));
    em_iter->second.insert(pair<string,double>("mgmmr_dihedral",mgmmr_dihedral));
    em_iter->second.insert(pair<string,double>("mgmmr_improper",mgmmr_improper));

   }

   // search for minimum ligand-site/model energy instance of each form of the site

   map<string,double> min_gmml;

   for (map<InstanceID,map<string, double> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    double tmpf = 0;
    map<string, double>::const_iterator titr;
    string formname(eitr->first.get_instname());

    // ligand-site terms

    if (molecule.get_use_term_l_e()){
      titr = eitr->second.find(string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_lj()){
      titr = eitr->second.find(string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_bond()){
      titr = eitr->second.find(string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_angle()){
      titr = eitr->second.find(string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_dihedral()){
      titr = eitr->second.find(string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_l_improper()){
      titr = eitr->second.find(string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // ligand-model terms

    if (molecule.get_use_term_ml_e()){
      titr = eitr->second.find(string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_lj()){
     titr = eitr->second.find(string("mgmml_lj"));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
    else
     tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_bond()){
      titr = eitr->second.find(string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_angle()){
      titr = eitr->second.find(string("mgmml_angle"));
      if (titr== eitr->second.end())
       ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
      else
       tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_dihedral()){
      titr = eitr->second.find(string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (molecule.get_use_term_ml_improper()){
      titr = eitr->second.find(string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // create or update minimum energy entry for this form of the site

    map<string,double>::iterator meitr(min_gmml.find(formname));
    if (meitr == min_gmml.end()){
     pair<string,double> tmppair(formname,tmpf);
     min_gmml.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }
   }

   // search for minimum site-site/model electrostatic energy instance of each form of the site

   map<string,double> min_gmm_e;

   for (map<InstanceID,map<string, double> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    double tmpf = 0;
    map<string, double>::const_iterator titr;
    string formname(eitr->first.get_instname());

    // site-site terms

    if (molecule.get_use_term_s_e()){
     titr = eitr->second.find(string("gmmr_e"));
     if (titr == eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // create or update minimum energy entry for this form of the site

    map<string,double>::iterator meitr(min_gmm_e.find(formname));
    if (meitr == min_gmm_e.end()){
     pair<string,double> tmppair(formname,tmpf);
     min_gmm_e.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }

   }

   // now collect all energy terms in the final terms

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    double gmmint = 0;
    double gmmback = 0;
    double gmmsite = 0;
    double gmmmodelback = 0;

    map<InstanceID,map<string, double> >::const_iterator eitr(emap.find(j->get_instid()));
    if (eitr == emap.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint instance missing in energy term map\n");
    map<string, double>::const_iterator titr(eitr->second.find(string("gmmback")));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");

    gmmback = titr->second;

    double min_l = 0;
    double min_se = 0;

    map<string, double>::const_iterator ml_itr(min_gmml.find(j->get_name()));
    if (ml_itr != min_gmml.end()) min_l = ml_itr->second;

    map<string, double>::const_iterator ms_itr(min_gmm_e.find(j->get_name()));
    if (ms_itr != min_gmm_e.end()) min_se = ms_itr->second;

    double gmml = 0;

    // ligand-site terms

    if (molecule.get_use_term_l_e()){
      titr = eitr->second.find(string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_lj()){
      titr = eitr->second.find(string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_bond()){
      titr = eitr->second.find(string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_angle()){
      titr = eitr->second.find(string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_dihedral()){
      titr = eitr->second.find(string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (molecule.get_use_term_l_improper()){
      titr = eitr->second.find(string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    double mgmml = 0;

    // ligand-modelback terms

    if (molecule.get_use_term_ml_e()){
      titr = eitr->second.find(string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_lj()){
      titr = eitr->second.find(string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_bond()){
      titr = eitr->second.find(string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_angle()){
      titr = eitr->second.find(string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_dihedral()){
      titr = eitr->second.find(string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (molecule.get_use_term_ml_improper()){
      titr = eitr->second.find(string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    double gmme = 0;

    // site-site terms

    if (molecule.get_use_term_s_e()){
      titr = eitr->second.find(string("gmmr_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (molecule.get_use_term_s_lj()){
      titr = eitr->second.find(string("gmmr_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (molecule.get_use_term_s_bond()){
      titr = eitr->second.find(string("gmmr_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (molecule.get_use_term_s_angle()){
      titr = eitr->second.find(string("gmmr_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (molecule.get_use_term_s_dihedral()){
      titr = eitr->second.find(string("gmmr_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (molecule.get_use_term_s_improper()){
      titr = eitr->second.find(string("gmmr_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    double mgmme = 0;

    // site-model terms

    if (molecule.get_use_term_ms_e()){
      titr = eitr->second.find(string("mgmmr_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (molecule.get_use_term_ms_lj()){
      titr = eitr->second.find(string("mgmmr_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (molecule.get_use_term_ms_bond()){
      titr = eitr->second.find(string("mgmmr_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (molecule.get_use_term_ms_angle()){
      titr = eitr->second.find(string("mgmmr_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (molecule.get_use_term_ms_dihedral()){
     titr = eitr->second.find(string("mgmmr_dihedral"));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
    else
     mgmme += titr->second;
    }

    if (molecule.get_use_term_ms_improper()){
      titr = eitr->second.find(string("mgmmr_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    // minimum ligand-site interaction should be included in model energy
    // correct for this by considering only relative ligand site interactions within 1 form
    // equivalent for intra site and site-model electrostatics
    gmmsite      = gmme  + gmml - min_l - min_se;
    gmmmodelback = mgmme + mgmml;

    j->set_gmmsite(gmmsite);
    j->set_gmmmodelback(gmmmodelback);
    j->set_gmmback(gmmback);

   }




  }

  if (molecule.get_print_chm()) molecule.print_mmint();

  blab1 << "Exiting GcemCharmmEnergyCalculator::calc_mmint_charmm" << endl;

}

void GcemCharmmEnergyCalculator::calc_mmint_charmm_qm(GcemMolecule &molecule) const
{

  blab2 << "Entering GcemCharmmEnergyCalculator::calc_mmint_charmm_qm" << endl;
  
  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   int stype(molecule.get_rotmap_value(i->get_sid()));
   if (stype != 4){
    continue;
   }
   if (!molecule.chmpara.contains_Topology(i->get_name())){
    cerr << "GcemCharmmEnergyCalculator::calc_mmint_charmm_qm Warning: could not find topology entry in CHARMM topology for site name "
         << i->get_name() << ".\nWill not calculate intrinsic MM energy for this residue."
         << i->get_sid()  << endl;
    continue;
   }
   // retrieve topology
   CharmmParaResTop top(molecule.chmpara.get_Topology(i->get_name()));
   list<AtomID> aidlist;
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   list<string> ligats;
   if (stype == 1 || stype == 3){
    // this is a titratable and flexible site
    // collect all ligand atoms
    GcemSiteDef sdef(molecule.get_sitedef(i->get_name()));
    for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
     for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
      // search atom with zero occupancy
      if (l->second.second == 0){ligats.push_back(l->first.get_atname());}
     }
    }
    ligats.sort();
    ligats.unique();
   }

   map<InstanceID, map<string, double> > emap;
   map<string, double> min_e_map;

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    // all terms involving ligand + site atoms and background - model atoms
    double gmmback = 0;

    // all terms involving ligand atoms and only site atoms
    double gmml = 0;
    double gmml_e = 0;
    double gmml_lj = 0;
    double gmml_bond = 0;
    double gmml_angle = 0;
    double gmml_dihedral = 0;
    double gmml_improper = 0;
    // all terms not involving ligand atoms and only site atoms
    double gmmr = 0;
    double gmmr_e = 0;
    double gmmr_lj = 0;
    double gmmr_bond = 0;
    double gmmr_angle = 0;
    double gmmr_dihedral = 0;
    double gmmr_improper = 0;

    // all terms involving ligand atoms and also model atoms
    double mgmml = 0;
    double mgmml_e = 0;
    double mgmml_lj = 0;
    double mgmml_bond = 0;
    double mgmml_angle = 0;
    double mgmml_dihedral = 0;
    double mgmml_improper = 0;
    // all terms not involving ligand atoms but model_atoms
    double mgmmr = 0;
    double mgmmr_e = 0;
    double mgmmr_lj = 0;
    double mgmmr_bond = 0;
    double mgmmr_angle = 0;
    double mgmmr_dihedral = 0;
    double mgmmr_improper = 0;

    // calculate gmm_back
    if (molecule.get_print_debug()) cout << "GcemCharmmEnergyCalculator::calc_mmint_charmm_qm computing molecule.background terms" << endl;

    for (AtomSet::const_iterator k=j->site_struct.begin(); k!=j->site_struct.end(); ++k){
     if (!molecule.chmpara.contains_attype(CharmmID(k->second)) && !molecule.chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm: attype not found.");
     }
     CharmmParaNB nbk;
     if (molecule.chmpara.contains_NB(CharmmID(k->second))){
      nbk = molecule.chmpara.get_NB(CharmmID(k->second));
     }
     else if (k->second.atname == "OT1" || k->second.atname == "OT2" || k->second.atname == "OY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
     }
     else if (k->second.atname == "CAT" || k->second.atname == "CT" || k->second.atname == "CAY"){
      nbk = molecule.chmpara.get_NB(string("CT3"));
     }
     else if (k->second.atname == "CY"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("C")));
     }
     else if (k->second.atname == "NT"){
      nbk = molecule.chmpara.get_NB(string("NH2"));
     }
     else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT" || k->second.atname == "HY1" || k->second.atname == "HY2" || k->second.atname == "HY3"){
      nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
     }
     else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
      nbk = molecule.chmpara.get_NB(string("H"));
     }
     else {
      cerr << k->second << endl;
      ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm: Lennard Jones parameter missing.");
     }
     for (AtomSet::const_iterator l=molecule.background.begin(); l!=molecule.background.end(); ++l){

      if (!molecule.chmpara.contains_attype(CharmmID(l->second)) && !molecule.chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm: attype not found.");
      }
      // skip model atoms, interactions within site are covered below
      if(j->model_struct.contains(l->first)){continue;}
      // get LJ parameters
      CharmmParaNB nbl;
      // molecule.background could contain unknown atom types
      if (molecule.chmpara.contains_attype(CharmmID(l->second))){
       if (molecule.chmpara.contains_NB(CharmmID(l->second))){
        nbl = molecule.chmpara.get_NB(CharmmID(l->second));
       }
      }
      else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
      }
      else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
      }
      else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
       nbl = molecule.chmpara.get_NB(string("CT3"));
      }
      else if (l->second.atname == "CY"){
       nbl = molecule.chmpara.get_NB(string("C"));
      }
      else if (l->second.atname == "NT"){
       nbl = molecule.chmpara.get_NB(string("NH2"));
      }
      else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
       nbl = molecule.chmpara.get_NB(string("H"));
      }
      // sloppy search for matching atom name in patch residues
      else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
       nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
      }
      else if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname))){
       nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
      }
      else {
       cerr << l->second << endl;
       ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm: Lennard Jones parameter missing.");
      }
      double dist = geom_calc.abs_value(k->second.coord - l->second.coord);
      double ljek = nbk.lje;
      double ljrk = nbk.ljr;
      double ljel = nbl.lje;
      double ljrl = nbl.ljr;
      if (ljek <= 0 && ljel <= 0){
       if ( dist != 0 ){
        double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
        gmmback += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
       } else {
        cerr << k->second << endl;
        cerr << l->second << endl;
        ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm: two atoms have identical coordinates");
       }
      }
      else {
       if (ljek > 0 || ljel > 0){
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm atom has non-negative VDW well depth.");
       }
      }
     }
    }

    if (molecule.get_print_chm()){

     cout << endl << "---------------------------------------" << endl;

     cout << "Energies for instance " << (j->get_instid()).get_fn_prefix() << endl;

     cout << "gmmback "       << gmmback << endl;

     cout << endl << "---------------------------------------" << endl;

    }

    map<InstanceID, map<string, double> >::iterator em_iter = emap.find(j->get_instid());
    // should never happen
    if (em_iter != emap.end()) {
      ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm non-unique InstanceID\n");
    }

    map<string,double> tmpmap;
    tmpmap.insert(pair<string,double>("gmmback",gmmback));
    pair<InstanceID,map<string, double> > tmppair1(j->get_instid(),tmpmap);

    pair<map<InstanceID, map<string, double> >::iterator, bool> tmpair2 = emap.insert(tmppair1);

   }

   // now collect all energy terms in the final terms

   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

    double gmmint = 0;
    double gmmback = 0;
    double gmmsite = 0;
    double gmmmodelback = 0;

    map<InstanceID,map<string, double> >::const_iterator eitr(emap.find(j->get_instid()));
    if (eitr == emap.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm instance missing in energy term map\n");
    map<string, double>::const_iterator titr(eitr->second.find(string("gmmback")));
    if (titr== eitr->second.end())
     ::error("ERROR in GcemCharmmEnergyCalculator::calc_mmint_charmm_qm energy term missing\n");

    gmmback = titr->second;

    j->set_gmmback(gmmback);

   }




  }

  if (molecule.get_print_chm()) molecule.print_mmint();

  blab1 << "Exiting GcemCharmmEnergyCalculator::calc_mmint_charmm_qm" << endl;

}

void GcemCharmmEnergyCalculator::calc_mmg(GcemMolecule &molecule) const
{

  blab2 << "Entering GcemCharmmEnergyCalculator::calc_mmg" << endl;

  for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   list<GcemSite>::const_iterator iplus = i; iplus++;
   for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    for (list<GcemSite>::const_iterator ii=iplus; ii!=molecule.sites.end(); ++ii){
     // dont need interactions among two non-flexible sites
     int stypei(molecule.get_rotmap_value(i->get_sid()));
     int stypeii(molecule.get_rotmap_value(ii->get_sid()));
     if (molecule.get_skip_stat_pair_mmg() && (stypei != 1 && stypei != 2 && stypeii != 1 && stypeii != 2)){continue;}
     for (list<GcemInstance>::const_iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
      double tmpf = 0;
      for (AtomSet::const_iterator k=j->site_struct.begin(); k!=j->site_struct.end(); ++k){
       CharmmParaNB nbk;
       if (molecule.chmpara.contains_attype(k->second)){
        if (molecule.chmpara.contains_NB(CharmmID(k->second))){
         nbk = molecule.chmpara.get_NB(CharmmID(k->second));
        }
       }
       else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3" || k->second.atname == "HNT" || k->second.atname == "HY1" || k->second.atname == "HY2" || k->second.atname == "HY3"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("HN")));
       }
       else if (k->second.atname == "OT1" || k->second.atname == "OT2" || k->second.atname == "OY"){
        nbk = molecule.chmpara.get_NB(molecule.chmpara.get_attype(k->second.resname,string("O")));
       }
       else if (k->second.atname == "CAT" || k->second.atname == "CT" || k->second.atname == "CAY"){
        nbk = molecule.chmpara.get_NB(string("CT3"));
       }
       else if (k->second.atname == "CY"){
        nbk = molecule.chmpara.get_NB(string("C"));
       }
       else if (k->second.atname == "NT"){
        nbk = molecule.chmpara.get_NB(string("NH2"));
       }
       else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
        nbk = molecule.chmpara.get_NB(string("H"));
       }
       else {
        CharmmParaResTop top;
        if (molecule.chmpara.contains_element_NB(top.get_element(k->second.atname)))
         nbk = molecule.chmpara.get_element_NB(top.get_element(k->second.atname));
        else {
         cerr << k->second << endl;
         ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmg: Lennard Jones parameter missing.");
        }
       }
       for (AtomSet::const_iterator l=jj->site_struct.begin(); l!=jj->site_struct.end(); ++l){
        CharmmParaNB nbl;
        if (molecule.chmpara.contains_attype(l->second)){
         if (molecule.chmpara.contains_NB(CharmmID(l->second))){
          nbl = molecule.chmpara.get_NB(CharmmID(l->second));
         }
        }
        else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3" || l->second.atname == "HNT" || l->second.atname == "HY1" || l->second.atname == "HY2" || l->second.atname == "HY3"){
         nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("HN")));
        }
        else if (l->second.atname == "OT1" || l->second.atname == "OT2" || l->second.atname == "OY"){
         nbl = molecule.chmpara.get_NB(molecule.chmpara.get_attype(l->second.resname,string("O")));
        }
        else if (l->second.atname == "CAT" || l->second.atname == "CT" || l->second.atname == "CAY"){
         nbl = molecule.chmpara.get_NB(string("CT3"));
        }
        else if (l->second.atname == "CY"){
         nbl = molecule.chmpara.get_NB(string("C"));
        }
        else if (l->second.atname == "NT"){
         nbl = molecule.chmpara.get_NB(string("NH2"));
        }
        else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
         nbl = molecule.chmpara.get_NB(string("H"));
        }
        // sloppy search for matching atom name in patch residues
        else if (molecule.chmpara.contains_pres_attype(l->second.atname)){
         nbl = molecule.chmpara.get_NB(molecule.chmpara.get_pres_attype(l->second.atname));
        }
        else {
         CharmmParaResTop top;
         if (molecule.chmpara.contains_element_NB(top.get_element(l->second.atname)))
          nbl = molecule.chmpara.get_element_NB(top.get_element(l->second.atname));
         else {
          cerr << l->second << endl;
          ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmg: Lennard Jones parameter missing.");
         }
        }

        double ljek = nbk.lje;
        double ljrk = nbk.ljr;
        double ljel = nbl.lje;
        double ljrl = nbl.ljr;
        Coord dvec(k->second.coord - l->second.coord);
        double distsq = dvec.x * dvec.x + dvec.y * dvec.y + dvec.z * dvec.z;
        if (ljek <= 0 && ljel <= 0){
         if ( distsq != 0 ){
          if (distsq < 100){
           double dist = sqrt(distsq);
           double r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
           tmpf += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
          }
         } else {
          cerr << k->second << endl;
          cerr << l->second << endl;
          ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmg: two atoms have identical coordinates");
         }
        }
        else {
         if (ljek > 0 || ljel > 0){
           ::error ("ERROR in GcemCharmmEnergyCalculator::calc_mmg atom has non-negative VDW well depth.");
         }
        }

       }
      }

      // very small interaction energies are set to zero
      if ((tmpf > 5.0E-05 || tmpf < -5.0E-05) && !molecule.contains_mmg(j->get_instid(),jj->get_instid()))
       molecule.insert_mmg(j->get_instid(),jj->get_instid(), tmpf);

     }
    }
   }
  }

  if (molecule.get_print_chm()) molecule.print_mmg();

  blab1 << "Exiting GcemCharmmEnergyCalculator::calc_mmg" << endl;

}

// GcemCharmmEnergyCalculator.cc ends here
