/*  A form of a GCEM Site

    label, bound ligands, charge and occupancy of atoms

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemFormDef_h
#define _GcemFormDef_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/globals.h"
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/map.hpp>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class GcemFormDef {
public:
  GcemFormDef();
  GcemFormDef (const string& nm);
  GcemFormDef (const string& nm, const float gm, const float er, const float ir);
  GcemFormDef (const GcemFormDef& a);
  ~GcemFormDef () {}
  //!nowrap!+
  GcemFormDef& operator=(const GcemFormDef& a);

  //!nowrap!-

  void print_ligandmap();
  void print_atommap();
  void print_linkmap();

  string get_name() const {return name;}
  float get_gmodel() const {return gmodel;}
  void set_gmodel(const float f){gmodel = f;}
  float get_epsref() const {return epsref;}
  void set_epsref(const float f){epsref = f;}
  float get_istrref() const {return istrref;}
  void set_istrref(const float f){istrref = f;}

  std::map<string,int> ligandmap;
  std::map<CharmmID,pair<float, int> > atommap;
  std::multimap<AtomID,AtomID> linkmap;

  // lookup attype by CharmmID or resname/atname pair

  // check if ligand is in list by ligand label
  bool contains_ligand (const string&) const;
  // check if atom is in list by CharmmID(s) or resname atname pair(s)
  bool contains_atom (const string&, const string&) const;
  bool contains_atom (const CharmmID&) const;
  // check if atom pairs is a link / if atom is a link atom / retrieve bond partner of link atom
  bool contains_link(const AtomID& k, const AtomID& l) const;
  bool is_link_atom(const AtomID& k) const;
  list<AtomID> get_link_atom_partners(const AtomID& k) const;

  // retrieve parameter by CharmmID(s) or attype(s) or resname atname pair(s)
  // retrieve ligand number by ligand label
  pair<float,int> get_atom(const string&, const string&) const;
  pair<float,int> get_atom(const CharmmID&) const;

  int get_ligand (const string&) const;

  // for QM sites

  // insert in list, return true if successfull
  std::pair<std::map<string,int>::iterator,bool> insert_ligand(const string&, const int);
  std::pair<std::map<CharmmID,pair<float,int> >::iterator,bool> insert_atom(const string&, const string&, const float, const int);
  std::pair<std::map<CharmmID,pair<float,int> >::iterator,bool> insert_atom(const CharmmID&, const float, const int);
  std::multimap<AtomID,AtomID>::iterator insert_link(const AtomID&, const AtomID&);

  //!nowrap!-
private:
  friend class boost::serialization::access;
  friend class GcemSiteDef;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & ligandmap;
    ar & atommap;
    ar & linkmap;
    ar & gmodel;
    ar & epsref;
    ar & istrref;
    ar & name;
  }
  string name;
  float gmodel, epsref, istrref;
};

// atom and ligand map functions

inline pair<float,int> GcemFormDef::get_atom(const CharmmID& k) const
{
  std::map<CharmmID,pair<float, int> >::const_iterator i=atommap.find(k);
  if (i == atommap.end())
    ::error("GcemFormDef::get_atom: key not found");
  return i->second;
}

inline pair<float,int> GcemFormDef::get_atom(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return GcemFormDef::get_atom(k);
}

inline int GcemFormDef::get_ligand(const string& k) const
{
  std::map<string,int>::const_iterator i=ligandmap.find(k);
  if (i == ligandmap.end())
    ::error("GcemFormDef::get_ligand: key not found");
  return i->second;
}

// Existence of element (not STL-like

inline bool GcemFormDef::contains_atom(const CharmmID& k) const
{
  std::map<CharmmID,std::pair<float,int> >::const_iterator i=atommap.find(k);
  return i != atommap.end();
}

inline bool GcemFormDef::contains_atom(const string& rn, const string& an) const
{
  CharmmID k(rn, an);
  return GcemFormDef::contains_atom(k);
}

inline bool GcemFormDef::contains_ligand(const string& k) const
{
  std::map<string,int>::const_iterator i=ligandmap.find(k);
  return i != ligandmap.end();
}

inline bool GcemFormDef::contains_link(const AtomID& k, const AtomID& l) const
{
  for (std::multimap<AtomID,AtomID>::const_iterator i=linkmap.begin(); i!=linkmap.end(); i++){
   if ((k == i->first && l == i->second) || (l == i->first && k == i->second)){
    return true;
   }
  }
  return false;
}

inline bool GcemFormDef::is_link_atom(const AtomID& k) const
{
  for (std::multimap<AtomID,AtomID>::const_iterator i=linkmap.begin(); i!=linkmap.end(); i++){
   if (k == i->first || k == i->second){
    return true;
   }
  }
  return false;
}

inline list<AtomID> GcemFormDef::get_link_atom_partners(const AtomID& k) const
{
  list<AtomID> result;
  for (std::multimap<AtomID,AtomID>::const_iterator i=linkmap.begin(); i!=linkmap.end(); i++){
   if (k == i->first){
    result.push_back(i->first);
   } else if (k == i->second){
    result.push_back(i->second);
   }
  }
  return result;
}

// easier to use insert (not quite STL-like)
inline pair<std::map<CharmmID,pair<float,int> >::iterator,bool> GcemFormDef::insert_atom (const string& rn, const string& an, const float q, const int o) {
  CharmmID k(rn, an);
  if (contains_atom(k)) {
    cerr << k << ":" << name << endl;
    ::error("GcemFormDef::insert_atom: attempt to insert a duplicate atom for form\n");
  }
  return atommap.insert(std::pair<CharmmID,pair<float, int> >(k, std::pair<float,int>(q,o)));
}

inline pair<std::map<CharmmID,pair<float,int> >::iterator,bool> GcemFormDef::insert_atom (const CharmmID& k, const float q, const int o){
  if (contains_atom(k)) {
    cerr << k << ":" << name << endl;
    ::error("GcemFormDef::insert_atom: attempt to insert a duplicate attype for atom\n");
  }
  return atommap.insert(std::pair<CharmmID,pair<float,int> >(k, std::pair<float,int>(q,o)));
}

inline pair<std::map<string,int>::iterator,bool> GcemFormDef::insert_ligand (const string& k, const int n){
  if (contains_ligand(k)) {
    cerr << k << ":" << name << endl;
    ::error("GcemFormDef::insert_ligand: attempt to insert a duplicate ligand\n");
  }
  return ligandmap.insert(std::pair<string,int>(k, n));
}

inline std::multimap<AtomID, AtomID>::iterator GcemFormDef::insert_link (const AtomID &aid1, const AtomID &aid2){
  if (contains_link(aid1,aid2)) {
    cerr << aid1 << ":" << aid2 << endl;
    ::error("GcemFormDef::insert_link: attempt to insert a duplicate link\n");
  }
  return linkmap.insert(std::pair<AtomID,AtomID>(aid1, aid2));
}

#endif

// GcemFormDef.h ends here
