/*  
    helper object for GcemMolecule
    provides geometric functions, vector maths

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemGeometryCalculator_h
#define _GcemGeometryCalculator_h 1

#include "MEAD_e/globals.h"
#include "MEAD_e/Coord.h"


//!wrap!
class GcemGeometryCalculator {
public:
  GcemGeometryCalculator();
  ~GcemGeometryCalculator () {}

  // geometry functions
  // i,j,k,l, target torsion angle
  //!nowrap!+
  Coord set_tor(const Coord&, const Coord&, const Coord&, const Coord&, double) const;
  // i,j,k,l, change in torsion angle
  Coord change_tor(const Coord&, const Coord&, const Coord&, const Coord&, double) const;
  // i,j,k,l,
  double get_tor(const Coord&, const Coord&, const Coord&, const Coord&) const;
  // i,j,k
  double get_angle(const Coord&, const Coord&, const Coord&) const;
  // point, point
  double get_dist(const Coord&, const Coord&) const;
  double get_sqdist(const Coord&, const Coord&) const;
  // vector
  double abs_value(const Coord&) const;
  // vector, vector
  double dot_p(const Coord&, const Coord&) const;
  // point, point point
  Coord get_normal(const Coord&, const Coord&, const Coord&) const;
  // vector, vector
  Coord cross_p(const Coord&, const Coord&) const;
  // unit normal_vector of plane, point to project, point in plane
  Coord project_on_plane (const Coord&, const Coord&, const Coord&) const;
  // point, point in plane, normal vector
  double get_dist_p_plane (const Coord&, const Coord&, const Coord&) const;
  // point, point on line, point on line
  double get_dist_p_line (const Coord&, const Coord&, const Coord&) const;
  // get integer torsion bin (rotamer library) from torsion angle
  int get_tor_bin(const double) const;
  //!nowrap!-
};


////////////////////////// geometry functions //////////////////////////

// i,j,k,l, target torsion angle
inline Coord GcemGeometryCalculator::set_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const double t) const
{
  Coord a(k-j);
  Coord b(l-k);
  double d(abs_value(a));
  a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  double tmptor = get_tor(i,j,k,l) - t;
  double cos_t(cos(tmptor));
  double sin_t(sin(tmptor));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l, change in torsion angle
inline Coord GcemGeometryCalculator::change_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l, const double t) const
{
  Coord a(k-j);
  Coord b(l-k);
  double d(abs_value(a));
  if (d==0)
   ::error("Error in GcemGeometryCalculator::get_tor(): points j and k have identical coordinates.");
   a /= d;
  // need inverse sign here since sign of torsion in matrix seems to be reversed
  double cos_t(cos(-t));
  double sin_t(sin(-t));
  Coord A1(((a.x * a.x * (1 - cos_t)) + cos_t),
           ((a.x * a.y * (1 - cos_t)) - (a.z * sin_t)),
           ((a.x * a.z * (1 - cos_t)) + (a.y * sin_t)));
  Coord A2(((a.x * a.y * (1 - cos_t)) + (a.z * sin_t)),
           ((a.y * a.y * (1 - cos_t)) + cos_t),
           ((a.y * a.z * (1 - cos_t)) - (a.x * sin_t)));
  Coord A3(((a.x * a.z * (1 - cos_t)) - (a.y * sin_t)),
           ((a.y * a.z * (1 - cos_t)) + (a.x * sin_t)),
           ((a.z * a.z * (1 - cos_t)) + cos_t));
  Coord r  ((k.x + (A1.x * b.x) + (A2.x * b.y) + (A3.x * b.z)),
            (k.y + (A1.y * b.x) + (A2.y * b.y) + (A3.y * b.z)),
            (k.z + (A1.z * b.x) + (A2.z * b.y) + (A3.z * b.z)));
  return r;
}

// i,j,k,l
inline double GcemGeometryCalculator::get_tor(const Coord& i, const Coord& j, const Coord& k, const Coord& l) const
{
 Coord a(get_normal(i,j,k));
 Coord b(get_normal(j,k,l));
 double dp = dot_p(a,b);
 if      ((1-dp) < 1.0E-14) return 0;
 else if ((1+dp) < 1.0E-14) return pi;
 double t(acos(dp));
 double v(dot_p(k-j,cross_p(a,b)));
 if (v < (double) 0) t *= -1;
 return t;
}

// i,j,k
inline double GcemGeometryCalculator::get_angle(const Coord& i, const Coord& j, const Coord& k) const
{
  Coord a(i-j); a /= abs_value(a);
  Coord b(k-j); b /= abs_value(b);
  return acos(dot_p(a,b));
}

// point, point
inline double GcemGeometryCalculator::get_dist(const Coord& a, const Coord& b) const
{return sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// point, point
inline double GcemGeometryCalculator::get_sqdist(const Coord& a, const Coord& b) const
{return ((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z));}

// vector
inline double GcemGeometryCalculator::abs_value(const Coord& a) const
{return sqrt(a.x*a.x + a.y*a.y + a.z*a.z);}

// vector, vector
inline double GcemGeometryCalculator::dot_p(const Coord& a, const Coord& b) const
{return a.x*b.x + a.y*b.y + a.z*b.z;}

// vector, vector
inline Coord GcemGeometryCalculator::cross_p(const Coord& a, const Coord& b) const
{Coord prod(a.y*b.z - a.z*b.y, b.x*a.z - a.x*b.z, a.x*b.y - b.x*a.y); return prod;}

// point in plane, point in plane, point in plane
inline Coord GcemGeometryCalculator::get_normal (const Coord& i, const Coord& j, const Coord& k) const
{Coord n(cross_p(j-i,k-i)); n /= abs_value(n); return n;}

// unit normal vector of plane, point to project, point in plane
inline Coord GcemGeometryCalculator::project_on_plane (const Coord& n, const Coord& a, const Coord& b) const
{return Coord(a - (n * get_dist_p_plane(a, b, n)));}

// point, point in plane, normal vector
inline double GcemGeometryCalculator::get_dist_p_plane (const Coord& a, const Coord& b, const Coord& n) const
{return dot_p(a-b,n);}

// point, point on line, point on line
inline double GcemGeometryCalculator::get_dist_p_line (const Coord& a, const Coord& b, const Coord& c) const
{return abs_value(cross_p(c-b,b-a)) / abs_value(c-b);}

inline int GcemGeometryCalculator::get_tor_bin(const double z) const
{return static_cast<int>((floor(((z*(180/3.141593)) + 5) / 10) * 10));}

#endif

// GcemGeometryCalculator.h ends here
