/*  A GCEM Instance

    label, forms

    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemInstance_h
#define _GcemInstance_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/globals.h"
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/map.hpp>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class GcemInstance {
public:
  GcemInstance ();
  GcemInstance (const GcemInstance& a);
  GcemInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&);
  GcemInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int);
  GcemInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float);
  GcemInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float, const float);
  GcemInstance (const InstanceID&, const map<string,int>&, const AtomSet&, const AtomSet&, const int, const float, const float, const float);
  ~GcemInstance () {}
  //!nowrap!+
  GcemInstance& operator=(const GcemInstance& a);

  //!nowrap!-
  void print() const;
  void write_struct();
  void write_struct(const string&, const string&);
  void write_model_struct();
  void write_model_struct(const string&);
  void write_site_struct();
  void write_site_struct(const string&);

  string get_name() const {return name;}
  InstanceID get_instid() const {return instid;}
  void set_instid(const InstanceID& id){instid = id;}

  float get_gmodel() const {return gmodel;}
  void set_gmodel(const float f){gmodel = f;}
  float get_epsref() const {return epsref;}
  void set_epsref(const float f){epsref = f;}
  float get_istrref() const {return istrref;}
  void set_istrref(const float f){istrref = f;}

  // sums of MM and electrostatic contributions (without membrane potential contribution)
  float get_gmmint() const {return gmmint;}
  float get_gpbint() const {return gpbint;}
  float get_gmmint_ref() const {return gmmint_ref;}
  void set_gmmint_ref(const float f){gmmint_ref = f;}

  // electrostatic contributions (ref refers to reference conformer herein)
  float get_gpbvmemb() const {return gpbvmemb;}
  void set_gpbvmemb(const float f){gpbvmemb = f;}
  float get_gpbborn_ffe_ref() const {return gpbborn_ffe_ref;}
  void set_gpbborn_ffe_ref(const float f){gpbborn_ffe_ref = f; calculate_gpbint();}
  float get_gpbborn_prot_ref() const {return gpbborn_prot_ref;}
  void set_gpbborn_prot_ref(const float f){gpbborn_prot_ref = f; calculate_gpbint();}
  float get_gpbborn_sol_ref() const {return gpbborn_sol_ref;}
  void set_gpbborn_sol_ref(const float f){gpbborn_sol_ref = f; calculate_gpbint();}
  float get_gpbborn_ffe() const {return gpbborn_ffe;}
  void set_gpbborn_ffe(const float f){gpbborn_ffe = f; calculate_gpbint();}
  float get_gpbborn_prot() const {return gpbborn_prot;}
  void set_gpbborn_prot(const float f){gpbborn_prot = f; calculate_gpbint();}
  float get_gpbborn_sol() const {return gpbborn_sol;}
  void set_gpbborn_sol(const float f){gpbborn_sol = f; calculate_gpbint();}
  float get_gpbback_prot() const {return gpbback_prot;}
  void set_gpbback_prot(const float f){gpbback_prot = f; calculate_gpbint();}
  float get_gpbback_sol() const {return gpbback_sol;}
  void set_gpbback_sol(const float f){gpbback_sol = f; calculate_gpbint();}

  // transfer solvent (reference environment) -> receptor:
  //  (gpbborn_prot - gpbborn_sol + gpbback_prot - gpbback_sol reference)
  // transfer force field environment -> solvent (reference environment) relative to reference conformer
  //  add missing conformation dependence to model energies of MM sites
  //  (QM sites have all the terms appearing in the expression set to zero):
  //  ((gpbborn_sol - gpbborn_sol_ref) - (gpbborn_ffe - gpbborn_ffe_ref))
  inline void calculate_gpbint(){gpbint = gpbborn_prot - gpbborn_sol + gpbback_prot - gpbback_sol + ((gpbborn_sol - gpbborn_sol_ref) - (gpbborn_ffe - gpbborn_ffe_ref));}

  // MM contributions
  float get_gmmmodelback() const {return gmmmodelback;}
  void set_gmmmodelback(const float f){gmmmodelback = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_gmmsite() const {return gmmsite;}
  void set_gmmsite(const float f){gmmsite = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_gmmback() const {return gmmback;}
  void set_gmmback(const float f){gmmback = f; gmmint = gmmsite + gmmmodelback + gmmback;}
  float get_propensity() const {return propensity;}

  // empirical energy function instead of Force Field derived one
  // rotamer propensities from backbone dependent rotamer library
  void set_propensity(const float f){propensity = f;}
  bool get_has_propensity() const {return has_propensity;}
  void set_has_propensity(const bool f){has_propensity = f;}
  bool get_is_predef() const {return is_predef;}
  void set_is_predef(const bool f){is_predef = f;}

  // each conformer can have multiple charge/binding forms
  int get_confnum() const {return confnum;}
  void set_confnum(const int i){confnum = i;}
  int get_confnum_ref() const {return confnum_ref;}
  void set_confnum_ref(const int i){confnum_ref = i;}

  AtomSet model_struct;
  AtomSet site_struct;
  map<string,int> ligandmap;

  // lookup attype by CharmmID or resname/atname pair

  // check by label if form or ligand are already listed
  bool contains_ligand (const string&) const;
  // return number of bound ligands of a certain type
  int ligands_bound (const string&) const;

  bool operator== (const GcemInstance& a) const {return instid == a.get_instid();}
  bool operator!= (const GcemInstance& a) const {return instid != a.get_instid();}
  bool operator<  (const GcemInstance& a) const {return instid <  a.get_instid();}
  bool operator>  (const GcemInstance& a) const {return instid >  a.get_instid();}

  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & gpbint;
    ar & gpbborn_prot_ref;
    ar & gpbborn_sol_ref;
    ar & gpbborn_ffe_ref;
    ar & gpbborn_ffe;
    ar & gpbborn_prot;
    ar & gpbborn_sol;
    ar & gpbback_prot;
    ar & gpbback_sol;
    ar & gpbvmemb;
    ar & gmmint_ref;
    ar & gmmint;
    ar & gmmback;
    ar & gmmsite;
    ar & gmmmodelback;
    ar & gmodel;
    ar & epsref;
    ar & istrref;
    ar & propensity;
    ar & has_propensity;
    ar & is_predef;
    ar & model_struct;
    ar & site_struct;
    ar & ligandmap;
    ar & confnum;
    ar & confnum_ref;
    ar & instid;
    ar & sitename;
    ar & name;
  }
  string sitename;
  string name;
  InstanceID instid;
  // total intrinsic PB-energy without membrane potential contribution
  float gpbint;
  // born energies and intrinsic MM energy of reference instance
  // (same form, minimum (gpbborn_sol - gpbborn_ffe + gmmint))
  float gpbborn_prot_ref;
  float gpbborn_sol_ref;
  float gpbborn_ffe_ref;
  float gmmint_ref;
  // born energy in reference environment used during force field parametrization
  float gpbborn_ffe;
  // born and background energy in proteins and solution
  float gpbborn_prot;
  float gpbborn_sol;
  float gpbback_prot;
  float gpbback_sol;
  // membrane potential contribution to intrinsic energy normalized to 1 V
  float gpbvmemb;
  // total intrinsic MM-energy
  float gmmint;
  // interactions (site+ligand)-(background without background-model)
  float gmmback;
  // interactions (site+ligand)-(site+ligand)
  float gmmsite;
  // interactions (site+ligand)-(background-model)
  float gmmmodelback;
  float gmodel;
  float epsref;
  float istrref;
  float propensity;
  bool has_propensity;
  bool is_predef;
  int confnum;
  int confnum_ref;
};

// form list functions

inline bool GcemInstance::contains_ligand(const string& k) const
{
  std::map<string,int>::const_iterator j(ligandmap.find(k));
  return j!=ligandmap.end();
}

inline int GcemInstance::ligands_bound(const string& k) const
{
  int result;
  std::map<string,int>::const_iterator j(ligandmap.find(k));
  if (j!=ligandmap.end())
   result = j->second;
  else
   result = 0;
  return result;
}

#endif

// GcemInstance.h ends here
