/*
    helper object for GcemInstanceFilter
    creates Gcemmolecule.sites from GcemSiteDefs and AtomSets

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemInstanceFilter.h"
#include "MEAD_e/globals.h"
#include <stdio.h>
#include <fstream>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <sstream>
#include <iostream>
#include <string>

using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::stringstream;
using std::string;

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

// GMCT instance index in char --> only 255 possible, remove complete conformer instances with highest Gint
int GcemInstanceFilter::cap_instances(const int cap, GcemMolecule &molecule) const
{
  if (cap < 1) return 0;

  blab2 << "Entering GcemInstanceFilter::cap_instances" << endl;
  int total   = 0;
  int removed = 0;
  int remconf = 0;
  int max_nin = 0;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gint  = numeric_limits<double>::max();
   double this_gint = numeric_limits<double>::max();
   double tmpsum    = 0;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    tmpsum = j->get_gmmint() + j->get_gpbint();
    for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
     if (i==ii) continue;
     double mininter  = numeric_limits<double>::max();
     double thisinter = 0;
     for (list<GcemInstance>::iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
      thisinter = molecule.get_mmg(j->get_instid(),jj->get_instid()) + molecule.get_pbg(j->get_instid(),jj->get_instid());
      if (thisinter < mininter)
       mininter = thisinter;
     }
     tmpsum += mininter;
    }
    if (tmpsum < this_gint) this_gint= tmpsum;
   }

   // collect conformers in map keyed by confnum
   map<int,list<GcemInstance> > confmap;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    total++;
    map<int,list<GcemInstance> >::iterator citer=confmap.find(j->get_confnum());
    if (citer!=confmap.end())
     citer->second.push_back(*j);
    else {
     list<GcemInstance> tl; tl.push_back(*j);
     confmap.insert(map<int,list<GcemInstance> >::value_type(j->get_confnum(),tl));
    }
   }
   map<double,int> conf_gintmap;
   for (map<int,list<GcemInstance> >::const_iterator j=confmap.begin(); j!=confmap.end(); ++j){
    // find minimum interaction of this conformer
    double this_gint = numeric_limits<double>::max();
    double tmpsum    = 0;
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
     tmpsum = li->get_gmmint() + li->get_gpbint();
     for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
      if (i==ii) continue;
      double mininter  = numeric_limits<double>::max();
      double thisinter = 0;
      for (list<GcemInstance>::iterator jj=i->instances.begin(); jj!=i->instances.end(); ++jj){
       thisinter = molecule.get_mmg(li->get_instid(),jj->get_instid()) + molecule.get_pbg(li->get_instid(),jj->get_instid());
       if (thisinter < mininter)
        mininter = thisinter;
      }
      tmpsum += mininter;
     }
     if (tmpsum < this_gint) this_gint = tmpsum;
    }
    conf_gintmap.insert(pair<double,int>(this_gint, j->first));
   }

   // successively remove conformers with highest gint until ninst <= cap
   while (i->get_ninst() > cap){
    list<InstanceID> to_delete;
    map<double,int>::reverse_iterator cgmi = conf_gintmap.rbegin();
    map<int,list<GcemInstance> >::const_iterator j=confmap.find(cgmi->second);
    // should never happen
    if (j==confmap.end()){
     cerr << "Error in GcemInstanceFilter::cap_instances(): Could not not find conformer." << endl;
     exit(-1);
    }
    // collect all instances belonging to this conformer
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end(); li++){
      to_delete.push_back(li->get_instid());
    }
    // remove collected instances
    remconf++;
    for (list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
     removed++;
     i->delete_instance(*k);
    }
    conf_gintmap.erase(cgmi->first);
   }
   if (i->get_ninst() > max_nin) max_nin = i->get_ninst();

  }

  blab2 << "GcemInstanceFilter::cap_instances: removed " << removed
        << " of " << total << " total instances comprising " << remconf << " site conformers. "
        << (total - removed) << " instances remain. Maximum number of instances of a site: "
        << max_nin << endl;

  blab1 << "Exiting GcemInstanceFilter::cap_instances" << endl;

  return removed;
}

void GcemInstanceFilter::filter_instances(const bool keep_predef, GcemMolecule &molecule) const
{

  blab2 << "Entering GcemInstanceFilter::filter_instances" << endl;

  int total   = 0;
  int removed = 0;
  int remconf = 0;
  int max_nin = 0;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gmmint = numeric_limits<double>::max();
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    if (j->get_gmmint() < min_gmmint){
     min_gmmint = j->get_gmmint();
    }
   }
   list<InstanceID> to_delete;
   // collect conformers in map keyed by confnum
   map<int,list<GcemInstance> > confmap;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    total++;
    map<int,list<GcemInstance> >::iterator citer=confmap.find(j->get_confnum());
    if (citer!=confmap.end())
     citer->second.push_back(*j);
    else {
     list<GcemInstance> tl; tl.push_back(*j);
     confmap.insert(map<int,list<GcemInstance> >::value_type(j->get_confnum(),tl));
    }
   }
   int nempty = 0;
   for (map<int,list<GcemInstance> >::const_iterator j=confmap.begin(); j!=confmap.end(); ++j){
    // find minimum gmmint of this conformer
    double this_gmmint = numeric_limits<double>::max();
    // keep only one instance named (PD)EMPTY/empty
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
     if (li->get_gmmint()<this_gmmint) this_gmmint = li->get_gmmint();
     if (li->get_name() == string("EMPTY")   ||
         li->get_name() == string("empty")   ||
         li->get_name() == string("PDEMPTY") ||
         li->get_name() == string("PDempty")){
      // need only one empty instance
      if (nempty > 0){
       to_delete.push_back(li->get_instid());
       cerr << "WARNING will remove ocurrence " << (nempty + 1) << " of an instance called EMPTY/empty "
            << "of site " << i->get_sid() << " with InstanceID " << li->get_instid() << endl;
      }
      nempty++;
     }
    }
    if ((this_gmmint - min_gmmint) > molecule.get_filter_cutoff()){
     remconf++;
     for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
      if (li->get_is_predef()){
       remconf--;
       break;
      }
      to_delete.push_back(li->get_instid());
     }
    }
   }
   for (list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
    removed++;
    i->delete_instance(*k);
   }
   i->renumber_instances();
   if (i->get_ninst() > max_nin) max_nin = i->get_ninst();
  }

  blab2 << "GcemInstanceFilter::filter_instances: removed " << removed
        << " of " << total << " total instances comprising " << remconf << " site conformers. "
        << (total - removed) << " instances remain. Maximum number of instances of a site: "
        << max_nin << endl;

  blab1 << "Exiting GcemInstanceFilter::filter_instances" << endl;

}

void GcemInstanceFilter::filter_instances_gs(const bool keep_predef, GcemMolecule &molecule) const
{

  blab2 << "Entering GcemInstanceFilter::filter_instances_gs" << endl;

  int total   = 0;
  int removed = 0;
  int remconf = 0;
  int max_nin = 0;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gint  = numeric_limits<double>::max();
   double this_gint = numeric_limits<double>::max();
   double tmpsum    = 0;
   list<GcemInstance>::iterator inst_ref = i->instances.begin();
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    tmpsum = j->get_gmmint();
    for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
     if (i==ii) continue;
     double mininter  = numeric_limits<double>::max();
     double thisinter = 0;
     for (list<GcemInstance>::iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
      thisinter = molecule.get_mmg(j->get_instid(),jj->get_instid());
      if (thisinter < mininter){
       mininter = thisinter;
      }
     }
     tmpsum += mininter;
    }
    if (tmpsum < min_gint){
     min_gint= tmpsum;
     inst_ref = j;
    }
   }
   list<InstanceID> to_delete;
   // collect conformers in map keyed by confnum
   map<int,list<GcemInstance> > confmap;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    total++;
    map<int,list<GcemInstance> >::iterator citer=confmap.find(j->get_confnum());
    if (citer!=confmap.end())
     citer->second.push_back(*j);
    else {
     list<GcemInstance> tl; tl.push_back(*j);
     confmap.insert(map<int,list<GcemInstance> >::value_type(j->get_confnum(),tl));
    }
   }
   for (map<int,list<GcemInstance> >::const_iterator j=confmap.begin(); j!=confmap.end(); ++j){
    // find minimum energy of any instance of this conformer relative to reference instance
    double this_rel_ene = numeric_limits<double>::max();
    double tmpsum    = 0;
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
     // intrinsic energy relative to reference instance !
     tmpsum = li->get_gmmint() - inst_ref->get_gmmint();
     for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
      if (i==ii) continue;
      double rel_mininter  = numeric_limits<double>::max();
      double this_rel_inter = 0;
      for (list<GcemInstance>::iterator jj=i->instances.begin(); jj!=i->instances.end(); ++jj){
       // interaction energy relative to reference instance !
       this_rel_inter = molecule.get_mmg(li->get_instid(),jj->get_instid())
                      - molecule.get_mmg(inst_ref->get_instid(),jj->get_instid());
       if (this_rel_inter < rel_mininter)
        rel_mininter = this_rel_inter;
      }
      tmpsum += rel_mininter;
     }
     if (tmpsum < this_rel_ene) this_rel_ene = tmpsum;
    }
    if (this_rel_ene > molecule.get_filter_cutoff_gs()){
     remconf++;
     for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
      if (li->get_is_predef()){
       remconf--;
       break;
      }
      to_delete.push_back(li->get_instid());
     }
    }
   }
   for (list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
    removed++;
    i->delete_instance(*k);
   }
   if (i->get_ninst() > max_nin) max_nin = i->get_ninst();
  }


  blab2 << "GcemInstanceFilter::filter_instances_gs: removed " << removed
        << " of " << total << " total instances comprising " << remconf << " site conformers. "
        << (total - removed) << " instances remain. Maximum number of instances of a site: "
        << max_nin << endl;

  blab1 << "Exiting GcemInstanceFilter::filter_instances_gs" << endl;

}

// with electrostatic terms

int GcemInstanceFilter::filter_instances_full(const bool keep_predef, GcemMolecule &molecule) const
{

  blab2 << "Entering GcemInstanceFilter::filter_instances_full" << endl;

  int total   = 0;
  int removed = 0;
  int remconf = 0;
  int max_nin = 0;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gint = numeric_limits<double>::max();
   double tmpd     = 0;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    tmpd = j->get_gmmint() + j->get_gpbint();
    if (tmpd < min_gint) min_gint = tmpd;
   }
   list<InstanceID> to_delete;
   // collect conformers in map keyed by confnum
   map<int,list<GcemInstance> > confmap;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    total++;
    map<int,list<GcemInstance> >::iterator citer=confmap.find(j->get_confnum());
    if (citer!=confmap.end())
     citer->second.push_back(*j);
    else {
     list<GcemInstance> tl; tl.push_back(*j);
     confmap.insert(map<int,list<GcemInstance> >::value_type(j->get_confnum(),tl));
    }
   }
   int nempty = 0;
   for (map<int,list<GcemInstance> >::const_iterator j=confmap.begin(); j!=confmap.end(); ++j){
    // find minimum gmmint of this conformer
    double this_gint = numeric_limits<double>::max();
    double tmpd      = 0;
    // keep only one instance named (PD)EMPTY/empty
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
     tmpd = li->get_gmmint() + li->get_gpbint();
     if (tmpd < this_gint) this_gint = tmpd;
     if (li->get_name() == string("EMPTY")   ||
         li->get_name() == string("empty")   ||
         li->get_name() == string("PDEMPTY") ||
         li->get_name() == string("PDempty")){
      // need only one empty instance
      if (nempty > 0){
       to_delete.push_back(li->get_instid());
       cerr << "WARNING will remove ocurrence " << (nempty + 1) << " of an instance called EMPTY/empty "
            << "of site " << i->get_sid() << " with InstanceID " << li->get_instid() << endl;
      }
      nempty++;
     }
    }
    if ((this_gint - min_gint) > molecule.get_filter_cutoff()){
     remconf++;
     for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
      if (li->get_is_predef()){
       remconf--;
       break;
      }
      to_delete.push_back(li->get_instid());
     }
    }
   }
   for (list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
    removed++;
    i->delete_instance(*k);
   }
   if (i->get_ninst() > max_nin) max_nin = i->get_ninst();
  }

  blab2 << "GcemInstanceFilter::filter_instances_full: removed " << removed
        << " of " << total << " total instances comprising " << remconf << " site conformers. "
        << (total - removed) << " instances remain. Maximum number of instances of a site: "
        << max_nin << endl;

  blab1 << "Exiting GcemInstanceFilter::filter_instances_full" << endl;

  return removed;
}

int GcemInstanceFilter::filter_instances_gs_full(const bool keep_predef, GcemMolecule &molecule) const
{

  blab2 << "Entering GcemInstanceFilter::filter_instances_gs_full" << endl;

  int total   = 0;
  int removed = 0;
  int remconf = 0;
  int max_nin = 0;

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gint  = numeric_limits<double>::max();
   double this_gint = numeric_limits<double>::max();
   double tmpsum    = 0;
   list<GcemInstance>::iterator inst_ref = i->instances.begin();
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    tmpsum = j->get_gmmint() + j->get_gpbint();
    for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
     if (i==ii) continue;
     double mininter  = numeric_limits<double>::max();
     double thisinter = 0;
     for (list<GcemInstance>::iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
      thisinter = molecule.get_mmg(j->get_instid(),jj->get_instid()) + molecule.get_pbg(j->get_instid(),jj->get_instid());
      if (thisinter < mininter){
       mininter = thisinter;
      }
     }
     tmpsum += mininter;
    }
    if (tmpsum < min_gint){
     min_gint= tmpsum;
     inst_ref = j;
    }
   }
   list<InstanceID> to_delete;
   // collect conformers in map keyed by confnum
   map<int,list<GcemInstance> > confmap;
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    total++;
    map<int,list<GcemInstance> >::iterator citer=confmap.find(j->get_confnum());
    if (citer!=confmap.end())
     citer->second.push_back(*j);
    else {
     list<GcemInstance> tl; tl.push_back(*j);
     confmap.insert(map<int,list<GcemInstance> >::value_type(j->get_confnum(),tl));
    }
   }
   for (map<int,list<GcemInstance> >::const_iterator j=confmap.begin(); j!=confmap.end(); ++j){
    // find minimum energy of any instance of this conformer relative to reference instance
    double this_rel_ene = numeric_limits<double>::max();
    double tmpsum    = 0;
    for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
     // intrinsic energy relative to reference instance !
     tmpsum = li->get_gmmint()
            + li->get_gpbint()
            - inst_ref->get_gmmint()
            - inst_ref->get_gpbint();
     for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
      if (i==ii) continue;
      double rel_mininter  = numeric_limits<double>::max();
      double this_rel_inter = 0;
      for (list<GcemInstance>::iterator jj=i->instances.begin(); jj!=i->instances.end(); ++jj){
       // interaction energy relative to reference instance !
       this_rel_inter = molecule.get_mmg(li->get_instid(),jj->get_instid())
                      + molecule.get_pbg(li->get_instid(),jj->get_instid())
                      - molecule.get_mmg(inst_ref->get_instid(),jj->get_instid())
                      - molecule.get_pbg(inst_ref->get_instid(),jj->get_instid());
       if (this_rel_inter < rel_mininter)
        rel_mininter = this_rel_inter;
      }
      tmpsum += rel_mininter;
     }
     if (tmpsum < this_rel_ene) this_rel_ene = tmpsum;
    }
    if (this_rel_ene > molecule.get_filter_cutoff_gs()){
     remconf++;
     for (list<GcemInstance>::const_iterator li=j->second.begin(); li!=j->second.end();li++){
      if (li->get_is_predef()){
       remconf--;
       break;
      }
      to_delete.push_back(li->get_instid());
     }
    }
   }
   for (list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
    removed++;
    i->delete_instance(*k);
   }
   if (i->get_ninst() > max_nin) max_nin = i->get_ninst();
  }

  blab2 << "GcemInstanceFilter::filter_instances_gs_full: removed " << removed
        << " of " << total << " total instances comprising "  << remconf << " site conformers. "
        << (total - removed) << " instances remain. Maximum number of instances of a site: "
        << max_nin << endl;

  blab1 << "Exiting GcemInstanceFilter::filter_instances_gs_full" << endl;

  return removed;
}

/*
FIXME : this is under construction, purpose get reference instances with lowest energy, solve otherwise?
*/
/*
void GcemInstanceFilter::get_InstanceIDs_min_gs(map<SiteID,InstanceID> &ref_insts, GcemMolecule &molecule) const
{

  blab2 << "Entering GcemInstanceFilter::get_InstanceIDs_min_gs" << endl;

  ref_insts.clear();

  for (list<GcemSite>::iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   if (i->get_nconf()==1){
    total += i->get_ninst();
    continue;
   }
   double min_gint  = numeric_limits<double>::max();
   double this_gint = numeric_limits<double>::max();
   double tmpsum    = 0;
   list<GcemInstance>::iterator inst_ref = i->instances.begin();
   for (list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    tmpsum = j->get_gmmint() + j->get_gpbint();
    for (list<GcemSite>::iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
     if (i==ii) continue;
     double mininter  = numeric_limits<double>::max();
     double thisinter = 0;
     for (list<GcemInstance>::iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
      thisinter = molecule.get_mmg(j->get_instid(),jj->get_instid()) + molecule.get_pbg(j->get_instid(),jj->get_instid());
      if (thisinter < mininter){
       mininter = thisinter;
      }
     }
     tmpsum += mininter;
    }
    if (tmpsum < min_gint){
     min_gint= tmpsum;
     inst_ref = j;
    }
    ref_insts.insert(s);
   }
  }

  blab1 << "Exiting GcemInstanceFilter::get_InstanceIDs_min_gs" << endl;

  return removed;
}
*/

// GcemInstanceFilter.cc ends here
