/*   A GCEM molecule containing titratable and/or flexible sites

    uses Rotlib object for defining sidechain conformations
    uses force field parameters from GcemMolecule object
    uses initial geometry from an AtomSet

    reads .rot file to determine sites and site types
    reads .est files for description of titratable sites
    reads .qst files for description of QM sites

    calculates non-electrostatic contributions to intrinsic and
    interaction energies

    based purely on CHARMM energy function or in combination of
    knowledge based potential with CHARMM energy function
    
    writes input and job script for electrostatic potential
    solvers for the calculation of electrostatic contributions
    to intrinsic and interaction energies
    
    reads output of electrostatic potential solvers

    writes energy matrices to files, e.g., for GMCT

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemMolecule_h
#define _GcemMolecule_h 1

/* physical and other constants plus definitions */
#include "MEAD_e/globals.h"
#include "MEAD_e/PhysCond.h"
/* elementary objects to handle structures */
#include "MEAD_e/Coord.h"
#include "MEAD_e/AtomSet.h"
/* GCEM specific data storage objects */
#include "MEAD_e/GcemSiteDef.h"
#include "MEAD_e/GcemSite.h"
#include "MEAD_e/CharmmPara.h"
#include "MEAD_e/Rotlib.h"
/* GCEM specific object identifiers */
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/SiteID.h"
#include "MEAD_e/InstanceID.h"
/* these helper objects provide work functions for GcemMolecule */
#include "MEAD_e/GcemSiteCreator.h"
#include "MEAD_e/GcemGeometryCalculator.h"
#include "MEAD_e/GcemCharmmEnergyCalculator.h"
#include "MEAD_e/GcemInstanceFilter.h"
#include "MEAD_e/GcemSolverInput.h"
#include "MEAD_e/GcemSolverOutput.h"
/* STL containers and functions */
#include <vector>
#include <set>
#include <map>
#include <list>
#include <string>
#include <cmath>
#include <limits>
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <fstream>  // file stream
#include <iostream> // ::std::cout
#include <ostream>  // ::std::ostream (<<)
#include <iomanip>  // ::std::setw, ::std::setfill
#include <ios>      // ::std::left, ::std::right, stream status bits
/* Boost serialization library */
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/vector.hpp>
/* Boost regex library */
#include <boost/regex.hpp>
/* Boost filesystem library */
#include <boost/filesystem.hpp>

using std::vector;
using std::list;
using std::set;
using std::map;
using std::pair;
using std::string;
using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::stringstream;

class GcemSiteCreator;
class GcemCharmmEnergyCalculator;
class GcemInstanceFilter;
class GcemSolverInput;
class GcemSolverOutput;

//!wrap!
class GcemMolecule {
public:
  GcemMolecule();
  GcemMolecule (const string&);
  GcemMolecule (const GcemMolecule&);
  GcemMolecule (const string&, const string&, const string&, const string&);
  GcemMolecule (const string&, const string&, const string&, const string&, const bool);
  GcemMolecule (const string&, const string&, const string&, const string&, const double);
  GcemMolecule (const string&, const string&, const string&, const string&, const double, const bool);
  ~GcemMolecule () {}
  //!nowrap!+
  GcemMolecule& operator=(const GcemMolecule& a);

  // set default behaviour and variables
  void set_default();

  // these subroutines control the work flow
  // read data
  // create structures of background and instances of sites,
  // compute MM background and interaction energies
  // write structure files and GCEM input file
  // the master subroutine
  void prepare_gcem();
  void prepare_gcem(const string&, const string&, const string&, const string&, const bool);
  void prepare_filtered_gcem();
  void prepare_filtered_gcem(const int cap_max_ninst, const bool);
  // reread prepared GCEM molecule from binary restart file,
  // read electrostatic energies from MEAD ourtput files
  // sum energy contributions
  // write GMCT compatible input
  void postproc_gcem();
  // read site definitions for titratable sites from .est files
  // create instances from rotamers and or charge forms of sites
  void create_sites();
  /*
  // round all site coordinates to 3 decimal places for
  // comparison of energies to CHARMM calculations with the written structures
  void round_coordinates();
  */
  // calculate MM part of background and interaction energies
  // use knowledge based potential for intrinsic energies or use pure CHARMM
  // call writing of matrices to file
  void calc_mmint();
  void calc_mmg();
  // write MM energy matrices to files .mmint and .mmg
  void write_mmint(const string&);
  void write_mmg(const string&);
  // write energy matrices to files .gint .vmemb and .inter
  void write_gmct_input();
  void write_gint(const string&);
  void write_vmemb(const string&);
  void write_inter(const string&);
  void write_conf(const string&);
  void write_sum(const string&);
  // print energy matrices
  void print_mmint() const;
  void print_mmg() const;

  // helper function
  list<string> get_ligands() const;

  // read .rot file to find out which sites are and how to treat --> rotmap
  void read_rot();
  void read_rot(const string&);
  // read .con file to find out connectivity
  void read_con();
  void read_con(const string&);
  // read output files of MEAD electrostatics solvers
  void read_mead_out();
  // read .est files
  void read_site_defs();

  void write_restart_file(const string&) const;
  void  read_restart_file(const string&);

  // list functions
  bool in_list(const string&, const list<string>&) const;

  // get/set variables determining properties or behaviour of the object
  string get_meadpath() const {return meadpath;}
  void set_meadpath(const string& n) {meadpath = n;}
  string get_eps1set() const {return eps1set;}
  void set_eps1set(const string& n) {eps1set = n;}
  string get_eps2set() const {return eps2set;}
  void set_eps2set(const string& n) {eps2set = n;}
  string get_eps3set() const {return eps3set;}
  void set_eps3set(const string& n) {eps3set = n;}
  string get_channelset() const {return channelset;}
  void set_channelset(const string& n) {channelset = n;}
  string get_elycavset() const {return elycavset;}
  void set_elycavset(const string& n) {elycavset = n;}
  string get_name() const {return name;}
  void set_name(const string& n) {name = n;}
  string get_gcemdir() const {return gcemdir;}
  void set_gcemdir(const string& n) {gcemdir = n;}
  // query/set usage of membrane
  bool get_has_memb() const {return has_memb;}
  void set_has_memb(const bool n) {has_memb = n;}
  // query/set usage of separate potentials for the membrane sides
  bool get_has_con() const {return has_con;}
  void set_has_con(const bool n) {has_con = n;}
  // include capacitance in membrane potential calculation
  bool get_do_cap() const {return do_cap;}
  void set_do_cap(const bool n) {do_cap = n;}
  // query/set post-processing mode
  bool get_do_postproc() const {return do_postproc;}
  void set_do_postproc(const bool n) {do_postproc = n;}
  // query/set usage of knowledge based potential
  bool get_use_empirical() const {return use_empirical;}
  void set_use_empirical(const bool n) {use_empirical = n;}
  // query/set usage of intrinsic energy filter
  bool get_use_filter() const {return use_filter;}
  void set_use_filter(const bool n) {use_filter = n;}
  // print separate energy terms
  bool get_print_chm() const {return print_chm;}
  void set_print_chm(const bool n) {print_chm = n;}
  bool get_print_debug() const {return print_debug;}
  void set_print_debug(const bool n) {print_debug = n; if (n) print_chm = n;}
  // scaling factor for coulomb interaction between atoms in a 1-4 interaction / dihedral
  double get_coulomb_14_scalefac() const {return coulomb_14_scalefac;}
  void set_coulomb_14_scalefac (const double f) {coulomb_14_scalefac = f;}
  // get/set energy cutoff for intrinsic energies
  double get_filter_cutoff() const {return filter_cutoff;}
  void set_filter_cutoff (const double f) {filter_cutoff = f;}
  double get_filter_cutoff_gs() const {return filter_cutoff_gs;}
  void set_filter_cutoff_gs (const double f) {filter_cutoff_gs = f;}
  double get_capacitance() const {return capacitance;}
  void set_capacitance (const double f) {capacitance = f;}
  double get_vmemb_inter() const {return vmemb_inter;}
  void set_vmemb_inter (const double f) {vmemb_inter = f;}
  double get_zmaxh() const {return zmaxh;}
  void set_zmaxh (const double f) {zmaxh = f; has_memb = true;}
  double get_zmaxc() const {return zmaxc;}
  void set_zmaxc (const double f) {zmaxc = f; has_memb = true;}
  double get_zminc() const {return zminc;}
  void set_zminc (const double f) {zminc = f; has_memb = true;}
  double get_zminh() const {return zminh;}
  void set_zminh (const double f) {zminh = f; has_memb = true;}
  double get_membholer() const {return membholer;}
  void set_membholer (const double f) {membholer = f; has_memb = true;}
  double get_membholex() const {return membholex;}
  void set_membholex (const double f) {membholex = f; has_memb = true;}
  double get_membholey() const {return membholey;}
  void set_membholey (const double f) {membholey = f; has_memb = true;}
  double get_ionicstr() const {return ionicstr;}
  void set_ionicstr (const double f) {ionicstr = f;}
  double get_ionicstr1() const {return ionicstr1;}
  void set_ionicstr1 (const double f) {ionicstr1 = f; has_memb = true;}
  double get_ionicstr2() const {return ionicstr2;}
  void set_ionicstr2 (const double f) {ionicstr2 = f; has_memb = true;}
  double get_eps_ff() const {return eps_ff;}
  void set_eps_ff (const double f) {eps_ff = f;}
  double get_epsin1() const {return epsin1;}
  void set_epsin1 (const double f) {epsin1 = f;}
  double get_epsin2() const {return epsin2;}
  void set_epsin2 (const double f) {epsin2 = f;}
  double get_epsin3() const {return epsin3;}
  void set_epsin3 (const double f) {epsin3 = f;}
  double get_epsext() const {return epsext;}
  void set_epsext (const double f) {epsext = f;}
  double get_epscore() const {return epscore;}
  void set_epscore (const double f) {epscore = f; has_memb = true;}
  double get_epshead() const {return epshead;}
  void set_epshead (const double f) {epshead = f; has_memb = true;}
  int get_inside() const {return inside;}
  void set_inside (const int i) {inside = i;}
  int get_grid_detail() const {return grid_detail;}
  void set_grid_detail (const int i) {grid_detail = i;}
  // get/set which energy terms to use
  bool get_use_term_l_e() const {return use_term_l_e;}
  void set_use_term_l_e(const bool b) {use_term_l_e = b;}
  bool get_use_term_l_lj() const {return use_term_l_lj;}
  void set_use_term_l_lj(const bool b) {use_term_l_lj = b;}
  bool get_use_term_l_bond() const {return use_term_l_bond;}
  void set_use_term_l_bond(const bool b) {use_term_l_bond = b;}
  bool get_use_term_l_angle() const {return use_term_l_angle;}
  void set_use_term_l_angle(const bool b) {use_term_l_angle = b;}
  bool get_use_term_l_dihedral() const {return use_term_l_dihedral;}
  void set_use_term_l_dihedral(const bool b) {use_term_l_dihedral = b;}
  bool get_use_term_l_improper() const {return use_term_l_improper;}
  void set_use_term_l_improper(const bool b) {use_term_l_improper = b;}
  bool get_use_term_ml_e() const {return use_term_ml_e;}
  void set_use_term_ml_e(const bool b) {use_term_ml_e = b;}
  bool get_use_term_ml_lj() const {return use_term_ml_lj;}
  void set_use_term_ml_lj(const bool b) {use_term_ml_lj = b;}
  bool get_use_term_ml_bond() const {return use_term_ml_bond;}
  void set_use_term_ml_bond(const bool b) {use_term_ml_bond = b;}
  bool get_use_term_ml_angle() const {return use_term_ml_angle;}
  void set_use_term_ml_angle(const bool b) {use_term_ml_angle = b;}
  bool get_use_term_ml_dihedral() const {return use_term_ml_dihedral;}
  void set_use_term_ml_dihedral(const bool b) {use_term_ml_dihedral = b;}
  bool get_use_term_ml_improper() const {return use_term_ml_improper;}
  void set_use_term_ml_improper(const bool b) {use_term_ml_improper = b;}
  bool get_use_term_s_e() const {return use_term_s_e;}
  void set_use_term_s_e(const bool b) {use_term_s_e = b;}
  bool get_use_term_s_lj() const {return use_term_s_lj;}
  void set_use_term_s_lj(const bool b) {use_term_s_lj = b;}
  bool get_use_term_s_bond() const {return use_term_s_bond;}
  void set_use_term_s_bond(const bool b) {use_term_s_bond = b;}
  bool get_use_term_s_angle() const {return use_term_s_angle;}
  void set_use_term_s_angle(const bool b) {use_term_s_angle = b;}
  bool get_use_term_s_dihedral() const {return use_term_s_dihedral;}
  void set_use_term_s_dihedral(const bool b) {use_term_s_dihedral = b;}
  bool get_use_term_s_improper() const {return use_term_s_improper;}
  void set_use_term_s_improper(const bool b) {use_term_s_improper = b;}
  bool get_use_term_ms_e() const {return use_term_ms_e;}
  void set_use_term_ms_e(const bool b) {use_term_ms_e = b;}
  bool get_use_term_ms_lj() const {return use_term_ms_lj;}
  void set_use_term_ms_lj(const bool b) {use_term_ms_lj = b;}
  bool get_use_term_ms_bond() const {return use_term_ms_bond;}
  void set_use_term_ms_bond(const bool b) {use_term_ms_bond = b;}
  bool get_use_term_ms_angle() const {return use_term_ms_angle;}
  void set_use_term_ms_angle(const bool b) {use_term_ms_angle = b;}
  bool get_use_term_ms_dihedral() const {return use_term_ms_dihedral;}
  void set_use_term_ms_dihedral(const bool b) {use_term_ms_dihedral = b;}
  bool get_use_term_ms_improper() const {return use_term_ms_improper;}
  void set_use_term_ms_improper(const bool b) {use_term_ms_improper = b;}
  bool get_skip_stat_pair_mmg() const {return skip_stat_pair_mmg;}
  void set_skip_stat_pair_mmg(const bool b) {skip_stat_pair_mmg = b;}
  bool get_skip_stat_mmint() const {return skip_stat_mmint;}
  void set_skip_stat_mmint(const bool b) {skip_stat_mmint = b;}

  // holds contents of .con file, defines connectivity to membrane sides
  map<SiteID,int> conmap;
  // holds contents of .rot file, defines sites and their type
  map<SiteID,int> rotmap;

  // this is parameter data read from file
  list<GcemSiteDef> sitedef;
  CharmmPara chmpara;
  Rotlib rlib;

  // this is the derived data - the GCEM representation of our molecule
  AtomSet input_pqr;
  AtomSet background;
  list<GcemSite> sites;
  map<InstanceID,map<InstanceID, double> > mmg;
  map<InstanceID,map<InstanceID, double> > pbg;

  // does only apply if the map does really contain an entry
  bool contains_mmg (const InstanceID&, const InstanceID&) const;
  bool contains_pbg (const InstanceID&, const InstanceID&) const;
  // functions for mmg supporting primitive sparseness
  // by simply returning zero if an element does not exist
  double get_mmg (const InstanceID&, const InstanceID&) const;
  double get_pbg (const InstanceID&, const InstanceID&) const;
  // this one is complaining if no real entry is found
  double get_mmg_strict (const InstanceID&, const InstanceID&) const;
  double get_pbg_strict (const InstanceID&, const InstanceID&) const;
  // insert / update entries
  pair<map<InstanceID,map<InstanceID, double> >::iterator,bool> insert_mmg (const InstanceID&, const InstanceID&, const double);
  pair<map<InstanceID,map<InstanceID, double> >::iterator,bool> insert_pbg (const InstanceID&, const InstanceID&, const double);

  // verify uniqueness and insert in conmap/rotmap retrieve by SiteID
  // at the moment sites are not allowed to change connectivity, easily changeable by using InstanceIDs
  pair<map<SiteID,int>::iterator,bool> insert_rotmap_entry(const SiteID&, const int);
  pair<map<SiteID,int>::iterator,bool> insert_conmap_entry(const SiteID&, const int);
  bool contains_rotmap_entry(const SiteID&) const;
  bool contains_rotmap_value(const int) const;
  bool contains_conmap_entry(const SiteID&) const;
  int get_rotmap_value(const SiteID&) const;
  int get_conmap_value(const SiteID&) const;

  // sitedef functions
  inline bool contains_sitedef(const string&) const;
  inline GcemSiteDef get_sitedef(const string&) const;
  inline bool contains_sitedef(const string&, const SiteID&) const;
  inline GcemSiteDef get_sitedef(const string&, const SiteID&) const;
  inline bool contains_sitedef(const SiteID&) const;
  inline GcemSiteDef get_sitedef(const SiteID&) const;
  inline list<GcemSiteDef>::iterator insert_sitedef(const GcemSiteDef&);

  // sites functions
  bool contains_site(const SiteID&) const;
  GcemSite get_site(const SiteID&) const;
  list<GcemSite>::iterator insert_site(const GcemSite&);
  pair<map<CharmmID,string>::iterator,bool> insert_site(const SiteID&, const string&);
  
  // provides vector math functions, does not store data
  GcemGeometryCalculator geom_calc;

  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & conmap;
    ar & rotmap;
    ar & sitedef;
    ar & chmpara;
    ar & rlib;
    ar & input_pqr;
    ar & background;
    ar & sites;
    ar & mmg;
    ar & pbg;
    // string
    ar & name & gcemdir & meadpath & eps1set & eps2set & eps3set & channelset & elycavset;
    // bool
    ar & use_term_l_e & use_term_l_lj & use_term_l_bond & use_term_l_angle
       & use_term_l_dihedral & use_term_l_improper & use_term_ml_e
       & use_term_ml_lj & use_term_ml_bond & use_term_ml_angle
       & use_term_ml_dihedral & use_term_ml_improper & use_term_s_e
       & use_term_s_lj & use_term_s_bond & use_term_s_angle
       & use_term_s_dihedral & use_term_s_improper
       & use_term_ms_e & use_term_ms_lj & use_term_ms_bond
       & use_term_ms_angle & use_term_ms_dihedral & use_term_ms_improper
       & use_empirical & use_filter & skip_stat_pair_mmg & skip_stat_mmint
       & do_postproc & print_chm & print_debug & has_memb & has_con & do_cap;
    // int
    ar & inside & grid_detail;
    // double
    ar & capacitance & vmemb_inter & filter_cutoff & filter_cutoff_gs & eps_ff
       & coulomb_14_scalefac & epsin1 & epsin2 & epsin3 & epscore
       & epshead & epsext & zminh & zminc & zmaxc & zmaxh & membholer
       & membholex & membholey & ionicstr & ionicstr1 & ionicstr2;
  }
  // terms involving ligand atoms and no model atoms
  bool use_term_l_e;
  bool use_term_l_lj;
  bool use_term_l_bond;
  bool use_term_l_angle;
  bool use_term_l_dihedral;
  bool use_term_l_improper;
  // terms involving ligand atoms and model atoms (ligand model interactions belongs to mmback)
  bool use_term_ml_e;
  bool use_term_ml_lj;
  bool use_term_ml_bond;
  bool use_term_ml_angle;
  bool use_term_ml_dihedral;
  bool use_term_ml_improper;
  // terms involving no ligand and no model atoms (internal site)
  bool use_term_s_e;
  bool use_term_s_lj;
  bool use_term_s_bond;
  bool use_term_s_angle;
  bool use_term_s_dihedral;
  bool use_term_s_improper;
  // terms involving no ligand but model atoms (site/model interaction mmback)
  bool use_term_ms_e;
  bool use_term_ms_lj;
  bool use_term_ms_bond;
  bool use_term_ms_angle;
  bool use_term_ms_dihedral;
  bool use_term_ms_improper;
  bool use_empirical;
  bool use_filter;
  bool skip_stat_pair_mmg;
  bool skip_stat_mmint;
  bool do_postproc;
  bool print_chm;
  bool print_debug;
  // capacitance for  1V membrane potential (CAUTION: grid size dependent!)
  double capacitance, vmemb_inter;
  // energy cutoff for exclusion of instances with high intrinsic energy
  // prior to interaction energy calculation, should be > filter_cutoff_gs
  double filter_cutoff;
  // filter cutoff for use with goldstein like filter criterium, applied after
  // interaction energy calculation
  double filter_cutoff_gs;
  // relative dielectric permittivity for which the forcefield was parametrized
  double eps_ff, coulomb_14_scalefac;
  double epsin1,epsin2,epsin3,epscore,epshead,epsext,zminh,zminc,zmaxc,zmaxh,
         membholer,membholex,membholey,ionicstr,ionicstr1,ionicstr2;
  bool has_memb, has_con, do_cap;
  // which membrane side is inside def. 1 = upper phase (def), 0 = lower phase
  int inside;
  // detail of autogenerated electrostatics grids
  // grid size of innermost grid (always), grid spacing of inner levels (if no user defined grids arfe supplied)
  int grid_detail;
  string name,gcemdir,meadpath,eps1set,eps2set,eps3set,channelset,elycavset;
};

// mmg / pbg functions

inline bool GcemMolecule::contains_mmg(const InstanceID& k, const InstanceID& l) const
{
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=mmg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=mmg.find(l);
  if (i!=mmg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j!=i->second.end()){
    found = true;
   }
  }
  if (!found && ii != mmg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j!=ii->second.end()){
    found = true;
   }
  }
  return found;
}

inline bool GcemMolecule::contains_pbg(const InstanceID& k, const InstanceID& l) const
{
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=pbg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=pbg.find(l);
  if (i!=pbg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j!=i->second.end()){
    found = true;
   }
  }
  if (!found && ii != pbg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j!=ii->second.end()){
    found = true;
   }
  }
  return found;
}

// BEWARE: this function does not complain about non-existing elements but
// simply returns zero to get the mmg map sparse
// calculation of all terms must be ensured by calc mmg !!!
inline double GcemMolecule::get_mmg(const InstanceID& k, const InstanceID& l) const
{
  double result = 0;
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=mmg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=mmg.find(l);
  if (i  != mmg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != mmg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  return result;
}

// BEWARE: this function does not complain about non-existing elements but
// simply returns zero to get the pbg map sparse
// calculation/reading of all terms must be ensured by GcemSolverOutput::read_mead_out()
inline double GcemMolecule::get_pbg(const InstanceID& k, const InstanceID& l) const
{
  double result = 0;
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=pbg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=pbg.find(l);
  if (i  != pbg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != pbg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  return result;
}

// this get function complains about a non-existing entry
inline double GcemMolecule::get_mmg_strict(const InstanceID& k, const InstanceID& l) const
{
  double result;
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=mmg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=mmg.find(l);
  if (i  != mmg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != mmg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found)
    ::error ("GcemMolecule::get_mmg_strict: could not find entry.\n");
  return result;
}

// this get function complains about a non-existing entry
inline double GcemMolecule::get_pbg_strict(const InstanceID& k, const InstanceID& l) const
{
  double result;
  bool found = false;
  map<InstanceID,map<InstanceID,double> >::const_iterator i=pbg.find(k);
  map<InstanceID,map<InstanceID,double> >::const_iterator ii=pbg.find(l);
  if (i  != pbg.end()){
   map<InstanceID,double>::const_iterator j=i->second.find(l);
   if (j != i->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found && ii != pbg.end()){
   map<InstanceID,double>::const_iterator j=ii->second.find(k);
   if (j != ii->second.end()){
    found = true;
    result = j->second;
   }
  }
  if (!found)
    ::error ("GcemMolecule::get_pbg_strict: could not find entry.\n");
  return result;
}

inline pair<map<InstanceID,map<InstanceID,double> >::iterator,bool> GcemMolecule::insert_mmg(const InstanceID& k, const InstanceID& l, const double f)
{
  pair<map<InstanceID,map<InstanceID,double> >::iterator,bool> result;
  result.first = mmg.end();
  result.second = false;
  bool foundj  = false;
  bool foundjj = false;
  bool foundi  = false;
  bool foundii = false;
  map<InstanceID,map<InstanceID,double> >::iterator i=mmg.find(k);
  map<InstanceID,map<InstanceID,double> >::iterator ii=mmg.find(l);
  map<InstanceID,double>::iterator j,jj;
  if (i != mmg.end()){
   foundi = true;
   j=i->second.find(l);
   if (j != i->second.end()){
    foundj = true;
   }
  }
  if (ii != mmg.end()){
   foundii = true;
   jj=ii->second.find(k);
   if (jj != ii->second.end()){
    foundjj = true;
   }
  }
  // replace both possible preexisting occurrences
  if (foundj || foundjj){
   if (foundjj)
    if (foundi)   j->second = f;
   if (foundjj)
    if (foundii) jj->second = f;
  } // creation of one entry suffices
  else if (foundi) {
   i->second.insert(map<InstanceID,double>::value_type(l,f));
   result.first = i; result.second = true;
  }
  else if (foundii){
   ii->second.insert(map<InstanceID,double>::value_type(k,f));
   result.first = ii; result.second = true;
  }
  // create completely new entry
  else {
   map<InstanceID,double> tmpmap;
   tmpmap.insert(map<InstanceID,double>::value_type(l,f));
   result = mmg.insert(map<InstanceID,map<InstanceID,double> >::value_type(k,tmpmap));
  }
  return result;
}


inline pair<map<InstanceID,map<InstanceID,double> >::iterator,bool> GcemMolecule::insert_pbg(const InstanceID& k, const InstanceID& l, const double f)
{
  pair<map<InstanceID,map<InstanceID,double> >::iterator,bool> result;
  result.first = pbg.end();
  result.second = false;
  bool foundj  = false;
  bool foundjj = false;
  bool foundi  = false;
  bool foundii = false;
  map<InstanceID,map<InstanceID,double> >::iterator i=pbg.find(k);
  map<InstanceID,map<InstanceID,double> >::iterator ii=pbg.find(l);
  map<InstanceID,double>::iterator j,jj;
  if (i != pbg.end()){
   foundi = true;
   j=i->second.find(l);
   if (j != i->second.end()){
    foundj = true;
   }
  }
  if (ii != pbg.end()){
   foundii = true;
   jj=ii->second.find(k);
   if (jj != ii->second.end()){
    foundjj = true;
   }
  }
  // replace both possible preexisting occurrences
  if (foundj || foundjj){
   if (foundjj)
    if (foundi)   j->second = f;
   if (foundjj)
    if (foundii) jj->second = f;
  } // creation of one entry suffices
  else if (foundi) {
   i->second.insert(map<InstanceID,double>::value_type(l,f));
   result.first = i; result.second = true;
  }
  else if (foundii){
   ii->second.insert(map<InstanceID,double>::value_type(k,f));
   result.first = ii; result.second = true;
  }
  // create completely new entry
  else {
   map<InstanceID,double> tmpmap;
   tmpmap.insert(map<InstanceID,double>::value_type(l,f));
   result = pbg.insert(map<InstanceID,map<InstanceID,double> >::value_type(k,tmpmap));
  }
  return result;
}

// conmap and rotmap functions
inline bool GcemMolecule::contains_conmap_entry(const SiteID& k) const
{
  map<SiteID,int>::const_iterator i=conmap.find(k);
  return i != conmap.end();
}

inline int GcemMolecule::get_conmap_value(const SiteID& k) const
{
  map<SiteID,int>::const_iterator i=conmap.find(k);
  if (i != conmap.end())
  {
    return i->second;
  }
  else
  {
    ::error ("GcemMolecule::get_conmap_entry: could not find entry.\n");
  }
  return 0; // quite clang
}

inline pair<map<SiteID,int>::iterator,bool> GcemMolecule::insert_conmap_entry(const SiteID& sid, const int i1)
{
  pair<map<SiteID,int>::iterator,bool> tp2;
  if (!contains_conmap_entry(sid)){
    pair<SiteID,int> tp(sid, i1);
    tp2 = conmap.insert(tp);
  }
  else {
    ::error ("GcemMolecule::insert_conmap_entry: attempt to insert duplicate key.\n");
  }
  return tp2;
}

inline bool GcemMolecule::contains_rotmap_entry(const SiteID& k) const
{
  map<SiteID,int>::const_iterator i=rotmap.find(k);
  return i != rotmap.end();
}

inline bool GcemMolecule::contains_rotmap_value(const int value) const
{
  for (map<SiteID,int>::const_iterator i=rotmap.begin(); i!=rotmap.end(); i++){
   if (i->second == value) return true;
  }
  return false;
}

inline int GcemMolecule::get_rotmap_value(const SiteID& k) const
{
  map<SiteID,int>::const_iterator i=rotmap.find(k);
  if (i != rotmap.end())
  {
    return i->second;
  }
  else
  {
    ::error ("GcemMolecule::get_rotmap_entry: could not find entry.\n");
  }
  return 0; // quite clang
}

inline pair<map<SiteID,int>::iterator,bool> GcemMolecule::insert_rotmap_entry(const SiteID& sid, const int i1)
{
  pair<map<SiteID,int>::iterator,bool> tp2;
  if (!contains_rotmap_entry(sid)){
    pair<SiteID,int> tp(sid, i1);
    tp2 = rotmap.insert(tp);
  }
  else {
    ::error ("GcemMolecule::insert_rotmap_entry: attempt to insert duplicate key.\n");
  }
  return tp2;
}

// sitedef functions

inline bool GcemMolecule::contains_sitedef(const string& k) const
{
  bool found = false;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_name() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline bool GcemMolecule::contains_sitedef(const string& k, const SiteID& s) const
{
  bool found = false;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_name() == k && i->get_sid() == s){
    found = true;
    break;
   }
  }
  return found;
}

inline bool GcemMolecule::contains_sitedef(const SiteID& s) const
{
  bool found = false;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_sid() == s){
    found = true;
    break;
   }
  }
  return found;
}

inline GcemSiteDef GcemMolecule::get_sitedef(const SiteID& s) const
{
  bool found = false;
  list<GcemSiteDef>::const_iterator j;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_sid() == s){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
  {
    ::error ("GcemMolecule::get_sitedef: could not find key.\n");
    return GcemSiteDef(); // quite clang
  }
  else
  {
    return *j;
  }
}

inline GcemSiteDef GcemMolecule::get_sitedef(const string& k) const
{
  bool found = false;
  list<GcemSiteDef>::const_iterator j;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_name() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
  {
    ::error ("GcemMolecule::get_sitedef: could not find key.\n");
    return GcemSiteDef(); // quite clang
  }
  else
  {
    return *j;
  }
}

inline GcemSiteDef GcemMolecule::get_sitedef(const string& k, const SiteID& s) const
{
  bool found = false;
  list<GcemSiteDef>::const_iterator j;
  for (list<GcemSiteDef>::const_iterator i = sitedef.begin(); i != sitedef.end(); ++i){
   if (i->get_name() == k && i->get_sid() == s){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
  {
    ::error ("GcemMolecule::get_sitedef: could not find key.\n");
    return GcemSiteDef(); // quite clang
  }
  else
  {
    return *j;
  }
}

inline list<GcemSiteDef>::iterator GcemMolecule::insert_sitedef(const GcemSiteDef& a)
{
  if (contains_sitedef(a.get_name(),a.get_sid()))
    ::error ("GcemMolecule::insert_sitedef: attempt to insert duplicate entry.\n");
  sitedef.push_back(a);
  return sitedef.end();
}

// site functions

inline bool GcemMolecule::contains_site(const SiteID& s) const
{
  bool found = false;
  for (list<GcemSite>::const_iterator i = sites.begin(); i != sites.end(); ++i){
   if (i->get_sid() == s){
    found = true;
    break;
   }
  }
  return found;
}

inline GcemSite GcemMolecule::get_site(const SiteID& s) const
{
  bool found = false;
  list<GcemSite>::const_iterator j;
  for (list<GcemSite>::const_iterator i = sites.begin(); i != sites.end(); ++i){
   if (i->get_sid() == s){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
  {
    ::error("GcemMolecule::get_site: could not find key.\n");
    return GcemSite(); // quite clang
  }
  else
  {
    return *j;
  }
}

inline list<GcemSite>::iterator GcemMolecule::insert_site(const GcemSite& a)
{
  if (contains_site(a.get_sid()))
    ::error ("GcemMolecule::insert_site: attempt to insert duplicate entry.\n");
  sites.push_back(a);
  return sites.end();
}

//////////////////////// list helper functions /////////////////////////

inline bool GcemMolecule::in_list(const string& s, const list<string>& l) const
{
  bool found = false;
  for (list<string>::const_iterator i = l.begin(); i != l.end(); ++i){
   if (*i == s){
    found = true;
    break;
   }
  }
  return found;
}

#endif

// GcemMolecule.h ends here
