/*  A GCEM Site

    site label, instances (defined by charge form and conformer)

    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemSite_h
#define _GcemSite_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/SiteID.h"
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/GcemInstance.h"
#include "MEAD_e/globals.h"
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/list.hpp>

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

//!wrap!
class GcemSite {
public:
  GcemSite();
  GcemSite (const SiteID& id);
  GcemSite (const GcemSite& a);
  ~GcemSite () {}
  //!nowrap!+
  GcemSite& operator=(const GcemSite& a);

  //!nowrap!-
  void print() const;
  void print_instances() const;
  void renumber_instances();

  list<GcemInstance> instances;

  string get_name() const {return name;}
  SiteID get_sid() const {return sid;}
  int get_max_confnum() const;
  int get_nconf() const;
  int get_ninst() const {return instances.size();}

  // check by label if form or ligand are already listed
  bool contains_form (const string&) const;
  bool contains_instance (const InstanceID&) const;
  bool contains_ligand (const string&) const;

  // retrieve list of instances by form label
  list<GcemInstance> get_form(const string&) const;
  // retrieve instances sorted by form label
  map<string,list<GcemInstance> > get_form_map() const;
  // retrieve specific instance by InstanceID
  GcemInstance get_instance(const InstanceID&) const;

  // insert in map, InstanceID is generated automatically
  list<GcemInstance>::iterator insert_instance(const GcemInstance&);
  list<GcemInstance>::iterator delete_instance(const InstanceID&);
  list<GcemInstance>::iterator delete_confnum(const int);
  InstanceID get_free_instid (const GcemInstance&) const;

  bool operator== (const GcemSite& a) const {return sid == a.get_sid();}
  bool operator!= (const GcemSite& a) const {return sid != a.get_sid();}
  bool operator< (const GcemSite& a) const {return sid < a.get_sid();}
  bool operator> (const GcemSite& a) const {return sid > a.get_sid();}

  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & instances;
    ar & sid;
    ar & name;
  }
  string name;
  SiteID sid;
};

// form list functions

inline std::list<GcemInstance> GcemSite::get_form(const string& k) const
{
  bool found = false;
  std::list<GcemInstance> result;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (i->get_name() == k){
    found = true;
    result.push_back(*i);
   }
  }
  if (!found)
  {
    ::error("GcemSite::get_form: key not found");
  }
  return result;
}

inline std::map<std::string, std::list<GcemInstance> > GcemSite::get_form_map() const
{
  bool found = false;
  std::map<std::string, std::list<GcemInstance> > result;
  std::map<std::string, std::list<GcemInstance> >::iterator m;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   string name = i->get_name();
   m = result.find(name);
   if (m != result.end()){
    m->second.push_back(*i);
   } else {
    std::list<GcemInstance> tmplist;
    tmplist.push_back(*i);
    result.insert(std::pair<std::string, std::list<GcemInstance> >(name,tmplist));
   }
  }
  return result;
}

inline GcemInstance GcemSite::get_instance(const InstanceID& k) const
{
  bool found = false;
  std::list<GcemInstance>::const_iterator j;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (i->get_instid() == k){
    j = i;
    found = true;
    break;
   }
  }
  if (!found)
  {
    return (*j);
  }
  else
  {
    ::error("GcemSite::get_instance: key not found");
    return GcemInstance(); // dummy to quite Clang
  } 
}

inline std::list<GcemInstance>::iterator GcemSite::delete_instance(const InstanceID& k)
{
  bool found = false;
  std::list<GcemInstance>::iterator j;
  for (list<GcemInstance>::iterator i = instances.begin(); i != instances.end(); ++i){
    if (i->get_instid() == k)
    {
      j = i;
      found = true;
      break;
    }
  }
  if (!found)
  {
    ::error("GcemSite::delete_instance: key not found");
    return instances.end(); // dummy to quite Clang
  }
  else
  {
    instances.erase(j);
    return instances.begin();
  }
}

inline std::list<GcemInstance>::iterator GcemSite::delete_confnum(const int k)
{
  int found=0;
  for (list<GcemInstance>::iterator i = instances.begin(); i != instances.end(); ++i){
   if (i->get_confnum() == k){
    instances.erase(i);
    found++;
   }
  }
  if (found==0)
    ::error("GcemSite::delete_conf: conformation not found");
  return instances.begin();
}

// Existence of element (not STL-like

inline bool GcemSite::contains_form(const string& k) const
{
  bool found = false;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (i->get_name() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline bool GcemSite::contains_instance(const InstanceID& k) const
{
  bool found = false;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (i->get_instid() == k){
    found = true;
    break;
   }
  }
  return found;
}

inline bool GcemSite::contains_ligand(const string& k) const
{
  bool found = false;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   for (map<string,int>::const_iterator j = i->ligandmap.begin(); j != i->ligandmap.end(); ++j){
    if (j->first == k && j->second > 0){
     found = true;
     break;
    }
   }
  }
  return found;
}

inline int GcemSite::get_max_confnum() const
{
  int maxn = 0;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   if (maxn < i->get_confnum())
    maxn = i->get_confnum();
  }
  return maxn;
}

inline int GcemSite::get_nconf() const
{
  std::list<int> cl;
  for (list<GcemInstance>::const_iterator i = instances.begin(); i != instances.end(); ++i){
   cl.push_back(i->get_confnum());
  }
  cl.sort();
  cl.unique();
  return cl.size();
}

// easier to use insert (not quite STL-like)
inline std::list<GcemInstance>::iterator GcemSite::insert_instance (const GcemInstance& a){
  GcemInstance b(a);
  InstanceID id = get_free_instid(b);
  b.set_instid(id);
  if (!contains_instance(id))
    instances.push_back(b);
  else
    ::error("GcemSite::insert_instance: attempt to insert a duplicate instance\n");
  return instances.end();
}

inline InstanceID GcemSite::get_free_instid(const GcemInstance& a) const {
  // start from 1 because instid 0 is meant to mark background parts
  int maxinst = int(1);
  InstanceID id(a.get_instid());
  while (contains_instance(id)){
   id = InstanceID(sid.get_sitename(), id.get_instname(), sid.get_conformer(), sid.get_chain(), sid.get_site(), maxinst);
   maxinst++;
  }
  return id;
}

#endif

// GcemSite.h ends here
