/*  
    helper object for GcemMolecule
    creates GcemSites from GcemSiteDefs and AtomSets

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemSiteCreator.h"
namespace fs = boost::filesystem;

void GcemSiteCreator::create_sites(GcemMolecule& molecule) const
{
  blab2 << "Entering GcemSiteCreator::create_sites" << endl;

  map<SiteID,pair<AtomSet,AtomSet> > tmpsites;
  map<SiteID,map<int,AtomSet> > predef_inst;
  map<SiteID,map<int,pair<double, AtomSet> > > new_inst;

  // look for predefined conformers/forms marked by non-zero instance IDs
  // collect atoms with zero instance ID for each site as basis for generation of rotamers
  for (map<SiteID,int>::const_iterator i = molecule.rotmap.begin(); i!=molecule.rotmap.end(); ++i){
   AtomSet tmpset1, tmpset2, tmpset3, tmpset4, tmpset5;
   bool site_has_predef = false;
   for (AtomSet::const_iterator j = molecule.input_pqr.begin(); j!=molecule.input_pqr.end(); ++j) {
    // atoms belonging to site
    if (i->first.get_chain() == j->second.chain && i->first.get_site() == j->second.site){
     if (j->second.instance == 0){
      tmpset1.insert(j->second);
     }
     // use predefined instances only for sites marked flexible or QM (1 || 2 || 4 in .rot)
     else if (i->second == 1 || i->second == 2 || i->second == 4){
      tmpset2.insert(j->second);
      site_has_predef = true;
     }
    }

    // atoms of neighbouring residues needed later for blocking groups of model compound
    if (i->first.get_chain() == j->second.chain && i->first.get_site() == (j->second.site + 1)){
     if (j->second.instance == 0 && (j->second.atname == "C"  || j->second.atname == "O")){
      tmpset3.insert(j->second);
     }
    }
    else if (i->first.get_chain() == j->second.chain && i->first.get_site() == (j->second.site - 1)){
     // pyroglutamate should anyway appear only at the N-terminus
     if (j->second.resname != "PRO"){
      if (j->second.instance == 0 && (j->second.atname == "N" || j->second.atname == "NH" || j->second.atname == "HN" ||
                                      j->second.atname == "H" || j->second.atname == "CA" ||
                                      j->second.atname == "HA")){
       tmpset3.insert(j->second);
      }
     }
     else {
      if (j->second.instance == 0 && (j->second.atname == "N"   || j->second.atname == "NH"  || j->second.atname == "CD"  ||
                                      j->second.atname == "HD1" || j->second.atname == "HD2" ||
                                      j->second.atname == "CA"  || j->second.atname == "HA")){
       tmpset3.insert(j->second);
      }
     }
    }
   }
   
   // QM site without predefined instances in the input structure
   // try to read structure data from separate files
   if (i->second == 4 && !site_has_predef){
    GcemSiteDef sdef(molecule.get_sitedef(i->first));
    // arrange atoms of predefined instances sorted by input instance nr
    map<int,AtomSet> tmpmap;
    int count = 0;
    for (list<GcemFormDef>::const_iterator f = sdef.forms.begin(); f != sdef.forms.end(); f++){
     count++;
     stringstream ss1;
     string iid_tmpstr;
     ss1 << count;
     ss1 >> iid_tmpstr;
     string struct_fn(i->first.get_fn_prefix());
     struct_fn += "_" + f->get_name() + "-" + iid_tmpstr;
     AtomSet ta(struct_fn);
     ta.read();
     // make sure that the instance and site IDs are correctly assigned
     for (AtomSet::iterator taiter = ta.begin(); taiter != ta.end(); taiter++){
      taiter->second.instance = count;
      taiter->second.site     = i->first.get_site();
      tmpset2.insert(taiter->second);
     }
    }
   }
   
   // look if our site appears to be an amino acid - i.e. if the blocking group rule applies at all
   // otherwise purge the blocking group atoms, QM sites get no blocking groups
   bool b1 = false; bool b2 = false; bool b3 = false; bool b4 = false;
   for (AtomSet::const_iterator j = tmpset1.begin(); j!=tmpset1.end(); ++j) {
    if      (j->second.atname == "C") {b1 = true;}
    else if (j->second.atname == "O") {b2 = true;}
    else if (j->second.atname == "N" || j->second.atname == "NH") {b3 = true;}
    else if (j->second.atname == "CA"){b4 = true;}
   }
   if (b1 && b2 && b3 && b4 && i->second != 4){
    // check zero net charge of blocking groups - otherwise reset charges with CHARMM standard
    if (tmpset3.total_charge() != 0){
     AtomSet tmpset3b;
     for (AtomSet::const_iterator j = tmpset3.begin(); j!=tmpset3.end(); ++j) {
      if      (j->second.atname == "C") {
      Atom ta(j->second);
      ta.charge = double(0.51);
      tmpset3b.insert(ta);
      }
      else if (j->second.atname == "O") {
       Atom ta(j->second); ta.charge = double(-0.51); tmpset3b.insert(ta);
      }
      // pyroglutamate should anyway appear only at the N-terminus
      if (j->second.resname != "PRO"){
       if      (j->second.atname == "H" || j->second.atname == "HN"){Atom ta(j->second); ta.charge = double(0.31);
                                                                     tmpset3b.insert(ta);}
       else if (j->second.atname == "N") {Atom ta(j->second); ta.charge = double(-0.47); tmpset3b.insert(ta);}
       else if (j->second.atname == "CA"){Atom ta(j->second); ta.charge = double(0.07); tmpset3b.insert(ta);}
       else if (j->second.atname == "HA"){Atom ta(j->second); ta.charge = double(0.09); tmpset3b.insert(ta);}
      }
      else {
       if      (j->second.atname == "N"){Atom ta(j->second); ta.charge = double(-0.29); tmpset3b.insert(ta);}
       else if (j->second.atname == "CD") {Atom ta(j->second); ta.charge = double(0.00); tmpset3b.insert(ta);}
       else if (j->second.atname == "HD1"){Atom ta(j->second); ta.charge = double(0.09); tmpset3b.insert(ta);}
       else if (j->second.atname == "HD2"){Atom ta(j->second); ta.charge = double(0.09); tmpset3b.insert(ta);}
       else if (j->second.atname == "CA"){Atom ta(j->second); ta.charge = double(0.02); tmpset3b.insert(ta);}
       else if (j->second.atname == "HA"){Atom ta(j->second); ta.charge = double(0.09); tmpset3b.insert(ta);}
      }
     }
     // if still no integer charge (most likely incomplete atoms) purge blocking groups
     if (tmpset3b.total_charge() != 0){tmpset3.clear();}
     else {tmpset3 = tmpset3b;}
    }
   }
   else {
    tmpset3.clear();
   }
   // from tmpset1 search atoms occurring in SiteDef or moving atoms in ResTop
   // (actual site) and put in tmpset4, rest + tmpset3 goes into tmpset5
   // (additional atoms of  model compound)
   for (AtomSet::const_iterator j = tmpset1.begin(); j!=tmpset1.end(); ++j){
    bool in_sdef = false;
    if (i->second == 1 || i->second == 3){
     GcemSiteDef sdef(molecule.get_sitedef(i->first.get_sitename()));
     in_sdef = sdef.contains_atom(j->second.resname,j->second.atname);
    }
    // second check ResTop for atoms to move in sidechain torsions
    bool in_sctor = false;
    if (i->second == 1 || i->second == 2){
     CharmmParaResTop restop(molecule.chmpara.get_Topology(i->first.get_sitename()));
     in_sctor = restop.moved_by_sctor(j->second.atname);
    }
    // FIXME: hack to get complete charge group and thus integer charge of site
    // is this necessary?
    if (j->second.atname == "CB" && (i->first.get_sitename() == "ASP"
        || i->first.get_sitename() == "ASN" || i->first.get_sitename() == "GLU" || i->first.get_sitename() == "GLN"
        || i->first.get_sitename() == "ARG" || i->first.get_sitename() == "LYS" || i->first.get_sitename() == "HIS"
        || i->first.get_sitename() == "HSP" || i->first.get_sitename() == "HSD" || i->first.get_sitename() == "HSE"
        || i->first.get_sitename() == "SER" || i->first.get_sitename() == "THR" || i->first.get_sitename() == "CYS"
        || i->first.get_sitename() == "MET" || i->first.get_sitename() == "VAL" || i->first.get_sitename() == "ILE"
        || i->first.get_sitename() == "LEU" || i->first.get_sitename() == "PHE" || i->first.get_sitename() == "TRP"
        || i->first.get_sitename() == "TYR") && (i->second == 1 || i->second == 2)){in_sctor = true;}
    // third check for atoms appearing in predefined instances
    AtomID aid(int residue_number, const string& atom_name, const int confomer = 0, const int chain = 0, const int site = 0, const int instance = 0);
    bool in_pdinst = false;
    for (AtomSet::const_iterator k = tmpset2.begin(); k!=tmpset2.end(); ++k){
     if (j->second.conformer == k->second.conformer && j->second.chain == k->second.chain &&
         j->second.site == k->second.site && j->second.resname == k->second.resname &&
         j->second.atname == k->second.atname){
      in_pdinst = true;
      break;
     }
    }
    if (in_sdef || in_sctor || in_pdinst){
     tmpset4.insert(j->second);
    } else {
      tmpset5.insert(j->second);
    }
   }
   for (AtomSet::const_iterator j = tmpset3.begin(); j!=tmpset3.end(); ++j){
    tmpset5.insert(j->second);
   }
   // save collected atoms into tmpsites map
   pair<AtomSet,AtomSet> tmppair1(tmpset4,tmpset5);
   tmpsites.insert(pair<SiteID,pair<AtomSet,AtomSet> >(i->first,tmppair1));

   // flexible MM, titratable or QM sites
   if (i->second == 1 || i->second == 2 || i->second == 4){
    // arrange atoms of predefined instances sorted by input instance nr
    map<int,AtomSet> tmpmap, tmpmap2;
    // check if variable atoms in non-zero instances are also found among non-variable atoms
    for (AtomSet::const_iterator j = tmpset2.begin(); j!=tmpset2.end(); ++j) {
     map<int,AtomSet>::iterator k = tmpmap.find(j->second.instance);
     if (k != tmpmap.end()){
      k->second.insert(j->second);
     } else {
      AtomSet ta;
      ta.insert(j->second);
      tmpmap.insert(pair<int,AtomSet>(j->second.instance,ta));
     }
    }
    // renumber predefined instances,

    int ii = 0;
    for (map<int,AtomSet>::const_iterator j = tmpmap.begin(); j!=tmpmap.end(); ++j) {
     ii++;
     tmpmap2.insert(pair<int,AtomSet>(ii,j->second));
    }
    if (i->second == 4){
     GcemSiteDef sdef(molecule.get_sitedef(i->first));
     if (ii!=sdef.forms.size() && ii!=0){
      cerr << "site " << i->first << ": number of instances in structure = " << ii
           << ", while in .qst " << sdef.forms.size() << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: number of instances for QM site differs between structure and .qst.");      
     }
    }
    
    // save sorted predefined instances to predef_inst
    predef_inst.insert(pair<SiteID,map<int,AtomSet> >(i->first,tmpmap2));
   }

  }

  // cross check exclusiveness of real site atoms
  for (map<SiteID,pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
   map<SiteID,pair<AtomSet,AtomSet> >::const_iterator ti(i); ti++;
   for (map<SiteID,pair<AtomSet,AtomSet> >::const_iterator j=ti; j!=tmpsites.end(); ++j){
    for (AtomSet::const_iterator k = i->second.first.begin(); k!=i->second.first.end(); ++k){
     if (j->second.first.contains(k->first)){
      cerr << k->first << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: Atom occurred in two sites.");
     }
    }
   }
  }

  // if a real site atom occurs in the model compound of another site set its charge to zero
  for (map<SiteID,pair<AtomSet,AtomSet> >::iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
   for (map<SiteID,pair<AtomSet,AtomSet> >::iterator j=tmpsites.begin(); j!=tmpsites.end(); ++j){
    if (j == i) continue;
    for (AtomSet::iterator k = i->second.second.begin(); k!=i->second.second.end(); ++k){
     if (j->second.first.contains(k->first)){
      k->second.charge = double(0.0);
     }
    }
   }
  }

  // construct molecule.background
  for (AtomSet::const_iterator k = molecule.input_pqr.begin(); k!=molecule.input_pqr.end(); ++k){
   if (k->second.instance != int(0)) continue;
   bool found = false;
   for (map<SiteID,pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
    if (i->second.first.contains(k->first)){
     found = true;
     break;
    }
   }
   if (!found && !molecule.background.contains(k->first)){
    molecule.background.insert(k->second);
   }
  }
  // check if molecule.background has integer charge
  float sum_back_charge   = molecule.background.total_charge();
  double diff = 1.0E-07*((1.0E+07*sum_back_charge) - static_cast<double>( static_cast<int>(1.0E+07*sum_back_charge)));
  if (diff<0) diff *= -1;
  if (diff > 1E-06){
    cerr << "Total charge of molecule.background AtomSet " << sum_back_charge << endl;
    ::error ("ERROR in GcemSiteCreator::create_sites: molecule.background AtomSet has non-integer charge");
  }

   //////// debug ////////
   if (molecule.get_print_debug()){
   for (map<SiteID,pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
    cout << "!//////////////////////////////////////////////////////" << endl;
    cout << "\nsite atoms of " << i->first << " :" << endl;
    i->second.first.print();
    cout << "......................................................." << endl;
    cout << "\nmolecule.background atoms of " << i->first << " :" << endl;
    i->second.second.print();
   }
   }
   //////// debug ////////

  // now we are ready to construct the rotamers

  for (map<SiteID,int>::const_iterator i = molecule.rotmap.begin(); i!=molecule.rotmap.end(); ++i){
   // flexible sites and flexible + titratable sites
   if (i->second == 1 || i->second == 2){
    map<SiteID,pair<AtomSet,AtomSet> >::const_iterator k = tmpsites.find(i->first);
    // this should never occur
    if (k == tmpsites.end()){
     cerr << i->first << endl;
     ::error ("ERROR in GcemSiteCreator::create_sites: site is missing while generating rotamers.");
    }
    if (!molecule.rlib.contains_residue(i->first.get_sitename())){
     cerr << "Warning: could not find entry in rotamer database for site name "
          << i->first.get_sitename() << ".\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }
    if (!molecule.chmpara.contains_Topology(i->first.get_sitename())){
     cerr << "GcemSiteCreator::create_sites Warning: could not find topology entry in CHARMM topology for site name "
          << i->first.get_sitename() << ".\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }
    // retrieve topology
    CharmmParaResTop top(molecule.chmpara.get_Topology(i->first.get_sitename()));
    if (top.sidechain_tor.size() == 0){
     cerr << "GcemSiteCreator::create_sites Warning: topology entry in CHARMM topology for site name "
          << i->first.get_sitename() << " contains no sidechain torsions.\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }


    // find coordinates of backbone atoms for this site, compute backbone torsion angles
    bool bb_cplt = true;
    bool got_a = true;
    bool got_b = true;
    bool got_c = true;
    bool got_d = true;
    bool got_e = true;
    double phi, psi;
    Coord a, b, c, d, e;
    AtomID id_a(i->first.get_site()-1,"C",i->first.get_conformer(),i->first.get_chain(),i->first.get_site()-1,int(0));
    AtomID id_b(i->first.get_site(),"N",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_c(i->first.get_site(),"CA",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_d(i->first.get_site(),"C",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_e(i->first.get_site()+1,"N",i->first.get_conformer(),i->first.get_chain(),i->first.get_site()+1,int(0));
    if (molecule.input_pqr.contains(id_a)){a = molecule.input_pqr[id_a].coord;} else {bb_cplt = false; got_a = false;
     blab1 << "GcemSiteCreator::create_sites Warning: could not find atom " << id_a << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (molecule.input_pqr.contains(id_b)){b = molecule.input_pqr[id_b].coord;} else {bb_cplt = false; got_b = false;
     blab1 << "GcemSiteCreator::create_sites Warning: could not find atom " << id_b << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (molecule.input_pqr.contains(id_c)){c = molecule.input_pqr[id_c].coord;} else {bb_cplt = false; got_c = false;
     blab1 << "GcemSiteCreator::create_sites Warning: could not find atom " << id_c << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (molecule.input_pqr.contains(id_d)){d = molecule.input_pqr[id_d].coord;} else {bb_cplt = false; got_d = false;
     blab1 << "GcemSiteCreator::create_sites Warning: could not find atom " << id_d << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (molecule.input_pqr.contains(id_e)){e = molecule.input_pqr[id_e].coord;} else {bb_cplt = false; got_e = false;
     blab1 << "GcemSiteCreator::create_sites Warning: could not find atom " << id_e << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (!got_b || !got_c){
     blab1 << "GcemSiteCreator::create_sites Warning could not find all atoms necessary for defining sidechain torsions" << endl;
     blab1 << "                           Will not generate rotamers for this residue." << endl;
    }
    ///////////////////////////////////////////////////////////////////////////////////
    // FIXME: this is an ugly hack to be commented out in the trunk version
    // it allows to include my custom peptide flip residue in azurin
    if (!got_a && got_b){
      for (AtomSet::const_iterator ii = molecule.input_pqr.begin(); ii != molecule.input_pqr.end(); ++ii){
       if (ii->second.resname == "PEP" && ii->second.atname == "3C" && double(geom_calc.abs_value(ii->second.coord-b)) <= double(1.8)){
        a = ii->second.coord; got_a = true;
        blab1 << "GcemSiteCreator::create_sites: Found nearby residue with resname PEPT and atom named 3C in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom C of preceding residue." << endl;
        break;
       }
       if (ii->second.resname == "PEP" && ii->second.atname == "6C" && double(geom_calc.abs_value(ii->second.coord-b)) <= double(1.8)){
        a = ii->second.coord; got_a = true;
        blab1 << "GcemSiteCreator::create_sites: Found nearby residue with resname PEPT and atom named 6C in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom C of preceding residue." << endl;
        break;
       }
      }
    }
    if (!got_e && got_d){
      for (AtomSet::const_iterator ii = molecule.input_pqr.begin(); ii != molecule.input_pqr.end(); ++ii){
       if (ii->second.resname == "PEP" && ii->second.atname == "1N" && double(geom_calc.abs_value(ii->second.coord-d)) <= double(1.8)){
        a = ii->second.coord; got_e = true;
        blab1 << "GcemSiteCreator::create_sites: Found nearby residue with resname PEPT and atom named 1N in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom N of succeeding residue." << endl;
        break;
       }
       if (ii->second.resname == "PEP" && ii->second.atname == "4N" && double(geom_calc.abs_value(ii->second.coord-d)) <= double(1.8)){
        a = ii->second.coord; got_e = true;
        blab1 << "GcemSiteCreator::create_sites: Found nearby residue with resname PEPT and atom named 4N in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom N of succeeding residue." << endl;
        break;
       }
      }
    }
    if (got_a && got_b && got_c && got_d && got_e) bb_cplt = true;
    else                                           bb_cplt = false;
    ///////////////////////////////////////////////////////////////////////////////////
    if (bb_cplt){
     phi = geom_calc.get_tor(a,b,c,d);
     psi = geom_calc.get_tor(b,c,d,e);
    } else if (!got_a && got_b && got_c && got_d && got_e){
     // take alpha helix parameters
     phi = 57*(pi/180);
     psi = geom_calc.get_tor(b,c,d,e);
    } else if (got_a && got_b && got_c && got_d && !got_e){
     // take alpha helix parameters
     phi = geom_calc.get_tor(a,b,c,d);
     psi = 57*(pi/180);
    } else if (!got_a && got_b && got_c && got_d && !got_e){
     // take alpha helix parameters
     phi = 57*(pi/180);
     psi = 57*(pi/180);
    } else {
     cerr << "GcemSiteCreator::create_sites: backbone atom is missing while generating rotamers for residue " << endl;
     if (!got_a)   cerr << endl << id_a << " is missing." << endl;
     if (!got_b)   cerr << endl << id_b << " is missing." << endl;
     if (!got_c)   cerr << endl << id_c << " is missing." << endl;
     if (!got_d)   cerr << endl << id_d << " is missing." << endl;
     if (!got_e)   cerr << endl << id_e << " is missing." << endl;
     if (!bb_cplt) cerr << " !bbcplt" << endl;
     continue;
    }


    // prepare temporary atomset with all sidechain torsion angles set to zero
    AtomSet zerochi = k->second.first;
    int nchi = top.sidechain_tor.size() - 1;
    if (nchi > int(3)) nchi = int(3);
    for (int ii=nchi; ii >= 0; --ii){
     Coord o, p, q, r;
     AtomID aido(i->first.get_site(),top.sidechain_tor[ii].first[0],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidp(i->first.get_site(),top.sidechain_tor[ii].first[1],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidq(i->first.get_site(),top.sidechain_tor[ii].first[2],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidr(i->first.get_site(),top.sidechain_tor[ii].first[3],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     if      (k->second.first.contains(aido)){o = zerochi.find(aido)->second.coord;}
     else if (k->second.second.contains(aido)){o = k->second.second.find(aido)->second.coord;}
     else {
      cerr << "GcemSiteCreator::create_sites: Atom " << aido << " is missing." << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidp)){p = zerochi.find(aidp)->second.coord;}
     else if (k->second.second.contains(aidp)){p = k->second.second.find(aidp)->second.coord;}
     else {
      cerr << "GcemSiteCreator::create_sites: Atom " << aidp << " is missing." << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidq)){q = zerochi.find(aidq)->second.coord;}
     else if (k->second.second.contains(aidq)){q = k->second.second.find(aidq)->second.coord;}
     else {
      cerr << "GcemSiteCreator::create_sites: Atom " << aidq << " is missing." << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidr)){r = zerochi.find(aidr)->second.coord;}
     else if (k->second.second.contains(aidr)){r = k->second.second.find(aidr)->second.coord;}
     else {
      cerr << "GcemSiteCreator::create_sites: Atom " << aidr << " is missing." << endl;
      ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
     }
     double chi = geom_calc.get_tor(o,p,q,r);
     // move all atoms indicated by CharmmParaResTop sidechain_tor
     for (list<string>::const_iterator jj=top.sidechain_tor[ii].second.begin();jj!=top.sidechain_tor[ii].second.end();++jj){
      AtomID aid(i->first.get_site(),*jj,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomSet::iterator kk;
      if (zerochi.contains(aid)) kk = zerochi.find(aid);
      else{
       cerr << "\nGcemSiteCreator::create_sites: Atom " << aid << " is missing." << endl;
       cerr << ("WARNING GcemSiteCreator::create_sites: atom is missing while generating rotamers.\n");
      }
      kk->second.coord = geom_calc.change_tor(o,p,q,kk->second.coord,-chi);
     }
    }

    /// debug ///
    if (molecule.get_print_debug()){
     if (!fs::is_directory("zerochi")){
      if (fs::exists("zerochi")) fs::remove_all("zerochi");
      fs::create_directory("zerochi");
     }
     stringstream ss,sc;
     string sstr, cstr;
     sc << i->first.get_chain(); sc >> cstr;
     ss << i->first.get_site(); ss >> sstr;
     zerochi.write_pqr(string("zerochi/" + i->first.get_sitename() + "-" + cstr
                                              + "-" + sstr + ".zerochi.pqr"));
    }
    /////////////


    // generate all the rotamers given in molecule.rlib, save them in newinst

    list<RotData> rlist(molecule.rlib[RotID(i->first.get_sitename(),geom_calc.get_tor_bin(phi),geom_calc.get_tor_bin(psi))]);

    int ninst = 0;

    for (list<RotData>::const_iterator ll = rlist.begin(); ll != rlist.end(); ++ll){
     ninst++;
     AtomSet tmpinst = zerochi;
     vector<double> chi;
     chi.push_back((ll->x1 * (pi / 180)));
     chi.push_back((ll->x2 * (pi / 180)));
     chi.push_back((ll->x3 * (pi / 180)));
     chi.push_back((ll->x4 * (pi / 180)));
     double propensity = ll->p;
     int nchi = top.sidechain_tor.size() - 1;
     if (nchi > int(3)) nchi = int(3);
     for (int ii=nchi; ii >= 0; --ii){
      Coord o, p, q, r;
      AtomID aido(i->first.get_site(),top.sidechain_tor[ii].first[0],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidp(i->first.get_site(),top.sidechain_tor[ii].first[1],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidq(i->first.get_site(),top.sidechain_tor[ii].first[2],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidr(i->first.get_site(),top.sidechain_tor[ii].first[3],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      if      (tmpinst.contains(aido)){o = tmpinst.find(aido)->second.coord;}
      else if (k->second.second.contains(aido)){o = k->second.second.find(aido)->second.coord;}
      else {
       cerr << "GcemSiteCreator::create_sites: Atom " << aido << " is missing." << endl;
       ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
      }
      if      (tmpinst.contains(aidp)){p = tmpinst.find(aidp)->second.coord;}
      else if (k->second.second.contains(aidp)){p = k->second.second.find(aidp)->second.coord;}
      else {
       cerr << "GcemSiteCreator::create_sites: Atom " << aidp << " is missing." << endl;
       ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
      }
      if      (tmpinst.contains(aidq)){q = tmpinst.find(aidq)->second.coord;}
      else if (k->second.second.contains(aidq)){q = k->second.second.find(aidq)->second.coord;}
      else {
       cerr << "GcemSiteCreator::create_sites: Atom " << aidq << " is missing." << endl;
       ::error ("ERROR in GcemSiteCreator::create_sites: atom is missing while generating rotamers.");
      }
      // move all atoms indicated by CharmmParaResTop sidechain_tor
      for (list<string>::const_iterator jj=top.sidechain_tor[ii].second.begin();jj!=top.sidechain_tor[ii].second.end();++jj){
       AtomID aid(i->first.get_site(),*jj,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
       AtomSet::iterator kk;
       if (tmpinst.contains(aid)) kk = tmpinst.find(aid);
       else{
        cerr << "\nGcemSiteCreator::create_sites: Atom " << aid << " is missing." << endl;
        cerr << ("WARNING GcemSiteCreator::create_sites: atom is missing while generating rotamers.\n");
       }
       kk->second.coord = geom_calc.change_tor(o,p,q,kk->second.coord,chi[ii]);
      }
     }
     // save tmpinst
     map<SiteID,map<int,pair<double, AtomSet> > >::iterator ni(new_inst.find(i->first));
     if (ni == new_inst.end()){
      pair<double,AtomSet> tmppair1(propensity, tmpinst);
      pair<int,pair<double,AtomSet> > tmppair2(ninst, tmppair1);
      map<int,pair<double,AtomSet> > tmpmap;
      tmpmap.insert(tmppair2);
      pair<SiteID,map<int,pair<double, AtomSet> > > tmppair3(i->first,tmpmap);
      new_inst.insert(tmppair3);
     } else {
      pair<double,AtomSet> tmppair1(propensity, tmpinst);
      pair<int,pair<double,AtomSet> > tmppair2(ninst, tmppair1);
      ni->second.insert(tmppair2);
     }

     /// debug ///
     if (molecule.get_print_debug()){
      if (!fs::is_directory("tmpinst")){
       if (fs::exists("tmpinst")) fs::remove_all("tmpinst");
       fs::create_directory("tmpinst");
      }
      stringstream sn,ss,sc;
      string nstr, sstr, cstr;
      sc << i->first.get_chain(); sc >> cstr;
      ss << i->first.get_site(); ss >> sstr;
      sn << (ninst-1); sn >> nstr;
      tmpinst.write_pqr(string("tmpinst/" + i->first.get_sitename() + "-" + cstr
                                              + "-" + sstr + "-" + nstr + ".tmpinst.pqr"));
     }
     /////////////


    }



   }
  }

  // finally create sites

  for (map<SiteID,int>::const_iterator i = molecule.rotmap.begin(); i!=molecule.rotmap.end(); ++i){

   bool has_rot = false;
   bool has_pdi = false;
   bool has_tsi = false;

   map<SiteID,map<int,pair<double, AtomSet> > >::const_iterator ni(new_inst.find(i->first));
   if (ni != new_inst.end()) has_rot = true;
   map<SiteID,map<int,AtomSet> >::const_iterator pdi(predef_inst.find(i->first));
   if (pdi != predef_inst.end()) has_pdi = true;
   map<SiteID,pair<AtomSet,AtomSet> >::const_iterator ts(tmpsites.find(i->first));
   if (ts != tmpsites.end() && ts->second.first.size() > 0) has_tsi = true;

   if (!has_tsi && !has_pdi){
     cerr << "GcemSiteCreator::create_sites:  Could not find site " << i->first << " in structure." << endl;
     ::error ("ERROR in GcemSiteCreator::create_sites: site is missing.");
   }

   GcemSite tmps(i->first);

   // first insert predefined instances

   if (has_pdi){
    if (i->second == 1){
     int ic = 0;
     int cn = 0;
     for(map<int,AtomSet>::const_iterator j=pdi->second.begin(); j!=pdi->second.end(); ++j){
      cn++;
      GcemSiteDef sdef(molecule.get_sitedef(i->first.get_sitename()));
      for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
       ic++;
       InstanceID tmpid(i->first.get_sitename(),"PD" + k->get_name(),i->first.get_conformer(),
                        i->first.get_chain(),i->first.get_site(),ic);
       map<string,int> tmpmap(k->ligandmap);
       AtomSet tmpa1;
       tmpa1.clear();
       // insert atoms found in GcemFormDef with listed charge in site struct
       for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
        // do not include atom with zero occupancy
        if (l->second.second == 0){continue;}
        AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),j->first);
        AtomID aid0(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),0);
        AtomSet::const_iterator aitr;
        if        (j->second.contains(aid)){
         aitr = j->second.find(aid);
        } else if (molecule.input_pqr.contains(aid0)){
         aitr = molecule.input_pqr.find(aid0);
        } else {
         cerr << i->first << "   " << aid  << endl;
         ::error ("ERROR in GcemSiteCreator::create_sites: atom in GcemFormDef missing in predefined site conformer.");
        }
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        tmpatom.charge = l->second.first;
        if (!tmpa1.contains(AtomID(tmpatom)))
         tmpa1.insert(tmpatom);
        // this will not happen
        else
         ::error ("ERROR in GcemSiteCreator::create_sites: duplicate atom in GcemFormDef.");
       }
       // also insert atoms not found in GcemFormDef but in the predefined instance in the site struct
       for (AtomSet::const_iterator aitr=j->second.begin(); aitr!=j->second.end(); ++aitr){
        if (aitr->second.instance!=cn) continue;
        AtomID aid(i->first.get_site(),aitr->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),ic);
        if (!tmpa1.contains(aid) && !k->contains_atom(aitr->second.resname,aitr->second.atname)){
         Atom tmpatom(aitr->second);
         tmpatom.instance = ic;
         tmpa1.insert(tmpatom);
        }
       }
       AtomSet tmpa2;
       if (has_tsi){
        tmpa2 = ts->second.second;
       }
       GcemInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,cn,k->get_gmodel());
       tmpi.set_is_predef((bool)true);
       tmpi.set_epsref(k->get_epsref());
       tmpi.set_istrref(k->get_istrref());
       tmps.insert_instance(tmpi);
      }
     }
    }

    if (i->second == 4){
     GcemSiteDef sdef(molecule.get_sitedef(i->first));
     list<GcemFormDef>::const_iterator k=sdef.forms.begin();     
     int ic = 0;
     for(map<int,AtomSet>::const_iterator j=pdi->second.begin(); j!=pdi->second.end(); ++j){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),k->get_name(),i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      map<string,int> tmpmap;
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      GcemInstance tmpi(tmpid,tmpmap,j->second,tmpa2,ic,k->get_gmodel());
      tmpi.set_is_predef((bool)true);
      tmpi.set_epsref(k->get_epsref());
      tmpi.set_istrref(k->get_istrref());
      tmpi.set_gmodel(k->get_gmodel());
      tmps.insert_instance(tmpi);
      k++;
     }
    }

    if (i->second == 2){
     int ic = 0;
     for(map<int,AtomSet>::const_iterator j=pdi->second.begin(); j!=pdi->second.end(); ++j){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),"predef",i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      map<string,int> tmpmap;
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      GcemInstance tmpi(tmpid,tmpmap,j->second,tmpa2,ic);
      tmpi.set_is_predef((bool)true);
      tmps.insert_instance(tmpi);
     }
    }

   }

   // now insert new instances

   if (has_rot){
    if (i->second == 1){
     int ic = 0;
     int cn = 0;
     if (has_pdi){
      ic = tmps.get_ninst();
      cn = tmps.get_max_confnum();
     }
     for(map<int,pair<double,AtomSet> >::const_iterator j=ni->second.begin(); j!=ni->second.end(); ++j){
      cn++;
      GcemSiteDef sdef(molecule.get_sitedef(i->first.get_sitename()));
      for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
       ic++;
       InstanceID tmpid(i->first.get_sitename(),k->get_name(),i->first.get_conformer(),
                        i->first.get_chain(),i->first.get_site(),ic);
       map<string,int> tmpmap(k->ligandmap);
       AtomSet tmpa1;
       // first insert all atoms found in GcemFormDef in site_struct
       for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
        // do not include atom with zero occupancy
        if (l->second.second == 0){continue;}
        AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),j->first);
        AtomID aid2(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
        AtomSet::const_iterator aitr;
        if (j->second.second.contains(aid2)){
         aitr = j->second.second.find(aid2);
        }
        //else if (j->second.second.contains(aid)){
        // aitr = j->second.second.find(aid);
        //}
        else if (has_tsi && ts->second.first.contains(aid2)){
         aitr = ts->second.first.find(aid2);
        }
        else {
         cerr << i->first << " " << aid << endl;
         ::error ("ERROR in GcemSiteCreator::create_sites: atom in GcemFormDef missing in generated site conformer.");
        }
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        tmpatom.charge = l->second.first;
        if (!tmpa1.contains(aid))
         tmpa1.insert(tmpatom);
        // this will not happen
        else
         ::error ("ERROR in GcemSiteCreator::create_sites: duplicate atom in GcemFormDef.");
       }
       // second insert all atoms not found in GcemFormDef but in the new instance
       // (moving atoms with constant charge) in site_struct
       for (AtomSet::const_iterator aitr=j->second.second.begin(); aitr!=j->second.second.end(); ++aitr){
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        AtomID aid(i->first.get_site(),tmpatom.atname,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),ic);
        if (!tmpa1.contains(aid) && !k->contains_atom(CharmmID(tmpatom)))
         tmpa1.insert(aitr->second);
       }
       AtomSet tmpa2;
       if (has_tsi){
        tmpa2 = ts->second.second;
       }
       GcemInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,cn,k->get_gmodel(),j->second.first);
       tmpi.set_epsref(k->get_epsref());
       tmpi.set_istrref(k->get_istrref());
       tmps.insert_instance(tmpi);
      }
     }
    }
    if (i->second == 2){
     int ic = 0;
     if (has_pdi) ic = tmps.get_ninst();
     for(map<int,pair<double,AtomSet> >::const_iterator j=ni->second.begin(); j!=ni->second.end(); ++j){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),"rotamer",i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      map<string,int> tmpmap;
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      GcemInstance tmpi(tmpid,tmpmap,j->second.second,tmpa2,ic,double(0),j->second.first);
      tmps.insert_instance(tmpi);
     }
    }

   }

   // unflexible, titratable sites

   if (!has_rot && has_tsi){
    int ic = 1;
    if (i->second == 3){
     GcemSiteDef sdef(molecule.get_sitedef(i->first.get_sitename()));
     for (list<GcemFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
//      ic++;
      InstanceID tmpid(i->first.get_sitename(),k->get_name(),i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      map<string,int> tmpmap(k->ligandmap);
      AtomSet tmpa1;
      for (map<CharmmID,pair<float, int> >::const_iterator l = k->atommap.begin(); l!=k->atommap.end();++l){
       // do not include atom with zero occupancy
       if (l->second.second == 0){continue;}
       AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
       AtomSet::const_iterator aitr;
       if (ts->second.first.contains(aid)){
        aitr = ts->second.first.find(aid);
       } else {
        cerr << i->first << " " << aid << endl;
        ::error ("ERROR in GcemSiteCreator::create_sites: atom in GcemFormDef missing in structure.");
       }
       Atom tmpatom(aitr->second);
       tmpatom.instance = ic;
       tmpatom.charge = l->second.first;
       if (!tmpa1.contains(aid))
        tmpa1.insert(tmpatom);
       // this will not happen
       else
        ::error ("ERROR in GcemSiteCreator::create_sites: duplicate atom in GcemFormDef.");
      }
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      GcemInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,ic,k->get_gmodel());
      tmpi.set_epsref(k->get_epsref());
      tmpi.set_istrref(k->get_istrref());
      tmps.insert_instance(tmpi);
     }
    }
   }

   // all instances should have been added, now insert site in list
   molecule.insert_site(tmps);

   //////////////// debug ////////////////
   if (molecule.get_print_debug()){
    tmps.print();
   }
   ///////////////////////////////////////

  }



  blab1 << "Exiting GcemSiteCreator::create_sites" << endl;
}


// GcemSiteCreator.cc ends here
