/*  A GCEM Site

    label, forms

    I/O functions for .est files

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemSiteDef.h"
#include "MEAD_e/globals.h"
#include <stdio.h>
#include <iomanip>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <fstream>
#include <sstream>
#include <iostream>

using std::istringstream;

using namespace std;
using namespace boost;

GcemSiteDef::GcemSiteDef(const GcemSiteDef& a)
: name(a.name), forms(a.forms), sid(a.sid)
{
}


GcemSiteDef::GcemSiteDef()
: name(""), sid()
{
}
GcemSiteDef::GcemSiteDef (const string& nm)
: name(nm), sid()
{
}
GcemSiteDef::GcemSiteDef (const string& nm, const SiteID& s)
: name(nm), sid(s)
{
}

void GcemSiteDef::read_est (){
 string fn = name + ".est";
 read_est(fn);
}

void GcemSiteDef::read_est (const string& fn_string){

  blab2 << "GcemSiteDef::read_est: starting for site named " << name << endl;

  const char *fn = fn_string.c_str();
  string line, tmpstr1, tmpstr2;
  string atname, resname;

  vector<string> form_labels;
  vector<string> ligand_labels;
  vector<vector<int> > ligands;
  vector<float> gmodels;
  vector<float> epsrefs;
  vector<float> istrrefs;
  vector<vector<float> > charges;
  vector<vector<int> > occups;
  vector<CharmmID> atoms;

  boost::regex re_comment("^(!|#|//|\\s*$|\\s*center)");
  boost::regex re_int("\\A\\d+\\z");
  boost::regex re_number("^([\\d.+-]+[eE]{1}[+-][\\d]{2}|[\\d.+-]+)$");
  boost::regex re_label("^\\s*(label|LABEL)\\s+\\w+");
  boost::regex re_gmod("^\\s*(Gmodel|gmodel|GMODEL)\\s+[\\d.+-eE]+");
  boost::regex re_epsref("^\\s*(epsref|Epsref|EpsRef|EPSREF)\\s+[\\d.+-eE]+");
  boost::regex re_istrref("^\\s*(istrref|Istrref|IstrRef|ISTRREF)\\s+[\\d.+-eE]+");
  boost::regex re_ligand("^\\s*\\w+\\s+\\d+(\\s+|\\n)");
  boost::regex re_atom("^\\s*\\w{1,4}\\s+\\w{1,4}\\s+([\\d.+-eE]+|NaN|[^\\d^\\s]+)(\\s+|\\n)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  ifstream topfile(fn);

  if (!topfile)
    ::error ("GcemSiteDef::read_est: Cannot open input file, ", fn, "\n");

  int atcount  = 0;
  int ligcount = 0;

  // start of parsing
  while (! topfile.eof() ){

   bool inlabel   = false;
   bool ingmod    = false;
   bool inepsref  = false;
   bool inistrref = false;
   bool inlig     = false;
   bool inatoms   = false;
   bool matched   = false;

   std::string::const_iterator start, end;

   getline (topfile, line);
   start = line.begin();
   end   = line.end();

   if (boost::regex_search(start, end, match, re_atom, flags)){
    inlabel = false;  ingmod = false; inepsref = false; inistrref = false; inlig = false; inatoms = true;  matched = true;
   }
   if (boost::regex_search(start, end, match, re_label, flags)){
    inlabel = true;   ingmod = false; inepsref = false; inistrref = false; inlig = false; inatoms = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_gmod, flags)){
    inlabel = false;  ingmod = true;  inepsref = false; inistrref = false; inlig = false; inatoms = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_epsref, flags)){
    inlabel = false;  ingmod = false; inepsref = true; inistrref = false;  inlig = false; inatoms = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_istrref, flags)){
    inlabel = false;  ingmod = false; inepsref = false; inistrref = true;  inlig = false; inatoms = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_ligand, flags)){
    if (!ingmod && !inlabel && !inepsref){
     inlabel = false; ingmod = false; inepsref = false; inistrref = false; inlig = true;  inatoms = false; matched = true;
    }
   }
   if (boost::regex_search(start, end, match, re_comment, flags)){
    inlabel = false; ingmod = false;  inepsref = false; inistrref = false; inlig = false; inatoms = false; matched = true;
   }
   if (!matched){
    cerr << "Error: GcemSiteDef::read_est: could not match line, use # or ! or // for comments\n"
         << line << endl;
    ::error ("GcemSiteDef::read_est: error in parsing of extended sites file, ", fn, "\n");
   }

   if (inlabel){
    stringstream ss1(line);
    string tmpstr;
    int i = 0;
    while (ss1){
     ss1 >> tmpstr;
     if (ss1.fail()){break;}
    // first token is "label"
     if (i > 0){
      for (vector<string>::const_iterator fitr = form_labels.begin(); fitr != form_labels.end(); ++fitr){
       if (*fitr == tmpstr){
        cerr << "Error: GcemSiteDef::read_est: form label " << tmpstr << " occurred multiple times." << endl;
        ::error ("GcemSiteDef::read_est: error in parsing of extended sites file, ", fn, "\n");
       }
      }
      form_labels.push_back(tmpstr);
     }
     i++;
    }
    inlabel = false;
   }

   if (ingmod){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "gmodel"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      gmodels.push_back(tmpf);
     } else {
       cerr << "Error: GcemSiteDef::read_est: could not read Gmodel in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_est: error in reading Gmodel in parsing of extended sites file, ", fn, "\n");
     }
    }
    ingmod = false;
   }

   if (inepsref){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "epsref"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      epsrefs.push_back(tmpf);
     } else {
       cerr << "Error: GcemSiteDef::read_est: could not read EpsRef in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_est: error in reading EpsRef in parsing of extended sites file, ", fn, "\n");
     }
    }
    inepsref = false;
   }

   if (inistrref){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "istrref"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      istrrefs.push_back(tmpf);
     } else {
       cerr << "Error: GcemSiteDef::read_est: could not read IstrRef in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_est: error in reading IstrRef in parsing of extended sites file, ", fn, "\n");
     }
    }
    inistrref = false;
   }

   if (inlig){
    vector<string> tmpvec;
    vector<int> tmpvec2;
    stringstream ss1(line);
    string tmpstr1, tmpstr2;
    ss1 >> tmpstr1;
    ligand_labels.push_back(tmpstr1);
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start2, end2;
     start2 = tmpstr2.begin();
     end2   = tmpstr2.end();
     if (boost::regex_search(start2, end2, match, re_int, flags)){
      int tmpi;
      stringstream ss2(tmpstr2);
      ss2 >> tmpi;
      tmpvec2.push_back(tmpi);
     } else {
       int tmpi;
       stringstream ss2(tmpstr2);
       ss2 >> tmpi;
       cerr << "Error: GcemSiteDef::read_est: could not read number of ligands in file "
            << fn_string << " for ligand " << tmpstr1 << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_est: error in reading number of ligands in parsing of extended sites file, ", fn, "\n");
     }
    }
    ligands.push_back(tmpvec2);
    ligcount++;
    inlig = false;
   }

   if (inatoms){
    vector<float> tmpvec2;
    vector<int> tmpvec3;
    string tmpstr1, tmpstr2, tmpstr3;
    stringstream ss1(line);
    // first two tokens are resname atname
    ss1 >> tmpstr1 >> tmpstr2;
    atoms.push_back(CharmmID(tmpstr1, tmpstr2));
    while(ss1){
     ss1 >> tmpstr3;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr3.begin();
     end1   = tmpstr3.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr3);
      float tmpf;
      ss2 >> tmpf;
      tmpvec2.push_back(tmpf);
      tmpvec3.push_back(int(1));
     } else {
      tmpvec2.push_back(float(0));
      tmpvec3.push_back(int(0));
     }
    }
    charges.push_back(tmpvec2);
    occups.push_back(tmpvec3);
    atcount++;
    inatoms = false;
   }

  matched = false;

  } // end of parsing

  topfile.close();

  // begin debug
  /*
    cout << "form_labels : ";
    for (int i=0; i<form_labels.size(); i++){
     cout << form_labels[i] << " ";
    }    cout << endl;
    cout << "Gmodels : ";
    for (int i=0; i<gmodels.size(); i++){
     cout << gmodels[i] << " ";
    }    cout << endl;
    cout << "EpsRefs : ";
    for (int i=0; i<epsrefs.size(); i++){
     cout << epsrefs[i] << " ";
    }    cout << endl;
    cout << "IstrRefs : ";
    for (int i=0; i<istrrefs.size(); i++){
     cout << istrrefs[i] << " ";
    }    cout << endl;
    cout << "ligands : " << endl;
    for (int i=0; i<ligand_labels.size(); i++){
     cout << ligand_labels[i] << " ";
     for (int j=0; j<form_labels.size(); j++){
       cout << (ligands[i])[j] << " ";
     } cout << endl;
    }

    cout << "charges : " << endl;

    int tmpi1 = atoms.size() * form_labels.size();
    int tmpi2 = 0;
    for (int i=0; i<charges.size(); i++){
     for (int j=0; j<(charges[i]).size(); j++){
      tmpi2++;
      printf ("%.5f ", (charges[i])[j]);
     }
     cout << endl;
    }

    cout << "occupancies : " << endl;

    int tmpi3 = 0;
    for (int i=0; i<occups.size(); i++){
     for (int j=0; j<(occups[i]).size(); j++){
      tmpi3++;
      cout << (occups[i])[j] << " ";
     }
     cout << endl;
    }
  */
  // end debug

  bool verified = true;
  if (form_labels.size() != gmodels.size()) {
    cerr << "ERROR: GcemSiteDef::read_est: number of form labels = " << form_labels.size()
         << ", while number of model energies = " << gmodels.size() << endl;
    verified = false;
  }
  if (form_labels.size() != epsrefs.size() && epsrefs.size() != 0) {
    cerr << "ERROR: GcemSiteDef::read_est: number of form labels = " << form_labels.size()
         << ", while number of reference dielectric constants = " << epsrefs.size()
         << ", EpsRef must be defined for all forms or for none" << endl;
    verified = false;
  }
  if (form_labels.size() != istrrefs.size() && istrrefs.size() != 0) {
    cerr << "ERROR: GcemSiteDef::read_est: number of form labels = " << form_labels.size()
         << ", while number of reference ionic strengths = " << istrrefs.size()
         << ", IstrRef must be defined for all forms or for none" << endl;
    verified = false;
  }
  if (atcount != atoms.size()) {
    cerr << "ERROR: GcemSiteDef::read_est: Map length of atoms = " << atoms.size()
         << ", while in reading atom count = " << atcount << endl;
    verified = false;
  }
  if (atcount != charges.size()) {
    cerr << "ERROR: GcemSiteDef::read_est: number of charges = " << charges.size()
         << ", while in reading atom count = " << atcount << endl;
    verified = false;
  }
  for (int i=0; i < charges.size(); i++){
   if ((charges[i]).size() != form_labels.size()){
    cerr << "ERROR: GcemSiteDef::read_est: number of charge records for atom "
         << atoms[i] << " = " << (charges[i]).size()
         << " while number of forms = " << form_labels.size() << endl;
    verified = false;
   }
  }
  for (int i=0; i < ligands.size(); i++){
   if ((ligands[i]).size() != form_labels.size()){
    cerr << "ERROR: GcemSiteDef::read_est: number of records of number of ligands bound = "
         << (ligands[i]).size() << " while number of forms = " << form_labels.size() << endl;
    verified = false;
   }
  }
  if (! verified){
   ::error ("GcemSiteDef::read_est: error in size verification, ", fn, "\n");
  }

  // epsref must be defined for all forms or for none, if undefined set it to -1
  if (epsrefs.size() != gmodels.size()){
   epsrefs.clear();
   for (int i=0; i<gmodels.size(); i++){
    epsrefs.push_back(-1);
   }
  }

  // istrref must be defined for all forms or for none, if undefined set it to -1
  if (istrrefs.size() != gmodels.size()){
   istrrefs.clear();
   for (int i=0; i<gmodels.size(); i++){
    istrrefs.push_back(-1);
   }
  }

  // now insert everything in a GcemSiteDef and its GcemFormDefs

  for (int i=0; i<gmodels.size(); i++){

   GcemFormDef tmpform(form_labels[i], gmodels[i], epsrefs[i], istrrefs[i]);

   for (int j=0; j<ligand_labels.size(); j++){
    tmpform.insert_ligand(ligand_labels[j], (ligands[j])[i]);
   }

   for (int j=0; j<atoms.size(); j++){
    tmpform.insert_atom(atoms[j], (charges[j])[i], (occups[j])[i]);
   }

   forms.push_back(tmpform);

  }

  blab1 << "GcemSiteDef::read_est: finished file " << fn_string
        << " with " << forms.size() << " forms, " << atcount
        << " atoms and " << ligcount << " ligands read." << endl;

}


void GcemSiteDef::print_est(){

  blab2 << "GcemSiteDef::print_est: started " << endl;

     list<string> form_lbl;
     list<string> ligand_lbl;
     list<float> gmodels;
     list<float> epsrefs;
     list<float> istrrefs;
     list<CharmmID> atoms;
     string resname, atname;
     // first collect ligands, atoms model energies

     for (std::list<GcemFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i) {
      form_lbl.push_back(i->get_name());
      gmodels.push_back(i->get_gmodel());
      epsrefs.push_back(i->get_epsref());
      istrrefs.push_back(i->get_istrref());
      if (i == forms.begin()){
       for (std::map<string,int>::const_iterator j = i->ligandmap.begin(); j != i->ligandmap.end(); ++j) {
        ligand_lbl.push_back(j->first);
       }
       for (std::map<CharmmID,pair<float,int> >::const_iterator j = i->atommap.begin(); j != i->atommap.end(); ++j) {
        atoms.push_back(j->first);
       }
      }
     }
     // with known sizes declare arrays for ligands, atomic partial charges and atomic occupancy
     float charges[atoms.size()][form_lbl.size()];
     int occupancy[atoms.size()][form_lbl.size()];
     int ligands[ligand_lbl.size()][form_lbl.size()];
     // fill arrays
     int jj = 0;
     for (std::list<GcemFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i) {
      int ii = 0;
      for (std::map<string,int>::const_iterator j = i->ligandmap.begin(); j != i->ligandmap.end(); ++j) {
       ligands[ii][jj] = j->second;
       ii++;
      }
      ii = 0;
      for (std::map<CharmmID,pair<float,int> >::const_iterator j = i->atommap.begin(); j != i->atommap.end(); ++j) {
       charges[ii][jj]   = (j->second).first;
       occupancy[ii][jj] = (j->second).second;
       ii++;
      }
      jj++;
     }
     // print data like in .est file
     // form labels
     string line = "label        ";
     for (std::list<std::string>::const_iterator i = form_lbl.begin(); i != form_lbl.end(); ++i){
      line = line + " " + (*i);
      for (int ii=0; ii < (11 - i->length()); ++ii){
       line = line + " ";
      }
     }
     cout << line << endl;
     line.erase();

     // Gmodel line
     line = "Gmodel    ";
     for (std::list<float>::const_iterator i = gmodels.begin(); i != gmodels.end(); ++i){
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     cout << line << endl;
     line.erase();

     // EpsRef line
     line = "EpsRef    ";
     for (std::list<float>::const_iterator i = epsrefs.begin(); i != epsrefs.end(); ++i){
      if ((*i) == -1){
       line.erase();
       break;
      }
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     if (line.length() != 0){
      cout << line << endl;
      line.erase();
     }

     // IstrRef line
     line = "EpsRef    ";
     for (std::list<float>::const_iterator i = istrrefs.begin(); i != istrrefs.end(); ++i){
      if ((*i) == -1){
       line.erase();
       break;
      }
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     if (line.length() != 0){
      cout << line << endl;
      line.erase();
     }

     // ligand binding lines
     jj = 0;
     for (std::list<std::string>::const_iterator j = ligand_lbl.begin(); j != ligand_lbl.end(); ++j){
      line = (*j);
      for (int ii=0; ii < (10 - j->length()); ++ii){
       line = line + " ";
      }
      int kk = 0;
      for (std::list<std::string>::const_iterator i = form_lbl.begin(); i != form_lbl.end(); ++i){
       char buffer[512];
       sprintf(buffer, "%12d", ligands[jj][kk]);
       line = line + string(buffer);
       kk++;
      }
      cout << line << endl;
      line.erase();
     }

     // atom lines
     jj = 0;
     for (std::list<CharmmID>::const_iterator j = atoms.begin(); j != atoms.end(); ++j){
      char buf[256];
      sprintf(buf, "%-4s %-4s ", (j->get_resname()).c_str(), (j->get_atname()).c_str());
      line = string(buf);
      int kk = 0;
      for (std::list<std::string>::const_iterator i = form_lbl.begin(); i != form_lbl.end(); ++i){
       if (occupancy[jj][kk] == int(1)){
        char buffer[512];
        sprintf(buffer, "%12.6f", charges[jj][kk]);
        line = line + string(buffer);
       } else {line = line + "         NaN";}
       kk++;
      }
      cout << line << endl;
      line.erase();
      jj++;
     }

  blab1 << "GcemSiteDef::print_est: finished " << endl;

}

void GcemSiteDef::read_qst (){
 string fn = name + ".qst";
 // check if name.qst exists, otherwise remove confid
 read_qst(fn);
}

void GcemSiteDef::read_qst (const string &fn_string){

  blab2 << "GcemSiteDef::read_qst: starting for site named " << name << endl;

  const char *fn = fn_string.c_str();
  string line, tmpstr1, tmpstr2;
  string atname, resname;

  vector<string> form_labels;
  vector<string> ligand_labels;
  vector<vector<int> > ligands;
  vector<float> gmodels;
  vector<float> epsrefs;
  vector<float> istrrefs;
  multimap<AtomID, AtomID> links;

  boost::regex re_comment("^(!|#|//|\\s*$|\\s*center)");
  boost::regex re_int("\\A\\d+\\z");
  boost::regex re_number("^([\\d.+-]+[eE]{1}[+-][\\d]{2}|[\\d.+-]+)$");
  boost::regex re_label("^\\s*(label|LABEL)\\s+\\w+");
  boost::regex re_gmod("^\\s*(Gmodel|gmodel|GMODEL)\\s+[\\d.+-eE]+");
  boost::regex re_link("^\\s*(link|Link|LINK)\\s+([^\\s]+)\\s+([^\\s]+)");
  boost::regex re_epsref("^\\s*(epsref|Epsref|EpsRef|EPSREF)\\s+[\\d.+-eE]+");
  boost::regex re_istrref("^\\s*(istrref|Istrref|IstrRef|ISTRREF)\\s+[\\d.+-eE]+");
  boost::regex re_ligand("^\\s*\\w+\\s+\\d+(\\s+|\\n)");
  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  ifstream topfile(fn);

  if (!topfile)
    ::error ("GcemSiteDef::read_qst: Cannot open input file, ", fn, "\n");

  int atcount  = 0;
  int ligcount = 0;

  // start of parsing
  while (! topfile.eof() ){

   bool inlabel   = false;
   bool ingmod    = false;
   bool inepsref  = false;
   bool inistrref = false;
   bool inlig     = false;
   bool inlink    = false;
   bool matched   = false;

   std::string::const_iterator start, end;

   getline (topfile, line);
   start = line.begin();
   end   = line.end();

   if (boost::regex_search(start, end, match, re_link, flags)){
    inlabel = false;  ingmod = false; inepsref = false; inistrref = false; inlig = false; inlink = true;  matched = true;
   }
   if (boost::regex_search(start, end, match, re_label, flags)){
    inlabel = true;   ingmod = false; inepsref = false; inistrref = false; inlig = false; inlink = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_gmod, flags)){
    inlabel = false;  ingmod = true;  inepsref = false; inistrref = false; inlig = false; inlink = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_epsref, flags)){
    inlabel = false;  ingmod = false; inepsref = true; inistrref = false;  inlig = false; inlink = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_istrref, flags)){
    inlabel = false;  ingmod = false; inepsref = false; inistrref = true;  inlig = false; inlink = false; matched = true;
   }
   if (boost::regex_search(start, end, match, re_ligand, flags)){
    if (!ingmod && !inlabel && !inepsref){
     inlabel = false; ingmod = false; inepsref = false; inistrref = false; inlig = true;  inlink = false; matched = true;
    }
   }
   if (boost::regex_search(start, end, match, re_comment, flags)){
    inlabel = false; ingmod = false;  inepsref = false; inistrref = false; inlig = false; inlink = false; matched = true;
   }
   if (!matched){
    blab3 << "WARNING GcemSiteDef::read_qst: could not match line, use # or ! or // for comments\n"
         << line << endl;
    ::error ("GcemSiteDef::read_qst: error in parsing of QM sites file, ", fn, "\n");
   }

   if (inlabel){
    stringstream ss1(line);
    string tmpstr;
    int i = 0;
    while (ss1){
     ss1 >> tmpstr;
     if (ss1.fail()){break;}
    // first token is "label"
     if (i > 0){
      form_labels.push_back(tmpstr);
     }
     i++;
    }
    inlabel = false;
   }

   if (ingmod){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "gmodel"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      gmodels.push_back(tmpf);
     } else {
       blab3 << "WARNING GcemSiteDef::read_qst: could not read Gmodel in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_qst: error in reading Gmodel in parsing of QM sites file, ", fn, "\n");
     }
    }
    ingmod = false;
   }

   if (inepsref){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "epsref"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      istringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      epsrefs.push_back(tmpf);
     } else {
       blab3 << "WARNING GcemSiteDef::read_qst: could not read EpsRef in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_qst: error in reading EpsRef in parsing of QM sites file, ", fn, "\n");
     }
    }
    inepsref = false;
   }

   if (inistrref){
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    float tmpf;
    // first token is "istrref"
    ss1 >> tmpstr1;
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start1, end1;
     start1 = tmpstr2.begin();
     end1   = tmpstr2.end();
     if (boost::regex_search(start1, end1, match, re_number, flags)){
      stringstream ss2(tmpstr2);
      float tmpf;
      ss2 >> tmpf;
      istrrefs.push_back(tmpf);
     } else {
       blab3 << "WARNING GcemSiteDef::read_qst: could not read IstrRef in file "
            << fn_string << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_qst: error in reading IstrRef in parsing of QM sites file, ", fn, "\n");
     }
    }
    inistrref = false;
   }

   if (inlig){
    vector<string> tmpvec;
    vector<int> tmpvec2;
    istringstream ss1(line);
    string tmpstr1, tmpstr2;
    ss1 >> tmpstr1;
    ligand_labels.push_back(tmpstr1);
    while (ss1){
     ss1 >> tmpstr2;
     if (ss1.fail()){break;}
     std::string::const_iterator start2, end2;
     start2 = tmpstr2.begin();
     end2   = tmpstr2.end();
     if (boost::regex_search(start2, end2, match, re_int, flags)){
      int tmpi;
      istringstream ss2(tmpstr2);
      ss2 >> tmpi;
      tmpvec2.push_back(tmpi);
     } else {
       int tmpi;
       stringstream ss2(tmpstr2);
       ss2 >> tmpi;
       cerr << "WARNING GcemSiteDef::read_qst: could not read number of ligands in file "
            << fn_string << " for ligand " << tmpstr1 << " input " << tmpstr2 << endl;
       ::error ("GcemSiteDef::read_qst: error in reading number of ligands in parsing of QM sites file, ", fn, "\n");
     }
    }
    ligands.push_back(tmpvec2);
    ligcount++;
    inlig = false;
   }

   if (inlink){
    // pattern : link AtomID1 AtomID2
    istringstream ss1(line);
    // the link keyword
    string tmpstr;
    // the atoms forming the linking bond
    AtomID aid1, aid2;
    ss1 >> tmpstr >> aid1 >> aid2;
    pair<AtomID, AtomID> tmplink(aid1, aid2);
    if (!ss1.fail() && !ss1.bad()){
     links.insert(tmplink);
    }
    inlink = false;
   }

   matched = false;

  } // end of parsing

  topfile.close();

  // begin debug
  /*
    cout << "form_labels : ";
    for (int i=0; i<form_labels.size(); i++){
     cout << form_labels[i] << " ";
    }    cout << endl;
    cout << "Gmodels : ";
    for (int i=0; i<gmodels.size(); i++){
     cout << gmodels[i] << " ";
    }    cout << endl;
    cout << "EpsRefs : ";
    for (int i=0; i<epsrefs.size(); i++){
     cout << epsrefs[i] << " ";
    }    cout << endl;
    cout << "IstrRefs : ";
    for (int i=0; i<istrrefs.size(); i++){
     cout << istrrefs[i] << " ";
    }    cout << endl;
    cout << "ligands : " << endl;
    for (int i=0; i<ligand_labels.size(); i++){
     cout << ligand_labels[i] << " ";
     for (int j=0; j<form_labels.size(); j++){
       cout << (ligands[i])[j] << " ";
     } cout << endl;
    }

  */
  // end debug

  bool verified = true;
  if (form_labels.size() != gmodels.size()) {
    cerr << "WARNING GcemSiteDef::read_qst: number of form labels = " << form_labels.size()
         << ", while number of model energies = " << gmodels.size() << endl;
    verified = false;
  }
  if (form_labels.size() != epsrefs.size() && epsrefs.size() != 0) {
    cerr << "WARNING GcemSiteDef::read_qst: number of form labels = " << form_labels.size()
         << ", while number of reference dielectric constants = " << epsrefs.size()
         << ", EpsRef must be defined for all forms or for none" << endl;
    verified = false;
  }
  if (form_labels.size() != istrrefs.size() && istrrefs.size() != 0) {
    cerr << "WARNING GcemSiteDef::read_qst: number of form labels = " << form_labels.size()
         << ", while number of reference ionic strenths = " << istrrefs.size()
         << ", IstrRef must be defined for all forms or for none" << endl;
    verified = false;
  }
  for (int i=0; i < ligands.size(); i++){
   if ((ligands[i]).size() != form_labels.size()){
    cerr << "WARNING GcemSiteDef::read_qst: number of records of number of ligands bound = "
         << (ligands[i]).size() << " while number of forms = " << form_labels.size() << endl;
    verified = false;
   }
  }
  if (! verified){
   ::error ("GcemSiteDef::read_qst: error in size verification, ", fn, "\n");
  }

  // epsref must be defined for all forms or for none, if undefined set it to -1
  if (epsrefs.size() != gmodels.size()){
   epsrefs.clear();
   for (int i=0; i<gmodels.size(); i++){
    epsrefs.push_back(-1);
   }
  }
  // istrref must be defined for all forms or for none, if undefined set it to -1
  if (istrrefs.size() != gmodels.size()){
   istrrefs.clear();
   for (int i=0; i<gmodels.size(); i++){
    istrrefs.push_back(-1);
   }
  }

  // now insert everything in a GcemSiteDef and its GcemFormDefs

  for (int i=0; i<gmodels.size(); i++){

   GcemFormDef tmpform(form_labels[i], gmodels[i], epsrefs[i], istrrefs[i]);

   for (multimap<AtomID, AtomID>::const_iterator j=links.begin(); j!=links.end(); j++){
    tmpform.insert_link(j->first,j->second);
   }

   for (int j=0; j<ligand_labels.size(); j++){
    tmpform.insert_ligand(ligand_labels[j], (ligands[j])[i]);
   }

   forms.push_back(tmpform);

  }

  blab1 << "GcemSiteDef::read_qst: finished file " << fn_string << " with "
        << forms.size() << " forms, " << links.size() << " links and "
        << ligcount << " ligands read." << endl;

}

void GcemSiteDef::print_qst(){

  blab2 << "GcemSiteDef::print_qst: started " << endl;

     list<string> form_lbl;
     list<string> ligand_lbl;
     list<float> gmodels;
     list<float> epsrefs;
     list<float> istrrefs;
     // first collect ligands, atoms model energies

     for (std::list<GcemFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i) {
      form_lbl.push_back(i->get_name());
      gmodels.push_back(i->get_gmodel());
      epsrefs.push_back(i->get_epsref());
      if (i == forms.begin()){
       for (std::map<string,int>::const_iterator j = i->ligandmap.begin(); j != i->ligandmap.end(); ++j) {
        ligand_lbl.push_back(j->first);
       }
      }
     }
     // with known sizes declare arrays for ligands, atomic partial charges and atomic occupancy
     int ligands[ligand_lbl.size()][form_lbl.size()];
     // fill arrays
     int jj = 0;
     for (std::list<GcemFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i) {
      int ii = 0;
      for (std::map<string,int>::const_iterator j = i->ligandmap.begin(); j != i->ligandmap.end(); ++j) {
       ligands[ii][jj] = j->second;
       ii++;
      }
      jj++;
     }
     // print data like in .qst file
     // form labels
     string line = "label        ";
     for (std::list<std::string>::const_iterator i = form_lbl.begin(); i != form_lbl.end(); ++i){
      line = line + " " + (*i);
      for (int ii=0; ii < (11 - i->length()); ++ii){
       line = line + " ";
      }
     }
     cout << line << endl;
     line.erase();

     // Gmodel line
     line = "Gmodel    ";
     for (std::list<float>::const_iterator i = gmodels.begin(); i != gmodels.end(); ++i){
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     cout << line << endl;
     line.erase();

     // EpsRef line
     line = "EpsRef    ";
     for (std::list<float>::const_iterator i = epsrefs.begin(); i != epsrefs.end(); ++i){
      if ((*i) == -1){
       line.erase();
       break;
      }
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     if (line.length() != 0){
      cout << line << endl;
      line.erase();
     }

     // IstrRef line
     line = "IstrRef    ";
     for (std::list<float>::const_iterator i = istrrefs.begin(); i != istrrefs.end(); ++i){
      if ((*i) == -1){
       line.erase();
       break;
      }
      char buffer[512];
      sprintf(buffer, "%12.6f", (*i));
      line = line + string(buffer);
     }
     if (line.length() != 0){
      cout << line << endl;
      line.erase();
     }

     // ligand binding lines
     jj = 0;
     for (std::list<std::string>::const_iterator j = ligand_lbl.begin(); j != ligand_lbl.end(); ++j){
      line = (*j);
      for (int ii=0; ii < (10 - j->length()); ++ii){
       line = line + " ";
      }
      int kk = 0;
      for (std::list<std::string>::const_iterator i = form_lbl.begin(); i != form_lbl.end(); ++i){
       char buffer[512];
       sprintf(buffer, "%12d", ligands[jj][kk]);
       line = line + string(buffer);
       kk++;
      }
      cout << line << endl;
      line.erase();
     }
     
     // not formated in form columns
     for (std::list<GcemFormDef>::const_iterator i = forms.begin(); i != forms.end(); ++i){
      for (std::multimap<AtomID, AtomID>::const_iterator j = i->linkmap.begin(); j != i->linkmap.end(); ++j){
       cout << "link      " << j->first << "   " << j->second << endl;
      }
      break;
     }

  blab1 << "GcemSiteDef::print_qst: finished " << endl;

}

// GcemSiteDef.cc ends here
