/*  
    helper object for GcemMolecule
    creates the MEAD solver input for the calculation
    of electrostatic energies

    parsing functions for output files of
    my_2diel_solver
    my_Ndiel_solver, my_3diel_solver, my_memb_solver
    my_membpot_solver

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemSolverInput.h"

using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::stringstream;
using std::string;

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;


pair<FinDiffMethod,FinDiffMethod> GcemSolverInput::get_fdm(const string& fn, const GcemMolecule &molecule) const
{
  string ogm_name(fn + ".ogm");
  string mgm_name(fn + ".mgm");
  // first check existence of files .ogm .mgm
  ifstream ogm_fs(ogm_name.c_str()); bool got_ogm = ogm_fs.good(); ogm_fs.close();
  ifstream mgm_fs(mgm_name.c_str()); bool got_mgm = mgm_fs.good(); mgm_fs.close();
  FinDiffMethod fdm_ogm, fdm_mgm;
  // read existing files
  if (got_ogm && got_mgm){
   fdm_ogm.read(ogm_name);
   fdm_mgm.read(mgm_name);
  }
  //otherwise create grids
  else {

   // dimensions of whole molecule
   Coord gmax(-std::numeric_limits<float>::max(),
              -std::numeric_limits<float>::max(),
              -std::numeric_limits<float>::max());
   Coord gmin( std::numeric_limits<float>::max(),
               std::numeric_limits<float>::max(),
               std::numeric_limits<float>::max());

   for (AtomSet::const_iterator i=molecule.input_pqr.begin(); i!=molecule.input_pqr.end();++i){
    if (i->second.coord.x < gmin.x) gmin.x = i->second.coord.x;
    if (i->second.coord.y < gmin.y) gmin.y = i->second.coord.y;
    if (i->second.coord.z < gmin.z) gmin.z = i->second.coord.z;
    if (i->second.coord.x > gmax.x) gmax.x = i->second.coord.x;
    if (i->second.coord.y > gmax.y) gmax.y = i->second.coord.y;
    if (i->second.coord.z > gmax.z) gmax.z = i->second.coord.z;
   }

   Coord gdim = gmax - gmin;

   // dimensions of largest site (model_struct)

   Coord mdim(0,0,0);

   for (list<GcemSite>::const_iterator k=molecule.sites.begin(); k!=molecule.sites.end(); ++k){
    for (list<GcemInstance>::const_iterator j=k->instances.begin(); j!=k->instances.end(); ++j){
     Coord mmax(-std::numeric_limits<float>::max(),
                -std::numeric_limits<float>::max(),
                -std::numeric_limits<float>::max());
     Coord mmin( std::numeric_limits<float>::max(),
                 std::numeric_limits<float>::max(),
                 std::numeric_limits<float>::max());
     for (AtomSet::const_iterator i=j->model_struct.begin(); i!=j->model_struct.end();++i){
      if (i->second.coord.x < mmin.x) mmin.x = i->second.coord.x;
      if (i->second.coord.y < mmin.y) mmin.y = i->second.coord.y;
      if (i->second.coord.z < mmin.z) mmin.z = i->second.coord.z;
      if (i->second.coord.x > mmax.x) mmax.x = i->second.coord.x;
      if (i->second.coord.y > mmax.y) mmax.y = i->second.coord.y;
      if (i->second.coord.z > mmax.z) mmax.z = i->second.coord.z;
     }
     Coord tmpdim = mmax - mmin;
     if (geom_calc.abs_value(mdim) < geom_calc.abs_value(tmpdim)) mdim = tmpdim;
    }
   }

   double gmaxdim = gdim.x;
   if (gdim.y > gmaxdim) gmaxdim = gdim.y;
   if (gdim.z > gmaxdim) gmaxdim = gdim.z;

   double mmaxdim = mdim.x;
   if (mdim.y > mmaxdim) mmaxdim = mdim.y;
   if (mdim.z > mmaxdim) mmaxdim = mdim.z;

   // now add 3 or 4 fdm levels to mgm and ogm
   if        (molecule.get_grid_detail() == 0) {
    // coarse, solvation and interaction energies will be afflicted with grid discretization error
    double s1 = 2.00;
    double s2 = 1.00;
    double s3 = 0.50;

    double d1 = gmaxdim + (mmaxdim * 0.5f) + 40;
    double d2 = mmaxdim + 30;
    double d3 = mmaxdim + 20;

    // outer grids must be larger than inner grids
    if (d3 > (d2 + s2)) {d2 = d3 + s2;}
    if (d2 > (d1 + s1)) {d1 = d2 + s1;}

    int n1 = static_cast<int>(floor(d1/s1));
    int n2 = static_cast<int>(floor(d2/s2));
    int n3 = static_cast<int>(floor(d3/s3));

    // make grid size odd
    if (n1 % 2 == 0) n1++;
    if (n2 % 2 == 0) n2++;
    if (n3 % 2 == 0) n3++;

    // add grid levels to fdm
    fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);
   } else if (molecule.get_grid_detail() == 1) {
    // settings similar to old multiflex examples, ensures fairly converged solvation and interaction energies
    double s1 = 2.00;
    double s2 = 1.00;
    double s3 = 0.25;

    double d1 = gmaxdim + (mmaxdim * 0.5f) + 40;
    double d2 = mmaxdim + 30;
    double d3 = mmaxdim + 20;

    // outer grids must be larger than inner grids
    if (d3 > (d2 + s2)) {d2 = d3 + s2;}
    if (d2 > (d1 + s1)) {d1 = d2 + s1;}

    int n1 = static_cast<int>(floor(d1/s1));
    int n2 = static_cast<int>(floor(d2/s2));
    int n3 = static_cast<int>(floor(d3/s3));

    // make grid size odd
    if (n1 % 2 == 0) n1++;
    if (n2 % 2 == 0) n2++;
    if (n3 % 2 == 0) n3++;

    // add grid levels to fdm
    fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);
   } else if (molecule.get_grid_detail() == 2) {
    // somewhat finer than before, ensures high quality solvation and interaction energies 
    double s1 = 2.00;
    double s2 = 0.50;
    double s3 = 0.20;

    double d1 = gmaxdim + (mmaxdim * 0.5f) + 80;
    double d2 = mmaxdim + 60;
    double d3 = mmaxdim + 40;

    // outer grids must be larger than inner grids
    if (d3 > (d2 + s2)) {d2 = d3 + s2;}
    if (d2 > (d1 + s1)) {d1 = d2 + s1;}

    int n1 = static_cast<int>(floor(d1/s1));
    int n2 = static_cast<int>(floor(d2/s2));
    int n3 = static_cast<int>(floor(d3/s3));

    // make grid size odd
    if (n1 % 2 == 0) n1++;
    if (n2 % 2 == 0) n2++;
    if (n3 % 2 == 0) n3++;

    // add grid levels to fdm
    fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);
   } else if (molecule.get_grid_detail() == 3) {
    // very fine and large inner grids, ensures very high quality solvation and interaction energies
    // also in case of membrane proteins with far reaching interactions
    // due to low screening in the apolar membrane environment
    double s1 = 2.00;
    double s2 = 0.50;
    double s3 = 0.20;
    double s4 = 0.15;

    double d1 = gmaxdim + (mmaxdim * 0.5f) + 150;
    double d2 = mmaxdim + 100;
    double d3 = mmaxdim + 60;
    double d4 = mmaxdim + 30;

    // settings chosen to get excellent boundary conditions (effectively zero potentials) and
    // converged interaction energies on large distances (important for membrane proteins)
    if (d1 < 240) {d1 = 240;}
    if (d2 < 120) {d2 = 120;}
    if (d3 < 60)  {d3 = 60;}
    if (d4 < 40)  {d4 = 40;}

    // outer grids must be larger than inner grids
    if (d4 > (d3 + s3)) {d3 = d4 + s3;}
    if (d3 > (d2 + s2)) {d2 = d3 + s2;}
    if (d2 > (d1 + s1)) {d1 = d2 + s1;}

    int n1 = static_cast<int>(floor(d1/s1));
    int n2 = static_cast<int>(floor(d2/s2));
    int n3 = static_cast<int>(floor(d3/s3));
    int n4 = static_cast<int>(floor(d4/s4));

    // make grid size odd
    if (n1 % 2 == 0) n1++;
    if (n2 % 2 == 0) n2++;
    if (n3 % 2 == 0) n3++;
    if (n4 % 2 == 0) n4++;

    // add grid levels to fdm
    fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n4, s4, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n4, s4, ON_CENT_OF_INTR);
   } else if (molecule.get_grid_detail() == 4) {
    // ultra fine and large inner grids, should not give any significant difference to molecule.get_grid_detail() == 3
    double s1 = 1.00;
    double s2 = 0.25;
    double s3 = 0.15;
    double s4 = 0.05;

    double d1 = gmaxdim + (mmaxdim * 0.5f) + 150;
    double d2 = mmaxdim + 100;
    double d3 = mmaxdim + 60;
    double d4 = mmaxdim + 30;

    // settings chosen to get excellent boundary conditions (effectively zero potentials) and
    // converged interaction energies on large distances (important for membrane proteins)
    if (d1 < 240) {d1 = 240;}
    if (d2 < 120) {d2 = 120;}
    if (d3 < 60)  {d3 = 60;}
    if (d4 < 40)  {d4 = 40;}

    // outer grids must be larger than inner grids
    if (d4 > (d3 + s3)) {d3 = d4 + s3;}
    if (d3 > (d2 + s2)) {d2 = d3 + s2;}
    if (d2 > (d1 + s1)) {d1 = d2 + s1;}

    int n1 = static_cast<int>(floor(d1/s1));
    int n2 = static_cast<int>(floor(d2/s2));
    int n3 = static_cast<int>(floor(d3/s3));
    int n4 = static_cast<int>(floor(d4/s4));

    // make grid size odd
    if (n1 % 2 == 0) n1++;
    if (n2 % 2 == 0) n2++;
    if (n3 % 2 == 0) n3++;
    if (n4 % 2 == 0) n4++;

    // add grid levels to fdm
    fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
    fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);
    fdm_ogm.add_level(n4, s4, ON_CENT_OF_INTR);
    fdm_mgm.add_level(n4, s4, ON_CENT_OF_INTR);
   } else {
    ::error("GcemSolverInputget_fdm unknown value for molecule.get_grid_detail() : must be in {0, 1, 2, 3, 4} but is ");
    cerr << molecule.get_grid_detail() << "." << endl;
   }

  }

  // need to do this because iteration over levels is only allowed if resolved
  fdm_ogm.resolve(Coord(0,0,0),Coord(0,0,0));
  fdm_mgm.resolve(Coord(0,0,0),Coord(0,0,0));

  return pair<FinDiffMethod,FinDiffMethod>(fdm_ogm,fdm_mgm);
}

void GcemSolverInput::write_gcem_input(const string& dir, const string& fn, const GcemMolecule &molecule) const
{

  blab2 << "Entering GcemSolverInput::write_gcem_input" << endl;


  // molecule.get_ligands() takes care that protons and electrons precede any other ligands
  list<string> ligands = molecule.get_ligands();

  // FIXME? the whole grid data retrieval is quite cumbersome
  // due to all the private items in the FinDiffMethod

  pair<FinDiffMethod,FinDiffMethod> fdm_pair = get_fdm(molecule.get_name(),molecule);
  list<CenteringStyle> mcstylelist, ocstylelist;
  list<int>            mgrddimlist, ogrddimlist;
  list<double>         mgrdspclist, ogrdspclist;
  vector<Coord>        mgrdcenlist, ogrdcenlist;

  CubeLatSpec *clso = fdm_pair.first.get_coarsest();

  do {
   ocstylelist.push_back((*clso).get_centering_style());
   ogrddimlist.push_back((*clso).get_grid_dim());
   ogrdspclist.push_back((*clso).get_spacing());
   ogrdcenlist.push_back((*clso).get_center());
  } while ((clso = fdm_pair.first.get_finer()));


  CubeLatSpec *clsm = fdm_pair.second.get_coarsest();

  do {
   mcstylelist.push_back((*clsm).get_centering_style());
   mgrddimlist.push_back((*clsm).get_grid_dim());
   mgrdspclist.push_back((*clsm).get_spacing());
   mgrdcenlist.push_back((*clsm).get_center());
  } while ((clsm = fdm_pair.second.get_finer()));

  blab3 << "Writing the GCEM input script \"" << fn << "\" and structures to directory "
        << "\"" << dir << "\"." << endl;

  ////// check if the gcem_input directory exists, otherwise create //////

  if (!fs::is_directory(dir)){
   if (fs::exists(dir)) fs::remove_all(dir);
   fs::create_directory(dir);
  }
  // subdirectory for solver jobs and their input
  if (fs::exists(dir + "/gcem")) fs::remove_all(dir + "/gcem");
  if (!fs::is_directory(dir + "/gcem")){
   if (fs::exists(dir + "/gcem")) fs::remove_all(dir + "/gcem");
   fs::create_directory(dir + "/gcem");
  }
  // subdirectory for membrane potential
  if (molecule.get_has_memb() && !fs::is_directory(dir + "/membpot")){
   if (fs::exists(dir + "/membpot")) fs::remove_all(dir + "/membpot");
   fs::create_directory(dir + "/membpot");
  }

  ////// write the GCEM input script, the structure files and the job script(s) //////

  // prepare string representations of solver options
  stringstream ss09, ss10, ss11, ss12, ss13, ss14, ss15, ss16, ss17, ss18,
               ss19, ss20, ss21, ss22, ss23, ss24, ss25, ss26, ss27, ss28,
               ss29;
  string epsin1str, epsin2str, epsin3str, epscorestr, epsheadstr, epsextstr, epsvacstr,
         eps_ffstr, istrstr, istr1str, istr2str, zminhstr, zmaxhstr, zmincstr, zmaxcstr,
         membholerstr, membholexstr, membholeystr, Tstr, sternlstr, solradstr;
  ss09 << PhysCond::get_T();                   ss09 >> Tstr;
  ss10 << molecule.get_eps_ff();               ss10 >> eps_ffstr;
  ss11 << molecule.get_epsin1();               ss11 >> epsin1str;
  ss12 << molecule.get_epsin2();               ss12 >> epsin2str;
  ss13 << molecule.get_epsin3();               ss13 >> epsin3str;
  ss14 << molecule.get_epscore();              ss14 >> epscorestr;
  ss15 << molecule.get_epshead();              ss15 >> epsheadstr;
  ss16 << molecule.get_epsext();               ss16 >> epsextstr;
  ss17 << PhysCond::get_vacuum_permittivity(); ss17 >> epsvacstr;
  ss18 << molecule.get_ionicstr();             ss18 >> istrstr;
  ss19 << molecule.get_ionicstr1();            ss19 >> istr1str;
  ss20 << molecule.get_ionicstr2();            ss20 >> istr2str;
  ss21 << molecule.get_zminh();                ss21 >> zminhstr;
  ss22 << molecule.get_zmaxh();                ss22 >> zmaxhstr;
  ss23 << molecule.get_zminc();                ss23 >> zmincstr;
  ss24 << molecule.get_zmaxc();                ss24 >> zmaxcstr;
  ss25 << molecule.get_membholer();            ss25 >> membholerstr;
  ss26 << molecule.get_membholex();            ss26 >> membholexstr;
  ss27 << molecule.get_membholey();            ss27 >> membholeystr;
  ss28 << PhysCond::get_sterln();              ss28 >> sternlstr;
  ss29 << PhysCond::get_solrad();              ss29 >> solradstr;

  // check if we have any flexible site or site with dissociable
  // ligand atoms, if so we cannot use the same dielectric and
  // electrolyte regions for all but need separate ones for each
  // calculation/ the electrolyte region is less critical, could
  // even use one if only dissociable hydrogens are present
  bool has_flex = false;
  for (map<SiteID,int>::const_iterator ri=molecule.rotmap.begin(); ri!=molecule.rotmap.end();ri++){
   if (ri->second==1 || ri->second==2){
    has_flex = true;
    break;
   }
  }
  // the ligand atoms are collected and written to a .pqr file, the user has to decide
  // if and to which region they should belong (cav/molecule.get_elycavset())
  bool has_ligats = false;
  AtomSet ligats;
  for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); i++){
   int stype(molecule.get_rotmap_value(i->get_sid()));
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   if (stype == 1 || stype == 3){
    GcemSiteDef sdef(molecule.get_sitedef(i->get_name()));
    // this is a titratable and/or flexible site
    // collect all ligand atoms
    for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
     // check all other forms for absence of any atom of this form
     for (list<GcemFormDef>::iterator fitr = sdef.forms.begin(); fitr != sdef.forms.end(); fitr++){
      for (AtomSet::const_iterator itr=j->site_struct.begin();itr!=j->site_struct.end();++itr){
       // check if atom is in GcemFormDef (occupancy and/or charge varies among forms)
       if (fitr->contains_atom(CharmmID(itr->second))){
        if (fitr->get_atom(CharmmID(itr->second)).second == 0){
         // this atom belongs to a dissociable ligand
         has_ligats = true;
         if (!ligats.contains(itr->first)) ligats.insert(itr->second);
        }
       }
      }
     }
    }
   }
  }

  if(has_ligats) ligats.write_pqr(string(dir + "/ligats.pqr"));
  
  // it can happen that ligand atoms of another site end up in the
  // model compound of a site, e.g., if a titratable amino acid appears at the terminus
  // In this case, we need to include these ligand atoms also in the eps2set envelope
  // to ensure that the Born solvation energy is not artifically made more favorable.
  AtomSet ligats_in_model_of_other_site;
  for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); i++){
   if (!has_ligats) break;
   int stype(molecule.get_rotmap_value(i->get_sid()));
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   if (stype == 1 || stype == 3){
    // this is a titratable and/or flexible site
    // collect all ligand atoms of other sites appearing in the model_struct
    for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); j++){
     // if the model struct contains a ligand atom but the site atom does not,
     // we have found a ligand atom of another site in the model struct
     for (AtomSet::const_iterator k=ligats.begin(); k!=ligats.end(); k++){
      if (!j->site_struct.contains_any_inst(k->first) && j->model_struct.contains_any_inst(k->first)){
       if (!ligats_in_model_of_other_site.contains(k->first)){
        ligats_in_model_of_other_site.insert(k->second);
       }
      }
     }
    }
   }
  }

  molecule.background.write_pqr(string(dir + "/" + molecule.get_name() + ".background.pqr"));

  // create structure for IAV/SAV envelopes

  // all instances for IAV
  AtomSet tmpset(molecule.background);
  for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
   for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
    for (AtomSet::const_iterator itr=j->site_struct.begin();itr!=j->site_struct.end();++itr){
     if (!tmpset.contains(itr->first)) tmpset.insert(itr->second);
    }
   }
  }

  tmpset.write_pqr(dir + "/" + molecule.get_name() + ".all.pqr");

  // all instances in their appropriate eps region for SAVs
  AtomSet e1set;
  AtomSet e2set(molecule.background);
  if ((!has_flex) || molecule.get_has_memb()){
   for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
    bool is_qm = false;
    if (molecule.get_rotmap_value(i->get_sid()) == 4) is_qm = true;
    for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
     for (AtomSet::const_iterator itr=j->site_struct.begin(); itr!=j->site_struct.end(); ++itr){
      if (is_qm){
       if ((!ligats.contains(itr->first) || ligats_in_model_of_other_site.contains_any_inst(itr->first))
           && !e1set.contains(itr->first)){
        e1set.insert(itr->second);
       }
      }
      else {
       if ((!ligats.contains(itr->first) || ligats_in_model_of_other_site.contains_any_inst(itr->first))
           && !e2set.contains(itr->first)){
        e2set.insert(itr->second);
       }
      }
     }
    }
   }
   // if (e1set.size()!=0) e1set.write_pqr(string(dir + "/gcem/epsin1_region.pqr"));
   // if (e2set.size()!=0) e2set.write_pqr(string(dir + "/gcem/epsin2_region.pqr"));
   e1set.write_pqr(string(dir + "/gcem/epsin1_region.pqr"));
   e2set.write_pqr(string(dir + "/gcem/epsin2_region.pqr"));
  }

  //// write membrane potential job script ////
  if (molecule.get_has_memb()){
   // membpot job script
   string gcem_mjobscript_fn(dir + "/membpot/job.sh");
   ofstream mjf(gcem_mjobscript_fn.c_str());
   if (!mjf.good())
     ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
             gcem_mjobscript_fn.c_str());

   mjf << "#!/bin/tcsh" << endl << endl;
   mjf << "# precalculation of volumes speeds up the calculation," << endl
       << "# one can link those of ../gcem/here if already calculated" << endl;
   if (e1set.size()!=0) mjf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << " epsin1_region" << endl;
   if (e2set.size()!=0) mjf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << " epsin2_region" << endl;
   if (molecule.get_eps3set()!=""){
    if (molecule.get_has_memb() && has_ligats){
                        mjf << "## user must decide if ligand atoms should be added to molecule.get_eps3set(), molecule.get_elycavset(), molecule.get_channelset() "
                            << "(e.g., to avoid spurious occurrence of membrane dielectric within the molecule)" << endl;
                        mjf << "#rm " << molecule.get_eps3set() << ".pqr && cat ../../" << molecule.get_eps3set()
                            << ".pqr ../ligats.pqr | perl -p -e 's/(ATOM\\s+\\d+)\\s{1,4}[^\\s]{1,4}\\s{1,3}[^\\s]{1,4} /$1  DUM DUM /g' >! "
                            << molecule.get_eps3set() << ".pqr " << endl;
    }
                        mjf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << " " << molecule.get_eps3set()   << endl;
   }
   if (molecule.get_elycavset()!="")   mjf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_elycavset() << endl;
   if (molecule.get_channelset()!="")  mjf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_channelset() << endl;
                        mjf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_name() << ".all"
                            << " && mv " << molecule.get_name() << ".all.iav.dat " << molecule.get_name() << ".background.iav.dat" << endl;
   mjf << molecule.get_meadpath() + "/apps/my_membpot_solver/my_membpot_solver ";
   mjf << "-sterln " << sternlstr << " -solrad " << solradstr;
   if (e1set.size()!=0) mjf << " -eps1set epsin1_region -epsin1 " << epsin1str;
   if (e2set.size()!=0) mjf << " -eps2set epsin2_region -epsin2 " << epsin2str;
   if (molecule.get_eps3set()!="")
    mjf << " -eps3set " << molecule.get_eps3set() << " -epsin3 " << epsin3str;
   if (molecule.get_elycavset()!="")
    mjf << " -elycavset " << molecule.get_elycavset();
   if (molecule.get_channelset()!="")
    mjf << " -channelset " << molecule.get_channelset();
   mjf << " -epshead " << epsheadstr << " -epscore " << epscorestr << " -epsext " << epsextstr;
   mjf << " -membz " << zmincstr << " " << zmaxcstr << " " << zminhstr << " " << zmaxhstr;
   if (molecule.get_ionicstr1()!=0 || molecule.get_ionicstr2()!=0)
    mjf << " -ionicstr1 " << istr1str << " -ionicstr2 " << istr2str;
   else
    mjf << " -ionicstr " << istrstr;
   if (molecule.get_membholer()!=0)
    mjf << " -membhole " << membholerstr << " " << membholexstr << " " << membholeystr;
   if (molecule.get_inside()!=1)
    mjf << " -inside " << molecule.get_inside();
   if (molecule.get_do_cap())
    mjf << " -capacitance";
   mjf << " -fpt site -membpot 1 " << molecule.get_name() << ".background > membpot.out" << endl << endl;

   if (!mjf.good())
     ::error("GcemSolverInput::write_gcem_input failed to write file, ",
             gcem_mjobscript_fn.c_str());
   mjf.close();
  }
  //// end membrane potential job script ////

  // need also grid for membrane potential solver
  if (molecule.get_has_memb()){
     // write grids for molecule.sites
     list<CenteringStyle> socstylelist = ocstylelist;
     list<int>            sogrddimlist = ogrddimlist;
     list<double>         sogrdspclist = ogrdspclist;
     vector<Coord>        sogrdcenlist = ogrdcenlist;

     // the default of 0.2 A grid spacing is hardly achievable for large
     // bioenergetic complexes, membrane potential is also not very
     // sensitive to grid spacing ---> use coarser inner grid if the user
     // did not provide an own .ogm
     string ogm_name(molecule.get_name() + ".ogm");
     ifstream ogm_fs(ogm_name.c_str()); bool got_ogm = ogm_fs.good(); ogm_fs.close();
     if (!got_ogm){
      sogrdspclist.clear();
      sogrdspclist.push_back((double)2);
      if      (molecule.get_grid_detail() == 0) {
       sogrdspclist.push_back((double)1.50);
       sogrdspclist.push_back((double)1.00);
      }
      else if (molecule.get_grid_detail() == 1) {
       sogrdspclist.push_back((double)1.00);
       sogrdspclist.push_back((double)0.50);
      }
      else if (molecule.get_grid_detail() == 2) {
       sogrdspclist.push_back((double)0.75);
       sogrdspclist.push_back((double)0.30);
      }
      else if (molecule.get_grid_detail() == 3) {
       sogrdspclist.push_back((double)0.50);
       sogrdspclist.push_back((double)0.35);
       sogrdspclist.push_back((double)0.20);
      }
      else if (molecule.get_grid_detail() == 4) {
       sogrdspclist.push_back((double)0.40);
       sogrdspclist.push_back((double)0.25);
       sogrdspclist.push_back((double)0.15);
      }
      else {
       ::error("GcemSolverInput::write_gcem_input unknown value for molecule.get_grid_detail() : must be in {0, 1, 2, 3, 4} but is ");
       cerr << molecule.get_grid_detail() << endl;
      }
     }

     // modify center and dim for the innermost grid for a user defined grid
     // or center, spacing and dim for all grid levels of an autogenerated grid
     Coord thiscent, maxd, mind;
     thiscent.x = thiscent.y = thiscent.z = 0;
     maxd.x = maxd.y = maxd.z = -std::numeric_limits<float>::max();
     mind.x = mind.y = mind.z = +std::numeric_limits<float>::max();
     int nats = 0;
     for (AtomSet::const_iterator iitr=tmpset.begin();iitr!=tmpset.end();++iitr){
      nats++;
      thiscent += iitr->second.coord;
      if (iitr->second.coord.x > maxd.x) maxd.x = iitr->second.coord.x;
      if (iitr->second.coord.x < mind.x) mind.x = iitr->second.coord.x;
      if (iitr->second.coord.y > maxd.y) maxd.y = iitr->second.coord.y;
      if (iitr->second.coord.y < mind.y) mind.y = iitr->second.coord.y;
      if (iitr->second.coord.z > maxd.z) maxd.z = iitr->second.coord.z;
      if (iitr->second.coord.z < mind.z) mind.z = iitr->second.coord.z;
     }
     if (nats > 0) thiscent /= static_cast<double>(nats);
     blab3 << "resolved center of " << molecule.get_name() << " to " << thiscent << endl;
     for (int b=0; b<sogrdcenlist.size(); b++) sogrdcenlist[b] = thiscent;
     double md = maxd.x - mind.x;
     if ((maxd.y - mind.y) > md) md = maxd.y - mind.y;
     if ((maxd.z - mind.z) > md) md = maxd.z - mind.z;
     md += 20;
     
     // adjust grid dimension to molecule size
     if (got_ogm){
      // adjust only inner grid
      list<double>::iterator tsi = sogrdspclist.end(); tsi--;
      tsi = sogrdspclist.end(); tsi--;
      double thismd = 1;
      if      (molecule.get_grid_detail() == 0) {
       thismd = md + 10 + (2 * (*tsi));
      }
      else if (molecule.get_grid_detail() == 1) {
       thismd = md + 20 + (2 * (*tsi));
      }
      else if (molecule.get_grid_detail() == 2) {
       thismd = md + 30 + (2 * (*tsi));
      }
      else if (molecule.get_grid_detail() == 3) {
       thismd = md + 50 + (2 * (*tsi));
      }
      else if (molecule.get_grid_detail() == 4) {
       thismd = md + 50 + (2 * (*tsi));
      }
      else {
       ::error("GcemSolverInput::write_gcem_input unknown value for molecule.get_grid_detail() : must be in {0, 1, 2, 3, 4} but is ");
       cerr << molecule.get_grid_detail() << endl;
      }
      list<int>::iterator tdi = sogrddimlist.end(); tdi--;
      *tdi = static_cast<int>(ceil(thismd / (*tsi)));
      if ((*tdi) % 2 == 0) (*tdi)++;
     } else {
      list<int>::iterator tdi = sogrddimlist.begin();
      int thisi = sogrddimlist.size() - 1;
      for (list<double>::const_iterator tsi = sogrdspclist.begin(); tsi != sogrdspclist.end(); ++tsi){
       double thismd = 0;
       if      (molecule.get_grid_detail() == 0) {
        thismd = ceil(md + (thisi * 10) + (2 * (*tsi)));
       }
       else if (molecule.get_grid_detail() == 1) {
        thismd = ceil(md + (thisi * 20) + (2 * (*tsi)));
       }
       else if (molecule.get_grid_detail() == 2) {
        thismd = ceil(md + (thisi * 30) + (2 * (*tsi)));
       }
       else if (molecule.get_grid_detail() == 3) {
        thismd = ceil(md + (thisi * 50) + (2 * (*tsi)));
       }
       else if (molecule.get_grid_detail() == 4) {
        thismd = ceil(md + (thisi * 50) + (2 * (*tsi)));
       }
       else {
        ::error("GcemSolverInput::write_gcem_input unknown value for molecule.get_grid_detail() : must be in {0, 1, 2, 3, 4} but is ");
        cerr << molecule.get_grid_detail() << endl;
       }
       (*tdi) = static_cast<int>(ceil(thismd / (*tsi)));
       if ((*tdi) % 2 == 0) (*tdi)++;
       tdi++;
       thisi--;
      }      
     }

     // molecule.background ogm file
     string sogm_fn(dir + "/membpot/" + molecule.get_name() + ".background.ogm");
     ofstream sogm(sogm_fn.c_str());
     if (!sogm.good())
      ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ", sogm_fn.c_str());

     list<int>::const_iterator     di = sogrddimlist.begin();
     list<double>::const_iterator  si = sogrdspclist.begin();
     vector<Coord>::const_iterator bi = sogrdcenlist.begin();

     for (list<CenteringStyle>::const_iterator ci=socstylelist.begin(); ci!=socstylelist.end(); ci++){
      switch (*ci) {
       case ON_ORIGIN:
         sogm << " ON_ORIGIN";
         break;
       case ON_CENT_OF_INTR:
         sogm << " " << *bi;
         break;
       case ON_GEOM_CENT:
         sogm << " " << *bi;
         break;
       case SPECIFIC:
         sogm << " " << *bi;
       break;
      }
      sogm << " " << *di << " " << setprecision(3) << fixed << *si << endl;
      di++; si++; bi++;
     }
     if (!sogm.good())
      ::error("GcemSolverInput::write_gcem_input failed to write file, ", sogm_fn.c_str());
  }

  // QMPB input script
  string gcem_input_fn(dir + "/" + fn);
  ofstream of(gcem_input_fn.c_str());
  if (!of.good())
    ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
            gcem_input_fn.c_str());

  // job script
  string gcem_jobscript_fn(dir + "/gcem/job.sh");
  ofstream jf(gcem_jobscript_fn.c_str());
  if (!jf.good())
    ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
            gcem_jobscript_fn.c_str());

  // site fpt file
  string gcem_fpt_fn(dir + "/gcem/site.fpt");
  ofstream fpt(gcem_fpt_fn.c_str());
  if (!fpt.good())
    ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
            gcem_fpt_fn.c_str());

   // write header of job script

   jf << "#!/bin/tcsh" << endl << endl;

   // add jobs for precalculation of analytical volumes
   if (!has_flex || molecule.get_has_memb()){
    if (e1set.size()!=0) jf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << " epsin1_region" << endl;
    if (e2set.size()!=0) jf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << "  epsin2_region" << endl;
    if (molecule.get_eps3set()!=""){
     if (molecule.get_has_memb() && has_ligats){
                         jf << "## user must decide if ligand atoms should be added to molecule.get_eps3set(), molecule.get_elycavset(), molecule.get_channelset() "
                            << "(e.g., to avoid spurious occurrence of membrane dielectric within the molecule)" << endl;
                         jf << "#rm " << molecule.get_eps3set() << ".pqr && cat ../../" << molecule.get_eps3set()
                            << ".pqr ../ligats.pqr | perl -p -e 's/(ATOM\\s+\\d+)\\s{1,4}[^\\s]{1,4}\\s{1,3}[^\\s]{1,4} /$1  DUM DUM /g' >! "
                            << molecule.get_eps3set() << ".pqr " << endl;
     }
                         jf << "#" << molecule.get_meadpath() << "/apps/pqr2SolvAccVol/pqr2SolvAccVol -solrad " << solradstr << " " << molecule.get_eps3set() << endl;
    }
    if (molecule.get_elycavset()!="")   jf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_elycavset() << endl;
    if (molecule.get_channelset()!="")  jf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_channelset() << endl;
                         jf << "#" << molecule.get_meadpath() << "/apps/pqr2IonAccVol/pqr2IonAccVol -sterln " << sternlstr << " " << molecule.get_name() << ".all" << endl;
   }

   // write header of QMPB input script
   // the QMPB script is anyway not using all features (correctly), remove it ?

   of << "# QMPB run with multiple instances  "  << endl;
   of << "# Input generated by MEAD GCEM V 0.6"  << endl;

   of << endl << endl;

   of << "molecule.get_meadpath() = " << molecule.get_meadpath() << endl;
   of << "I = " << PhysCond::get_ionicstr() << endl;
   of << "T = " << PhysCond::get_T() << endl;
   of << "backfile = " << molecule.get_name() << ".background.pqr" << endl;
   of << "epsin1 = " << molecule.get_epsin1() << endl;
   of << "epsin2 = " << molecule.get_epsin2() << endl;


   of << "OGMcenter =";
   short ix = 0;
   for (list<CenteringStyle>::const_iterator i=ocstylelist.begin(); i!=ocstylelist.end(); ++i){
    switch (*i) {
     case ON_ORIGIN:
       of << " ON_ORIGIN";
       break;
     case ON_CENT_OF_INTR:
       of << " ON_CENT_OF_INTR";
       break;
     case ON_GEOM_CENT:
       of << " ON_GEOM_CENT";
       break;
     case SPECIFIC:
       of << " " << ogrdcenlist[ix];
     break;
    }
   } of << endl;


   of << "OGMpoints =";
   for (list<int>::const_iterator i=ogrddimlist.begin(); i!=ogrddimlist.end(); ++i){
     of << " " << *i;
   } of << endl;


   of << "OGMspace =";
   for (list<double>::const_iterator i=ogrdspclist.begin(); i!=ogrdspclist.end(); ++i){
     of << " " << setprecision(2) << fixed << *i;
   } of << endl;


   of << "MGMcenter =";
   ix = 0;
   for (list<CenteringStyle>::const_iterator i=mcstylelist.begin(); i!=mcstylelist.end(); ++i){
    switch (*i) {
     case ON_ORIGIN:
       of << " ON_ORIGIN";
       break;
     case ON_CENT_OF_INTR:
       of << " ON_CENT_OF_INTR";
       break;
     case ON_GEOM_CENT:
       of << " ON_GEOM_CENT";
       break;
     case SPECIFIC:
       of << " " << mgrdcenlist[ix];
     break;
    }
   } of << endl;


   of << "MGMpoints =";
   for (list<int>::const_iterator i=mgrddimlist.begin(); i!=mgrddimlist.end(); ++i){
     of << " " << *i;
   } of << endl;


   of << "MGMspace =";
   for (list<double>::const_iterator i=mgrdspclist.begin(); i!=mgrdspclist.end(); ++i){
     of << " " << setprecision(2) << fixed << *i;
   } of << endl;


   of << "Ligand_Labels =";
   for (list<string>::const_iterator i=ligands.begin(); i!=ligands.end(); ++i){
    of << " " << *i;
   } of << endl;

   of << endl << endl << "#####################" << endl << endl;



   // now write the sites to the QMPB input script, their structure files and their GCEM jobs to the job script

   int sid = -1;

   for (list<GcemSite>::const_iterator i=molecule.sites.begin(); i!=molecule.sites.end(); ++i){
    sid++;
    int iid = -1;
    int c((i->get_sid()).get_chain());
    int s((i->get_sid()).get_site());
    string str_c, str_s, epsrefstr, istrrefstr;
    stringstream ss1, ss2;
    ss1 << c; ss1 >> str_c;
    ss2 << s; ss2 >> str_s;
    string slbl = (i->get_sid()).get_sitename() + "-" + str_c + "-" + str_s;

    of << endl << "#####################" << endl << endl;

    for (list<GcemInstance>::const_iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
     iid++;
     // write site atom coordinates and charges to site.fpt
     for (AtomSet::const_iterator ai=j->site_struct.begin(); ai!=j->site_struct.end(); ai++){
      fpt << sid << " " << iid  << " "
          << setprecision(4) << fixed << ai->second.coord.x << " "
          << setprecision(4) << fixed << ai->second.coord.y << " "
          << setprecision(4) << fixed << ai->second.coord.z << " "
          << setprecision(6) << fixed << ai->second.charge  << endl;
     }

     string str_in, ilbl, milbl;
     double gmmint(j->get_gmmint());
     stringstream ss3, ss4, ss5;
     //int in((j->get_instid()).get_instance());
     //ss3 << in; ss3 >> str_in;
     ss3 << iid; ss3 >> str_in;
     ilbl = (j->get_instid()).get_instname() + "-" + str_in;
     milbl = (j->get_instid()).get_instname() + "-" + str_in + "_model";

     // dielectric constant of reference environment
     double tmpepsref = j->get_epsref();
     // default fallback epsref = epsext
     if (tmpepsref < 1.0f){tmpepsref = molecule.get_epsext();}
     ss4 << tmpepsref; ss4 >> epsrefstr;     

     // ionic strength of reference environment
     double tmpistrref = j->get_istrref();
     // default fallback istrref = epsext
     if (tmpistrref < 0.0f){tmpistrref = molecule.get_ionicstr();}
     ss5 << tmpistrref; ss5 >> istrrefstr;     

     // for now let the MM energies aside, we anyway need to merge the interaction energy matrices

     of << endl;

     of << "MMsite " << slbl << " " << ilbl << endl;
     of << "sid=" << sid << endl;
     of << "iid=" << iid << endl;
     of << "  file=" << slbl << "_" << ilbl << ".pqr" << endl;
     of << "  ref=" << milbl << endl;
     of << "  Gmm=0" << endl;

     of << endl;

     of << "Modelsite " << slbl << " " << milbl << endl;
     of << "sid=" << sid << endl;
     of << "iid=" << iid << endl;
     of << "  file=" << slbl << "_" << milbl << ".pqr" << endl;
     of << "  ref=" << ilbl << endl;
     of << "  Gmodel=" << j->get_gmodel() << endl;
     of << "  eps=" << "4" << endl;
     of << "  N=";
     for (list<string>::const_iterator k=ligands.begin(); k!=ligands.end(); ++k){
       of << " " << j->ligands_bound(*k);
     } of << endl;

     // write structure files ...
     // ... and entries in job script
     string fsitestr = dir + "/" + slbl + "_" + ilbl + ".pqr";
     string fsitestr2 = "../" + slbl + "_" + ilbl + ".pqr";
     j->site_struct.write_pqr(fsitestr);
     string tsitestr = dir + "/gcem/" + slbl + "_" + ilbl + ".pqr";
     if (!fs::exists(fsitestr)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fsitestr.c_str());
     if (fs::exists(tsitestr)) fs::remove_all(tsitestr);
     fs::create_symlink(fsitestr2, tsitestr);
     j->model_struct.write_pqr(string(dir + "/" + slbl + "_" + milbl + ".pqr"));
     // need model_back2d structure file with all site atoms set to zero charge
     AtomSet b2d;
     for (AtomSet::const_iterator ai=j->model_struct.begin(); ai!=j->model_struct.end(); ai++){
      AtomID k = ai->first;
      if (!j->site_struct.contains(k)){
       const Atom& a = ai->second;
       b2d.insert(a);
      } else {
       // set charge of site atoms zero
       Atom a(ai->second);
       a.charge = 0;
       // set  instance to zero, not necessary but avoids croaking of solvers about atoms occuring in protein and solute
       a.instance = 0;
       b2d.insert(a);
      }
     }
     b2d.write_pqr(string(dir + "/gcem/" + slbl + "_" + ilbl + "_model_back2d.pqr"));
     bool is_qm = false;
     if (molecule.get_rotmap_value(i->get_sid()) == 4) is_qm = true;
     // need dielectric/electrolyte envelope plus molecule.background with just one rotamer per flexible site
     // and correct instance structure for this site
     AtomSet ei1;
     AtomSet ei2(molecule.background);
     if (has_flex || has_ligats){
      // insert this instance of this site into envelope
      for (AtomSet::const_iterator ai=j->site_struct.begin(); ai!=j->site_struct.end(); ai++){
       AtomID k = ai->first;
       const Atom& a = ai->second;
       if (is_qm)
        ei1.insert(a);
       else
        ei2.insert(a);
      }
      // insert all instances of other sites into envelope except dissociable ligand atoms
      // FIXME ? this becomes a problem if the model compound contains ligand atoms of another site,
      // eg if a terminus and the terminal residue are titratable
      for (list<GcemSite>::const_iterator ii=molecule.sites.begin(); ii!=molecule.sites.end(); ++ii){
       if (ii==i) continue;
       bool is_qm2 = false;
       if (molecule.get_rotmap_value(ii->get_sid()) == 4) is_qm2 = true;
       for (list<GcemInstance>::const_iterator jj=ii->instances.begin(); jj!=ii->instances.end(); ++jj){
        for (AtomSet::const_iterator iitr=jj->site_struct.begin();iitr!=jj->site_struct.end();++iitr){
         if (is_qm2){
          if ((!ligats.contains(iitr->first) || ligats_in_model_of_other_site.contains_any_inst(iitr->first))
              && !ei1.contains(iitr->first)){
           ei1.insert(iitr->second);
          }
         }
         else {
          if ((!ligats.contains(iitr->first) || ligats_in_model_of_other_site.contains_any_inst(iitr->first))
              && !ei2.contains(iitr->first)){
           ei2.insert(iitr->second);
          }
         }
        }
       }
      }
     }
     if (ei1.size()!=0 && (has_flex || has_ligats)){
      string ei1str = dir + "/gcem/epsin1_region_" + slbl + "_" + ilbl + ".pqr";
      ei1.write_pqr(ei1str);
     }
     if (ei2.size()!=0 && (has_flex || has_ligats)){
      string ei2str = dir + "/gcem/epsin2_region_" + slbl + "_" + ilbl + ".pqr";
      ei2.write_pqr(ei2str);
     }

     // calls to continuum electrostatics solvers

     // reference environment of a titratable or QM site
     string jobline_2d_re(molecule.get_meadpath() + "/apps/my_2diel_solver/my_2diel_solver");
     jobline_2d_re += " -T " + Tstr + " -ionicstr " + istrrefstr + " -epsext " + epsrefstr;
     jobline_2d_re += " -sterln " + sternlstr + " -solrad " + solradstr;
     if (is_qm) jobline_2d_re += " -epsin " + epsin1str;
     else       jobline_2d_re += " -epsin " + epsin2str;
     jobline_2d_re += " " + slbl + "_" + ilbl + " " + slbl + "_" + milbl + "_back2d";
     jobline_2d_re += " > " + slbl + "_" + ilbl + "_2d.out";

     // force field environment of a flexible MM site
     string jobline_2d_ff(molecule.get_meadpath() + "/apps/my_2diel_solver/my_2diel_solver");
     jobline_2d_ff += " -T " + Tstr + " -epsext " + eps_ffstr;
     jobline_2d_ff += " -sterln " + sternlstr + " -solrad " + solradstr;
     jobline_2d_ff += " -epsin " + epsin2str;
     jobline_2d_ff += " " + slbl + "_" + ilbl + " " + slbl + "_" + milbl + "_back2d";
     jobline_2d_ff += " > " + slbl + "_" + ilbl + "_2d_ff.out";

     // receptor environment
     string jobline_nd;
     if (molecule.get_has_memb()){
      jobline_nd = molecule.get_meadpath() + "/apps/my_memb_solver/my_memb_solver";
      jobline_nd += " -T " + Tstr;
      jobline_nd += " -sterln " + sternlstr + " -solrad " + solradstr;
      if (ei1.size()!=0 || is_qm){
       if (has_flex || has_ligats)
        jobline_nd += " -eps1set epsin1_region_" + slbl + "_" + ilbl + " -epsin1 " + epsin1str;
       else
        jobline_nd += " -eps1set epsin1_region -epsin1 " + epsin1str;
      }
      if (ei2.size()!=0){
       if (has_flex || has_ligats)
        jobline_nd += " -eps2set epsin2_region_" + slbl + "_" + ilbl + " -epsin2 " + epsin2str;
       else
        jobline_nd += " -eps2set epsin2_region -epsin2 " + epsin2str;
      }
      if (molecule.get_eps3set()!="")
       jobline_nd += " -eps3set " + molecule.get_eps3set() + " -epsin3 " + epsin3str;
      if (molecule.get_elycavset()!="")
       jobline_nd += " -elycavset " + molecule.get_elycavset();
      if (molecule.get_channelset()!="")
       jobline_nd += " -channelset " + molecule.get_channelset();
      jobline_nd += " -epshead " + epsheadstr + " -epscore " + epscorestr + " -epsext " + epsextstr;
      jobline_nd += " -membz " + zmincstr + " " + zmaxcstr + " " + zminhstr + " " + zmaxhstr;
      if (molecule.get_membholer()!=0)
       jobline_nd += " -membholer " + membholerstr + " -membholex " + membholexstr + " -membholey " + membholeystr;
      if (molecule.get_ionicstr1()!=0 || molecule.get_ionicstr2()!=0)
       jobline_nd += " -ionicstr1 " + istr1str + " -ionicstr2 " + istr2str;
      else
       jobline_nd += " -ionicstr " + istrstr;
     }
     else {
      jobline_nd = molecule.get_meadpath() + "/apps/my_3diel_solver/my_3diel_solver";
      jobline_nd += " -T " + Tstr;
      jobline_nd += " -sterln " + sternlstr + " -solrad " + solradstr;
      if (ei1.size()!=0 || is_qm){
       if (has_flex || has_ligats)
        jobline_nd += " -eps1set epsin1_region_" + slbl + "_" + ilbl + " -epsin1 " + epsin1str;
       else
        jobline_nd += " -eps1set epsin1_region -epsin1 " + epsin1str;
      }
      if (ei2.size()!=0){
       if (has_flex || has_ligats)
        jobline_nd += " -eps2set epsin2_region_" + slbl + "_" + ilbl + " -epsin2 " + epsin2str;
       else
        jobline_nd += " -eps2set epsin2_region -epsin2 " + epsin2str;
      }
      jobline_nd += " -epsext " + epsextstr;
      jobline_nd += " -ionicstr " + istrstr;
     }
     jobline_nd += " -fpt site " + slbl + "_" + ilbl + " " + molecule.get_name() + ".background > " + slbl + "_" + ilbl + ".out";

     // for flexible, non-titratable sites we don't need the first thermodynamic cycle
     // (intrinsic energy shift relative to known gmodel of binding form in reference environment)
     if (molecule.get_rotmap_value(i->get_sid()) != 2){
      jf << jobline_2d_re << endl;
     }
     // second thermodynamic cycle only needed for flexible MM sites
     // (MM energy shift relative to reference environment used in force field parametrization)
     if (molecule.get_rotmap_value(i->get_sid()) == 1 || molecule.get_rotmap_value(i->get_sid()) == 2){
      jf << jobline_2d_ff << endl;
     }
     // site in the receptor environment for both thermodynamic cycles
     jf << jobline_nd << endl;

     // write grids for molecule.sites

     list<CenteringStyle> smcstylelist = mcstylelist;
     list<CenteringStyle> socstylelist = ocstylelist;
     list<int>            smgrddimlist = mgrddimlist;
     list<int>            sogrddimlist = ogrddimlist;
     list<double>         smgrdspclist = mgrdspclist;
     list<double>         sogrdspclist = ogrdspclist;
     vector<Coord>        smgrdcenlist = mgrdcenlist;
     vector<Coord>        sogrdcenlist = ogrdcenlist;

     // modify center and dim for the innermost grid
     Coord thiscent, maxd, mind;
     thiscent.x = thiscent.y = thiscent.z = 0;
     maxd.x = maxd.y = maxd.z = -std::numeric_limits<float>::max();
     mind.x = mind.y = mind.z = +std::numeric_limits<float>::max();
     int nats = 0;
     for (AtomSet::const_iterator iitr=j->site_struct.begin();iitr!=j->site_struct.end();++iitr){
      nats++;
      thiscent += iitr->second.coord;
      if (iitr->second.coord.x > maxd.x) maxd.x = iitr->second.coord.x;
      if (iitr->second.coord.x < mind.x) mind.x = iitr->second.coord.x;
      if (iitr->second.coord.y > maxd.y) maxd.y = iitr->second.coord.y;
      if (iitr->second.coord.y < mind.y) mind.y = iitr->second.coord.y;
      if (iitr->second.coord.z > maxd.z) maxd.z = iitr->second.coord.z;
      if (iitr->second.coord.z < mind.z) mind.z = iitr->second.coord.z;
     }
     if (nats > 0) thiscent /= nats;
     blab3 << "resolved center of " << slbl << "_" << milbl << " to " << thiscent << endl;
     for (int b=1; b<smgrdcenlist.size(); b++) smgrdcenlist[b] = thiscent;
     for (int b=1; b<sogrdcenlist.size(); b++) sogrdcenlist[b] = thiscent;
     double md = maxd.x - mind.x;
     if ((maxd.y - mind.y) > md) md = maxd.y - mind.y;
     if ((maxd.z - mind.z) > md) md = maxd.z - mind.z;
     if        (molecule.get_grid_detail() == 0) {
      // coarse, solvation and interaction energies will be afflicted with grid discretization error
      md += 10;
     } else if (molecule.get_grid_detail() == 1) {
      // settings similar to old multiflex examples, ensures fairly converged solvation and interaction energies
      md += 20;
     } else if (molecule.get_grid_detail() == 2) {
      // somewhat finer than before, ensures high quality solvation and interaction energies 
      md += 30;
     } else if (molecule.get_grid_detail() == 3) {
      // very fine and large inner grids, ensures very high quality solvation and interaction energies
      md += 50;
     } else if (molecule.get_grid_detail() == 4) {
      // ultra fine and large inner grids, should not give any significant difference to molecule.get_grid_detail() == 3
      md += 50;
     } else {
      ::error("GcemSolverInput::write_gcem_input unknown value for molecule.get_grid_detail() : must be in {0, 1, 2, 3, 4} but is ");
      cerr << molecule.get_grid_detail() << "." << endl;
     }
     list<double>::iterator si = smgrdspclist.end(); si--;
     int imd = static_cast<int>(floor(md/(*si)));
     if (imd % 2 == 0) imd++;
     list<int>::iterator di = smgrddimlist.end(); di--;
     *di = imd;
     si = sogrdspclist.end(); si--;
     imd = static_cast<int>(floor(md/(*si)));
     if (imd % 2 == 0) imd++;
     di = sogrddimlist.end(); di--;
     *di = imd;

     // site ogm/mgm file
     string sogm_fn(dir + "/gcem/" + slbl + "_" + ilbl + ".ogm");
     string smgm_fn(dir + "/gcem/" + slbl + "_" + milbl + ".mgm");
     ofstream sogm(sogm_fn.c_str());
     if (!sogm.good())
      ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
            sogm_fn.c_str());
     ofstream smgm(smgm_fn.c_str());
     if (!smgm.good())
      ::error("GcemSolverInput::write_gcem_input failed to open file for writing, ",
            smgm_fn.c_str());

                                               di = sogrddimlist.begin();
                                               si = sogrdspclist.begin();
     vector<Coord>::const_iterator        bi = sogrdcenlist.begin();

     for (list<CenteringStyle>::const_iterator ci=socstylelist.begin(); ci!=socstylelist.end(); ci++){
      switch (*ci) {
       case ON_ORIGIN:
         sogm << " ON_ORIGIN";
         break;
       case ON_CENT_OF_INTR:
         sogm << " " << *bi;
         break;
       case ON_GEOM_CENT:
         sogm << " ON_GEOM_CENT";
         break;
       case SPECIFIC:
         sogm << " " << *bi;
       break;
      }
      sogm << " " << *di << " " << setprecision(3) << fixed << *si << endl;
      di++; si++; bi++;
     }

     di = smgrddimlist.begin();
     si = smgrdspclist.begin();
     bi = smgrdcenlist.begin();

     for (list<CenteringStyle>::const_iterator ci=smcstylelist.begin(); ci!=smcstylelist.end(); ci++){
      switch (*ci) {
       case ON_ORIGIN:
         smgm << " ON_ORIGIN";
         break;
       case ON_CENT_OF_INTR:
         smgm << " " << *bi;
         break;
       case ON_GEOM_CENT:
         smgm << " ON_GEOM_CENT";
         break;
       case SPECIFIC:
         smgm << " " << *bi;
       break;
      }
      smgm << " " << *di << " " << setprecision(3) << fixed << *si << endl;
      di++; si++; bi++;
     }

     if (!sogm.good())
      ::error("GcemSolverInput::write_gcem_input failed to write file, ",
            sogm_fn.c_str());
     if (!smgm.good())
      ::error("GcemSolverInput::write_gcem_input failed to write file, ",
            smgm_fn.c_str());
    }
   }

  // need a newline for GCEM parser at file end
  of << endl;

  if (!of.good())
    ::error("GcemSolverInput::write_gcem_input failed to write file, ",
            gcem_input_fn.c_str());
  of.close();
  if (!jf.good())
    ::error("GcemSolverInput::write_gcem_input failed to write file, ",
            gcem_jobscript_fn.c_str());
  jf.close();
  if (!fpt.good())
    ::error("GcemSolverInput::write_gcem_input failed to write file, ",
            gcem_fpt_fn.c_str());
  fpt.close();

  ///// create symlinks for files needed in multiple directories /////
  string fromname, fromname2, linkname;

  if ((!has_flex  && !has_ligats) || molecule.get_has_memb()){
   fromname = dir + "/" + molecule.get_name() + ".all.pqr";
   fromname2 = "../" + molecule.get_name() + ".all.pqr";
   linkname = dir + "/gcem/" + molecule.get_name() + ".all.pqr";
   if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   if (molecule.get_has_memb()){
    linkname = dir + "/membpot/" + molecule.get_name() + ".all.pqr";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
   // this is an IonAccVol for everything in the molecule, including the eps(1|2|3)set and all instances of all molecule.sites
   fromname2 = "../gcem/" + molecule.get_name() + ".all.iav.dat";
   linkname = dir + "/gcem/" + molecule.get_name() + ".background.iav.dat";
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
  }

  if (molecule.get_has_memb()){
   fromname = dir + "/gcem/site.fpt";
   fromname2 = "../gcem/site.fpt";
   linkname = dir + "/membpot/site.fpt";
   if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   //
   fromname2 = "../gcem/" + molecule.get_name() + ".all.iav.dat";
   linkname = dir + "/gcem/" + molecule.get_name() + ".background.iav.dat";
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   fromname2 = "../gcem/" + molecule.get_name() + ".all.iav.dat";
   linkname = dir + "/membpot/" + molecule.get_name() + ".background.iav.dat";
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   //
   if (molecule.get_elycavset() != "") {
    fromname2 = "../gcem/" + molecule.get_elycavset() + ".iav.dat";
    linkname = dir + "/membpot/" + molecule.get_elycavset() + ".iav.dat";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
   //
   if (molecule.get_channelset() != "") {
    fromname2 = "../gcem/" + molecule.get_channelset() + ".dat";
    linkname = dir + "/membpot/" + molecule.get_channelset() + ".iav.dat";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
   //
   if (molecule.get_eps3set() != "") {
    fromname2 = "../gcem/" + molecule.get_eps3set() + ".dat";
    linkname = dir + "/membpot/" + molecule.get_eps3set() + ".dat";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
  }

  fromname = dir + "/" + molecule.get_name() + ".background.pqr";
  fromname2 = "../" + molecule.get_name() + ".background.pqr";
  linkname = dir + "/gcem/" + molecule.get_name() + ".background.pqr";
  if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
  if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
  fs::create_symlink(fromname2, linkname);
  if (molecule.get_has_memb()){
   linkname = dir + "/membpot/" + molecule.get_name() + ".background.pqr";
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
  }
  fromname = dir + "/" + molecule.get_name() + ".all.pqr";
  fromname2 = "../" + molecule.get_name() + ".all.pqr";
  linkname = dir + "/gcem/" + molecule.get_name() + ".all.pqr";
  if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
  if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
  fs::create_symlink(fromname2, linkname);

  if (molecule.get_has_memb()){
   fromname = dir + "/gcem/epsin1_region.pqr";
   fromname2 = "../gcem/epsin1_region.pqr";
   linkname = dir + "/membpot/epsin1_region.pqr";
   if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   fromname = dir + "/gcem/epsin2_region.pqr";
   fromname2 = "../gcem/epsin2_region.pqr";
   linkname = dir + "/membpot/epsin2_region.pqr";
   if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
   if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
   fs::create_symlink(fromname2, linkname);
   if (molecule.get_eps3set()!=""){
    fromname = molecule.get_eps3set() + ".pqr";
    fromname2 = "../../" + molecule.get_eps3set() + ".pqr";
    linkname = dir + "/gcem/" + molecule.get_eps3set() + ".pqr";
    if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
    linkname = dir + "/membpot/" + molecule.get_eps3set() + ".pqr";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
   if (molecule.get_elycavset()!=""){
    fromname = molecule.get_elycavset() + ".pqr";
    fromname2 = "../../" + molecule.get_elycavset() + ".pqr";
    linkname = dir + "/gcem/" + molecule.get_elycavset() + ".pqr";
    if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
    linkname = dir + "/membpot/" + molecule.get_elycavset() + ".pqr";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
   else if (molecule.get_channelset()!=""){
    fromname = molecule.get_channelset() + ".pqr";
    fromname2 = "../../" + molecule.get_channelset() + ".pqr";
    linkname = dir + "/gcem/" + molecule.get_channelset();
    if (!fs::exists(fromname)) ::error("GcemSolverInput::write_gcem_input failed to find file for linking, ", fromname.c_str());
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
    linkname = dir + "/membpot/" + molecule.get_channelset() + ".pqr";
    if (fs::exists(linkname) || fs::is_symlink(linkname)) fs::remove_all(linkname);
    fs::create_symlink(fromname2, linkname);
   }
  }
  /////                    end create symlinks                   /////

  blab1 << "Exiting GcemSolverInput::write_gcem_input" << endl;

}


// GcemSolverInput.cc ends here
