/*  
    helper object for GcemMolecule
    handles the MEAD solver output containing
    the electrostatic energies

    parsing functions for output files of
    my_2diel_solver
    my_Ndiel_solver, my_3diel_solver, my_memb_solver
    my_membpot_solver

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/GcemSolverOutput.h"
#include "MEAD_e/globals.h"
#include <stdio.h>
#include <fstream>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <sstream>
#include <iostream>
#include <string>

using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::stringstream;
using std::string;

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

GcemSolverOutput::GcemSolverOutput(const GcemSolverOutput& a)
: name(a.name), inter(a.inter), nsites(a.nsites), nin(a.nin)
{
}

GcemSolverOutput::GcemSolverOutput()
: name("")
{
}

GcemSolverOutput::GcemSolverOutput (const string& nm)
: name(nm)
{
}

bool GcemSolverOutput::read_mead_out_vmemb(const string& gcemdir, const string& name, std::list<GcemSite> &sites, const bool do_cap, double *capacitance, double *vmemb_inter, const int nsites, vec<int> &nin)
{
  string fn_string = gcemdir + "/membpot/membpot.out";
  const char *fn = fn_string.c_str();
  string line;
  double gborn = 0, gback = 0, tmpd = 0;
  int sid2 = 0, iid2 = 0;

  matrix<double> site_vmemb_inter((int)0,nsites-1,nin,(double)0);

  boost::regex re_vmemb_inter("^Total energy of " + name + ".background in \\w+\\s*[dD]ielectric environment with membrane potential [eE.+-\\d]+ V = ([eE.+-\\d]+)\\s*");
  boost::regex re_capacitance("^The corresponding capacitance is = ([eE.+-\\d]+)\\s+\\(in Farad = As/V\\)");
  boost::regex re_ibgn("^Interaction energies of membrane potential with charges of given instances of sites");
  boost::regex re_inte("^\\s*(\\d+)\\s+(\\d+)\\s+:\\s+([eE.+-\\d]+)");
  boost::regex re_time("^Total runtime");

  boost::match_flag_type flags = boost::match_default;
  boost::match_results<std::string::const_iterator> match;

  bool got_capacitance = false, got_vmemb_inter = false, got_time = false, got_ibgn = false;
  std::string::const_iterator start, end;

  ifstream infile(fn);

  if (!infile)
    ::error ("GcemSolverOutput::read_mead_out_vmemb: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (!infile.eof() ){
    getline (infile, line);
    start = line.begin();
    end   = line.end();
    if        (boost::regex_search(start, end, match, re_time, flags)){
     got_time = true;
     break;
    } else if (boost::regex_search(start, end, match, re_vmemb_inter, flags)){
     got_vmemb_inter = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> *vmemb_inter;
    } else if (boost::regex_search(start, end, match, re_capacitance, flags)){
     got_capacitance = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> *capacitance;
    } else if (boost::regex_search(start, end, match, re_ibgn, flags)){
     got_ibgn = true;
    } else if (got_ibgn){
     if       (boost::regex_search(start, end, match, re_inte, flags)){
      istringstream ist1(std::string(match[1].first, match[1].second));
      ist1 >> sid2;
      istringstream ist2(std::string(match[2].first, match[2].second));
      ist2 >> iid2;
      istringstream ist3(std::string(match[3].first, match[3].second));
      ist3 >> tmpd;
      site_vmemb_inter(sid2,iid2) = tmpd;
     }
    }
  } // end of parsing

  infile.close();

  int sid = 0, iid=0;
  if (got_ibgn && got_time){
   for (std::list<GcemSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
    iid = 0;
    for (std::list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){
     j->set_gpbvmemb(site_vmemb_inter(sid,iid));
     iid++;
    }
    sid++;
   }
  }

  if (!got_time)
    cerr <<  "GcemSolverOutput::read_mead_out_vmemb: Incomplete input file, " << fn_string << endl;

  if (!do_cap) got_capacitance = true;

  return (got_capacitance && got_vmemb_inter && got_ibgn && got_time);
}


bool GcemSolverOutput::read_mead_out_xd(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator j, const int sid1, const int iid1, d4_tensor<double> &inter)
{
  string fn_string = gcemdir + "/gcem/" + label + ".out";
  const char *fn = fn_string.c_str();
  string line;
  double gborn = 0, gback = 0, tmpd = 0;
  int sid2 = 0, iid2 = 0;

  boost::regex re_born("^Self energy of " + label + " in \\w+\\s*[dD]ielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_back("^Interaction energy of " + label + " with background charge set in \\w+\\s*[dD]ielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_ibgn("^Interaction energies of " + label + " with charges of given instances of sites");
  boost::regex re_inte("^\\s*(\\d+)\\s+(\\d+)\\s+:\\s+([eE.+-\\d]+)");
  boost::regex re_time("^Total runtime");

  boost::match_flag_type flags = boost::match_default;
  boost::match_results<std::string::const_iterator> match;

  bool got_born = false, got_back = false, got_time = false, got_ibgn = false;
  std::string::const_iterator start, end;

  ifstream infile(fn);

  if (!infile)
    ::error ("GcemSolverOutput::read_mead_out_xd: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (!infile.eof() ){
    getline (infile, line);
    start = line.begin();
    end   = line.end();
    if        (boost::regex_search(start, end, match, re_time, flags)){
     got_time = true;
     break;
    } else if (boost::regex_search(start, end, match, re_born, flags)){
     got_born = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gborn;
    } else if (boost::regex_search(start, end, match, re_back, flags)){
     got_back = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gback;
    } else if (boost::regex_search(start, end, match, re_ibgn, flags)){
     got_ibgn = true;
    } else if (got_ibgn){
     if       (boost::regex_search(start, end, match, re_inte, flags)){
      istringstream ist1(std::string(match[1].first, match[1].second));
      ist1 >> sid2;
      istringstream ist2(std::string(match[2].first, match[2].second));
      ist2 >> iid2;
      istringstream ist3(std::string(match[3].first, match[3].second));
      ist3 >> tmpd;
      inter(sid1,iid1,sid2,iid2) = tmpd;
     }
    }
  } // end of parsing

  infile.close();

  if (got_born) j->set_gpbborn_prot(gborn);
  if (got_back) j->set_gpbback_prot(gback);
  if (!got_time)
    cerr <<  "GcemSolverOutput::read_mead_out_xd: Incomplete input file, " << fn_string << endl;

  return (got_born && got_back && got_ibgn && got_time);
}

bool GcemSolverOutput::read_mead_out_2d(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator j)
{
  string fn_string = gcemdir + "/gcem/" + label + "_2d.out";
  const char *fn = fn_string.c_str();
  string line;
  double gborn = 0, gback = 0;

  boost::regex re_born("^Self energy of " + label + " in two dielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_back("^Interaction energy of " + label + " with background charge set in two dielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_time("^Total runtime");

  boost::match_flag_type flags = boost::match_default;
  boost::match_results<std::string::const_iterator> match;

  bool got_born = false; bool got_back = false; bool got_time = false;
  std::string::const_iterator start, end;

  ifstream infile(fn);

  if (!infile)
    ::error ("GcemSolverOutput::read_mead_out_2d: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (!infile.eof() ){
    getline (infile, line);
    start = line.begin();
    end   = line.end();
    if        (boost::regex_search(start, end, match, re_time, flags)){
     got_time = true;
     break;
    } else if (boost::regex_search(start, end, match, re_born, flags)){
     got_born = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gborn;
    } else if (boost::regex_search(start, end, match, re_back, flags)){
     got_back = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gback;
    }
  } // end of parsing

  infile.close();

  if (got_born) j->set_gpbborn_sol(gborn);
  if (got_back) j->set_gpbback_sol(gback);
  if (!got_time)
    cerr <<  "GcemSolverOutput::read_mead_out_2d: Incomplete input file, " << fn_string << endl;

  return (got_born && got_back && got_time);
}

bool GcemSolverOutput::read_mead_out_2d_ff(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator j)
{
  string fn_string = gcemdir + "/gcem/" + label + "_2d_ff.out";
  const char *fn = fn_string.c_str();
  string line;
  double gborn = 0, gback = 0;

  boost::regex re_born("^Self energy of " + label + " in two dielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_back("^Interaction energy of " + label + " with background charge set in two dielectric environment = ([eE.+-\\d]+)\\s*");
  boost::regex re_time("^Total runtime");

  boost::match_flag_type flags = boost::match_default;
  boost::match_results<std::string::const_iterator> match;

  bool got_born = false; bool got_back = false; bool got_time = false;
  std::string::const_iterator start, end;

  ifstream infile(fn);

  if (!infile)
    ::error ("GcemSolverOutput::read_mead_out_2d: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (!infile.eof() ){
    getline (infile, line);
    start = line.begin();
    end   = line.end();
    if        (boost::regex_search(start, end, match, re_time, flags)){
     got_time = true;
     break;
    } else if (boost::regex_search(start, end, match, re_born, flags)){
     got_born = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gborn;
    } else if (boost::regex_search(start, end, match, re_back, flags)){
     got_back = true;
     istringstream ist1(std::string(match[1].first, match[1].second));
     ist1 >> gback;
    }
  } // end of parsing

  infile.close();

  if (got_born) j->set_gpbborn_ffe(gborn);
  if (!got_time)
    cerr <<  "GcemSolverOutput::read_mead_out_2d: Incomplete input file, " << fn_string << endl;

  return (got_born && got_back && got_time);
}

bool GcemSolverOutput::read_mead_out(const GcemMolecule& molecule, const string& gcemdir, const string& molname, std::list<GcemSite> &sites, const bool has_memb, const bool has_con, const bool do_cap, double *capacitance, double *vmemb_inter)
{
  blab2 << "GcemSolverOutput::read_mead_out: starting\n";

   std::string str_in, ilbl, name;
   int nout_ok = 0, nout_tot = 0, sid = 0, iid=0, c=0;
   bool success = false;

   // prepare interaction energy array
   nsites = sites.size();
   nin.init_arr(0,nsites-1,0);
   for (std::list<GcemSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
    nin(c) = i->get_ninst();
    c++;
   }
   inter.init_arr((int)0,nsites-1,nin,(int)0,nsites-1,nin,(double)0);

   // membrane and defined connectivities imply that membrane potentials are to be considered
   if (has_memb && has_con){
    success = read_mead_out_vmemb(gcemdir, molecule.get_name(), sites, do_cap, capacitance, vmemb_inter, nsites, nin);
    if (success) nout_ok++;
    else
      cerr << "GcemSolverOutput::read_mead_out: input file membpot.out incomplete." << endl;
    nout_tot++;
   }

   for (std::list<GcemSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
    int c(((*i).get_sid()).get_chain());
    int s(((*i).get_sid()).get_site());
    std::string str_c, str_s;
    stringstream ss1,ss2;
    ss1 << c; ss1 >> str_c;
    ss2 << s; ss2 >> str_s;
    string slbl = ((*i).get_sid()).get_sitename() + "-" + str_c + "-" + str_s;

    iid=0;
    for (std::list<GcemInstance>::iterator j=i->instances.begin(); j!=i->instances.end(); ++j){

     stringstream ss3;
     ss3 << iid; ss3 >> str_in;
     ilbl = ((*j).get_instid()).get_instname() + "-" + str_in;

     name = slbl + "_" + ilbl;
     
     // my_memb_solver, my_3diel_solver, my_Ndiel_solver
     success = read_mead_out_xd(gcemdir, name, j, sid, iid, inter);
     if (success)
     {
      nout_ok++;
     }
     else
     {
      cerr << "GcemSolverOutput::read_mead_out: input file " << name << ".out incomplete." << endl;
     }
     nout_tot++;

     // my_2diel_solver, model environment (epsref||epsext outside model compound)
     if (molecule.get_rotmap_value(i->get_sid()) != 2){
      success = read_mead_out_2d(gcemdir, name, j);
      if (success)
      {
       nout_ok++;
      }
      else
      {
       cerr << "GcemSolverOutput::read_mead_out: input file " << name << "_2d.out incomplete." << endl;
      }
      nout_tot++;
     } else {
      // set dummy solvent Born energy equal to Born energy in receptor environment
      j->set_gpbborn_sol(j->get_gpbborn_prot());
     }

     // my_2diel_solver, force field environment (eps_ff outside model compound)
     if (molecule.get_rotmap_value(i->get_sid()) == 1 || molecule.get_rotmap_value(i->get_sid()) == 2){
      success = read_mead_out_2d_ff(gcemdir, name, j);
      if (success) nout_ok++;
      else
       cerr << "GcemSolverOutput::read_mead_out: input file " << name << "_2d_ff.out incomplete." << endl;
      nout_tot++;
     }

     iid++;
    }
    sid++;
   }

   if (nout_ok!=nout_tot)
    ::error("GcemSolverOutput::read_mead_out: MEAD solver output files incomplete\n");

  blab2  << "GcemSolverOutput::read_mead_out: finished." << endl;
  blab2  << "Successfully read " << nout_ok << " of " << nout_tot << " MEAD solver output files." << endl;

   if (nout_ok!=nout_tot)
    ::error ("GcemSolverOutput::read_mead_out: MEAD solver output files incomplete\n");


  blab1 << "GcemSolverOutput::read_mead_out: exiting\n";

  return success;
}


// GcemSolverOutput.cc ends here
