/*  
    helper object for GcemMolecule
    handles the MEAD solver output containing
    the electrostatic energies

    parsing functions for output files of
    my_2diel_solver
    my_Ndiel_solver, my_3diel_solver, my_memb_solver
    my_membpot_solver

    Copyright (c) 2007-2011 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#ifndef _GcemSolverOutput_h
#define _GcemSolverOutput_h 1

#include <vector>
#include <map>
#include <list>
#include <string>
#include "MEAD_e/CharmmID.h"
#include "MEAD_e/InstanceID.h"
#include "MEAD_e/GcemMolecule.h"
#include "MEAD_e/GcemSite.h"
#include "MEAD_e/globals.h"
#include <fstream>
#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/list.hpp>
// these are the irregularly shaped n-dimensional matrices
#include "./irreg_matrices/vector.hpp"    // 1D
#include "./irreg_matrices/matrix.hpp"    // 2D
#include "./irreg_matrices/tensor.hpp"    // 3D
#include "./irreg_matrices/d4_tensor.hpp" // 4D

using std::vector;
using std::list;
using std::map;
using std::pair;
using std::string;

class GcemMolecule;

//!wrap!
class GcemSolverOutput {
public:
  GcemSolverOutput();
  GcemSolverOutput (const string& name);
  GcemSolverOutput (const GcemSolverOutput& a);
  ~GcemSolverOutput () {}

  string get_name() const {return name;}
  void set_name(std::string n) {name = n;}

  // read output files, return true if all output files could be read */
  bool read_mead_out(const GcemMolecule &molecule, const string& gcemdir, const string& name, std::list<GcemSite> &sites, const bool has_memb, const bool has_con, const bool do_cap, double *capacitance, double *vmemb_inter);

private:
  friend class GcemMolecule;
  friend class boost::serialization::access;

  bool read_mead_out_xd(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator iiter, const int sid, const int iid, d4_tensor<double> &inter);
  bool read_mead_out_2d(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator iiter);
  bool read_mead_out_2d_ff(const string& gcemdir, const string& label, std::list<GcemInstance>::iterator iiter);
  bool read_mead_out_vmemb(const string& gcemdir, const string& name, std::list<GcemSite> &sites, const bool do_cap, double *capacitance, double *vmemb_inter, const int nsites, vec<int> &nin);

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & nsites;
    ar & nin;
    ar & inter;
    ar & name;
  }
  int nsites;
  vec<int> nin;
  d4_tensor<double> inter;
  string name;
};



#endif

// GcemSolverOutput.h ends here
