/* class for calculations relating to Ion-accessible volumes of molecules
    Copyright (c) 1993--1995 by Donald Bashford and Tony You.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: IonAccVol.cc,v 2.14 2004/12/06 17:58:39 bashford Exp $
*/

#include "MEAD_e/IonAccVol.h"
#include "MEAD_e/PhysCond.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/Shell.h"
#include "MEAD_e/Sausage.h"
#include <sstream>
#include <string>
using std::ios;
using std::string;
using std::vector;

IonAccVol::IonAccVol()
{
  AtomSet a;
  rep = new IonAccVolRep(a, PhysCond::get_sterln());
}

IonAccVol::IonAccVol(const AtomSet& a)
{
  rep = new IonAccVolRep(a, PhysCond::get_sterln());
}

IonAccVol::IonAccVol(const AtomSet& a, float probe_radius)
{
  rep = new IonAccVolRep(a, probe_radius);
}

IonAccVol::IonAccVol(const IonAccVol& s)
{
  s.rep->referenceCount++;
  rep = s.rep;
}

IonAccVol&
IonAccVol::operator= (const IonAccVol& s)
{
  if (this != &s) {
    s.rep->referenceCount++;
    if (rep->referenceCount == 1)
      delete rep;
    else
      --rep->referenceCount;
    rep = s.rep;
  }
  return *this;
}

IonAccVol::~IonAccVol()
{
  if (rep->referenceCount == 1)
    delete rep;
  else
    --rep->referenceCount;
}

void
IonAccVol::anal_calc()
{
  rep->anal_calc();
}

void
IonAccVol::calc_cuberep(const CubeLatSpec& cls, AccTag* atg)
{
  rep->calc_cuberep(cls, atg);
}

void
IonAccVol::tag_points(int npts, const Coord* pt, AccTag* acc_array)
{
  rep->tag_points(npts, pt, acc_array);
}

int
IonAccVol::accessible(const Coord& r)
{
  return rep->accessible(r);
}

int
IonAccVol::check_is_calculated() const
{
  return rep->check_is_calculated();
}

void IonAccVol::write_top_in_binary() const
{
 rep->write_top_in_binary();
}

void IonAccVol::write_top_in_ascii() const
{
 rep->write_top_in_ascii();
}

int IonAccVol::read_top_in_binary()
{
 return rep->read_top_in_binary();
}

int IonAccVol::read_top_in_ascii()
{
 return rep->read_top_in_ascii();
}

void IonAccVol::write_top_in_binary(ostream& ost) const
{
 rep->write_top_in_binary(ost);
}

void IonAccVol::write_top_in_ascii(ostream& ost) const
{
 rep->write_top_in_ascii(ost);
}

int IonAccVol::read_top_in_binary(istream& ist)
{
 return rep->read_top_in_binary(ist);
}

int IonAccVol::read_top_in_ascii(istream& ist)
{
 return rep->read_top_in_ascii(ist);
}

void IonAccVol::write_top_in_binary(const string& filename) const
{
 rep->write_top_in_binary(filename);
}

void IonAccVol::write_top_in_ascii(const string& filename) const
{
 rep->write_top_in_ascii(filename);
}

int IonAccVol::read_top_in_binary(const string& filename)
{
 return rep->read_top_in_binary(filename);
}

int IonAccVol::read_top_in_ascii(const string& filename)
{
 return rep->read_top_in_ascii(filename);
}

//-------- IonAccVolRep, where the real work happens -----------

int IonAccVolRep::instanceCount = 0;

IonAccVolRep::IonAccVolRep(const AtomSet& a, float probe_radius)
: Rprob(probe_radius)
{
  blab2 << "IonAccVolRep constructor called with AtomSet" << endl;
  ++instanceCount;
  referenceCount = 1;
  sph_count = 0;
  sausage_count = 0;
  is_calculated = 0;
  sglist = 0;
  atsph = 0;
  // Create a name from the AtomSet and radius.
  // This will be used for save-file name so must be unique for a given
  // molecule coords, radii and probe radii.
  // Assume name of the AtomSet is already unique for coords and atom radii.
  {
    std::ostringstream ost;
    ost << a.get_name() << "_Rprob_" << Rprob << "_anal_IonAccVol"
	<< std::flush;
    name = ost.str();
  }

  // Transfer atom data to the shell list.
  // Later in anal_calc it will be necessary to eliminate completely
  // buried atoms in a rewrite of shell list.
  if( a.size() == 0 ) { // No atoms is not a problem!
    blab2 << "IonAccVol constructor No atoms so empty data in this IonAccVol"
      << "\nAll space will be free"  << endl;
    sph_count = 0;
    is_calculated = 1;
    return;
  }
  atsph = new Shell [a.size()];
  int atcount = 0;
  for(AtomSet::const_iterator ip = a.begin(); ip!=a.end(); ++ip ) {
    // TRU 2007, FixMe:
    // this line calling an altered Shell constructor is the only real
    // difference of IonAccVol and SolvAccVol,
    // better organization in one class would be nice
    Shell sp(ip->second, Rprob, Rprob);
    if (sp.get_inner_rad() <= 0.0) continue;  // Zero rad atoms not seen.
    atsph[atcount] = sp;
    atcount++;
  }
  sph_count = atcount;
  blab3 << "Finishing copy atomic information from AtomSet\t# of atom ="
    << atcount << endl;;
}

IonAccVolRep::~IonAccVolRep()
{
  if (--referenceCount)
    cerr << "WARNING: IonAccVolRep destructor called with referenceCount = "
      << referenceCount << endl;
  if (sph_count) delete [] atsph;
  if (sausage_count) delete [] sglist;
  --instanceCount;
}

int
IonAccVolRep::check_is_calculated() const
{return is_calculated;}

int
IonAccVolRep::instances()
{return instanceCount;}

void IonAccVolRep::write_top_in_binary() const
{
  string filename = name + ".dat";
  write_top_in_binary(filename);
}

void IonAccVolRep::write_top_in_binary(const string& filename_string) const
{
  const char *filename = filename_string.c_str();
  ofstream mol_top_in_binary(filename, ios::out | ios::binary);
  if( !mol_top_in_binary ) {
    cerr << "ERROR: file_name: " << filename << "\n";
    cerr <<"IonAccVolRep::write_top_in_binary: couldn't open for witing\n";
  }
  else
    write_top_in_binary(mol_top_in_binary);
}

void IonAccVolRep::write_top_in_binary(ostream& mol_top_in_binary) const
{
  if (!is_calculated) {
    cerr << "WARNING: IonAccVolRep::write_top_in_binary called\n"
      << "when analytical representation not yet calculated" << endl;
    return;
  }
  mol_top_in_binary.write((char *) &Rprob, sizeof(float));
  mol_top_in_binary.write((char *) &sph_count, sizeof(float));
  int n;
  for(n=0; n<sph_count; n++) {
    atsph[n].write_top_in_binary(mol_top_in_binary);
  }
  mol_top_in_binary.write((char *) &sausage_count, sizeof(float));
  for( n=0; n<sausage_count; n++) {
    sglist[n].write_top_in_binary(mol_top_in_binary);
  }
  return;
}

void IonAccVolRep::write_top_in_ascii() const
{
  string filename = name + ".txt";
  write_top_in_ascii(filename);
}

void IonAccVolRep::write_top_in_ascii(const string& filename_string) const
{
  const char *filename = filename_string.c_str();
  ofstream mol_top_in_ascii(filename);
  if( !mol_top_in_ascii.good() ) {
    cout << "file_name: " << filename << "\n";
    cerr <<"IonAccVolRep::write_trop_in_binary: couldn't open for witing\n";
  }
  else
    write_top_in_ascii(mol_top_in_ascii);
}
void IonAccVolRep::write_top_in_ascii(ostream& mol_top_in_ascii) const
{
  if (!is_calculated) {
    cerr << "WARNING: IonAccVolRep::write_top_in_binary called\n"
      << "when analytical representation not yet calculated" << endl;
    return;
  }
  mol_top_in_ascii << "Ion Accessible Volume Elements of " << name << endl;
  mol_top_in_ascii << endl;
  mol_top_in_ascii << "Probe radius is " << Rprob << endl;
  mol_top_in_ascii << endl;
  mol_top_in_ascii << "Number of shells is " << sph_count << endl;
  mol_top_in_ascii << endl;
  int n;
  for(n=0; n<sph_count; n++)
     {
      mol_top_in_ascii << "Shell " << n+1 << "   ";
      atsph[n].write_top_in_ascii(mol_top_in_ascii);
      mol_top_in_ascii << endl;
     }
  mol_top_in_ascii << "Number of sausages is " << sausage_count << endl;
  for( n=0; n<sausage_count; n++)
     {
      mol_top_in_ascii << "Sausage " << n+1 << "   ";
      sglist[n].write_top_in_ascii(mol_top_in_ascii);
      mol_top_in_ascii << endl;
     }
}

int IonAccVolRep::read_top_in_binary()
{
  string filename = name + ".dat";
  return read_top_in_binary(filename);
}
int IonAccVolRep::read_top_in_binary(const string& filename_string)
{
  const char *filename = filename_string.c_str();
  blab2 << "IonAccVol::read_top_in_binary(" << filename << ") called" << endl;
  ifstream ftop_file(filename, ios::in | ios::binary);
  if ( !ftop_file.good() )
    return 0;
  return read_top_in_binary(ftop_file);
}

int IonAccVolRep::read_top_in_binary(istream& ftop_file)
{
 float sol_prob;
 ftop_file.read( (char*) &sol_prob, sizeof(float) );
 if( !ftop_file )
   {
    cerr << "ERROR : IonAccVolRep::read_top_in_binary: " << endl;
    cerr << "Unable to read Ion probe radius" << endl;
    return 0;
   }
 if (Rprob != sol_prob)
   cerr << "ERROR: IonAccVolRep::read_top_in_binary:\n"
     << "Ion probe radius from file does not match current data.\n"
       << "Will ignore value from file and continue." << endl;
 ftop_file.read( (char*) &sph_count, sizeof(float) );
 blab2 << "sph_count = " << sph_count << endl;
 if( !ftop_file )
   {
    cerr << "ERROR : IonAccVolRep::read_top_in_binary: " << endl;
    cerr << "Unable to read shell_count" << endl;
    return 0;
   }
 if( sph_count <= 0 )
   {
    cerr << "WARNING : IonAccVolRep::read_top_in_binary: " << endl;
    cerr << "sph_count <= zero is detected." << endl;
   }
 Shell *shell_list = new Shell[sph_count];
 int n;
 for(n=0; n<sph_count; n++)
    {
     if( shell_list[n].read_top_in_binary(ftop_file) == 0 )
         return 0;
    }
 blab2 << "Number of shells read in = " << sph_count << endl;
 ftop_file.read( (char*) &sausage_count, sizeof(int) );
 if( !ftop_file )
   {
    cerr << "ERROR : IonAccVolRep::read_top_in_binary: " << endl;
    cerr << "Unable to read sausage_count " << endl;
    return 0;
   }
 if( sausage_count <= 0 )
   {
    cerr << "WARNING : IonAccVolRep::read_top_in_binary: " << endl;
    cerr << "sausage_count <= zero is detected." << endl;
   }
 Sausage *sausage_list = new Sausage[sausage_count];
 for( n=0; n<sausage_count; n++)
    {
     if( sausage_list[n].read_top_in_binary(ftop_file) == 0 )
         return 0;
    }
 blab2 << "Number of sausages read in = " << sausage_count << endl;
 delete [] atsph;
 delete [] sglist;
 atsph = shell_list;
 sglist = sausage_list;
 return is_calculated = 1;
}

int IonAccVolRep::read_top_in_ascii()
{
  string filename = name + ".txt";
  return read_top_in_ascii(filename);
}
int IonAccVolRep::read_top_in_ascii(const string& filename_string)
{
  const char *filename = filename_string.c_str();
  blab2 << "IonAccVol::read_top_in_ascii(" << filename << ") called" << endl;
  ifstream ftop_file(filename, ios::in);
  if ( !ftop_file.good() )
    return 0;
  return read_top_in_ascii(ftop_file);
}

int IonAccVolRep::read_top_in_ascii(istream& ftop_file)
{
 string word;
 int n;
 for(n=0; n<9; n++ )
      ftop_file >> word;
 float sol_prob;
 ftop_file >> sol_prob;
 if (Rprob != sol_prob)
   cerr << "ERROR: IonAccVolRep::read_top_in_binary:\n"
     << "Ionent probe radius from file does not match current data.\n"
       << "Will ignore value from file and continue." << endl;
 for(n=0; n<4; n++ )
      ftop_file >> word;
 ftop_file >> sph_count;
 Shell *shell_list = new Shell[sph_count];
 for( n=0; n<sph_count; n++ )
    {
     if( shell_list[n].read_top_in_ascii(ftop_file) == 0 )
       {
        cerr << "ERROR : IonAccVolRep::read_top_in_ascii : " << endl;
        return 0;
       }
    }
 blab2 << "Number of shells read in = " << sph_count << endl;
 for( n=0; n<4; n++ )
      ftop_file >> word;
 ftop_file >> sausage_count;
 Sausage *sausage_list = new Sausage[sausage_count];
 for( n=0; n<sausage_count; n++ )
    {
     if( sausage_list[n].read_top_in_ascii(ftop_file) == 0 )
       {
        cerr << "ERROR : IonAccVolRep::read_top_in_ascii : " << endl;
        return 0;
       }
    }
 blab2 << "Number of sausages read in = " << sausage_count << endl;
 delete [] atsph;
 delete [] sglist;
 atsph = shell_list;
 sglist = sausage_list;
 return is_calculated = 1;
}

// IonAccVol.cc ends here
