/* ChargeDist subclass for ion gradients, membrane potential simulation

    Modified 2006--2007 from AtomChargeSet

    Thomas Ullmann, BISB Group,
    Dep. of Biopolymers, University of Bayreuth,
    Universitaetsstrasse 30 BGI, 95447 Bayreuth, Germany
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: MembChargeSet.h,v 2.14 2004/12/06 17:57:53 bashford Exp $
*/

#ifndef _MembChargeSet_h
#define _MembChargeSet_h 1

#include "MEAD_e/ChargeDist.h"
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/AtomChargeSet.h"
#include "MEAD_e/IonAccVol.h"
#include "MEAD_e/PhysCond.h"
#include "MEAD_e/ElstatPot.h"

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   1 atom set for the united QM, normal + cavity regions
   It is meant to complement the SixValueDielMembThreeAtoms/ElectrolyteMembAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

class OnePointCharge;
class ManyPointCharge;

//!wrap!
class MembChargeSetMembAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSetMembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSetMembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  MembChargeSetMembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSetMembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  ~MembChargeSetMembAtoms(){}
  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSetMembAtoms& operator=(const MembChargeSetMembAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  virtual int has_charges() const;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!+

  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}
  //virtual ChargeDist::const_iterator pc_begin() const {return 0;}
  //virtual ChargeDist::const_iterator pc_end() const {return 0;}
  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-
  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSetMembAtoms&) {return 0;}

private:
  IonAccVol ion_acc;
  AtomSet _ats;   // For GB/DL's sake
  bool _ats_defined;
  Coord hole_center;
  float zlowerhead, zupperhead, zlowercore, zuppercore, holerad,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSetMembAtoms * multiplyChargeDist (const MembChargeSetMembAtoms& acs, float a);
MembChargeSetMembAtoms * divideChargeDist (const MembChargeSetMembAtoms& acs, float a);
MembChargeSetMembAtoms * addChargeDist (const MembChargeSetMembAtoms& acs1, const MembChargeSetMembAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSetMembAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSetMembAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSetMembAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSetMembAtoms
  MembChargeSetMembAtoms* __mul__ (float a);
  MembChargeSetMembAtoms* __rmul__ (float a);
  MembChargeSetMembAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSetMembAtoms * MembChargeSetMembAtoms___mul__ (const MembChargeSetMembAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembAtoms * MembChargeSetMembAtoms___rmul__ (const MembChargeSetMembAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembAtoms * MembChargeSetMembAtoms___div__ (const MembChargeSetMembAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembTwoAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

//!wrap!
class MembChargeSetMembTwoAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSetMembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSetMembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  MembChargeSetMembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSetMembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  ~MembChargeSetMembTwoAtoms(){}

  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSetMembTwoAtoms& operator=(const MembChargeSetMembTwoAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  // this factor can then be multiplied by the ionic strength at the lattice point to obtain a charge
  virtual int has_charges() const ;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!+

  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}
  //virtual ChargeDist::const_iterator pc_begin() const {return 0;}
  //virtual ChargeDist::const_iterator pc_end() const {return 0;}
  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-
  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSetMembTwoAtoms&) {return 0;}
private:
  IonAccVol ion_acc1, ion_acc2;
  AtomSet _ats1, _ats2;   // For GB/DL's sake
  bool _ats_defined;
  Coord hole_center;
  float zlowerhead, zupperhead, zlowercore, zuppercore, holerad,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const ;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSetMembTwoAtoms * multiplyChargeDist (const MembChargeSetMembTwoAtoms& acs, float a);
MembChargeSetMembTwoAtoms * divideChargeDist (const MembChargeSetMembTwoAtoms& acs, float a);
MembChargeSetMembTwoAtoms * addChargeDist (const MembChargeSetMembTwoAtoms& acs1, const MembChargeSetMembTwoAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSetMembTwoAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSetMembTwoAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSetMembTwoAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSetMembTwoAtoms
  MembChargeSetMembTwoAtoms* __mul__ (float a);
  MembChargeSetMembTwoAtoms* __rmul__ (float a);
  MembChargeSetMembTwoAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSetMembTwoAtoms * MembChargeSetMembTwoAtoms___mul__ (const MembChargeSetMembTwoAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembTwoAtoms * MembChargeSetMembTwoAtoms___rmul__ (const MembChargeSetMembTwoAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembTwoAtoms * MembChargeSetMembTwoAtoms___div__ (const MembChargeSetMembTwoAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE


/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane boundaries
   atom set 3 for a ion accessible region traversing the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembThreeAtoms
*/


//!wrap!
class MembChargeSetMembThreeAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSetMembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot);
  MembChargeSetMembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc);
  MembChargeSetMembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot, float stl);
  MembChargeSetMembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot);
  MembChargeSetMembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc);
  ~MembChargeSetMembThreeAtoms(){}

  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSetMembThreeAtoms& operator=(const MembChargeSetMembThreeAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  virtual int has_charges() const ;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!+

  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}

  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}


  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSetMembThreeAtoms&) {return 0;}

private:
  IonAccVol ion_acc1, ion_acc2, ion_acc3;
  AtomSet _ats1, _ats2, _ats3;   // For GB/DL's sake
  bool _ats_defined;
  float zlowerhead, zupperhead, zlowercore, zuppercore,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const ;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSetMembThreeAtoms * multiplyChargeDist (const MembChargeSetMembThreeAtoms& acs, float a);
MembChargeSetMembThreeAtoms * divideChargeDist (const MembChargeSetMembThreeAtoms& acs, float a);
MembChargeSetMembThreeAtoms * addChargeDist (const MembChargeSetMembThreeAtoms& acs1, const MembChargeSetMembThreeAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSetMembThreeAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSetMembThreeAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSetMembThreeAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSetMembThreeAtoms
  MembChargeSetMembThreeAtoms* __mul__ (float a);
  MembChargeSetMembThreeAtoms* __rmul__ (float a);
  MembChargeSetMembThreeAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSetMembThreeAtoms * MembChargeSetMembThreeAtoms___mul__ (const MembChargeSetMembThreeAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembThreeAtoms * MembChargeSetMembThreeAtoms___rmul__ (const MembChargeSetMembThreeAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSetMembThreeAtoms * MembChargeSetMembThreeAtoms___div__ (const MembChargeSetMembThreeAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE

/*
   The following charge distributions MembChargeSet2xyz are similar to
   the ones before termed MembChargeSet2xyz, but the charge distribution
   generating the membrane potential is only assigned to the headgroup region
*/



//!wrap!
class MembChargeSet2MembAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSet2MembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSet2MembAtoms (const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  MembChargeSet2MembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSet2MembAtoms (const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  ~MembChargeSet2MembAtoms(){}

  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSet2MembAtoms& operator=(const MembChargeSet2MembAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  virtual int has_charges() const ;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!-

  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}
  //virtual ChargeDist::const_iterator pc_begin() const {return 0;}
  //virtual ChargeDist::const_iterator pc_end() const {return 0;}
  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-
  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSet2MembAtoms&) {return 0;}

private:
  IonAccVol ion_acc;
  AtomSet _ats;   // For GB/DL's sake
  bool _ats_defined;
  Coord hole_center;
  float zlowerhead, zupperhead, zlowercore, zuppercore, holerad,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const ;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSet2MembAtoms * multiplyChargeDist (const MembChargeSet2MembAtoms& acs, float a);
MembChargeSet2MembAtoms * divideChargeDist (const MembChargeSet2MembAtoms& acs, float a);
MembChargeSet2MembAtoms * addChargeDist (const MembChargeSet2MembAtoms& acs1, const MembChargeSet2MembAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSet2MembAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSet2MembAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSet2MembAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSet2MembAtoms
  MembChargeSet2MembAtoms* __mul__ (float a);
  MembChargeSet2MembAtoms* __rmul__ (float a);
  MembChargeSet2MembAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSet2MembAtoms * MembChargeSet2MembAtoms___mul__ (const MembChargeSet2MembAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembAtoms * MembChargeSet2MembAtoms___rmul__ (const MembChargeSet2MembAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembAtoms * MembChargeSet2MembAtoms___div__ (const MembChargeSet2MembAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembTwoAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

//!wrap!
class MembChargeSet2MembTwoAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSet2MembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSet2MembTwoAtoms (const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  MembChargeSet2MembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr, float ionic1, float ionic2, float membpot);
  MembChargeSet2MembTwoAtoms (const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           Coord c, float hr);
  ~MembChargeSet2MembTwoAtoms(){}

  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSet2MembTwoAtoms& operator=(const MembChargeSet2MembTwoAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  virtual int has_charges() const ;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!-

  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}
  //virtual ChargeDist::const_iterator pc_begin() const {return 0;}
  //virtual ChargeDist::const_iterator pc_end() const {return 0;}
  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-
  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSet2MembTwoAtoms&) {return 0;}

private:
  IonAccVol ion_acc1, ion_acc2;
  AtomSet _ats1, _ats2;   // For GB/DL's sake
  bool _ats_defined;
  Coord hole_center;
  float zlowerhead, zupperhead, zlowercore, zuppercore, holerad,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const ;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSet2MembTwoAtoms * multiplyChargeDist (const MembChargeSet2MembTwoAtoms& acs, float a);
MembChargeSet2MembTwoAtoms * divideChargeDist (const MembChargeSet2MembTwoAtoms& acs, float a);
MembChargeSet2MembTwoAtoms * addChargeDist (const MembChargeSet2MembTwoAtoms& acs1, const MembChargeSet2MembTwoAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSet2MembTwoAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSet2MembTwoAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSet2MembTwoAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSet2MembTwoAtoms
  MembChargeSet2MembTwoAtoms* __mul__ (float a);
  MembChargeSet2MembTwoAtoms* __rmul__ (float a);
  MembChargeSet2MembTwoAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSet2MembTwoAtoms * MembChargeSet2MembTwoAtoms___mul__ (const MembChargeSet2MembTwoAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembTwoAtoms * MembChargeSet2MembTwoAtoms___rmul__ (const MembChargeSet2MembTwoAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembTwoAtoms * MembChargeSet2MembTwoAtoms___div__ (const MembChargeSet2MembTwoAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE


/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane boundaries
   atom set 3 for a ion accessible region traversing the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembThreeAtoms
*/


//!wrap!
class MembChargeSet2MembThreeAtoms : public ChargeDist_lett, public AtomSet {
public:
  MembChargeSet2MembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot);
  MembChargeSet2MembThreeAtoms (const AtomSet&, const AtomSet&, const AtomSet&, float zlh, float zuh, float zlc, float zuc);
  MembChargeSet2MembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot, float stl);
  MembChargeSet2MembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc,
                           float ionic1, float ionic2, float membpot);
  MembChargeSet2MembThreeAtoms (const IonAccVol&, const IonAccVol&, const IonAccVol&, float zlh, float zuh, float zlc, float zuc);
  ~MembChargeSet2MembThreeAtoms(){}

  //!nowrap!+
  float ionic_strength1() const {return ionic_str1;}
  float ionic_strength2() const {return ionic_str2;}
  float ionic_strength() const {return ionic_str1;}
  float membrane_potential() const {return (memb_pot * PhysCond::get_voltconv());}
  MembChargeSet2MembThreeAtoms& operator=(const MembChargeSet2MembThreeAtoms&);
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&);
  // this get_cube_rep returns the sum of membrane potential generating
  // and induced mobile charge density (total mobile charge density) or
  // the induced mobile charge density only if bool is true/false, repectively
  virtual ChargeCubeRep* get_cuberep(const CubeLatSpec&, const ElstatPot&, const bool);
  const AtomSet& get_atomset() const; // For GB/DL's sake
  //!nowrap!-
  // this factor can then be multiplied by the ionic strength at the lattice point to obtain a charge
  virtual int has_charges() const ;
  virtual float vacuum_coulomb(const Coord& c) const;
  virtual int number_points() const;
  //!nowrap!+


  // These functions seem to be mandatory due to ChargeDist
  virtual float total_charge() const {return 0;}

  //virtual ChargeDist::const_iterator pc_begin() const {return 0;}
  //virtual ChargeDist::const_iterator pc_end() const {return 0;}
  virtual ChargeDist::const_iterator pc_begin() const
    {return make_PointCharge_const_iterator(begin());}
  virtual ChargeDist::const_iterator pc_end() const
    {return make_PointCharge_const_iterator(end());}

  virtual const_iterator begin() const {return AtomSet::begin();}
  virtual const_iterator end() const {return AtomSet::end();}

  //    {return STL_iter_to_CDL_iter_ptr<pair_to_point>(end());}
  //!nowrap!-

  // Functions inherited from AtomSet that update might the charges
  // must be redefined so as not to screw up the charge statistics FIXME!

  virtual bool dispatch_addChargeDist (const ChargeDist_lett& c,
				       ChargeDist &result) const {return 0;}
  ChargeDist addChargeDist (const ManyPointCharge&) {return 0;}
  ChargeDist addChargeDist (const OnePointCharge&) {return 0;}
  ChargeDist addChargeDist (const MembChargeSet2MembThreeAtoms&) {return 0;}

private:
  IonAccVol ion_acc1, ion_acc2, ion_acc3;
  AtomSet _ats1, _ats2, _ats3;   // For GB/DL's sake
  bool _ats_defined;
  float zlowerhead, zupperhead, zlowercore, zuppercore,
        ionic_str1, ionic_str2, memb_pot, sternl;
  void update_charge_stats() const ;
  class charge_accumulator {
    float charge;
  public:
    charge_accumulator() : charge(0.0) {}
    void operator() (const value_type& p) {charge += p.second.charge;}
    float get_sum() const {return charge;}
    void reset() {charge=0;}
  };
  struct is_charged_t {
    bool operator() (const value_type& p) {return p.second.charge != 0.0;}
  } is_charged;
};

// Added to enable the Python interface
// bergsma 5/21/01
MembChargeSet2MembThreeAtoms * multiplyChargeDist (const MembChargeSet2MembThreeAtoms& acs, float a);
MembChargeSet2MembThreeAtoms * divideChargeDist (const MembChargeSet2MembThreeAtoms& acs, float a);
MembChargeSet2MembThreeAtoms * addChargeDist (const MembChargeSet2MembThreeAtoms& acs1, const MembChargeSet2MembThreeAtoms& acs2);

#if SWIGPP_LITERAL_INCLUDE

%addmethods MembChargeSet2MembThreeAtoms {
//
// The operation for multiplying by an ElstatPot_lett derived class.
// The general __rmul__ version is defined in the ElstatPot_lett class.
   float operator* (const MembChargeSet2MembThreeAtoms& acs, const ElstatPot_lett& ep);
//
// The operation for multiplying by an ElstatPotCombination.
// The __ruml__ version is defined in ElstatPotCombination.
   float operator* (const MembChargeSet2MembThreeAtoms& acs, const ElstatPotCombination& epc);
//
// These unusual operators return a pointer to a newly constructed
// object in part because we don't want an expensive copy operation
// everytime __op__ is invoked, but also because they take advantage of
// the addChargeDist routines that are normally used for constructing
// new letter objects for the ChargeDist envelope.
//
// The ways of scaling MembChargeSet2MembThreeAtoms
  MembChargeSet2MembThreeAtoms* __mul__ (float a);
  MembChargeSet2MembThreeAtoms* __rmul__ (float a);
  MembChargeSet2MembThreeAtoms* __div__ (float a);
};

%wrapper %{

BEGIN_CPLUSPLUS_SECTION

// Define our special __mul__ functions here
MembChargeSet2MembThreeAtoms * MembChargeSet2MembThreeAtoms___mul__ (const MembChargeSet2MembThreeAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembThreeAtoms * MembChargeSet2MembThreeAtoms___rmul__ (const MembChargeSet2MembThreeAtoms *acs, float a)
{ return ::multiplyChargeDist(*acs, a); }
MembChargeSet2MembThreeAtoms * MembChargeSet2MembThreeAtoms___div__ (const MembChargeSet2MembThreeAtoms *acs, float a)
{ return ::divideChargeDist(*acs, a); }

END_CPLUSPLUS_SECTION

%}

#endif // SWIGPP_LITERAL_INCLUDE

#endif

// MembChargeSet.h ends here
