/*
    Read electrolyte regions from file
    filename is given with constructor
    file with name filename.ely is read by function read

    file format:
       prefix ionic_strength ion_radius
       "      "              "

    For each ion inaccessible volume (IAV) it is tried to read a precalculated
    analytical representation from file.
    
    Binary file is read from prefix.dat if the file exists
    otherwise, ASCII file is read from prefix.txt if the file exists
    otherwise, the analytical IAV is recalculated from prefix.pqr
    
    Priority increases from the beginning to the end of the list.
    Volumes overide previous volumes on the list.

    A gridpoint is ion accessible if is outside one of gthe IAVs.
    
    e.g.
    membrane_plus_solvent_phase_2   ionic_strength_1   1.4
    membrane_plus_solvent_phase_1   ionic_strength_2   1.4

    This source code file is part of the MEAD_e (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD_e is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/

#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
using std::string;
using std::vector;

#include "MEAD_e/NElectrolyteByFile.h"

NElectrolyteByFile::NElectrolyteByFile (const string& nm) : ElectrolyteEnvironment_lett()
{
  blab2 << "Contructing an NElectrolyteByFile from file for " << nm << endl;
  name = nm;
  read();
}

NElectrolyteByFile::NElectrolyteByFile (const NElectrolyteByFile& a) : ElectrolyteEnvironment_lett(), ionic_str(a.ionic_str), ion_acc_vol(a.ion_acc_vol), name (a.name)
{
}

void NElectrolyteByFile::read(){
  string filename = name + DOT + "ely";
  read(filename);
}

void NElectrolyteByFile::read (const string& filename)
{
  // clear previous definitions
  ionic_str.clear();
  ion_acc_vol.clear();

  float istr, radius;
  string elyfile;
  
  // open file
  ifstream surfin (filename.c_str());
  if (!surfin)
   ::error ("NElectrolyteByFile:: Cannot open input file, ", filename.c_str(), "\n");
  
  // read lines with definitions
  while (surfin >> elyfile >> istr >> radius) {
   AtomSet ats(elyfile);
   ats.read(); // read pqr
   ionic_str.push_back(istr);
   
   IonAccVol tmpiav(ats, radius);
   ion_acc_vol.push_back(tmpiav);
   vector<IonAccVol>::reverse_iterator a = ion_acc_vol.rbegin();

   // try to read precomputed analytical volume, otherwise recalculate
   blab1 << "Try to read in IonAccVol for " << elyfile << ", if impossible calculate it" << endl;
   if (a->read_top_in_binary(elyfile + DOT + "iav" + DOT + "dat") == 0) {
    blab2 << "Unable to read in IonAccVol from binary file for " << elyfile << endl;
    if (a->read_top_in_ascii(elyfile + DOT + "iav" + DOT + "txt") == 0) {
     blab2 << "Unable to read in IonAccVol from ASCII file for " << elyfile << endl;
     blab2 << "Calculating IonAccVol from AtomSet for " << elyfile << endl;
     a->anal_calc(); // this is a bit costly for big molecules
    }
   }

  }
}

ElyCubeRep NElectrolyteByFile::get_cuberep(const CubeLatSpec& cls)
{
  blab3 << "ElectrolyteByAtoms::get_cuberep entered" << endl;

  ElyCubeRep ecr(cls);
  int grid_dim = cls.get_grid_dim();
  int ncube = grid_dim*grid_dim*grid_dim;

  for (int h=0; h<ncube; ++h) {
   ecr[h] = 0.0f;
  }

  // each IonAccVol is mapped to the grid, next overrides previous,
  // only exterior is mapped to grid
  vector<float>::const_iterator b = ionic_str.begin();

  for (vector<IonAccVol>::iterator a = ion_acc_vol.begin(); a!=ion_acc_vol.end(); ++a){

   AccTag* acc_array = new AccTag[ncube];
   a->calc_cuberep(cls, acc_array);

   // acc_array holds the ion accessible volume
   for (int h=0; h<ncube; ++h) {
     if (acc_array[h] == exterior){
       ecr[h] = *b;
     } else if (acc_array[h] != interior) {
       cerr << "acc_array[" << h << "] has unexpected value, "
	 << acc_array[h] << endl;
       ::error("ERROR in ");
     }
   }
   delete [] acc_array;

   b++;
  }

  ecr.declare_defined();
  blab3 << "exit from ElectrolyteByAtoms::get_cuberep\n" << endl;
  
  return ecr;
}
