/*  class RotID is a key type for an associative array representation
    of the Squirell backbone dependent rotamer library

    TRU 2007

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

*/
#ifndef _RotID_h
#define _RotID_h 1

#include <string>
#include <iostream>
#include "MEAD_e/Atom.h"
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>

using std::string;

//!wrap!
class RotID {
public:
  RotID ();
  RotID (const string& residue_name, const int tor_phi, const int tor_psi);
  RotID (const RotID &a) : resname(a.resname), phi(a.phi), psi(a.psi){}
  ~RotID () {}
  //!nowrap!
  RotID& operator= (const RotID&);
  bool operator== (const RotID&) const;
  bool operator!= (const RotID& a) const;
  bool operator<(const RotID&) const;
  bool operator>(const RotID&) const;
  //!nowrap!+
  ostream& print(ostream&);
  string get_resname() const {return resname;}
  int get_phi() const {return phi;}
  int get_psi() const {return psi;}
  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & phi;
    ar & psi;
    ar & resname;
  }
  string resname;
  int phi, psi;
};

#if SWIGPP_LITERAL_INCLUDE

%addmethods RotID {
// Wrap these private members as read-only attributes for Python, but
// this is a bit of a hack (see below).
// It'd probably be better to just wrap the get_member() functions,
// but for Python's use it seems more natural to treat these simple types
// as attributes.
  string resname;
//  string chainid;
  int phi;
  int psi;

// swigpp.el strips the default value of the string argument in the
// above constructor. So, wrap this constructor with the default
// argument removed and...
  RotID (int tor_phi, int tor_psi);
};

%wrapper %{
// ...provide the wrapped constructor with the default argument supplied
RotID * new_RotID(int tor_phi, int tor_psi){
  return new RotID("", tor_phi, tor_psi);
}
// Provide the read-only get/set functions for the aboved wrapped attributes
void RotID_resname_set(RotID *self, string *resname){
   _SWIG_exception(SWIG_ValueError, "The resname attribute is read-only");
}
void RotID_phi_set(RotID *self, int phi){
   _SWIG_exception(SWIG_ValueError, "The phi attribute is read-only");
}
void RotID_psi_set(RotID *self, int psi){
   _SWIG_exception(SWIG_ValueError, "The psi attribute is read-only");
}
string RotID_resname_get(RotID *self){return self->get_resname();}
int RotID_phi_get(RotID *self){return self->get_phi();}
int RotID_phi_get(RotID *self){return self->get_psi();}
// These strings are returned by reference, so create global placeholders
// to stash the results to be returned (read-only).
// Note that this works fine from Python since assignment of
// simple objects like numbers and strings effectively creates a copy.
// However, using these global references certainly is not thread safe!
string _RotID_resname_readonly;
string * RotID_resname_get(RotID *self){_RotID_resname_readonly = self->get_resname(); return &_RotID_resname_readonly;}
%}

#endif // SWIGPP_LITERAL_INCLUDE

inline bool RotID::operator== (const RotID& a) const
{return resname == a.resname && phi == a.phi && psi == a.psi;}


inline bool RotID::operator!= (const RotID& a) const
{return !(*this == a);}

inline RotID& RotID::operator= (const RotID& a)
{ resname = a.resname; phi = a.phi; psi = a.psi;
 return *this;}

inline bool RotID::operator<(const RotID& x) const
{
  if (resname == x.resname) {
    if (phi == x.phi)
      return psi < x.psi;
    else {return phi < x.phi;}
  }
  else {return resname < x.resname;}
}

inline bool RotID::operator>(const RotID& x) const
{
  if (resname == x.resname) {
    if (phi == x.phi)
      return psi > x.psi;
    else {return phi > x.phi;}
  }
  else {return resname > x.resname;}
}


inline ostream& operator<<(ostream& ost, RotID aid)
{return aid.print(ost);}

#endif

// RotID.h ends here
