/*  SiteSet is an associative array of AtomSets
    keyed by SiteID -*- C++ -*-

    TRU, 2007

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/

#ifndef _SiteSet_h
#define _SiteSet_h 1

#include <map>
#include <list>
#include <string>
using std::string;
#include "MEAD_e/AtomSet.h"
#include "MEAD_e/Coord.h"
#include "MEAD_e/SiteID.h"
#include "MEAD_e/globals.h"
#include <vector>
#include <boost/archive/basic_archive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/base_object.hpp>

using std::vector;
using std::list;
using std::map;
using std::pair;

//!wrap!
class SiteSet: public map<SiteID, AtomSet > {
public:
  SiteSet();
  SiteSet (const string& name);
  SiteSet (const SiteSet& a);
  ~SiteSet () {}
  //!nowrap!+
  SiteSet& operator=(const SiteSet& a);
  //!nowrap!-
  void read();
  void create (const AtomSet& a);
  //!nowrap!+
  string get_name() const {return name;}

  // some deviations from STL container functionality:

  // associative lookup, but don't create if missing
  const AtomSet& operator[] (const SiteID&) const;
  //!nowrap!-
  AtomSet& operator[] (const SiteID&);

  //!nowrap!+
  bool contains(const SiteID&) const; // Existence of element (not STL-like)
  pair<iterator,bool> insert (const SiteID&, const AtomSet&); // easier-to-use insert
  //!nowrap!-
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & boost::serialization::base_object<std::map<SiteID, AtomSet> >(*this);
    ar & name;
  }
  string name;
};

#if SWIGPP_LITERAL_INCLUDE

%addmethods SiteSet {
// Wrap this private member as a read-only attribute for Python, but
// this is a bit of a hack (see below).
// It'd probably be better to just wrap the get_name() function,
// but for Python it seems more natural to treat this simple type
// as an attribute.

  string name;
//
// Not really dict-like, but it's the only way to add new atoms
  void insert (const SiteID& i, const AtomSet& a) { self->insert(i, a); }
//
// Functions to emulate a Python dict
//
  void clear () { dict_clear(self); }
  SiteSet * copy() { return new SiteSet(*self); }
  int has_key (const AtomID& k) { return dict_has_key(self, k); }
  int __len__ () { return dict___len__(self); }
  void __delitem__ (const AtomID& k) { dict___delitem__(self, k); }
// Adds all elements from ats
  void update (const SiteSet& ats) { dict_update(self, ats); }
// This function will return a list of SiteID keys from map_SiteID_AtomSet,
// which later is converted to a python list of keys by a typemap
  list_SiteID * keys() { return dict_keys(self); }
// This function will return a list of Atom values from map_SiteID_AtomSet,
// which later is converted to a python list of values by a typemap
  list_AtomPtr * values() { return dict_values(self); }

};

%wrapper %{
//
// Provide the read-only get/set functions for the aboved wrapped attribute
void SiteSet_name_set(SiteSet *self, string *name){
   _SWIG_exception(SWIG_ValueError, "The name attribute is read-only");
}
// This type is returned by reference, so create a global placeholder
// to stash the result to be returned (read-only)
// Note that this works OK from Python since assignment of
// simple objects like numbers and strings effectively creates a copy.
// However, using this global reference certainly is not thread safe!
string _SiteSet_name_readonly;
string * SiteSet_name_get(SiteSet *self){_SiteSet_name_readonly = self->get_name(); return &_SiteSet_name_readonly;}

%}

#endif // SWIGPP_LITERAL_INCLUDE

// associative lookup, but don't create if missing
inline const AtomSet& SiteSet::operator[] (const SiteID& k) const
{
  const_iterator i = find(k);
  if (i == end())
    ::error("SiteSet::operator[]: key not found");
  return i->second;
}

inline AtomSet& SiteSet::operator[] (const SiteID& k)
{
  iterator i = find(k);
  if (i == end())
    ::error("SiteSet::operator[]: key not found");
  return i->second;
}

// Existence of element (not STL-like
inline bool SiteSet::contains(const SiteID& k) const
{
  return find(k) != end();
}

// easier to use insert (not quite STL-like)
inline pair<SiteSet::iterator,bool> SiteSet::insert (const SiteID& k, const AtomSet& a){
  if (contains(k)) {
    ::error("SiteSet::insert: attempt to insert a duplicate atom\n");
  }
  return map<SiteID,AtomSet>::insert(value_type(k, a));
}

#endif

// SiteSet.h ends here
