/*
   R. Thomas Ullmann 2009, based on Mirco Tills objects
   irregularly shaped 1 dimensional array
   boost serialization function
*/
#ifndef VEC_
#define VEC_
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <boost/archive/basic_archive.hpp> 
#include <boost/serialization/split_member.hpp>

using namespace std;
/* Vector template

  Template class for a one dimensional vector. Cares about all the memory allocation things and initialisatio if asked.
  Any given range is possible, no need to start at 0!
  Examples: Integer vector:
  vec<int> example; followed by example.init_arr(1,100) gives you an interger vector from 1 to 100 with 100 elemets.
  vec<int> eample(1,100); Same vector as above;
  vec<int> example; followed by example.init_arr(1,100,0); Same vector as above, but all elements are initialised with 0;
  example(1) = 1; Assignes 1 to the first element.

  TRU 02.08 added new constructors from standard C array

*/
template<class T>
class vec{
private:
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive & ar, const unsigned int file_version){
     boost::serialization::split_member(ar, *this, file_version);
    }
    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
     ar << dim1;
     ar << start1;
     for(int i=start1; i < start1 + dim1; i++){
       ar << arr[i];
     }
    }
    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
     ar >> dim1;
     ar >> start1;
     end1 = start1 + dim1 - 1;
     kill_arr_only();
     init_arr(start1, end1);
     for(int i=start1; i < start1 + dim1; i++){
       ar >> arr[i];
     }
    }
    int   dim1;   /*  Dimension of the vector: rhs - lhs + 1 */
    short start1; /*  Starting point of the vector */
    short end1;   /*  End point of the vector */
    float space;  /*  memory usage */
public:
    T *arr; /*< The vector itself */
    /* Constructor without allocation and initialisation

        initial has to be called for allocating memory
    */
    vec(): dim1(0), start1(0), space(sizeof(vec)), arr(NULL){
    };
    /* Constructor with allocation but without initialisation

        Memory is allocated by init_arr(int, int)
        \param st Starting index
        \param stop Last index
    */
    vec(const int st, const int stop):dim1(stop-st+1),start1(st),end1(stop),space(sizeof(vec)), arr(NULL){
        init_arr(st,stop);
    };
    /* Constructor with allocation and initialisation

        Memory is allocated by init_arr(int, int, T) and all elements are set to ini
        \param st Starting index
        \param stop Last index
        \param ini Initialising value for all elements
    */
    vec(const int st, const int stop, const int ini):dim1(stop-st+1),start1(st),end1(stop),space(sizeof(vec)), arr(NULL){
        init_arr(st,stop,ini);
    };
    /* Copyconstructor

        \param &ini_vec Constant reference of the vector used for initialisation
    */
    vec(const vec<T> &ini_vec){
        init_arr(ini_vec);
    };
    /* Copyconstructor from C array

        \param &ini_vec Constant reference of the vector used for initialisation
    */
    vec(const T *ini_vec, const int istart, const int isize){
        init_arr(ini_vec, istart, isize);
    };
    /* Initialise vector from another vector

        Takes the start value and the dimension from another vector of the same type and calls init_arr(int, int)
        for the new vector. Afterwards copies the values of the passed vector to the new one.
        \param &ini_vec Constant reference of the vector used for initialisation
    */
    void init_arr(const vec<T> &ini_vec){
        arr = NULL;
        init_arr(ini_vec.get_start(),ini_vec.get_end());
        for(int i=ini_vec.get_start(); i<=ini_vec.get_end();i++){
            arr[i] = ini_vec(i);
        }
    };
    /* Initialise vector from a standard C array

        Takes the start value and the dimension from another vector of the same type and calls init_arr(int, int)
        for the new vector. Afterwards copies the values of the passed vector to the new one.
        \param &ini_vec Constant reference of the vector used for initialisation
    */
    void init_arr(const T *ini_vec, const int istart, const int isize){
        arr = NULL;
        int iend = isize + istart - 1;
        init_arr(istart, iend);
        for(int i=istart; i<=iend;i++){
            arr[i] = ini_vec[i];
        }
    };
    /* Initialises the supporting values and calls memory allocator

        \param x Starting index
        \param x2 Last index
        */
    void init_arr(const int x, const int x2){
        if (arr) kill_arr(); // prevent memory leaks
        dim1 = x2 - x + 1;
        start1 = x;
        end1 = x2;
        try{
            arr = new T[dim1];
            arr -= start1;
            space += dim1*sizeof(T);
        }catch (std::bad_alloc&){
            cerr << "Memory allocation problem in vec. dim1:" << dim1 << endl;
            exit(0);
        };
    };
    /* Initialises the supporting values and calls memory allocator

        After allocation, all elements are set to ini
        \param x Starting index
        \param x2 Last index
        \param ini Initialising value for all elemetns
    */
    void init_arr(const int x, const int x2, const T ini){
        init_arr(x,x2);
        for(int i=x; i<=x2;i++){
            arr[i] = ini;
        }
    };
    /* frees memory
       The memory allocated for the array is freed and pointers and space is set back.
    */
    void kill_arr(){
        if(arr != NULL){
            arr += start1;
            delete [] arr;
        };
        dim1 = start1 = end1 = 0;
        space = sizeof(vec);
        arr = NULL;
    };
    void kill_arr_only(){
        if(arr != NULL){
            arr += start1;
            delete [] arr;
        };
        arr = NULL;
    };
    /* Adds one element at the end of the array */
    void add_element(T a){expand_arr(1); arr[end1] = a;}
    /* Adds memory for n elements at the end of the array.

        Allocates memory for an array with the new size and copies the values of the original array to the new one.
        \param n number of elements to add.
    */
    void expand_arr(const int n){
        if(arr == NULL){
            init_arr(1,n);
            return;
        };
        T *tmp_arr = NULL;
        tmp_arr = allocate_arr(tmp_arr,dim1+n,start1);
        for(int i = start1; i<start1 + dim1; i++){
            tmp_arr[i] = arr[i];
        };
        kill_arr_only();
        arr = tmp_arr;
        tmp_arr = NULL;
        dim1 += n;
        end1 += n;
        space = sizeof(vec) + dim1*sizeof(T);
    };
    /* Memory allocator

        Tries to allocate memory in any asked size for the array
        \param *arr Pointer to the core array
        \param dim1 Number of elements
        \param start1 First index
    */

    T* allocate_arr(T *arr, const int dim1, const int start1){
        try{
            arr = new T[dim1];
            arr -= start1;
        }catch (std::bad_alloc&){
            cerr << "Memory allocation problem in vec. dim1:" << dim1 << endl;
            exit(0);
        };
        return arr;
    };
    /* Dimension of the array
    */
    int get_dim() const {return dim1;};
    /* Index of the first element */
    int get_start() const{return start1;}
    int get_end() const{return end1;}
    /* Memory usage of the array */
    double get_space(){return space;};

    T max(T start) const {
      T m = start;
      for(int i=1; i<=dim1 + start1 -1; i++){
        if(m < arr[i]){m = arr[i];}
      };
      return m;
    }

    double get_sum() const {
	T sum = 0;
	for(int i=start1; i<= dim1 + start1 -1; i++){
	    sum += arr[i];
	}
	return sum;
    }
    double is_inside(T &search) const {
      for(int i=start1; i<= dim1 + start1 -1; i++){
        if(arr[i] == search){
          return 1;
        }
      }
      return 0;
    }
    int search_inside(T &search) const {
      for(int i=start1; i<= dim1 + start1 -1; i++){
        if(arr[i] == search){
          return i;
        }
      }
      return 0;
    }
    /* Destructor
        Frees the memory by calling kill_arr()
    */
    ~vec(){
        // dellocate array
        kill_arr();
        if (space != sizeof(vec)){
            cerr << "Could not free all memory of vec. Remaining space = " << space << endl;
            exit(0);
        }
    };
    /* Index operator

        Overloaded operator for easy way of indexing
        \param x Index
    */
    T &operator[](const int x){return arr[x];};
    T &operator()(const int x){return arr[x];};
    T &operator()(const int x) const {return arr[x];};
    /* Stream output operator */
    friend ostream& operator<<(ostream &output, vec &v){
        for(int i=v.start1;i<=v.start1+v.dim1-1;i++){
            output << setprecision(2) << fixed << v(i) << " ";
        }
        output << endl;
        return output;
    };
    /* Write the whole vector in binary format */
    void write_binary(ostream &out){
        out.write(reinterpret_cast<char *>(&arr[start1]), sizeof(T)*dim1);
    }
    /* Write the whole vector in binary format rest n items */
    void write_binary(ostream &out, int ni){
      if(ni <= dim1){
        out.write(reinterpret_cast<char *>(&arr[start1]), sizeof(T)*ni);
      }else{
        cerr << "WARNING: ni > dim1 in write_binary!" << endl;
      }
    }
    /* Read a binary vector directly into arr */
    void read_binary(istream &in){
        in.read(reinterpret_cast<char *>(&arr[start1]), sizeof(T)*dim1);
    };
    // assignment operator
    vec<T>& operator=(const vec<T> &ref){
        if (ref.dim1 == dim1){
          if (ref.start1!=start1){
            arr -= (ref.start1-start1);
          }
          for(int i=start1; i<=dim1 + start1 -1; i++){
            arr[i] = ref.arr[i];
          }
        } else {
          kill_arr();
          init_arr(ref);
        }
        return *this;
    };
};

#endif
