#!/usr/bin/perl

use Getopt::EvaP;
use mead;

my($OPT) = {};
my $epsin1;
my $epsin2;

&parse_command_line();
&check_parameter();

$epsvac = 1;


print "Starting solinprot for molecule named $solute in $protein\n";
print "using the following physical conditions\n\n";
mead::PhysCond_print();
print "\n";
print "Solute interior dielectric constant, epsin1 = $epsin1\n";
print "Protein interior dielectric constant, epsin2 = $epsin2\n";
print "Vacuum dielectric constant, epsvac = $epsvac\n";
if(defined($OPT{blab1})) {print "Blab level set to 1\n";}
elsif (defined($OPT{blab2})) {print "Blab level set to 2\n";}
elsif (defined($OPT{blab3})) {print "Blab level set to 3\n";}
else {print "No blab level set.\n";}
print "\n";

#*****************************************************************************
#                            solinprot starts here
#*****************************************************************************

$a1 = new mead::AtomSet($solute);
$a1->read();
$acs1 = new mead::AtomChargeSet($a1);
$rhogen = new mead::ChargeDist($acs1);

$a2 = new mead::AtomSet($protein);
$a2->read();
$acs2 = new mead::AtomChargeSet($a2);
$rhofeel = new mead::ChargeDist($acs2);

$delp = new mead::ThreeValueDielectricByAtoms($a1, $epsin1, $a2, $epsin2);
$eps = new mead::DielectricEnvironment($delp);

$u = mead::unite_as_as($a1, $a2);

$eba = new mead::ElectrolyteByAtoms($u);
$ely = new mead::ElectrolyteEnvironment($eba);

$fdm = new mead::FinDiffMethod();
$fdm->read($solute.".ogm");

$interesting = new mead::Coord(0, 0, 0);
$fdm->resolve($a1->geom_cent(), $interesting);
print "Using finite difference method with lattice levels:\n"; mead::write_fdm($fdm);

$phi = new mead::ElstatPot($fdm,  $eps, $rhogen, $ely);
$phi->solve();

$prod_sol = mead::mul_cd_ep($rhogen, $phi);
print "\nprod_sol = $prod_sol\n";

$protein_interaction = mead::mul_ep_cd($phi, $rhofeel);
$protein_interaction = $protein_interaction * mead::PhysCond_econv_get();
print "Interaction of solute with protein charges = ";
print "$protein_interaction\n(probably in kcal/mole)\n";

mead::PhysCond_epsext_set($epsvac);

$vac_ely = new mead::ElectrolyteEnvironment;
$delp2 = new mead::TwoValueDielectricByAtoms($a1,  $epsin1);
$vac_eps = new mead::DielectricEnvironment($delp2);

$vac_phi = new mead::ElstatPot($fdm, $vac_eps, $rhogen, $vac_ely);
$vac_phi->solve();

$prod_vac = mead::mul_ep_cd($vac_phi, $rhogen);
print "\nprod_vac = $prod_vac\n";
$reac_energy = ($prod_sol - $prod_vac) / 2 * mead::PhysCond_econv_get();
print "\n\nReaction field component of solvation = $reac_energy\n";
print "(probably in kcal/mole)\n";
$solvation_energy = $reac_energy + $protein_interaction;
print "\n\nSOLVATION ENERGY IN PROTEIN = $solvation_energy\n";
print "(probably in kcal/mole)\n";


#*****************************************************************************
#                            solinprot ends here
#*****************************************************************************


sub parse_command_line {

my($OPT) = @_;
my(@PDT,@MM);

$Getopt::EvaP::evap_Help_Hooks{'P_HHBRFL'} = "\n\"Solute.pqr\", \"solute.ogm\" and \"protein.pqr\" are needed.\n\n";
$Getopt::EvaP::evap_Help_Hooks{'P_HHERFL'} = "\nFilenames for solute and protein missing!\n\n";

@PDT = split /\n/, <<'end-of-PDT';
PDT sample
  epsin1, e1: real
  epsin2, e2: real
  epsext, ee: real
  epsvac, ev: real = "1.0"
  solrad, sr: real
  sterln, sl: real
  ionicstr, is: real
  T, t: real
  kBolt, kb: real
  conconv, cc: real
  econv, ec: real
  bohr_radius, br: real
  proton_charge, pc: real
  ReactionField, rf: switch
  ProteinField, pf: switch
  converge_oldway, co: switch
  blab1, b1: switch
  blab2, b2: switch
  blab3, b3: switch
PDTEND required_file_list
end-of-PDT

@MM = split /\n/, <<'end-of-MM';
sample

	    ***************Solinprot******************
	
	Usage:
	
	  ./solinprot.pl -epsin1 n -epsin2 m solute protein

.epsin1
	Dielectric constant of solute interior.
        In old versions of MEAD this had a default value,
        but this caused confusion, so now you must
        specify epsin explicitly.
.epsin2
	Dielectric constant of the regions interior to the 
	protein but exterior to the solute.
        In old versions of MEAD this had a realdefault value,
        but this caused confusion, so now you must
        specify epsin explicitly.
.epsext
	Exterior dielectric constant.
.epsvac
	The dielectric constant of the other
        external media in the solvate program,
        presumably the vacuum.
.solrad
	Solvent probe radius used in rolling ball
	procedure to determine contact surface
	which is taken as the boundary between
	epsin and epsext.  (Angstroms, by default)
.sterln
	Ion exclusion layer thickness which is added
	to atomic radii to determine region inaccessible
	to salt so that kappa term in PB equation
	is zero.  (Angstroms, by default)
.ionicstr
	Ionic strength (moles/liter, by default).
.T
	Temperature (Kelvins by default)
.kBolt
	The Boltzmann constant in units of charge unit
	squared per length unit per temperature unit.
	You must adjust this if you do not use the
	default units of length, charge and temperature
	(protons squared)/(Angstrom * Kelvin, by default)
.conconv
	Conversion factor for going from concentration
	units used for ionic strength to units of
	particles per cubic length unit.  You must adjust
	this if you do not use the default units of length
	and concentration.
	((particles/cubic Angstrom)/(moles/liter), by default)
.econv
	Conversion factor for going from energy units
        of (charge units squared)/(length unit) to
        to whatever energy units you want for your output.
        You must adjust this if you do not use the
	default units of length and charge, or if you
	want output in different energy units.
	((kcal/mole)/(protons squared/Angstrom), by default)
.ReactionField
	Flag to write values of the
        solvent reaction field potential at space
        points specified by the user.  The point should
        be in an input file named MolName.fpt in which
        points are listed in the format, "(x, y, z)"
        Corresponding reaction field potential values
        will be written to a file MolName.rf.
.converge_oldway
	Revert to the old way of testing for convergence
        of the SOR method of solving the finite-difference
        representation of the Poisson--Boltzmann equation.
        The new method gives improved long-range accuracy
        for large lattices, but at a sometimes substantial
        computational cost.  See the NEWS file for further
        discussion.
.blab1
	Verbose level of the program. Now the program gives
	rough information about what´s going on.
.blab2
	Verbose level of the program. The program tells you
	quite a lot.
.blab3
	Verbose level of the program. You should know
	everything you wanted to know now.
end-of-MM

EvaP \@PDT, \@MM, \%OPT;

}

sub check_parameter() {

if(defined($OPT{epsin1})) {
  #print "epsin1 ok\n"
} else {
  die("epsin1 not ok\n");
}
$epsin1 = $OPT{epsin1};

if(defined($OPT{epsin2})) {
  #print "epsin2 ok\n";
} else {
  die("epsin2 not ok\n");
}
$epsin2 = $OPT{epsin2};

($solute, $protein) = @ARGV;
if(-e $solute.".pqr") {
  #print $solute.".pqr found\n";
} else {
  die("Solute pqr-file missing\n");
}
if(-e $solute.".ogm") {
  #print $solute.".ogm found\n";
} else {
  die("Solute ogm-file missing\n");
}
if(-e $protein.".pqr") {
  #print $protein.".pqr found\n";
} else {
  die("Protein pqr-file missing\n");
}

if(defined($OPT{epsext})) {mead::PhysCond_epsext_set($OPT{epsext})};
if(defined($OPT{epsvac})) {$epsvac = $OPT{epsvac};}
if(defined($OPT{solrad})) {mead::PhysCond_solrad_set($OPT{solrad})};
if(defined($OPT{sterln})) {mead::PhysCond_sterln_set($OPT{sterln})};
if(defined($OPT{ionicstr})) {mead::PhysCond_ionicstr_set($OPT{ionicstr})};
if(defined($OPT{T})) {mead::PhysCond_T_set($OPT{T})};
if(defined($OPT{kBolt})) {mead::PhysCond_kBolt_set($OPT{kBolt})};
if(defined($OPT{conconv})) {mead::PhysCond_conconv_set($OPT{conconv})};
if(defined($OPT{econv})) {mead::PhysCond_econv_set($OPT{econv})};
if(defined($OPT{bohr_radius})) {mead::PhysCond_bohr_radius_set($OPT{bohr_radius})};
if(defined($OPT{proton_charge})) {mead::PhysCond_proton_charge_set($OPT{proton_charge})};

if(defined($OPT{ReactionField})) {$doReactionField = 1;};
if(defined($OPT{ProteinField})) {$doProteinField = 1;};
if(defined($OPT{converge_oldway})) {$doconverge_oldway = 1;};

if(defined($OPT{blab1})) {mead::blab1_set()}
if(defined($OPT{blab2})) {mead::blab2_set()}
if(defined($OPT{blab3})) {mead::blab3_set()}

print "Parameters OK\n\n";

}