/*  A QMPB molecule containing titratable and/or flexible sites

    uses Rotlib object for defining sidechain conformations
    uses force field parameters from QmpbMolecule object
    uses initial geometry from an AtomSet

    reads .rot file to determine sites and site types
    reads .est files for titratable sites to get charges and
     bound ligands of forms

    calculates non-electrostatic contributions to intrinsic and
    interaction energies

    based purely on CHARMM energy function or on combination of
    knowledge based potential with CHARMM energy function

    writes energy matrices to files
    writes QMPB input (site structures, input script)

    Copyright (C) 2007 Thomas Ullmann
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford.

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.
*/
#include "MEAD_e/QmpbMolecule.h"
#include "MEAD_e/globals.h"
#include <fstream>
#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <sstream>
#include <stdio.h>
#include <iostream> // ::std::cout 
#include <ostream>  // ::std::ostream (<<) 
#include <iomanip>  // ::std::setw, ::std::setfill 
#include <ios>      // ::std::left, ::std::right
#include <cmath>

using std::istringstream;
using std::stringstream;

using namespace std;
using namespace boost;
namespace fs = boost::filesystem;

QmpbMolecule::QmpbMolecule(const QmpbMolecule& a)
: name(a.name), rotmap(a.rotmap), sitemap(a.sitemap),
  sitedef(a.sitedef), chmpara(a.chmpara), rlib(a.rlib),
  use_filter(a.use_filter), use_prop(a.use_prop),
  filter_cutoff(a.filter_cutoff), eps_ff(a.eps_ff),
  use_term_l_e(a.use_term_l_e), use_term_l_lj(a.use_term_l_lj),
  use_term_l_bond(a.use_term_l_bond), use_term_l_angle(a.use_term_l_angle),
  use_term_l_dihedral(a.use_term_l_dihedral), use_term_l_improper(a.use_term_l_improper),
  use_term_ml_e(a.use_term_ml_e), use_term_ml_lj(a.use_term_ml_lj),
  use_term_ml_bond(a.use_term_ml_bond), use_term_ml_angle(a.use_term_ml_angle),
  use_term_ml_dihedral(a.use_term_ml_dihedral), use_term_ml_improper(a.use_term_ml_improper),
  use_term_s_e(a.use_term_s_e), use_term_s_lj(a.use_term_s_lj),
  use_term_s_bond(a.use_term_s_bond), use_term_s_angle(a.use_term_s_angle),
  use_term_s_dihedral(a.use_term_s_dihedral), use_term_s_improper(a.use_term_s_improper),
  use_term_ms_e(a.use_term_ms_e), use_term_ms_lj(a.use_term_ms_lj),
  use_term_ms_bond(a.use_term_ms_bond), use_term_ms_angle(a.use_term_ms_angle),
  use_term_ms_dihedral(a.use_term_ms_dihedral), use_term_ms_improper(a.use_term_ms_improper)
{
}

QmpbMolecule::QmpbMolecule()
: name("")
{
 set_default();
}

QmpbMolecule::QmpbMolecule (const string& nm)
: name(nm)
{
 set_default();
}

QmpbMolecule::QmpbMolecule (const string& nm, const string& rlibn, const string& chmtopn, const string& chmparn)
: name(nm)
{
 set_default();
 //prepare_qmpb(nm, rlibn, chmtopn, chmparn);
}

QmpbMolecule::QmpbMolecule (const string& nm, const string& rlibn, const string& chmtopn, const string& chmparn, const bool up)
: name(nm)
{
 set_default();
 use_prop = up;
 //prepare_qmpb(nm, rlibn, chmtopn, chmparn);
}

QmpbMolecule::QmpbMolecule (const string& nm, const string& rlibn, const string& chmtopn, const string& chmparn, const float ctof)
: name(nm)
{
 set_default();
 use_filter = true;
 filter_cutoff = ctof;
 //prepare_qmpb(nm, rlibn, chmtopn, chmparn);
}

QmpbMolecule::QmpbMolecule (const string& nm, const string& rlibn, const string& chmtopn, const string& chmparn, const float ctof, const bool up)
: name(nm)
{
 set_default();
 use_prop = up;
 use_filter = true;
 filter_cutoff = ctof;
 //prepare_qmpb(nm, rlibn, chmtopn, chmparn);
}

void QmpbMolecule::set_default()
{
  qmpbdir              = std::string("qmpb_input");
  eps_ff               = 1;
  filter_cutoff        = 30;
  epsin1               = 1;
  epsin2               = 4;
  epsin3               = 80;
  epsext               = 80;
  epscore              = 2;
  epshead              = 15;
  ionicstr             = 0.15;
  ionicstr1            = 0.15;
  ionicstr2            = 0.15;
  has_memb             = false;
  zmaxh                = 0;
  zmaxc                = 0;
  zminc                = 0;
  zminh                = 0;
  meadpath             = "";
  // output flags
  print_chm            = false;
  print_debug          = false;
  // use knowledge based potential for intrinsic energy of Site
  use_prop             = false;
  use_filter           = false;
  // terms involving ligand atoms and no model atoms
  // are floored to lowest energy instance of same form to avoid double counting in Gmodel
  use_term_l_e         = true;
  use_term_l_lj        = true;
  use_term_l_bond      = false;
  use_term_l_angle     = false;
  use_term_l_dihedral  = true;
  use_term_l_improper  = true;
  // terms involving ligand atoms and model atoms (ligand model interactions belongs to mmback)
  // are floored to lowest energy instance of same form to avoid double counting in Gmodel
  use_term_ml_e        = false; // done later with LPBE (G_model_back in QMPB)
  use_term_ml_lj       = true;
  use_term_ml_bond     = false;
  use_term_ml_angle    = false;
  use_term_ml_dihedral = true;
  use_term_ml_improper = true;
  // terms involving no ligand and no model atoms (internal site)
  use_term_s_e         = true; // are floored to lowest energy instance of same form to avoid double counting in Gmodel
  use_term_s_lj        = true;
  use_term_s_bond      = false;
  use_term_s_angle     = false;
  use_term_s_dihedral  = true;
  use_term_s_improper  = true;
  // terms involving no ligand but model atoms (site/model interaction mmback)
  use_term_ms_e        = false; // done later with LPBE (G_model_back in QMPB), and then should be floored again to be consistent!?
  use_term_ms_lj       = true;
  use_term_ms_bond     = false;
  use_term_ms_angle    = false;
  use_term_ms_dihedral = true;
  use_term_ms_improper = true;
}

void QmpbMolecule::prepare_qmpb()
{
  prepare_qmpb(name, "rotlib.txt", "top.inp", "par.inp");
}

void QmpbMolecule::prepare_qmpb(const string& nm, const string& rlibn, const string& chmtopn, const string& chmparn)
{
  // read structure from pqr file
  input_pqr = AtomSet(nm);
  input_pqr.read();
  // read site definitions and parameters
  read_rot(string(nm + ".rot"));
  // read rotamer library
  rlib.read(rlibn);
  chmpara.read(chmtopn, chmparn);
  if (print_chm) chmpara.print();
  // read .est files
  read_site_defs();
  // coordinate construction of sites
  create_sites();
  // calculate MM part of intrinsic and interaction energies
  calc_mmint();
  if (use_filter) filter_instances();
  calc_mmg();
  // write energie matrices to .mmint and .mmg
  write_mmint(nm + ".mmint");
  write_mmg(nm + ".mmg");
  //print_mmint();
  //print_mmg();
  // write input script and structure files for qmpb
  write_qmpb_input("qmpb_input", nm + ".qmpb.in");
}

void QmpbMolecule::create_sites()
{
  blab2 << "Entering QmpbMolecule::create_sites" << endl;

  std::map<SiteID,std::pair<AtomSet,AtomSet> > tmpsites;
  std::map<SiteID,std::map<int,AtomSet> > predef_inst;
  std::map<SiteID,std::map<int,std::pair<float, AtomSet> > > new_inst;

  // look for predefined conformers/forms marked by non-zero instance IDs
  // collect atoms with zero instance ID for each site as basis for generation of rotamers
  for (map<SiteID,int>::const_iterator i = rotmap.begin(); i!=rotmap.end(); ++i){
   AtomSet tmpset1, tmpset2, tmpset3, tmpset4, tmpset5;
   for (AtomSet::const_iterator j = input_pqr.begin(); j!=input_pqr.end(); ++j) {
    // atoms belonging to site
    if (i->first.get_chain() == j->second.chain && i->first.get_site() == j->second.site){
     if (j->second.instance == 0){
      tmpset1.insert(j->second);
     }
     // use predefined instances only for sites marked flexible (1 || 2 in .rot)
     else if (i->second == 1 || i->second == 2){
      tmpset2.insert(j->second);
     }
    }
    // atoms of neighbouring residues needed later for blocking groups of model compound
    if (i->first.get_chain() == j->second.chain && i->first.get_site() == (j->second.site + 1)){
     if (j->second.instance == 0 && (j->second.atname == "C"  || j->second.atname == "O")){
      tmpset3.insert(j->second);
     }
    }
    else if (i->first.get_chain() == j->second.chain && i->first.get_site() == (j->second.site - 1)){
     // pyroglutamate should anyway appear only at the N-terminus
     if (j->second.resname != "PRO"){
      if (j->second.instance == 0 && (j->second.atname == "N" || j->second.atname == "HN" ||
                                      j->second.atname == "H" || j->second.atname == "CA" ||
                                      j->second.atname == "HA")){
       tmpset3.insert(j->second);
      }
     }
     else {
      if (j->second.instance == 0 && (j->second.atname == "N"   || j->second.atname == "CD"  ||
                                      j->second.atname == "HD1" || j->second.atname == "HD2" ||
                                      j->second.atname == "CA"  || j->second.atname == "HA")){
       tmpset3.insert(j->second);
      }
     }
    }
   }
   // look if our site appears to be an amino acid - i.e. if the blocking group rule applies at all
   // otherwise purge the blocking group atoms
   bool b1 = false; bool b2 = false; bool b3 = false; bool b4 = false;
   for (AtomSet::const_iterator j = tmpset1.begin(); j!=tmpset1.end(); ++j) {
    if      (j->second.atname == "C") {b1 = true;}
    else if (j->second.atname == "O") {b2 = true;}
    else if (j->second.atname == "N") {b3 = true;}
    else if (j->second.atname == "CA"){b4 = true;}
   }
   if (b1 && b2 && b3 && b4){
    // check zero net charge of blocking groups - otherwise reset charges with CHARMM standard
    if (tmpset3.total_charge() != 0){
     AtomSet tmpset3b;
     for (AtomSet::const_iterator j = tmpset3.begin(); j!=tmpset3.end(); ++j) {
      if      (j->second.atname == "C") {
      Atom ta(j->second);
      ta.charge = float(0.51);
      tmpset3b.insert(ta);
      }
      else if (j->second.atname == "O") {
       Atom ta(j->second); ta.charge = float(-0.51); tmpset3b.insert(ta);
      }
      // pyroglutamate should anyway appear only at the N-terminus
      if (j->second.resname != "PRO"){
       if      (j->second.atname == "H" || j->second.atname == "HN"){Atom ta(j->second); ta.charge = float(0.31);
                                                                     tmpset3b.insert(ta);}
       else if (j->second.atname == "N") {Atom ta(j->second); ta.charge = float(-0.47); tmpset3b.insert(ta);}
       else if (j->second.atname == "CA"){Atom ta(j->second); ta.charge = float(0.07); tmpset3b.insert(ta);}
       else if (j->second.atname == "HA"){Atom ta(j->second); ta.charge = float(0.09); tmpset3b.insert(ta);}
      }
      else {
       if      (j->second.atname == "N"){Atom ta(j->second); ta.charge = float(-0.29); tmpset3b.insert(ta);}
       else if (j->second.atname == "CD") {Atom ta(j->second); ta.charge = float(0.00); tmpset3b.insert(ta);}
       else if (j->second.atname == "HD1"){Atom ta(j->second); ta.charge = float(0.09); tmpset3b.insert(ta);}
       else if (j->second.atname == "HD2"){Atom ta(j->second); ta.charge = float(0.09); tmpset3b.insert(ta);}
       else if (j->second.atname == "CA"){Atom ta(j->second); ta.charge = float(0.02); tmpset3b.insert(ta);}
       else if (j->second.atname == "HA"){Atom ta(j->second); ta.charge = float(0.09); tmpset3b.insert(ta);}
      }
     }
     // if still no integer charge (most likely incomplete atoms) purge blocking groups
     if (tmpset3b.total_charge() != 0){tmpset3.clear();}
     else {tmpset3 = tmpset3b;}
    }
   }
   else {
    tmpset3.clear();
   }
   // from tmpset1 search atoms occurring in SiteDef or moving atoms in ResTop
   // (actual site) and put in tmpset4, rest + tmpset3 goes into tmpset5
   // (additional atoms of  model compound)
   for (AtomSet::const_iterator j = tmpset1.begin(); j!=tmpset1.end(); ++j){
    bool in_sdef = false;
    if (i->second == 1 || i->second == 3){
     QmpbSiteDef sdef(get_sitedef(i->first.get_sitename()));
     in_sdef = sdef.contains_atom(j->second.resname,j->second.atname);
    }
    // second check ResTop for atoms to move in sidechain torsions
    bool in_sctor = false;
    if (i->second == 1 || i->second == 2){
     CharmmParaResTop restop(chmpara.get_Topology(i->first.get_sitename()));
     in_sctor = restop.moved_by_sctor(j->second.atname);
    }
    // third check for atoms appearing in predefined instances
    AtomID aid(int residue_number, const string& atom_name, const int confomer = 0, const int chain = 0, const int site = 0, const int instance = 0);
    bool in_pdinst = false;
    for (AtomSet::const_iterator k = tmpset2.begin(); k!=tmpset2.end(); ++k){
     if (j->second.conformer == k->second.conformer && j->second.chain == k->second.chain &&
         j->second.site == k->second.site && j->second.resname == k->second.resname &&
         j->second.atname == k->second.atname){
      in_pdinst = true;
      break;
     }
    }
    if (in_sdef || in_sctor || in_pdinst){
     tmpset4.insert(j->second);
    } else {
      tmpset5.insert(j->second);
    }
   }
   for (AtomSet::const_iterator j = tmpset3.begin(); j!=tmpset3.end(); ++j){
    tmpset5.insert(j->second);
   }
   // save collected atoms into tmpsites map
   std::pair<AtomSet,AtomSet> tmppair1(tmpset4,tmpset5);
   tmpsites.insert(std::pair<SiteID,std::pair<AtomSet,AtomSet> >(i->first,tmppair1));
   // flexible MM or titratable sites
   if (i->second == 1 || i->second == 2){
    // arrange atoms of predefined instances sorted by input instance nr
    std::map<int,AtomSet> tmpmap, tmpmap2;
    // check if variable atoms in non-zero instances are also found among non-variable atoms
    for (AtomSet::const_iterator j = tmpset2.begin(); j!=tmpset2.end(); ++j) {
     std::map<int,AtomSet>::iterator k = tmpmap.find(j->second.instance);
     if (k != tmpmap.end()){
      k->second.insert(j->second);
     } else {
      AtomSet ta;
      ta.insert(j->second);
      tmpmap.insert(std::pair<int,AtomSet>(j->second.instance,ta));
     }
    }
    // renumber predefined instances
    int ii = 1;
    for (std::map<int,AtomSet>::const_iterator j = tmpmap.begin(); j!=tmpmap.end(); ++j) {
     tmpmap2.insert(std::pair<int,AtomSet>(ii,j->second));
     ii++;
    }
    // save sorted predefined instances to predef_inst
    predef_inst.insert(std::pair<SiteID,std::map<int,AtomSet> >(i->first,tmpmap2));
   }
  }

  // cross check exclusiveness of real site atoms
  for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
   std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator ti(i); ti++;
   for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator j=ti; j!=tmpsites.end(); ++j){
    for (AtomSet::const_iterator k = i->second.first.begin(); k!=i->second.first.end(); ++k){
     if (j->second.first.contains(k->first)){
      cerr << k->first << endl;
      ::error ("ERROR in QmpbMolecule::create_sites: Atom occurred in two sites.");
     }
    }
   }
  }

  // if a real site atom occurs in the model compound of another site set its charge to zero
  for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
   for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::iterator j=tmpsites.begin(); j!=tmpsites.end(); ++j){
    if (j == i) continue;
    for (AtomSet::iterator k = i->second.second.begin(); k!=i->second.second.end(); ++k){
     if (j->second.first.contains(k->first)){
      k->second.charge = float(0.0);
     }
    }
   }
  }

  // construct background
  for (AtomSet::const_iterator k = input_pqr.begin(); k!=input_pqr.end(); ++k){
   if (k->second.instance != int(0)) continue;
   bool found = false;
   for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
    if (i->second.first.contains(k->first)){
     found = true;
     break;
    }
   }
   if (!found && !background.contains(k->first)){
    background.insert(k->second);
   }
  }

   //////// debug ////////
   if (print_debug){
   for (std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator i=tmpsites.begin(); i!=tmpsites.end(); ++i){
    cout << "!//////////////////////////////////////////////////////" << endl;
    cout << "\nsite atoms of " << i->first << " :" << endl;
    i->second.first.print();
    cout << "......................................................." << endl;
    cout << "\nbackground atoms of " << i->first << " :" << endl;
    i->second.second.print();
   }
   }
   //////// debug ////////

  // now we are ready to construct the rotamers

  for (map<SiteID,int>::const_iterator i = rotmap.begin(); i!=rotmap.end(); ++i){
   // flexible sites and flexible + titratable sites
   if (i->second == 1 || i->second == 2){
    std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator k = tmpsites.find(i->first);
    // this should never occur
    if (k == tmpsites.end()){
     cerr << i->first << endl;
     ::error ("ERROR in QmpbMolecule::create_sites: site is missing while generating rotamers.");
    }
    if (!rlib.contains_residue(i->first.get_sitename())){
     cerr << "Warning: could not find entry in rotamer database for site name "
          << i->first.get_sitename() << ".\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }
    if (!chmpara.contains_Topology(i->first.get_sitename())){
     cerr << "QmpbMolecule::create_sites Warning: could not find topology entry in CHARMM topology for site name "
          << i->first.get_sitename() << ".\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }
    // retrieve topology
    CharmmParaResTop top(chmpara.get_Topology(i->first.get_sitename()));
    if (top.sidechain_tor.size() == 0){
     cerr << "QmpbMolecule::create_sites Warning: topology entry in CHARMM topology for site name "
          << i->first.get_sitename() << " contains no sidechain torsions.\nWill not generate rotamers for flexible site "
          << i->first << endl;
     continue;
    }


    // find coordinates of backbone atoms for this site, compute backbone torsion angles
    bool bb_cplt = true;
    bool got_a = true;
    bool got_b = true;
    bool got_c = true;
    bool got_d = true;
    bool got_e = true;
    float phi, psi;
    Coord a,b,c,d,e;
    AtomID id_a(i->first.get_site()-1,"C",i->first.get_conformer(),i->first.get_chain(),i->first.get_site()-1,int(0));
    AtomID id_b(i->first.get_site(),"N",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_c(i->first.get_site(),"CA",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_d(i->first.get_site(),"C",i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
    AtomID id_e(i->first.get_site()+1,"N",i->first.get_conformer(),i->first.get_chain(),i->first.get_site()+1,int(0));
    if (input_pqr.contains(id_a)){a = input_pqr[id_a].coord;} else {bb_cplt = false; got_a = false;
     blab1 << "QmpbMolecule::create_sites Warning: could not find atom " << id_a << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (input_pqr.contains(id_b)){b = input_pqr[id_b].coord;} else {bb_cplt = false; got_b = false;
     blab1 << "QmpbMolecule::create_sites Warning: could not find atom " << id_b << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (input_pqr.contains(id_c)){c = input_pqr[id_c].coord;} else {bb_cplt = false; got_c = false;
     blab1 << "QmpbMolecule::create_sites Warning: could not find atom " << id_c << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (input_pqr.contains(id_d)){d = input_pqr[id_d].coord;} else {bb_cplt = false; got_d = false;
     blab1 << "QmpbMolecule::create_sites Warning: could not find atom " << id_d << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (input_pqr.contains(id_e)){e = input_pqr[id_e].coord;} else {bb_cplt = false; got_e = false;
     blab1 << "QmpbMolecule::create_sites Warning: could not find atom " << id_e << " for defining backbone torsions of residue " << i->first << endl;
    }
    if (!got_b || !got_c){
     blab1 << "QmpbMolecule::create_sites Warning could not find all atoms necessary for defining sidechain torsions" << endl;
     blab1 << "                           Will not generate rotamers for this residue." << endl;
    }
    ///////////////////////////////////////////////////////////////////////////////////
    // FIXME: this is an ugly hack to be commented out in the trunk version
    // it allows to include my custom peptide flip residue in azurin
    if (!got_a && got_b){
      for (AtomSet::const_iterator ii = input_pqr.begin(); ii != input_pqr.end(); ++ii){
       if (ii->second.resname == "PEP" && ii->second.atname == "2C" && float(abs_value(ii->second.coord-b)) <= float(1.8)){
        a = ii->second.coord; got_a = true;
        blab1 << "QmpbMolecule::create_sites: Found nearby residue with resname PEPT and atom named 2C in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom C of preceding residue." << endl;
        break;
       }
       if (ii->second.resname == "PEP" && ii->second.atname == "6C" && float(abs_value(ii->second.coord-b)) <= float(1.8)){
        a = ii->second.coord; got_a = true;
        blab1 << "QmpbMolecule::create_sites: Found nearby residue with resname PEPT and atom named 6C in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom C of preceding residue." << endl;
        break;
       }       
      }
    }
    if (!got_e && got_d){
      for (AtomSet::const_iterator ii = input_pqr.begin(); ii != input_pqr.end(); ++ii){
       if (ii->second.resname == "PEP" && ii->second.atname == "1N" && float(abs_value(ii->second.coord-d)) <= float(1.8)){
        a = ii->second.coord; got_e = true;
        blab1 << "QmpbMolecule::create_sites: Found nearby residue with resname PEPT and atom named 1N in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom N of succeeding residue." << endl;
        break;
       }
       if (ii->second.resname == "PEP" && ii->second.atname == "4N" && float(abs_value(ii->second.coord-d)) <= float(1.8)){
        a = ii->second.coord; got_e = true;
        blab1 << "QmpbMolecule::create_sites: Found nearby residue with resname PEPT and atom named 4N in appropriate distance" << endl;
        blab1 << "                            Will use coords for atom N of succeeding residue." << endl;
        break;
       }
      }
    }
    ///////////////////////////////////////////////////////////////////////////////////
    if (bb_cplt){
     phi = get_tor(a,b,c,d);
     psi = get_tor(b,c,d,e);
    } else if (!got_a && got_b && got_c && got_d && got_e){
     // take alpha helix parameters
     phi = 57*(pi/180);
     psi = get_tor(b,c,d,e);
    } else if (got_a && got_b && got_c && got_d && !got_e){
     // take alpha helix parameters
     phi = get_tor(a,b,c,d);
     psi = 57*(pi/180);
    } else if (!got_a && got_b && got_c && got_d && !got_e){
     // take alpha helix parameters
     phi = 57*(pi/180);
     psi = 57*(pi/180);
    } else {continue;}


    // prepare temporary atomset with all sidechain torsion angles set to zero
    AtomSet zerochi = k->second.first;
    int nchi = top.sidechain_tor.size() - 1;
    if (nchi > int(3)) nchi = int(3);
    for (int ii=nchi; ii >= 0; --ii){
     Coord o, p, q, r;
     AtomID aido(i->first.get_site(),top.sidechain_tor[ii].first[0],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidp(i->first.get_site(),top.sidechain_tor[ii].first[1],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidq(i->first.get_site(),top.sidechain_tor[ii].first[2],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     AtomID aidr(i->first.get_site(),top.sidechain_tor[ii].first[3],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
     if      (k->second.first.contains(aido)){o = zerochi.find(aido)->second.coord;}
     else if (k->second.second.contains(aido)){o = k->second.second.find(aido)->second.coord;}
     else {
      cerr << "QmpbMolecule::create_sites: Atom " << aido << " is missing." << endl;
      ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidp)){p = zerochi.find(aidp)->second.coord;}
     else if (k->second.second.contains(aidp)){p = k->second.second.find(aidp)->second.coord;}
     else {
      cerr << "QmpbMolecule::create_sites: Atom " << aidp << " is missing." << endl;
      ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidq)){q = zerochi.find(aidq)->second.coord;}
     else if (k->second.second.contains(aidq)){q = k->second.second.find(aidq)->second.coord;}
     else {
      cerr << "QmpbMolecule::create_sites: Atom " << aidq << " is missing." << endl;
      ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
     }
     if      (k->second.first.contains(aidr)){r = zerochi.find(aidr)->second.coord;}
     else if (k->second.second.contains(aidr)){r = k->second.second.find(aidr)->second.coord;}
     else {
      cerr << "QmpbMolecule::create_sites: Atom " << aidr << " is missing." << endl;
      ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
     }
     float chi = get_tor(o,p,q,r);
     // move all atoms indicated by CharmmParaResTop sidechain_tor
     for (std::list<string>::const_iterator jj=top.sidechain_tor[ii].second.begin();jj!=top.sidechain_tor[ii].second.end();++jj){
      AtomID aid(i->first.get_site(),*jj,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomSet::iterator kk;
      if (zerochi.contains(aid)) kk = zerochi.find(aid);
      else{
       cerr << "\nQmpbMolecule::create_sites: Atom " << aid << " is missing." << endl;
       cerr << ("WARNING QmpbMolecule::create_sites: atom is missing while generating rotamers.\n");
      }
      (*kk).second.coord = change_tor(o,p,q,(*kk).second.coord,-chi);
     }
    }
    
    /// debug ///
    if (print_debug){
    if (!fs::is_directory("zerochi")){
     if (fs::exists("zerochi")) fs::remove_all("zerochi");
     fs::create_directory("zerochi");
    }
    stringstream ss,sc;
    std::string sstr, cstr;
    sc << i->first.get_chain(); sc >> cstr;
    ss << i->first.get_site(); ss >> sstr;
    zerochi.write_pqr(std::string("zerochi/" + i->first.get_sitename() + "-" + cstr    
                                             + "-" + sstr + ".zerochi.pqr"));
    }
    /////////////


    // generate all the rotamers given in rlib, save them in newinst

    list<RotData> rlist(rlib[RotID(i->first.get_sitename(),get_tor_bin(phi),get_tor_bin(psi))]);

    int ninst = 0;

    for (std::list<RotData>::const_iterator ll = rlist.begin(); ll != rlist.end(); ++ll){
     ninst++;
     AtomSet tmpinst = zerochi;
     std::vector<float> chi;
     chi.push_back(((*ll).x1 * (pi / 180)));
     chi.push_back(((*ll).x2 * (pi / 180)));
     chi.push_back(((*ll).x3 * (pi / 180)));
     chi.push_back(((*ll).x4 * (pi / 180)));
     float propensity = (*ll).p;
     int nchi = top.sidechain_tor.size() - 1;
     if (nchi > int(3)) nchi = int(3);
     for (int ii=nchi; ii >= 0; --ii){
      Coord o, p, q, r;
      AtomID aido(i->first.get_site(),top.sidechain_tor[ii].first[0],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidp(i->first.get_site(),top.sidechain_tor[ii].first[1],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidq(i->first.get_site(),top.sidechain_tor[ii].first[2],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      AtomID aidr(i->first.get_site(),top.sidechain_tor[ii].first[3],i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
      if      (tmpinst.contains(aido)){o = tmpinst.find(aido)->second.coord;}
      else if (k->second.second.contains(aido)){o = k->second.second.find(aido)->second.coord;}
      else {
       cerr << "QmpbMolecule::create_sites: Atom " << aido << " is missing." << endl;
       ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
      }
      if      (tmpinst.contains(aidp)){p = tmpinst.find(aidp)->second.coord;}
      else if (k->second.second.contains(aidp)){p = k->second.second.find(aidp)->second.coord;}
      else {
       cerr << "QmpbMolecule::create_sites: Atom " << aidp << " is missing." << endl;
       ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
      }
      if      (tmpinst.contains(aidq)){q = tmpinst.find(aidq)->second.coord;}
      else if (k->second.second.contains(aidq)){q = k->second.second.find(aidq)->second.coord;}
      else {
       cerr << "QmpbMolecule::create_sites: Atom " << aidq << " is missing." << endl;
       ::error ("ERROR in QmpbMolecule::create_sites: atom is missing while generating rotamers.");
      }
      // move all atoms indicated by CharmmParaResTop sidechain_tor
      for (std::list<string>::const_iterator jj=top.sidechain_tor[ii].second.begin();jj!=top.sidechain_tor[ii].second.end();++jj){
       AtomID aid(i->first.get_site(),*jj,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
       AtomSet::iterator kk;
       if (tmpinst.contains(aid)) kk = tmpinst.find(aid);
       else{
        cerr << "\nQmpbMolecule::create_sites: Atom " << aid << " is missing." << endl;
        cerr << ("WARNING QmpbMolecule::create_sites: atom is missing while generating rotamers.\n");
       }
       (*kk).second.coord = change_tor(o,p,q,(*kk).second.coord,chi[ii]);
      }
     }
     // save tmpinst
     std::map<SiteID,std::map<int,std::pair<float, AtomSet> > >::iterator ni(new_inst.find(i->first));
     if (ni == new_inst.end()){
      std::pair<float,AtomSet> tmppair1(propensity, tmpinst);
      std::pair<int,std::pair<float,AtomSet> > tmppair2(ninst, tmppair1);
      std::map<int,std::pair<float,AtomSet> > tmpmap;
      tmpmap.insert(tmppair2);
      std::pair<SiteID,std::map<int,std::pair<float, AtomSet> > > tmppair3(i->first,tmpmap);
      new_inst.insert(tmppair3);
     } else {
      std::pair<float,AtomSet> tmppair1(propensity, tmpinst);
      std::pair<int,std::pair<float,AtomSet> > tmppair2(ninst, tmppair1);
      ni->second.insert(tmppair2);
     }

     /// debug ///
     if (print_debug){
     if (!fs::is_directory("tmpinst")){
      if (fs::exists("tmpinst")) fs::remove_all("tmpinst");
      fs::create_directory("tmpinst");
     }
     stringstream sn,ss,sc;
     std::string nstr, sstr, cstr;
     sc << i->first.get_chain(); sc >> cstr;
     ss << i->first.get_site(); ss >> sstr;
     sn << (ninst-1); sn >> nstr;
     tmpinst.write_pqr(std::string("tmpinst/" + i->first.get_sitename() + "-" + cstr
                                             + "-" + sstr + "-" + nstr + ".tmpinst.pqr"));
     }
     /////////////


    }



   }
  }

  // finally create sites

  for (map<SiteID,int>::const_iterator i = rotmap.begin(); i!=rotmap.end(); ++i){

   bool has_rot = false;
   bool has_pdi = false;
   bool has_tsi = false;

   std::map<SiteID,std::map<int,std::pair<float, AtomSet> > >::const_iterator ni(new_inst.find(i->first));
   if (ni != new_inst.end()) has_rot = true;
   std::map<SiteID,std::map<int,AtomSet> >::const_iterator pdi(predef_inst.find(i->first));
   if (pdi != predef_inst.end()) has_pdi = true;
   std::map<SiteID,std::pair<AtomSet,AtomSet> >::const_iterator ts(tmpsites.find(i->first));
   if (ts != tmpsites.end()) has_tsi = true;

   if (!has_tsi && !has_pdi){
     cerr << "QmpbMolecule::create_sites:  Could not find site " << i->first << " in structure." << endl;
     ::error ("ERROR in QmpbMolecule::create_sites: site is missing.");
   }

   QmpbSite tmps(i->first);

   // first insert predefined instances

   if (has_pdi){
    if (i->second == 1){
     int ic = 0;
     int cn = 0;
     for(std::map<int,AtomSet>::const_iterator j=pdi->second.begin(); j!=pdi->second.end(); ++j){
      cn++;
      QmpbSiteDef sdef(get_sitedef(i->first.get_sitename()));
      for (std::list<QmpbFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
       ic++;
       InstanceID tmpid(i->first.get_sitename(),"predef",i->first.get_conformer(),
                        i->first.get_chain(),i->first.get_site(),ic);
       std::map<std::string,int> tmpmap((*k).ligandmap);
       AtomSet tmpa1;
       // insert atoms found in QmpbFormDef with listed charge in site struct
       for (std::map<CharmmID,pair<float, int> >::const_iterator l = (*k).atommap.begin(); l!=(*k).atommap.end();++l){
        // do not include atom with zero occupancy
        if (l->second.second == 0){continue;}
        AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),j->first);
        AtomSet::const_iterator aitr;
        if (j->second.contains(aid)){
         aitr = j->second.find(aid);
        } else {
         cerr << i->first << " " << aid << endl;
         ::error ("ERROR in QmpbMolecule::create_sites: atom in QmpbFormDef missing in predefined site conformer.");
        }
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        tmpatom.charge = l->second.first;
        if (!tmpa1.contains(aid))
         tmpa1.insert(tmpatom);
        // this will not happen
        else
         ::error ("ERROR in QmpbMolecule::create_sites: duplicate atom in QmpbFormDef.");
       }
       // also insert atoms not found in QmpbFormDef but in the predefined instance in the site struct
       for (AtomSet::const_iterator aitr=j->second.begin(); aitr!=j->second.end(); ++aitr){
        if (!tmpa1.contains(aitr->first))
         tmpa1.insert(aitr->second);
       }
       AtomSet tmpa2;
       if (has_tsi){
        tmpa2 = ts->second.second;
       }
       QmpbInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,cn,(*k).get_gmodel());
       tmps.insert_instance(tmpi);
      }
     }
    }
    if (i->second == 2){
     int ic = 0;
     for(std::map<int,AtomSet>::const_iterator j=pdi->second.begin(); j!=pdi->second.end(); ++j){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),"predef",i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      std::map<std::string,int> tmpmap;
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      QmpbInstance tmpi(tmpid,tmpmap,j->second,tmpa2,ic);
      tmps.insert_instance(tmpi);
     }
    }

   }

   // now insert new instances

   if (has_rot){
    if (i->second == 1){
     int ic = 0;
     int cn = 0;
     if (has_pdi) cn = tmps.get_max_confnum();
     for(std::map<int,std::pair<float,AtomSet> >::const_iterator j=ni->second.begin(); j!=ni->second.end(); ++j){
      cn++;
      QmpbSiteDef sdef(get_sitedef(i->first.get_sitename()));
      for (std::list<QmpbFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
       ic++;
       InstanceID tmpid(i->first.get_sitename(),(*k).get_name(),i->first.get_conformer(),
                        i->first.get_chain(),i->first.get_site(),ic);
       std::map<std::string,int> tmpmap((*k).ligandmap);
       AtomSet tmpa1;
       // first insert all atoms found in QmpbFormDef in site_struct
       for (std::map<CharmmID,pair<float, int> >::const_iterator l = (*k).atommap.begin(); l!=(*k).atommap.end();++l){
        // do not include atom with zero occupancy
        if (l->second.second == 0){continue;}
        AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),j->first);
        AtomID aid2(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
        AtomSet::const_iterator aitr;
        if (j->second.second.contains(aid2)){
         aitr = j->second.second.find(aid2);
        }
        //else if (j->second.second.contains(aid)){
        // aitr = j->second.second.find(aid);
        //}
        else if (has_tsi && ts->second.first.contains(aid2)){
         aitr = ts->second.first.find(aid2);
        }
        else {
         cerr << i->first << " " << aid << endl;
         ::error ("ERROR in QmpbMolecule::create_sites: atom in QmpbFormDef missing in generated site conformer.");
        }
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        tmpatom.charge = l->second.first;
        if (!tmpa1.contains(aid))
         tmpa1.insert(tmpatom);
        // this will not happen
        else
         ::error ("ERROR in QmpbMolecule::create_sites: duplicate atom in QmpbFormDef.");
       }
       // second insert all atoms not found in QmpbFormDef but in the new instance
       // (moving atoms with constant charge) in site_struct
       for (AtomSet::const_iterator aitr=j->second.second.begin(); aitr!=j->second.second.end(); ++aitr){
        Atom tmpatom(aitr->second);
        tmpatom.instance = ic;
        AtomID aid(i->first.get_site(),tmpatom.atname,i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),ic);
        if (!tmpa1.contains(aid) && !(*k).contains_atom(CharmmID(tmpatom)))
         tmpa1.insert(aitr->second);
       }
       AtomSet tmpa2;
       if (has_tsi){
        tmpa2 = ts->second.second;
       }
       QmpbInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,cn,(*k).get_gmodel(),j->second.first);
       tmps.insert_instance(tmpi);
      }
     }
    }
    if (i->second == 2){
     int ic = 0;
     for(std::map<int,std::pair<float,AtomSet> >::const_iterator j=ni->second.begin(); j!=ni->second.end(); ++j){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),"rotamer",i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      std::map<std::string,int> tmpmap;
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      QmpbInstance tmpi(tmpid,tmpmap,j->second.second,tmpa2,ic,float(0),j->second.first);
      tmps.insert_instance(tmpi);
     }
    }

   }

   // unflexible, titratable sites

   if (!has_rot && has_tsi){
    int ic = 0;
    if (i->second == 3){
     QmpbSiteDef sdef(get_sitedef(i->first.get_sitename()));
     for (std::list<QmpbFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
      ic++;
      InstanceID tmpid(i->first.get_sitename(),(*k).get_name(),i->first.get_conformer(),
                       i->first.get_chain(),i->first.get_site(),ic);
      std::map<std::string,int> tmpmap((*k).ligandmap);
      AtomSet tmpa1;
      for (std::map<CharmmID,pair<float, int> >::const_iterator l = (*k).atommap.begin(); l!=(*k).atommap.end();++l){
       // do not include atom with zero occupancy
       if (l->second.second == 0){continue;}
       AtomID aid(i->first.get_site(),l->first.get_atname(),i->first.get_conformer(),i->first.get_chain(),i->first.get_site(),int(0));
       AtomSet::const_iterator aitr;
       if (ts->second.first.contains(aid)){
        aitr = ts->second.first.find(aid);
       } else {
        cerr << i->first << " " << aid << endl;
        ::error ("ERROR in QmpbMolecule::create_sites: atom in QmpbFormDef missing in structure.");
       }
       Atom tmpatom(aitr->second);
       tmpatom.instance = ic;
       tmpatom.charge = l->second.first;
       if (!tmpa1.contains(aid))
        tmpa1.insert(tmpatom);
       // this will not happen
       else
        ::error ("ERROR in QmpbMolecule::create_sites: duplicate atom in QmpbFormDef.");
      }
      AtomSet tmpa2;
      if (has_tsi){
       tmpa2 = ts->second.second;
      }
      QmpbInstance tmpi(tmpid,tmpmap,tmpa1,tmpa2,ic,(*k).get_gmodel());
      tmps.insert_instance(tmpi);
     }
    }
   }

   // all instances should have been added, now insert site in list
   insert_site(tmps);

   //////////////// debug ////////////////
   if (print_debug){
   tmps.print();
   }
   ///////////////////////////////////////


  }

   //////////////// debug ////////////////
   if (print_debug){
   create_sitemap();
   write_sites_file();
   }
   ///////////////////////////////////////


  blab1 << "Exiting QmpbMolecule::create_sites" << endl;
}

void QmpbMolecule::calc_mmint()
{
  // calculate intrinsic energies
  // we need to loop over all instances of all sites
  // and calculate Gint from the pure CHARMM energy function or a combination
  // of CHARMM energy function and a knowledge based potential derived from the
  // relative propensities in the rotamer database
  if (use_prop)
   calc_mmint_propensities();
  else
   calc_mmint_charmm();
}

void QmpbMolecule::calc_mmint_propensities()
{

  blab2 << "Entering QmpbMolecule::calc_mmint_propensities" << endl;

  const float one_over_beta = PhysCond::get_Boltzmann_constant() * PhysCond::get_T() 
                              * PhysCond::get_Avogadro_Number() / 4184;
                              
  cout << "RT in kcal/mol = " << one_over_beta << endl;

  for (std::list<QmpbSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
   int stype(get_rotmap_value((*i).get_sid()));
   if (stype != 1 && stype != 2){
   //if (1 == 2){
    blab1 << "QmpbMolecule::calc_mmint_charmm: site with name "
          << (*i).get_name() << " and ID "
          << (*i).get_sid()  << " is not flexible or flexible and titratable.\n"
          << "Will not calculate intrinsic MM energies for this site" << endl;
    continue;
   }
   if (!chmpara.contains_Topology((*i).get_name())){
    cerr << "QmpbMolecule::calc_mmint_charmm Warning: could not find topology entry in CHARMM topology for site name "
         << (*i).get_name() << ".\nWill not calculate intrinsic MM energy for this residue."
         << (*i).get_sid()  << endl;
    continue;
   }
   // retrieve topology
   CharmmParaResTop top(chmpara.get_Topology((*i).get_name()));
   std::list<AtomID> aidlist;
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   std::list<std::string> ligats;
   if (stype == 1 || stype == 3){
    // this is a titratable and flexible site
    // collect all ligand atoms
    QmpbSiteDef sdef(get_sitedef((*i).get_name()));
    for (std::list<QmpbFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
     for (std::map<CharmmID,pair<float, int> >::const_iterator l = (*k).atommap.begin(); l!=(*k).atommap.end();++l){
      // search atom with zero occupancy
      if (l->second.second == 0){ligats.push_back(l->first.get_atname());}
     }
    }
    ligats.sort();
    ligats.unique();
   }

   std::map<InstanceID, std::map<std::string, float> > emap;
   std::map<std::string, float> min_e_map;

   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){

    // all terms involving ligand + site atoms and background - model atoms
    float gmmback = 0;

    // all terms involving ligand atoms and only site atoms
    float gmml = 0;
    float gmml_e = 0;
    float gmml_lj = 0;
    float gmml_bond = 0;
    float gmml_angle = 0;
    float gmml_dihedral = 0;
    float gmml_improper = 0;

    // all terms not involving ligand atoms and only site+model atoms
    // are assumed to be accumulated in the rotamer propensity 
    
    float gprop = 0; 
    
    if ((*j).get_has_propensity()){
     // if the propensity is zero we would get an infinitly large gprop
     // Squirell has propensities in .6f so assume 5.0E-07
     float p = (*j).get_propensity();
     if (p > 0)
      gprop = -one_over_beta * log(p);
     else if (p == 0){
      gprop = -one_over_beta * log(5.0E-07);
      blab3 << (*j).get_instid() << endl;
      blab3 << "QmpbMolecule::calc_mmint_propensities WARNING: Instance has zero propensity.\n" 
            << "5.0E-07 is used instead." << endl; 
     }
     else {
      cerr << (*j).get_instid() << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Instance has negative propensity.");
     }
    }
    else 
     cerr << "QmpbMolecule::calc_mmint_propensities: WARNING instance " << (*j).get_instid() 
          << " has no propensity attached. Will not calculate complete intrinsic energy for this site" << endl;

    // all terms involving ligand atoms and also model atoms
    float mgmml = 0;
    float mgmml_e = 0;
    float mgmml_lj = 0;
    float mgmml_bond = 0;
    float mgmml_angle = 0;
    float mgmml_dihedral = 0;
    float mgmml_improper = 0;

    // calculate gmm_back

    for (AtomSet::const_iterator k=(*j).site_struct.begin(); k!=(*j).site_struct.end(); ++k){
     if (!chmpara.contains_attype(CharmmID(k->second)) && !chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: attype not found.");       
     }
     CharmmParaNB nbk;
     if (chmpara.contains_NB(CharmmID(k->second))){
      nbk = chmpara.get_NB(CharmmID(k->second));
     }
     else if (k->second.atname == "OT1" || k->second.atname == "OT2"){
      nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("O")));
     }
     else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3"){
      nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("HN")));
     }
     else {
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Lennard Jones parameter missing.");
     }     
     for (AtomSet::const_iterator l=background.begin(); l!=background.end(); ++l){
      // skip model atoms, interactions within site are covered below
      if((*j).model_struct.contains(l->first)){continue;}
      // get LJ parameters
      CharmmParaNB nbl;
      // background could contain unknown atom types
      if (chmpara.contains_attype(CharmmID(l->second))){
       if (chmpara.contains_NB(CharmmID(l->second))){
        nbl = chmpara.get_NB(CharmmID(l->second));
       }
      }
      else if (l->second.atname == "OT1" || l->second.atname == "OT2"){
       nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("O")));
      }
      else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3"){
       nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("HN")));
      }
      // sloppy search for matching atom name in patch residues
      else if (chmpara.contains_pres_attype(l->second.atname)){
       nbk = chmpara.get_NB(chmpara.get_pres_attype(l->second.atname));
      }
      else if (chmpara.contains_element_NB(top.get_element(l->second.atname))){
       nbk = chmpara.get_element_NB(top.get_element(l->second.atname));
      }
      else {
       cerr << l->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Lennard Jones parameter missing.");
      }
      float dist = abs_value(k->second.coord - l->second.coord);
      float ljek = nbk.lje;
      float ljrk = nbk.ljr;
      float ljel = nbl.lje;
      float ljrl = nbl.ljr;
      if (ljek <= 0 && ljel <= 0){
       if ( dist != 0 ){
        float r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
        gmmback += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
       } else {
        cerr << k->second << endl;
        cerr << l->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: two atoms have identical coordinates");
       }
      }
      else {
       if (ljek > 0 || ljel > 0){
         ::error ("ERROR in QmpbMolecule::calc_mmint_propensities atom has non-negative VDW well depth.");
       }
      }
     }
    }

    // calculate internal terms

    // first non-bonded terms

    for (AtomSet::const_iterator k=(*j).model_struct.begin(); k!=(*j).model_struct.end(); ++k){
     if (!chmpara.contains_attype(CharmmID(k->second)) && !chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: attype not found.");       
     }
     AtomSet::const_iterator kplus = k; kplus++;
     for (AtomSet::const_iterator l=kplus; l!=(*j).model_struct.end(); ++l){
      if (!chmpara.contains_attype(CharmmID(l->second)) && !chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: attype not found.");       
      }
      // skip term if no ligand atom is involved
      if (!in_list(k->second.atname,ligats) && !in_list(l->second.atname,ligats)) {continue;}
 
      // explore bonding situation
      bool i12 = top.in_12(k->second.atname,l->second.atname);
      bool i13 = top.in_13(k->second.atname,l->second.atname);
      bool i14 = top.in_14(k->second.atname,l->second.atname);
      if (!i12 && !i13){
       CharmmParaNB nbk, nbl;
       if (chmpara.contains_NB(CharmmID(k->second))){
        nbk = chmpara.get_NB(CharmmID(k->second));
       }
       else if (k->second.atname == "OT1" || k->second.atname == "OT2"){
        nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("O")));
       }
       else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3"){
        nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("HN")));
       }
       else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
        nbl = chmpara.get_NB(std::string("H"));
       }       
       else {
        cerr << k->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Lennard Jones parameter missing.");
       }
       if (chmpara.contains_NB(CharmmID(l->second))){
        nbl = chmpara.get_NB(CharmmID(l->second));
       }
       else if (l->second.atname == "OT1" || l->second.atname == "OT2"){
        nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("O")));
       }
       else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3"){
        nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("HN")));
       }
       else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
        nbl = chmpara.get_NB(std::string("H"));
       }       
       else {
        cerr << l->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Lennard Jones parameter missing.");
       }
       float ljek, ljrk, ljel, ljrl;
       if (i14){
        ljek = nbk.lje14;
        ljrk = nbk.ljr14;
        ljel = nbl.lje14;
        ljrl = nbl.ljr14;
       } else {
        ljek = nbk.lje;
        ljrk = nbk.ljr;
        ljel = nbl.lje;
        ljrl = nbl.ljr;
       }
       float tmpf1 = 0;
       float tmpf2 = 0;
       float dist = abs_value(k->second.coord - l->second.coord);
       if (ljek <= 0 && ljel <= 0){
        if ( dist != 0 ){
         float r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
         tmpf1 = sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
        } else {
         cerr << k->second << endl;
         cerr << l->second << endl;
         ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: two atoms have identical coordinates");
        }
       }
       else {
        if (ljek > 0 || ljel > 0){
          ::error ("ERROR in QmpbMolecule::calc_mmint_propensities atom has non-negative VDW well depth.");
        }
       }
       // coulomb electrostatics
       if (!i14)
        tmpf2 = PhysCond::get_econv() * (k->second.charge * l->second.charge) / (dist * eps_ff);

       if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first)){
        mgmml_lj += tmpf1;
        mgmml_e  += tmpf2;
       }
       else {
        gmml_lj += tmpf1;
        gmml_e  += tmpf2;
       }

      }
     }
    }

    // bond terms

    for (std::map<std::pair<std::string,std::string>,int>::const_iterator ii = top.bonds.begin(); ii != top.bonds.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!in_list(ii->first.first,ligats) && !in_list(ii->first.second,ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l;

     if (((ii->first).first).find("+") != 0 && ((ii->first).first).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.first,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((ii->first).first,ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.first,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && ii->first.first == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && ii->first.first == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((ii->first).first,ligats)){
       cerr << "non-ligand atom named " << (ii->first).first << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (((ii->first).second).find("+") != 0 && ((ii->first).second).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.second,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((ii->first).second,ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.second,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && ii->first.second == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && ii->first.second == "HN"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((ii->first).second,ligats)){
       cerr << "non-ligand atom named " << (ii->first).second << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy

     CharmmParaBond tmppar;
     if (chmpara.contains_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second))){
      tmppar = chmpara.get_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second));
     }
     else if (chmpara.contains_Bond(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)))){
      tmppar = chmpara.get_Bond(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)));
     }
     else if (chmpara.contains_Bond(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname))){
      tmppar = chmpara.get_Bond(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Bond parameter missing.");
     }

     float dist = abs_value(k->second.coord - l->second.coord);
     float tmpf = tmppar.k * (dist - tmppar.r) * (dist - tmppar.r);

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first)){
      mgmml_bond += tmpf;
     }
     else {
      gmml_bond += tmpf;
     }

    }

    // angle terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.angles.begin(); ii != top.angles.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!in_list((*ii)[0],ligats) && !in_list((*ii)[1],ligats) && !in_list((*ii)[2],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){

      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "HN"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy

     CharmmParaAngle tmppar;

     if (chmpara.contains_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second))){
      tmppar = chmpara.get_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second));
     }
     else if (chmpara.contains_Angle(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)))){
      tmppar = chmpara.get_Angle(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)));
     }
     else if (chmpara.contains_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)))){
      tmppar = chmpara.get_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)));
     }
     else if (chmpara.contains_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname))){
      tmppar = chmpara.get_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl << m->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Angle parameter missing.");
     }

     float angle  = get_angle(k->second.coord ,l->second.coord, m->second.coord);
     float phirad = tmppar.phi * (pi / 180);
     if (phirad < 0){phirad = -1 * phirad;}
     float tmpf = tmppar.kphi * (angle - phirad) * (angle - phirad);
     // Urey-Bradley term
     if (tmppar.kub != 0){
      float dist = abs_value(k->second.coord - m->second.coord);
      tmpf += tmppar.kub * (dist - tmppar.rub) * (dist - tmppar.rub);
     }

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first)){
       mgmml_angle += tmpf;
     }
     else {
      gmml_angle += tmpf;
     }

    }

    // dihedral terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.dihedrals.begin(); ii != top.dihedrals.end(); ++ii){

     // skip term if no ligand atom is involved
     if (!in_list((*ii)[0],ligats) && !in_list((*ii)[1],ligats) && !in_list((*ii)[2],ligats) && !in_list((*ii)[3],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (n == (*j).model_struct.end() && in_list((*ii)[3],ligats)) {continue;}
      if (n == (*j).model_struct.end()){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "O"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "HN"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && !in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy
     std::list<CharmmParaDihedral> tmppar;

     if (chmpara.contains_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = chmpara.get_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Dihedral parameter missing.");
     }

     float dihe = get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     float tmpf = 0;
     for (std::list<CharmmParaDihedral>::const_iterator jj=tmppar.begin(); jj != tmppar.end(); ++jj){
      tmpf += (*jj).k * (1 - cos(((*jj).n * dihe) - ((*jj).delta * (pi / 180))));
     }

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first) || !(*j).site_struct.contains(n->first)){
      mgmml_dihedral += tmpf;
     }
     else {
      gmml_dihedral += tmpf;
     }

    }

    // improper terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.impropers.begin(); ii != top.impropers.end(); ++ii){
     
     // skip term if no ligand atom is involved
     if (!in_list((*ii)[0],ligats) && !in_list((*ii)[1],ligats) && !in_list((*ii)[2],ligats) && !in_list((*ii)[3],ligats)) {continue;}

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "HN"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "HN"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (n == (*j).model_struct.end() && in_list((*ii)[3],ligats)) {continue;}
      if (n == (*j).model_struct.end()){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "O"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "HN"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && !in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint_propensities IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // retrieve parameters and calculate energy
     CharmmParaImproper tmppar;

     if (chmpara.contains_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = chmpara.get_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (chmpara.contains_Improper(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint_propensities: Improper parameter missing.");
     }

     float imp = get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     float psirad = tmppar.psi * (pi / 180);
     float tmpf = tmppar.kpsi * (imp - psirad) * (imp - psirad);

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first) || !(*j).site_struct.contains(n->first)){
      mgmml_improper += tmpf;
     }
     else {
      gmml_improper += tmpf;
     }
    }

    gmml = gmml_e + gmml_lj + gmml_bond + gmml_angle + gmml_dihedral + gmml_improper;
    mgmml = mgmml_e + mgmml_lj + mgmml_bond + mgmml_angle + mgmml_dihedral + mgmml_improper;
 
    if (print_chm){
    
     cout << endl << "---------------------------------------" << endl;

     cout << "Energies for instance " << ((*j).get_instid()).get_fn_prefix() << endl;

     cout << "gmmback "       << gmmback << endl;

     cout << "gmml "          << gmml << endl;
     cout << "gmml_e "        << gmml_e << endl;
     cout << "gmml_lj "       << gmml_lj << endl;
     cout << "gmml_bond "     << gmml_bond << endl;
     cout << "gmml_angle "    << gmml_angle << endl;
     cout << "gmml_dihedral " << gmml_dihedral << endl;
     cout << "gmml_improper " << gmml_improper << endl;

     cout << "mgmml "          << mgmml << endl;
     cout << "mgmml_e "        << mgmml_e << endl;
     cout << "mgmml_lj "       << mgmml_lj << endl;
     cout << "mgmml_bond "     << mgmml_bond << endl;
     cout << "mgmml_angle "    << mgmml_angle << endl;
     cout << "mgmml_dihedral " << mgmml_dihedral << endl;
     cout << "mgmml_improper " << mgmml_improper << endl;

     cout << "gprop "          << gprop << endl;

     cout << endl << "---------------------------------------" << endl;
    
    }

    std::map<InstanceID, std::map<std::string, float> >::iterator em_iter = emap.find((*j).get_instid());
    // should never happen
    if (em_iter != emap.end()) {
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities non-unique InstanceID\n");
    }

    std::map<std::string,float> tmpmap;
    tmpmap.insert(std::pair<std::string,float>("gmmback",gmmback));
    std::pair<InstanceID,std::map<std::string, float> > tmppair1((*j).get_instid(),tmpmap);

    std::pair<std::map<InstanceID, std::map<std::string, float> >::iterator, bool> tmpair2 = emap.insert(tmppair1);

    em_iter = tmpair2.first;

    em_iter->second.insert(std::pair<std::string,float>("gmml_e",gmml_e));
    em_iter->second.insert(std::pair<std::string,float>("gmml_lj",gmml_lj));
    em_iter->second.insert(std::pair<std::string,float>("gmml_bond",gmml_bond));
    em_iter->second.insert(std::pair<std::string,float>("gmml_angle",gmml_angle));
    em_iter->second.insert(std::pair<std::string,float>("gmml_dihedral",gmml_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("gmml_improper",gmml_improper));

    em_iter->second.insert(std::pair<std::string,float>("mgmml_e",mgmml_e));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_lj",mgmml_lj));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_bond",mgmml_bond));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_angle",mgmml_angle));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_dihedral",mgmml_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_improper",mgmml_improper));

    em_iter->second.insert(std::pair<std::string,float>("gprop",gprop));

   }

   // search for minimum ligand-site/model energy instance of each form of the site

   std::map<std::string,float> min_gmml;

   for (std::map<InstanceID,std::map<std::string, float> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    float tmpf = 0;
    std::map<std::string, float>::const_iterator titr;
    std::string formname(eitr->first.get_instname());

    // ligand-site terms

    if (use_term_l_e){
      titr = eitr->second.find(std::string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_lj){
      titr = eitr->second.find(std::string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_bond){
      titr = eitr->second.find(std::string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_angle){
      titr = eitr->second.find(std::string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_dihedral){
      titr = eitr->second.find(std::string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_improper){
      titr = eitr->second.find(std::string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      tmpf += titr->second;
    }

    // ligand-model terms

    if (use_term_ml_e){
      titr = eitr->second.find(std::string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_lj){
      titr = eitr->second.find(std::string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_bond){
      titr = eitr->second.find(std::string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_angle){
      titr = eitr->second.find(std::string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_dihedral){
      titr = eitr->second.find(std::string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_improper){
      titr = eitr->second.find(std::string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }
     
    // create or update minimum energy entry for this form of the site

    std::map<std::string,float>::iterator meitr(min_gmml.find(formname));
    if (meitr == min_gmml.end()){
     std::pair<std::string,float> tmppair(formname,tmpf);
     min_gmml.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }
   }

   // now collect all energy terms in the final terms

   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){

    float gmmint = 0;
    float gmmback = 0;
    float gmmsite = 0;
    float gprop = 0;
    float gmmmodelback = 0;

    std::map<InstanceID,std::map<std::string, float> >::const_iterator eitr(emap.find((*j).get_instid()));
    if (eitr == emap.end())
     ::error("ERROR in QmpbMolecule::calc_mmint_propensities instance missing in energy term map\n");
    std::map<std::string, float>::const_iterator titr(eitr->second.find(std::string("gmmback")));
    if (titr== eitr->second.end())
     ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");

    gmmback = titr->second;
    
    titr = eitr->second.find(std::string("gprop"));
    if (titr== eitr->second.end())
     ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
    else 
     gprop = titr->second; 
     
    float min_l = 0;

    std::map<std::string, float>::const_iterator ml_itr(min_gmml.find((*j).get_name()));
    if (ml_itr != min_gmml.end()) min_l = ml_itr->second;

    float gmml = 0;

    // ligand-site terms

    if (use_term_l_e){
      titr = eitr->second.find(std::string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_lj){
      titr = eitr->second.find(std::string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_bond){
      titr = eitr->second.find(std::string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_angle){
      titr = eitr->second.find(std::string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_dihedral){
      titr = eitr->second.find(std::string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_improper){
      titr = eitr->second.find(std::string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      gmml += titr->second;
    }

    float mgmml = 0;

    // ligand-modelback terms

    if (use_term_ml_e){
      titr = eitr->second.find(std::string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_lj){
      titr = eitr->second.find(std::string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_bond){
      titr = eitr->second.find(std::string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_angle){
      titr = eitr->second.find(std::string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_dihedral){
      titr = eitr->second.find(std::string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_improper){
      titr = eitr->second.find(std::string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint_propensities energy term missing\n");
     else
      mgmml += titr->second;
    }
     
    // minimum ligand-site interaction should be included in model energy
    // correct for this by considering only relative ligand site interactions within 1 form
    gmmsite      = gprop  + gmml - min_l;
    // interactions of non-ligand site atoms with modelback are assumed to be included in gprop
    gmmmodelback = mgmml;

    (*j).set_gmmsite(gmmsite);
    (*j).set_gmmmodelback(gmmmodelback);
    (*j).set_gmmback(gmmback);
 
   }



  }


  blab1 << "Exiting QmpbMolecule::calc_mmint_propensities" << endl;

}

void QmpbMolecule::calc_mmint_charmm()
{

  blab2 << "Entering QmpbMolecule::calc_mmint_charmm" << endl;

  for (std::list<QmpbSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
   int stype(get_rotmap_value((*i).get_sid()));
   if (stype != 1 && stype != 2){
   //if (1 == 2){
    blab1 << "QmpbMolecule::calc_mmint_charmm: site with name "
          << (*i).get_name() << " and ID "
          << (*i).get_sid()  << " is not flexible or flexible and titratable.\n"
          << "Will not calculate intrinsic MM energies for this site" << endl;
    continue;
   }
   if (!chmpara.contains_Topology((*i).get_name())){
    cerr << "QmpbMolecule::calc_mmint_charmm Warning: could not find topology entry in CHARMM topology for site name "
         << (*i).get_name() << ".\nWill not calculate intrinsic MM energy for this residue."
         << (*i).get_sid()  << endl;
    continue;
   }
   // retrieve topology
   CharmmParaResTop top(chmpara.get_Topology((*i).get_name()));
   std::list<AtomID> aidlist;
   // retrieve list of atoms which have zero occupancy in any form (ligand atoms)
   std::list<std::string> ligats;
   if (stype == 1 || stype == 3){
    // this is a titratable and flexible site
    // collect all ligand atoms
    QmpbSiteDef sdef(get_sitedef((*i).get_name()));
    for (std::list<QmpbFormDef>::const_iterator k=sdef.forms.begin(); k!=sdef.forms.end(); ++k){
     for (std::map<CharmmID,pair<float, int> >::const_iterator l = (*k).atommap.begin(); l!=(*k).atommap.end();++l){
      // search atom with zero occupancy
      if (l->second.second == 0){ligats.push_back(l->first.get_atname());}
     }
    }
    ligats.sort();
    ligats.unique();
   }

   std::map<InstanceID, std::map<std::string, float> > emap;
   std::map<std::string, float> min_e_map;

   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){

    // all terms involving ligand + site atoms and background - model atoms
    float gmmback = 0;

    // all terms involving ligand atoms and only site atoms
    float gmml = 0;
    float gmml_e = 0;
    float gmml_lj = 0;
    float gmml_bond = 0;
    float gmml_angle = 0;
    float gmml_dihedral = 0;
    float gmml_improper = 0;
    // all terms not involving ligand atoms and only site atoms
    float gmmr = 0;
    float gmmr_e = 0;
    float gmmr_lj = 0;
    float gmmr_bond = 0;
    float gmmr_angle = 0;
    float gmmr_dihedral = 0;
    float gmmr_improper = 0;

    // all terms involving ligand atoms and also model atoms
    float mgmml = 0;
    float mgmml_e = 0;
    float mgmml_lj = 0;
    float mgmml_bond = 0;
    float mgmml_angle = 0;
    float mgmml_dihedral = 0;
    float mgmml_improper = 0;
    // all terms not involving ligand atoms but model_atoms
    float mgmmr = 0;
    float mgmmr_e = 0;
    float mgmmr_lj = 0;
    float mgmmr_bond = 0;
    float mgmmr_angle = 0;
    float mgmmr_dihedral = 0;
    float mgmmr_improper = 0;

    // calculate gmm_back

    for (AtomSet::const_iterator k=(*j).site_struct.begin(); k!=(*j).site_struct.end(); ++k){
     if (!chmpara.contains_attype(CharmmID(k->second)) && !chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint: attype not found.");
     }
     CharmmParaNB nbk;
     if (chmpara.contains_attype(CharmmID(k->second))){
      if (chmpara.contains_NB(CharmmID(k->second))){
       nbk = chmpara.get_NB(CharmmID(k->second));
      }
     }
     else if (k->second.atname == "OT1" || k->second.atname == "OT2"){
      nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("O")));
     }
     else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3"){
      nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("HN")));
     }
     else {
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint: Lennard Jones parameter missing.");
     }     
     for (AtomSet::const_iterator l=background.begin(); l!=background.end(); ++l){
      if (!chmpara.contains_attype(CharmmID(l->second)) && !chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint: attype not found.");
      }
      // skip model atoms, interactions within site are covered below
      if((*j).model_struct.contains(l->first)){continue;}
      // get LJ parameters
      CharmmParaNB nbl;
      // background could contain unknown atom types
      if (chmpara.contains_attype(CharmmID(l->second))){
       if (chmpara.contains_NB(CharmmID(l->second))){
        nbl = chmpara.get_NB(CharmmID(l->second));
       }
      }
      else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3"){
       nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("HN")));
      }
      else if (l->second.atname == "OT1"){
       nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("O")));
      }
      // sloppy search for matching atom name in patch residues
      else if (chmpara.contains_pres_attype(l->second.atname)){
       nbk = chmpara.get_NB(chmpara.get_pres_attype(l->second.atname));
      }
      else if (chmpara.contains_element_NB(top.get_element(l->second.atname))){
       nbk = chmpara.get_element_NB(top.get_element(l->second.atname));
      }
      else {
       cerr << l->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint: Lennard Jones parameter missing.");
      }
      float dist = abs_value(k->second.coord - l->second.coord);
      float ljek = nbk.lje;
      float ljrk = nbk.ljr;
      float ljel = nbl.lje;
      float ljrl = nbl.ljr;
      if (ljek <= 0 && ljel <= 0){
       if ( dist != 0 ){
        float r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
        gmmback += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
       } else {
        cerr << k->second << endl;
        cerr << l->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint: two atoms have identical coordinates");
       }
      }
      else {
       if (ljek > 0 || ljel > 0){
         ::error ("ERROR in QmpbMolecule::calc_mmint atom has non-negative VDW well depth.");
       }
      }
     }
    }

    // calculate internal terms

    // first non-bonded terms

    for (AtomSet::const_iterator k=(*j).model_struct.begin(); k!=(*j).model_struct.end(); ++k){
     if (!chmpara.contains_attype(CharmmID(k->second)) && !chmpara.contains_pres_attype(k->second.atname)){
      cerr << k->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint: attype not found.");
     }
     AtomSet::const_iterator kplus = k; kplus++;
     for (AtomSet::const_iterator l=kplus; l!=(*j).model_struct.end(); ++l){
      if (!chmpara.contains_attype(CharmmID(l->second)) && !chmpara.contains_pres_attype(l->second.atname)){
       cerr << l->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint: attype not found.");
      }
      if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first)) {continue;}
      // explore bonding situation
      bool i12 = top.in_12(k->second.atname,l->second.atname);
      bool i13 = top.in_13(k->second.atname,l->second.atname);
      bool i14 = top.in_14(k->second.atname,l->second.atname);
      if (!i12 && !i13){
       CharmmParaNB nbk, nbl;
       if (chmpara.contains_attype(CharmmID(k->second))){
        if (chmpara.contains_NB(CharmmID(k->second))){
         nbk = chmpara.get_NB(CharmmID(k->second));
        }
       }
       else if (k->second.atname == "OT1" || k->second.atname == "OT2"){
        nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("O")));
       }
       else if (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3"){
        nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("HN")));
       }
       else if (k->second.atname == "HT11" || k->second.atname == "HT12" || k->second.atname == "HT21" || k->second.atname == "HT22"){
        nbk = chmpara.get_NB(std::string("H"));
       }       
       else {
        cerr << k->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint: Lennard Jones parameter missing.");
       }
       if (chmpara.contains_attype(CharmmID(l->second))){
        if (chmpara.contains_NB(CharmmID(l->second))){
         nbl = chmpara.get_NB(CharmmID(l->second));
        }
       }
       else if (l->second.atname == "OT1" || l->second.atname == "OT2"){
        nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("O")));
       }
       else if (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3"){
        nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("HN")));
       }
       else if (l->second.atname == "HT11" || l->second.atname == "HT12" || l->second.atname == "HT21" || l->second.atname == "HT22"){
        nbl = chmpara.get_NB(std::string("H"));
       }       
       else {
        cerr << l->second << endl;
        ::error ("ERROR in QmpbMolecule::calc_mmint: Lennard Jones parameter missing.");
       }
       float ljek, ljrk, ljel, ljrl;
       if (i14){
        ljek = nbk.lje14;
        ljrk = nbk.ljr14;
        ljel = nbl.lje14;
        ljrl = nbl.ljr14;
       } else {
        ljek = nbk.lje;
        ljrk = nbk.ljr;
        ljel = nbl.lje;
        ljrl = nbl.ljr;
       }
       float tmpf1 = 0;
       float tmpf2 = 0;
       float dist = abs_value(k->second.coord - l->second.coord);
       if (ljek <= 0 && ljel <= 0){
        if ( dist != 0 ){
         float r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
         tmpf1 = sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
        } else {
         cerr << k->second << endl;
         cerr << l->second << endl;
         ::error ("ERROR in QmpbMolecule::calc_mmint: two atoms have identical coordinates");
        }
       }
       else {
        if (ljek > 0 || ljel > 0){
          ::error ("ERROR in QmpbMolecule::calc_mmint atom has non-negative VDW well depth.");
        }
       }
       // coulomb electrostatics
       if (!i14)
        tmpf2 = PhysCond::get_econv() * (k->second.charge * l->second.charge) / (dist * eps_ff);

       if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first)){
        if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats)){
         mgmml_lj += tmpf1;
         mgmml_e  += tmpf2;
        } else {
         if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first)){continue;}
         mgmmr_lj += tmpf1;
         mgmmr_e  += tmpf2;
        }
       }
       else {
        if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats)){
         gmml_lj += tmpf1;
         gmml_e  += tmpf2;
        } else {
         gmmr_lj += tmpf1;
         gmmr_e  += tmpf2;
        }
       }

      }
     }
    }

    // bond terms

    for (std::map<std::pair<std::string,std::string>,int>::const_iterator ii = top.bonds.begin(); ii != top.bonds.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l;

     if (((ii->first).first).find("+") != 0 && ((ii->first).first).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.first,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((ii->first).first,ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.first,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && ii->first.first == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && ii->first.first == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((ii->first).first,ligats)){
       cerr << "non-ligand atom named " << (ii->first).first << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if (((ii->first).second).find("+") != 0 && ((ii->first).second).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.second,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((ii->first).second,ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), ii->first.second,
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && ii->first.second == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && ii->first.second == "HN"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((ii->first).second,ligats)){
       cerr << "non-ligand atom named " << (ii->first).second << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint BOND non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first)) {continue;}

     // retrieve parameters and calculate energy

     CharmmParaBond tmppar;
     if (chmpara.contains_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second))){
      tmppar = chmpara.get_Bond(CharmmID(k->second.resname, ii->first.first), CharmmID(l->second.resname, ii->first.second));
     }
     else if (chmpara.contains_Bond(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)))){
      tmppar = chmpara.get_Bond(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)));
     }
     else if (chmpara.contains_Bond(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname))){
      tmppar = chmpara.get_Bond(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint: Bond parameter missing.");
     }

     float dist = abs_value(k->second.coord - l->second.coord);
     float tmpf = tmppar.k * (dist - tmppar.r) * (dist - tmppar.r);

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first)){
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats)){
       mgmml_bond += tmpf;
      } else {
       if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first)){continue;}
       mgmmr_bond += tmpf;
      }
     }
     else {
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats)){
       gmml_bond += tmpf;
      } else {
       gmmr_bond += tmpf;
      }
     }

    }

    // angle terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.angles.begin(); ii != top.angles.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){

      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "H"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "HN"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint ANGLE non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first)) {continue;}

     // retrieve parameters and calculate energy

     CharmmParaAngle tmppar;

     if (chmpara.contains_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second))){
      tmppar = chmpara.get_Angle(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second));
     }
     else if (chmpara.contains_Angle(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)))){
      tmppar = chmpara.get_Angle(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)));
     }
     else if (chmpara.contains_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)))){
      tmppar = chmpara.get_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)));
     }
     else if (chmpara.contains_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname))){
      tmppar = chmpara.get_Angle(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname));
     }
     else {
      cerr << k->second << endl << l->second << endl << m->second << endl;
      ::error ("ERROR in QmpbMolecule::calc_mmint: Angle parameter missing.");
     }

     float angle  = get_angle(k->second.coord ,l->second.coord, m->second.coord);
     float phirad = tmppar.phi * (pi / 180);
     if (phirad < 0){phirad = -1 * phirad;}
     float tmpf = tmppar.kphi * (angle - phirad) * (angle - phirad);
     // Urey-Bradley term
     if (tmppar.kub != 0){
      float dist = abs_value(k->second.coord - m->second.coord);
      tmpf += tmppar.kub * (dist - tmppar.rub) * (dist - tmppar.rub);
     }

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first)){
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) || in_list(m->second.atname, ligats)){
       mgmml_angle += tmpf;
      } else {
       if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first)){continue;}
       mgmmr_angle += tmpf;
      }
     }
     else {
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) || in_list(m->second.atname, ligats)){
       gmml_angle += tmpf;
      } else {
       gmmr_angle += tmpf;
      }
     }

    }

    // dihedral terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.dihedrals.begin(); ii != top.dihedrals.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (n == (*j).model_struct.end() && in_list((*ii)[3],ligats)) {continue;}
      if (n == (*j).model_struct.end()){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "O"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "HN"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && !in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint DIHEDRAL non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first) && !(*j).site_struct.contains(n->first)) {continue;}

     // retrieve parameters and calculate energy
     std::list<CharmmParaDihedral> tmppar;

     if (chmpara.contains_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = chmpara.get_Dihedral(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname))){
      tmppar = chmpara.get_Dihedral(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint: Dihedral parameter missing.");
     }

     float dihe = get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     float tmpf = 0;
     for (std::list<CharmmParaDihedral>::const_iterator jj=tmppar.begin(); jj != tmppar.end(); ++jj){
      tmpf += (*jj).k * (1 - cos(((*jj).n * dihe) - ((*jj).delta * (pi / 180))));
     }

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first) || !(*j).site_struct.contains(n->first)){
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) ||
          in_list(m->second.atname, ligats) || in_list(n->second.atname, ligats)){
       mgmml_dihedral += tmpf;
      } else {
       if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first) && !(*j).site_struct.contains(n->first)){continue;}
       mgmmr_dihedral += tmpf;
      }
     }
     else {
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) ||
          in_list(m->second.atname, ligats) || in_list(n->second.atname, ligats)){
       gmml_dihedral += tmpf;
      } else {
       gmmr_dihedral += tmpf;
      }
     }

    }

    // improper terms

    for (std::list<std::vector<std::string> >::const_iterator ii = top.impropers.begin(); ii != top.impropers.end(); ++ii){

     // find atoms

     AtomSet::const_iterator k,l,m,n;

     if ( ((*ii)[0]).find("+") != 0 && ((*ii)[0]).find("-") != 0){
      k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (k == (*j).model_struct.end() && in_list((*ii)[0],ligats)) {continue;}
      if (k == (*j).model_struct.end()){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[0],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "O"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && (*ii)[0] == "HN"){
       k = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (k == (*j).model_struct.end() && !in_list((*ii)[0],ligats)){
       cerr << "non-ligand atom named " << (*ii)[0] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[1]).find("+") != 0 && ((*ii)[1]).find("-") != 0){
      l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (l == (*j).model_struct.end() && in_list((*ii)[1],ligats)) {continue;}
      if (l == (*j).model_struct.end()){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[1],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "O"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && (*ii)[1] == "HN"){
       l = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (l == (*j).model_struct.end() && !in_list((*ii)[1],ligats)){
       cerr << "non-ligand atom named " << (*ii)[1] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[2]).find("+") != 0 && ((*ii)[2]).find("-") != 0){
      m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (m == (*j).model_struct.end() && in_list((*ii)[2],ligats)) {continue;}
      if (m == (*j).model_struct.end()){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[2],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "O"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && (*ii)[2] == "HN"){
       m = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (m == (*j).model_struct.end() && !in_list((*ii)[2],ligats)){
       cerr << "non-ligand atom named " << (*ii)[2] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     if ( ((*ii)[3]).find("+") != 0 && ((*ii)[3]).find("-") != 0){
      n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), ((*j).get_instid()).get_instance()));
      if (n == (*j).model_struct.end() && in_list((*ii)[3],ligats)) {continue;}
      if (n == (*j).model_struct.end()){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), (*ii)[3],
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(), int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "O"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "OT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && (*ii)[3] == "HN"){
       n = (*j).model_struct.find(
          AtomID(((*j).get_instid()).get_site(), "HT1",
                 ((*j).get_instid()).get_conformer(), ((*j).get_instid()).get_chain(), ((*j).get_instid()).get_site(),
                 int(0)));
      }
      if (n == (*j).model_struct.end() && !in_list((*ii)[3],ligats)){
       cerr << "non-ligand atom named " << (*ii)[3] << " is missing in site " << (*i).get_sid() << endl;
       ::error("ERROR in QmpbMolecule::calc_mmint IMPROPER non-ligand atom is missing in site\n");
      }
     } else {continue;}

     // skip term if no site atom is involved
     if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first) && !(*j).site_struct.contains(n->first)) {continue;}

     // retrieve parameters and calculate energy
     CharmmParaImproper tmppar;

     if (chmpara.contains_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second))){
      tmppar = chmpara.get_Improper(CharmmID(k->second), CharmmID(l->second), CharmmID(m->second), CharmmID(n->second));
     }
     else if (chmpara.contains_Improper(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_pres_attype(k->second.atname), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_pres_attype(l->second.atname), chmpara.get_attype(CharmmID(m->second)), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_pres_attype(m->second.atname), chmpara.get_attype(CharmmID(n->second)));
     }
     else if (chmpara.contains_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname))){
      tmppar = chmpara.get_Improper(chmpara.get_attype(CharmmID(k->second)), chmpara.get_attype(CharmmID(l->second)), chmpara.get_attype(CharmmID(m->second)), chmpara.get_pres_attype(n->second.atname));
     }
     else {
       cerr << k->second << endl << l->second << endl << m->second << endl << n->second << endl;
       ::error ("ERROR in QmpbMolecule::calc_mmint: Improper parameter missing.");
     }

     float imp = get_tor(k->second.coord ,l->second.coord, m->second.coord, n->second.coord);
     float psirad = tmppar.psi * (pi / 180);
     float tmpf = tmppar.kpsi * (imp - psirad) * (imp - psirad);

     if (!(*j).site_struct.contains(k->first) || !(*j).site_struct.contains(l->first) || !(*j).site_struct.contains(m->first) || !(*j).site_struct.contains(n->first)){
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) ||
          in_list(m->second.atname, ligats) || in_list(n->second.atname, ligats)){
       mgmml_improper += tmpf;
      } else {
       if (!(*j).site_struct.contains(k->first) && !(*j).site_struct.contains(l->first) && !(*j).site_struct.contains(m->first) && !(*j).site_struct.contains(n->first)){continue;}
       mgmmr_improper += tmpf;
      }
     }
     else {
      if (in_list(k->second.atname, ligats) || in_list(l->second.atname, ligats) ||
          in_list(m->second.atname, ligats) || in_list(n->second.atname, ligats)){
       gmml_improper += tmpf;
      } else {
       gmmr_improper += tmpf;
      }
     }
    }

    gmml = gmml_e + gmml_lj + gmml_bond + gmml_angle + gmml_dihedral + gmml_improper;
    gmmr = gmmr_e + gmmr_lj + gmmr_bond + gmmr_angle + gmmr_dihedral + gmmr_improper;
    mgmml = mgmml_e + mgmml_lj + mgmml_bond + mgmml_angle + mgmml_dihedral + mgmml_improper;
    mgmmr = mgmmr_e + mgmmr_lj + mgmmr_bond + mgmmr_angle + mgmmr_dihedral + mgmmr_improper;

    if (print_chm){
    
     cout << endl << "---------------------------------------" << endl;

     cout << "Energies for instance " << ((*j).get_instid()).get_fn_prefix() << endl;

     cout << "gmmback "       << gmmback << endl;

     cout << "gmml "          << gmml << endl;
     cout << "gmml_e "        << gmml_e << endl;
     cout << "gmml_lj "       << gmml_lj << endl;
     cout << "gmml_bond "     << gmml_bond << endl;
     cout << "gmml_angle "    << gmml_angle << endl;
     cout << "gmml_dihedral " << gmml_dihedral << endl;
     cout << "gmml_improper " << gmml_improper << endl;

     cout << "gmmr "          << gmmr << endl;
     cout << "gmmr_e "        << gmmr_e << endl;
     cout << "gmmr_lj "       << gmmr_lj << endl;
     cout << "gmmr_bond "     << gmmr_bond << endl;
     cout << "gmmr_angle "    << gmmr_angle << endl;
     cout << "gmmr_dihedral " << gmmr_dihedral << endl;
     cout << "gmmr_improper " << gmmr_improper << endl;

     cout << "mgmml "          << mgmml << endl;
     cout << "mgmml_e "        << mgmml_e << endl;
     cout << "mgmml_lj "       << mgmml_lj << endl;
     cout << "mgmml_bond "     << mgmml_bond << endl;
     cout << "mgmml_angle "    << mgmml_angle << endl;
     cout << "mgmml_dihedral " << mgmml_dihedral << endl;
     cout << "mgmml_improper " << mgmml_improper << endl;

     cout << "mgmmr "          << mgmmr << endl;
     cout << "mgmmr_e "        << mgmmr_e << endl;
     cout << "mgmmr_lj "       << mgmmr_lj << endl;
     cout << "mgmmr_bond "     << mgmmr_bond << endl;
     cout << "mgmmr_angle "    << mgmmr_angle << endl;
     cout << "mgmmr_dihedral " << mgmmr_dihedral << endl;
     cout << "mgmmr_improper " << mgmmr_improper << endl;

     cout << endl << "---------------------------------------" << endl;
  
    }

    std::map<InstanceID, std::map<std::string, float> >::iterator em_iter = emap.find((*j).get_instid());
    // should never happen
    if (em_iter != emap.end()) {
      ::error("ERROR in QmpbMolecule::calc_mmint non-unique InstanceID\n");
    }

    std::map<std::string,float> tmpmap;
    tmpmap.insert(std::pair<std::string,float>("gmmback",gmmback));
    std::pair<InstanceID,std::map<std::string, float> > tmppair1((*j).get_instid(),tmpmap);

    std::pair<std::map<InstanceID, std::map<std::string, float> >::iterator, bool> tmpair2 = emap.insert(tmppair1);

    em_iter = tmpair2.first;

    em_iter->second.insert(std::pair<std::string,float>("gmml_e",gmml_e));
    em_iter->second.insert(std::pair<std::string,float>("gmml_lj",gmml_lj));
    em_iter->second.insert(std::pair<std::string,float>("gmml_bond",gmml_bond));
    em_iter->second.insert(std::pair<std::string,float>("gmml_angle",gmml_angle));
    em_iter->second.insert(std::pair<std::string,float>("gmml_dihedral",gmml_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("gmml_improper",gmml_improper));

    em_iter->second.insert(std::pair<std::string,float>("gmmr_e",gmmr_e));
    em_iter->second.insert(std::pair<std::string,float>("gmmr_lj",gmmr_lj));
    em_iter->second.insert(std::pair<std::string,float>("gmmr_bond",gmmr_bond));
    em_iter->second.insert(std::pair<std::string,float>("gmmr_angle",gmmr_angle));
    em_iter->second.insert(std::pair<std::string,float>("gmmr_dihedral",gmmr_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("gmmr_improper",gmmr_improper));

    em_iter->second.insert(std::pair<std::string,float>("mgmml_e",mgmml_e));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_lj",mgmml_lj));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_bond",mgmml_bond));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_angle",mgmml_angle));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_dihedral",mgmml_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("mgmml_improper",mgmml_improper));

    em_iter->second.insert(std::pair<std::string,float>("mgmmr_e",mgmmr_e));
    em_iter->second.insert(std::pair<std::string,float>("mgmmr_lj",mgmmr_lj));
    em_iter->second.insert(std::pair<std::string,float>("mgmmr_bond",mgmmr_bond));
    em_iter->second.insert(std::pair<std::string,float>("mgmmr_angle",mgmmr_angle));
    em_iter->second.insert(std::pair<std::string,float>("mgmmr_dihedral",mgmmr_dihedral));
    em_iter->second.insert(std::pair<std::string,float>("mgmmr_improper",mgmmr_improper));

   }

   // search for minimum ligand-site/model energy instance of each form of the site

   std::map<std::string,float> min_gmml;

   for (std::map<InstanceID,std::map<std::string, float> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    float tmpf = 0;
    std::map<std::string, float>::const_iterator titr;
    std::string formname(eitr->first.get_instname());

    // ligand-site terms

    if (use_term_l_e){
      titr = eitr->second.find(std::string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }
    
    if (use_term_l_lj){
      titr = eitr->second.find(std::string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_bond){
      titr = eitr->second.find(std::string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_angle){
      titr = eitr->second.find(std::string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_dihedral){
      titr = eitr->second.find(std::string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_l_improper){
      titr = eitr->second.find(std::string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // ligand-model terms

    if (use_term_ml_e){
      titr = eitr->second.find(std::string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_lj){
     titr = eitr->second.find(std::string("mgmml_lj"));
    if (titr== eitr->second.end())
     ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
    else
     tmpf += titr->second;
    }

    if (use_term_ml_bond){
      titr = eitr->second.find(std::string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_angle){
      titr = eitr->second.find(std::string("mgmml_angle"));
      if (titr== eitr->second.end())
       ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
      else
       tmpf += titr->second;
    }
    
    if (use_term_ml_dihedral){
      titr = eitr->second.find(std::string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    if (use_term_ml_improper){
      titr = eitr->second.find(std::string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // create or update minimum energy entry for this form of the site

    std::map<std::string,float>::iterator meitr(min_gmml.find(formname));
    if (meitr == min_gmml.end()){
     std::pair<std::string,float> tmppair(formname,tmpf);
     min_gmml.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }
   }

   // search for minimum site-site/model electrostatic energy instance of each form of the site

   std::map<std::string,float> min_gmm_e;

   for (std::map<InstanceID,std::map<std::string, float> >::const_iterator eitr=emap.begin(); eitr!=emap.end();++eitr){

    float tmpf = 0;
    std::map<std::string, float>::const_iterator titr;
    std::string formname(eitr->first.get_instname());

    // site-site terms

    if (use_term_s_e){
     titr = eitr->second.find(std::string("gmmr_e"));
     if (titr == eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      tmpf += titr->second;
    }

    // create or update minimum energy entry for this form of the site

    std::map<std::string,float>::iterator meitr(min_gmm_e.find(formname));
    if (meitr == min_gmm_e.end()){
     std::pair<std::string,float> tmppair(formname,tmpf);
     min_gmm_e.insert(tmppair);
    } else if (meitr->second > tmpf){
     meitr->second = tmpf;
    }

   }

   // now collect all energy terms in the final terms

   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){

    float gmmint = 0;
    float gmmback = 0;
    float gmmsite = 0;
    float gmmmodelback = 0;

    std::map<InstanceID,std::map<std::string, float> >::const_iterator eitr(emap.find((*j).get_instid()));
    if (eitr == emap.end())
     ::error("ERROR in QmpbMolecule::calc_mmint instance missing in energy term map\n");
    std::map<std::string, float>::const_iterator titr(eitr->second.find(std::string("gmmback")));
    if (titr== eitr->second.end())
     ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");

    gmmback = titr->second;

    float min_l = 0;
    float min_se = 0;
  
    std::map<std::string, float>::const_iterator ml_itr(min_gmml.find((*j).get_name()));
    if (ml_itr != min_gmml.end()) min_l = ml_itr->second;
    
    std::map<std::string, float>::const_iterator ms_itr(min_gmm_e.find((*j).get_name()));
    if (ms_itr != min_gmm_e.end()) min_se = ms_itr->second;

    float gmml = 0;

    // ligand-site terms

    if (use_term_l_e){
      titr = eitr->second.find(std::string("gmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_lj){
      titr = eitr->second.find(std::string("gmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_bond){
      titr = eitr->second.find(std::string("gmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_angle){
      titr = eitr->second.find(std::string("gmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_dihedral){
      titr = eitr->second.find(std::string("gmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    if (use_term_l_improper){
      titr = eitr->second.find(std::string("gmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmml += titr->second;
    }

    float mgmml = 0;

    // ligand-modelback terms

    if (use_term_ml_e){
      titr = eitr->second.find(std::string("mgmml_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_lj){
      titr = eitr->second.find(std::string("mgmml_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_bond){
      titr = eitr->second.find(std::string("mgmml_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_angle){
      titr = eitr->second.find(std::string("mgmml_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_dihedral){
      titr = eitr->second.find(std::string("mgmml_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    if (use_term_ml_improper){
      titr = eitr->second.find(std::string("mgmml_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmml += titr->second;
    }

    float gmme = 0;

    // site-site terms

    if (use_term_s_e){
      titr = eitr->second.find(std::string("gmmr_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (use_term_s_lj){
      titr = eitr->second.find(std::string("gmmr_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (use_term_s_bond){
      titr = eitr->second.find(std::string("gmmr_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (use_term_s_angle){
      titr = eitr->second.find(std::string("gmmr_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (use_term_s_dihedral){
      titr = eitr->second.find(std::string("gmmr_dihedral"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    if (use_term_s_improper){
      titr = eitr->second.find(std::string("gmmr_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      gmme += titr->second;
    }

    float mgmme = 0;

    // site-model terms

    if (use_term_ms_e){
      titr = eitr->second.find(std::string("mgmmr_e"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (use_term_ms_lj){
      titr = eitr->second.find(std::string("mgmmr_lj"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (use_term_ms_bond){
      titr = eitr->second.find(std::string("mgmmr_bond"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }
    
    if (use_term_ms_angle){
      titr = eitr->second.find(std::string("mgmmr_angle"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    if (use_term_ms_dihedral){
     titr = eitr->second.find(std::string("mgmmr_dihedral"));
    if (titr== eitr->second.end())
     ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
    else
     mgmme += titr->second;
    }

    if (use_term_ms_improper){
      titr = eitr->second.find(std::string("mgmmr_improper"));
     if (titr== eitr->second.end())
      ::error("ERROR in QmpbMolecule::calc_mmint energy term missing\n");
     else
      mgmme += titr->second;
    }

    // minimum ligand-site interaction should be included in model energy
    // correct for this by considering only relative ligand site interactions within 1 form
    // equivalent for intra site and site-model electrostatics
    gmmsite      = gmme  + gmml - min_l - min_se;
    gmmmodelback = mgmme + mgmml;

    (*j).set_gmmsite(gmmsite);
    (*j).set_gmmmodelback(gmmmodelback);
    (*j).set_gmmback(gmmback);
 
   }




  }

  blab1 << "Exiting QmpbMolecule::calc_mmint_charmm" << endl;  

}

void QmpbMolecule::calc_mmg()
{

  blab2 << "Entering QmpbMolecule::calc_mmg" << endl;
  
  for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
   std::list<QmpbSite>::const_iterator iplus = i; iplus++;
   for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    for (std::list<QmpbSite>::const_iterator ii=iplus; ii!=sites.end(); ++ii){
     // dont need interactions among two non-flexible sites
     int stypei(get_rotmap_value((*i).get_sid()));
     int stypeii(get_rotmap_value((*ii).get_sid()));
     if (stypei != 1 && stypei != 2 && stypeii != 1 && stypeii != 2){continue;}     
     for (std::list<QmpbInstance>::const_iterator jj=(*ii).instances.begin(); jj!=(*ii).instances.end(); ++jj){
      float tmpf = 0;
      for (AtomSet::const_iterator k=(*j).site_struct.begin(); k!=(*j).site_struct.end(); ++k){
       for (AtomSet::const_iterator l=(*jj).site_struct.begin(); l!=(*jj).site_struct.end(); ++l){
        CharmmParaNB nbk, nbl;
 
        if (chmpara.contains_attype(k->second)){
         if (chmpara.contains_NB(CharmmID(k->second))){
          nbk = chmpara.get_NB(CharmmID(k->second));
         }
        }
        else if (chmpara.contains_attype(k->second.resname,std::string("O")) && 
                 (k->second.atname == "OT1" || k->second.atname == "OT2")){
         nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("O")));
        }
        else if (chmpara.contains_attype(k->second.resname,std::string("H")) &&
                 (k->second.atname == "HT1" || k->second.atname == "HT2" || k->second.atname == "HT3")){
         nbk = chmpara.get_NB(chmpara.get_attype(k->second.resname,std::string("H")));
        }
        else {
         CharmmParaResTop top;
         if (chmpara.contains_element_NB(top.get_element(k->second.atname)))
          nbk = chmpara.get_element_NB(top.get_element(k->second.atname));
         else { 
          cerr << k->second << endl;
          ::error ("ERROR in QmpbMolecule::calc_mmg: Lennard Jones parameter missing.");
         }
        }
 
        if (chmpara.contains_attype(l->second)){
         if (chmpara.contains_NB(CharmmID(l->second))){
          nbl = chmpara.get_NB(CharmmID(l->second));
         }
        }
        else if (chmpara.contains_attype(l->second.resname,std::string("O")) && 
                 (l->second.atname == "OT1" || l->second.atname == "OT2")){
         nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("O")));
        }
        else if (chmpara.contains_attype(l->second.resname,std::string("H")) &&
                 (l->second.atname == "HT1" || l->second.atname == "HT2" || l->second.atname == "HT3")){
         nbl = chmpara.get_NB(chmpara.get_attype(l->second.resname,std::string("H")));
        }
        else {
         CharmmParaResTop top;
         if (chmpara.contains_element_NB(top.get_element(l->second.atname)))
          nbl = chmpara.get_element_NB(top.get_element(l->second.atname));
         else { 
          cerr << l->second << endl;
          ::error ("ERROR in QmpbMolecule::calc_mmg: Lennard Jones parameter missing.");
         }
        }
      
        float ljek = nbk.lje;
        float ljrk = nbk.ljr;
        float ljel = nbl.lje;
        float ljrl = nbl.ljr;
        float dist = abs_value(k->second.coord - l->second.coord);
        if (ljek <= 0 && ljel <= 0){
         if ( dist != 0 ){
          float r_over_dist_pow_6 = pow(((ljrk + ljrl) / dist ),6);
          tmpf += sqrt(ljek * ljel) * ((r_over_dist_pow_6 * r_over_dist_pow_6) - (2 * r_over_dist_pow_6));
         } else {
          cerr << k->second << endl;
          cerr << l->second << endl;
          ::error ("ERROR in QmpbMolecule::calc_mmg: two atoms have identical coordinates");
         }
        }
        else {
         if (ljek > 0 || ljel > 0){
           ::error ("ERROR in QmpbMolecule::calc_mmg atom has non-negative VDW well depth.");
         }
        }
        
       }
      }
 
      // very small interaction energies are set to zero
      if (tmpf > 5.0E-04 && !contains_mmg((*j).get_instid(),(*jj).get_instid()))
       insert_mmg((*j).get_instid(),(*jj).get_instid(), tmpf);
      
     }
    }
   }
  }

  blab1 << "Exiting QmpbMolecule::calc_mmg" << endl;

}

void QmpbMolecule::filter_instances()
{

  blab2 << "Entering QmpbMolecule::filter_instances" << endl;

  int total   = 0;
  int removed = 0;

  for (std::list<QmpbSite>::iterator i=sites.begin(); i!=sites.end(); ++i){
   float min_gmmint = 1.0E99;
   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    if ((*j).get_gmmint() < min_gmmint){
     min_gmmint = (*j).get_gmmint();
    }
   }
   std::list<InstanceID> to_delete;
   for (std::list<QmpbInstance>::iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    total++;
    if (((*j).get_gmmint() - min_gmmint) > filter_cutoff){
     to_delete.push_back((*j).get_instid());
    }
   }
   for (std::list<InstanceID>::iterator k=to_delete.begin(); k!=to_delete.end();++k){
    removed++;
    (*i).delete_instance(*k);
   }
   (*i).renumber_instances();   
  }

  blab2 << "QmpbMolecule::filter_instances: removed " << removed 
        << " of " << total << " total instances. " 
        << (total - removed) << " instances remain." << endl;
  
  blab1 << "Exiting QmpbMolecule::filter_instances" << endl;

}

std::pair<FinDiffMethod,FinDiffMethod> QmpbMolecule::get_fdm(const std::string& fn)
{
  std::string ogm_name(fn + ".ogm");
  std::string mgm_name(fn + ".mgm");
  // first check existence of files .ogm .mgm
  ifstream ogm_fs(ogm_name.c_str()); bool got_ogm = ogm_fs.good(); ogm_fs.close();
  ifstream mgm_fs(mgm_name.c_str()); bool got_mgm = mgm_fs.good(); mgm_fs.close();
  FinDiffMethod fdm_ogm, fdm_mgm;  
  // read existing files
  if (got_ogm && got_mgm){
   fdm_ogm.read(ogm_name);
   fdm_mgm.read(mgm_name);
  }
  //otherwise create grids
  else {
   
   // dimensions of whole molecule
   Coord gmax(-1.0E+99,-1.0E+99,-1.0E+99);
   Coord gmin(1.0E+99,1.0E+99,1.0E+99);
      
   for (AtomSet::const_iterator i=input_pqr.begin(); i!=input_pqr.end();++i){
    if (i->second.coord.x < gmin.x) gmin.x = i->second.coord.x;
    if (i->second.coord.y < gmin.y) gmin.y = i->second.coord.y;
    if (i->second.coord.z < gmin.z) gmin.z = i->second.coord.z;
    if (i->second.coord.x > gmax.x) gmax.x = i->second.coord.x;
    if (i->second.coord.y > gmax.y) gmax.y = i->second.coord.y;
    if (i->second.coord.z > gmax.z) gmax.z = i->second.coord.z;   
   }
   
   Coord gdim = gmax - gmin;   
   
   // dimensions of largest site (model_struct)

   Coord mdim(0,0,0);
   
   for (std::list<QmpbSite>::const_iterator k=sites.begin(); k!=sites.end(); ++k){
    for (std::list<QmpbInstance>::const_iterator j=(*k).instances.begin(); j!=(*k).instances.end(); ++j){
     Coord mmax(-1.0E+99,-1.0E+99,-1.0E+99);
     Coord mmin(1.0E+99,1.0E+99,1.0E+99);
     for (AtomSet::const_iterator i=(*j).model_struct.begin(); i!=(*j).model_struct.end();++i){
      if (i->second.coord.x < mmin.x) mmin.x = i->second.coord.x;
      if (i->second.coord.y < mmin.y) mmin.y = i->second.coord.y;
      if (i->second.coord.z < mmin.z) mmin.z = i->second.coord.z;
      if (i->second.coord.x > mmax.x) mmax.x = i->second.coord.x;
      if (i->second.coord.y > mmax.y) mmax.y = i->second.coord.y;
      if (i->second.coord.z > mmax.z) mmax.z = i->second.coord.z;      
     }
     Coord tmpdim = mmax - mmin;
     if (abs_value(mdim) < abs_value(tmpdim)) mdim = tmpdim; 
    }
   }
   
   float gmaxdim = gdim.x;
   if (gdim.y > gmaxdim) gmaxdim = gdim.y;  
   if (gdim.z > gmaxdim) gmaxdim = gdim.z;  
   
   float mmaxdim = mdim.x;
   if (mdim.y > mmaxdim) mmaxdim = mdim.y;  
   if (mdim.z > mmaxdim) mmaxdim = mdim.z;  
   
   // now add 3 fdm levels to mgm and ogm

   float s1 = 2.00;
   float s2 = 0.50;
   float s3 = 0.20;
   
   float d1 = gmaxdim + (mmaxdim / 2) + 100; 
   float d2 = mmaxdim + 80;
   float d3 = mmaxdim + 30;
   
   if (d1 < 200) {d1 = 200;}
   if (d2 < 100) {d2 = 100;}
   if (d3 < 50)  {d3 = 50;}
  
   int n1 = static_cast<int>(floor(d1/s1));
   int n2 = static_cast<int>(floor(d2/s2));
   int n3 = static_cast<int>(floor(d3/s3));
   
   // make grid size odd
   if (n1 % 2 == 0) n1++;
   if (n2 % 2 == 0) n2++;
   if (n3 % 2 == 0) n3++;
   
   fdm_ogm.add_level(n1, s1, ON_GEOM_CENT);
   fdm_mgm.add_level(n1, s1, ON_GEOM_CENT);
   fdm_ogm.add_level(n2, s2, ON_CENT_OF_INTR);
   fdm_mgm.add_level(n2, s2, ON_CENT_OF_INTR);
   fdm_ogm.add_level(n3, s3, ON_CENT_OF_INTR);
   fdm_mgm.add_level(n3, s3, ON_CENT_OF_INTR);   
  
  }
  
  // need to do this because iteration over levels is only allowed if resolved
  fdm_ogm.resolve(Coord(0,0,0),Coord(0,0,0));
  fdm_mgm.resolve(Coord(0,0,0),Coord(0,0,0));
  
  return std::pair<FinDiffMethod,FinDiffMethod>(fdm_ogm,fdm_mgm);
}

std::list<std::string> QmpbMolecule::get_ligands()
{
  std::list<std::string> ligands;
   for (std::list<QmpbSite>::const_iterator k=sites.begin(); k!=sites.end(); ++k){
    for (std::list<QmpbInstance>::const_iterator j=(*k).instances.begin(); j!=(*k).instances.end(); ++j){
     for (std::map<std::string,int>::const_iterator i=(*j).ligandmap.begin(); i!=(*j).ligandmap.end(); ++i){
      ligands.push_back(i->first);
     }
     ligands.sort(); ligands.unique();     
    }
   }
  return ligands;
}

void QmpbMolecule::write_qmpb_input(const string& dir, const string& fn)
{

  blab2 << "Entering QmpbMolecule::write_qmpb_input" << endl;


  std::list<std::string> ligands = get_ligands();
  // want the protons first so sort and reverse
  ligands.sort(); ligands.reverse();

  
  // FIXME? the whole grid data retrieval is quite cumbersome
  // due to all the private items in the FinDiffMethod
  
  std::pair<FinDiffMethod,FinDiffMethod> fdm_pair = get_fdm(name);
  std::list<CenteringStyle> mcstylelist, ocstylelist;
  std::list<int>            mgrddimlist, ogrddimlist;
  std::list<float>          mgrdspclist, ogrdspclist;
  std::vector<Coord>        mgrdcenlist, ogrdcenlist;

   
  CubeLatSpec *clso = fdm_pair.first.get_coarsest();

  
  do {
   ocstylelist.push_back((*clso).get_centering_style());
   ogrddimlist.push_back((*clso).get_grid_dim());
   ogrdspclist.push_back((*clso).get_spacing());
   ogrdcenlist.push_back((*clso).get_center());
  } while (clso = fdm_pair.first.get_finer());


  CubeLatSpec *clsm = fdm_pair.second.get_coarsest();

  do {
   mcstylelist.push_back((*clsm).get_centering_style());
   mgrddimlist.push_back((*clsm).get_grid_dim());
   mgrdspclist.push_back((*clsm).get_spacing());
   mgrdcenlist.push_back((*clsm).get_center());
  } while (clsm = fdm_pair.second.get_finer());
  
  blab3 << "Writing the QMPB input script \"" << fn << "\" and structures to directory " 
        << "\"" << dir<< "\"." << endl;

  ////// check if the qmpb_input directory exists, otherwise create //////  

  if (!fs::is_directory(qmpbdir)){
   if (fs::exists(qmpbdir)) fs::remove_all(qmpbdir);
   fs::create_directory(qmpbdir);
  }
  
  ////// write the QMPB input script //////  
  
  std::string qmpb_input_fn(dir + "/" + fn);
  
  ofstream of(qmpb_input_fn.c_str());
  if (!of.good())
    ::error("QmpbMolecule::write_mmint failed to open file for writing, ",
            qmpb_input_fn.c_str());
 
   // write header

   
   of << "# QMPB run with multiple rotameric instances" << endl;
   of << "# Input generated by MEAD make-rotamers V 0.1" << endl;
   
   of << endl << endl;
   
   of << "meadpath = " << meadpath << endl;
   of << "I = " << PhysCond::get_ionicstr() << endl;
   of << "T = " << PhysCond::get_T() << endl;
   of << "backfile = background.pqr" << endl;
   of << "epsin1 = " << epsin1 << endl;
   of << "epsin2 = " << epsin2 << endl;

 
   of << "OGMcenter =";
   short ix = 0;
   for (std::list<CenteringStyle>::const_iterator i=ocstylelist.begin(); i!=ocstylelist.end(); ++i){
    switch (*i) {
     case ON_ORIGIN:
       of << " ON_ORIGIN";
       break;
     case ON_CENT_OF_INTR:
       of << " ON_CENT_OF_INTR";
       break;
     case ON_GEOM_CENT:
       of << " ON_GEOM_CENT";
       break;
     case SPECIFIC:
       of << " " << ogrdcenlist[ix];
     break;
    }        
   } of << endl;

   
   of << "OGMpoints ="; 
   for (std::list<int>::const_iterator i=ogrddimlist.begin(); i!=ogrddimlist.end(); ++i){
     of << " " << *i;
   } of << endl;

   
   of << "OGMspace ="; 
   for (std::list<float>::const_iterator i=ogrdspclist.begin(); i!=ogrdspclist.end(); ++i){
     of << " " << setprecision(2) << fixed << *i;
   } of << endl;


   of << "MGMcenter =";
   ix = 0;
   for (std::list<CenteringStyle>::const_iterator i=mcstylelist.begin(); i!=mcstylelist.end(); ++i){
    switch (*i) {
     case ON_ORIGIN:
       of << " ON_ORIGIN";
       break;
     case ON_CENT_OF_INTR:
       of << " ON_CENT_OF_INTR";
       break;
     case ON_GEOM_CENT:
       of << " ON_GEOM_CENT";
       break;
     case SPECIFIC:
       of << " " << mgrdcenlist[ix];
     break;
    }        
   } of << endl;

   
   of << "MGMpoints ="; 
   for (std::list<int>::const_iterator i=mgrddimlist.begin(); i!=mgrddimlist.end(); ++i){
     of << " " << *i;
   } of << endl;

   
   of << "MGMspace ="; 
   for (std::list<float>::const_iterator i=mgrdspclist.begin(); i!=mgrdspclist.end(); ++i){
     of << " " << setprecision(2) << fixed << *i;
   } of << endl;

   
   of << "Ligand_Labels =";   
   for (std::list<std::string>::const_iterator i=ligands.begin(); i!=ligands.end(); ++i){
    of << " " << *i;
   } of << endl;
   
   of << endl << endl << "#####################" << endl << endl;


 
   // now write the sites and their structure files
   
   int sid = -1;
   
   for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
    sid++;
    int iid = -1;
    int c(((*i).get_sid()).get_chain());
    int s(((*i).get_sid()).get_site());
    std::string str_c, str_s;
    stringstream ss1,ss2;
    ss1 << c; ss1 >> str_c;
    ss2 << s; ss2 >> str_s;    
    
    of << endl << "#####################" << endl << endl;
    
    for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
     iid++;
     int in(((*j).get_instid()).get_instance());
     std::string str_in, slbl, ilbl, milbl;
     float gmmint((*j).get_gmmint());
     int protons((*j).ligands_bound(std::string("proton")));
     int electrons((*j).ligands_bound(std::string("electron")));
     stringstream ss3;
     ss3 << in; ss3 >> str_in;
     slbl = ((*i).get_sid()).get_sitename() + "-" + str_c + "-" + str_s;
     ilbl = ((*j).get_instid()).get_instname() + "-" + str_in;
     milbl = ((*j).get_instid()).get_instname() + "-" + str_in + "_model";
     
     // for now let the MM energies aside, we anyway need to merge the interaction energy matrices
     
     of << endl;
          
     of << "MMsite " << slbl << " " << ilbl << endl;
     of << "sid=" << sid << endl;
     of << "iid=" << iid << endl;
     of << "  file=" << slbl << "-" << ilbl << ".pqr" << endl;   
     of << "  ref=" << milbl << endl;   
     of << "  Gmm=0" << endl;   

     of << endl;

     of << "Modelsite " << slbl << " " << milbl << endl;
     of << "sid=" << sid << endl;
     of << "iid=" << iid << endl;
     of << "  file=" << slbl << "-" << milbl << ".pqr" << endl;   
     of << "  ref=" << ilbl << endl;   
     of << "  Gmodel=" << (*j).get_gmodel() << endl;   
     of << "  eps=" << "4" << endl;   
     of << "  N=";
     for (std::list<std::string>::const_iterator k=ligands.begin(); k!=ligands.end(); ++k){
       of << " " << (*j).ligands_bound(*k);
     } of << endl;
     
     // write structure files
     (*j).site_struct.write_pqr(std::string(dir + "/" + slbl + "-" + ilbl + ".pqr"));
     (*j).model_struct.write_pqr(std::string(dir + "/" + slbl + "-" + milbl + ".pqr"));
     
    }
   }
   
  // need a newline for QMPB parser at file end 
  of << endl;
  
  if (! of.good())
    ::error("QmpbMolecule::write_mmint failed to write file, ",
            qmpb_input_fn.c_str());
  of.close();
  
  background.write_pqr(std::string(dir + "/background.pqr"));

  blab1 << "Exiting QmpbMolecule::write_qmpb_input" << endl;

}

void QmpbMolecule::write_mmint(const string& fn)
{

  blab2 << "Entering QmpbMolecule::write_mmint" << endl;

  ofstream of(fn.c_str());
  if (!of.good())
    ::error("QmpbMolecule::write_mmint failed to open file for writing, ",
            fn.c_str());
 
   for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
    int c(((*i).get_sid()).get_chain());
    int s(((*i).get_sid()).get_site());
    std::string str_c, str_s;
    stringstream ss1,ss2;
    ss1 << c; ss1 >> str_c;
    ss2 << s; ss2 >> str_s;    
    for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
     int in(((*j).get_instid()).get_instance());
     std::string str_in, slbl, ilbl;
     float gmmint((*j).get_gmmint());
     int protons((*j).ligands_bound(std::string("proton")));
     int electrons((*j).ligands_bound(std::string("electron")));
     stringstream ss3;
     ss3 << in; ss3 >> str_in;
     slbl = ((*i).get_sid()).get_sitename() + "-" + str_c + "-" + str_s;
     ilbl = ((*j).get_instid()).get_instname() + "-" + str_in;
     int w1 = slbl.size();
     if (w1<13){w1=13;}
     int w2 = ilbl.size();
     if (w2<10){w2=10;}
     of << right << setw(w1) << slbl << "  " << setw(w2) << ilbl
        << setw(9) << setprecision(3) << fixed;
     
     if (gmmint > -10000 && gmmint < 10000)
      of << setw(9) << setprecision(3) << fixed << gmmint;
     else
      of << " " << setprecision(6) << scientific << gmmint;     
     
     of << setw(3) << protons << setw(3) << electrons << "  0" << endl;
    }
   }
  
  if (! of.good())
    ::error("QmpbMolecule::write_mmint failed to write file, ",
            fn.c_str());
  of.close();

  blab1 << "Exiting QmpbMolecule::write_mmint" << endl;

}

void QmpbMolecule::write_mmg(const string& fn)
{

  blab2 << "Entering QmpbMolecule::write_mmg" << endl;

  ofstream of(fn.c_str());
  if (!of.good())
    ::error("QmpbMolecule::write_mmg failed to open file for writing, ",
            fn.c_str());

   for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
    int ci(((*i).get_sid()).get_chain());
    int si(((*i).get_sid()).get_site());
    std::string str_ci, str_si;
    stringstream ss1i,ss2i;
    ss1i << ci; ss1i >> str_ci;
    ss2i << si; ss2i >> str_si;    
    for (std::list<QmpbSite>::const_iterator ii=sites.begin(); ii!=sites.end(); ++ii){
     int cii(((*ii).get_sid()).get_chain());
     int sii(((*ii).get_sid()).get_site());
     std::string str_cii, str_sii;
     stringstream ss1ii,ss2ii;
     ss1ii << cii; ss1ii >> str_cii;
     ss2ii << sii; ss2ii >> str_sii;
     for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
      int inj(((*j).get_instid()).get_instance());
      std::string str_inj;
      stringstream ss3j;
      ss3j << inj; ss3j >> str_inj;
      for (std::list<QmpbInstance>::const_iterator jj=(*ii).instances.begin(); jj!=(*ii).instances.end(); ++jj){
       int injj(((*jj).get_instid()).get_instance());
       std::string str_injj;
       stringstream ss3jj;
       ss3jj << injj; ss3jj >> str_injj;
       std::string slbl1, slbl2, ilbl1, ilbl2;
       slbl1 = ((*i).get_sid()).get_sitename() + "-" + str_ci + "-" + str_si;
       ilbl1 = ((*j).get_instid()).get_instname() + "-" + str_inj;
       slbl2 = ((*ii).get_sid()).get_sitename() + "-" + str_cii + "-" + str_sii;
       ilbl2 = ((*jj).get_instid()).get_instname() + "-" + str_injj;
       int w11 = slbl1.size();
       if (w11<13){w11=13;}
       int w21 = ilbl1.size();
       if (w21<10){w21=10;}
       int w12 = slbl2.size();
       if (w12<13){w12=13;}
       int w22 = ilbl2.size();
       if (w22<10){w22=10;}       
       float gmmg(get_mmg((*j).get_instid(),(*jj).get_instid()));
       int protons((*j).ligands_bound(std::string("proton")));
       int electrons((*j).ligands_bound(std::string("electron")));
       of << right << setw(w11) << slbl1 << "  " << setw(w21) << ilbl1
          << right << setw(w12) << slbl2 << "  " << setw(w22) << ilbl2;
       if (gmmg > -10000 && gmmg < 10000)
        of << setw(9) << setprecision(3) << fixed << gmmg << endl;
       else
        of << " " << setprecision(6) << scientific << gmmg << endl;       
      }
     }
    }
   }
  
  if (! of.good())
    ::error("QmpbMolecule::write_mmg failed to write file, ",
            fn.c_str());
  of.close();


  blab1 << "Exiting QmpbMolecule::write_mmg" << endl;

}

void QmpbMolecule::print_mmint() const
{

  blab2 << "Entering QmpbMolecule::print_mmint" << endl;

  for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
   cout << "\n---------------------------------------------------------------------\n" << endl; 
   cout << "printing intrinsic MM energy of site " << (*i).get_sid() << " :\n" << endl; 
   cout << "InstanceID      Gmmint      Gmmsite   Gmmmodelback   Gmmback" << endl; 
   for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    cout << (*j).get_instid() << "      " << (*j).get_gmmint() << "      "
         << (*j).get_gmmsite()  << "   " << (*j).get_gmmmodelback()  << "   " << (*j).get_gmmback() << endl;
   }
   cout << "\n---------------------------------------------------------------------\n" << endl; 
  }

  blab1 << "Exiting QmpbMolecule::print_mmint" << endl;

}

void QmpbMolecule::print_mmg() const
{

  blab2 << "Entering QmpbMolecule::print_mmg" << endl;

  cout << "\n---------------------------------------------------------------------\n" << endl; 
  cout << "printing interaction energies " << ":\n" << endl; 
  cout << "SiteID   InstanceID      SiteID   InstanceID      Gmmg" << endl; 
 
  for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
   for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    for (std::list<QmpbSite>::const_iterator ii=sites.begin(); ii!=sites.end(); ++ii){
     for (std::list<QmpbInstance>::const_iterator jj=(*ii).instances.begin(); jj!=(*ii).instances.end(); ++jj){
      cout << (*i).get_sid()  << "   " << (*j).get_instid()  << "      "
           << (*ii).get_sid() << "   " << (*jj).get_instid() << "      " 
           << get_mmg((*j).get_instid(), (*jj).get_instid()) << endl;
     }
    }
   }
  }

  cout << "\n---------------------------------------------------------------------\n" << endl; 

  blab1 << "Exiting QmpbMolecule::print_mmg" << endl;

}

void QmpbMolecule::read_site_defs(){

  blab2 << "Entering QmpbMolecule::read_site_defs" << endl;

  // read .est files for all titratable sites
  for (map<SiteID,int>::const_iterator i = rotmap.begin(); i!= rotmap.end(); ++i){
   // titratable and flexible + titratable sites
   if (i->second == 1 || i->second == 3){
    string nm = i->first.get_sitename();
    if(!contains_sitedef(nm)){
     QmpbSiteDef sd(nm);
     sd.read_est();
     insert_sitedef(sd);
     //////// debug ////////
     if(print_debug){
     sd.print_est();
     }
     ///////////////////////
    }
   }
  }

  blab1 << "Exiting QmpbMolecule::read_site_defs" << endl;
}

void QmpbMolecule::read_rot ()
{
  string rot_fn = name + DOT + "rot";
  read_rot(rot_fn);
}

void QmpbMolecule::read_rot (const string& fn_string){

  blab2 << "QmpbMolecule::read_rot: starting\n";

  const char *fn = fn_string.c_str();
  string line;
  int nsites = 0; int ntflex = 0; int nt = 0; int nflex = 0; int nign = 0;

  boost::regex re_rot("^\\s*(\\w+)-(\\d+)-(\\d+)\\s+(\\d+)");

  boost::match_results<std::string::const_iterator> match;
  boost::match_flag_type flags = boost::match_default;

  ifstream rotfile(fn);

  if (!rotfile)
    ::error ("QmpbMolecule::read_rot: Cannot open input file, ", fn, "\n");

  // start of parsing
  while (!rotfile.eof() ){

   getline (rotfile, line);
   std::string::const_iterator start, end;
   start = line.begin();
   end   = line.end();

    if (boost::regex_search(start, end, match, re_rot, flags)){
     int cn, sn, sitetype;
     string sitelabel = std::string(match[1].first, match[1].second);
     istringstream ist1(std::string(match[2].first, match[2].second));
     ist1 >> cn;
     istringstream ist2(std::string(match[3].first, match[3].second));
     ist2 >> sn;
     istringstream ist3(std::string(match[4].first, match[4].second));
     ist3 >> sitetype;
     if (sitetype == 0){
      // dont need to consider non-flexible and non-titrateable sites
      nign++;
      continue;
     }
     SiteID(sitelabel, cn, sn);

     pair<map<SiteID,int>::iterator,bool> tp;
     tp = insert_rotmap_entry(SiteID(sitelabel, 0, cn, sn), sitetype);
     if (tp.second){
      nsites++;
      if      (sitetype == 1){ntflex++;}
      else if (sitetype == 2){nflex++;}
      else if (sitetype == 3){nt++;}
      else {
       cerr << "QmpbMolecule::read_rot: Sitetype must be one of:"
            << "   0 = non-titratable non-flexible\n"
            << "   1 =     titratable     flexible\n"
            << "   2 = non titratable     flexible\n"
            << "   3 =     titratable non-flexible\n" << endl;
       ::error ("QmpbMolecule::read_rot: unknown site type, ", line.c_str(), "\n");
      }
     }
    }


  } // end of parsing

  rotfile.close();

  ///// debug /////
  if (print_debug){
  for (map<SiteID,int>::const_iterator i = rotmap.begin(); i!= rotmap.end(); ++i){
   cout << i->first << "   " << i->second << endl;
  }
  }
  /////////////////

  blab2  << "QmpbMolecule::read_rot: finished with " <<
  nsites << " sites of which " <<
  ntflex << " are titratable and flexible, " <<
  nflex  << " are non-titratable and flexible, " <<
  nt     << " are titratable and non-flexible\n" <<
  nign   << " non-titratable and non-flexible sites were ignored." << endl;

  blab1 << "QmpbMolecule::read_rot: exiting\n";
}

void QmpbMolecule::create_sitemap()
{
  // FIXME: HOW best to define types in .sites file
  // till now 0 for H+ binding, 1 for e- binding, 2 for non-titratable but flexible
  // with generalized titration theory multiple ligands in 1 site/form and more than
  // just e-, H+ so what to do here?, append ligandlist?

  // save type and number of instances of each site to sitemap

  cout << "create sitemap" << endl;

  sitemap.clear();
  sites.sort();

  for (std::list<QmpbSite>::const_iterator i=sites.begin(); i!=sites.end(); ++i){
   std::list<std::string> liglist;
   for (std::list<QmpbInstance>::const_iterator j=(*i).instances.begin(); j!=(*i).instances.end(); ++j){
    for (std::map<std::string,int>::const_iterator k = (*j).ligandmap.begin(); k!= (*j).ligandmap.end(); ++k){
     if (k->second > int(0)){
      liglist.push_back(k->first);
     }
    }
   }
   liglist.sort(); liglist.unique();
   int nlig = liglist.size();
   int nrot = (*i).get_nconf();
   int type = 0;
   if (nlig > 0){
    type = 0;
    if (nlig == 1 && (*i).contains_ligand("proton")){
     type = 0;
    }
    else if (nlig == 1 && (*i).contains_ligand("electron")){
     type = 1;
    }
    else if (nlig > 1){
     type = 3;
    }
   }
   else if (nlig == 0){
    type = 2;
   }
   if (!contains_sitemap_entry((*i).get_sid()))
    insert_sitemap_entry((*i).get_sid(), type, nrot);
   else // should never happen
    ::error ("QmpbMolecule::create_sitemap: attempt to insert duplicate site into sitemap.\n");
  }

}

void QmpbMolecule::write_sites_file() const {
 string fn = name + ".sites";
 write_sites_file(fn);
}

void QmpbMolecule::write_sites_file(const string& fn) const {
  FILE * outfile;
  outfile = fopen (fn.c_str(),"w");
  if (outfile == NULL)
    ::error ("QmpbMolecule::write_sites_file: Cannot open file, ", fn.c_str(), "\n");
  for (std::map<SiteID,pair<int,int> >::const_iterator i = sitemap.begin(); i!=sitemap.end(); ++i) {
    std::stringstream ss1, ss2;
    std::string tmpstr1, tmpstr2, tmpstr3;
    ss1 << i->first.get_chain();
    ss1 >> tmpstr1;
    ss2 << i->first.get_site();
    ss2 >> tmpstr2;
    tmpstr3 = i->first.get_sitename() + "-" + tmpstr1 + "-" + tmpstr2;
    fprintf (outfile, "%-18s %d %3d\n", tmpstr3.c_str(), i->second.first, i->second.second);
  }
  fclose(outfile);

}



// QmpbMolecule.cc ends here
