/* ChargeDist subclass for ion gradients, membrane potential simulation

    Modified 2006--2007 from AtomChargeSet

    Thomas Ullmann, BISB Group,
    Dep. of Biopolymers, University of Bayreuth,
    Universitaetsstrasse 30 BGI, 95447 Bayreuth, Germany
    thomas.ullmann@uni-bayreuth.de

    Copyright (c) 1993--1995 by Donald Bashford

    This source code file is part of the MEAD (Macroscopic
    Electrostatics with Atomic Detail) package of objects and
    programs.  MEAD is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 1, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
    02139, USA.

    Donald Bashford can be contacted by electronic mail by the address,
    bashford@scripps.edu, or by paper mail at Department of Molecular
    Biology, The Scripps Research Institute, 10666 North Torrey Pines
    Road, La Jolla, California 92037.

$Id: MembChargeSet.cc,v 2.14 2004/12/06 17:57:53 bashford Exp $
*/

#include "MEAD_e/MembChargeSet.h"

#include "MEAD_e/OnePointCharge.h"
#include "MEAD_e/ManyPointCharge.h"
#include "MEAD_e/globals.h"
#include "MEAD_e/CubeLatSpec.h"
#include "MEAD_e/ChargeCubeRep.h"
#include "MEAD_e/Bigmem.h"
#include <string.h>
#include <math.h>
#include <algorithm>
#include <stdio.h>
#include <iostream>


/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   1 atom set for the united QM, normal + cavity regions
   It is meant to complement the SixValueDielMembThreeAtoms/ElectrolyteMembAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

MembChargeSetMembAtoms::MembChargeSetMembAtoms(const AtomSet& a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, stl), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2),
    memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }

}

MembChargeSetMembAtoms::MembChargeSetMembAtoms (const AtomSet & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, PhysCond::get_sterln()), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSetMembAtoms::MembChargeSetMembAtoms (const AtomSet & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, PhysCond::get_sterln()), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembAtoms::MembChargeSetMembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembAtoms::MembChargeSetMembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembAtoms::MembChargeSetMembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSetMembAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSetMembAtoms::get_atomset: atom set not defined");
  return _ats;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted

int MembChargeSetMembAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSetMembAtoms::number_points() const
{
    return 0;
}
// function on basis of cuberep needed ?

float MembChargeSetMembAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}

ChargeCubeRep*
MembChargeSetMembAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSetMembAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag* acc_array = new AccTag[ncube];
  AccTag* tmp_array = new AccTag[ncube];

  if (ion_acc.check_is_calculated()) {
    blab2 << "MembChargeSetMembAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc.calc_cuberep(cls, tmp_array);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc.calc_cuberep(cls, tmp_array);
  } else {
    blab2 << "MembChargeSetMembAtoms::get_cuberep calculating "
      << " simple ion accessible volume representation to make a cubic representation" << endl;

    for (int h=0; h<ncube; ++h) {
      tmp_array[h] = exterior;
    }

    for(AtomSet::const_iterator ip = _ats.begin(); ip!=_ats.end(); ++ip ) {

      int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
      int radstlsq = rad * rad;
      Coord atom_coord = ip->second.coord - grid_origin_in_space;
      int k = (int) round((atom_coord.z/spacing));
      int i = (int) round((atom_coord.x/spacing));
      int j = (int) round((atom_coord.y/spacing));
      int h = (i * nsq) + (j * grid_dim) + k;
      // skip atom if not in grid
      if (k < stlin || i < stlin || j < stlin ||
          k > grid_dim_plus_stli ||
          i > grid_dim_plus_stli ||
          j > grid_dim_plus_stli){continue;}
      int lower_l = k - rad;
      int lower_m = i - rad;
      int lower_n = j - rad;
      int upper_l = k + rad + 1;
      int upper_m = i + rad + 1;
      int upper_n = j + rad + 1;
      bool fastswitch = true;
      if (lower_l < 0) {lower_l = 0; fastswitch = false;}
      if (lower_m < 0) {lower_m = 0; fastswitch = false;}
      if (lower_n < 0) {lower_n = 0; fastswitch = false;}
      if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
      if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
      if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
      if (fastswitch) {
        // in the inner grid we have a full sphere around the point
        // and can save time by doing loops and distance
        // only in +,+,+ octand of sphere
        // tried to improve by computing dist only for 1/48th sphere
        // also permutating x,y,z and doing loops and distance only
        // in in x >= y >= z but result is slower
        int dl = -1;
        int lterm = -grid_dim;
        for (int l=k; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          lterm += grid_dim;
          int lsqterm = dl*nsq;
          int dm = -1;
          int mterm = -grid_dim;
          for (int m=i; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            // out of sphere already in plane
            if (dmlsq > radstlsq){break;}
            mterm += grid_dim;
            int msqterm = dm * nsq;
            int dn = -1;
            int nterm = -grid_dim;
            for (int n=j; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq > radstlsq){break;}
              int nsqterm = dn*nsq;
              nterm += grid_dim;
              // 1: x, y, z
              // here we permutate +/- signs in the three components
              // + + +
              int hh = h + msqterm + nterm + dl;
              tmp_array[hh] = interior;
              // + + -
              hh = hh - dl  - dl;
              tmp_array[hh] = interior;
              // + - -
              hh = hh - nterm - nterm;
              tmp_array[hh] = interior;
              // + - +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
              // - - +
              hh = hh - msqterm - msqterm;
              tmp_array[hh] = interior;
              // - - -
              hh = hh - dl - dl;
              tmp_array[hh] = interior;
              // - + -
              hh = hh + nterm + nterm;
              tmp_array[hh] = interior;
              // - + +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
            }
          }
        }
      }
      else {
        // these are atoms at the grid border
        int dl = -1;
        for (int l=lower_l; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          int dm = -1;
          for (int m=lower_m; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            if (dmlsq > radstlsq){break;}
            int msqterm = m*nsq;
            int dn = -1;
            for (int n=lower_n; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq <= radstlsq){
                int hh = msqterm + n*grid_dim + l;
                tmp_array[hh] = interior;
              }
            }
          }
        }
      }
    }


  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
          else if (rsq > holesq) {
            acc_array[h] = interior;
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float corewidth = zucbou - zlcbou;
  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // linear gradient of ionic strength and membrane potential within membrane pore
            chtmp[h] = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
                        * (qmemb_fac * membrane_pos);
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = ionic_str1 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = - 1.0 * ionic_str2 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSetMembAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembAtoms::get_cuberep" << endl;
  return ccrp;
}


ChargeCubeRep*
MembChargeSetMembAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSetMembAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag* acc_array = new AccTag[ncube];
  AccTag* tmp_array = new AccTag[ncube];

  if (ion_acc.check_is_calculated()) {
    blab2 << "MembChargeSetMembAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc.calc_cuberep(cls, tmp_array);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc.calc_cuberep(cls, tmp_array);
  } else {
    blab2 << "MembChargeSetMembAtoms::get_cuberep calculating "
      << " simple ion accessible volume representation to make a cubic representation" << endl;

    for (int h=0; h<ncube; ++h) {
      tmp_array[h] = exterior;
    }

    for(AtomSet::const_iterator ip = _ats.begin(); ip!=_ats.end(); ++ip ) {

      int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
      int radstlsq = rad * rad;
      Coord atom_coord = ip->second.coord - grid_origin_in_space;
      int k = (int) round((atom_coord.z/spacing));
      int i = (int) round((atom_coord.x/spacing));
      int j = (int) round((atom_coord.y/spacing));
      int h = (i * nsq) + (j * grid_dim) + k;
      // skip atom if not in grid
      if (k < stlin || i < stlin || j < stlin ||
          k > grid_dim_plus_stli ||
          i > grid_dim_plus_stli ||
          j > grid_dim_plus_stli){continue;}
      int lower_l = k - rad;
      int lower_m = i - rad;
      int lower_n = j - rad;
      int upper_l = k + rad + 1;
      int upper_m = i + rad + 1;
      int upper_n = j + rad + 1;
      bool fastswitch = true;
      if (lower_l < 0) {lower_l = 0; fastswitch = false;}
      if (lower_m < 0) {lower_m = 0; fastswitch = false;}
      if (lower_n < 0) {lower_n = 0; fastswitch = false;}
      if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
      if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
      if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
      if (fastswitch) {
        // in the inner grid we have a full sphere around the point
        // and can save time by doing loops and distance
        // only in +,+,+ octand of sphere
        // tried to improve by computing dist only for 1/48th sphere
        // also permutating x,y,z and doing loops and distance only
        // in in x >= y >= z but result is slower
        int dl = -1;
        int lterm = -grid_dim;
        for (int l=k; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          lterm += grid_dim;
          int lsqterm = dl*nsq;
          int dm = -1;
          int mterm = -grid_dim;
          for (int m=i; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            // out of sphere already in plane
            if (dmlsq > radstlsq){break;}
            mterm += grid_dim;
            int msqterm = dm * nsq;
            int dn = -1;
            int nterm = -grid_dim;
            for (int n=j; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq > radstlsq){break;}
              int nsqterm = dn*nsq;
              nterm += grid_dim;
              // 1: x, y, z
              // here we permutate +/- signs in the three components
              // + + +
              int hh = h + msqterm + nterm + dl;
              tmp_array[hh] = interior;
              // + + -
              hh = hh - dl  - dl;
              tmp_array[hh] = interior;
              // + - -
              hh = hh - nterm - nterm;
              tmp_array[hh] = interior;
              // + - +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
              // - - +
              hh = hh - msqterm - msqterm;
              tmp_array[hh] = interior;
              // - - -
              hh = hh - dl - dl;
              tmp_array[hh] = interior;
              // - + -
              hh = hh + nterm + nterm;
              tmp_array[hh] = interior;
              // - + +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
            }
          }
        }
      }
      else {
        // these are atoms at the grid border
        int dl = -1;
        for (int l=lower_l; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          int dm = -1;
          for (int m=lower_m; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            if (dmlsq > radstlsq){break;}
            int msqterm = m*nsq;
            int dn = -1;
            for (int n=lower_n; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq <= radstlsq){
                int hh = msqterm + n*grid_dim + l;
                tmp_array[hh] = interior;
              }
            }
          }
        }
      }
    }


  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
          else if (rsq > holesq) {
            acc_array[h] = interior;
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  // double charge0 = 0, charge1 = 0;

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // linear gradient of ionic strength and membrane potential within membrane pore
            if (do_sum)
             chtmp[h]  = ((ionic_str1 * membrane_pos * (maxpot_half - phi.value(tmpvec)))
                        - (ionic_str2 * (1 - membrane_pos) * (phi.value(tmpvec) + maxpot_half)))
                         * -qindu_fac;
            else
             chtmp[h]  = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
                        *  phi.value(tmpvec) * qindu_fac;
            //if      (chtmp[h] > 0) charge1 += chtmp[h];
            //else if (chtmp[h] < 0) charge0 += chtmp[h];
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            //charge1 += chtmp[h];
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            //charge0 += chtmp[h];
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  /*

  blab3 << "charge1 " << charge1 << endl
        << "charge0 " << charge0 << endl;

  float mean_charge = (float) 0.5 * (charge1 - charge0);
  if (mean_charge < 0) mean_charge = -mean_charge;

  float abs_mp = memb_pot;
  if (abs_mp < 0) abs_mp = -abs_mp;

  float capacitance = mean_charge * PhysCond::get_elementary_charge() / abs_mp;
  // capacitive energy G = 1/2 * C * U**2
  // 1/energy_unit converts [J] to atomic units which are the converted to user
  // requested unit by econv
  float e_capac     = 0.5 * capacitance * abs_mp * abs_mp / PhysCond::get_energy_unit();

  cout  << "mean absolute value of charge = " << mean_charge << endl;
  cout  << "Capacitance                   = " << capacitance << " (in [Farad])" << endl;
  cout  << "Capacitive Energy             = " << e_capac * PhysCond::get_econv()
         << " (probably in [kcal/mol])" << endl;
  */

  blab3 << "MembChargeSetMembAtoms::get_cuberep(cls,phi,do_sum): finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembTwoAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms(const AtomSet& a1, const AtomSet& a2,
 float zlh, float zuh, float zlc, float zuc, Coord c, float hr, float ionic1, float ionic2,
 float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, stl), ion_acc2(a2, stl),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }

}

MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms (const AtomSet & a1, const AtomSet & a2,
 float zlh, float zuh, float zlc, float zuc, Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms (const AtomSet & a1, const AtomSet & a2, float zlh, float zuh,
                                                      float zlc, float zuc, Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms (const IonAccVol & a1, const IonAccVol & a2, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms (const IonAccVol& a1, const IonAccVol& a2,
 float zlh, float zuh, float zlc, float zuc, Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembTwoAtoms::MembChargeSetMembTwoAtoms (const IonAccVol & a1, const IonAccVol & a2,
   float zlh, float zuh, float zlc, float zuc, Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSetMembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSetMembTwoAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSetMembTwoAtoms::get_atomset: atom set not defined");
  return _ats1;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted


int MembChargeSetMembTwoAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSetMembTwoAtoms::number_points() const
{
    return 0;
}

// function on basis of cuberep needed ?

float MembChargeSetMembTwoAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}


ChargeCubeRep*
MembChargeSetMembTwoAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSetMembTwoAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[2];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated()) {
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else {
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<2; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

    for (int g=0; g<2; ++g){
      if        (g==0){
       startats = _ats1.begin(); endats = _ats1.end();
      } else if (g==1){
       startats = _ats2.begin(); endats = _ats2.end();
      }
      for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          else {
            // membrane core
            // case 1: ion accessible volume given by inside of
            //  dummy atom set elycavset
            //  use AccTag exterior
            // case 2: ion accessible volume given by cylindrical hole
            //  use AccTag undecided to distinguish channel volume
            //  from bulk electrolyte
            if (tmp_array[1][h] == interior) {
              // inside elycav
              acc_array[h] = exterior;
            }
            else if (rsq <= holesq) {
              // outside elycav and inside cylinder
              acc_array[h] = undecided;
            }
            else {
              // outside elycav and outside cylinder
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float corewidth = zucbou - zlcbou;
  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior){
            chtmp[h] = 0.0;
          }
          else if (acc_array[h] == exterior){
            // ionic strength of closest region;
            // ambigous if a large ion accessible cavity
            // happens to lie in the center of the
            // membrane core - seems very improbable
            if (membrane_pos >= 0.5){
              chtmp[h] = ionic_str1 * qmemb_fac;
            }
            else{
              chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
            }
          }
          else if (acc_array[h] == undecided){
            // linear gradient of ionic strength and membrane potential within membrane pore
            chtmp[h] = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
                        * (qmemb_fac * membrane_pos);
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = ionic_str1 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSetMembTwoAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembTwoAtoms::get_cuberep" << endl;
  return ccrp;
}

ChargeCubeRep*
MembChargeSetMembTwoAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSetMembTwoAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[2];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated()) {
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else {
    blab2 << "MembChargeSetMembTwoAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<2; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

    for (int g=0; g<2; ++g){
      if        (g==0){
       startats = _ats1.begin(); endats = _ats1.end();
      } else if (g==1){
       startats = _ats2.begin(); endats = _ats2.end();
      }
      for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          else {
            // membrane core
            // case 1: ion accessible volume given by inside of
            //  dummy atom set elycavset
            //  use AccTag exterior
            // case 2: ion accessible volume given by cylindrical hole
            //  use AccTag undecided to distinguish channel volume
            //  from bulk electrolyte
            if (tmp_array[1][h] == interior) {
              // inside elycav
              acc_array[h] = exterior;
            }
            else if (rsq <= holesq) {
              // outside elycav and inside cylinder
              acc_array[h] = undecided;
            }
            else {
              // outside elycav and outside cylinder
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior){
            chtmp[h] = 0.0;
          }
          else if (acc_array[h] == exterior){
            // ionic strength of closest region;
            // ambigous if a large ion accessible cavity
            // happens to lie in the center of the
            // membrane core - seems very improbable
            if (membrane_pos >= 0.5){
              if (do_sum)
               chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
              else
               chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            }
            else{
              if (do_sum)
               chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
              else
               chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            }
          }
          else if (acc_array[h] == undecided){
            // linear gradient of ionic strength and membrane potential within membrane pore
            if (do_sum)
             chtmp[h]  = ((ionic_str1 * membrane_pos * (maxpot_half - phi.value(tmpvec)))
                        - (ionic_str2 * (1 - membrane_pos) * (phi.value(tmpvec) + maxpot_half)))
                         * -qindu_fac;
            else
             chtmp[h]  = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
                        *  phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  blab3 << "MembChargeSetMembTwoAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembTwoAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}


/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane boundaries
   atom set 3 for a ion accessible region traversing the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembThreeAtoms
*/


MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms(const AtomSet& a1, const AtomSet& a2,
 const AtomSet& a3, float zlh, float zuh, float zlc, float zuc,
 float ionic1, float ionic2, float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, stl), ion_acc2(a2, stl), ion_acc3(a3, stl),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms (const AtomSet & a1, const AtomSet & a2,
 const AtomSet & a3, float zlh, float zuh, float zlc, float zuc, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    ion_acc3(a3, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms (const AtomSet & a1, const AtomSet & a2,
 const AtomSet & a3, float zlh, float zuh, float zlc, float zuc)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    ion_acc3(a3, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms (const IonAccVol & a1, const IonAccVol & a2,
 const IonAccVol & a3, float zlh, float zuh, float zlc, float zuc,
 float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms (const IonAccVol& a1, const IonAccVol& a2,
 const IonAccVol& a3, float zlh, float zuh, float zlc, float zuc)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()), memb_pot(PhysCond::get_membpot()),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSetMembThreeAtoms::MembChargeSetMembThreeAtoms (const IonAccVol & a1, const IonAccVol & a2,
 const IonAccVol & a3, float zlh, float zuh, float zlc, float zuc, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSetMembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSetMembThreeAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSetMembThreeAtoms::get_atomset: atom set not defined");
  return _ats1;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted


int MembChargeSetMembThreeAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSetMembThreeAtoms::number_points() const
{
    return 0;
}

// function on basis of cuberep needed ?

float MembChargeSetMembThreeAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}


ChargeCubeRep*
MembChargeSetMembThreeAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSetMembThreeAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[3];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];
           tmp_array[2] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated() && ion_acc3.check_is_calculated()) {
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else {
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<3; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

      for (int g=0; g<3; ++g){
       if        (g==0){
        startats = _ats1.begin(); endats = _ats1.end();
       } else if (g==1){
        startats = _ats2.begin(); endats = _ats2.end();
       } else if (g==2){
        startats = _ats3.begin(); endats = _ats3.end();
       }

       for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          // within the membrane core ions are not allowed
          // if we are outside IAV 2 and outside IAV3
          // or inside the IAV 1 + sternlayer
          // within IAV 3 the "undecided" AccTag is used
          // for assignment of an electrolyte gradient
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          // points not already set to interior for IAV 1
          else {
            // membrane core
            // ion accessible volume given by IAV 3 overrides IAV 2
            // use AccTag undecided to distinguish channel volume
            // from bulk electrolyte
            if (tmp_array[2][h] == interior) {
              // inside channelset
              acc_array[h] = undecided;
            }
            else if (tmp_array[1][h] == interior) {
              // inside elycavset and outside channelset
              acc_array[h] = exterior;
            }
            else {
              // outside elycavset and outside channelset
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array[2];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float corewidth  = zucbou - zlcbou;
  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior){
            chtmp[h] = 0.0;
          }
          else if (acc_array[h] == exterior){
            // ionic strength of closest region;
            // ambigous if a large ion accessible cavity
            // happens to lie in the center of the
            // membrane core - seems very improbable
            if (membrane_pos >= 0.5){
              chtmp[h] = ionic_str1 * qmemb_fac;
            }
            else{
              chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
            }
          }
          else if (acc_array[h] == undecided){
            // linear gradient of ionic strength / membrane potential within membrane pore;
            chtmp[h] = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
            * (qmemb_fac * membrane_pos);
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = ionic_str1 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSetMembThreeAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done.  Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembThreeAtoms::get_cuberep" << endl;
  return ccrp;
}

ChargeCubeRep*
MembChargeSetMembThreeAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSetMembThreeAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[3];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];
           tmp_array[2] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated() && ion_acc3.check_is_calculated()) {
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else {
    blab2 << "MembChargeSetMembThreeAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<3; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

      for (int g=0; g<3; ++g){
       if        (g==0){
        startats = _ats1.begin(); endats = _ats1.end();
       } else if (g==1){
        startats = _ats2.begin(); endats = _ats2.end();
       } else if (g==2){
        startats = _ats3.begin(); endats = _ats3.end();
       }

       for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          // within the membrane core ions are not allowed
          // if we are outside IAV 2 and outside IAV3
          // or inside the IAV 1 + sternlayer
          // within IAV 3 the "undecided" AccTag is used
          // for assignment of an electrolyte gradient
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          // points not already set to interior for IAV 1
          else {
            // membrane core
            // ion accessible volume given by IAV 3 overrides IAV 2
            // use AccTag undecided to distinguish channel volume
            // from bulk electrolyte
            if (tmp_array[2][h] == interior) {
              // inside channelset
              acc_array[h] = undecided;
            }
            else if (tmp_array[1][h] == interior) {
              // inside elycavset and outside channelset
              acc_array[h] = exterior;
            }
            else {
              // outside elycavset and outside channelset
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array[2];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      float membrane_pos = (fk - zlcbou) / corewidth;
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior){
            chtmp[h] = 0.0;
          }
          else if (acc_array[h] == exterior){
            // ionic strength of closest region;
            // ambigous if a large ion accessible cavity
            // happens to lie in the center of the
            // membrane core - seems very improbable
            if (membrane_pos >= 0.5){
              if (do_sum)
               chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
              else
               chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            }
            else{
              if (do_sum)
               chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
              else
               chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            }
          }
          else if (acc_array[h] == undecided){
            // linear gradient of ionic strength and membrane potential within membrane pore
            if (do_sum)
             chtmp[h]  = ((ionic_str1 * membrane_pos * (maxpot_half - phi.value(tmpvec)))
                        - (ionic_str2 * (1 - membrane_pos) * (phi.value(tmpvec) + maxpot_half)))
                         * -qindu_fac;
            else
             chtmp[h]  = ((ionic_str1 * membrane_pos) - (ionic_str2 * (1 - membrane_pos)))
                        *  phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            if (do_sum)
             chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
            else
             chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  blab3 << "MembChargeSetMembThreeAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done.  Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSetMembThreeAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}

/*
   The following charge distributions MembChargeSet2xyz are similar to
   the ones before termed MembChargeSet2xyz, but the charge distribution
   generating the membrane potential is only assigned to the headgroup region
*/

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   1 atom set for the united QM, normal + cavity regions
   It is meant to complement the SixValueDielMembThreeAtoms/ElectrolyteMembAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

MembChargeSet2MembAtoms::MembChargeSet2MembAtoms(const AtomSet& a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, stl), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2),
    memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }

}

MembChargeSet2MembAtoms::MembChargeSet2MembAtoms (const AtomSet & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, PhysCond::get_sterln()), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSet2MembAtoms::MembChargeSet2MembAtoms (const AtomSet & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats(a), _ats_defined(true),
    ion_acc(a, PhysCond::get_sterln()), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembAtoms::MembChargeSet2MembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembAtoms::MembChargeSet2MembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembAtoms::MembChargeSet2MembAtoms (const IonAccVol & a, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats(), _ats_defined(false),
    ion_acc(a), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSet2MembAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSet2MembAtoms::get_atomset: atom set not defined");
  return _ats;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted

int MembChargeSet2MembAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSet2MembAtoms::number_points() const
{
    return 0;
}

// function on basis of cuberep needed ?

float MembChargeSet2MembAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}


ChargeCubeRep*
MembChargeSet2MembAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSet2MembAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag* acc_array = new AccTag[ncube];
  AccTag* tmp_array = new AccTag[ncube];

  if (ion_acc.check_is_calculated()) {
    blab2 << "MembChargeSet2MembAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc.calc_cuberep(cls, tmp_array);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc.calc_cuberep(cls, tmp_array);
  } else {
    blab2 << "MembChargeSet2MembAtoms::get_cuberep calculating "
      << " simple ion accessible volume representation to make a cubic representation" << endl;

    for (int h=0; h<ncube; ++h) {
      tmp_array[h] = exterior;
    }

    for(AtomSet::const_iterator ip = _ats.begin(); ip!=_ats.end(); ++ip ) {

      int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
      int radstlsq = rad * rad;
      Coord atom_coord = ip->second.coord - grid_origin_in_space;
      int k = (int) round((atom_coord.z/spacing));
      int i = (int) round((atom_coord.x/spacing));
      int j = (int) round((atom_coord.y/spacing));
      int h = (i * nsq) + (j * grid_dim) + k;
      // skip atom if not in grid
      if (k < stlin || i < stlin || j < stlin ||
          k > grid_dim_plus_stli ||
          i > grid_dim_plus_stli ||
          j > grid_dim_plus_stli){continue;}
      int lower_l = k - rad;
      int lower_m = i - rad;
      int lower_n = j - rad;
      int upper_l = k + rad + 1;
      int upper_m = i + rad + 1;
      int upper_n = j + rad + 1;
      bool fastswitch = true;
      if (lower_l < 0) {lower_l = 0; fastswitch = false;}
      if (lower_m < 0) {lower_m = 0; fastswitch = false;}
      if (lower_n < 0) {lower_n = 0; fastswitch = false;}
      if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
      if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
      if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
      if (fastswitch) {
        // in the inner grid we have a full sphere around the point
        // and can save time by doing loops and distance
        // only in +,+,+ octand of sphere
        // tried to improve by computing dist only for 1/48th sphere
        // also permutating x,y,z and doing loops and distance only
        // in in x >= y >= z but result is slower
        int dl = -1;
        int lterm = -grid_dim;
        for (int l=k; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          lterm += grid_dim;
          int lsqterm = dl*nsq;
          int dm = -1;
          int mterm = -grid_dim;
          for (int m=i; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            // out of sphere already in plane
            if (dmlsq > radstlsq){break;}
            mterm += grid_dim;
            int msqterm = dm * nsq;
            int dn = -1;
            int nterm = -grid_dim;
            for (int n=j; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq > radstlsq){break;}
              int nsqterm = dn*nsq;
              nterm += grid_dim;
              // 1: x, y, z
              // here we permutate +/- signs in the three components
              // + + +
              int hh = h + msqterm + nterm + dl;
              tmp_array[hh] = interior;
              // + + -
              hh = hh - dl  - dl;
              tmp_array[hh] = interior;
              // + - -
              hh = hh - nterm - nterm;
              tmp_array[hh] = interior;
              // + - +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
              // - - +
              hh = hh - msqterm - msqterm;
              tmp_array[hh] = interior;
              // - - -
              hh = hh - dl - dl;
              tmp_array[hh] = interior;
              // - + -
              hh = hh + nterm + nterm;
              tmp_array[hh] = interior;
              // - + +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
            }
          }
        }
      }
      else {
        // these are atoms at the grid border
        int dl = -1;
        for (int l=lower_l; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          int dm = -1;
          for (int m=lower_m; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            if (dmlsq > radstlsq){break;}
            int msqterm = m*nsq;
            int dn = -1;
            for (int n=lower_n; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq <= radstlsq){
                int hh = msqterm + n*grid_dim + l;
                tmp_array[hh] = interior;
              }
            }
          }
        }
      }
    }


  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
          else if (rsq > holesq) {
            acc_array[h] = interior;
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
         int h = iterm + j*grid_dim + k;
         chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou)
              chtmp[h] = ionic_str1 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
            if (fk >= zlhbou)
              chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembAtoms::get_cuberep" << endl;
  return ccrp;
}

ChargeCubeRep*
MembChargeSet2MembAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSet2MembAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag* acc_array = new AccTag[ncube];
  AccTag* tmp_array = new AccTag[ncube];

  if (ion_acc.check_is_calculated()) {
    blab2 << "MembChargeSet2MembAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc.calc_cuberep(cls, tmp_array);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc.calc_cuberep(cls, tmp_array);
  } else {
    blab2 << "MembChargeSet2MembAtoms::get_cuberep calculating "
      << " simple ion accessible volume representation to make a cubic representation" << endl;

    for (int h=0; h<ncube; ++h) {
      tmp_array[h] = exterior;
    }

    for(AtomSet::const_iterator ip = _ats.begin(); ip!=_ats.end(); ++ip ) {

      int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
      int radstlsq = rad * rad;
      Coord atom_coord = ip->second.coord - grid_origin_in_space;
      int k = (int) round((atom_coord.z/spacing));
      int i = (int) round((atom_coord.x/spacing));
      int j = (int) round((atom_coord.y/spacing));
      int h = (i * nsq) + (j * grid_dim) + k;
      // skip atom if not in grid
      if (k < stlin || i < stlin || j < stlin ||
          k > grid_dim_plus_stli ||
          i > grid_dim_plus_stli ||
          j > grid_dim_plus_stli){continue;}
      int lower_l = k - rad;
      int lower_m = i - rad;
      int lower_n = j - rad;
      int upper_l = k + rad + 1;
      int upper_m = i + rad + 1;
      int upper_n = j + rad + 1;
      bool fastswitch = true;
      if (lower_l < 0) {lower_l = 0; fastswitch = false;}
      if (lower_m < 0) {lower_m = 0; fastswitch = false;}
      if (lower_n < 0) {lower_n = 0; fastswitch = false;}
      if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
      if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
      if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
      if (fastswitch) {
        // in the inner grid we have a full sphere around the point
        // and can save time by doing loops and distance
        // only in +,+,+ octand of sphere
        // tried to improve by computing dist only for 1/48th sphere
        // also permutating x,y,z and doing loops and distance only
        // in in x >= y >= z but result is slower
        int dl = -1;
        int lterm = -grid_dim;
        for (int l=k; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          lterm += grid_dim;
          int lsqterm = dl*nsq;
          int dm = -1;
          int mterm = -grid_dim;
          for (int m=i; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            // out of sphere already in plane
            if (dmlsq > radstlsq){break;}
            mterm += grid_dim;
            int msqterm = dm * nsq;
            int dn = -1;
            int nterm = -grid_dim;
            for (int n=j; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq > radstlsq){break;}
              int nsqterm = dn*nsq;
              nterm += grid_dim;
              // 1: x, y, z
              // here we permutate +/- signs in the three components
              // + + +
              int hh = h + msqterm + nterm + dl;
              tmp_array[hh] = interior;
              // + + -
              hh = hh - dl  - dl;
              tmp_array[hh] = interior;
              // + - -
              hh = hh - nterm - nterm;
              tmp_array[hh] = interior;
              // + - +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
              // - - +
              hh = hh - msqterm - msqterm;
              tmp_array[hh] = interior;
              // - - -
              hh = hh - dl - dl;
              tmp_array[hh] = interior;
              // - + -
              hh = hh + nterm + nterm;
              tmp_array[hh] = interior;
              // - + +
              hh = hh + dl + dl;
              tmp_array[hh] = interior;
            }
          }
        }
      }
      else {
        // these are atoms at the grid border
        int dl = -1;
        for (int l=lower_l; l<upper_l; ++l) {
          dl++;
          int dlsq = dl * dl;
          int dm = -1;
          for (int m=lower_m; m<upper_m; ++m) {
            dm++;
            int dmsq = dm * dm;
            int dmlsq = dmsq + dlsq;
            if (dmlsq > radstlsq){break;}
            int msqterm = m*nsq;
            int dn = -1;
            for (int n=lower_n; n<upper_n; ++n) {
              dn++;
              int dissq = (dn * dn) + dmlsq;
              if (dissq <= radstlsq){
                int hh = msqterm + n*grid_dim + l;
                tmp_array[hh] = interior;
              }
            }
          }
        }
      }
    }


  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
          else if (rsq > holesq) {
            acc_array[h] = interior;
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
         int h = iterm + j*grid_dim + k;
         chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou){
             if (do_sum)
              chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
            if (fk >= zlhbou){
             if (do_sum)
              chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembTwoAtoms
   A cylindrical hole through the membrane may be used to mimick a channel
*/

MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms(const AtomSet& a1, const AtomSet& a2, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, stl), ion_acc2(a2, stl),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }

}

MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms (const AtomSet & a1, const AtomSet & a2, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms (const AtomSet & a1, const AtomSet & a2, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms (const IonAccVol & a1, const IonAccVol & a2,
 float zlh, float zuh, float zlc, float zuc, Coord c, float hr, float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms (const IonAccVol& a1, const IonAccVol& a2,
 float zlh, float zuh, float zlc, float zuc, Coord c, float hr)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    hole_center(c), holerad(hr), ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembTwoAtoms::MembChargeSet2MembTwoAtoms (const IonAccVol & a1, const IonAccVol & a2, float zlh, float zuh, float zlc, float zuc,
 Coord c, float hr, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    holerad(hr), ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (holerad < 0.00) {
    cerr << "ERROR: MembChargeSet2MembTwoAtoms ctor called with sternl < 0.0\n"
      << "(" << holerad << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSet2MembTwoAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSet2MembTwoAtoms::get_atomset: atom set not defined");
  return _ats1;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted


int MembChargeSet2MembTwoAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSet2MembTwoAtoms::number_points() const
{
    return 0;
}

// function on basis of cuberep needed ?

float MembChargeSet2MembTwoAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}


ChargeCubeRep*
MembChargeSet2MembTwoAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSet2MembTwoAtoms::get_cuberep entered" << endl;


  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[2];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated()) {
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else {
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<2; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

      for (int g=0; g<2; ++g){
       if        (g==0){
        startats = _ats1.begin(); endats = _ats1.end();
       } else if (g==1){
        startats = _ats2.begin(); endats = _ats2.end();
       }
       for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          else {
            // membrane core
            // case 1: ion accessible volume given by inside of
            //  dummy atom set elycavset
            //  use AccTag exterior
            // case 2: ion accessible volume given by cylindrical hole
            //  use AccTag undecided to distinguish channel volume
            //  from bulk electrolyte
            if (tmp_array[1][h] == interior) {
              // inside elycav
              acc_array[h] = exterior;
            }
            else if (rsq <= holesq) {
              // outside elycav and inside cylinder
              acc_array[h] = undecided;
            }
            else {
              // outside elycav and outside cylinder
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou)
              chtmp[h] = ionic_str1 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
	    if (fk >= zlhbou)
              chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembTwoAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembTwoAtoms::get_cuberep" << endl;
  return ccrp;
}

ChargeCubeRep*
MembChargeSet2MembTwoAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSetMembTwoAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  Coord holecen_in_grid = (hole_center - grid_center_in_space)/spacing
    + grid_center_in_grid;
  float holerad_in_grid = holerad/spacing;
  float holesq = holerad_in_grid*holerad_in_grid;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[2];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated()) {
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
  } else {
    blab2 << "MembChargeSet2MembTwoAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<2; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

      for (int g=0; g<2; ++g){
       if        (g==0){
        startats = _ats1.begin(); endats = _ats1.end();
       } else if (g==1){
        startats = _ats2.begin(); endats = _ats2.end();
       }
       for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
	float xr = ((float) i) - holecen_in_grid.x;
	float xrsq = xr*xr;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          float yr = ((float) j) - holecen_in_grid.y;
	  float rsq = xrsq + (yr*yr);
          // within the membrane core ions are not allowed
          // if we are outside the hole or inside the
          // solvent inaccessible surface + sternlayer
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          else {
            // membrane core
            // case 1: ion accessible volume given by inside of
            //  dummy atom set elycavset
            //  use AccTag exterior
            // case 2: ion accessible volume given by cylindrical hole
            //  use AccTag undecided to distinguish channel volume
            //  from bulk electrolyte
            if (tmp_array[1][h] == interior) {
              // inside elycav
              acc_array[h] = exterior;
            }
            else if (rsq <= holesq) {
              // outside elycav and inside cylinder
              acc_array[h] = undecided;
            }
            else {
              // outside elycav and outside cylinder
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou){
             if (do_sum)
              chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
	    if (fk >= zlhbou){
             if (do_sum)
              chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembTwoAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done. Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembTwoAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}

/*
   added, TRU 2007
   This is a charge distribution generating a membrane potential
   across a 3-slab membrane with embedded atoms.
   atom set 1 for the united QM, normal + cavity regions and
   atom set 2 for a ion accessible region within the membrane boundaries
   atom set 3 for a ion accessible region traversing the membrane
   It is meant to complement the SixValueDielMembThreeAtoms/TwoElectrolyteMembThreeAtoms
*/


MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms(const AtomSet& a1, const AtomSet& a2,
 const AtomSet& a3, float zlh, float zuh, float zlc, float zuc,
 float ionic1, float ionic2, float membpot, float stl)
 :ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, stl), ion_acc2(a2, stl), ion_acc3(a3, stl),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms (const AtomSet & a1, const AtomSet & a2,
 const AtomSet & a3, float zlh, float zuh, float zlc, float zuc, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    ion_acc3(a3, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}


MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms (const AtomSet & a1, const AtomSet & a2,
 const AtomSet & a3, float zlh, float zuh, float zlc, float zuc)
 : ChargeDist_lett(), AtomSet(), _ats1(a1), _ats2(a2), _ats3(a3), _ats_defined(true),
    ion_acc1(a1, PhysCond::get_sterln()), ion_acc2(a2, PhysCond::get_sterln()),
    ion_acc3(a3, PhysCond::get_sterln()),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()),
    memb_pot(PhysCond::get_membpot()), sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms (const IonAccVol & a1, const IonAccVol & a2,
 const IonAccVol & a3, float zlh, float zuh, float zlc, float zuc,
 float ionic1, float ionic2, float membpot, float stl)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot), sternl(stl)
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms (const IonAccVol& a1, const IonAccVol& a2,
 const IonAccVol& a3, float zlh, float zuh, float zlc, float zuc)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(PhysCond::get_ionicstr()), ionic_str2(PhysCond::get_ionicstr()), memb_pot(PhysCond::get_membpot()),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " <= 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

MembChargeSet2MembThreeAtoms::MembChargeSet2MembThreeAtoms (const IonAccVol & a1, const IonAccVol & a2,
 const IonAccVol & a3, float zlh, float zuh, float zlc, float zuc, float ionic1, float ionic2, float membpot)
 : ChargeDist_lett(), AtomSet(), _ats1(), _ats2(), _ats3(), _ats_defined(false),
    ion_acc1(a1), ion_acc2(a2), ion_acc3(a3),
    zlowerhead(zlh), zupperhead(zuh), zlowercore(zlc), zuppercore(zuc),
    ionic_str1(ionic1), ionic_str2(ionic2), memb_pot(membpot),
    sternl(PhysCond::get_sterln())
{
  // Sanity checks
  if (zlowercore >= zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowercore >= zuppercore\n"
      << "(" << zlowercore << " >= " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead >= zupperhead) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead >= zupperhead\n"
      << "(" << zlowerhead << " >= " << zupperhead << ")" << endl;
    ::error("ERROR in ");
  }
  if (zlowerhead > zlowercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zlowerhead > zlowercore\n"
      << "(" << zlowerhead << " > " << zlowercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (zupperhead < zuppercore) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with zupperhead < zuppercore\n"
      << "(" << zupperhead << " > " << zuppercore << ")" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str1 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str1 <= 0.0\n"
      << "(" << ionic_str1 << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (ionic_str2 <= 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with ionic_str2 <= 0.0\n"
      << "(" << ionic_str2 << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
  if (sternl < 0.00) {
    cerr << "ERROR: MembChargeSet2MembThreeAtoms ctor called with sternl < 0.0\n"
      << "(" << sternl << " < 0.00)" << endl;
    ::error("ERROR in ");
  }
}

// For GB/DL's sake
const AtomSet& MembChargeSet2MembThreeAtoms::get_atomset() const
{
  if (!_ats_defined)
    ::error("MembChargeSet2MembThreeAtoms::get_atomset: atom set not defined");
  return _ats1;
}

// FIXME the following functions were included in AtomChargeSet,
// but are useless here, maybe they can be omitted


int MembChargeSet2MembThreeAtoms::has_charges() const
{
  if (memb_pot != 0){
    return 1;
  }
  else {
    return 0;
  }
}

int MembChargeSet2MembThreeAtoms::number_points() const
{
    return 0;
}

// function on basis of cuberep needed ?

float MembChargeSet2MembThreeAtoms::vacuum_coulomb(const Coord& c) const
{
  float coul = 0;
  if (c.z > zuppercore){coul = membrane_potential() *  0.5;}
  if (c.z < zuppercore){coul = membrane_potential() * -0.5;}
  return coul;
}


ChargeCubeRep*
MembChargeSet2MembThreeAtoms::get_cuberep (const CubeLatSpec& cls)
{
  blab3 << "MembChargeSet2MembThreeAtoms::get_cuberep entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[3];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];
           tmp_array[2] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated() && ion_acc3.check_is_calculated()) {
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else {
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<3; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

    for (int g=0; g<3; ++g){
      if        (g==0){
       startats = _ats1.begin(); endats = _ats1.end();
      } else if (g==1){
       startats = _ats2.begin(); endats = _ats2.end();
      } else if (g==2){
       startats = _ats3.begin(); endats = _ats3.end();
      }

      for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array

  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          // within the membrane core ions are not allowed
          // if we are outside IAV 2 and outside IAV3
          // or inside the IAV 1 + sternlayer
          // within IAV 3 the "undecided" AccTag is used
          // for assignment of an electrolyte gradient
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          // points not already set to interior for IAV 1
          else {
            // membrane core
            // ion accessible volume given by IAV 3 overrides IAV 2
            // use AccTag undecided to distinguish channel volume
            // from bulk electrolyte
            if (tmp_array[2][h] == interior) {
              // inside channelset
              acc_array[h] = undecided;
            }
            else if (tmp_array[1][h] == interior) {
              // inside elycavset and outside channelset
              acc_array[h] = exterior;
            }
            else {
              // outside elycavset and outside channelset
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array[2];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  float qmemb_fac =  memb_pot * spacing * spacing * spacing
                     * 1e-27 * PhysCond::get_Avogadro_Number()
                     * PhysCond::get_elementary_charge()
                     / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
            chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou)
              chtmp[h] = ionic_str1 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
            if (fk >= zlhbou)
              chtmp[h] = -1.0 * ionic_str2 * qmemb_fac;
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
        }
      }
    }

  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembThreeAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done.  Get ready to make the actual ChargeCubeRep.

  // Note that charges on outer boundary points will be dropped.
  // This is as it should be.

  // FIXME?  But this costs a bit to do the skipping.  Is it worth it?
  // Perhaps it is only really important to skip the -x, x sides of the box,
  // which would be cheap.

  // First count the number of charged grid points.
  int chcount=0;
  int nm1 = grid_dim - 1;
  int i;
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i = 1; i<nm1; ++i) {
    int iterm = i*nsq;
    for (int j=1; j<nm1; ++j) {
      int jterm = j*grid_dim;
      for (int k=1; k<nm1; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembThreeAtoms::get_cuberep" << endl;
  return ccrp;
}

ChargeCubeRep*
MembChargeSet2MembThreeAtoms::get_cuberep (const CubeLatSpec& cls, const ElstatPot& phi, const bool do_sum)
{
  blab3 << "MembChargeSet2MembThreeAtoms::get_cuberep(cls,phi,do_sum) entered" << endl;

  // we have to extend the solvent inaccessible volume by the Stern layer
  float spacing = cls.get_spacing();
  Coord grid_center_in_space = cls.get_center();
  int grid_dim = cls.get_grid_dim();
  int nsq = grid_dim*grid_dim;
  int ncube = grid_dim*grid_dim*grid_dim;
  float grlen = (float) (grid_dim - 1);
  float halfgrlen = grlen/2;
  Coord grid_center_in_grid (halfgrlen, halfgrlen, halfgrlen);
  float sternl_in_grid = sternl/spacing;
  float stlsq = sternl_in_grid * sternl_in_grid;
  int stli = (int) ceil(sternl_in_grid);
  int stlin = -stli;
  int grid_dim_plus_stli = grid_dim + stli;
  float halfgrlen_in_space = halfgrlen * spacing;
  Coord grid_diag_in_space;
  grid_diag_in_space.x = halfgrlen_in_space;
  grid_diag_in_space.y = halfgrlen_in_space;
  grid_diag_in_space.z = halfgrlen_in_space;
  Coord grid_origin_in_space = grid_center_in_space - grid_diag_in_space;
  float zlhbou = (zlowerhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zuhbou = (zupperhead - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zlcbou = (zlowercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;
  float zucbou = (zuppercore - grid_center_in_space.z)/spacing
    + grid_center_in_grid.z;

  AccTag  *acc_array    = new AccTag[ncube];
  AccTag **tmp_array    = new AccTag*[3];
           tmp_array[0] = new AccTag[ncube];
           tmp_array[1] = new AccTag[ncube];
           tmp_array[2] = new AccTag[ncube];

  if (ion_acc1.check_is_calculated() && ion_acc2.check_is_calculated() && ion_acc3.check_is_calculated()) {
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep using existing "
      << "analytical ion accessible volume representation to make a cubic representation" << endl;
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else if (!_ats_defined){
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep calculating "
      << "analytical ion accessible volume representations to make a cubic representation" << endl;
    // anal_calc is automatically called by calc_cuberep if flag is_calculated is false
    ion_acc1.calc_cuberep(cls, tmp_array[0]);
    ion_acc2.calc_cuberep(cls, tmp_array[1]);
    ion_acc3.calc_cuberep(cls, tmp_array[2]);
  } else {
    blab2 << "MembChargeSet2MembThreeAtoms::get_cuberep calculating "
      << " simple ion accessible volume representations to make a cubic representation" << endl;

    // initialize tmp_array
    for (int g=0; g<3; ++g)
     for (int h=0; h<ncube; ++h)
      tmp_array[g][h] = exterior;

    AtomSet:: const_iterator startats;
    AtomSet:: const_iterator endats;

    for (int g=0; g<3; ++g){
      if        (g==0){
       startats = _ats1.begin(); endats = _ats1.end();
      } else if (g==1){
       startats = _ats2.begin(); endats = _ats2.end();
      } else if (g==2){
       startats = _ats3.begin(); endats = _ats3.end();
      }

      for(AtomSet::const_iterator ip = startats; ip!=endats; ++ip ) {

        int rad = (int) ceil ((ip->second.rad + sternl)/spacing);
        int radstlsq = rad * rad;
        Coord atom_coord = ip->second.coord - grid_origin_in_space;
        int k = (int) round((atom_coord.z/spacing));
        int i = (int) round((atom_coord.x/spacing));
        int j = (int) round((atom_coord.y/spacing));
        int h = (i * nsq) + (j * grid_dim) + k;
        // skip atom if not in grid
        if (k < stlin || i < stlin || j < stlin ||
            k > grid_dim_plus_stli ||
            i > grid_dim_plus_stli ||
            j > grid_dim_plus_stli){continue;}
        int lower_l = k - rad;
        int lower_m = i - rad;
        int lower_n = j - rad;
        int upper_l = k + rad + 1;
        int upper_m = i + rad + 1;
        int upper_n = j + rad + 1;
        bool fastswitch = true;
        if (lower_l < 0) {lower_l = 0; fastswitch = false;}
        if (lower_m < 0) {lower_m = 0; fastswitch = false;}
        if (lower_n < 0) {lower_n = 0; fastswitch = false;}
        if (upper_l > grid_dim) {upper_l = grid_dim; fastswitch = false;}
        if (upper_m > grid_dim) {upper_m = grid_dim; fastswitch = false;}
        if (upper_n > grid_dim) {upper_n = grid_dim; fastswitch = false;}
        if (fastswitch) {
          // in the inner grid we have a full sphere around the point
          // and can save time by doing loops and distance
          // only in +,+,+ octand of sphere
          // tried to improve by computing dist only for 1/48th sphere
          // also permutating x,y,z and doing loops and distance only
          // in in x >= y >= z but result is slower
          int dl = -1;
          int lterm = -grid_dim;
          for (int l=k; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            lterm += grid_dim;
            int lsqterm = dl*nsq;
            int dm = -1;
            int mterm = -grid_dim;
            for (int m=i; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              // out of sphere already in plane
              if (dmlsq > radstlsq){break;}
              mterm += grid_dim;
              int msqterm = dm * nsq;
              int dn = -1;
              int nterm = -grid_dim;
              for (int n=j; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq > radstlsq){break;}
                int nsqterm = dn*nsq;
                nterm += grid_dim;
                // 1: x, y, z
                // here we permutate +/- signs in the three components
                // + + +
                int hh = h + msqterm + nterm + dl;
                tmp_array[g][hh] = interior;
                // + + -
                hh = hh - dl  - dl;
                tmp_array[g][hh] = interior;
                // + - -
                hh = hh - nterm - nterm;
                tmp_array[g][hh] = interior;
                // + - +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
                // - - +
                hh = hh - msqterm - msqterm;
                tmp_array[g][hh] = interior;
                // - - -
                hh = hh - dl - dl;
                tmp_array[g][hh] = interior;
                // - + -
                hh = hh + nterm + nterm;
                tmp_array[g][hh] = interior;
                // - + +
                hh = hh + dl + dl;
                tmp_array[g][hh] = interior;
              }
            }
          }
        }
        else {
          // these are atoms at the grid border
          int dl = -1;
          for (int l=lower_l; l<upper_l; ++l) {
            dl++;
            int dlsq = dl * dl;
            int dm = -1;
            for (int m=lower_m; m<upper_m; ++m) {
              dm++;
              int dmsq = dm * dm;
              int dmlsq = dmsq + dlsq;
              if (dmlsq > radstlsq){break;}
              int msqterm = m*nsq;
              int dn = -1;
              for (int n=lower_n; n<upper_n; ++n) {
                dn++;
                int dissq = (dn * dn) + dmlsq;
                if (dissq <= radstlsq){
                  int hh = msqterm + n*grid_dim + l;
                  tmp_array[g][hh] = interior;
                }
              }
            }
          }
        }
      }
    }

  }

  // initialize acc_array
  for (int h=0; h<ncube; ++h) {
    acc_array[h] = exterior;
  }

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          // within the membrane core ions are not allowed
          // if we are outside IAV 2 and outside IAV3
          // or inside the IAV 1 + sternlayer
          // within IAV 3 the "undecided" AccTag is used
          // for assignment of an electrolyte gradient
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
          // points not already set to interior for IAV 1
          else {
            // membrane core
            // ion accessible volume given by IAV 3 overrides IAV 2
            // use AccTag undecided to distinguish channel volume
            // from bulk electrolyte
            if (tmp_array[2][h] == interior) {
              // inside channelset
              acc_array[h] = undecided;
            }
            else if (tmp_array[1][h] == interior) {
              // inside elycavset and outside channelset
              acc_array[h] = exterior;
            }
            else {
              // outside elycavset and outside channelset
              acc_array[h] = interior;
            }
          }
        }
      }
    } // end inside membrane core
    else {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (tmp_array[0][h] == interior) {
            acc_array[h] = interior;
          }
        }
      }
    } // end outside membrane core
  }
  delete [] tmp_array[0];
  delete [] tmp_array[1];
  delete [] tmp_array[2];
  delete [] tmp_array;

  // Allocate a temp array for charges of grid points;

  float *chtmp = (float *) big_rigid_malloc ((sizeof (float)) * ncube);
  memset ((char *) chtmp, 0, ncube * sizeof (float)); // zero the array

  // now we are ready to construct the ChargeCubeRep*,
  // acc_array holds the ion accessible volume

  Coord tmpvec = grid_origin_in_space;
  float corewidth = zucbou - zlcbou;
  float maxpot_half = 0.5 * memb_pot / PhysCond::get_voltconv();
  float qindu_fac =        -2 * spacing * spacing * spacing
                      * 1e-27 * PhysCond::get_Avogadro_Number()
                      * PhysCond::get_elementary_charge() * PhysCond::get_voltconv()
                      / (PhysCond::get_Boltzmann_constant() * PhysCond::get_T());
        // gives membrane potential induced mobile charge to be added to each ion accessible grid point
        // when multiplied by its ionic strength and potential

  for (int k=0; k<grid_dim; ++k) {
    float fk = (float) k;
    // check if we are inside the membrane core
    if (fk >= zlcbou && fk <= zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i*nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
            chtmp[h] = 0.0;
        }
      }
    }
    // ~ above the membrane core
    else if (fk > zucbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior){
            // check if in upper headgroup region
            if (fk <= zuhbou){
             if (do_sum)
              chtmp[h] = -ionic_str1 * (maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str1 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          }
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    // ~ below the membrane core
    else if (fk < zlcbou) {
      for (int i = 0; i<grid_dim; ++i) {
        int iterm = i * nsq;
        for (int j=0; j<grid_dim; ++j) {
          int h = iterm + j*grid_dim + k;
          if (acc_array[h] == interior)
            chtmp[h] = 0.0;
          else if (acc_array[h] == exterior)
            // check if in lower headgroup region
            if (fk >= zlhbou){
             if (do_sum)
              chtmp[h] = -ionic_str2 * (-maxpot_half - phi.value(tmpvec)) * qindu_fac;
             else
              chtmp[h] =  ionic_str2 * phi.value(tmpvec) * qindu_fac;
            }
            else
              chtmp[h] = 0.0;
          else {
            cerr << "acc_array[" << h << "] has unexpected value, "
            << acc_array[h] << endl;
            ::error("ERROR in ");
          }
          tmpvec.y += spacing;
        }
        tmpvec.x += spacing;
        tmpvec.y  = grid_origin_in_space.y;
      }
    }
    tmpvec.x  = grid_origin_in_space.x;
    tmpvec.y  = grid_origin_in_space.y;
    tmpvec.z += spacing;
  }

  delete [] acc_array;

  blab3 << "MembChargeSet2MembThreeAtoms::get_cuberep: finished chtmp\n" << endl;

  // chtmp done.  Get ready to make the actual ChargeCubeRep.

  // ATTENTION CHARGES ON OUTER BOUNDARY POINTS WILL NOT BE DROPPED HERE
  // THIS ChargeCubeRep is not suitable for an FD calculation

  // First count the number of charged grid points.
  int chcount=0;
  int i;
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ++chcount;
      }
    }
  }

  // Create and fill the (Sparse) ChargeCubeRep
  ChargeCubeRep *ccrp = new SparseChargeCubeRep(cls, chcount);
  for (i=0; i<grid_dim; ++i) {
    int iterm = i*nsq;
    for (int j=0; j<grid_dim; ++j) {
      int jterm = j*grid_dim;
      for (int k=0; k<grid_dim; ++k) {
	int h = iterm + jterm + k;
	if (chtmp[h] != 0.0)
	  ccrp->add(h, chtmp[h]);
      }
    }
  }

  big_rigid_free(chtmp);
  blab2 << "Exiting MembChargeSet2MembThreeAtoms::get_cuberep(cls,phi,do_sum)" << endl;
  return ccrp;
}

// MembChargeSet.cc ends here
